#!/usr/bin/env bash

set -e

verbose=1
release=""
project_dir="$(pwd)"
iqportal_dir="$HOME/workspace/iqportal"
servers="ws2.clusternet ws3.clusternet ws4.clusternet ws5.clusternet db4.clusternet"
package=""
tag=""
version=""
scheduler_host="db4.clusternet"
valid_release_names="nightly prototype demo sandbox testing acceptance stable"
post_only=0
downgrade=0

confirm () {
    # call with a prompt string or use a default
    read -r -p "${1:-Are you sure?} [y/N] " response
    case $response in
        [yY][eE][sS]|[yY])
            true
            ;;
        *)
            false
            ;;
    esac
}

log() {
	[ $verbose -eq 1 ] && echo $1
}


parallel_ssh() {
	cmd=$1

	log "parallel-ssh: $servers: $cmd"

	args=""
	for host in $servers; do
		args="$args -H $host"
	done

	parallel-ssh --timeout 0 --inline $args "$cmd"

}
main() {

	# check requirements
	[ -z $release ] && usage && exit 1
	[[ $valid_release_names =~ $(echo $release | sed 's/-virtual//') ]] || (echo "Release must be in $valid_release_names" && echo && usage && exit 1)

	[ ! -e '/usr/bin/parallel-ssh' ] && echo 'please install pssh' && exit 1;

	if [[ $release == "nightly" ]] || [[ $release == "prototype" ]]; then
		tag="$(date +%y%m%d%H%M)"
		if [[ $release == "nightly" ]]; then
		    servers="vm.office"
		    scheduler_host="vm.office"
		fi
	else

		# if no tag given, get tag based on last tag created
		if [[ $tag == "" ]]; then
			tag=$(git describe --tags --abbrev=0)
		fi

	fi

	project_name=$(basename $project_dir)
	package_name="$project_name-$release"

	# versions in debian cant have underscores
	version=$(echo $tag | sed 's/_/-/g')
	package="$package_name=$version"

	echo
	echo "Tag n bagging for $package"
	if [[ $release != "nightly" ]] && [[ $release != "prototype" ]]; then
		echo " - tag $tag"
	fi
	echo " - release $release"
	echo " - project_dir $project_dir"
	echo " - to servers: $servers"
	echo " - post_only: $post_only"
	echo " - downgrade: $downgrade"
	echo
	confirm 'Looky ok?' || exit

	if [[ $post_only != 1 ]]; then

		if [[ $release != "nightly" ]] && [[ $release != "prototype" ]]; then
			tag
		fi

		make_deb
		do_install
	fi

	do_post

	echo 'tagnbag donner!'

}

tag() {

	echo 'tag?'

	if [ ! -z "$(git diff)" ]; then
		echo 'git is dirty, aborting.. (need to commit?)'
		exit 1
	fi

	log "Fetching existing tags.."
	git fetch --tags  > /dev/null

	if [ -z "$(git tag | grep -e ^$(echo $tag | sed 's/\./\\./g')$)" ]; then
		echo 'Existing tag not found.. Last commit:'
		echo
		echo $(git status | head -n 1)
		echo
		git log | head -7
		echo
		echo

		confirm "Create and push tag $tag?" || exit

		git tag $tag
		git push origin-ssh --tags

	else
		echo " - Tag $tag already exists"
	fi

	# push commits if on master
	if [ "$(git branch | grep '*')" = "* master" ]; then git push origin-ssh; fi

}

make_deb() {


	log "Checking if package exists in apt repo.."
	sudo apt-get update -o Dir::Etc::sourcelist="sources.list.d/basesoft.list" -o Dir::Etc::sourceparts="-" -o APT::Get::List-Cleanup="0" > /dev/null

	set +e
	apt-cache show $package &>/dev/null
	exit_code=$?
	set -e

	if [ $exit_code -eq 0 ]; then
		echo " - found package $package"
	else
		echo ' - package not found'

		log "Creating .deb for $package"
		find . -maxdepth 1 -name '*.deb' -exec rm {} \;

		include_tests=0;
		if [[ $release == "nightly" ]] || [[ $release == "prototype" ]]; then
			include_tests=1;
		fi

		$iqportal_dir/related/bin/build_deb.sh -r $release -d $project_dir -q $iqportal_dir -t $tag -i $include_tests

		if [[ $downgrade == 1 ]]; then
			unpublish_deb.sh $package_name
		fi

		log "Publishing .deb.."
		publish_deb.sh *.deb

	fi

}

remote_www_exec() {
	ssh $scheduler_host "sudo su www-data -s /bin/bash -c 'cd /var/www/iqportal-$release.basesoft.com && source vars && $1'"
}

manage_command() {
	remote_www_exec "python manage.py $1"
}

do_install() {
	parallel_ssh "sudo apt-get update -o Dir::Etc::sourcelist="sources.list.d/basesoft.list" -o Dir::Etc::sourceparts="-" -o APT::Get::List-Cleanup="0" > /dev/null"
	force=""
	if [[ $downgrade == 1 ]]; then
		force="--force-yes"
	fi
	parallel_ssh "sudo DEBIAN_FRONTEND=noninteractive apt-get -q -y $force install $package"
}

sync_all() {
	manage_command "iqubesync"
	manage_command "iqsyncdb"
}

do_post() {
	if [[ $release == "nightly" && $project_name == "iqportal" ]] || [[ $release == "prototype" && $project_name == "iqportal" ]]; then
		confirm "Run resetdb for $release on $scheduler_host?" && manage_command "iqresetdb"
		confirm "run fme populate?" && remote_www_exec "./bin/iqportal/fme_populate_responses.py"
		confirm "run tnt populate?" && remote_www_exec "./bin/iqportal/tnt_populate_responses.py"
		confirm "run pvcycle populate?" && remote_www_exec "./bin/pvcycle/pvcycle_populate.py"
		confirm "run weee-eu populate?" && remote_www_exec "./bin/weee-eu/we_populate.py"
	fi

	confirm "Run iqubesync and syncdb for $release on $scheduler_host?" && sync_all
	confirm "Restart the $release scheduler on $scheduler_host?" && ssh $scheduler_host "sudo /etc/init.d/iqportal-scheduler-$release restart"
}

usage() {
	echo "Usage: tag_n_bag.sh -r release [ -d project_dir ] [ -q iqportal_dir ] [ -t tag ] [ -s webserver1,webserver2 ] [ -p ] [ -g ]"
	echo "Example: tag_n_bag.sh -r acceptance"
	echo
	echo "tag_n_bag.sh deploys"
}


while getopts "h?vr:d:q:t:s:pg" opt; do
		case "$opt" in
		h|\?)
		usage
			;;
		v)  verbose=1
			;;
		r)  release=$OPTARG
			;;
		d)  project_dir=$OPTARG
			;;
		q)  iqportal_dir=$OPTARG
			;;
		t)  tag=$OPTARG
			;;
		s)  servers=$OPTARG
			;;
		p)
			post_only=1
			;;
		g)
			downgrade=1
			;;
				esac
done

shift $((OPTIND-1))

[ "$1" = "--" ] && shift

main

