"""
Python Logging
See (see http://docs.python.org/lib/module-logging.html for more)

--

Logging works like normal python logging but with custom settings for iqportal. The logger should be imported
as following:

from basesoft import logging
logger = logging.getLogger('application.name')

It can also be used with a request argument, which contains ip and user info, so the Log entries from the db have more
info:
logger = logging.getLogger('application.name', request)

Another way to use the logger is in controllers, where a controller already has a logger variable. This way, no more
imports need to be done.
# need to add the aplication name in the controller, to override the default logger name. also, when the controller is
# instantiated, besides the regular sa_session and account_id, we can (optionally) pass request for having extra info in
# the database, like ip and user who was logged when that logging call was made. Eg:
class MyController(AbstractController):
    logger_name = 'application.name'
    ...
    # calling is done like:
    self.logger.debug('message') # or warning, critical, info, error
---

By default log on INFO level.. you can enable debug level for specific applications in personal settings like this:

from basesoft.logging import logger
logging.getLogger('application.name').setLevel(logging.DEBUG)

"""

import sys
import logging

from logging.handlers import SysLogHandler

from django.conf import settings

# configure basic logging
logging.basicConfig(
    level = logging.INFO,
    format = '%(asctime)s - %(levelname)s - %(message)s',
)


class _Logger(logging.Logger):
    """
    Custom logger that also adds extra information to the database. This is private, used only in this module
    """
    sa_session = None

    def __init__(self, name, level=0):
        """
        Initialize the logger object
        @param name: the name of the logger
        """
        logging.Logger.__init__(self, name, level)
        self.logger_name = name

    def _process(self, level, message, kwargs):
        """
        Process the extra information given to the log calls
        @param level: the log level (DEBUG, INFO, etc)
        @param message: the message to be logged
        @param kwargs: the keyword argument param, where we look to see if we have an extra context (with request)
        @return: the message to be logged, prepended with user id and ip info
        """
        request = None

        ip = None
        uid = None
        name = None

        if kwargs.has_key('extra'):

            if kwargs['extra'].has_key('request') and kwargs['extra']['request'] is not None:
                request = kwargs['extra']['request']

                # user_id
                if request.session.has_key('user_id'):
                    try:
                        uid = int(request.session['user_id'])
                    except ValueError:
                        pass

                # ip
                ip = request.META.get('REMOTE_ADDR', None)

            name = None
            if kwargs['extra'].has_key('name'):
                name = kwargs['extra']['name']

        return "%s: ip: %s, uid: %s, %s: %s" % (settings.SYSLOG_IDENT, ip, uid, name, message)

    def debug(self, message, *args, **kwargs):
        """
        Override for the debug method
        """
        message = self._process(logging.DEBUG, message, kwargs)
        logging.Logger.debug(self, message)

    def info(self, message, *args, **kwargs):
        """
        Override for the info method
        """
        message = self._process(logging.INFO, message, kwargs)
        logging.Logger.info(self, message)

    def warn(self, message, *args, **kwargs):
        """
        Override for the warn method
        """
        message = self._process(logging.WARN, message, kwargs)
        logging.Logger.warn(self, message)

    def warning(self, message, *args, **kwargs):
        """
        Override for the warning method
        """
        message = self._process(logging.WARN, message, kwargs)
        logging.Logger.warning(self, message)

    def error(self, message, *args, **kwargs):
        """
        Override for the error method
        """
        message = self._process(logging.ERROR, message, kwargs)
        logging.Logger.error(self, message)

    def critical(self, message, *args, **kwargs):
        """
        Override for the critical method
        """
        message = self._process(logging.CRITICAL, message, kwargs)
        logging.Logger.critical(self, message)

logging.setLoggerClass(_Logger)

# configure the root logger to output to stdout and syslog
# syslog can be configured thru settings and maybe we can add other configs here later too
logger = logging.getLogger()
logger.setLevel(logging.INFO)

# stdout logging to debug by default
logger.handlers[0].setLevel(settings.LOGGER_STDOUT_LEVEL)

# add syslog handler which logs level from settings (info normally, or debug for development)
if settings.LOGGER_SYSLOG_ADDRESS is not None:
    sh = SysLogHandler(address=settings.LOGGER_SYSLOG_ADDRESS)
    sh.setLevel(settings.LOGGER_SYSLOG_LEVEL)
    logger.addHandler(sh)


class Logger:
    """
    deprecated: use the self.logger from AbstractController
    """

    def __init__(self, name, request=None):
        """
        Initialize a logger container that can do the regular methods (debug, info, etc)

        @param name: the name of the logger
        @param request: optional request object that contains additional info that we might use (default None)
            eg. if provided/available, we use the user id and the ip
        """
        self._logger = logging.getLogger(name)
        self._extra = {'name': name, 'request': request}

    def update(self, request=None):
        """
        Update the logger context. (eg. the account controller doesn't have the user in the beginning)
        @param request: if request is provided, the extra context is updated with this new value
        """
        if request is not None:
            self._extra['request'] = request

    def debug(self, message):
        """
        Log a debug message.
        @param message: the message that is logged
        """
        self._logger.debug(message, extra=self._extra)

    def info(self, message):
        """
        Log an info message.
        @param message: the message that is logged
        """
        self._logger.info(message, extra=self._extra)

    def warn(self, message):
        """
        Log a warning message.
        @param message: the message that is logged
        """
        self._logger.warn(message, extra=self._extra)

    def warning(self, message):
        """
        Log a warning message.
        @param message: the message that is logged
        """
        self._logger.warning(message, extra=self._extra)

    def error(self, message):
        """
        Log an error message.
        @param message: the message that is logged
        """
        self._logger.error(message, extra=self._extra)

    def critical(self, message):
        """
        Log a critical message.
        @param message: the message that is logged
        """
        self._logger.critical(message, extra=self._extra)

def getLogger(name, request=None):
    """
    Extends the pythong getlogger function so we can add custom behaviour to it that applies for all
    applications.

    Name is required so we force all apps to write to a logger. The name must exist in the lognames list
    else we error out. This way we keep track of which apps do logging to where so when you make a new logname
    do it in style.

    @param request: optional. if provided, the user and ip information from request are also stored in the db log entry
    """
    return Logger(name, request)

