# Copyright (c) 2008 BaseSoft Solutions B.V.
# See LICENSE for details.

"""
Contains ScheduleTask object for the Twisted Scheduler Server
"""
from basesoft.web.core.util.db import TaskSession
from basesoft import logging

# these signals can be send to the task and can be implemented by execute() to act accordingly
SIGNAL_STOP = 1
SIGNAL_PAUSE = 2
SIGNAL_CONTINUE = 3

class ScheduleTask:
    """
    Abstract scheduled task class. This can be subclassed in a django application' schedule module and appended to its
    tasks to be loaded by the scheduler.
    """

    """
    The interval (seconds) in which the execute() method is executed.
    (Run every $INTERVAL seconds)

    INTERVAL can also be a CronTab instance, similar to linux cron jobs.
    For example, to run a task every night at 03:00:
    INTERVAL = CronTab('3 * * * *')

    For the CronTab format, see
    https://github.com/josiahcarlson/parse-crontab/blob/master/README

    @static
    """
    INTERVAL = 3600

    """
    set ENABLED to False if it should not run by default. This can be changed by the scheduler server api
    @static
    """
    ENABLED = True

    """
    set ALLOW_DUPLICATES to true to allow more than one task of this type to be running or queued to run
    @static
    """
    ALLOW_DUPLICATES = False

    """
    set QUEUE_IF_EXISTS to true for tasks which should be processed, but not run simultaneously
    """
    QUEUE_IF_EXISTS = False

    """
    the MAX_COMPARABLE_TASKS_IN_QUEUE is used if QUEUE_IF_EXISTS is true. This will be the maximum number of tasks queued.
    """
    MAX_COMPARABLE_TASKS_IN_QUEUE = 10

    process_id = None

    def __init__(self, process_id=None, start_time=None, params=None):
        """
        Constructor of the ScheduleTask
        @param sa_session: sqlalchemy session (assigned by server), thread safe and should only be used in the execute() method!
        @param process_id: running process_id (assigned by the scheduler server)
        @param start_time: datetime when the process started (assigned by the server)
        """

        if process_id is not None:
            assert isinstance(process_id, (int ))

        self.sa_session = TaskSession()

        self.process_id = process_id
        self.start_time = start_time
        self.params = params

        self._signal = None

        self.logger = logging.Logger("%s.%s" % (self.__module__, self.__class__.__name__))

    def execute_wrapper(self):
        """
        wrapper called by the scheduler server to be able to catch exceptions
        """
        try:
            return self.execute()
        except TaskStoppedException:
            return 'killed by SIGNAL_STOP'

    def execute(self):
        """
        Gets executed for evey INTERVAL seconds in its own thread.
        If the task can take a long time it might be good idea to implement signal checking and status returning.

        If the task did something (like updated the database or sent out mails) it should return a result to a
        not-None (or non-empty list) value so it will go into the result history. If this methods returns None (or empty list)
        the result is ignored for history.
        """

        return None

    def get_status(self):
        """
        This method can be used to return status during execution. Be sure to implement a thread-safe solution:
        http://effbot.org/zone/thread-synchronization.htm
        """
        return None

    def set_signal(self, signal):
        """
        Execute can implement checking of signals and act accordingly. Like while execute() is running the server can
        send a SIGNAL_STOP which the execute() method can check (if its in a loop). Then if its set simply return. But
        again, this depends on the task.

        @final
        @param signal: int
        """
        assert isinstance(signal, int)
        self._signal = signal

    def loop_validator(self):
        """
        Look validator can be passed as a corben to controllers and executed after each loop. If in the mean time a SIGNAL_STOP
        is set on the task it calls a TaskStoppedException and the task execution gets stopped without an error
        """
        if self._signal == SIGNAL_STOP:
            raise TaskStoppedException()

    def queue_full_notification(self):
        """
        This is executed if QUEUE_IF_EXISTS is True and the MAX_COMPARABLE_TASKS_IN_QUEUE is reached.
        """
        pass

    def __repr__(self):
        return '<%s pid=%s>' % (self.__class__.__name__, self.process_id)

class TaskStoppedException(Exception):
    pass
