# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

"""
Classes for managing Attributes. They can be used by any class that needs attributes.

Trying to achieve something similar to what YUI has:
http://developer.yahoo.com/yui/docs/Attribute.js.html
http://developer.yahoo.com/yui/docs/AttributeProvider.js.html
"""

class Attribute:
    """
    Provides Attribute configurations.
    """

    def __init__(self, map=None):
        """
        Init an Attribute object.

        @param map: if map is a dictionary, set the Attribute to have the parameters the values from map
            and also configure the initial config to the values from map
        """

        # name of the attribute
        self.name = None

        # the value of the attribute
        self.value = None

        # whether or not the attribute is read only
        self.readOnly = False

        # whether or not the attribute can only be written once
        self.writeOnce = False

        # the attribute's initial configuration
        self.__initConfig = None

        # whether or not the attribute's value has been set
        self.__written = False

        # the method to use when setting the attribute's value
        self.onChange = None

        # the validator to use when setting the attribute's value
        self.validator = None

        if map:
            self.configure(map) # set the attributes
            self.configure(map, True) # set the initial config too

    def getValue(self):
        """
        Retrieves the current value of the attribute.
        """
        return self.value

    def setValue(self, value):
        """
        Sets the value of the attribute after it's validated and fires the onChange method.

        @param value: The value to apply to the attribute.
        @return: True if the value was set, or raise Exceptions otherwise
        """
        name = self.name

        if self.readOnly or ( self.writeOnce and self.__written):
            raise Exception("Can't set value '%s' for the '%s' attribute. Value not allowed" % (str(value), name))

        if self.validator and not self.validator(value):
            raise Exception("Validation failed for value '%s' for the '%s' attribute." % (str(value), name))

        if self.onChange:
            self.onChange(value)

        self.value = value;
        self.__written = True;

        return True;

    def __setConfig(self, **kwargs):
        """
        Set the configuration properties for the attribute.
        Each of the parameters is used to set the appropriate value. This is called by the configure function.
        Only set the parameters that are present in kw_args
        """
        for key in kwargs:
            if key == 'name': self.name = kwargs[key]
            if key == 'value': self.value = kwargs[key]
            if key == 'readOnly': self.readOnly = kwargs[key]
            if key == 'writeOnce': self.writeOnce = kwargs[key]
            if key == 'onChange': self.onChange = kwargs[key]
            if key == 'validator': self.validator = kwargs[key]

    def getConfig(self):
        """
        Return a dictionary of the attribute's properties
        """
        return {
            'name': self.name,
            'value': self.value,
            'readOnly': self.readOnly,
            'writeOnce': self.writeOnce,
            'onChange': self.onChange,
            'validator': self.validator
        }

    def configure(self, map=None, init=False):
        """
        Allows for configuring the Attribute's properties.

        @param map: A key-value map of Attribute properties.
        @param init: True if the map contains properties for the initial configuration
        """
        if not map: map = {}

        self.__written = False # reset writeOnce
        if not self.__initConfig: self.__initConfig = {}

        for key in map:
            if init:
                self.__initConfig[key] = map[key]
            else:
                self.__setConfig(**map)

    def resetValue(self):
        """
        Resets the value to the initial config value.
        @return: Whether or not the value was set.
        """
        return self.setValue(self.__initConfig['value']);

    def resetConfig(self):
        """
        Resets the attribute's config to the initial config state.
        """
        self.configure(self.__initConfig)


class AttributeProvider:
    """
    Provides and manages Attribute instances
    """

    def __init__(self):
        # a key-value map of Attribute configurations
        self.__configs = None

    def get(self, key):
        """
        Returns the current value of the attribute

        @param key: The attribute whose value will be returned
        """

        if not self.__configs: self.__configs = {}

        if not self.exists(key):
            raise KeyError("Attribute %s doesn't exist." % key)

        return self.__configs[key].getValue()

    def set(self, key, value):
        """
        Sets the value of a config.

        @param key: The name of the attribute
        @param value: The value to apply to the attribute
        @return: Whether or not the value was set
        """
        if not self.__configs: self.__configs = {}

        if not self.exists(key):
            raise KeyError("Attribute %s doesn't exist." % key)

        return self.__configs[key].setValue(value)

    def getAttributeKeys(self):
        """
        Returns a list of attribute names.
        """
        keys = []
        for item in self.__configs:
            keys.append(item)

        return keys

    def setAttributes(self, map):
        """
        Sets multiple attribute values.
        @param map: A key-value dictionary of attributes
        """
        for key in map:
            if self.exists(key):
                self.set(key, map[key])

    def resetValue(self, key):
        """
        Resets the specified attribute's value to its initial value.
        """
        if not self.__configs: self.__configs = {}

        if not self.exists(key):
            raise KeyError("Attribute %s doesn't exist." % key)

        self.__configs[key].resetValue()


    def exists(self, key):
        """
        Find out if an attribute exists or not

        @param key: the name of the attribute
        @return: True if the attribute exists, False otherwise
        """

        return self.__configs.has_key(key)

    def getAttributeConfig(self, key):
        """
        Returns the attribute's properties.

        @param key: the attribute's name
        @return: A dictionary containing all the attribute's properties
        """
        if not self.__configs: self.__configs = {}

        if not self.exists(key):
            raise KeyError("Attribute %s doesn't exist." % key)

        return self.__config[key].getConfig()

    def setAttributeConfig(self, key, map=None, init=False):
        """
        Sets or updates an Attribute instance's properties.

        @param key: the attribute's name
        @param map: a dictionary of the attribute's properties
        @param init: Whether or not this should become the intial config.
        """
        if not self.__configs: self.__configs = {}
        if not map: map = {}

        if not self.exists(key):
            map['name'] = key
            self.__configs[key] = self.createAttribute(map)
        else:
            self.__configs[key].configure(map, init)

    def resetAttributeConfig(self, key):
        """
        Resets an attribute to its intial configuration.

        @param key: the attribute's name
        """
        if not self.__configs: self.__configs = {}

        if not self.exists(key):
            raise KeyError("Attribute %s doesn't exist." % key)

        self.__config[key].resetConfig()

    def createAttribute(self, map):
        """
        Create an attribute
        """
        return Attribute(map)
