# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

"""
General purpose functions that can be used for password manipulations.
"""
from django.utils.translation import ugettext as _
from random import choice, randint
import cracklib
import string

punctuation = '!#$%&*+,-.:;=?@_~'

def generate_password(min_length=8, max_length=16,
                      use_lowercase=True, use_uppercase=True, use_digits=True, use_punctuation=False):
    """
    Generate a random password.

    ascii_lowercase = 'abcdefghijklmnopqrstuvwxyz'
    ascii_uppercase = 'ABCDEFGHIJKLMNOPQRSTUVWXYZ'
    digits = '0123456789'
    punctuation = '!"#$%&()*+,-.:;<=>?@[]^_~'

    @param min_length: an integer specifying the minimum password length. default 8
    @param max_length: an integer specifying the maximum password length. default 16
    @param use_*: indicates whether the desired character sets should be used

    @return: the generated password
    """
    chars = '' # chars that make up the password
    if use_lowercase:
        chars = chars + string.ascii_lowercase

    if use_uppercase:
        chars = chars + string.ascii_uppercase

    if use_digits:
        chars = chars + string.digits

    if use_punctuation:
        chars = chars + punctuation

    return "".join(choice(chars) for x in range(randint(min_length, max_length)))

def password_pronunciation(password, separator=' - '):
    """"
    Return the pronunciation for the given password
    @param password
    @param separator: what should be used as a separator for the pronunciation
    @return the pronunciation string: eg. password='a2D' => 'alpha - 2 - DELTA'
    """
    hash = {
        'a': 'alpha',   'b': 'bravo',   'c': 'charlie', 'd': 'delta',   'e': 'echo',    'f': 'foxtrot',
        'g': 'golf',    'h': 'hotel',   'i': 'india',   'j': 'juliet',  'k': 'kilo',    'l': 'lima',
        'm': 'mike',    'n': 'november','o': 'oscar',   'p': 'papa',    'q': 'quebec',  'r': 'romeo',
        's': 'siera',   't': 'tango',   'u': 'uniform', 'v': 'victor',  'w': 'whiskey', 'x': 'xray',
        'y': 'yankee',  'z': 'zulu',

        '0': '0',       '1': '1',       '2': '2',       '3': '3',       '4': '4',
        '5': '5',       '6': '6',       '7': '7',       '8': '8',       '9': '9',

        '!': _('Exclamation'),          '#': _('Hash'),                 '$': _('Dollar'),
        '%': _('Percent'),              '&': _('Ampersand'),            '*': _('Asterisk'),
        '+': _('Plus'),                 ',': _('Comma'),                '-': _('Dash'),
        '.': _('Period'),               ':': _('Colon'),                ';': _('Semicolon'),
        '=': _('Equals'),               '?': _('Question'),             '@': _('At'),
        '_': _('Underscore'),           '~': _('Tilde'),
    }

    pronunciation = []

    for c in password:
        if c in string.digits:
            pronunciation.append(hash[c])
            continue

        if c in string.ascii_letters:
            if c.islower():
                pronunciation.append(hash[c])
            else:
                pronunciation.append(hash[c.lower()].upper())
            continue

        if c in punctuation:
            pronunciation.append(hash[c])

    return separator.join(pronunciation)

def is_password_fascist_valid(password):
    """
    Indicates if the given password is valid by using the cracklib.FascistCheck method
    @param password: the password we want to check
    @return: a boolean that indicates if the password is valid
    """
    try:
        return cracklib.FascistCheck(password) == password # if password is valid, it's returned by the check method
    except:
        return False
