# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

"""
models for account app
"""

import datetime
from basesoft.web.contrib.i18n.models import Locale
from sqlalchemy.orm.session import Session

from django.conf import settings
from django.utils.translation import ugettext, ugettext_lazy as _

from basesoft.web.contrib.mailing.models import Mail

class Account(object):
    """
    This model represents a user account for IQPortal.
    @DynamicAttrs
    """

    permissions = ['create_account']

    NAME_I18N = _('Account')
    JS_MODULE = None

    def __get_default_group(self):
        """
        This retrieves the default group of this account.
        """
        session = Session.object_session(self)
        return session.query(GroupAccount).filter(
            GroupAccount.account_id==self.id).filter(
            GroupAccount.is_default==True).one().group

    default_group = property(__get_default_group)

    def __get_domain(self):
        return self.__get_default_group().license.domain

    domain = property(__get_domain)

    def get_config(self, config):
        from basesoft.web.core.util.controllers import AccountConfigController
        session = Session.object_session(self)

        acc = AccountConfigController(session, account_id=self.id)
        return acc.get(config)

    def set_config(self, config, value):
        from basesoft.web.core.util.controllers import AccountConfigController
        session = Session.object_session(self)

        acc = AccountConfigController(session, account_id=self.id)
        acc.set(config, value)

    def get_full_name(self):
        if self.person is None:
            return self.name
        return self.person.full_name

    def __repr__(self):
        return '<Account id="%s" name="%s">' % (self.id, self.name)

class AccountActivationMail(Mail):
    """
    This model represents the Account Activation Mail.
    @DynamicAttrs
    """

    NAME_I18N = _("Account activation mail")

    PRIV_USE_FOR_SENDING = 64

    JS_MODULE = 'iq_account_mail_plugin'

    def get_privileges(self):
        return Mail.get_privileges(self) + [
            (self.PRIV_USE_FOR_SENDING, ugettext('Use for sending'))
        ]

    def __repr__(self):
        return '<AccountActivationMail id="%s">' % self.id

    @staticmethod
    def replacement_fields():
        """
        account activation emails dont have person information so just override it and dont call parent
        """
        return {
            ugettext('Account'): {
                'account.name': ugettext('Account name'),
                'account.activation.link': ugettext('Link to activation')
            }
        }

    def _replace_field(self, person, field_name, context):

        # context is set by self.execute passing to _build_message
        activation = context['activation']
        url = 'activate'
        if activation.activation_type == AccountActivation.ACTIVATION_TYPE_NEW_ACCOUNT_FROM_LICENSE:
            url = 'change_password'

        if field_name == 'account.name':
            return activation.account.name

        if field_name == 'account.activation.link':
            link = self._base_url(context) + '#m=%s-%s' % (url, activation.activation_key)
            return '<a href="%s">%s</a>' % (link, link)

        return Mail._replace_field(self, person, field_name, context)

class LostPasswordMail(Mail):
    """
    This model represents the Lost Password Mail.
    @DynamicAttrs
    """

    NAME_I18N = _("Lost password mail")

    PRIV_USE_FOR_SENDING = 64

    JS_MODULE = 'iq_account_mail_plugin'

    def get_privileges(self):
        return Mail.get_privileges(self) + [
            (self.PRIV_USE_FOR_SENDING, ugettext('Use for sending'))
        ]

    def __repr__(self):
        return '<LostPasswordMail id="%s">' % self.id

    @staticmethod
    def replacement_fields():
        """
        lost password emails replacement fields
        """
        # this emails can contain the person info
        fields = Mail.replacement_fields()
        fields[ugettext('Account')] = {
            'account.name': ugettext('Account name'),
            'account.activation.link': ugettext('Link to activation')
        }
        return fields

    def _replace_field(self, person, field_name, context):

        # context is set by self.execute passing to _build_message
        activation = context['activation']

        if field_name == 'account.name':
            return activation.account.name

        if field_name == 'account.activation.link':
            link = self._base_url(context) + '#m=change_password-%s' % activation.activation_key
            return '<a href="%s">%s</a>' % (link, link)

        return Mail._replace_field(self, person, field_name, context)

class AccountActivation(object):
    """
    A simple profile which stores an activation key for use during
    user account registration.

    Generally, you will not want to interact directly with instances
    of this model; the provided manager includes methods
    for creating and activating new accounts, as well as for cleaning
    out accounts which have never been activated.

    While it is possible to use this model as the value of the
    ``AUTH_PROFILE_MODULE`` setting, it's not recommended that you do
    so. This model's sole purpose is to store data temporarily during
    account registration and activation, and a mechanism for
    automatically creating an instance of a site-specific profile
    model is provided via the ``create_inactive_user`` on
    ``RegistrationManager``.
    @DynamicAttrs
    """
    ACTIVATION_TYPE_NEW_ACCOUNT = 1
    ACTIVATION_TYPE_CHANGE_PASSWORD = 2
    ACTIVATION_TYPE_NEW_ACCOUNT_FROM_LICENSE = 3

    # New activation entry.
    ACTIVATION_STAGE_NEW = 0
    # The activation is currently being send (the email is in the queue)
    ACTIVATION_STAGE_SENDING = 1
    # The activation has been sent to the user and requires action.
    ACTIVATION_STAGE_PENDING = 2
    # The key was used and this activation is no longer valid.
    ACTIVATION_STAGE_USED = 3

    # db properties
    activation_key = None
    activation_type = None
    activation_stage = None

    def activation_key_expired(self):
        """
        Determines whether this activation key has expired.

        Returns ``True`` if the key has expired, ``False`` otherwise.

        Key expiration is determined by the setting
        ``ACTIVATION_KEY_VALID_DAYS``, which should be the number of
        days a key should remain valid after an account is registered.

        """
        expiration_date = datetime.timedelta(days=settings.ACTIVATION_KEY_VALID_DAYS)
        return self.created + expiration_date <= datetime.datetime.now()

    def __repr__(self):
        return '<AccountActivation account="%s", type="%s">' % \
                (self.account, self.activation_type)

class GroupAccount(object):
    """
    This model represents a group and account link
    @DynamicAttrs
    """
    def __repr__(self):
        return '<GroupAccount group=%s, account=%s>' % (self.group, self.account)

class Config(object):
    """
    Contains data about account config options
    @DynamicAttrs
    """

class AccountConfig(object):
    """
    Overwrites default values for account configs
    @DynamicAttrs
    """

class AccountCustomQueryPreset(object):
    """
    Presets
    @DynamicAttrs
    """
