# Copyright (c) 2009 BaseSoft B.V.
# See LICENSE for details.

"""
This file contains methods which will be executed by the scheduler on a regular
interval.

This is used for the cleanup of expired keys
"""


from sqlalchemy.sql import and_

from basesoft.scheduler.task import ScheduleTask
from basesoft.web.contrib.mailing.schedule import HtmlMailTask
from basesoft.web.contrib.account.models import Account, AccountActivation, AccountActivationMail, LostPasswordMail

from basesoft import logging
from basesoft.web.core.util.controllers import PermissionController
from basesoft.web.contrib.mailing.models import MailGroup
from basesoft.web.contrib.account.controllers import AccountController
logger = logging.getLogger('basesoft.account.schedule')

class ExpiredKeysCleanupTask(ScheduleTask):
    """
    Removes expired instances of ``AccountActivation`` and their
    associated ``Account``s.

    Accounts to be deleted are identified by searching for
    instances of ``AccountActivation`` with expired activation
    keys, and then checking to see if their associated ``Account``
    instances have the field ``is_active`` set to ``False``; any
    ``Account`` who is both inactive and has an expired activation
    key will be deleted.

    Regularly clearing out accounts which have never been
    activated serves two useful purposes:

    1. It alleviates the occasional need to reset a
       ``AccountActivation`` and/or re-send an activation email
       when a user does not receive or does not act upon the
       initial activation email; since the account will be
       deleted, the user will be able to simply re-register and
       receive a new activation key.

    2. It prevents the possibility of a malicious user registering
       one or more accounts and never activating them (thus
       denying the use of those account names to anyone else); since
       those accounts will be deleted, the account names will become
       available for use again.

    If you have a troublesome ``Account`` and wish to disable their
    account while keeping it in the database, simply delete the
    associated ``AccountActivation``; an inactive ``Account`` which
    does not have an associated ``AccountActivation`` will not
    be deleted.
    """

    # run every 6 hours
    INTERVAL = 21600

    def execute(self):
        """
        TODO: This code is copied from the original account controller but needs to be handled by a task. It needs
              to be properly tested and rely on the database cascading deletion. Also make sure its logging and
              returning properly to the scheduler server. ticket: #428

        #get only the activation keys used for activating an account
        q = self.sa_session.query(AccountActivation).filter(
                and_(AccountActivation.activation_type == ACTIVATION_TYPE_NEW_ACCOUNT))
        for profile in q.all():
            if profile.activation_key_expired():
                account = profile.account
                q = self.sa_session.query(GroupAccount).filter(
                     and_(GroupAccount.account_id == account.id))
                group_account = q.one()
                if not group_account.active:
                    self.sa_session.delete(profile)
                    self.sa_session.delete(account.person)
                    self.sa_session.delete(account)
                    self.sa_session.delete(group_account.group.license.domain)
                    self.sa_session.delete(group_account.group.license)
                    self.sa_session.delete(group_account.group)
                    self.sa_session.delete(group_account)
        """

class AccountActivationMailTask(HtmlMailTask):
    """
    I query all accounts that need to be activated and send out an email with activation key.
    """
    # run every 30 secs
    INTERVAL = 30

    def execute(self):

        """
        if not mail:
            self.logger.warn('AccountActivationMailTask: no mail available, aborting')
            return
        """

        # query all activation mails to be send out
        activations = self._get_activations()

        for activation in  activations:

            # change activation status to 'sending'.. flush so the changes get written to the db
            activation.activation_stage = AccountActivation.ACTIVATION_STAGE_SENDING
            self.sa_session.add(activation)

        self.sa_session.flush()

        for activation in activations:

            # build message and send mail
            person = activation.account.person
            license = activation.account.default_group.license
            mail = self._get_mail(activation.account)
            if mail is None:
                self.logger.warn('AccountActivationMailTask: no activation mail available for account (%s), aborting' % activation.account)
                continue

            msg, result = self._build_message(person, mail, context={
                'activation': activation,
                'license': license
            }, mail_from=mail.mail_from())
            self._send_mail(person.email, msg, result, mail_from=mail.mail_from(), return_path=mail.mail_return_path())

            # change activation status to 'pending'
            activation.activation_stage = AccountActivation.ACTIVATION_STAGE_PENDING
            self.sa_session.add(activation)
            self.sa_session.flush()

            logger.info('%s: Activation mail sent to person_id [%s]' % (self, person.id) )

    def _get_mail(self, account):
        """
        return the activation mail object
        """
        ac = AccountController(self.sa_session, account_id=account.id)
        return ac.get_account_activation_mail(account.default_group)

    def _get_activations(self):
        """
        return the list with account activation objects
        """
        # query all activation mails to be send out
        return self.sa_session.query(AccountActivation).join(Account).filter(and_(
            AccountActivation.activation_stage == AccountActivation.ACTIVATION_STAGE_NEW,
            AccountActivation.activation_type == AccountActivation.ACTIVATION_TYPE_NEW_ACCOUNT
        )).all()

class AccountActivationFromLicenseMailTask(AccountActivationMailTask):
    """
    Sends mails for new accounts linked to a license contact (on license creating)
    """
    def _get_activations(self):
        """
        return the list with account activation objects
        """
        # query all activation mails to be send out
        return self.sa_session.query(AccountActivation).join(Account).filter(and_(
            AccountActivation.activation_stage == AccountActivation.ACTIVATION_STAGE_NEW,
            AccountActivation.activation_type == AccountActivation.ACTIVATION_TYPE_NEW_ACCOUNT_FROM_LICENSE
        )).all()

class LostPasswordActivationMailTask(AccountActivationFromLicenseMailTask):
    """
    I query all accounts that requested password change and send out an email with activation key.
    """

    def _get_mail(self, account):
        """
        return the activation mail object
        """
        ac = AccountController(self.sa_session, account_id=account.id)
        return ac.get_lost_password_mail(account.default_group)

    def _get_activations(self):
        """
        return the list with account activation objects
        """
        # query all activation mails to be send out
        return self.sa_session.query(AccountActivation).join(Account).filter(and_(
            AccountActivation.activation_stage == AccountActivation.ACTIVATION_STAGE_NEW,
            AccountActivation.activation_type == AccountActivation.ACTIVATION_TYPE_CHANGE_PASSWORD
        )).all()

# tasks so the scheduler can pick them up
tasks = (ExpiredKeysCleanupTask, AccountActivationMailTask, AccountActivationFromLicenseMailTask,
         LostPasswordActivationMailTask)
