# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

import datetime

from sqlalchemy.sql.expression import and_, or_
from types import NoneType

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator, LicenseConfigController
from basesoft.web.contrib.file.models import ImageFile
from basesoft.web.contrib.announce.models import Announcement, AnnouncementAccount, AnnouncementGroup
from basesoft.web.contrib.wiki.controllers import WikiController
from basesoft.web.contrib.license.controllers import LicenseController


class AnnounceController(AbstractController, WikiController):

    def get_announcement(self, announcement_id, priv=Announcement.PRIV_VIEW):
        return self.get_privileged_object_by_id(Announcement, announcement_id, priv)[0]

    def get_wiki_from_instance(self, announcement_id):
        """
        implementation for superclass WikiController
        """
        return self.get_announcement(announcement_id).wiki

    def get_wiki_from_instance_for_reading(self, announcement_id):
        """
        implementation for superclass WikiController
        """
        return self.get_announcement(announcement_id, priv=Announcement.PRIV_RECIPIENT).wiki

    def get_logo(self, announcement_id):
        group = self.get_object_by_priv_in_hierarchy(self.get_default_group(), Announcement, priv=Announcement.PRIV_UPDATE, _where=Announcement.id==announcement_id, return_group=True)
        if group is None:
            return None

        lcc = LicenseConfigController(self.sa_session, scope=group.license)
        img = lcc.get('basesoft.license.company.logo')
        if img is None:
            return None

        return self.sa_session.query(ImageFile).get(img).file_hash

    def load_announcement(self, announcement_id):
        """
        called by the details panel for editing the announcement
        """

        a = self.get_announcement(announcement_id)

        rv = {
            'announce': {
                'label': a.label,
                'publish_date': a.publish_date,
                'expire_date': a.expire_date,
                'status': a.status,
                'must_read': a.must_read
            }
        }

        recipients = []
        for g in a.groups:
            if g.privilege!=Announcement.PRIV_RECIPIENT:
                continue
            if g.license_has_privilege:
                recipients.append('license_%s' % g.group.license.id)
                #continue
            if g.domain_has_privilege:
                recipients.append('domain_%s' % g.group.license.domain.id)
                continue

        rv['announce']['recipients'] = recipients
        rv['wiki'] = self._load_wiki(a.wiki)

        return rv

    @transaction_decorator
    def save_announcement(self, announcement_id, state):
        """
        called by the announcement details panel for saving an announcement
        """

        self.assert_any_permissions('manage_announcements')

        a_state = state['announce']
        w_state = state['wiki']

        if a_state['expire_date'] == '':
            a_state['expire_date'] = None

        assert isinstance(a_state['publish_date'], datetime.datetime)
        assert isinstance(a_state['expire_date'], (datetime.datetime, NoneType))
        assert a_state['status'] in (Announcement.STATUS_PENDING, Announcement.STATUS_PUBLISHED)

        if announcement_id is None:
            a = Announcement()
        else:
            a = self.get_announcement(announcement_id, priv=Announcement.PRIV_UPDATE)

        a.label = a_state['label']
        a.publish_date = a_state['publish_date']
        a.expire_date = a_state['expire_date']
        a.status = a_state['status']
        a.must_read = a_state['must_read']
        a.wiki = self._save_wiki(a.wiki, w_state)
        self.sa_session.add(a)

        if announcement_id is None:
            self.save_default_privilege_object_group(a)

        # recipients

        lc = LicenseController(self.sa_session, account_id=self.get_account().id)

        for g in a.groups:
            lc.revoke_privilege_from(a, Announcement.PRIV_RECIPIENT, g, license=True, domain=True, check_own_privs=False)

        recipients = a_state['recipients']
        for r in recipients:
            if r[:8] == 'license_':
                print r[8:]
                license_ = lc.get_licenses(r[8:]).pop()
                self.grant_privilege_to(a, Announcement.PRIV_RECIPIENT, group=False, license=True, check_own_privs=False, target_group=license_.default_group)
                continue
            if r[:7] == 'domain_':

                domain = lc.get_domains(r[7:]).pop()
                self.grant_privilege_to(a, Announcement.PRIV_RECIPIENT, group=False, domain=True, check_own_privs=False, target_group=domain.default_group)
                continue

        return a


    @transaction_decorator
    def mark_as_read(self, announcement_id):
        """
        called when the recipient marks an announcement as read
        """

        assert isinstance(announcement_id, (int, long))

        announcement = self.get_announcement(announcement_id, priv=Announcement.PRIV_RECIPIENT)

        # return if already marked as read
        if len(self.sa_session.query(AnnouncementAccount).filter(AnnouncementAccount.announcement == announcement, AnnouncementAccount.account == self.get_account()).all()) > 0:
            return

        aa = AnnouncementAccount()
        aa.announcement = announcement
        aa.account = self.get_account()
        self.sa_session.add(aa)

        self.logger.info("Account %s marked announcement %s as read" % (self.get_account().id, announcement.id))


    def query_unread_announcements(self, must_read=False):
        """
        returns all unread announcements
        """

        filter_, fo = self.get_query_filter_on_privileges(Announcement, AnnouncementGroup, Announcement.PRIV_RECIPIENT)
        filter_ = and_(filter_, and_(Announcement.deleted==False, Announcement.archived==False))

        fo = fo.outerjoin(AnnouncementAccount, and_(AnnouncementAccount.announcement_id==Announcement.id, AnnouncementAccount.account_id==self.get_account().id))

        where = and_(
            Announcement.must_read == must_read,
            Announcement.status == Announcement.STATUS_PUBLISHED,
            or_(Announcement.expire_date == None, self.now() < Announcement.expire_date),
            Announcement.publish_date >= self.get_account().created,
            AnnouncementAccount.account_id == None
        )
        filter_ = and_(filter_, where)

        query = self.sa_session.query(Announcement).select_from(fo).filter(filter_)

        return query.all()
