# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.
import re

from sqlalchemy import and_

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator,\
    LicenseConfigController
from basesoft.web.contrib.branding.models import BrandingHeaderFooter,\
    LicenseTheme, Site, Theme
from basesoft.web.contrib.license.controllers import LicenseController
from basesoft.web.core.exceptions import SystemException, SecurityException
from basesoft.web.contrib.license.models import License
from basesoft.web.contrib.file.models import ImageFile

class HeaderFooterController(AbstractController):

    def get_header_footers(self, id=None):
        self.assert_permissions('branding_view_header_footer')
        return self.get_privileged_object_by_id(BrandingHeaderFooter, id, BrandingHeaderFooter.PRIV_VIEW)

    def load_state(self, id):
        """
        Loads a header/footer state
        """
        hf = self.get_header_footers(id)[0]

        return {
            'dbID': hf.id,
            'name': hf.name,
            'content': hf.content
        }

    def save_state(self, state):
        """
        Saves or updates a h/f with the given state
        @param state: the state of the h/f to be saved/updated
        @return: the new/updated hf object
        """
        assert isinstance(state, dict)
        assert isinstance(state['name'], unicode)
        assert isinstance(state['content'], unicode)

        if state['dbID']:
            self.assert_permissions('branding_update_header_footer')
            hf = self.get_header_footers(state['dbID'])[0]
        else:
            self.assert_permissions('branding_create_header_footer')
            hf = BrandingHeaderFooter()

        hf.name = state['name']
        hf.content = state['content']

        if not state['dbID']:
            self.save_default_privilege_object_group(hf)

        self.sa_session.add(hf)
        self.sa_session.flush()
        return hf

class SiteController(AbstractController):

    def get_current_site(self):
        """
        returns the Site object matching the current hostname, fallback to ''
        """

        if self.request is not None:
            http_host = self.request.META['HTTP_HOST']
        else:
            self.logger.warn('No request available for site lookup')
            http_host = ''

        sites = self.sa_session.query(Site).order_by(Site.order)
        for site in sites:
            if re.match(site.host_match, http_host):
                return site

        raise SystemException("no site matching hostname: %s" % self.request.META['HTTP_HOST'])

    @transaction_decorator
    def save_or_update_site(self, host_match, order, domain_id, theme_id,
                            home_app='website', analytics_id='UA-5140190-2', trac_conn_str=None):
        self.assert_permissions('site_manager')

        site = self.get_site(host_match)
        if site is None:
            site = Site()
            site.host_match = host_match

        site.order = order
        if domain_id is not None:
            lc = LicenseController(self.sa_session, account_id=self.get_account().id, request=self.request)
            site.domain = lc.get_domains(domain_id).pop()
        else:
            site.domain = None

        site_licensee = None if self.request is not None else self.get_default_group().license

        ltc = LicenseThemeController(self.sa_session, account_id=self.get_account().id, request=self.request, site_licensee=site_licensee)
        site.theme = ltc.get_theme(theme_id)

        site.home_app = home_app
        site.trac_conn_str = trac_conn_str
        site.analytics_id = analytics_id
        self.sa_session.add(site)
        return site

    def get_site(self, host_match=None):
        self.assert_permissions('site_manager')
        if host_match is None:
            self.sa_session.query(Site).all()

        return self.sa_session.query(Site).filter(Site.host_match==host_match).first()

    def get_site_logo(self, site=None):
        if site is None:
            site = self.get_current_site()

        if site.domain is None:
            return None

        lcc = LicenseConfigController(self.sa_session, scope=site.domain.default_license)
        img = lcc.get('basesoft.license.company.logo')
        if img is None:
            return None

        return self.sa_session.query(ImageFile).get(img)

    @transaction_decorator
    def delete_site(self, host_match):
        self.assert_permissions('site_manager')

        if type(host_match) != list:
            host_match = [host_match]

        for site in self.sa_session.query(Site).filter(Site.host_match.in_(host_match)).all():
            self.sa_session.delete(site)

        return True

    def load_site_state(self, host_match):
        site = self.get_site(host_match)

        return {
            'host_match': site.host_match,
            'order': site.order,
            'domain_id': None if site.domain is None else site.domain.id,
            'theme_id': site.theme.id,
            'home_app': site.home_app,
            'analytics_id': site.analytics_id
        }

class LicenseThemeController(AbstractController):
    """
    Controller that handles the retrieval of themes associated with an account, etc.
    """

    def __init__(self, sa_session=None, locale=None, account_id=None, request=None, site_licensee=None):
        AbstractController.__init__(self, sa_session, locale, account_id, request)

        if site_licensee is None:
            lc = LicenseController(sa_session, locale, account_id)
            site_licensee = lc.get_license_for_request(request)

        self.site_licensee = site_licensee

    def get_all_themes(self):
        self.assert_permissions('view_all_domains')
        return self.sa_session.query(Theme).all()

    @transaction_decorator
    def save_themes_for_licenses(self, themes, licenses):
        self.assert_permissions('view_all_domains')

        lcs = self.sa_session.query(License).filter(License.id.in_(licenses)).all()
        themes = self.sa_session.query(Theme).filter(Theme.id.in_(themes)).all()

        for theme in themes:
            for _license in lcs:
                lt = self.sa_session.query(LicenseTheme).filter(and_(LicenseTheme.license==_license, LicenseTheme.theme==theme)).first()
                if lt is not None:
                    continue

                lt = LicenseTheme()
                lt.theme = theme
                lt.license = _license
                self.sa_session.add(lt)

        self.sa_session.flush()

    def delete_license_theme(self, license_id, theme_id):
        self.assert_permissions('view_all_domains')
        lt = self.sa_session.query(LicenseTheme).filter(and_(LicenseTheme.license_id==license_id, LicenseTheme.theme_id==theme_id)).first()
        if lt is None:
            return True


        self.sa_session.delete(lt)
        self.sa_session.flush()
        return True

    def get_available_themes(self):
        """
        Get the available themes based on the site_licensee, as a list of
        """
        available_license_themes = self.sa_session.query(LicenseTheme).filter(LicenseTheme.license==self.site_licensee).all()
        available_themes = [e.theme for e in available_license_themes]
        return available_themes

    def get_current_theme(self):
        """
        Return the Theme object that needs to be used for rendering
        """

        available_themes = self.get_available_themes()

        if self._account is not None:
            theme_config = self.config.get('basesoft.web.core.media.themes')

            # return theme for user if its in available themes
            if theme_config in [t.module_name for t in available_themes]:
                for t in available_themes:
                    if t.module_name == theme_config:
                        print 'returning account config theme: %s' % theme_config
                        return t

        # if we come here either there is no account or config available
        # get the theme based on the request host name from tblSite
        sc = SiteController(self.sa_session, request=self.request)
        site = sc.get_current_site()
        return site.theme

    def get_theme(self, _id):
        themes = self.get_available_themes()
        theme = self.sa_session.query(Theme).filter(Theme.id==_id).first()
        if theme not in themes:
            raise SecurityException('No permission to supplied theme id.')
        return theme
