# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

import datetime

from django.conf import settings
from django.utils.translation import ugettext as _

from sqlalchemy.orm import mapper, relation
from sqlalchemy import Column, Integer, ForeignKey, DateTime, UnicodeText, SmallInteger
from sqlalchemy.types import String

from basesoft.sqlalchemy.table import IQTable
from basesoft.web.contrib.license.hooks import PrivilegeObjectMapper
from basesoft.web.contrib.license import models as license_models

from basesoft.web.contrib.graph.models import GraphTheme
from basesoft.web.contrib.branding.models import Theme, LicenseTheme, Site
from basesoft.web.contrib.license.models import Domain
from basesoft.web.contrib.dashboard.models import Dashboard

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    return tables

def do_tables(metadata):
    global __tables_defined
    if not __tables_defined:

        tblBrandingHeaderFooter = IQTable('tblBrandingHeaderFooter', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('name', UnicodeText(128), nullable=False),
            Column('content', UnicodeText, nullable=False),

            Column('deleted', SmallInteger, nullable=False, default=0),
            Column('archived', SmallInteger, nullable=False, default=0),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
        )

        tblBrandingHeaderFooterGroup = IQTable('tblBrandingHeaderFooterGroup', metadata,
            Column('header_footer_id', ForeignKey('tblBrandingHeaderFooter.id'), primary_key=True),
            Column('group_id', Integer, ForeignKey('tblGroup.id'), primary_key=True),
            Column('privilege', Integer, nullable=False, primary_key=True, autoincrement=False),
            Column('group_has_privilege', SmallInteger, default=0),
            Column('license_has_privilege', SmallInteger, default=0),
            Column('domain_has_privilege', SmallInteger, default=0),
            Column('world_has_privilege', SmallInteger, default=0),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now),
        )

        # theme
        tblTheme = IQTable('tblTheme', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),

            Column('name', String(50), nullable=False), # name of the theme, shown in the preferences dropdown
            Column('title', String(200), nullable=False), # title shown in the main header section
            Column('module_name', String(50), nullable=False), # name of the css module, eg. 'iq_theme_icecream'
            Column('css_class', String(50), nullable=False), # class that's  used for the body, eg. 'icecream tnt'

            Column('default_graph_theme_id', Integer, ForeignKey('tblGraphTheme.id')),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        # holds available themes per license
        # the default theme is taken from tblSite
        tblLicenseTheme = IQTable('tblLicenseTheme', metadata,
            Column('license_id', Integer, ForeignKey('tblLicense.id', ondelete='CASCADE'), primary_key=True),
            Column('theme_id', Integer, ForeignKey('tblTheme.id', ondelete='CASCADE'), primary_key=True),
            #Column('is_default', SmallInteger, default=0, nullable=False),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        # site
        tblSite = IQTable('tblSite', metadata,
            Column('host_match', String(255), nullable=False, primary_key=True),
            Column('order', Integer, nullable=False),
            Column('domain_id', Integer, ForeignKey('tblDomain.id', ondelete='CASCADE'), nullable=True),
            Column('theme_id', Integer, ForeignKey('tblTheme.id', ondelete='CASCADE'), nullable=False),
            Column('home_app', String(100), nullable=False),
            Column('trac_conn_str', String(255), nullable=True),
            Column('analytics_id', String(32), nullable=True, default=settings.GOOGLE_ANALYTICS_ID),
        )


        tables['tblBrandingHeaderFooter'] = tblBrandingHeaderFooter
        tables['tblBrandingHeaderFooterGroup'] = tblBrandingHeaderFooterGroup

        tables['tblTheme'] = tblTheme
        tables['tblLicenseTheme'] = tblLicenseTheme
        tables['tblSite'] = tblSite

        __tables_defined = True

def do_mappings():
    """
    This function is only called by the django middleware
    when it is needed
    """
    global __mapping_defined
    if not __mapping_defined:
        from basesoft.web.contrib.branding import models as branding_models

        mapper(branding_models.BrandingHeaderFooter, tables['tblBrandingHeaderFooter'],
            extension = PrivilegeObjectMapper(),
            properties = {
                'groups': relation(branding_models.BrandingHeaderFooterGroup, backref='header_footer', cascade="all", passive_deletes=True)
            }
        )

        mapper(Theme, tables['tblTheme'],
            properties = {
                'default_graph_theme': relation(GraphTheme),
            }
        )

        mapper(LicenseTheme, tables['tblLicenseTheme'],
            properties = {
                'theme': relation(Theme),
                'license': relation(license_models.License),
            }
        )


        mapper(branding_models.BrandingHeaderFooterGroup, tables['tblBrandingHeaderFooterGroup'],
            properties = {
                'group': relation(license_models.Group)
            }
        )

        mapper(Site, tables['tblSite'],
            properties = {
                'domain': relation(Domain),
                'theme': relation(Theme),
            }
        )


        __mapping_defined = True

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'Branding')

def app_permissions():
    """
    The permissions used within this application. Can be sync'd with the
    `iqsyncpermissions` command.
    """
    return [
        license_models.Permission(u"branding_view_header_footer", u"""
            This gives the right to view global header/footer definitions.
            """, _("View header/footer")),
        license_models.Permission(u"branding_create_header_footer", u"""
            This gives the right to create global header/footer definitions.
            """, _("Create header/footer")),
        license_models.Permission(u"branding_update_header_footer", u"""
            This gives the right to update global header/footer definitions.
            """, _("Update header/footer")),
        license_models.Permission(u"branding_delete_header_footer", u"""
            This gives the right to delete global header/footer definitions.
            """, _("Delete header/footer")),

        license_models.Permission(u"site_manager", u"""
            This gives the right to manage Sites.
            """, _('Site management'))
    ]
