# Copyright (c) 2012 BaseSoft B.V.
# See LICENSE for details.

"""
This file contains methods which will be executed by the scheduler on a regular interval. They take care of time related
issues like publishing and finishing of compose modules.

Notice that they do not run in a 'web session' so they should not depend on anything like that. Think of it that they
run as a 'root' user.

This file contains two subclasses of ScheduleTask and are added to the 'tasks' list in this same module. The scheduler
loads these tasks and instanciates them for each interval.
"""

from datetime import datetime

from sqlalchemy.sql import and_

from basesoft.scheduler.task import ScheduleTask
from basesoft.web.contrib.mailing.schedule import MailingTask
from basesoft.web.contrib.mailing.models import Mailing, MailResult
from basesoft.web.contrib.compose.models import ComposePublication, ComposeInvitee, ComposePublicationMailing
from basesoft.web.contrib.compose.controllers import ComposeInviteeController
from basesoft import logging
logger = logging.getLogger("basesoft.web.contrib.compose.schedule")

class PublishComposeTask(ScheduleTask):
    INTERVAL = 600 # every ten minutes
    model = ComposePublication

    def _get_publications(self):
        # get all scheduled publications in the current period
        return self.sa_session.query(self.model).filter(
            (self.model.status==self.model.STATUS_SCHEDULED) &
            (self.model.archived==0) &
            (self.model.deleted==0) &
            (self.model.begin_date <= datetime.now()) &
                ((self.model.end_date > datetime.now()) |
                (self.model.end_date==None))).all()

    def _create_anonymous_invitee(self, publication):
        # check if a survey has anonymous flag set, then create one anonymous Invitee
        if publication.anonymous:
            i = ComposeInvitee()
            # get the already generated hash from ComposePublication
            i.invitee_hash = publication.anonymous_hash

            publication.invitees.append(i)

            logger.info('PublishComposeTask: Publication [%s] is anonymous.. created anonymous ComposeInvitee: [%s]'
                        % (publication.id, i.invitee_hash))

    def execute(self):

        rv = []

        # get all scheduled publications in the current period
        result = self._get_publications()

        for publication in result:
            logger.info('PublishComposeTask: Publication [%s] changing status to STATUS_ONLINE.' % (publication.id))
            rv.append({'publication_id': publication.id, 'msg': 'changed status to STATUS_ONLINE'})

            publication.status = self.model.STATUS_ONLINE

            self._create_anonymous_invitee(publication)

            self.sa_session.add(publication)

        return rv

class CreatePublicationInviteesTask(ScheduleTask):
    INTERVAL = 90
    model = ComposePublication
    controller = ComposeInviteeController

    def execute(self):

        # get all publications that are online
        result = self.sa_session.query(self.model).filter(
            and_(self.model.status==self.model.STATUS_ONLINE,
                 self.model.archived==0,
                 self.model.deleted==0,
            )).all()

        if len(result) > 0:
            rc = self.controller(self.sa_session)

        affected = []

        for w in result:
            # now we add invitees to the publication.
            invitees = rc.save_or_update_invitees(w)
            if len(invitees):
                affected.append(w)
                logger.info('CreatePublicationInviteesTask: Added [%s] Invitees to publication [%s]' % (len(invitees), w.id))

        return affected

class FinishPublicationsTask(ScheduleTask):

    INTERVAL = 60 # seconds
    model = ComposePublication

    def execute(self):
        """
        Query all publications that should not currently be online
        """
        # result is the return value with data of surveys that went offline
        rv = []

        # get all surveys that are supposed to be/stay online
        query = self.sa_session.query(self.model).filter(and_(
                         self.model.status==self.model.STATUS_ONLINE,
                         self.model.end_date <= datetime.now(),
                        ))

        # set them to finished
        for publication in query.all():
            logger.info('FinishComposeTask: Publication [%s] changing status to STATUS_FINISHED' % publication.id)
            rv.append({'survey_id': publication.id, 'msg': 'changed status to STATUS_FINISHED'})

            publication.status = self.model.STATUS_FINISHED
            self.sa_session.add(publication)

            # all attached mailings will be set to status archived
            pub_mailings = publication.mailings
            for mailing in pub_mailings:
                mailing.status = Mailing.STATUS_ARCHIVED
                mailing.number_of_recipients = self.sa_session.query(MailResult).filter(MailResult.mailing==mailing).count()
                self.sa_session.add(mailing)

        self.sa_session.flush()

        return rv

class PublicationMailing(MailingTask):
    """
    Implements a MailingTask that takes care of sending out mails to invitees.
    This will be ran immediately after the PublishTask has finished (it should not start automaticly).
    """
    INTERVAL = 600
    mailing_model = ComposePublicationMailing
    publication_model = None

    def get_mailings(self):
        if self.publication_model is None:
            return []
        return self.sa_session.query(self.mailing_model).join(self.publication_model).filter( \
            and_(self.mailing_model.status.in_([Mailing.STATUS_SCHEDULED, Mailing.STATUS_PREPARING_FOR_SENDING]), self.mailing_model.send_date <= datetime.now() )).all()

# all this tasks need to be overridden by children
tasks = ()
