# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.
import os

from datetime import datetime, date

from django.utils.translation import ugettext as _, ugettext_lazy

from django.utils import simplejson
from django.conf import settings

from sqlalchemy.sql import or_, and_, text
from sqlalchemy.orm import join
from sqlalchemy.orm.exc import NoResultFound

from basesoft.web.core.util.controllers import \
    AbstractController, transaction_decorator, LicenseConfigController, get_models_of_type, AccountConfigController

from basesoft.web.contrib.license.models import Group

from basesoft.web.contrib.i18n.models import Country, DomainCountry, Locale
from basesoft.web.contrib.i18n.controllers import I18nController

from basesoft.web.contrib.contact.models import ContactList, Contact, \
            ContactListGroup, ContactGroup, \
            Person, Address, BankAccount, \
            Company, CompanyAddress, CompanyBankAccount, ContactListContact, ContactFieldMapping

from basesoft.web.core.input.validation import plugins, group_plugins
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.exceptions import PublicUserException, SecurityException, SystemException

from basesoft.web.core.jsonrpc.dispatcher import JsonRpcJSONEncoder, json_object_hook

from basesoft import logging
from basesoft.web.contrib.file.models import ImportFileType, ImportFile
from basesoft.web.contrib.contact.importers import CsvImporter, XlsxImporter, XlsImporter
from basesoft.web.contrib.account.models import GroupAccount
from sqlalchemy.exc import IntegrityError


logger = logging.getLogger("basesoft.web.contrib.contact")

# options for when deleting a contact list or a list of contact lists:
# do not delete any contacts that are attached to the contact list(s) that will be removed
CONTACTS_DO_NOT_DELETE = 1
# delete only contacts that are attached only to the contact lists(s) that will be removed
CONTACTS_DELETE_ONLY_EXCLUSIVE = 2
# delete all contacts that are attached to the contact lists(s) that will be removed
CONTACTS_DELETE_ALL = 3

class SimpleContactController(AbstractController):
    """
    Controller for the Contact app representing the contact details: Address,
    Person, Company, BankAccount etc.
    Contains protected methods to be called in classes that extend it.
    """

    _instance_class = Contact

    def get_instance_class(self):
        return self._instance_class

    def _set_address(self, id, address_data):
        """
        Method for creating an Address object in order to save_or_update.
        It performs the arguments check and sets the attributes to the object
        """
        assert isinstance(address_data, dict)
        has_data = False
        for key in address_data:
            assert key in ["country", "country_name", "postcode", "street", "street_2", "city", "state", "country_iso_3166_a2", "country_iso_3166_a3"], key
            if address_data[key] not in ["", None]:
                has_data = True

        # return None if the state is empty
        if id is None and not has_data:
            return None

        if 'country' in address_data and not isinstance(address_data["country"], Country):
            countries = I18nController(self.sa_session).get_countries(id = address_data["country"])
            # remove the country property in case an invalid one was set
            if len(countries) > 0:
                address_data["country"] = I18nController(self.sa_session).get_countries(id = address_data["country"])[0]
            else:
                del address_data["country"]

        if id is None:
            address = Address()
        else:
            address = self.sa_session.query(Address).filter(Address.id==id).one()

        for key in address_data:
            address.__setattr__(key, address_data[key])
        return address

    def _load_address_state(self, id):
        """
        Returns dict with address state
        """
        address = self.sa_session.query(Address).filter(Address.id==id).one()
        country_name = None
        country_iso_3166_a2 = None
        country_iso_3166_a3 = None

        if address.country is not None:
            country_name = address.country.name
            country_iso_3166_a2 = address.country.iso_3166_a2
            country_iso_3166_a3 = address.country.iso_3166_a3

        return {'dbID': address.id,
                'street': address.street,
                'street_2': address.street_2,
                'postcode': address.postcode,
                'state': address.state,
                'city': address.city,
                'country': address.country_id,
                'country_name': country_name,
                'country_iso_3166_a2': country_iso_3166_a2,
                'country_iso_3166_a3': country_iso_3166_a3,
                }

    @transaction_decorator
    def _save_or_update_address(self, id, address_data):
        """
        Save address
        If id is not None, update the address with the provided id. If id is None, insert the address

        @param id: {int} or None
        @param address_data: {dict} with
                country: {int} country id or {Country} object
                postcode: {String}
                street: {String}
                city: {String}
                state: {String}
        """
        address = self._set_address(id, address_data)
        if address is None:
            return None
        self.sa_session.add(address)
        return address

    def _set_bank_account(self, id, bank_account_data):
        """
        @param id: {int} or None
        @param bank_account_data: {dict} with
                number: {String}
                iban: {String}
                bic: {String}
        """
        assert isinstance(bank_account_data, dict)
        has_data = False
        for key in bank_account_data:
            assert key in ['number', 'iban', 'bic', 'beneficiary', 'bank_name', 'note', 'address_data']
            if bank_account_data[key] != "":
                has_data = True

        # return None if the state is empty
        if id is None and not has_data:
            return None

        if id is None:
            bank_account = BankAccount()
        else:
            bank_account = self.sa_session.query(BankAccount).filter(
                               BankAccount.id==id).one()

        for key in bank_account_data:
            if key == 'address_data':
                _id = None
                if bank_account_data[key].has_key('dbID'):
                    del bank_account_data[key]['dbID']
                if bank_account.address is not None:
                    _id = bank_account.address.id
                bank_account.address = self._set_address(_id, bank_account_data[key])
                continue
            bank_account.__setattr__(key, bank_account_data[key])

        return bank_account

    def _load_bank_account_state(self, id):
        """
        Returns dict with address state
        """
        ba = self.sa_session.query(BankAccount).filter(BankAccount.id==id).one()
        rv = {'dbID': ba.id,
            'number': ba.number,
            'iban': ba.iban,
            'bic': ba.bic,
            'bank_name': ba.bank_name,
            'note': ba.note,
            'beneficiary': ba.beneficiary,
            'address_data': {},
        }
        if ba.address is not None:
            rv['address_data'] = self._load_address_state(ba.address.id)
        return rv

    @transaction_decorator
    def _save_or_update_bank_account(self, id, bank_account_data):
        """
        Save or update the bank account information
        @param id: {int} or None
        @param bank_account_data: {dict} with
                number: {String}
                iban: {String}
                bic: {String}
        """
        bank_account = self._set_bank_account(id, bank_account_data)
        if bank_account is None:
            return None
        self.sa_session.add(bank_account)
        return bank_account

    @transaction_decorator
    def _save_or_update_company_address(self, company, address, data):
        """
        @param company: {object} a Company model object
        @param address: {object} an Address model object
        @param data: {dict} with
                is_invoice: {Bool}
                is_delivery: {Bool}
        """
        assert isinstance(company, Company)
        assert isinstance(address, Address)
        for key in data:
            assert key in ['is_invoice', 'is_delivery']

        # lets see if there is an entry for this combination
        cas = self.sa_session.query(CompanyAddress).filter(
                     and_(CompanyAddress.company==company,
                          CompanyAddress.address==address)).all()
        if len(cas) == 1:
            ca = cas.pop()
        else:
            ca = CompanyAddress()
            ca.company = company
            ca.address = address

        ca.is_invoice = data['is_invoice']
        ca.is_delivery = data['is_delivery']
        self.sa_session.add(ca)
        return ca

    @transaction_decorator
    def _save_or_update_company_bank_account(self, company, bank_account, data):
        """
        @param company: {object} a Company model object
        @param bank_account: {object} a BankAccount model object
        @param data: {dict} with
                is_default: {Bool}
        """
        assert isinstance(company, Company)
        assert isinstance(bank_account, BankAccount)
        for key in data:
            assert key in ['is_default']

        cbas = self.sa_session.query(CompanyBankAccount).filter(
                     and_(CompanyBankAccount.company==company,
                          CompanyBankAccount.bank_account==bank_account)).all()
        if len(cbas) == 1:
            cba = cbas.pop()
        else:
            cba = CompanyBankAccount()
            cba.company = company
            cba.bank_account = bank_account

        cba.is_default = data['is_default']
        self.sa_session.add(cba)
        return cba

    def get_person(self, id):
        assert id is not None
        return self.sa_session.query(self._instance_class.person_instance_class).filter(
                self._instance_class.person_instance_class.id==id).one()

    def _get_import_person_object_filter(self, person_data, unique_fields):
        """
        Sets filters and joins for person unique fields
        """
        if len(unique_fields) == 0:
            return False, None, None, person_data

        qfrom = join(self._instance_class.person_instance_class, Contact).join(ContactGroup)
        filter, qfrom = self.get_query_filter_on_privileges(Contact, ContactGroup, Contact.PRIV_UPDATE, qfrom)

        address_join = False
        has_filter = False
        for uf in unique_fields:
            # at this point the state's country columns were checked and country obj was made
            if uf.startswith('country_') and uf[8:] in Country.cols_i18n.keys() and person_data.has_key('country'):
                filter = and_(filter, Address.country==person_data['country'])
                address_join = True
                has_filter = True
            elif uf in Person.cols_i18n.keys():
                filter = and_(filter, getattr(self._instance_class.person_instance_class, uf)==person_data[uf])
                has_filter = True
            elif uf in Address.cols_i18n.keys():
                filter = and_(filter, getattr(Address, uf)==person_data[uf])
                address_join = True
                has_filter = True

        if address_join:
            qfrom = qfrom.join(Address)

        return has_filter, filter, qfrom, person_data

    def set_import_person_state(self, person_data):
        """
        Sets the person state for the import. This only sets the correct fields (this will be used before calling
        validation). This method will also be called when setting person_for_import which will check for unique fields.
        """
        person_data['polymorphic_type'] = self._instance_class.person_instance_class.STR
        return person_data

    def set_person_for_import(self, person_data, unique_fields=[]):
        """
        Used for contact import. Returns person import state but it also checks for unique fields and returns the state
        containing the found object.
        """

        person_data = self.set_import_person_state(person_data)
        if len(unique_fields) == 0:
            return person_data

        has_filter, filter, qfrom, person_data = self._get_import_person_object_filter(person_data, unique_fields)

        if not has_filter:
            return person_data

        query = self.sa_session.query(self._instance_class.person_instance_class).select_from(qfrom).filter(filter).all()

        if len(query) == 0:
            return person_data

        if len(query) >= 1:
            if len(query) > 1:
                logger.warn('Found multiple instances for Person with unique set on '+str(unique_fields)+'. Will only get the first one!')
            person_data['person_object'] = query[0]
        return person_data

    def _set_person(self, id, person_data):
        """
        Method for creating a Person object in order to save_or_update.
        It performs the arguments check and sets the attributes to the object
        """
        assert isinstance(person_data, dict)
        if not person_data.has_key('polymorphic_type'):
            person_data['polymorphic_type'] = self._instance_class.person_instance_class.STR

        # used for contacts import, when updating (so we don't do another select)
        if person_data.has_key('person_object') and isinstance(person_data['person_object'], self._instance_class.person_instance_class):
            person = person_data['person_object']
            del person_data['person_object']
        else:
            if id is None:
                person = self._instance_class.person_instance_class()
            else:
                person = self.get_person(id)
                if person.polymorphic_type != person_data['polymorphic_type']:
                    raise SecurityException("%s trying to change person type for person id %s" % (self.account_id, id))

        for key in person_data:
            if key != "address_data":

                # strip the time from the birthdate in case it is a datetime object
                if key == "birth_date":
                    if isinstance(person_data[key], datetime):
                        person.birth_date = person_data[key].date()
                    elif isinstance(person_data[key], date):
                        person.birth_date = person_data[key]
                    else:
                        person.birth_date = None
                    continue

                if key == 'preferred_locale':
                    person.preferred_locale = self._set_person_preferred_locale(person_data, person_data[key])
                    if len(person.accounts) == 0 or person.preferred_locale is None:
                        continue
                    for account in person.accounts:
                        # we only set the account preferred locale if this locale is available for site translation.
                        i18nc = I18nController(self.sa_session)
                        locales = i18nc.get_domain_locales(domain_id=account.default_group.license.domain.id, availability=Locale.AVAILABLE_FOR_SITE_TRANSLATION)
                        if person.preferred_locale in locales:
                            acc = AccountConfigController(self.sa_session, account_id=account.id)
                            acc.set('basesoft.web.contrib.locale', person.preferred_locale.locale)
                    continue

                if key == "gender":

                    if person_data[key] == '':
                        person_data[key] = None

                    if person_data[key] is not None:
                        for gender in Person.gender_options:
                            if person_data[key].lower() in Person.gender_map[gender]:
                                person_data[key] = gender
                if hasattr(person, key):
                    setattr(person, key, person_data[key])
            else:
                if person_data['address_data'].has_key('dbID'):
                    del person_data['address_data']['dbID']
                if person.address is None: addr_id = None
                else: addr_id = person.address.id
                person.address = self._set_address(addr_id, person_data['address_data'])
        return person

    def _set_person_preferred_locale(self, person_data, preferred_locale):
        # already supplied as a locale instance.
        if isinstance(preferred_locale, Locale):
            return preferred_locale

        # check to see if we are dealing with an id.
        if preferred_locale is not None:
            try:
                id = int(preferred_locale)
                locale = self.sa_session.query(Locale).filter(Locale.id==id).first()
                if locale is not None:
                    return locale
                raise PublicUserException(_('Incorrect locale ID supplied.'))
            except ValueError: # we are not dealing with a Locale id
                pass

        # we are dealing with a string, we should make a locale based on the address country and the language code.
        if type(preferred_locale) in (str, unicode, ):
            if preferred_locale == '':
                return None
            if not person_data.has_key('address_data') and not person_data.has_key('country'):
                raise PublicUserException(_('A country is required in order to determine the correct language for this person.'))
            country = None
            if person_data.has_key('country'):
                country = person_data['country']
            else:
                address = self._set_address(None, person_data['address_data'])

                if address is not None:
                    if address.country is None:
                        raise PublicUserException(_('A country is required in order to determine the correct language for this person.'))
                    country = address.country

            if country is not None:
                locale_str = '%s_%s' % (preferred_locale.lower(), country.iso_3166_a2.upper())
                return self.sa_session.query(Locale).filter(Locale.locale==locale_str).first()
        return None

    def _load_person_state(self, p):
        """
        Returns dict with person state
        """
        if not isinstance(p, self._instance_class.person_instance_class):
            person = self.get_person(p)
        else:
            person = p

        state = {'dbID': person.id,
            'polymorphic_type': person.polymorphic_type,
            'first_name': person.first_name,
            'middle_name': person.middle_name,
            'last_name': person.last_name,
            'title': person.title,
            'gender': person.gender,
            'birth_date': person.birth_date,
            'email': person.email,
            'phone': person.phone,
            'mobile_phone': person.mobile_phone,
            'fax': person.fax,
            'department': person.department,
            'position': person.position,
            'skype_name': person.skype_name,
            'preferred_locale_code': None,
            'external_person_id': person.external_person_id
        }

        if person.preferred_locale is not None:
            state['preferred_locale'] = person.preferred_locale.id
            state['preferred_locale_str'] = person.preferred_locale.locale

        if person.address_id is not None:
            state['address_data'] = self._load_address_state(person.address_id)
        return state

    @transaction_decorator
    def _save_or_update_person(self, id, person_data):
        """
        Save a person
        If id is not None, update the person with the provided id. If id is None, insert the person

        @param id: {int} or None
        @param person_data: {Dict} with:
                    first_name: {String}
                    middle_name: {String}
                    last_name: {String}
                    title: {String}
                    gender: {String}
                    email: {String}
                    position: {String}
                    department: {String}
                    phone: {String}
                    mobile_phone: {String}
                    fax: {String}
                    birth_date: {Date}
                    address_data: {Dict} with
                        country: {Country}
                        postcode: {String}
                        street: {String}
                        city: {String}
                        state: {String}
        """
        person = self._set_person(id, person_data)
        self.sa_session.add(person)
        return person

    def get_person_cols(self):
        """
        Used for contact import. Returns a list with this person type columns.
        """
        p = []
        for col, l in self._instance_class.person_instance_class.cols_i18n.iteritems():
            p.append({"value": col, "label": l})
        return p

    def person_required_columns(self):
        """
        Return list with person columns that should be selected in the fields mapping stage.

        If there are combinations of options, this will return list of lists meaning at least one of the options (list)
        must be selected in field mappings.
        """
        return []

    def get_company(self, id):
        assert id is not None
        return self.sa_session.query(self._instance_class.company_instance_class).filter(
                self._instance_class.company_instance_class.id==id).one()

    def _set_company(self, id, company_data, instance_class=None):
        """
        @param id: {int} or None
        @param company_data: {dict} with
                name: {String}
                vat_no: {String}
                website: {String}
                contact_person_data: {dict}
        """
        assert isinstance(company_data, dict)
        if not company_data.has_key('polymorphic_type'):
            company_data['polymorphic_type'] = self._instance_class.company_instance_class.STR

        # used for contacts import, when updating (so we don't do another select)
        if company_data.has_key('company_object') and isinstance(company_data['company_object'], self._instance_class.company_instance_class):
            company = company_data['company_object']
            del company_data['company_object']
        else:
            if id is None:
                if instance_class is None:
                    company = self._instance_class.company_instance_class()
                else:
                    company = instance_class()
            else:
                company = self.get_company(id)
                if company.polymorphic_type != company_data['polymorphic_type']:
                    raise SecurityException("Trying to change company type for company id %s" % id)

        fields = [
            'name',
            'vat_no',
            'website',
            'chamber_of_commerce_nr',
            'de_tax_number',
            'es_cif_nr',
            'fr_rcs_number',
            'fr_rcs_location',
            'fr_company_type',
            'fr_naf_code',
            'it_fiscal_code',
            'it_pec_address',
            'polymorphic_type',
            'is_temp',
            'external_company_id'
        ]

        if self.has_permissions('manage_company_notes'):
            fields.append('notes')

        for key in fields:
            if company_data.has_key(key):
                if company_data[key] == '':
                    company_data[key] = None
                company.__setattr__(key, company_data[key])

        if company_data.has_key('address_data'):
            aid = None
            if company_data['address_data'].has_key('dbID'):
                aid = company_data['address_data']['dbID']
                del company_data['address_data']['dbID']

            address = self._set_address(aid, company_data['address_data'])
            if address is not None:
                try:
                    ca = self.sa_session.query(CompanyAddress).filter(CompanyAddress.address_id==address.id, CompanyAddress.company_id==company.id).one()
                except NoResultFound:
                    ca = CompanyAddress()
                    ca.company = company
                    ca.address = address
                    ca.is_invoice = True
                    ca.is_delivery = True
                    company.addresses.append(ca)

        if company_data.has_key('bank_data'):
            bid = None
            if company_data['bank_data'].has_key('dbID'):
                bid = company_data['bank_data']['dbID']
                del company_data['bank_data']['dbID']

            ba = self._set_bank_account(bid, company_data['bank_data'])

            if ba is not None:

                try:
                    ba = self.sa_session.query(CompanyBankAccount).filter(CompanyBankAccount.bank_account_id==ba.id, CompanyBankAccount.company_id==company.id).one()
                except NoResultFound:
                    cbk = CompanyBankAccount()
                    cbk.company = company
                    cbk.bank_account = ba
                    cbk.is_default = True
                    company.bank_accounts.append(cbk)

        # contact_person
        if company_data.has_key('contact_person_data'):
            if isinstance(company_data['contact_person_data'], Person):
                company.contact_person =  company_data['contact_person_data']
            else:
                if company.contact_person is None:
                    person_id = None
                else:
                    person_id = company.contact_person.id
                company.contact_person = self._set_person(person_id,
                                                          company_data['contact_person_data'])

        if company_data.has_key('invoice_person_data'):
            if isinstance(company_data['invoice_person_data'], Person):
                company.invoice_person =  company_data['invoice_person_data']
            else:
                person_id = None
                if company.invoice_person is not None:
                    if company.contact_person != company.invoice_person:
                        person_id = company.invoice_person.id
                if company_data['invoice_person_data'].has_key('is_same_person_as') and company_data['invoice_person_data']['is_same_person_as']:
                    company.invoice_person = company.contact_person
                else:
                    company.invoice_person = self._set_person(person_id,
                                                          company_data['invoice_person_data'])

        return company

    def _load_company_state(self, company):
        """
        Returns dict with company state
        """
        if company is None:
            return {}
        if not isinstance(company, self._instance_class.company_instance_class):
            c = self.get_company(company)
        else:
            c = company
        state = {
            'dbID': c.id,
            'polymorphic_type': c.polymorphic_type,
            'name': c.name,
            'vat_no': c.vat_no,
            'website': c.website,
            'chamber_of_commerce_nr': c.chamber_of_commerce_nr,
            'de_tax_number': c.de_tax_number,
            'es_cif_nr': c.es_cif_nr,
            'fr_rcs_number': c.fr_rcs_number,
            'fr_rcs_location': c.fr_rcs_location,
            'fr_company_type': c.fr_company_type,
            'fr_naf_code': c.fr_naf_code,
            'it_fiscal_code': c.it_fiscal_code,
            'it_pec_address': c.it_pec_address,
            'notes': c.notes,
            'external_company_id': c.external_company_id
        }

        if len(c.addresses) > 0:
            state['address_data'] = self._load_address_state(c.addresses[0].address_id)
        if len(c.bank_accounts) > 0:
            state['bank_data'] = self._load_bank_account_state(c.bank_accounts[0].bank_account_id)
        if c.contact_person is not None:
            state['contact_person_data'] = self._load_person_state(c.contact_person)
        if c.invoice_person is not None:
            state['invoice_person_data'] = self._load_person_state(c.invoice_person)
        if c.invoice_person is not None and c.invoice_person == c.contact_person:
            state['invoice_person_data']['is_same_person_as'] = True

        return state

    load_company_state = _load_company_state

    @transaction_decorator
    def _save_or_update_company(self, id, company_data, instance_class=None):
        """
        @param id: {int} or None
        @param company_data: {dict} with
                name: {String}
                vat_no: {String}
                website: {String}
                contact_person_data: {dict}
        """
        company = self._set_company(id, company_data, instance_class=instance_class)
        self.sa_session.add(company)
        return company

    def get_company_cols(self):
        """
        Used for contact import. Returns a list with this company type columns.
        """
        c = []
        for col, l in self._instance_class.company_instance_class.cols_i18n.iteritems():
            c.append({"value": col, "label": l})
        return c

    def get_contact_cols(self):
        c = []
        for col, l in Contact.cols_i18n.iteritems():
            c.append({"value": col, "label": l})
        return c

    def company_required_columns(self):
        """
        Return list with company columns that should be selected in the fields mapping stage

        If there are combinations of options, this will return list of lists meaning at least one of the options (list)
        must be selected in field mappings.
        """
        return []

    def get_field_plugins(self):
        """
        Used for contact import
        """
        domain = self.get_default_group().license.domain
        field_plugins = {
                #'first_name': [(plugins.IQPortalName, {})],
                #'middle_name': [(plugins.IQPortalName, {})],
                'gender': [(plugins.Gender, {})],
                'email': [(plugins.Email, {})],
                'contact_email': [(plugins.Email, {})],
                'id': [(plugins.Number, {})],
                'address_id': [(plugins.Number, {})],
                'country_name': [(plugins.Country, {"db_session": self.sa_session, "format": "name", "domain": domain})],
                'country_iso_3166_a2': [(plugins.Country, {"db_session": self.sa_session, "format": "iso_3166_a2"})],
                'country_iso_3166_a3': [(plugins.Country, {"db_session": self.sa_session, "format": "iso_3166_a3"})],
                'company_country_name': [(plugins.Country, {"db_session": self.sa_session, "format": "name", "domain": domain})],
                'company_country_iso_3166_a2': [(plugins.Country, {"db_session": self.sa_session, "format": "iso_3166_a2"})],
                'company_country_iso_3166_a3': [(plugins.Country, {"db_session": self.sa_session, "format": "iso_3166_a3"})]
            }
        gp = [
            (group_plugins.Required, {'contact_list_name': {'mail_settings': True}}), # if mail_settings has a value, contact list is required
        ]
        return field_plugins, gp

    def set_import_company_state(self, company_data):
        """
        Sets the company state for the import. This only sets the correct fields (this will be used before calling
        validation). This method will also be called when setting company_or_import which will check for unique fields.
        """
        company_data['polymorphic_type'] = self._instance_class.company_instance_class.STR
        return company_data

    def _get_import_company_object_filter(self, company_data, unique_fields, ret_base=False):
        """
        Sets filters and joins for company unique fields
        """

        if len(unique_fields) == 0:
            return False, None, None, company_data

        qfrom = join(self._instance_class.company_instance_class, Contact).join(ContactGroup)
        filter, qfrom = self.get_query_filter_on_privileges(Contact, ContactGroup, Contact.PRIV_UPDATE, qfrom)

        address_join = False
        bk_join = False
        has_filter = False

        for uf in unique_fields:
            # at this point the state's country columns were checked and country obj was made
            if uf.startswith('company_country_') and uf[16:] in Country.cols_i18n.keys() and company_data.has_key('country'):
                filter = and_(filter, Address.country==company_data['country'])
                address_join = True
                has_filter = True
            elif uf.startswith('company_') and uf[8:] in Address.cols_i18n.keys():
                filter = and_(filter, getattr(Address, uf[8:])==company_data[uf])
                address_join = True
                has_filter = True
            elif uf in BankAccount.cols_i18n.keys():
                filter = and_(filter, getattr(BankAccount, uf)==company_data[uf])
                bk_join = True
                has_filter = True
            elif uf in Company.cols_i18n.keys():
                filter = and_(filter, getattr(self._instance_class.company_instance_class, uf)==company_data[uf])
                has_filter = True

        if address_join:
            qfrom = qfrom.join(CompanyAddress).join(Address)
        if bk_join:
            qfrom = qfrom.join(CompanyBankAccount).join(BankAccount)

        return has_filter, filter, qfrom, company_data

    def set_company_for_import(self, company_data, unique_fields=[]):
        """
        Used for contact import. Returns company import state but it also checks for unique fields and returns the state
        containing the found object.
        """
        company_data = self.set_import_company_state(company_data)
        if len(unique_fields) == 0:
            return company_data

        has_filter, filter, qfrom, company_data = self._get_import_company_object_filter(company_data, unique_fields)

        if not has_filter:
            return company_data

        query = self.sa_session.query(self._instance_class.company_instance_class
                    ).select_from(qfrom).filter(filter).all()

        if len(query) == 0:
            return company_data

        if len(query) >= 1:
            if len(query) > 1:
                logger.warn('Found multiple instances for Company with unique set on '+str(unique_fields)+'. Will only get the first one!')
            company_data['company_object'] = query[0]
        return company_data

    def set_contact_for_import(self, contact_data, unique_fields=[]):
        """
        Used for contact import. Returns the contact state but also checks for unique fields.
        """
        if len(unique_fields) == 0:
            return contact_data

        qfrom = join(self._instance_class, ContactGroup)
        _filter, qfrom = self.get_query_filter_on_privileges(self._instance_class, ContactGroup, Contact.PRIV_UPDATE, qfrom)
        has_filter = False

        for uf in unique_fields:
            f = uf
            if uf == 'contact_email':
                f = 'email'
            elif uf not in self._instance_class.cols_i18n.keys():
                continue
            has_filter = True
            _filter = and_(_filter, getattr(self._instance_class, f)==contact_data[uf])

        if not has_filter:
            return contact_data

        query = self.sa_session.query(self._instance_class
                    ).select_from(qfrom).filter(_filter).all()

        if len(query) == 0:
            contact_data['contact_object'] = self._instance_class()
            return contact_data

        if len(query) >= 1:
            if len(query) > 1:
                logger.warn('Found multiple instances for Contact with unique set on '+str(unique_fields)+'. Will only get the first one!')
            contact_data['contact_object'] = query[0]

        return contact_data

    def validate_state(self, state):
        """
        Used for contact import. Validate company and/or person data. State contains person data plus company dict.
        """
        if not isinstance(state, list):
            states = [state]
        else:
            states = state
        ret = []

        # validate bank account related info
        bk_keys = ['number', 'iban', 'bic']
        for s in states:
            d = {}
            if s.has_key('company'):
                state = s['company']
                for key in bk_keys:
                    if state.has_key(key) and state[key] not in ['', None]:
                        for k in bk_keys:
                            if k != key:
                                if not state.has_key(k) or state[k] in ['', None]:
                                    d[key] = _("Data for 'account number', 'iban' and 'bic' required in order to save bank info")
            if s.has_key('preferred_locale'):
                try:
                    self._set_person_preferred_locale(s, s['preferred_locale'])
                except PublicUserException, e:
                    d['preferred_locale'] = e.message
            ret.append(d)
        return ret

    def cleanup_temporary_companies(self):
        for company in self.sa_session.query(Company).filter(Company.is_temp!=Company.TEMPORARY_STATUS_NONE).all():
            try:
                ip = company.invoice_person
                cc = company.contact_person
                self.sa_session.delete(company)
                if ip is not None:
                    self.sa_session.delete(ip)
                if cc is not None:
                    self.sa_session.delete(cc)
                self.sa_session.flush()
            except IntegrityError, e: # we will not delete temporary companies which are still referenced.
                continue

contact_controllers = {'Contact': SimpleContactController}

class ContactController(AbstractController):
    """
    Controller for the Contact app
    """
    # target for column names (column can be for person information, for
    # company information, contact list or will not be imported)
    COLUMN_IS_NOT_IMPORTED = 0
    COLUMN_IS_PERSON_INFO = 1
    COLUMN_IS_COMPANY_INFO = 2
    COLUMN_IS_CONTACT_LIST = 3
    COLUMN_IS_CONTACT = 4
    COLUMN_IS_CONTACT_LIST_CONTACT = 5


    def __init__(self, sa_session, account_id=None, contact_type='Contact'):
        AbstractController.__init__(self, sa_session, locale=None, account_id=account_id)
        self.set_instance_and_controller(contact_type)

    def get_simple_contact_controller(self, objectName):
        """
        Returns correspondent controller for the specified company type (objectName)
        """
        from twisted.python.reflect import namedAny
        apps = settings.INSTALLED_APPS
        for app in apps:
            try:
                controllers_module = namedAny(app+'.controllers')
            except AttributeError:
                continue

            if not hasattr(controllers_module, 'contact_controllers'):
                continue
            available_controllers = controllers_module.contact_controllers
            for key in available_controllers:
                if key == objectName:
                    return available_controllers[key]

        return False

    def set_instance_and_controller(self, contact_type):
        c = self.get_simple_contact_controller(contact_type)
        if not c:
            raise SecurityException('Invalid contact type %s' % contact_type)
        self.simple_contact_controller = c(self.sa_session, None, self.get_account().id)
        self._instance_class = self.simple_contact_controller.get_instance_class()

    def get_contacts(self, id = None, contact_list_id = None, priv=Contact.PRIV_VIEW, archived=False, permission='view_contact'):
        """
        @param id: None, list or int
        @return: Contact objects

        Returns a list with all Contacts in case id is None or a list of Contacts with supplied id/ids.
        (a list of one Contact in case a valid id is supplied).
        If group is set, the list returned will be with Contacts on which the specified group(s) has(have) privileges.
        """
        self.assert_permissions(permission)

        from basesoft.web.contrib.contact.sa_datamapper import get_tables
        filter, qfrom = self.get_query_filter_on_privileges(Contact, ContactGroup, priv)

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, Contact.id.in_(id))

        if contact_list_id is not None:
            qfrom = qfrom.outerjoin(get_tables()['tblContactListContact'])
            qfrom = qfrom.join(get_tables()['tblContactList'])
            if not isinstance(contact_list_id, list):
                contact_list_id = [contact_list_id]
            filter = and_(filter, ContactList.id.in_(contact_list_id))

        if archived is None:
            filter = and_(filter, Contact.deleted==False)
        else:
            filter = and_(filter, and_(Contact.deleted==False, Contact.archived==archived))

        query = self.sa_session.query(Contact
                    ).select_from(qfrom).filter(filter)

        return query.all()

    def get_contacts_for_person_and_company(self, person, company):
        filter, qfrom = self.get_query_filter_on_privileges(Contact, ContactGroup, Contact.PRIV_UPDATE)
        filter = and_(filter, Contact.deleted==False, Contact.person==person,
                      Contact.company==company)
        query = self.sa_session.query(Contact).select_from(qfrom).filter(filter)
        return query.all()

    def get_contact_lists(self, id = None, group = None):
        """
        @param id: None, list or int
        @param group: None, list with Group objects or Group object
        @return: ContactList objects

        Returns a list with all ContactLists in case id is None or a list of ContactLists with supplied id/ids.
        (a list of one ContactList in case a valid id is supplied).
        If group is set, the list returned will be with ContactLists on which the specified group(s) has(have) privileges.
        """
        self.assert_permissions('view_contact_list')

        filter, fo = self.get_query_filter_on_privileges(ContactList,
                                                      ContactListGroup,
                                                      ContactList.PRIV_VIEW)

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, ContactList.id.in_(id))

        #add the filter for group(s) on which the ContactLists have privileges
        if group is not None:
            if not isinstance(group, list):
                group = [group]
            g_ids = [g.id for g in group]
            filter = and_(filter,
                          or_(ContactList.group_id.in_(g_ids),
                              ContactListGroup.group_id.in_(g_ids))
                          )

        filter = and_(filter, and_(ContactList.deleted==False, ContactList.archived==False))

        query = self.sa_session.query(ContactList
                    ).select_from(fo).filter(filter)

        return query.all()

    @transaction_decorator
    def change_mail_settings(self, id, contact_list_id, setting):

        assert setting in [ContactListContact.CONTACT_RECEIVES_ALL_MAIL, ContactListContact.CONTACT_RECEIVES_NO_MAIL]

        contact = self.get_contacts(id, None, Contact.PRIV_UPDATE).pop()
        contact_list = self.get_contact_lists(contact_list_id).pop()

        clc = self.sa_session.query(ContactListContact).filter(
            and_(ContactListContact.contact==contact, ContactListContact.contact_list==contact_list)).one()

        clc.mail_settings = setting
        self.sa_session.add(clc)

    @transaction_decorator
    def save_or_update_contact(self, id, person_data, company_data=None, group_id=None, permissions_checks=True,
                               save_contact_privs=True, contact_data=None):
        """
        Save a contact (doesn't attach it to any contact list)
        If id is not None, update the contact with the provided id. If id is None, insert the contact

        @param id: {Int}; id of the contact or None
        @param person_data: {dict}; with
                    first_name: {String}
                    middle_name: {String}
                    last_name: {String}
                    title: {String}
                    gender: {String}
                    email: {String}
                    position: {String}
                    department: {String}
                    phone: {String}
                    mobile_phone: {String}
                    fax: {String}
                    birth_date: {Date}
                    preferred_locale: {Locale}
                    address_data: {Dict} with
                        country: {Country}
                        postcode: {String}
                        street: {String}
                        city: {String}
                        state: {String}

        @param company_data: {dict}; - always None for now
        @param permissions_checks: True/False weather or not to check for permissions and usage options
            (usefull when importing large number of contacts, so we don't have to check permissions for each)

        @return: {Contact}
        """
        if id is None:
            if permissions_checks:
                self.assert_permissions('create_contact', group_id=group_id)
                # check if sufficient credits
                self.assert_credits("max_number_of_contacts")
                # new contact, so remove one credit
                self.substract_usage_credits("max_number_of_contacts", 1)
            contact = self._instance_class()
        else:
            if permissions_checks:
                self.assert_permissions('update_contact')
            contact = self.sa_session.query(self._instance_class).filter(self._instance_class.id==id).one()
            self.assert_privileges_on_object(contact, Contact.PRIV_UPDATE)

        if person_data is not None and len(person_data) > 0:
            if contact.person is None: p_id = None
            else: p_id = contact.person.id
            contact.person = self.simple_contact_controller._set_person(p_id, person_data)
        if company_data is not None and len(company_data) > 0:
            if contact.company is None: c_id = None
            else: c_id = contact.company.id
            contact.company = self.simple_contact_controller._set_company(c_id, company_data)
        if contact_data is not None and len(contact_data) > 0:
            for k in [v['value'] for v in self.simple_contact_controller.get_contact_cols()]:
                if contact_data.has_key(k):
                    setattr(contact, k, contact_data[k])
            if contact_data.has_key('email'):
                contact.email = contact_data['email']

        if id is None and contact.person is not None and contact.person.id is not None and\
        contact.company is not None and contact.company.id is not None:
            # there is already a contact saved for this person and this company, so just return that
            c = self.sa_session.query(Contact).filter(and_(Contact.person_id==contact.person.id,
                                                           Contact.company_id==contact.company.id)).first()
            if c is not None:
                logger.warn('Duplicate contact %s' % c.id)
                # save privs for current group also (this contact might have privs for another group)
                if save_contact_privs:
                    g = self.sa_session.query(Group).filter(Group.id==group_id).first()
                    self.grant_privilege_to(c, Contact.all_privileges, group=True, target_group=g, check_own_privs=False)
                return c

        if id is None and save_contact_privs:
            self.save_default_privilege_object_group(contact, target_group=group_id)

        self.sa_session.add(contact)
        return contact

    def save_or_update_contacts_with_lists(self, contacts, group_id, contact_lists):

        if contact_lists is not None:
            if len(contact_lists) == 0:
                clos = []
            else:
                clos = self.get_contact_lists(id=contact_lists)

            for c in contacts:
                to_add = []
                to_remove = []
                for cl in c.contact_lists:
                    if cl not in clos:
                        to_remove.append(cl)
                for cl in clos:
                    if cl not in c.contact_lists:
                        to_add.append(cl)
                for cl in to_remove:
                    if not self.has_privileges_on_object(cl, ContactList.PRIV_UNLINK_CONTACTS):
                        continue
                    self.unlink_contacts([c], cl)
                for cl in to_add:
                    if not self.has_privileges_on_object(cl, ContactList.PRIV_LINK_CONTACTS):
                        continue
                    self.link_contacts([c], cl)

        return contacts

    def save_or_update_contacts_from_state(self, contacts_state, group_id=None, save_contacts_privs=True, loop_validator=None):
        """
        Saves or updates a list of contacts defined by state.
        This is the method called from rpc having as argument a list of contacts state (dictionary with the contact details).
        This method will call the save_or_update_contact, so all permissions and privileges checks are done in there.

        @param contacts_state: {list} with dictionaries representing a contact state: can contain contact id,
            person details (first_name, email etc), group_id, group_privs, license_privs, domain_privs
        @return: {list} with Contact objects
        """
        if len(contacts_state) == 0:
            return []

        assert isinstance(contacts_state, list)

        for_create = []
        for_update = []

        ret = []
        for c in contacts_state:
            person_data, company_data, contact_data = self._extract_states(c)
            if not contact_data.has_key('contact_object'):
                contacts = self._get_contacts_by_states(person_data, company_data)
                if len(contacts) > 0:
                    # update all found contacts
                    for fc in contacts:
                        if hasattr(self, 'archived_contacts') and fc.archived:
                            c['id'] = fc.id
                            self.archived_contacts.append(c)
                        else:
                            for_update.append((fc.id, person_data, company_data, group_id, False, save_contacts_privs, contact_data))
                else:
                    for_create.append((None, person_data, company_data, group_id, False, save_contacts_privs, contact_data))
            else:
                for_update.append((contact_data['contact_object'].id, person_data, company_data, group_id, False, save_contacts_privs, contact_data))

            # lets the task abort if stop signal was sent
            if loop_validator is not None:
                loop_validator()

        # save_or_update_contact is called with permission_checks False, so we check perms and credits here
        if len(for_create) > 0:
            self.assert_permissions('create_contact', group_id=group_id)
            # check if sufficient credits
            self.assert_credits("max_number_of_contacts", len(for_create))
            # new contact, so remove one credit
            self.substract_usage_credits("max_number_of_contacts", len(for_create))
        if len(for_update) > 0:
            self.assert_permissions('update_contact')

        for s in for_create:
            new_c = self.save_or_update_contact(*s)
            ret.append(new_c)
        for s in for_update:
            new_c = self.save_or_update_contact(*s)
            ret.append(new_c)

        return ret

    def _extract_states(self, state):
        id = None
        contact = None

        # FIXME: this is not correct as state is the contact state and not the person polymorphic type?
        if state.has_key('polymorphic_type') and state['polymorphic_type'] != self._instance_class.person_instance_class.STR:
            raise SecurityException('Persons type %s does not match contact type' % state['polymorphic_type'])

        if state.has_key("id"): #update
            contact = self.get_contacts(id = state["id"], archived=None)[0]
            id = state["id"]

        contact_data = {}
        contact_map = {}
        for k in self.simple_contact_controller.get_contact_cols():
            contact_map[k['value']] = k['value']
        contact_map['contact_email'] = 'email'
        if state.has_key('contact'):
            d = state['contact']
            if not contact and d.has_key('contact_object'):
                contact = d['contact_object']
                id = contact.id
            # we can have an email on Contact level as per ticket #3166
            # we can have a client_last_update on Contact level as per ticket #3169
            for key in d.keys():
                if contact_map.has_key(key):
                    contact_data[contact_map[key]] = d[key] != '' and d[key] or None

        if contact is not None:
            contact_data['contact_object'] = contact

        # the contact has a client last update date. we should convert the value to a Date object
        if contact_data.has_key('client_last_update'):

            # allow native date type (xls/xlsx) or check for user date_format
            if not isinstance(contact_data['client_last_update'], (datetime, date)):
                lcc = LicenseConfigController(self.sa_session, self.get_default_group().license)
                date_format = lcc.get('contact.client_last_update.date_format')
                contact_data['client_last_update'] = datetime.strptime(contact_data['client_last_update'], date_format).date()

        #set the person data for the contact from the contacts_state
        person_data = {}
        for key in state.keys():
            if key not in ['id', 'company', 'address_data', 'polymorphic_type', 'person_object', 'contact_email']:
                person_data[key] = state[key] != '' and state[key] or None
        if state.has_key('person_object'):
            person_data['person_object'] = state['person_object']

        #also include the address info in the person data
        person_data["address_data"] = {}
        for key in ["country", "postcode", "street", "city", "state"]:
            if state.has_key(key): person_data["address_data"][key] = state[key]
            elif id is not None and contact.person.address is not None:
                person_data["address_data"][key] = getattr(contact.person.address, key)
        if len(person_data["address_data"]) == 0:
            del person_data["address_data"]

        # company information
        company_data = {}
        if state.has_key("company"):
            company_data = state['company']

            if company_data.has_key('polymorphic_type') and\
            company_data['polymorphic_type'] != self._instance_class.company_instance_class.STR:
                raise SecurityException('Company type %s does not match contact type' % company_data['polymorphic_type'])

            #also include the address info in the company data
            company_data["address_data"] = {}
            # company address keys can have company_ prefix if state is built when import contact
            for key in ["company_country", "company_postcode", "company_street", "company_city", "company_state",
                        "country", "postcode", "street", "city", "state"]:
                if company_data.has_key(key):
                    company_data["address_data"][key.replace('company_', '')] = company_data[key]
                    del company_data[key]
            if len(company_data["address_data"]) == 0:
                del company_data["address_data"]
            #and bank account data
            company_data["bank_data"] = {}
            for key in ["number", "iban", "bic"]:
                if company_data.has_key(key):
                    company_data["bank_data"][key] = company_data[key]
                    del company_data[key]
            if len(company_data["bank_data"]) == 0:
                del company_data["bank_data"]

        return person_data, company_data, contact_data

    def _get_contacts_by_states(self, person_data, company_data):
        contacts = []
        # check if we need to update in case we have person_object and company_object
        if person_data.has_key('person_object') and company_data.has_key('company_object'):
            contacts = self.get_contacts_for_person_and_company(person_data['person_object'], company_data['company_object'])
        # search on contacts with person_object and None as company, but only in case no company data is provided
        # because in that case a new contact needs to be created.
        elif person_data.has_key('person_object') and len(company_data) == 0:
            contacts = self.get_contacts_for_person_and_company(person_data['person_object'], None)
        # search on contacts with company_object and None as person, but only in case no person data is provided
        # because in that case a new contact needs to be created.
        elif company_data.has_key('company_object') and len(person_data) == 0:
            contacts = self.get_contacts_for_person_and_company(None, company_data['company_object'])
        return contacts

    def validate_state(self, states):
        if type(states) != list:
            states = [states]

        lcc = LicenseConfigController(self.sa_session, self.get_default_group().license)
        format = lcc.get('contact.client_last_update.date_format')

        ret = self.simple_contact_controller.validate_state(states)

        i = -1
        for state in states:
            i = i +1

            if not state.has_key('contact'):
                continue

            s = state['contact']
            if not s.has_key('client_last_update'):
                continue

            # we validate the date format if not already a date(time) instance
            if not isinstance(s['client_last_update'], (datetime, date)):
                try:
                    datetime.strptime(s['client_last_update'], format)
                except ValueError:
                    ret[i]['client_last_update'] = _('Date does not match format %(format)s' % {'format': format})
                    continue

            person_data, company_data, contact_data = self._extract_states(state)
            contacts = self._get_contacts_by_states(person_data, company_data)
            if len(contacts) == 0:
                if contact_data.has_key('contact_object'):
                    contacts = [contact_data['contact_object']]

            for c in contacts:
                if c.client_last_update is not None and c.client_last_update >= contact_data['client_last_update']:
                    ret[i]['client_last_update'] = _('Contact data in database is newer than provided data.')

        return ret

    def load_contact_state(self, c):
        """
        Returns the contacts state
        """
        if not isinstance(c, Contact):
            contact = self.get_contacts(c)[0]
        else:
            contact = c
        if contact.polymorphic_type != self._instance_class.STR:
            self.set_instance_and_controller(contact.polymorphic_type)


        state =  {'dbID': contact.id,
                  'polymorphic_type': contact.polymorphic_type,
                  'contact_email': contact.email,
                  'person': self.simple_contact_controller._load_person_state(contact.person_id)
                 }
        if contact.company_id is not None:
            state['company'] = self.simple_contact_controller._load_company_state(contact.company_id)
        return state

    @transaction_decorator
    def save_or_update_contact_list(self, id, name, group_id=None):
        """
        Save a contact list (no contacts attached to it)
        If id is not None, update the contact list with the provided id. If id is None, insert the contact list

        @param id: {Int}; id of the contact List or None
        @param name: {String}; name of the contact list
        @param group: {Group}; the owner group of the contact list
        @param group_privs: {Int}; group privileges
        @param license_privs: {Int}; license privileges
        @param domain_privs: {Int}; domain privileges

        @return: {ContactList}
        """
        if id is None:
            self.assert_permissions('create_contact_list', group_id=group_id)
            contact_list = ContactList()
        else:
            self.assert_permissions('update_contact_list')
            contact_list = self.sa_session.query(ContactList).filter(ContactList.id==id).one()
            self.assert_privileges_on_object(contact_list, ContactList.PRIV_UPDATE)
        contact_list.name = name

        if id is None:
            self.save_default_privilege_object_group(contact_list, target_group=group_id)

        self.sa_session.add(contact_list)
        return contact_list

    def save_or_update_contact_lists_from_state(self, contact_lists_state, group_id = None):
        """
        Saves or updates a list of contact lists defined by state.
        This is the method called from rpc having as argument a list of contact lists state (dictionary with the contact list details).
        This method will call the save_or_update_contact_list, so all permissions and privileges checks are done in there.

        @param contact_lists_state: {list} with dictionaries representing a contact list state: can contain contact list id,
            contact list name, group_id, group_privs, license_privs, domain_privs
        @param default_group: {Group};
        @return: {list} with ContacList objects
        """
        assert isinstance(contact_lists_state, list)

        ret = []
        for cl in contact_lists_state:
            id = None
            contact_list = None
            if cl.has_key("id") and cl['id'] is not None:
                contact_list = self.get_contact_lists(id = cl["id"])[0]
                id = cl["id"]

            #name is required
            if cl.has_key("name"): name = cl["name"]
            else: name = contact_list.name

            #now call save_or_update_contact_list that will do the checks for permissions and privileges and commit
            new_cl = self.save_or_update_contact_list(id = id, name = name, group_id = group_id)
            ret.append(new_cl)
        return ret

    @transaction_decorator
    def delete_contacts(self, contacts):
        """
        @param contacts: {Contact} or {list} with Contact object

        Delete contact(s).
        All information about the contact (person details, address) will also be deleted
        """
        # TODO: add checks for case when contacts are also survey invitees and return a proper message to the user
        self.assert_permissions('delete_contact')
        if not isinstance(contacts, list):
            contacts = [contacts]
        return self.mark_as_deleted(Contact, [c.id for c in contacts])

    @transaction_decorator
    def link_contacts(self, contacts, contact_list, group_id=None, check_contacts_privs=True):
        """
        Links contacts to contact list
        @param contacts: {List} with Contact objects
        @param contact_list: {ContactList} object or name of the contact list
        @param check_contacts_privs: True/False whether or not to check privs on contacts that will be linked
        """
        if not isinstance(contact_list, ContactList):
            filter, fo = self.get_query_filter_on_privileges(ContactList,
                                                          ContactListGroup,
                                                          ContactList.PRIV_LINK_CONTACTS)
            filter = and_(ContactList.name==contact_list, ContactList.deleted==False, ContactList.archived==False)

            cl = self.sa_session.query(ContactList).select_from(fo).filter(filter).first()
            if cl is None:
                cl = self.save_or_update_contact_list(None, contact_list, group_id)
            contact_list = cl
        else:
            self.assert_privileges_on_object(contact_list, contact_list.PRIV_LINK_CONTACTS)

        assert isinstance(contacts, list)
        self.assert_permissions('link_to_contact_list')
        if len(contacts) == 0:
            return contact_list
        for c in contacts:
            assert isinstance(c, Contact)
            if check_contacts_privs:
                #need at least the view priv on each contact that will be linked
                self.assert_privileges_on_object(c, c.PRIV_VIEW)
        #only add the ones that are not already linked to the contact list
        distinct_contacts = contact_list.contacts
        for c in contacts:
            try:
                distinct_contacts.index(c)
            except ValueError:
                distinct_contacts.append(c)

        contact_list.contacts = distinct_contacts
        self.sa_session.add(contact_list)
        return contact_list

    @transaction_decorator
    def unlink_contacts(self, contacts, contact_list):
        """
        Unlinks contacts from contact list
        @param contacts: {List} with Contact objects
        @param contact_list: {ContactList} object
        """
        assert isinstance(contact_list, ContactList)
        assert isinstance(contacts, list)
        self.assert_permissions('unlink_from_contact_list')
        self.assert_privileges_on_object(contact_list, contact_list.PRIV_UNLINK_CONTACTS)
        if len(contacts) > 0:
            for c in contacts:
                assert isinstance(c, Contact)
                #need at least the view priv on each contact that will be unlinked
                self.assert_privileges_on_object(c, c.PRIV_VIEW)
            new_contacts = []
            for con in contact_list.contacts:
                if con not in contacts:
                    new_contacts.append(con)
            contact_list.contacts = new_contacts
            self.sa_session.add(contact_list)
            self.sa_session.flush()

    @transaction_decorator
    def delete_contact_lists(self, contact_lists, delete_contacts_option = CONTACTS_DO_NOT_DELETE):
        """
        @param contact_list: {ContactList} or {list} with ContactList objects
        @param delete_contacts_option: {int} specifies what to do with the contacts linked to the contact lists to be deleted

        This deletes the contact lists and depending on the second argument, the contacts linked to them.
        """
        self.assert_permissions('delete_contact_list')
        delete_contacts_option = int(delete_contacts_option)
        assert delete_contacts_option in [CONTACTS_DO_NOT_DELETE , CONTACTS_DELETE_ONLY_EXCLUSIVE, CONTACTS_DELETE_ALL]
        if not isinstance(contact_lists, list):
            contact_lists = [contact_lists]

        #remove the contacts (depending on the option of course)
        self._remove_contact_lists_contacts(contact_lists, delete_contacts_option)
        #remove the contact lists
        return self.mark_as_deleted(ContactList, [c.id for c in contact_lists])

    def _remove_contact_lists_contacts(self, contact_lists, contacts_option):
        """
        Potected method for removing the contacts linked to a list of contact lists
        This called when removing a contact list;
        """
        assert isinstance(contact_lists, list)
        if contacts_option == CONTACTS_DO_NOT_DELETE:
            return []
        # list with distinct Contact objects that are linked to the contact_lists
        contacts = []
        cl_ids = [cl.id for cl in contact_lists]
        for cl in contact_lists:
            for c in cl.contacts:
                # it is already marked as deleted
                if c.deleted:
                    continue
                if contacts_option == CONTACTS_DELETE_ALL:
                    try:
                        contacts.index(c)
                    except ValueError: contacts.append(c)
                elif contacts_option == CONTACTS_DELETE_ONLY_EXCLUSIVE:
                    ccls = [cc.id for cc in c.contact_lists]
                    delete = True
                    for ccl_id in ccls:
                        try:
                            # if in the contact's contact lists there is one that is not
                            # in the contact lists that will be removed, then do not delete the contact
                            cl_ids.index(ccl_id)
                        except ValueError:
                            delete = False
                    if delete:
                        try: contacts.index(c)
                        except ValueError: contacts.append(c)
            self.sa_session.expire(cl)
        if len(contacts) > 0:
            self.delete_contacts(contacts = contacts)

    def check_contact_lists_email(self, contact_lists):
        """
        Checks if there is at least one contact that has email address set on either person or contact.
        This check is done every time a mailing is saved.
        """
        if isinstance(contact_lists, ContactList):
            contact_lists = [contact_lists]
        assert isinstance(contact_lists, list)

        no_email_cls = []
        for contact_list in contact_lists:
            contacts = contact_list.contacts
            has = False
            for contact in contacts:
                if contact.email is not None or (contact.person is not None and contact.person.email is not None and contact.person.email != ""):
                    has = True
                    break
            if not has:
                no_email_cls.append(contact_list)
        if len(no_email_cls) > 0:
            message = _("The following contact lists have no contacts with an email address set")
            for cl in no_email_cls:
                message += ' "%s";' % cl.name
            raise PublicUserException(message)
        else:
            return True

    @transaction_decorator
    def duplicate_contact_list(self, contact_list):
        new = contact_list.clone(reference_children=['contacts'])
        self.sa_session.add(new)
        return new

    @transaction_decorator
    def merge(self, id_keep, ids, what):
        """
        id_keep: id of the object that will be kept
        ids: all selected object ids for merge
        what: company or person
        """
        assert what in ['company', 'person']
        self.assert_permissions('update_contact')

        # get the object we need to keep
        if what == 'company':
            obj = self.sa_session.query(Company).filter(Company.id==id_keep).all()
        else:
            obj = self.sa_session.query(Person).filter(Person.id==id_keep).all()
        if len(obj) == 0:
            raise SecurityException('Invalid %s id %s' % (what, id_keep))
        obj = obj[0]

        filter, qfrom = self.get_query_filter_on_privileges(Contact, ContactGroup, Contact.PRIV_UPDATE)
        if what == 'company':
            filter = and_(filter, Contact.company_id.in_(ids))
        else:
            filter = and_(filter, Contact.person_id.in_(ids))
        contacts = self.sa_session.query(Contact).select_from(qfrom).filter(filter).all()

        # except for the obj we need to keep, all rest need to be removed and contacts updated
        to_del = []
        for contact in contacts:
            o = getattr(contact, what)
            if o.id == obj.id:
                continue
            to_del.append(o)
            setattr(contact, what, obj)
            self.sa_session.add(contact)

        for d in to_del:
            self.sa_session.delete(d)

    def merge_contact_list_contacts(self, cl1, cl2, unique_identifiers):
        """
        This merges the companies/person from cl1 and cl2 based on the supplied unique identifiers.
        Contacts remain untouched for the moment (due to links with different kind of data).
        """
        person_map = {}
        company_map = {}
        if len(unique_identifiers) == 0:
            raise PublicUserException(_('Merge requires at least one unique field'))

        # first we make a map of the unique fields from contact list 1
        for contact in cl1.contacts:
            pd = ''
            cd = ''
            for id in unique_identifiers:
                if id['target'] == self.COLUMN_IS_PERSON_INFO:
                    pd += '%s:%s-' % (id['column'], getattr(contact.person, id['column']))
                elif id['target'] == self.COLUMN_IS_COMPANY_INFO:
                    cd += '%s:%s-' % (id['column'], getattr(contact.company, id['column']))
            if pd != '':
                person_map[pd] = contact.person.id
            if cd != '':
                company_map[cd] = contact.company.id

        merged = 0

        for contact in cl2.contacts:
            pd = ''
            cd = ''
            m = False
            for id in unique_identifiers:
                if id['target'] == self.COLUMN_IS_PERSON_INFO:
                    pd += '%s:%s-' % (id['column'], getattr(contact.person, id['column']))
                elif id['target'] == self.COLUMN_IS_COMPANY_INFO:
                    cd += '%s:%s-' % (id['column'], getattr(contact.company, id['column']))
            if person_map.has_key(pd):
                self.merge(person_map[pd], [contact.person.id], 'person')
                m = True
            if company_map.has_key(cd):
                self.merge(company_map[cd], [contact.company.id], 'company')
                m = True
            if m:
                merged += 1

        return merged

    def get_field_mapping_state(self, fm):
        return {
            'id': fm.id,
            'name': fm.name,
            'contact_type': fm.contact_type,
            'mapping': fm.mapping,
            'unique_fields': fm.unique_fields
        }

    def get_field_mappings(self, id=None, priv=ContactFieldMapping.PRIV_VIEW):
        rv = []
        for fm in self.get_privileged_object_by_id(ContactFieldMapping, id=id, priv=priv):

            # make sure the mapping and the possible unique field settings are deserialized.
            fm.mapping = simplejson.loads(fm.mapping, object_hook=json_object_hook)
            if fm.unique_fields is not None:
                fm.unique_fields = simplejson.loads(fm.unique_fields, object_hook=json_object_hook)

            rv.append(fm)

        return rv

    def get_field_mapping_states(self):
        rv = {}
        for fm in self.get_field_mappings():
            p = self.get_field_mapping_state(fm)
            rv[p['id']] = p
        return rv

    @transaction_decorator
    def save_or_update_field_mapping(self, id, name, contact_type, mapping, unique_fields=None):
        """
        These are presets of mappings between file headers and datbase fields which can be used when importing contacts.
        """

        self.assert_permissions('manage_field_mapping_presets')

        if contact_type not in [t['value'] for t in self.get_contact_types()]:
            raise SecurityException('No access to contact type [%s]' % contact_type)

        if id is not None:
            fm = self.get_field_mappings(id, priv=ContactFieldMapping.PRIV_UPDATE).pop()
        else:
            fm = ContactFieldMapping()

        # make sure any fields selected as unique fields are also selected as mapping
        # raise PublicUserException in case they didn't
        # TODO: this needs to be fixed correctly with ticket #3717
        if unique_fields is not None:
            assertable = ['%s+%s' % (i['target'].keys()[0], i['target'].values()[0]) for i in mapping]
            for i in unique_fields:
                if i not in assertable:
                    raise PublicUserException(_('The unique field %(field)s was not found in the mapping.' % {'field': i}))

            fm.unique_fields = simplejson.dumps(unique_fields, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        fm.name = name
        fm.contact_type = contact_type

        fm.mapping = simplejson.dumps(mapping, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        # attach default permissions in case this is a new preset.
        if id is None:
            self.save_default_privilege_object_group(fm)

        self.sa_session.add(fm)

        return fm

    def get_contact_types(self):
        rv = []
        for m in get_models_of_type(Contact, False, None, True):
            if m['model'].PERMISSIONS is not None and not self.has_permissions(*m['model'].PERMISSIONS):
                continue
            rv.append(m)
        return rv

class ImportContactController(ContactController):
    """
    Controller for importing contacts from file
    """

    # actions to take on import for unique identifiers
    ACTION_INSERT = 1 # insert as new
    ACTION_UPDATE = 2 # update if found
    ACTION_DELETE = 3 # remove old one and insert new

    # if number of records in the file exceeds this number, then import will be done in the background by a task
    MAX_RECORDS_FOR_AUTOMATICAL_IMPORT = 100

    # separator options (csv only)
    SEPARATOR_OPTIONS = [
        {"label": ugettext_lazy("Comma"), "value": ","},
        {"label": ugettext_lazy("Tab"), "value": "\t"},
        {"label": ugettext_lazy("Semi-colon"), "value": ";"},
    ]

    QUOTECHAR_OPTIONS = ['"', "'", '|', '']

    # weather or not to ask for which group to save for in case an account is linked to multiple groups
    privilege_save_handler = True

    def __init__(self, file_path, file_name=None, sa_session=None, account_id=None, contact_type='Contact'):
        """
        @param file_path: {String} - absolute file path
        """
        ContactController.__init__(self, sa_session, account_id, contact_type)
        self.assert_permissions('import_contacts')

        import os.path

        if file_path is not None:
            assert os.path.exists(file_path)

        if file_name is not None:
            base_filename, file_extension = os.path.splitext(file_name)

            if file_extension in ('.csv', '.txt'):
                self._importer = CsvImporter(file_path)

            elif file_extension == '.xlsx':
                self._importer = XlsxImporter(file_path)

            elif file_extension == '.xls':
                self._importer = XlsImporter(file_path)

            else:
                raise PublicUserException(_('Import type not supported: %s') % file_extension)

        # array of dictionaries with keys
        # "file_header", "column", "target"

        # e.g.:
        #  self.field_mappings = [
        #    {"file_header:"first name", ""column": "first_name", "target": COLUMN_IS_PERSON_INFO},
        #    {"file_header:"email address", "column": "email", "target": COLUMN_IS_PERSON_INFO}
        #  ]

        self.field_mappings = None

        # if true, first line of file will be imported
        self.import_row1 = False

        self.company_type = "Company"

        # dictionary with field plugins
        self.__field_plugins = {}


    def get_rows(self):
        return self._importer.get_rows()

    def set_file_lines(self, *args, **kwargs):
        self._importer.set_file_lines(*args, **kwargs)

    @transaction_decorator
    def import_contacts(self, updated_rows=None, excluded_rows=None, unique_fields=[],
                        action=ACTION_INSERT, group_id=None, save_contacts_privs=True, loop_validator=None):
        """
        Method for importing contacts from a csv file
        @param updated_rows: {Array} with dictionaries containing
            row indexes and new new contact state (column/value pairs)
            E.g.: [
                    { "row_index": 10,
                      "row_values": {"first_name": "john",
                                     "last_name": "smith",
                                     "email": "john@mail-test.basesoft.com",
                                     "phone": ""}
                    },
                    {"row_index": 15,
                      "row_values": {"first_name": "anne",
                                     "last_name": "franklin",
                                     "email": "anne@mail-test.basesoft.com",
                                     "phone":"123"}
                    }
                   ]
            Length of "values" must be the same as the length of the headers
        @param excluded_rows: {Array} with row indexes; rows that will not be
            imported in the db
        @param unique_fields: {Array} of dictionaries with target and column keys. Indicates the unique fields
            in order for the import to do an insert or an update of a record.
        @return: list with the imported Contact objects
        """
        assert self.field_mappings is not None

        # make sure all row indexes are ints
        if excluded_rows is not None:
            excluded_rows = [int(e) for e in excluded_rows]
        else: excluded_rows = []

        # set the unique fields per target (person or company):
        uf_per_target = {self.COLUMN_IS_PERSON_INFO: [], self.COLUMN_IS_COMPANY_INFO: [], self.COLUMN_IS_CONTACT: []}
        for uf in unique_fields:
            t = int(uf['target'])
            uf_per_target[t].append(uf['column'])

        # get int row indexes for rows that will be updated
        if updated_rows is not None:
            updated_rows_indexes = [int(ur["row_index"]) for ur in updated_rows]
            # set the updated rows contact state
            updated_rows_values = {}
            for ur in updated_rows:
                ri = int(ur["row_index"])
                assert ri >= 0 and ri <= len(self._importer.get_rows())
                person_data = {}
                for k,v in ur["row_values"].items():
                    person_data[k] = v
                person_data = self.__set_country_object(person_data)
                updated_rows_values[ri] = self.simple_contact_controller.set_person_for_import(
                                                person_data, uf_per_target[self.COLUMN_IS_PERSON_INFO])
                if updated_rows_values[ri].has_key("company"):
                    updated_rows_values[ri]["company"] = self.__set_country_object(updated_rows_values[ri]["company"],
                                                                                   "company_")
                    updated_rows_values[ri]['company'] = self.simple_contact_controller.set_company_for_import(
                                                            updated_rows_values[ri]['company'], uf_per_target[self.COLUMN_IS_COMPANY_INFO])
        else: updated_rows_indexes = []

        values = [v for v in self._importer.get_rows()]
        if len(values) == 0:
            return []
        #import or not the first line of the file
        if not self.import_row1:
            if len(values) == 1:
                return []
            values.pop(0)
        for index in updated_rows_indexes:
            values[int(index)] = updated_rows_values[index]
        if len(excluded_rows) > 0:
            vals = [values[i] for i in range(0, len(values)) if i not in excluded_rows]
        else:
            vals = [v for v in values]

        has_cl = False
        default_contact_list_name = None
        for f in self.field_mappings:
            if int(f["target"]) == self.COLUMN_IS_CONTACT_LIST or int(f["target"]) == self.COLUMN_IS_CONTACT_LIST_CONTACT:
                has_cl = True
                break
            lcc = LicenseConfigController(self.sa_session, scope=self.get_default_group().license)
            # Check if there is a default contact list set.
            contact_list_id = lcc.get('contact.import.default.contact.list.id')
            if contact_list_id != 0:
                default_contact_list_name = self.get_contact_lists(contact_list_id).pop().name
                has_cl = True


        # if no unique fields were set and no contact list column, we do the save in one go,
        # because it's faster (tnt case)
        if len(unique_fields) == 0 and not has_cl:
            contact_states = self.__create_contact_states(vals, uf_per_target)
            return self.save_or_update_contacts_from_state(contact_states, group_id, save_contacts_privs, loop_validator=loop_validator)

        # if unique fields were set, some contacts that have been archived might be found; don't update those
        self.archived_contacts = []
        self.archived_contact_states = []

        # else we do it one be one since it needs to check the db for unique fields and/or add them to a contact list
        contact_lists = {}
        contacts = []
        settings = {}
        for v in vals:
            cl_name = None

            # create the list with contacts state and then call the method to save them
            contact_state = self.__create_contact_states([v], uf_per_target)[0]

            if contact_state.has_key('contact_list_name') and contact_state['contact_list_name'] != '':
                cl_name = contact_state['contact_list_name']
                del contact_state['contact_list_name']

            mail_setting = 0
            if contact_state.has_key('mail_settings') and contact_state['mail_settings'] != '':
                try:
                    mail_setting = int(contact_state['mail_settings'])
                except ValueError:
                    mail_setting = 0

            before = len(self.archived_contacts)
            c = self.save_or_update_contacts_from_state([contact_state], group_id, save_contacts_privs, loop_validator=loop_validator)

            # should be just one contact
            if len(c) > 0:
                contacts.append(c[0])
            else:
                if len(self.archived_contacts) > before:
                    self.archived_contact_states.append(self.__create_contact_states([v], uf_per_target, for_validation=True)[0])
                continue

            if cl_name is None and default_contact_list_name is not None:
                cl_name = default_contact_list_name
            if cl_name is not None:
                if not contact_lists.has_key(cl_name):
                    contact_lists[cl_name] = []
                    settings[cl_name] = []
                contact_lists[cl_name].append(c[0])
                if mail_setting != 0:
                    settings[cl_name].append(c[0])

        if save_contacts_privs:
            for cl_name, cs in contact_lists.items():
                cl = self.link_contacts(cs, cl_name, group_id, False)
                for clc in cl.contact_list_contacts:
                    if clc.contact not in settings[cl_name]:
                        continue
                    clc.mail_settings = 1
                    self.sa_session.add(clc)
                self.sa_session.flush()

        return contacts

    def save_contacts_privs(self, contact_list_id, group_id=None):
        """
        This method should be called only if import was made without saving contact privs, imported contacts
        have been linked to a contact list and no unique fields were set.
        """
        if group_id is None:
            group_id = self._get_default_group().id


        sql = """
        INSERT INTO tblContactGroup (contact_id, group_id, privilege, group_has_privilege, created, last_updated)
        SELECT contact_id, %s, %s, 1, created, last_updated FROM tblContactListContact WHERE contact_list_id = %s
        ON DUPLICATE KEY UPDATE group_has_privilege = 1
        """

        for priv in Contact.all_privileges:
            self.sa_session.execute(text(sql % (group_id, priv, contact_list_id) ))

    def __create_contact_states(self, values, unique_fields=None, for_validation=False):
        """
        Returns a dictionary with contact data values
        (column name/value pairs)
        """
        if not isinstance(values, list):
            values = [values]

        if unique_fields is None:
            unique_fields = {}

        # make sure that if this is called for validation, we don't check unique_fields
        #if for_validation:
        #    unique_fields = {}

        states = []
        for v in values:

            if isinstance(v, dict): # this is an updated_row, it's state is already created
                states.append(v)
                continue

            #TODO: make sure that the values in field_mappings are in the same order as in the file

            state = {}
            company_data = {}
            person_data = {}
            contact_data = {}
            for i in range(len(self.field_mappings)):
                column = self.field_mappings[i]
                if int(column["target"]) == self.COLUMN_IS_NOT_IMPORTED:
                    continue
                val = ''
                if len(v) >= i+1:
                    val = v[i]
                if int(column['target']) in [self.COLUMN_IS_PERSON_INFO, self.COLUMN_IS_CONTACT_LIST, self.COLUMN_IS_CONTACT_LIST_CONTACT]:
                    person_data[column["column"]] = val
                elif int(column['target']) == self.COLUMN_IS_COMPANY_INFO:
                    company_data[column["column"]] = val
                elif int(column['target'])  == self.COLUMN_IS_CONTACT:
                    contact_data[column['column']] = val

            if for_validation:
                state = self.simple_contact_controller.set_import_person_state(person_data)
            else:
                person_data = self.__set_country_object(person_data)

                person_uf = []
                if unique_fields.has_key(self.COLUMN_IS_PERSON_INFO):
                    person_uf = unique_fields[self.COLUMN_IS_PERSON_INFO]

                state = self.simple_contact_controller.set_person_for_import(person_data, person_uf)

            if len(company_data) > 0:
                if for_validation:
                    state['company'] = self.simple_contact_controller.set_import_company_state(company_data)
                else:
                    company_data = self.__set_country_object(company_data, "company_")
                    company_uf = []
                    if unique_fields.has_key(self.COLUMN_IS_COMPANY_INFO):
                        company_uf = unique_fields[self.COLUMN_IS_COMPANY_INFO]
                    state['company'] = self.simple_contact_controller.set_company_for_import(company_data, company_uf)

            if len(contact_data) > 0:
                # TODO validation stuff (contact last update?)
                if for_validation:
                    state['contact'] = contact_data
                else:
                    contact_uf = []
                    if unique_fields.has_key(self.COLUMN_IS_CONTACT):
                        contact_uf = unique_fields[self.COLUMN_IS_CONTACT]
                    state['contact'] = self.simple_contact_controller.set_contact_for_import(contact_data, contact_uf)

            states.append(state)

        return states

    def __set_country_object(self, state, key_prefix=""):
        """
        For all country columns, creates the country object
        """
        country = None
        if state.has_key(key_prefix+"country_name"):
            country = self.sa_session.query(Country).filter(Country.name==state[key_prefix+"country_name"]).first()
            if country is None:
                country = self.sa_session.query(DomainCountry).filter(and_(DomainCountry.name==state[key_prefix+"country_name"],
                    DomainCountry.domain==self.get_default_group().license.domain, DomainCountry.active==1)).first()
                if country is not None:
                    country = country.country
            del state[key_prefix+"country_name"]
        if state.has_key(key_prefix+"country_iso_3166_a2"):
            country = self.sa_session.query(Country).filter(Country.iso_3166_a2==state[key_prefix+"country_iso_3166_a2"]).first()
            del state[key_prefix+"country_iso_3166_a2"]
        if state.has_key(key_prefix+"country_iso_3166_a3"):
            country = self.sa_session.query(Country).filter(Country.iso_3166_a3==state[key_prefix+"country_iso_3166_a3"]).first()
            del state[key_prefix+"country_iso_3166_a3"]
        if country is not None:
            state["country"] = country
        return state

    def get_row_values(self, row_index=0):
        """
        @param row_index: {Int} row index in the file
        """
        return self._importer.get_rows()[row_index]

    def set_fields_plugins(self, field_plugins=None):
        """
        Sets the validation plugins for each column
        """
        if field_plugins is None:
            self.__field_plugins , self.__group_plugins = self.simple_contact_controller.get_field_plugins()
            return
        assert isinstance(field_plugins, dict)
        self.__field_plugins = field_plugins

    def validate(self, updated_row=None, loop_validator=None, unique_fields=None):
        """
        Method for validating row values.
        @param row_values: {dictionary} with row_index and new values
        All file will be validated in case row_values is None
        @return: {Array} with dictionaries with
            row index, row values, columns and validation error messages
        E.g.: [{ "row_index": 10,
                 "row_values": {"first_name": "john",
                                "middle_name": "",
                                "last_name": "smith",
                                "email": "john@email",
                                "phone": "xx"},
                 errors: {"email": "Invalid email address",
                          "phone": "Invalid phone number"
                          }
               }
              ]
        """

        if len(self.__field_plugins) == 0:
            self.set_fields_plugins()
        if unique_fields is None:
            unique_fields = {}

        uf_per_target = {self.COLUMN_IS_PERSON_INFO: [], self.COLUMN_IS_COMPANY_INFO: [], self.COLUMN_IS_CONTACT: []}
        for uf in unique_fields:
            t = int(uf['target'])
            uf_per_target[t].append(uf['column'])

        validator = FormValidator(self.__field_plugins, self.__group_plugins)

        if updated_row is not None:
            assert updated_row["row_index"] >= 0 and updated_row["row_index"] < len(self._importer.get_rows())
            validation_result = validator.validate(updated_row["row_values"])
            #contact specific validation
            vs = self.validate_state(updated_row["row_values"])[0]
            for key in vs:
                validation_result[key] = vs[key]
            if len(validation_result) > 0:
                return [{"row_index": updated_row["row_index"],
                        "row_values": updated_row["row_values"],
                        "errors": validation_result}]
            return []

        #validate or not the first line of the file
        values = [v for v in self._importer.get_rows()]
        if len(values) == 0:
            return []
        if not self.import_row1:
            if len(values) == 1:
                return []
            values.pop(0)

        contact_states = self.__create_contact_states(values, None, for_validation=True)
        validate_state = self.__create_contact_states(values, uf_per_target, for_validation=False)
        # contact specific validation
        contact_validation = self.validate_state(validate_state)

        index = 0
        results = []
        for state in contact_states:

            if state.has_key('company'):
                for key in state['company']:
                    if key != 'polymorphic_type':
                        state[key] = state['company'][key]

            validation_result = validator.validate(state)
            if len(contact_validation) >= index+1:
                vs = contact_validation[index]
                for key in vs:
                    validation_result[key] = vs[key]

            if state.has_key('mail_settings') and state['mail_settings'] != '':
                try:
                    int(state['mail_settings'])
                except ValueError:
                    validation_result['mail_settings'] = 'Incorrect value for mail settings'

            #del state['company']
            #del state['polymorphic_type']
            if len(validation_result) > 0:
                results.append({"row_index": index,
                                "row_values": state,
                                "errors": validation_result})
            index = index + 1

            # lets the task abort if stop signal was sent
            if loop_validator is not None:
                loop_validator()

        return results


    def save_file(self, file, import_settings, props=None):
        """
        Save Import file to the db
        """
        from basesoft.web.contrib.file.controllers import ProcessFileController

        q = self.sa_session.query(ImportFileType).filter(ImportFileType.name=='basesoft.web.contrib.contact').all()
        if len(q) == 0:
            raise SystemException('Cannot find file import type "basesoft.web.contrib.contact".')
        type = q[0]

        fc = ProcessFileController(sa_session=self.sa_session, account_id=self.get_account().id)
        settings_str = simplejson.dumps(import_settings, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        fc.file_type_properties = {'settings': settings_str, 'type': type}
        if props is not None:
            for k,v in props.items():
                fc.file_type_properties[k] = v

        category = fc.get_category_by_label("Contact Import files")
        if category is None:
            category = fc.save_or_update_file_category(None, "Contact Import files")

        return fc.save_files_from_temp(file, category, ImportFile)[0]

    def get_available_columns(self):
        """
        Get the available tables columns for the field mappings
        """
        address_cols = ["postcode", "street", "city", "state"]
        country_cols = ["name", "iso_3166_a2", "iso_3166_a3"]
        bka_cols = ["number", "iban", "bic"]

        person = self.simple_contact_controller.get_person_cols()
        company = self.simple_contact_controller.get_company_cols()
        contact = self.simple_contact_controller.get_contact_cols()

        for col in bka_cols:
            company.append({"value": col, "label": BankAccount.cols_i18n[col]})
        for col in address_cols:
            person.append({"value": col, "label": Address.cols_i18n[col]})
            company.append({"value": "company_%s" % col, "label": Address.cols_i18n[col]})
        for col in country_cols:
            person.append({"value": "country_%s" % col, "label": Country.cols_i18n[col]})
            company.append({"value": "company_country_%s" % col, "label": Country.cols_i18n[col]})

        columns = [
            { "id": self.COLUMN_IS_NOT_IMPORTED,
              "label":_("Do not import"),
              "options": [{"value": "do_not_import", "label": _("Do not import")}],
            },
            { "id": self.COLUMN_IS_PERSON_INFO,
              "label":_("Person information"),
              "options": person,
            },
            { "id": self.COLUMN_IS_COMPANY_INFO,
              "label":_("Company information"),
              "options": company,
            },
            { "id": self.COLUMN_IS_CONTACT_LIST,
              "label":_("Contact list"),
              "options": [{"value": "contact_list_name", "label": _('Name')}],
            },
            { "id": self.COLUMN_IS_CONTACT,
              "label":_('Contact'),
              "options": contact
            },
            { "id": self.COLUMN_IS_CONTACT_LIST_CONTACT,
              "label":_('Contact settings'),
              "options": [{'value': "mail_settings", "label": _('Email setting')}]
            }
        ]

        return columns

    def get_required_columns(self):
        """
        Get the required tables columns for the field mappings
        """
        company = self.simple_contact_controller.company_required_columns()
        person = self.simple_contact_controller.person_required_columns()

        # need to check and return either list of lists or list of strings (columns names)
        # not mixed

        if len(company) == 0:
            return person
        if len(person) == 0:
            return company

        if isinstance(company[0], list) and isinstance(person[0], list):
            return person + company

        if isinstance(company[0], list):
            return [person] + company

        if isinstance(person[0], list):
            return person + [company]

        return person + company

    def get_actions(self):
        """
        Return dictionary with possible actions for unique identifiers
        """
        return [
            {"label": _("Insert new"), "value": self.ACTION_INSERT},
            {"label": _("Update"), "value": self.ACTION_UPDATE},
            {"label": _("Delete"), "value": self.ACTION_DELETE}
        ]

    def export_csv(self, data, contacts_filename, tmp=True):
        """
        Export invalid rows as csv
        @param data: validation state as sent to the client
        """
        if len(data) == 0:
            return None

        columns = {}
        for m in self.field_mappings:
            if int(m["target"]) == self.COLUMN_IS_NOT_IMPORTED:
                continue
            columns[m['column']] = m['file_header']
        # first line with headers
        headers = []
        for column in data[0]['row_values']:
            if columns.has_key(column):
                headers.append(columns[column].encode('utf-8'))
        # at the end add a column with row's error messages
        if data[0].has_key('errors'):
            headers.append(_('ERRORS'))

        def add_error_message(error_str, col, message):
            if error_str == '':
                error_str = '%s: ' % col
            else:
                error_str = '%s; %s: ' % (error_str, col)
            if isinstance(message, list):
                error_str = '%s [' % error_str
                for m in message:
                    if isinstance(m, list):
                        m = ', '.join(m)
                    error_str = '%s %s, ' % (error_str, m)
                error_str = '%s]' % error_str[0:-2]
            else:
                error_str = '%s %s' % (error_str, message)
            return error_str

        lines = [headers]
        for d in data:
            line = []
            for column, val in d['row_values'].items():
                if columns.has_key(column):
                    line.append(val.encode('utf-8'))
            if d.has_key('errors'):
                error_str = ''
                for column, mess in d['errors'].items():
                    if columns.has_key(column):
                        error_str = add_error_message(error_str, columns[column], mess)
                    else: # general message (not specific to a column)
                        error_str = add_error_message(error_str, _('General'), mess)
                line.append(error_str.encode('utf-8'))
            lines.append(line)

        from cStringIO import StringIO
        import csv
        from basesoft.web.contrib.file.controllers import ProcessFileController

        stream = StringIO()
        if self._importer.separator == "\t":
            validationCSV = csv.writer(stream, dialect='excel-tab', quoting=csv.QUOTE_ALL, quotechar=self._importer.quotechar)
        else:
            validationCSV = csv.writer(stream, delimiter=self._importer.separator, quoting=csv.QUOTE_ALL, quotechar=self._importer.quotechar)
        validationCSV.writerows(lines)

        name = '.'.join(contacts_filename.split('.')[:-1])
        extension = contacts_filename.split('.').pop()
        filename = "%s_%s_%s.%s" % (name, 'validation', datetime.now().strftime("%Y%m%d_%H%M"), extension)

        fc = ProcessFileController(self.sa_session, account_id=self.get_account())
        if tmp:
            file = fc.save_or_update_tmp_file_from_content(filename, stream.getvalue())
        else: # when import is done in the background, the file will not be saved as temp
            category = fc.get_category_by_label(u'Contact Import Validation')
            if category is None:
                category = fc.save_or_update_file_category(None, u'Contact Import Validation')
            file = fc.save_or_update_file_from_content(None, filename, None, category, stream.getvalue())

        stream.close()
        return file

class BackgroundImportContactController(AbstractController):

    @transaction_decorator
    def import_contact_file(self, file, loop_validator=None):
        """
        Import contacts from the specified contact import file using correspondent settings
        """
        assert isinstance(file, ImportFile)

        import_settings = simplejson.loads(file.settings, object_hook=json_object_hook)

        controller = self.get_import_contact_controller(import_settings['contact_type'])
        if not controller:
            controller = ImportContactController

        import_controller = controller(file_path=os.path.join(settings.FILE_REPOSITORY, file.file_hash),
                                     file_name = file.file_name,
                                     sa_session=self.sa_session, account_id=import_settings['account_id'],
                                     contact_type=import_settings['contact_type'])
        import_controller.encoding = import_settings['encoding']
        import_controller.set_file_lines(import_settings['separator'])
        import_controller.field_mappings = import_settings['field_mappings']
        import_controller.import_row1 = import_settings['import_row1']

        group_id = None
        if import_settings.has_key('group_id'):
            group_id = import_settings['group_id']

        validation_file = None
        validation_data = []
        if import_settings.has_key('do_validation') and import_settings['do_validation']:
            updated_rows = None
            validation_data = import_controller.validate(loop_validator=loop_validator)
            if len(validation_data) == 0:
                excluded_rows = None
            else:
                excluded_rows = [d['row_index'] for d in validation_data]
                # validation file will be exported after, because archived contacts might be found and they need
                # to be included in the file
        else:
            assert import_settings.has_key('updated_rows')
            assert import_settings.has_key('excluded_rows')
            updated_rows = import_settings['updated_rows']
            excluded_rows = import_settings['excluded_rows']
        assert import_settings.has_key('unique_fields')
        unique_fields = import_settings['unique_fields']

        contact_list = None
        save_privs = True # save contacts privs on import, per contact or all at the end

        has_cl = False
        for f in import_controller.field_mappings:
            if int(f["target"]) == import_controller.COLUMN_IS_CONTACT_LIST:
                # already has contact list set in the file
                has_cl = True
                break

        if not has_cl and import_settings.has_key('contact_list') and import_settings['contact_list'] is not None:
            cl = import_settings['contact_list']
            contact_controller = ContactController(self.sa_session, account_id=import_settings['account_id'],
                                                   contact_type=import_settings['contact_type'])
            if cl.has_key('id') and cl['id'] is not None:
                contact_list = contact_controller.get_contact_lists(id=cl['id'])[0]
            else:
                contact_list = contact_controller.save_or_update_contact_lists_from_state([cl], group_id)[0]
                # we need to flush after each contact in order to check unique fields
                if len(unique_fields) == 0:
                    save_privs = False
        contacts = import_controller.import_contacts(updated_rows, excluded_rows, unique_fields, group_id=group_id,
                                                     save_contacts_privs=save_privs, loop_validator=loop_validator)
        # check if there were any archived contacts found to be updated;
        if hasattr(import_controller, 'archived_contact_states') and len(import_controller.archived_contact_states) > 0:
            # those ones were not updated, they need to go in the validation file
            vd = []
            for ac in import_controller.archived_contact_states:
                if ac.has_key('company'):
                    for key in ac['company']:
                        if key != 'polymorphic_type':
                            ac[key] = ac['company'][key]
                del ac['company']
                del ac['polymorphic_type']
                vd.append({"row_index": None,
                           "row_values": ac,
                           "errors": {'': _('Archived contact found for this data')}})
            validation_data = validation_data + vd

        if len(validation_data) > 0:
            validation_file = import_controller.export_csv(validation_data, file.file_name, False)

        # link to contact list if one was provided
        if contact_list is not None and len(contacts) > 0:
            contact_controller.link_contacts(contacts, contact_list, group_id=group_id, check_contacts_privs=save_privs)
            if not save_privs:
                self.sa_session.flush()
                import_controller.save_contacts_privs(contact_list.id, group_id)

        return (contacts, validation_file)

    def get_import_contact_controller(self, contact_type):
        """
        Returns correspondent import controller for the specified contact type
        """
        from twisted.python.reflect import namedAny
        apps = settings.INSTALLED_APPS
        for app in apps:
            try:
                controllers_module = namedAny(app+'.controllers')
            except AttributeError:
                continue

            if not hasattr(controllers_module, 'import_contact_controllers'):
                continue
            available_controllers = controllers_module.import_contact_controllers
            for key in available_controllers:
                if key == contact_type:
                    return available_controllers[key]

        return False

    def scan_for_files(self):
        """
        This is used for automated import, by scanning a directory for files and adding them to the background importer.
        """
        from basesoft.web.contrib.notification.controllers import NotificationController

        # Get all licenses which have the correct permission.
        licenses = self.get_licenses_with_permissions('automated_imports')

        for license in licenses:

            d = os.path.join(settings.CONTACT_IMPORT_DIR, '%s/' % license.id)
            # Check the import directory for files for these licensees.
            if not os.path.exists(d):
                continue

            logger.info('Scanning for new automated import for license [%s]' % license.id)

            lcc = LicenseConfigController(self.sa_session, license)
            # Validate the default config for field mapping.
            mapping_id = lcc.get('contact.import.default.field.mapping.id')
            if mapping_id == 0:
                # Send a notification in case the config has not been set.
                nc = NotificationController(self.sa_session)
                notification = nc.set_notification(
                    'No field mappings defined',
                    'Automated contact import is impossible as there are no field mappings defined'
                )
                nc.create_system_notification(notification, license.default_group)
                logger.warn(' - Not scanning directory as no field mappings are defined.')
                continue

            # loop dir for new files:
            for root, bitbucket, files in os.walk(d):
                for f in files:
                    logger.info(' - Adding file %s to the import queue.' % f)
                    self._add_to_import(root, f, license, mapping_id)

    def _add_to_import(self, root, _file, license, mapping_id):

        file_path = os.path.join(root, _file)
        base_filename, file_extension = os.path.splitext(_file)

        if file_extension in ('.csv', '.txt'):
            _importer = CsvImporter(file_path)

        elif file_extension == '.xlsx':
            _importer = XlsxImporter(file_path)

        elif file_extension == '.xls':
            _importer = XlsImporter(file_path)

        else:
            raise PublicUserException('Import type not supported: %s' % file_extension)

        _importer.set_file_lines()

        fm = self.sa_session.query(ContactFieldMapping).filter(ContactFieldMapping.id==mapping_id).one()
        contact_type = fm.contact_type
        unique_fields = []
        if fm.unique_fields is not None:
            unique_fields = simplejson.loads(fm.unique_fields, object_hook=json_object_hook)

        mapping = []
        for fields in simplejson.loads(fm.mapping, object_hook=json_object_hook):
            target, column = fields['target'].popitem()
            mapping.append({
                'column': column,
                'file_header': fields['field'],
                'target': target
            })
        if unique_fields is not None: # TODO, ticket #3717 - based on fancy select
            uf = []
            for field in unique_fields:
                target, column = field.split('+')
                uf.append({'column': column, 'target': int(target)})
            unique_fields = uf

        # get account from default group for license. ???
        account_id = self.sa_session.query(GroupAccount).filter(GroupAccount.group==license.default_group).first().account.id

        settings = {
            "import_row1": 0,
            "updated_rows": [],
            "separator": _importer.separator,
            "encoding": _importer.encoding,
            "account_id": account_id,
            "do_validation": True,
            "excluded_rows": [],
            "field_mappings": mapping,
            "contact_list": None, # should be in the mapping
            "contact_type": contact_type,
            "unique_fields": unique_fields
        }

        from basesoft.web.contrib.file.controllers import AnonymousFileController
        afc = AnonymousFileController(self.sa_session)

        with file(file_path) as f:
            data = f.read()

        settings = simplejson.dumps(settings, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        _type = self.sa_session.query(ImportFileType).filter(ImportFileType.name=='basesoft.web.contrib.contact').one()

        afc.save_file_for_target_group(data, _file, 'Contact Import files', license.default_group, ImportFile, {
            'status': ImportFile.STATUS_NEW,
            'settings': settings,
            'type': _type
        })

        # remove the file.
        os.remove(file_path)

