# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _

from sqlalchemy import func, case, or_, desc

from basesoft.web.core.util.custom_query import CustomQuery, CustomQueryField, CustomQueryModel
from basesoft.web.contrib.contact.models import ContactList, Contact, ContactListContact, Person, Company,\
    ContactFieldMapping
from basesoft.web.contrib.file.models import File, ImportFile, ImportFileType
from basesoft.web.core.util.controllers import get_models_of_type
from basesoft.web.contrib.i18n.models import Locale
from basesoft.web.contrib.file.custom_query import AbstractFileImportCustomQuery

class ContactListCustomQuery(CustomQuery):

    def tables(self):
        return (
            CustomQueryModel(ContactList),
            CustomQueryModel(ContactListContact, onclause=ContactList.id==ContactListContact.contact_list_id, use_outerjoin=True),
            CustomQueryModel(Contact, onclause=Contact.id==ContactListContact.contact_id, use_outerjoin=True, priv=None),
        )
    def fields(self):
        return (
            CustomQueryField(ContactList.id, _('ID'), key='id'),
            CustomQueryField(ContactList.name, _('Name'), is_filter=True, key='name'),
            CustomQueryField(func.count(Contact.id), _('Contacts'), key="no_of_contacts"),
            CustomQueryField(ContactList.created, _('Created')),
            CustomQueryField(ContactList.last_updated, _('Last updated')),
        )

    group_by = (ContactList.id, )
    order_by = (desc(ContactList.id), )

class ContactCustomQuery(CustomQuery):

    def tables(self):
        return (
            CustomQueryModel(Contact),
            CustomQueryModel(Person, onclause=Contact.person_id==Person.id, use_in_count=False),
            CustomQueryModel(ContactListContact, onclause=Contact.id==ContactListContact.contact_id, use_outerjoin=True, use_in_count=False),
            CustomQueryModel(ContactList, onclause=ContactListContact.contact_list_id==ContactList.id, use_outerjoin=True, priv=None, use_in_count=False),
            CustomQueryModel(Company, onclause=Contact.company_id==Company.id, use_outerjoin=True, use_in_count=True),
            CustomQueryModel(Locale, onclause=Person.preferred_locale_id==Locale.id, use_in_count=False, use_outerjoin=True),
        )
    def fields(self):
        return (
            CustomQueryField(Contact.id, _('ID'), key='id'),
            CustomQueryField(func.concat(case([(Person.first_name!=None, Person.first_name)], else_=""),
                                        " ",
                                        case([(Person.middle_name!=None, Person.middle_name)], else_=""),
                                        " ",
                                        case([(Person.last_name!=None, Person.last_name)], else_=""),
                                        ), _('Full name'), key='contact'),
            CustomQueryField(Company.id, _('Company ID'), key='company_id'),
            CustomQueryField(Contact.polymorphic_type, _('Type'), key='contact_type', is_filter=True),
            CustomQueryField(Company.name, _('Company'), is_filter=True, key='company'),
            CustomQueryField(Company.vat_no, _('Company VAT'), is_filter=True, key='company_vat'),
            CustomQueryField(Person.id, _('Person ID'), key='person_id'),
            CustomQueryField(Person.email, _('Person email'), is_filter=True, key='person_email'),
            CustomQueryField(Contact.email, _('Contact email'), is_filter=True, key='contact_email'),
            CustomQueryField(Person.phone, _('Phone'), key='phone'),
            CustomQueryField(Person.mobile_phone, _('Mobile phone'), key='mobile_phone'),
            CustomQueryField(Locale.name, _('Preferred locale'), key='preferred_locale'),
            CustomQueryField(Contact.id, _('Mail settings'), key='mail_settings'),
            # removing this because it's making it too slow
            #CustomQueryField(func.group_concat(' ', ContactList.name), _('Contact lists'), key='contact_lists'),
            # this won't be shown in the column list, just add it to be used as filter (or static filter)
            CustomQueryField(ContactList.name, _('Contact list'), key='contact_list', is_filter=True, in_select=False),
            CustomQueryField(ContactList.id, _('Contact list ID'), key='contact_list_id', in_select=False),
            CustomQueryField(Person.last_name, _('Last name'), key='last_name', is_filter=True, in_select=False),
            CustomQueryField(Person.first_name, _('First name'), key='first_name', is_filter=True, in_select=False),
        )

    group_by = (Contact.id, )
    #order_by = ('contact', )

class ContactListContactCustomQuery(CustomQuery):
    def tables(self):
        return (
            CustomQueryModel(ContactList),
            CustomQueryModel(ContactListContact),
            CustomQueryModel(Contact),
        )

    def fields(self):
        return (
            CustomQueryField(Contact.id, _('Contact ID'), key='contact_id'),
            CustomQueryField(ContactList.id, _('Contact List ID'), key='cl_id'),
            CustomQueryField(ContactList.name, _('Contact list'), key='name'),
            CustomQueryField(ContactListContact.mail_settings, _('Mail settings'), key='mail_settings')
        )

class FileImportCustomQuery(AbstractFileImportCustomQuery):
    IMPORT_FILE_TYPE ='basesoft.web.contrib.contact'

class FieldMappingCustomQuery(CustomQuery):
    permissions = ('view_contact', 'use_field_mapping_presets', )

    def tables(self):
        return (
            CustomQueryModel(ContactFieldMapping),
        )

    def fields(self):
        return (
            CustomQueryField(ContactFieldMapping.id, _('ID'), key="id"),
            CustomQueryField(ContactFieldMapping.name, _('Name'), key="name"),
            CustomQueryField(ContactFieldMapping.contact_type, _('Type'), key="contact_type"),
            CustomQueryField(ContactFieldMapping.id, _('Default'), key="is_default"),
            CustomQueryField(ContactFieldMapping.created, _('Created'), key="created"),
            CustomQueryField(ContactFieldMapping.last_updated, _('Last updated'), key="last_updated"),
        )

custom_queries = (ContactListCustomQuery, ContactCustomQuery, FileImportCustomQuery, ContactListContactCustomQuery, FieldMappingCustomQuery, )
