# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.
import os

from django.conf import settings

from basesoft.web.core.jsonrpc.dispatcher import JsonRpcDispatcher
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.contrib.contact.controllers import ContactController, ImportContactController
from basesoft.web.contrib.contact.models import Contact
from basesoft.web.contrib.account.decorators import login_required
from basesoft.web.core.util.controllers import get_models_of_type
from basesoft.web.core.exceptions import PublicUserException, SystemException
from django.utils.translation import ugettext as _
from basesoft.web.contrib.file.models import ImportFile, ImportFileType, FileTmp

def person_form_rpc(request):
    """
    """
    from basesoft.web.core.input.validation import plugins
    form_handler = FormValidator({
        'email': [(plugins.Email, {})]
    })
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

def contact_form_rpc(request):
    """
    """
    from basesoft.web.core.input.validation import plugins
    form_handler = FormValidator({
        'contact_email': [(plugins.Email, {})]
    })
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

def address_form_rpc(request):
    """
    """
    form_handler = FormValidator({})
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

def company_form_rpc(request):
    """
    """
    from basesoft.web.core.input.validation import plugins
    form_handler = FormValidator({
        'vat_no': [(plugins.VATOffline, {})]
    })
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

def bank_account_form_rpc(request):
    """
    """
    form_handler = FormValidator({})
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

def get_contact_types(controller):
    """
    @return: Contact types
    """
    rv = []
    for m in get_models_of_type(Contact, False, None, True):
        if m['model'].PERMISSIONS is not None and not controller.has_permissions(*m['model'].PERMISSIONS):
            continue
        d = {
            'value': m['value'],
            'label': m['label'],
            'default': False,
            'can_set_unique_fields': m['model'].can_set_unique_fields,
            'js_module_company': m['model'].company_instance_class.JS_MODULE,
            'js_module_person': m['model'].person_instance_class.JS_MODULE
        }
        if m['value'] == "Contact":
            d['default']= True
        rv.append(d)
    return rv

@login_required
def contact_manager_rpc(request):
    """
    rpc view for contact manager dialog
    """

    sa_session = request.Session()
    if request.session.has_key('user_id'):
        account_id = request.session['user_id']
    else:
        account_id = None

    contact_controller = ContactController(sa_session, account_id)

    class ContactManagerRpc():

        def get_contact_types(self):
            """
            @return: Company types
            """
            return get_contact_types(contact_controller)

        def get_contact_lists(self, id=None, with_no_of_contacts=False):
            """
            @return: {list} with contact lists dict (id and name)
            """
            if id is not None and isinstance(id, list):
                cls = contact_controller.get_contact_lists(id=id)
            else:
                cls = contact_controller.get_contact_lists()
            if with_no_of_contacts:
                return [{'id': cl.id, 'name': cl.name,
                         'contacts': len(contact_controller.get_contacts(None, cl.id))}
                         for cl in cls]
            return [{'id': cl.id, 'name': cl.name} for cl in cls]

        def get_companies_in_contact_list(self, id=None):
            if id is None or id == '':
                return []

            companies = [c.company for c in contact_controller.get_contacts(contact_list_id=id)]
            return [{'label': c.name, 'value': c.id} for c in companies]

        def delete_contact_lists(self, cl_ids, contacts_delete):
            """
            method for deleting a list of contact lists
            @param cl_ids: {list} of int representing contact list ids
            @param contacts_delete: {int} when deleting the contact lists,
                what happends with the contacts attached to them
                CONTACTS_DO_NOT_DELETE = 1
                CONTACTS_DELETE_ONLY_EXCLUSIVE = 2
                CONTACTS_DELETE_ALL = 3
            """
            assert isinstance(cl_ids, list)
            # in case they are not all int, make them
            cl_ids = [int(cl_id) for cl_id in cl_ids]
            contact_controller.delete_contact_lists(contact_lists=contact_controller.get_contact_lists(id=cl_ids),
                                                    delete_contacts_option=contacts_delete)
        def duplicate_contact_lists(self, cl_ids):
            """
            method for duplicating the contact_lists.
            @param cl_ids: {list} of contact list ids
            """
            assert isinstance(cl_ids, list)
            #get all contact list objects
            cls = contact_controller.get_contact_lists(id=cl_ids)
            for cl in cls:
                contact_controller.duplicate_contact_list(cl)

        def save_contact_lists(self, contact_lists, group_id=None):
            """
            Save contact lists.
            This method is called from the ContactManagerPanel on insert (ContactListsComboPanel)
             or on update ( inline edit on ContactLists).
            @param contact_lists: {list} with contact lists dict
            """
            assert isinstance(contact_lists, list)

            contact_lists = contact_controller.save_or_update_contact_lists_from_state(contact_lists_state=contact_lists,
                                                                                       group_id=group_id)
            return [c.id for c in contact_lists]

        def delete_contacts(self, c_ids):
            """
            delete contacts
            @param c_ids: {list} with ids of contacts to be deleted
            """
            assert isinstance(c_ids, list)
            contacts = contact_controller.get_contacts(id=c_ids)
            contact_controller.delete_contacts(contacts=contacts)

        def save_contact_details(self, state):
            """
            Saves contact details received in state
            """
            if not state.has_key('id'):
                return None
            contact_controller = ContactController(sa_session, account_id, state['contact_type'])
            c_state = contact_controller.load_contact_state(state['id'])
            for key, value in state.items():
                if c_state.has_key(key):
                    c_state[key] = value
            contacts = contact_controller.save_or_update_contacts_from_state(contacts_state=[state])
            return [c.id for c in contacts]

        def save_contacts(self, contacts, contact_lists=None, group_id=None):
            """
            Save contacts.
            This method is called from the ContactManagerPanel on insert (ContactCreatePanel)
             or on update ( inline edit on Contacts).
            @param contacts: {list} with contacts dict
            @param contact_lists: {list} contact list ids; if this arg is present, the new contacts
             will be linked to this contact lists
            @return: list with contact ids
            """

            assert isinstance(contacts, list)
            if len(contacts) == 0:
                return []

            # all contacts in the list should have the same type, so get the first one
            contact_controller = ContactController(sa_session, account_id, contacts[0]['contact_type'])
            contacts = contact_controller.save_or_update_contacts_from_state(contacts_state=contacts, group_id=group_id)
            contacts = contact_controller.save_or_update_contacts_with_lists(contacts, group_id, contact_lists)
            return [c.id for c in contacts]

        def link_contacts(self, contact_ids, contact_list, group_id=None):
            """
            Link contacts to contact list
            @param contact_ids: {list} with contact ids
            @param contact_list: {dict} contact list state: id and name; if id is None, a new contactList will be created
            """
            if contact_list.has_key('id') and contact_list['id'] is not None:
                contact_list_id = contact_list['id']
            else:
                contact_list_id = self.save_contact_lists([contact_list], group_id)[0]

            contacts = contact_controller.get_contacts(id=contact_ids)
            contact_list = contact_controller.get_contact_lists(id=contact_list_id)[0]
            contact_controller.link_contacts(contacts, contact_list)

        def load_contact_state(self, contact_id, archived=False):
            """
            Returns the dict with the contact state
            """
            contact = contact_controller.get_contacts(contact_id, archived=archived)[0]
            return {
                'state': contact_controller.load_contact_state(contact),
                'contact_lists': [c.id for c in contact.contact_lists]
            }


        def has_permissions(self, *permissions):
            """
            Method called to check if user has certain permissions.
            Will return the list with the user's permissions (from the ones given as arg)
            """
            return [perm for perm in permissions if contact_controller.has_permissions(perm)]

        def change_mail_settings(self, contact_id, contact_list_id, setting):
            contact_controller.change_mail_settings(contact_id, contact_list_id, setting)

        def merge(self, id_keep, ids, what):
            contact_controller.merge(id_keep, ids, what)
            return True

    cm = ContactManagerRpc()

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(cm)
    return dispatcher.handle_request(request)

@login_required
def contacts_import_rpc(request):
    """
    rpc view for contacts import panel
    """
    # where to save the uploaded import file
    sa_session = request.Session()
    account_id = request.session['user_id']
    group = None

    class ContactsImportRpc:

        def get_import_contact_controller(self, contact_type):
            """
            Returns correspondent import controller for the specified contact type
            """
            from twisted.python.reflect import namedAny
            apps = settings.INSTALLED_APPS
            for app in apps:
                try:
                    controllers_module = namedAny(app+'.controllers')
                except AttributeError:
                    continue

                if not hasattr(controllers_module, 'import_contact_controllers'):
                    continue
                available_controllers = controllers_module.import_contact_controllers
                for key in available_controllers:
                    if key == contact_type:
                        return available_controllers[key]

            return False

        def __get_import_controller(self, files, contact_type='Contact'):
            """
            Return an import controller instance based on the file arg

            The client sends the file hashes and the dispatcher takes care of
            returning the FileTmp objects from that, so the files argument of
            this method is an array of FileTmp objects
            """

            # first search for app speciifc contact importer
            controller = self.get_import_contact_controller(contact_type)
            if not controller:
                controller = ImportContactController

            # since the user can only upload one file, the files list should
            # contain only one file
            assert len(files)==1
            file = files[0]
            assert isinstance(file, FileTmp)

            # get the import controller
            return controller(file_path=os.path.join(settings.TMP_FILE_REPOSITORY, file.file_hash), file_name=file.file_name,
                                         sa_session=sa_session,
                                         account_id=account_id,
                                         contact_type=contact_type)

        def get_field_mappings_data(self, file, separator = None, encoding = None):
            """
            It returns all we need from the server for second page of the wizard:
                * values for row 1 (most likely the file headers),
                * values for row 2 (as columns value examples)
                * available columns for field mappings
                * table fileds separator
                * possible separator options
                * possible actions for unique identifiers
                * contact types
            """

            import_controller = self.__get_import_controller(file)
            import_controller.encoding = encoding
            try:
                import_controller.set_file_lines(separator)
            except PublicUserException:
                if encoding is not None:
                    mes = '%s: "%s"' % (_('Invalid file encoding'), encoding)
                else:
                    mes = _('Failed to auto-determine file encoding')
                return {'exception': mes}

            ct = get_contact_types(import_controller)

            types = None
            columns = None
            if len(ct) == 1 and ct[0]['value'] == "Contact":
                columns = import_controller.get_available_columns()
            else:
                types = ct

            return {
                "row1": import_controller.get_row_values(0),
                "row2": import_controller.get_row_values(1),
                "columns": columns,
                #'extension': import_controller.file_extension, # FIXME
                "separator": import_controller._importer.separator,
                "encoding": import_controller._importer.encoding,
                "separator_options": import_controller.SEPARATOR_OPTIONS,
                "actions": import_controller.get_actions(),
                "contact_types": types,
                "required_columns": import_controller.get_required_columns(),
                "available_presets" : import_controller.get_field_mapping_states()
            }

        def get_columns(self, file, contact_type):
            """
            Returns list with available columns for mapping depending on company_type
            """
            import_controller = self.__get_import_controller(file, contact_type)
            return {"columns": import_controller.get_available_columns(),
                    "required_columns": import_controller.get_required_columns()}

        def get_grouped_select_columns(self, file, contact_type="Contact"):
            """
            Returns list with available columns for the grouped FS depending on contact_type
            """
            import_controller = self.__get_import_controller(file, contact_type)
            cols = import_controller.get_available_columns()
            opts = {}
            for c in cols:
                # only get selectable columns
                if c['id'] in [import_controller.COLUMN_IS_NOT_IMPORTED, import_controller.COLUMN_IS_CONTACT_LIST]:
                    continue
                opts[c['label']] = [{'label': o['label'],
                                     'value': '%s+%s' % (c['id'], o['value'])}
                                     for o in c['options']]
            return opts

        def get_grouped_select_available_fields(self, contact_type='Contact'):
            """
            TODO: the method above and the one below should be replaced with this one.
            """
            controller = self.get_import_contact_controller(contact_type)
            if not controller:
                controller = ImportContactController
            controller = controller(None, sa_session=sa_session, account_id=account_id)
            cols = controller.get_available_columns()

            rv = []

            for c in cols:
                if c['id'] in [controller.COLUMN_IS_NOT_IMPORTED]:
                    continue
                rv.append({
                    'id': c['id'],
                    'label': c['label'],
                    'options': [{'label': o['label'], 'value': o['value']} for o in c['options']]
                })
            return rv

        def get_unique_field_columns(self, contact_type='Contact', include=None):
            controller = self.get_import_contact_controller(contact_type)
            if not controller:
                controller = ImportContactController
            controller = controller(None, sa_session=sa_session, account_id=account_id)
            cols = controller.get_available_columns()
            opts = {}
            for c in cols:
                if include is not None and c['id'] not in include:
                    continue
                # only get selectable columns
                if c['id'] in [controller.COLUMN_IS_NOT_IMPORTED, controller.COLUMN_IS_CONTACT_LIST]:
                    continue
                opts[c['label']] = [{'label': o['label'],
                                     'value': '%s+%s' % (c['id'], o['value'])}
                                     for o in c['options']]
            return opts

        def validate(self, file, field_mappings, separator=None, encoding=None,\
                     import_row1=False, updated_row=None, contact_type="Contact", unique_fields=None):
            """
            Validates all file values
            """
            if contact_type is None:
                contact_type = 'Contact'
            if unique_fields is None:
                unique_fields = []

            import_controller = self.__get_import_controller(file, contact_type)
            import_controller.encoding = encoding
            import_controller.set_file_lines(separator)
            import_controller.field_mappings = field_mappings
            import_controller.import_row1 = import_row1

            # add it to background if too many rows
            import_rows = len(import_controller.get_rows())
            if not import_controller.import_row1:
                import_rows = import_rows - 1;
            if(import_rows > import_controller.MAX_RECORDS_FOR_AUTOMATICAL_IMPORT):
                return {'background': True}

            return import_controller.validate(updated_row, unique_fields=unique_fields)

        def get_overview(self, file, field_mappings, separator=None, encoding=None, import_row1=False, excluded_rows=None,
                         contact_type="Contact"):
            """
            return the overview before the import
            """
            import_controller = self.__get_import_controller(file, contact_type)
            import_controller.encoding = encoding
            import_controller.set_file_lines(separator)
            import_controller.field_mappings = field_mappings
            import_controller.import_row1 = import_row1

            # if there is no contact list name in the file, ask for a contact list at the end
            ask_for_contact_list = True

            #see if there is any column set for import
            cols = [f for f in import_controller.field_mappings
                    if int(f["target"])!=import_controller.COLUMN_IS_NOT_IMPORTED]
            if len(cols) == 0:
                return {"import_rows": 0,
                        "skip_rows": 0,
                        "background": False,
                        "ask_for_contact_list": ask_for_contact_list,
                        "privilege_save_handler": import_controller.privilege_save_handler
                        }

            skip_rows = 0
            import_rows = len(import_controller.get_rows())
            if not import_controller.import_row1:
                import_rows = import_rows - 1;
            if excluded_rows is not None:
                skip_rows = len(excluded_rows)
                import_rows = import_rows - skip_rows
            background = False
            if(import_rows > import_controller.MAX_RECORDS_FOR_AUTOMATICAL_IMPORT):
                background = True
            if background:
                skip_rows = None

            for f in import_controller.field_mappings:
                if int(f["target"]) == import_controller.COLUMN_IS_CONTACT_LIST:
                    ask_for_contact_list = False
                    break

            return {"import_rows": import_rows,
                    "skip_rows": skip_rows,
                    "background": background,
                    "ask_for_contact_list": ask_for_contact_list,
                    "privilege_save_handler": import_controller.privilege_save_handler
                    }

        def import_file(self, file, field_mappings, separator=None, encoding=None, import_row1=False,
                        updated_rows=None, excluded_rows=None, unique_fields=None, group_id=None,
                        contact_type="Contact", background=False, contact_list=None):
            """
            Import file rows
            """
            if unique_fields is None:
                unique_fields = []

            import_settings = {'separator': separator,
                               'encoding': encoding,
                               'field_mappings': field_mappings,
                               'import_row1': import_row1,
                               'contact_type': contact_type,
                               'contact_list': contact_list,
                               'updated_rows': updated_rows,
                               'excluded_rows': excluded_rows,
                               'unique_fields': unique_fields,
                               'account_id': account_id,
                               'group_id': group_id
                               }

            import_controller = self.__get_import_controller(file, contact_type)

            if background:
                import_settings['do_validation'] = True # maybe someday we implement this option
                import_controller.save_file(file, import_settings)
                return {'contacts': []}


            import_controller.encoding = encoding
            import_controller.set_file_lines(separator)
            import_controller.field_mappings = field_mappings
            import_controller.import_row1 = import_row1
            contacts = import_controller.import_contacts(updated_rows, excluded_rows, unique_fields, group_id=group_id)

            vd = []
            if hasattr(import_controller, 'archived_contact_states') and len(import_controller.archived_contact_states) > 0:
                # those ones were not updated, they need to go in the validation file
                for ac in import_controller.archived_contact_states:
                    if ac.has_key('company'):
                        for key in ac['company']:
                            if key != 'polymorphic_type':
                                ac[key] = ac['company'][key]
                    del ac['company']
                    del ac['polymorphic_type']
                    vd.append({"row_index": None,
                               "row_values": ac,
                               "errors": {'': _('Archived contact found for this data')}})

            props = {'status': ImportFile.STATUS_DO_NOT_PROCESS,
                     'result': str(len(contacts))}
            importf = import_controller.save_file(file, import_settings, props)

            if len(vd) > 0:
                validation_file = import_controller.export_csv(vd, file[0].file_name, False)
                from basesoft.web.contrib.file.controllers import ImportFileController
                impc = ImportFileController(sa_session)
                impc.set_data(importf, importf.status, failures_file=validation_file)

            return {'contacts': [c.id for c in contacts], 'archived': len(vd)}

        def export_csv(self, data, file, field_mappings, separator=None, encoding=None):
            import_controller = self.__get_import_controller(file)
            import_controller.encoding = encoding
            import_controller.set_file_lines(separator)
            import_controller.field_mappings = field_mappings
            return import_controller.export_csv(data, file[0].file_name).file_hash

        def get_encodings(self):
            return [
                {'value': 'ascii', 'label': 'ASCII'},
                {'value': 'big5', 'label': 'big5(Traditional Chinese)'},
                {'value': 'cp037', 'label': 'IBM037, IBM039'},

                {'value': 'iso-8859-1', 'label': 'ISO-8859-1(West Europe)'},
                {'value': 'iso-8859-2', 'label': 'ISO-8859-2(Central and Eastern Europe)'},
                {'value': 'iso-8859-3', 'label': 'ISO-8859-3(Esperanto, Maltese)'},
                {'value': 'iso-8859-4', 'label': 'ISO-8859-4(Baltic languages)'},
                {'value': 'iso-8859-5', 'label': 'ISO-8859-5(Bulgarian, Byelorussian, Macedonian, Russian, Serbian)'},
                {'value': 'iso-8859-6', 'label': 'ISO-8859-6(Arabic)'},
                {'value': 'iso-8859-7', 'label': 'ISO-8859-7(Greek)'},
                {'value': 'iso-8859-8', 'label': 'ISO-8859-8(Hebrew)'},
                {'value': 'iso-8859-9', 'label': 'ISO-8859-9(Turkish)'},
                {'value': 'iso-8859-10', 'label': 'ISO-8859-10(Nordic languages)'},
                {'value': 'iso-8859-13', 'label': 'ISO-8859-13(Baltic languages)'},
                {'value': 'iso-8859-14', 'label': 'ISO-8859-14(Celtic languages)'},
                {'value': 'iso-8859-15', 'label': 'ISO-8859-15(Western Europe)'},
                {'value': 'iso-8859-16', 'label': 'ISO-8859-16(South-Eastern Europe)'},

                {'value': 'mac_cyrillic', 'label': 'mac_cyrillic'},
                {'value': 'mac_greek', 'label': 'mac_greek'},
                {'value': 'mac_iceland', 'label': 'mac_iceland'},
                {'value': 'mac_latin2', 'label': 'mac_latin2'},
                {'value': 'mac_roman', 'label': 'mac_roman'},
                {'value': 'mac_turkish', 'label': 'mac_turkish'},

                {'value': 'shift_jis', 'label': 'shift_jis(Japanese)'},

                {'value': 'utf-7', 'label': 'UTF-7'},
                {'value': 'utf-8', 'label': 'UTF-8'},
                {'value': 'utf-16', 'label': 'UTF-16'},
                {'value': 'utf-32', 'label': 'UTF-32'},
            ]

    cm = ContactsImportRpc()

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(cm)
    return dispatcher.handle_request(request)
