# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _
from sqlalchemy import and_, func, or_
from sqlalchemy.orm import aliased
from sqlalchemy.types import Integer
from sqlalchemy.sql.expression import cast

from basesoft.web.core.util.custom_query import CustomQuery, CustomQueryField, CustomQueryModel, DATA_TYPE_BOOLEAN
from basesoft.web.contrib.dashboard.models import Dashboard, DashboardGroup, DashboardAccount
from basesoft.web.contrib.dashboard.controllers import DashboardStateController
from basesoft.web.core.util.controllers import StateManager
from basesoft.web.contrib.license.models import Group
from basesoft.web.contrib.account.models import GroupAccount, Account

class DashboardCustomQuery(CustomQuery):
    def tables(self):
        return (
            CustomQueryModel(Dashboard),
        )

    def fields(self):
        return (
            CustomQueryField(Dashboard.id, _('ID'), key="id"), # required for 'manage object privs'
            CustomQueryField(Dashboard.name, _('Name'), key="name", is_filter=True),
            CustomQueryField(Dashboard.id, _('Default'), key="is_default", type=DATA_TYPE_BOOLEAN),
            CustomQueryField(Dashboard.id, _('Sharing'), key="sharing"),
            CustomQueryField(Dashboard.created, _('Created')),
            CustomQueryField(Dashboard.last_updated, _('Last updated')),
        )

    def set_value_processors(self):

        def proc(row, key):
            value = getattr(row, key)
            if value == self.config.get('basesoft.dashboard.dashboard.id'):
                return True
            return False

        self._value_processors['is_default'] = proc

    def _get_query_filter_on_privileges(self, *args, **kwargs):
        _filter, fo = CustomQuery._get_query_filter_on_privileges(self, *args, **kwargs)
        dc = DashboardStateController(self.sa_session, StateManager(None, self.get_account().id, Dashboard.MODE_COMPOSE))
        _filter = and_(_filter, Dashboard.id.in_([d.id for d in dc.get_dashboards()]) )

        return _filter, fo

ga = aliased(GroupAccount)
da = aliased(DashboardAccount)

class DashboardShareCustomQuery(CustomQuery):
    """
    Query to show all the shares for a supplied dashboard. Crazy sqlalchemy
    """
    def tables(self):
        return (
            CustomQueryModel(DashboardGroup),
            CustomQueryModel(Group, onclause=Group.id==DashboardGroup.group_id),
            CustomQueryModel(ga, onclause=Group.id==ga.group_id, use_outerjoin=False),
            CustomQueryModel(da, onclause=and_(DashboardGroup.dashboard_id==da.dashboard_id, ga.account_id==da.account_id), use_outerjoin=True),
            CustomQueryModel(GroupAccount, onclause=Group.id==GroupAccount.group_id, use_outerjoin=False),
            CustomQueryModel(DashboardAccount, onclause=and_(DashboardGroup.dashboard_id==DashboardAccount.dashboard_id, GroupAccount.account_id==DashboardAccount.account_id), use_outerjoin=True),
            CustomQueryModel(Account, onclause=DashboardAccount.account_id==Account.id, use_outerjoin=True)
        )

    def fields(self):
        return (
            CustomQueryField(GroupAccount.account_id, 'account id', key='account_id'),
            CustomQueryField(DashboardAccount.account_id, 'account id', key='account_id2'),
            CustomQueryField(Group.id, 'group id', key='group_id'),
            CustomQueryField(Group.name, _('Group'), key='group_name', is_filter=True),
            CustomQueryField(func.ifnull(Account.name, _('All accounts')), _('Account'), key='account', is_filter=True),
            CustomQueryField(cast(func.bit_or(DashboardGroup.privilege), Integer).op('&')(Dashboard.PRIV_VIEW)==Dashboard.PRIV_VIEW, _('View privs'), key='view', type=DATA_TYPE_BOOLEAN),
            CustomQueryField(cast(func.bit_or(DashboardGroup.privilege), Integer).op('&')(Dashboard.PRIV_UPDATE)==Dashboard.PRIV_UPDATE, _('Update privs'), key='update', type=DATA_TYPE_BOOLEAN),
            CustomQueryField(DashboardGroup.dashboard_id, _('Dashboard ID'), key='dashboard_id', in_select=False),
        )

    def set_having(self):
        self.having = (or_(
            and_(
                GroupAccount.account_id!=None,
                func.count(da.account_id)==0
            ),
            and_(
                GroupAccount.account_id!=None,
                DashboardAccount.account_id!=None
            )
        ), )

    group_by = (DashboardGroup.dashboard_id, DashboardGroup.group_id, DashboardAccount.account_id)

custom_queries = (DashboardCustomQuery, DashboardShareCustomQuery, )

