# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

"""
Contains the objects, functions and variables to setup the dashboard database
"""
import datetime
from sqlalchemy.orm import mapper, relation
from sqlalchemy import Column, Integer, SmallInteger, String, DateTime, ForeignKey, UnicodeText
from basesoft.sqlalchemy.table import IQTable

from basesoft.web.contrib.license import models as license_models
from basesoft.web.contrib.account import models as account_models
from django.utils.translation import ugettext as _
from basesoft.web.contrib.license.hooks import PrivilegeObjectMapper

__tables_defined = False
__mapping_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    """
    return the list of all tables
    """

    if not __tables_defined:
        do_tables()

    return tables

def do_tables(metadata):
    global __tables_defined
    if not __tables_defined:

        tblDashboard = IQTable('tblDashboard', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('name', UnicodeText, nullable=False),
            Column('active_index', Integer, nullable=False),

            Column('deleted', SmallInteger, nullable=False, default=0),
            Column('archived', SmallInteger, nullable=False, default=0),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
            )

        tblDashboardAccount = IQTable('tblDashboardAccount', metadata,
            Column('dashboard_id', Integer, ForeignKey('tblDashboard.id', ondelete='CASCADE'), primary_key=True),
            Column('account_id', Integer, ForeignKey('tblAccount.id'), primary_key=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
        )

        tblDashboardGroup = IQTable('tblDashboardGroup', metadata,
            Column('dashboard_id', Integer, ForeignKey('tblDashboard.id', ondelete='CASCADE'), primary_key=True),
            Column('group_id', Integer, ForeignKey('tblGroup.id'), primary_key=True),
            Column('privilege', Integer, nullable=False, primary_key=True, autoincrement=False),
            Column('group_has_privilege', SmallInteger, default=0),
            Column('license_has_privilege', SmallInteger, default=0),
            Column('domain_has_privilege', SmallInteger, default=0),
            Column('world_has_privilege', SmallInteger, default=0),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now),
        )


        # Tabs go into a dashboard
        tblDashboardTab = IQTable('tblDashboardTab', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('label', UnicodeText, nullable=False),
            Column('dashboard_id', Integer, ForeignKey('tblDashboard.id', ondelete='CASCADE'), nullable=False),
            Column('position', Integer, nullable=False),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
            )

        # Blocks go into a dashboard tabs
        tblDashboardBlock = IQTable('tblDashboardBlock', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('tab_id', Integer, ForeignKey('tblDashboardTab.id', ondelete='CASCADE'), nullable=False),
            Column('position', Integer, nullable=False),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
            )

        # Columns go in blocks
        tblDashboardColumn = IQTable('tblDashboardColumn', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('block_id', Integer, ForeignKey('tblDashboardBlock.id', ondelete='CASCADE'), nullable=False),
            Column('position', Integer, nullable=False),
            Column('width_percentage', SmallInteger, nullable=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
            )

        # Items go in blocks. Items are abstract and need to be inherited (like CQ or Widget)
        tblDashboardWidget = IQTable('tblDashboardWidget', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('column_id', Integer, ForeignKey('tblDashboardColumn.id', ondelete='CASCADE'), nullable=False),
            Column('polymorphic_type', String(255), nullable=False),

            Column('title', UnicodeText, nullable=False),
            Column('show_header', SmallInteger, nullable=False, default=1),
            Column('hide_margins', SmallInteger, nullable=False, default=0),

            Column('text', UnicodeText, nullable=True, default=None),
            Column('text_location', SmallInteger, nullable=False, default=0),

            # js module where widget is defined
            Column('object_loader', String(255), nullable=False),
            # js object name (widget class)
            Column('object_name', String(255), nullable=False),

            Column('position', Integer, nullable=False),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
            )

        # CQ widget
        tblDashboardWidgetCQ = IQTable('tblDashboardWidgetCQ', metadata,
            Column('widget_id', Integer, ForeignKey('tblDashboardWidget.id', ondelete='CASCADE'), primary_key=True),
            Column('tables_state', UnicodeText),
            Column('filters_state', UnicodeText)
            )

        tables['tblDashboard'] = tblDashboard
        tables['tblDashboardAccount'] = tblDashboardAccount
        tables['tblDashboardGroup'] = tblDashboardGroup
        tables['tblDashboardTab'] = tblDashboardTab
        tables['tblDashboardBlock'] = tblDashboardBlock
        tables['tblDashboardColumn'] = tblDashboardColumn
        tables['tblDashboardWidget'] = tblDashboardWidget
        tables['tblDashboardWidgetCQ'] = tblDashboardWidgetCQ

        __tables_defined = True

def do_mappings():
    """
    This function is only called by the django middleware
    when it is needed
    """
    global __mapping_defined
    if not __mapping_defined:
        from basesoft.web.contrib.dashboard import models as dashboard_models

        mapper(dashboard_models.Dashboard, tables['tblDashboard'],
            extension = PrivilegeObjectMapper(),
            properties = {
                'groups': relation(dashboard_models.DashboardGroup, backref='dashboard', cascade="all", passive_deletes=True),
                'tabs': relation(dashboard_models.DashboardTab, backref='dashboard',
                    order_by=tables['tblDashboardTab'].c.position, passive_deletes=True)
            }
        )

        mapper(dashboard_models.DashboardAccount, tables['tblDashboardAccount'],
           properties = {
                'account': relation(account_models.Account),
                'dashboard': relation(dashboard_models.Dashboard),
            }
        )
        mapper(dashboard_models.DashboardGroup, tables['tblDashboardGroup'],
            properties = {
                'group': relation(license_models.Group)
            }
        )

        mapper(dashboard_models.DashboardTab, tables['tblDashboardTab'], properties = {
            'blocks': relation(dashboard_models.DashboardBlock, backref='tab',
                               order_by=tables['tblDashboardBlock'].c.position, passive_deletes=True)
        })

        mapper(dashboard_models.DashboardBlock, tables['tblDashboardBlock'], properties = {
            'columns': relation(dashboard_models.DashboardColumn, backref='block',
                                order_by=tables['tblDashboardColumn'].c.position, passive_deletes=True)
        })

        mapper(dashboard_models.DashboardColumn, tables['tblDashboardColumn'], properties = {
            'widgets': relation(dashboard_models.DashboardWidget, backref='column',
                                order_by=tables['tblDashboardWidget'].c.position, passive_deletes=True)
        })

        mapper(dashboard_models.DashboardWidget, tables['tblDashboardWidget'],
            polymorphic_on=tables['tblDashboardWidget'].c.polymorphic_type,
            polymorphic_identity='DashboardWidget'
        )

        mapper(dashboard_models.DashboardWidgetCQ, tables['tblDashboardWidgetCQ'],
            polymorphic_identity='DashboardWidgetCQ',
            inherits=dashboard_models.DashboardWidget
        )

        __mapping_defined = True

def app_configs():
    """
    The account configs available within all IQPortal. This can be sync'd with the
    `iqsyncpermissions` command.
    """
    configs = []

    config = account_models.Config()
    config.name = 'basesoft.dashboard.palette.hidden'
    config.type = 'boolean'
    config.default = 'False'
    config.description = u'Boolean for whether or not the Dashboard Palette is hidden.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'basesoft.dashboard.action.hidden'
    config.type = 'boolean'
    config.default = 'True'
    config.description = u'Boolean for whether or not the Dashboard Action panel is hidden.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'basesoft.dashboard.compose.right_layout_unit.width'
    config.type = 'integer'
    config.default = '250'
    config.description = u'Integer used to determine the default width (in px) of the Dashboard compose right panel.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'basesoft.dashboard.compose.right_layout_unit.collapsed'
    config.type = 'boolean'
    config.default = 'False'
    config.description = u'Boolean used to determine if the Dashboard compose right panel should be collapsed by default.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'basesoft.dashboard.compose.remove.confirmation'
    config.type = 'boolean'
    config.default = 'True'
    config.description = u'Boolean used to determine if confirmation should be requested when a module is removed from dashboard.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'basesoft.dashboard.mode.collapsed'
    config.type = 'boolean'
    config.default = 'False'
    config.description = u'Boolean used to determine if the mode change is collapsed.'
    configs.append(config)

    config = account_models.Config()
    config.name = 'basesoft.dashboard.mode.value'
    config.type = 'integer'
    config.default = 1
    config.description = u'Dashboard mode last used'
    configs.append(config)

    config = account_models.Config()
    config.name = 'basesoft.dashboard.dashboard.id'
    config.type = 'integer'
    config.default = 0
    config.description = u'Dashboard ID which will be loaded when getting to the dashboard app.'
    configs.append(config)

    return configs

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'Dashboard')

def app_permissions():
    """
    The permissions used within this application. Can be sync'd with the
    `iqsyncpermissions` command.
    """
    return [
        # dashboard
        license_models.Permission(u'view_dashboard', u"""
            This gives the right to view the dashboard.
            """, _("View the dashboard")),
        license_models.Permission(u'save_dashboard', u"""
            This gives the right to save a personalized dashboard.
            """, _("Save the dashboard")),
        license_models.Permission(u'share_dashboard', u"""
            This gives the right to share dashboard state wit other accounts.
            """, _("Share dashboard")),
        license_models.Permission(u'delete_dashboard', u"""
            This gives the right to delete dashboards
            """, _("Delete dashboard")),
        ]
