# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.
import codecs
import csv
import os
from hashlib import sha256 as sha
import random

import smtplib
from email import Encoders
from email.mime.multipart import MIMEMultipart
from email.utils import formatdate
from email.mime.text import MIMEText
from email.mime.base import MIMEBase

import cStringIO

import shutil
from twisted.python.reflect import namedAny

from django.utils.translation import ugettext as _
from django.utils.translation import ugettext_lazy
from django.conf import settings

from sqlalchemy import and_

from basesoft.web.core.util.controllers import \
    AbstractController, transaction_decorator, get_models_of_type, LicenseConfigController

from basesoft.web.contrib.file.models import File, FileGroup, FileTmp, FileCategory, FileCategoryGroup, ImportFile, ImportFileType, \
    FileImportNotification
from basesoft.web.contrib.license.models import Group
from basesoft.web.core.exceptions import SystemException, PublicUserException, PublicSystemException
from basesoft.web.contrib.notification.controllers import NotificationController

from basesoft import logging
from django.utils import simplejson

logger = logging.getLogger('basesoft.web.contrib.file')

import_file_type_map = None


def load_import_file_type_map():
    global import_file_type_map
    if import_file_type_map is not None:
        return import_file_type_map

    import_file_type_map = {}
    for app in settings.INSTALLED_APPS:
        try:
            module = namedAny(app + '.importers')
        except AttributeError:
            continue

        if not hasattr(module, 'import_file_map'):
            continue

        import_file_type_map.update(module.import_file_map)

    return import_file_type_map


class ImportFileException(PublicUserException):
    pass


class AbstractFileController(AbstractController):
    """
    Controller that holds common methods used by the file controllers
    """
    IMPORT_FILE_TYPE = 'basesoft.web.contrib.file'
    IMPORT_FILE_TYPE_I18N = ugettext_lazy('Import file')

    def _unique_hash(self):
        """
        Generates a unique hash for file
        """
        while True:
            hash = sha(str(random.random())).hexdigest()[:8]
            if self.sa_session.query(FileTmp).filter(FileTmp.file_hash == hash).count() == 0 \
                and self.sa_session.query(File).filter(File.file_hash == hash).count() == 0:
                break
        return hash

    def get_file_content(self, hash):
        """
        """
        from mimetypes import MimeTypes
        file = self.sa_session.query(File).filter(File.file_hash==hash).first()
        if file is not None:
            path = settings.FILE_REPOSITORY
            f = open(os.path.join(path, hash), 'r')
        else:
            file = self.sa_session.query(FileTmp).filter(FileTmp.file_hash==hash).first()
            f = open(os.path.join(settings.TMP_FILE_REPOSITORY, hash), 'r')

        mimetype = MimeTypes().guess_type(file.file_name)

        rv = {'content': f.read(), 'mimetype': mimetype[0], 'filename': file.file_name}
        f.close()
        return rv

    @transaction_decorator
    def get_or_create_file_category(self, label, target_group):
        filter = self.get_query_filter_on_privileges(FileCategory, FileCategoryGroup, FileCategory.PRIV_VIEW, groups=[target_group])
        filter = and_(FileCategory.label==label)
        cat = self.sa_session.query(FileCategory).filter(filter).first()
        if cat is not None:
            return cat
        cat = FileCategory()
        cat.label = label
        self.save_default_privilege_object_group(cat, target_group, (FileCategory.PRIV_VIEW,))
        return cat

class AnonymousFileController(AbstractFileController):

    @transaction_decorator
    def save_file_for_target_group(self, data, file_name, category, target_group, polymorphic_type, props=None, privileges=None):
        """
        No permission checks, this method can be used by schedule methods or similar non-user actions.
        """
        if not os.path.isdir(settings.FILE_REPOSITORY):
            os.mkdir(settings.FILE_REPOSITORY)

        assert isinstance(target_group, Group)

        file_hash = self._unique_hash()
        if polymorphic_type is None:
            polymorphic_type = self._get_type_for_file(file_name)

        file = polymorphic_type()
        file.file_hash = file_hash

        dest = os.path.join(settings.FILE_REPOSITORY, file_hash)

        # create the file and put in it the stream content
        f = open(dest, 'wb')
        if data is not None:
            f.write(data)
        f.close()

        # if we were not provided a file category object, retrieve/create one
        if not isinstance(category, FileCategory):
            category = self.get_or_create_file_category(category, target_group)

        file.file_name = file_name
        file.category = category

        # Additional properties for polymorphic files
        if props is not None:
            for a, v in props.iteritems():
                setattr(file, a, v)

        self.save_default_privilege_object_group(file, target_group, privileges)

        self.sa_session.add(file)

        return file

    def save_file_from_tmp_for_target_group(self, tmp_file, category, target_group, polymorphic_type, props=None, privileges=None):

        if not os.path.isdir(settings.FILE_REPOSITORY):
            os.mkdir(settings.FILE_REPOSITORY)

        assert isinstance(target_group, Group)

        file_hash = self._unique_hash()
        if polymorphic_type is None:
            polymorphic_type = self._get_type_for_file(tmp_file.file_name)

        file = polymorphic_type()
        file.file_hash = file_hash

        source_file_path = os.path.join(settings.TMP_FILE_REPOSITORY, tmp_file.file_hash)
        dest = os.path.join(settings.FILE_REPOSITORY, file_hash)

        shutil.move(source_file_path, dest)

        # if we were not provided a file category object, retrieve/create one
        if not isinstance(category, FileCategory):
            category = self.get_or_create_file_category(category, target_group)

        file.file_name = tmp_file.file_name
        file.category = category

        # Additional properties for polymorphic files
        if props is not None:
            for a, v in props.iteritems():
                setattr(file, a, v)

        self.save_default_privilege_object_group(file, target_group, privileges)

        self.sa_session.add(file)

        return file

class UploadFileController(AbstractFileController):
    """
    Controller for the file app (upload part).
    """

    def get_temp_files(self, id=None):
        """
        @param id: None, list or int
        @return: FileTmp objects

        Returns a list with all FileTmp in case id is None or a list of FileTmp with supplied id/ids.
        (a list of one FileTmp in case a valid id is supplied).
        """
        if id is not None:
            if not isinstance(id, list):
                id = [id]
            return self.sa_session.query(FileTmp).filter(FileTmp.id.in_(id)).all()
        return self.sa_session.query(FileTmp).all()

    def get_hash_temp_files(self, hash=None):
        """
        @param hash: None, list or int
        @return: FileTmp objects

        Returns a list with all FileTmp in case hash is None or a list of FileTmp with supplied hash/hashes.
        (a list of one FileTmp in case a valid hash is supplied).
        """

        if hash is not None:
            if not isinstance(hash, list):
                hash = [hash]

            f = [t for t in self.sa_session.query(FileTmp).filter(FileTmp.file_hash.in_(hash)).all()]
            if len(f) == len(hash):
                return f
            for h in hash:
                if h in [t.file_hash for t in f]:
                    continue
                t = self.sa_session.query(File).filter(File.file_hash==h).one()
                f.append(t)
            return f

        return self.sa_session.query(FileTmp).all()

    @transaction_decorator
    def save_temp_files(self, uploaded_files):
        """
        Called when the user uploads a file, creates a new FileTmp object with unique hash

        @param uploaded_files: uploaded files from request.FILES
        @return: array with FileTmp objects
        """

        if not os.path.isdir(settings.TMP_FILE_REPOSITORY):
            os.makedirs(settings.TMP_FILE_REPOSITORY) # creates intermediary folders too... eg /tmp/iqportal/tmp

        files = []
        for f in uploaded_files:
            # create the sha first so we have a unique identifier to use when storing the file
            file_hash = self._unique_hash()
            file = uploaded_files[f]
            dest = os.path.join(settings.TMP_FILE_REPOSITORY, file_hash)
            file_dest = open(dest, 'wb+')

            for chunk in file.chunks():
                file_dest.write(chunk)
            file_dest.close()

            # save the fileTmp obj
            fileTmp = FileTmp()
            fileTmp.file_name = file.name
            fileTmp.file_hash = file_hash

            self.sa_session.add(fileTmp)
            files.append(fileTmp)
        return files

    @transaction_decorator
    def delete_temp_files(self, files, remove_from_os=False):
        """
        Remove a temp file from the table and from the disk.
        """
        for f in files:
            assert isinstance(f, FileTmp)
            if os.path.isfile(os.path.join(settings.TMP_FILE_REPOSITORY, f.file_hash)):
                if not remove_from_os:
                    logger.warn('Cannot remove temp file %s from database because file still exists on file system' % f)
                    continue
                os.remove(os.path.join(settings.TMP_FILE_REPOSITORY, f.file_hash))
            self.sa_session.delete(f)

class ProcessFileController(AbstractFileController):
    """
    Controller for the file app (process file part)
    """

    # properties which will be set for different file polymorphic type
    file_type_properties = {}

    def get_files_by_type(self, type=File):
        """
        Returns the file categories with all the files of the provided type FileImage, Invoice, etc.
        @param type: the type of the needed files
        """
        from sqlalchemy.orm import join

        fo = join(type, FileGroup)
        filter, fo = self.get_query_filter_on_privileges(type, FileGroup, File.PRIV_VIEW, fo)
        filter = and_(filter, type.deleted==False) # get only files that are not marked as deleted
        filter = and_(filter, type.archived==False) # get only files that are not marked as archived

        return self.sa_session.query(type).select_from(fo).filter(filter).all()

    def _get_polymorphic_types(self):
        """
        Get all file models
        """
        return [i['model'] for i in get_models_of_type(File)]

    def _get_type_for_file(self, file_name):
        """
        """
        rv = File
        # TODO: make more elegant and better - #3181
        from mimetypes import MimeTypes
        mimetype = MimeTypes().guess_type(file_name)
        if mimetype is None:
            return rv
        mimetype = mimetype[0]
        for t in self._get_polymorphic_types():
            if t.MIMETYPES is None:
                continue
            if mimetype in t.MIMETYPES:
                return t
        return rv

    def get_files(self, id=None):
        """
        @param id: None, list or int
        @return: File objects

        Returns a list with all Files in case id is None or a list of Files with supplied id/ids.
        (a list of one File in case a valid id is supplied).
        """
        self.assert_permissions('view_file')
        return self.get_privileged_object_by_id(File, id)

    def get_file_by_hash(self, hash):
        """
        Gets a file by it's hash value
        """
        self.assert_permissions('view_file')
        result = self._get_privileged_object(File, File.file_hash==hash, File.PRIV_VIEW)
        if len(result)==0:
            return None
        return result.pop()

    def get_categories(self, id=None):
        """
        Get all the categories to which the current logged in user group has VIEW privileges.
        """
        self.assert_permissions('view_file')
        return self.get_privileged_object_by_id(FileCategory, id)

    def get_category_by_label(self, label):
        """
        Gets a category by the label for the current logged in user group. Used to determine
        whether or not the category is a new one or an existing one.
        """
        self.assert_permissions('view_file')
        result = self._get_privileged_object(FileCategory, FileCategory.label==label, FileCategory.PRIV_VIEW)
        if len(result)==0:
            return None
        return result.pop()

    @transaction_decorator
    def save_files_from_state(self, state):
        """
        A state includes the hashes of the files to be saved as well as the ID/label of the file category.
        """

        assert state.has_key('category')
        assert state.has_key('hashes')

        category = self.get_category_by_label(state['category'])
        if category is None:
            category = self.save_or_update_file_category(None, state['category'], None)

        hashes = state['hashes']

        ufc = UploadFileController(self.sa_session)
        temp_files = ufc.get_hash_temp_files(hashes)
        return self.save_files_from_temp(temp_files, category, None)

    @transaction_decorator
    def save_or_update_file_category(self, id, label, group=None):
        """
        Saves or updates a file category
        """

        if group is None:
            group = self._get_default_group()
        else:
            assert isinstance(group, Group)

        if id is None:
            self.assert_permissions('create_file')
            fc = FileCategory()
        else:
            self.assert_permissions('update_file')
            fc = self.sa_session.query(FileCategory).filter(FileCategory.id==id).one()
            self.assert_privileges_on_object(fc, FileCategory.PRIV_UPDATE)

        fc.label = label

        if id is None:
            self.save_default_privilege_object_group(fc)

        self.sa_session.add(fc)
        return fc

    @transaction_decorator
    def save_files_from_temp(self, temp_files, category, polymorphic_type=None, extra=None, remove=True):
        """
        Saves the temp uploaded files in the destination and saves them in the File table.

        @param temp_files: list FileTmp objects
        @param extra: dictionary of key/values that will be assigned to the new File object
        @return: array with File objects
        """
        files = []
        for f in temp_files:
            if isinstance(f, File):
                _id = f.id
                remove = False
            else:
                assert isinstance(f, FileTmp)
                _id = None
            # save the file obj
            _file = self.save_or_update_file(id=_id,
                                            file_name=f.file_name,
                                            file_hash=f.file_hash,
                                            remove_source = False,
                                            category=category,
                                            file_obj=None,
                                            polymorphic_type=polymorphic_type, extra=extra)

            if remove:
                UploadFileController(self.sa_session, None).delete_temp_files([f], True)

            files.append(_file)

        return files

    @transaction_decorator
    def save_or_update_file_from_content(self, id, file_name, file_hash, category, data=None,
                                         polymorphic_type=None, file_obj=None):
        """
        Saves or updates a file. It creates the file with the content from the file_stream in the destination_dir
        under the name <file_hash>_<file_name>

        !! A file hash cannot be updated

        @param id: {Int}; id of the file or None
        @param file_name: String with file name
        @param file_hash: String with a unique file hash or None if we're creating a new one
        @param data: file data: can come from a string, a stream (like stream.getvalue()), etc
            if None, it just creates the file (like 'touch') and doesn't write anything to it. it can later be opened
            with open(file.file_path, 'wb') for later updating
        @param polymorphic_type: {Class} the class name for the file we want to create. eg. ImageFile. If None,
            determine the class based on the mimetypes
        @param file_obj: {Object} a File type object. if given, a new File object is not created anymore

        @return: a File object
        """
        if not os.path.isdir(settings.FILE_REPOSITORY):
            os.mkdir(settings.FILE_REPOSITORY)

        assert isinstance(category, FileCategory)

        if id is None:
            self.assert_permissions('create_file')

            if file_hash is None:
                # create a new hash for the file
                file_hash = self._unique_hash()

            # check if file_hash is unique
            assert self.get_file_by_hash(file_hash) is None

            if file_obj is not None:
                file = file_obj
            else:
                if polymorphic_type is None:
                    polymorphic_type = self._get_type_for_file(file_name)

                file = polymorphic_type()

            file.file_hash = file_hash

        else:
            self.assert_permissions('update_file')
            file = self.sa_session.query(File).filter(File.id==id).one()
            self.assert_privileges_on_object(file, File.PRIV_UPDATE)

        # check if the destination_dir has a trailing slash
        dest = os.path.join(settings.FILE_REPOSITORY, file_hash)

        # create the file and put in it the stream content
        f = open(dest, 'wb')
        if data is not None:
            f.write(data)
        f.close()

        file.file_name = file_name
        file.category = category

        if id is None:
            self.save_default_privilege_object_group(file)

        self.sa_session.add(file)

        return file

    @transaction_decorator
    def save_or_update_tmp_file_from_content(self, file_name, data=None):
        """
        Saves or updates a tmp file.

        @param file_name: String with file name
        @param data: file data: can come from a string, a stream (like stream.getvalue()), etc
            if None, it just creates the file (like 'touch') and doesn't write anything to it. it can later be opened
            with open(file.file_path, 'wb') for later updating
        @return: a FileTmp object
        """
        if not os.path.isdir(settings.TMP_FILE_REPOSITORY):
            os.mkdir(settings.TMP_FILE_REPOSITORY)
        file_hash = self._unique_hash()
        file = FileTmp()
        file.file_hash = file_hash
        # check if the destination_dir has a trailing slash
        dest = os.path.join(settings.TMP_FILE_REPOSITORY, file_hash)

        # create the file and put in it the stream content
        f = open(dest, 'wb')
        if data is not None:
            f.write(data)
        f.close()

        file.file_name = file_name
        self.sa_session.add(file)

        return file

    @transaction_decorator
    def save_or_update_file(self, id, file_name, file_hash,
                            remove_source, category, file_obj=None, polymorphic_type=None, extra=None):
        """
        Saves or updates a file. It copies the file source_file_path in the destination_dir
        under the name <file_hash>_<file_name>

        !! A file hash cannot be updated

        @param id: {Int}; id of the file or None
        @param file_name: String with file name
        @param file_hash: String with a unique file hash
        @param file_obj: {Object} a File type object. if given, a new File object is not created anymore
        @param remove_source: boolean - if the source file should be removed or not
        @return: array with File objects
        """
        if not os.path.isdir(settings.FILE_REPOSITORY):
            os.mkdir(settings.FILE_REPOSITORY)

        assert isinstance(category, FileCategory)

        if polymorphic_type is None:
            polymorphic_type = self._get_type_for_file(file_name)

        if id is None:
            self.assert_permissions('create_file')
            # check if file_hash is unique
            assert self.get_file_by_hash(file_hash) is None, 'hash collision: %s' % file_hash

            if file_obj is not None:
                file = file_obj
            else:
                file = polymorphic_type()

            file.file_hash = file_hash
        else:
            self.assert_permissions('update_file')
            if polymorphic_type == File:
                file = self.sa_session.query(File).filter(File.id==id).one()
            else:
                file = self.sa_session.query(polymorphic_type).filter(polymorphic_type.file_id==id).one()
            self.assert_privileges_on_object(file, File.PRIV_UPDATE)

        dest = os.path.join(settings.FILE_REPOSITORY, file_hash)
        source_file_path = os.path.join(settings.TMP_FILE_REPOSITORY, file_hash)

        # copy/move the file
        if not remove_source:
            os.system ("cp %s %s" % (source_file_path, dest))
        else:
            os.system ("mv %s %s" % (source_file_path, dest))

        file.file_name = file_name
        file.category = category

        if extra is not None:
            for k,v in extra.iteritems():
                setattr(file, k, v)

        for key in self.file_type_properties:
            setattr(file, key, self.file_type_properties[key])

        if id is None:
            self.save_default_privilege_object_group(file)

        self.sa_session.add(file)

        return file

    @transaction_decorator
    def delete_files(self, files):
        """
        Remove a file from the table and from the disk.
        """
        self.assert_permissions('delete_file')

        for f in files:
            assert isinstance(f, File)
            self.mark_as_deleted(File, f.id)

class ImportFileController(AbstractFileController):
    """
    Controller for the import file polymorphic type
    """
    MAX_RECORDS_FOR_INSTANT_IMPORT = 100
    STOP_AFTER_NUMBER_EMPTY_ROWS = 1
    NUMBER_EXPECTED_COLUMNS = None

    _failure_lines = None
    _writer = None
    _file_headers = None

    def __init__(self, sa_session=None, locale=None, account_id=None, request=None, check_active=True):
        super(ImportFileController, self).__init__(sa_session, locale, account_id, request, check_active)
        self.reset()

    def _import_row(self, row, controller, dryrun, **kwargs):
        raise PublicSystemException('Implement in child class')

    def reset(self):
        self._file_rows = []
        self._failure_lines = []
        self._file_headers = []

    def reset_failure_lines(self):
        self._failure_lines = []

    def get_file_path(self, _file):
        file_path = _file

        if isinstance(_file, FileTmp):
            file_path = os.path.join(settings.TMP_FILE_REPOSITORY, _file.file_hash)
        elif type(_file) in [File, ImportFile]:
            file_path = os.path.join(settings.FILE_REPOSITORY, _file.file_hash)
        return file_path

    def get_column_headers(self):
        return self._file_headers

    def get_rows(self, file_path):
        """
        Return the file lines. Call from child.
        """
        rv = []
        for row in self.read_row(file_path):
            rv.append(row)
        return rv

    def get_row_count(self, file_path):
        rv = len(self.get_rows(file_path))
        self.reset()
        return rv

    def is_empty_row(self, row):
        for cell in row:
            if type(cell) in (str, unicode):
                cell = cell.strip()
            if cell not in [None, '']:
                return False
        return True

    def get_only_expected_columns(self, row):
        if len(row) > self.NUMBER_EXPECTED_COLUMNS:
            return row[:self.NUMBER_EXPECTED_COLUMNS]
        return row

    @staticmethod
    def get_notification_update_notify():
        return []

    def get_import_file_setting(self, _file):
        assert isinstance(_file, ImportFile)

        import_settings = simplejson.loads(_file.settings)
        _map = load_import_file_type_map()
        if not _map.has_key(_file.type.name):
            return {}

        c = _map[_file.type.name](self.sa_session, account_id=self.get_account().id)
        rv = {
            'file_name': _file.file_name
        }
        rv['settings'] = c.get_human_readable_settings(import_settings)
        return rv

    def get_human_readable_settings(self, import_settings):
        return {}

    @transaction_decorator
    def background_import_file(self, _file, controller, dryrun, **kwargs):
        file_path = self.get_file_path(_file)
        rv = self._import_file(file_path, controller, dryrun, **kwargs)

        validation_file = None
        if len(self._failure_lines) > 0:
            validation_file = self.write_failure_file(_file, self.get_column_headers(), self._failure_lines)
            self.set_import_result(_file, ImportFile.STATUS_SUCCESS, None, validation_file)
        return rv, validation_file

    @transaction_decorator
    def import_file(self, _file, controller, dryrun=False, **kwargs):

        file_path = self.get_file_path(_file)
        self.reset()

        row_count = self.get_row_count(file_path)

        # if account is None this is db_sync script
        if row_count > self.MAX_RECORDS_FOR_INSTANT_IMPORT and self._account is not None:
            self.save_import_file(_file, **kwargs)
            return None, None

        rv = self._import_file(file_path, controller, dryrun, **kwargs)

        return rv, self._failure_lines

    def _import_file(self, file_path, controller, dryrun, **kwargs):
        row_i = 0
        empty_rows = 0
        rv = []
        for row in self.read_row(file_path):
            is_empty = self.is_empty_row(row)
            if is_empty is True:
                empty_rows += 1
            else:
                empty_rows = 0

            if empty_rows >= self.STOP_AFTER_NUMBER_EMPTY_ROWS:
                break
            row_i += 1
            if row_i == 1: # Dont import the header line.
                self._file_headers = row
                continue
            if len(row) < self.NUMBER_EXPECTED_COLUMNS:
                r = row + ['']*(self.NUMBER_EXPECTED_COLUMNS-len(row))
                self._failure_lines.append([row_i] + r + [_('Found %(found)d columns, expected %(expected)d.' % {
                    'found': len(row),
                    'expected': self.NUMBER_EXPECTED_COLUMNS
                })])
                continue
            try:
                rv.append(self._import_row(row, controller, dryrun, **kwargs))
            except ImportFileException, e:
                self._failure_lines.append([row_i] + row + [e.message])

        # to do any post processing after rows have been read, useful for saving objects to db
        self._post_read_rows(controller, **kwargs)

        if len(self._failure_lines) > 0:
            self._file_headers.insert(0, 'Line number')
        return rv

    def _post_read_rows(self, *args, **kwargs):
        """
        called when the _import_file() is done reading all rows
        :return:
        """

    @transaction_decorator
    def save_import_file(self, tmp_file, **kwargs):
        self.assert_permissions('create_file')

        pfc = ProcessFileController(self.sa_session, account_id=self._account.id)
        category = pfc.get_or_create_file_category('File imports', self.get_default_group())

        _type = self.sa_session.query(ImportFileType).filter(ImportFileType.name == self.IMPORT_FILE_TYPE).one()

        settings = simplejson.dumps(kwargs)
        pfc.save_files_from_temp([tmp_file], category, ImportFile, {
            'status': ImportFile.STATUS_NEW,
            'settings': settings,
            'type': _type
        })

        return True

    def set_import_result(self, import_file, status, result, failure_file=None):
        import_file.status = status
        import_file.result = result
        import_file.failure_file = failure_file
        self.sa_session.add(import_file)
        self.sa_session.flush()

    def write_failure_file(self, import_file, column_headers, failures, polymorphic_type=File):
        column_headers.append('failure')
        failures.insert(0, column_headers)

        base_filename, file_extension = os.path.splitext(import_file.file_name)
        file_name = base_filename + '.csv'

        file_path, file_name = self.save_failure_file(file_name, failures)
        with file(file_path) as f:
            data = f.read()

        afc = AnonymousFileController(self.sa_session)
        return afc.save_file_for_target_group(data, file_name, 'Import Failures', import_file.groups[0].group, polymorphic_type)

    def save_failure_file(self, file_name, lines):
        file_name = 'failures_%s' % file_name
        file_path = os.path.join(settings.TMP_FILE_REPOSITORY, file_name)
        file_ = open(file_path, 'wb')

        self._queue = cStringIO.StringIO()
        self._writer = csv.writer(self._queue, delimiter="\t", quotechar='"', quoting=csv.QUOTE_ALL)
        self._stream = file_
        self._encoder = codecs.getincrementalencoder('utf-8')()

        for line in lines:
            self._write_row(line)

        file_.close()
        return file_path, file_name

    def _write_row(self, row):
        self._writer.writerow(['' if s is None else unicode(s).encode("utf-8") for s in row])
        data = self._queue.getvalue()
        data = data.decode("utf-8")
        data = self._encoder.encode(data)
        self._stream.write(data)
        self._queue.truncate(0)

    @transaction_decorator
    def set_data(self, file, status, result=None, failures_file=None, callback=None,
                 subject=None, text=None, notify=None):
        """
        Sets import file data. Used in different stages of import by apps specific import tasks.
        """
        from datetime import datetime
        assert isinstance(file, ImportFile)
        file.status = status
        if failures_file is not None:
            assert isinstance(failures_file, File)
            file.failure_file = failures_file
        # FIXME: need to find a better way of updating 'last_update' for 'parents' of polymorphic objects - #3183
        file.last_updated = datetime.now()
        if result is not None:
            file.result = unicode(result)
        self.sa_session.add(file)

        self._notify(file, callback, subject, text, notify)

    @transaction_decorator
    def _notify(self, file, callback, subject, text, notify):
        notification = None

        nc = NotificationController(self.sa_session)

        if file.status == ImportFile.STATUS_SUCCESS:
            if file.failure_file is None:
                notification = self._notify_success(file, nc, callback, subject, text, notify)
            else:
                notification = self._notify_failure(file, nc, callback, subject, text, notify)

        if notification is None:
            return

        nc.create_system_notification(notification, file.groups[0].group)

    def _notify_success(self, file, nc, callback, subject, text, notify):
        if subject is None:
            subject = 'File imported succesfully (%(file_name)s)' % {'file_name': file.file_name}
        if text is None:
            text = 'The file (%(file_name)s) was imported succesfully.' % {'file_name': file.file_name}
        return nc.set_notification(subject, text, instance_class=FileImportNotification, callback=callback, notify=notify)

    def _notify_failure(self, file, nc, callback, subject, text, notify):

        if subject is None:
            subject = 'File import contained errors (%(file_name)s)' % {'file_name': file.file_name}
        if text is None:
            text = 'The file (%(file_name)s) you uploaded for import contained errors, ' \
                   'follow the link to download the failure file. <a target="_blank" ' \
                   'href="/basesoft.web.contrib.file/view/%(failure_file_hash)s_%(failure_file_name)s/attachment/">%(failure_file_name)s</a>'\
                   % {
                      'file_name': file.file_name,
                      'failure_file_hash': file.failure_file.file_hash,
                      'failure_file_name': file.failure_file.file_name
                   }
        notification = nc.set_notification(subject, text, instance_class=FileImportNotification, callback=callback, notify=notify)

        # get the license config option for email notifications on import failures.
        lcc = LicenseConfigController(self.sa_session, file.groups[0].group.license)
        recipient = lcc.get('file.import.failure.email.notification.address')
        if recipient is None or recipient == '':
            return notification

        # create an email with the failure file as attachment.
        s = smtplib.SMTP(settings.EMAIL_HOST, settings.EMAIL_PORT)
        message = MIMEMultipart()
        message['From'] = settings.EMAIL_FROM
        message['To'] = recipient
        message['Date'] = formatdate(localtime=True)
        message['Subject'] = '[iqportal] File import contained errors (%(file_name)s)' % {'file_name': file.file_name}

        message.attach( MIMEText('The file (%(file_name)s) you uploaded for import contained errors, see attachment for specifics.' % {'file_name': file.file_name}) )

        content = self.get_file_content(file.failure_file.file_hash)

        part = MIMEBase('application', "octet-stream")
        part.set_payload( content['content'] )
        Encoders.encode_base64(part)
        part.add_header('Content-Disposition', 'attachment; filename="%s"' % file.file_name)
        message.attach(part)

        if settings.EMAIL_HOST is None:
            self.logger.warn('WARNING: not sending email, settings.EMAIL_HOST is None')
        else:
            s.sendmail(settings.EMAIL_FROM, recipient, message.as_string())

        return notification

class CSVFileController:

    QUOTECHAR_OPTIONS = ['"', "'", '|', '']
    SEPARATOR_OPTIONS = [
        {"label": ugettext_lazy("Comma"), "value": ","},
        {"label": ugettext_lazy("Tab"), "value": "\t"},
        {"label": ugettext_lazy("Semi-colon"), "value": ";"},
    ]

    def __init__(self, file_path, separator=None, quotechar=None, encoding=None):
        import os.path
        assert os.path.exists(file_path)

        self.__file_path = file_path

        self.separator = separator
        self.quotechar = quotechar
        self.encoding = encoding

        self.__file_rows = []
        self.set_file_rows(separator, quotechar)

    def get_file_rows(self):
        return self.__file_rows

    def set_file_rows(self, separator=None, quotechar=None):
        """
        Returns the file rows
        @return: {Array} with file rows
        """
        if separator is None:
            self.__autodetermine_separator()
        else:
            for s in self.SEPARATOR_OPTIONS:
                if separator == s["value"]:
                    self.separator = s["value"]
                    break
            if self.separator is None:
                raise SystemException('Invalid separator %s' % separator)
        if quotechar is None:
            # TODO: autodetermine quotechar - #3180
            self.quotechar = self.QUOTECHAR_OPTIONS[0]
        else:
            self.quotechar = quotechar
        import csv

        rows = csv.reader(open(self.__file_path, 'rU'), delimiter=self.separator, quotechar=self.quotechar)

        encodings = ['utf-8', 'iso-8859-1', 'iso-8859-2', 'ascii']
        chardet_used = False
        ret = []
        try:
            for r in rows:
                if len(r) > 0:
                    row = []
                    for v in r:
                        if self.encoding is not None:
                            try:
                                v = unicode(v, self.encoding)
                            except UnicodeDecodeError:
                                self.encoding = None
                        if self.encoding is None:
                            for enc in encodings:
                                try:
                                    v = unicode(v, enc)
                                    self.encoding = enc
                                    break
                                except UnicodeDecodeError:
                                    pass
                        if self.encoding is None and not chardet_used:
                            self.encoding = self.__detect_encoding()
                            chardet_used = True
                            try:
                                v = unicode(v, self.encoding)
                            except UnicodeDecodeError:
                                self.encoding = None
                        if self.encoding is None:
                            raise PublicUserException(_('Invalid file encoding'))
                        row.append(v.strip())
                    ret.append(row)
        except csv.Error, e:
            print 'file %s, line %d: %s' % (self.__file_path, rows.line_num, e.message)
            raise

        self.__file_rows = ret
        return ret

    def __detect_encoding(self):
        from chardet.universaldetector import UniversalDetector
        detector = UniversalDetector()
        infile = open(self.__file_path, "r")
        i = 0
        for line in infile:
            detector.feed(line)
            if detector.done: break
            # only check for the first 1000 rows otherwise it could timeout
            if i == 1000: break
            i = i + 1
        detector.close()
        return detector.result['encoding']

    def __autodetermine_separator(self):
        """
        Method to autodetermine the columns separator in a csv file
        """
        separators = [s["value"] for s in self.SEPARATOR_OPTIONS]

        infile = open(self.__file_path, "r")
        ret = []
        for line in infile:
            #remove line endings
            line = line.rstrip('\n')
            line = line.rstrip('\r')
            ret.append(line)
            if len(ret) == 3: # test for the first 3 rows(should be enough)
                break

        sep_count = {}
        if len(ret) == 3:
            for separator in separators:
                v1 = ret[0].count(separator)
                v2 = ret[1].count(separator)
                v3 = ret[2].count(separator)
                if v1==0 or v2==0 or v3==0:
                    continue
                if v1==v2 and v2==v3:
                    sep_count[separator] = v1

        # just pick the first one if nothing matched
        if len(sep_count)==0:
            self.separator = separators[0]
            return
        max = 0
        for key in sep_count:
            if sep_count[key] > max:
                max = sep_count[key]
                sep = key
        self.separator = sep
        return
