# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.
from basesoft.web.core.exceptions import PublicSecurityException

from basesoft.web.core.jsonrpc.dispatcher import JsonRpcResponse, \
    JsonRpcResponseException, JsonRpcDispatcher

from basesoft.web.contrib.file.controllers import UploadFileController, ProcessFileController
from django.http import HttpResponseNotFound, HttpResponseRedirect
from basesoft.web.contrib.file.models import File, ImageFile

from django.conf import settings
from django.utils.translation import ugettext as _

def validation(request):
    """
    """
    from basesoft.web.core.input.validation import plugins
    from basesoft.web.core.forms.handler import FormValidator
    form_handler = FormValidator({
        'category': [(plugins.Required, {})],
        'hashes': [(plugins.Required, {})]
        })
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(form_handler)
    return dispatcher.handle_request(request)

def upload_file(request):
    """
    View for file upload
    """
    sa_session = request.Session()
    try:
        fc = UploadFileController(sa_session)
        files = fc.save_temp_files(request.FILES)

        return JsonRpcResponse(request, {"upload_response": [f.file_hash for f in files],
                                         "files": [f.file_name for f in files]})
    except Exception, e:
        return JsonRpcResponseException(request, e)

def upload_file_rpc(request):
    """
    contains rpc methods called from the file uploader
    """

    class Handler:
        """
        Handler for file related rpc methods
        """
        def __init__(self, sa_session, account_id):
            """
            Instantiates the Upload and Process file controllers.
            """
            self._fc = UploadFileController(sa_session)
            self._pfc = ProcessFileController(sa_session, account_id=account_id)

        def remove_temp_files(self, hashes):
            """
            Remove a temp file. Delete from table and from the disk.
            """
            self._fc.delete_temp_files(self._fc.get_hash_temp_files(hashes))

        def get_categories(self):
            """
            Gets a list of categories for a dropdown/combobox
            """
            return [{'label': c.label, 'value': c.id} for c in self._pfc.get_categories()]

        def get_image_files(self, category_id=None):
            """
            """
            from basesoft.web.contrib.file.models import ImageFile
            rv = []
            for f in self._pfc.get_files_by_type(ImageFile):
                rv.append({'id': f.id, 'name': f.file_name, 'hash': f.file_hash, 'category': f.category.label})
            return rv

        def save_files(self, state):
            """
            Moves the files from temp status to permanent.
            """
            rv = []
            for f in self._pfc.save_files_from_state(state):
                rv.append({'id': f.id, 'name': f.file_name, 'hash': f.file_hash, 'category': f.category.label})
            return rv

        def delete_files(self, ids):
            """
            Moves the files from temp status to permanent.
            """
            self._pfc.delete_files(ids)

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(Handler(request.Session(), request.session['user_id']))
    return dispatcher.handle_request(request)

def view_file(request, hash, attachment=False, filename=None):

    from django.http import HttpResponse
    account_id = None
    if request.session.has_key('user_id'):
        account_id = request.session['user_id']
    pfc = ProcessFileController(request.Session(), request=request, account_id=account_id)

    file = request.Session().query(File).filter(File.file_hash==hash).first()

    if file is not None and not isinstance(file, ImageFile):
        if account_id is None:
            raise PublicSecurityException(_('You do not have the rights to view the requested file.'))
        # we should privilege check the file.
        files = pfc.get_files(file.id)
        if len(files) == 0:
            raise PublicSecurityException(_('You do not have the rights to view the requested file.'))
        file = files.pop()

    try:
        content = pfc.get_file_content(hash)
    except IOError:

        if settings.USE_MOCK_FILES:

            if isinstance(file, ImageFile):
                return HttpResponseRedirect('/basesoft.web.core/rte/make_image/%s' % _('Image not available'))
            else:
                return HttpResponse(_('The content of this file is not available'))

        # else notify by email
        raise


    response = HttpResponse(content['content'], mimetype=content['mimetype'])
    response['Content-Length'] = len(content['content'])
    response['Accept-Ranges'] = 'bytes'

    # If attachment is given (means the browser should present a 'download' dialog. Since filenames can contain
    # UTF-8 chars do a redirect to the same view but with the filename. Then set the content-disposition header,
    # the browser will then take the last url part for the filename. Fix for #4086
    if attachment and not filename:
        return HttpResponseRedirect(u'/basesoft.web.contrib.file/view/%s/attachment/%s' % (hash, content['filename']))
    elif attachment and filename:
        response['Content-Disposition'] = 'attachment'

    return response
