# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _

from sqlalchemy import exc as exceptions
from sqlalchemy.sql.expression import or_, and_
from sqlalchemy.orm import outerjoin

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from basesoft.web.contrib.i18n.models import Locale, Country, DomainCountry, DomainLocale, DomainLanguage, Language
from basesoft.web.contrib.license.models import Domain
from basesoft.web.core.exceptions import PublicSecurityException

class I18nController(AbstractController):

    CHECK_ACTIVE_ACCOUNT = False

    def get_locale_by_str(self, locale_str):
        """
        Returns a Locale object for the given locale_str. Throws an exception if not exists.
        """
        try:
            return self.sa_session.query(Locale).filter_by(locale=locale_str).one()
        except exceptions.InvalidRequestError:
            raise Exception('No locale found for %s' % locale_str)

    def get_locales(self, id=None, availability=None):
        """
        @param id: None, list or int
        @return: Locale objects

        Returns a list with all locales in case id is None or a list of locales with supplied id/ids.
        (a list of one locale in case a valid id is supplied)
        """
        filter = None
        if availability is not None:
            filter = Locale.availability.op('&')(availability)
        if id is not None:
            if not isinstance(id, list):
                id = [id]
            if filter is None:
                filter = Locale.id.in_(id)
            else:
                filter = and_(filter, Locale.id.in_(id))
        if filter is None:
            return self.sa_session.query(Locale).order_by(Locale.name).all()
        return self.sa_session.query(Locale).filter(filter).order_by(Locale.name).all()

    def get_domain_locales(self, id=None, domain_id=None, availability=None):
        """
        Returns all locales available for the current domain: if there are any locales in DomainLocale table for
        provided domain or for current domain, only check locales, else check for Locale table
        """
        if domain_id is not None:
            if self.get_account() is not None: # this is for getting domain locales for the public locales select
                self.assert_permissions('view_all_domains')
            domain = self.sa_session.query(Domain).filter(Domain.id==domain_id).one()
        else:
            domain = self._get_default_group().license.domain

        #first check if there are any domain locales
        filter = DomainLocale.domain_id == domain.id
        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, DomainLocale.locale_id.in_(id))
        q = self.sa_session.query(DomainLocale).filter(filter).all()
        if len(q) > 0:
            # always return Locale objects
            if availability is not None:
                return [dl.locale for dl in q if dl.availability & availability == availability]
            else:
                return [dl.locale for dl in q]

        return self.get_locales(id, availability)

    def save_or_update_domain_locales(self, domain_id, locales):
        """
        @param domain_id
        @param locales dict with locale_id-availability
        """
        self.assert_permissions('view_all_domains')
        domain = self.sa_session.query(Domain).filter(Domain.id==domain_id).one()

        locale_objs = self.get_locales(locales.keys())
        locale_ids = [l.id for l in locale_objs]

        domain_locale_ids = []
        # get all domain locales
        dl = self.sa_session.query(DomainLocale).filter(and_(DomainLocale.domain==domain)).all()
        for l in dl:
            # check what was removed
            if l.locale_id not in locale_ids:
                self.sa_session.delete(l)
            # what was updated
            else:
                l.availability = locales[str(l.locale_id)]
                self.sa_session.add(l)
                domain_locale_ids.append(l.locale_id)
        # and now what is new
        for l in locale_objs:
            if l.id not in domain_locale_ids:
                dl = DomainLocale()
                dl.locale = l
                dl.domain = domain
                dl.availability = locales[str(l.id)]
                self.sa_session.add(dl)

        self.sa_session.flush()

    def get_countries(self, id=None):
        """
        @param id: None, list or int
        @return: Country objects

        Returns a list with all countries in case id is None or a list of countries with supplied id/ids.
        (a list of one countrie in case a valid id is supplied)
        """

        if id == None:
            return self.sa_session.query(Country).all()

        if not isinstance(id, list):
            id = [id]

        return self.sa_session.query(Country).filter(Country.id.in_(id)).order_by(Country.name).all()


    def get_domain_countries(self, id=None, include_inactive=False):
        """
        returns all countries available for the current domain
        """
        domain = self._get_default_group().license.domain

        _filter = or_(DomainCountry.active == None, DomainCountry.active == 1)


        fo = outerjoin(Country, DomainCountry, onclause=(and_(Country.id == DomainCountry.country_id, DomainCountry.domain_id == domain.id)))
        q = self.sa_session.query(Country, DomainCountry).select_from(fo)
        if not include_inactive:
            q = q.filter(_filter)

        if id == None:
            return q.order_by(Country.name).all()

        if not isinstance(id, list):
            id = [id]

        return q.filter(Country.id.in_(id)).order_by(Country.name).all()

    @transaction_decorator
    def save_domain_country(self, country_id, local_name=False, active=None):
        self.assert_permissions('manage_domain_country_mapping')
        c, l = self.get_domain_countries(country_id, True).pop()

        if l is None:
            dc = DomainCountry()
            dc.country = c
            dc.domain = self.get_default_group().license.domain
            dc.active = True
            dc.name = None
        else:
            dc = l

        if active is not None:
            dc.active = active

        # if local name is the same as the original country name, set the field to None
        if c.name == local_name:
            dc.name = None
        elif local_name is not False:
            dc.name = local_name

        if l is not None and dc.active is True and dc.name is None:
            self.sa_session.delete(l)
        elif dc.active is False or dc.name is not None:
            self.sa_session.add(dc)

        self.sa_session.flush()

        return dc

    def to_locale(self, language_code):
        """
        @param language_code: string
        @return locale string

        Converts a language_code into a locale. Tries to get locale using django utils translation and in case that
        doesn't return a correct locale, it searches in the db (tblLocale) and returns the first find.
        """

        from django.utils import translation

        if len(language_code) == 2:
            from sqlalchemy import func
            l = self.sa_session.query(Locale).filter(func.substr(Locale.locale,1,2)==language_code)
            if l.count() == 0:
                raise Exception('Could not find a locale for language_code %s' % language_code)
            return l.all()[0].locale

        locale = translation.to_locale(language_code)
        l = self.sa_session.query(Locale).filter(Locale.locale==locale)
        if l.count() == 0:
            raise Exception('Could not find a locale for language_code %s' % language_code)

        return l.one().locale

    def get_language_by_alias(self, name):
        domain = self.get_default_group().license.domain
        alias = self.sa_session.query(DomainLanguage).filter(and_(DomainLanguage.name==name, DomainLanguage.domain==domain)).first()

        if alias is not None:
            return alias.language

        return self.sa_session.query(Language).filter(Language.name==name).first()

    @transaction_decorator
    def save_domain_language_alias(self, language_id, *aliases):
        self.assert_permissions('manage_domain_language_aliases')

        new_aliases = []
        # these aliases should not exist yet, or if they do the language id should match
        for alias in aliases:
            l = self.get_language_by_alias(alias)
            if l is None:
                new_aliases.append(alias)
                continue
            if l.id != language_id:
                raise PublicSecurityException(_('Alias already exists for another language: %(language)s' % {'language': l.name}))

        if len(new_aliases) == 0:
            return

        domain = self.get_default_group().license.domain

        for alias in new_aliases:

            dl = DomainLanguage()
            dl.domain = domain
            dl.language = self.sa_session.query(Language).filter(Language.id==language_id).one()
            dl.name = alias
            self.sa_session.add(dl)

        self.sa_session.flush()

    @transaction_decorator
    def remove_domain_language_aliases(self, language_id, *aliases):
        self.assert_permissions('manage_domain_language_aliases')

        for alias in aliases:
            self.sa_session.delete(alias)
        self.sa_session.flush()

    def get_language_aliases_by_language_id(self, language_id):
        domain = self.get_default_group().license.domain
        aliases = self.sa_session.query(DomainLanguage).filter(and_(DomainLanguage.language_id==language_id, DomainLanguage.domain==domain)).all()
        return aliases

    def get_language_by_id(self, language_id):
        return self.sa_session.query(Language).filter(Language.id==language_id).one()
