# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.
import operator

from twisted.python.reflect import namedAny

from basesoft.web.core.util.controllers import AbstractController
from basesoft.web.contrib.iconbar.models import IconBarIcon

class IconBarController(AbstractController):

    def __init__(self, sa_session, account_id=None, request=None):
        AbstractController.__init__(self, sa_session, account_id=account_id, request=request)
        self.icon_list = []

    def add_icons(self, icons):
        """
        Add an icon to the list of icons that will be used for generating the html code of the iconbar.
        If the list is empty, no icons will be added.

        @param icons: an IconBarIcon or a list of IconBarIcon
        """
        if not isinstance(icons, list):
            icons = [icons]

        for icon in icons:
            assert isinstance(icon, IconBarIcon)

            if icon.group is None:
                # the icon is not part of a stack, so we add it by itself
                self.icon_list.append({
                    "group": "",
                    "group_index": icon.icon_index,
                    "icons": [icon]
                })
                continue

            # the icon is part of a stack, so we check if it's the first icon added for that stack or not
            try:
                # try to find the position of the group of this icon
                index_position = [e['group'] for e in self.icon_list].index(icon.group)
                # we found the group, so add it to it's list of icons
                self.icon_list[index_position]['icons'].append(icon)
            except ValueError:
                # no icons were added yet with this group, so create a new one
                self.icon_list.append({
                    "group": icon.group,
                    # no group_index yet, it will be calculated when the icons are sorted
                    # this is important ! we check for None in the widget when we sort icons in each stack
                    "group_index": None,
                    "icons": [icon]
                })

    def get_grouped_icons(self):
        # sort each group of icons and set the group_index for each stacked group
        for i in range(len(self.icon_list)):
            g = self.icon_list[i]
            if g['group_index'] is None: # groups that have an group_index different then None contain only one icon
                g['icons'] = sorted(g['icons'], key=operator.attrgetter('icon_index')) # sort the icons based on index
                g['group_index'] = g['icons'][0].icon_index # set the index for the whole group as the first icon
                self.icon_list[i] = g

        # sort the list of groups according to the group_index
        self.icon_list.sort(key=operator.itemgetter('group_index'))

        # build the list of lists of icons for javascript
        rv = []
        for group in self.icon_list:
            rv.append([i.get_icon() for i in group['icons']])
        return rv

    def get_icons(self):
        from django.conf import settings
        # the list of applications that we check for icon modules
        apps = settings.INSTALLED_APPS

        # see if there is a icons module in the app
        for app in apps:
            try:
                icon_module = namedAny(app + '.icons')
            except AttributeError:
                # if we do not have an icons module then the app doesn't add icons
                #logger.debug('Exception "%s" in app %s when importing the icons module.' % (e, app) )
                continue

            if hasattr(icon_module, 'get_icons'):
                # this is the list of icons that the application defines
                icons_to_add = icon_module.get_icons()

            # we need to check if the icons returned have the proper permissions, in
            # order to decide if we want to add the icons to the bar or not
            for icon in icons_to_add:

                icon.request = self.request

                if icon.has_permissions(self):

                    # if a number_rpc_method is provided then an id is mandatory
                    if icon.number_rpc_method is not None and icon.id is None:
                        raise Exception("No icon id for icon %s" % icon.__class__)
                    icon.appforimage = app
                    self.add_icons(icon)

        return self.get_grouped_icons()

