# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.
from basesoft.web.core.exceptions import SecurityException, PublicSecurityException, IconException

class IconBarIcon:
    """
    One Icon for the iconbar.
    """

    # the html id attribute for the icon
    id = None # if not specified it won't be generated

    # the title of the icon (shown at hover)
    # if not specified, the "title" attribute won't be generated for the image
    title = None

    # the type of the icon (can be ['link', 'separator'])
    type = 'link'

    # the url for the click event
    href = None

    # the target window
    target = None

    # the path to the image that should be used for this icon
    image = None # mandatory

    # the js file that contains the code responsible for the behaviour
    js_file = None

    # the name of the js component that handles the behaviour
    component = None

    # the index of the icon, used for ordering. lowest index appears first
    icon_index = 1 # default 1, because home is 0 and we want it first

    # the group of the icon. icons having the same group will be stacked. if an icon doesn't have a group, it will
    # be displayed in the iconbar according to it's icon_index
    group = None

    # translatable group title
    group_title = None

    # when clicked 'app' tells the iq_bootloader to start that application.
    app = None
    # used for the icon image to determine the app where to load it from
    appforimage = None

    # attributes for the number icon
    # type of the number (determines the color of the number icon)
    # possible types: 'information' - blue, 'important' - red, 'warning' - yellow
    number_type = 'information'

    # rpc url where we check to see what value we should show
    number_rpc_url = None

    # rpc method used to check what value we should show. if this is set, a check will be made that an id is provided
    # also for this icon, for refresh purposes. the value returned by this method will be shown in the icon. if method
    # returns None, the icon will be hidden (until another update that might show a value)
    number_rpc_method = None

    # if the refresh value is provided (number of seconds), the rpc will check periodically for the value
    number_refresh = None

    # number value. if this is provided, it will override the rpc call and only this value is used. if this is set,
    # number_rpc_url, number_rpc_method and number_refresh are not taken into account
    number_value = None

    # if this value is true, the refresh is only done if the user is logged (for example, if user is autologged
    # out, the refresh is not done until the user logs back)
    number_rpc_login_required = False

    # this is a global event which gets fired when the number on the icon bar icon changes.
    number_rpc_global_change_event = None

    char_main = None
    char_sub = None
    css_class = None
    colorname = 'iconbar-empty'

    def get_icon(self):

        # either char_main can be set or image
        assert self.char_main is None or self.image is None

        return {
            'id': self.id,
            'type': self.type,
            'title': self.title,
            'href': self.href,
            'target': self.target,
            'image': self.image,
            'js_file': self.js_file,
            'component': self.component,
            'icon_index': self.icon_index,
            'group': self.group,
            'group_title': self.group if self.group_title is None else self.group_title,
            'app': self.app,
            'appforimage': self.appforimage,
            'number_type': self.number_type,
            'number_rpc_url': self.number_rpc_url,
            'number_rpc_method': self.number_rpc_method,
            'number_refresh': self.number_refresh,
            'number_value': self.number_value,
            'number_rpc_login_required': self.number_rpc_login_required,
            'number_rpc_global_change_event': self.number_rpc_global_change_event,

            'char_main': self.char_main,
            'char_sub': self.char_sub,
            'css_class': self.css_class,
            'colorname': self.colorname
        }

    def assert_permissions(self, *args):
        """
        This needs to be overwritten in children classes.
        """
        raise SecurityException('Icon doesn\'t have assert_permissions defined.')

    def has_permissions(self, controller):
        """
        Check if an icon has the needed permissions to be shown.

        @param request: the request that contains logged in users, etc.
        @return: bool True if icon has permissions, False if it doesn't.
        """
        try:
            self.assert_permissions(controller)
        except PublicSecurityException:
            return False
        except SecurityException:
            return False
        except IconException:
            return False
        except Exception, e:
            raise e
        return True

    def __str__(self):
        return '<IconBarIcon %s (id: %s)>' % (self.title, self.id)

    def __cmp__(self, other):
        """
        Used for sorting the list of icons in the iconbar, based on icon_index
        Eg. sorted_icons = my_list_of_icons.sort()
        """
        return cmp(self.icon_index, other.icon_index)
