# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from sqlalchemy.orm import interfaces
from basesoft.util.bitwise import get_bits

class MockupObjectHook(object):
    pass

class PrivilegeObjectMapper(interfaces.MapperExtension):
    """
    Our own mapper to add hooks to certain objects for specific domain/license/group.
    """
    __hooks = None
    __reload = False

    def __load_hook(self, instance):
        """
        Query information for a possible hook for the license/domain/group which owns the instance.
        """
        from basesoft.web.core.util.db import Session
        from basesoft.web.contrib.license.models import PrivilegeObjectHook
        from sqlalchemy.sql import and_, or_

        obj = instance.__class__
        object_name = obj.__name__

        session = Session()
        if self.__reload:
            instance = session.query(obj).filter(obj.id==instance.id).one()

        groups = []
        licenses = []
        domains = []
        # get all the attached groups (and it's licenses and domains)
        for g in instance.groups:
            if g.group.id not in groups: groups.append(g.group.id)
            if g.group.license.id not in licenses: licenses.append(g.group.license.id)
            if g.group.license.domain.id not in domains: domains.append(g.group.license.domain.id)

        # no groups attached? Nothing to do here.
        if len(groups) == 0:
            return

        # query hooks set for any of the groups, licenses, domains for this particular object
        q = session.query(PrivilegeObjectHook).filter(
            and_(PrivilegeObjectHook.object==object_name,
                 and_(
                      and_(
                           PrivilegeObjectHook.domain_id.in_(domains),
                           or_(PrivilegeObjectHook.license_id.in_(licenses),
                               PrivilegeObjectHook.license==None)),
                      or_(PrivilegeObjectHook.group_id.in_(groups),
                          PrivilegeObjectHook.group==None)
                 )
            )
        )
        self.__hooks = q.all()

    def __load_backwards_compatibility_hook(self, instance):
        """
        Backwards compatibility mode for old privilege setup.
        """
        hook = MockupObjectHook()
        # we used to have an "owner" group on the instance itself. If it's not there, it's using the new setup
        if not hasattr(instance, 'group'):
            return

        # Log which object is still using the old setup, look in it's controller to fix.
        from basesoft import logging
        logger = logging.getLogger("basesoft.core")
        logger.warning('Using backwards compat mode for object hook on %s' % instance)

        hook.target_group = instance.group
        hook.group = instance.group

        # possible attributes to be converted to a hook setup.
        for attr in ['group_privileges','license_privileges','domain_privileges','world_privileges']:
            if hasattr(instance, attr):
                value = getattr(instance, attr)
                if type(value) == tuple:
                    value = sum(value)
                setattr(hook, attr, value)

        if self.__hooks is None:
            self.__hooks = []
        self.__hooks.append(hook)

    def after_insert(self, mapper, connection, instance):
        """
        After we insert the Privilege object we have an ID and we can attach any groups defined in the hooks.
        """
        # TODO: when we are done converting old privilege objects remove this.
        self.__load_backwards_compatibility_hook(instance)
        self.__apply_hooks(mapper, connection, instance)

        self.__load_hook(instance)
        self.__apply_hooks(mapper, connection, instance)

        self.__reload = False

    def __apply_hooks(self, mapper, connection, instance):

        # no hooks, nothing to do.
        if self.__hooks is None:
            return

        from sqlalchemy.orm import object_mapper
        go_def, field = instance.get_group_object()

        # loop the hooks that we queried.
        for hook in self.__hooks:
            group_privs = []
            license_privs = []
            domain_privs = []
            world_privs = []
            # set group, license and domain privs.
            if hook.group_privileges is not None:
                group_privs = get_bits(hook.group_privileges)
            if hook.license_privileges is not None:
                license_privs = get_bits(hook.license_privileges)
            if hook.domain_privileges is not None:
                domain_privs = get_bits(hook.domain_privileges)
            if hasattr(hook, 'world_privileges') and hook.world_privileges is not None:
                world_privs = get_bits(hook.world_privileges)
            used_privs = group_privs
            [used_privs.append(p) for p in license_privs if p not in used_privs]
            [used_privs.append(p) for p in domain_privs if p not in used_privs]
            [used_privs.append(p) for p in world_privs if p not in used_privs]

            for go in instance.groups:
                if go.privilege not in used_privs or go.group != hook.target_group:
                    continue
                if go.privilege in group_privs:
                    go.group_has_privilege = True
                if go.privilege in license_privs:
                    go.license_has_privilege = True
                if go.privilege in domain_privs:
                    go.domain_has_privilege = True
                if go.privilege in world_privs:
                    go.world_has_privilege = True
                # TODO: add to session if possible?
                used_privs = [p for p in used_privs if p != go.privilege]
            if len(used_privs) == 0:
                continue

            # We still have some privs that are unsaved. Add new *Groups for those.
            for p in used_privs:
                go = go_def()
                values = {
                    '%s_id' % field: instance.id,
                    'group_id': hook.target_group.id,
                    'privilege': p
                }
                if p in group_privs:
                    values['group_has_privilege'] = True
                if p in license_privs:
                    values['license_has_privilege'] = True
                if p in domain_privs:
                    values['domain_has_privilege'] = True
                if p in world_privs:
                    values['world_has_privilege'] = True
                connection.execute(object_mapper(go).local_table.insert().values(**values))
                self.__reload = True

        self.__hooks = None
