# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.
import datetime

from basesoft.logging import getLogger
from django.conf import settings
from django.utils.translation import ugettext_lazy as _, ugettext
from sqlalchemy.orm.session import Session

from basesoft.util import strings
from basesoft.util.bitwise import get_bits
from basesoft.web.core.exceptions import PublicUserException, SecurityException


logger = getLogger('basesosft.web.contrib.license')

"""
models for license manager app
"""
class PermissionCategory(object):
    """
    This model represents a PermissionCategory.
    @DynamicAttrs
    """
    def __init__(self, name):
        self.name = name.strip()

    def __repr__(self):
        return '<PermissionCategory id=%s, name="%s">' % (self.id, self.name)

class Permission(object):
    """
    This model represents a permission
    @DynamicAttrs
    """
    def __init__(self, name=None, description=None, title=None):
        if name is None:
            self.name = None
        else:
            self.name = name.strip()

        if description is None:
            self.description = None
        else:
            self.description = strings.trim_multiline(description, True) # trim also the first/last lines if empty

        self.title = title

    def __repr__(self):
        return '<Permission id=%s, name="%s">' % (self.id, self.name)

class Domain(object):
    """
    This model represents a domain.
    @DynamicAttrs
    """

    # settings for the domain
    SETTING_SINGLE_LOGIN = 1 # if set, the users of this domain can be logged in only in one place at a time
    SETTING_REQUIRE_PERSON_DETAILS_ON_ACCOUNT_ACTIVATION = 2 # if set account activation can only be done when person details are provided.
    SETTING_REQUIRE_EULA_ON_REGISTRATION = 4 # if set an EULA is presented on registration and has to be accepted.

    # status for the domain
    STATUS_INACTIVE = 0
    STATUS_ACTIVE = 1
    STATUS_TEMP_DISABLED_DUE_TO_UPDATE = 2

    def has_setting(self, setting):
        """
        Returns True if the domain has the given setting, false otherwise
        @param setting: one of the SETTING_* constants
        @return: boolean
        """
        return setting in get_bits(self.settings)

    def get_dict(self):
        return {
            'domain_name': self.name,
            'single_login': self.has_setting(Domain.SETTING_SINGLE_LOGIN),
            'require_person_details': self.has_setting(Domain.SETTING_REQUIRE_PERSON_DETAILS_ON_ACCOUNT_ACTIVATION),
            'status': self.status
        }

    def is_active(self):
        if self.status == Domain.STATUS_ACTIVE:
            return True
        if self.status == Domain.STATUS_INACTIVE:
            raise PublicUserException(ugettext('The Domain your user account is linked to is Inactive.'))
        if self.status == Domain.STATUS_TEMP_DISABLED_DUE_TO_UPDATE:
            raise PublicUserException(ugettext('This action is temporarily disabled due to an update. Please try again later.'))

    def __get_default_license(self):
        from sqlalchemy import and_
        session = Session.object_session(self)
        return session.query(License).filter(and_(License.domain_id==self.id, License.is_default==True)).one()

    default_license = property(__get_default_license)

    def __get_default_group(self):
        from sqlalchemy import and_
        session = Session.object_session(self)
        return session.query(Group).join(License).filter(and_(License.domain_id==self.id, License.is_default==True,
                                                              Group.is_default==True)).one()
    default_group = property(__get_default_group)

    # Virtual date implementation per DOMAIN. This can only be used on instances which support it.
    def today(self):
        if settings.VIRTUAL_DATE is None:
            return datetime.datetime.now().date()

        if self.virtual_date is None:
            return datetime.datetime.now().date()

        return self.virtual_date

    def now(self):
        t = self.today()
        n = datetime.datetime.now()
        return datetime.datetime(t.year, t.month, t.day, n.hour, n.minute, n.second)

    def set_date(self, new_date):
        assert isinstance(new_date, datetime.date)
        if settings.VIRTUAL_DATE is None:
            raise SecurityException('Trying to set virtual date on an unsupported instance.')
        today = self.today()
        if self.virtual_date is not None and new_date < today:
            raise PublicUserException(ugettext('Cannot go backwards in time'))
        session = Session.object_session(self)
        self.virtual_date = new_date
        session.add(self)
        session.flush()
        logger.info('virtual date changed to %s' % new_date)

    def __repr__(self):
        return '<Domain id=%s, name="%s">' % (self.id, self.name)

class DomainGroup(object):
    """
    This model represents the link between domains and groups.
    @DynamicAttrs
    """

    def __repr__(self):
        return '<DomainGroup domain="%s", group="%s">' % (self.domain.name, self.group.name)

class License(object):
    """
    This model represents a license
    @DynamicAttrs
    """

    NAME_I18N = _("License")
    REQUIRED_PERMISSIONS = ['view_license']
    
    def __get_default_group(self):
        from sqlalchemy import and_
        session = Session.object_session(self)
        return session.query(Group).filter(and_(Group.license_id==self.id,Group.is_default==True)).one()

    default_group = property(__get_default_group)

    def __get_controller(self):
        from basesoft.web.contrib.license.controllers import LicenseController
        return LicenseController

    controller = property(__get_controller)

    def get_config(self, config):
        from basesoft.web.core.util.controllers import LicenseConfigController
        session = Session.object_session(self)

        lcc = LicenseConfigController(session, scope=self)
        return lcc.get(config)

    def set_config(self, config, value):
        from basesoft.web.core.util.controllers import LicenseConfigController
        session = Session.object_session(self)

        lcc = LicenseConfigController(session, scope=self)
        lcc.set(config, value)

    def __repr__(self):
        return '<License id=%s, name=%s>' % (self.id, self.name)

class LicenseConfig(object):
    """
    @DynamicAttrs
    """
    def __repr__(self):
        return '<LicenseConfig>'

class Reseller(License):
    """
    Model for a reseller license, this inherits from License.
    @DynamicAttrs
    """

    NAME_I18N = _("Reseller license")
    REQUIRED_PERMISSIONS = ['view_license', 'view_all_domains']

    def __repr__(self):
        return '<Reseller id=%s>' % self.id

class LicensePermission(object):
    """
    This model represents a license and permission link
    @DynamicAttrs
    """
    def __repr__(self):
        return '<LicensePermission license=%s, permission=%s>' \
                % (self.license, self.permission)

class Group(object):
    """
    This model represents a group
    @DynamicAttrs
    """
    DEFAULT_GROUP_NAME = 'administrators'

    def __repr__(self):
        return '<Group id=%s, name="%s">' % (self.id, self.name)

class GroupPermission(object):
    """
    This model represents a group and permission link
    @DynamicAttrs
    """

class UsageOption(object):
    """
    This model represents a usage option
    @DynamicAttrs
    """

    def __init__(self, name=None, description=None, title=None):
        if name is None:
            self.name = None
        else:
            self.name = name.strip()

        if description is None:
            self.description = None
        else:
            self.description = strings.trim_multiline(description, True) # trim also the first/last lines if empty

        self.title = title

    def __repr__(self):
        return '<UsageOption id=%s, name="%s">' % (self.id, self.name)

class LicenseUsage(object):
    """
    This model represents a license and usage option link
    @DynamicAttrs
    """
    def __repr__(self):
        return '<LicenseUsage license=%s, usage_option=%s, credits="%s">' \
                % (self.license, self.usage_option, self.credits)

class PrivilegeObjectHook(object):
    """
    @DynamicAttrs
    """
    def __repr__(self):
        return '<PrivilegeObjectHook>'

class Role(object):
    """
    @DynamicAttrs
    """
    ROLE_TYPE_MANUAL = 0
    ROLE_TYPE_AUTOMATIC = 1

class RolePermission(object):
    """
    @DynamicAttrs
    """
    pass

class RoleConfig(object):
    """
    @DynamicAttrs
    """
    def set_config(self, name, v, session=None, is_license=True):
        from basesoft.web.contrib.account.models import Config
        if session is None:
            session = Session.object_session(self)
        self.config = session.query(Config).filter(Config.name==name).one()
        self.value = v
        self.is_license = is_license

    def __repr__(self):
        return '<RoleConfig role=%s config=%s value=%s is_license=%s>' % (self.role, self.config, self.value, self.is_license)

