# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

from django.conf import settings

from basesoft.web.contrib.license.controllers import LicenseController, UniqueRoleName
from basesoft.web.contrib.license.models import Role
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.input.validation import plugins
from twisted.python.reflect import namedAny
from basesoft.web.contrib.branding.controllers import LicenseThemeController
from basesoft.web.core.exceptions import PublicSystemException, PublicSecurityException
from basesoft.util.dict import dict_merge
from basesoft.web.contrib.contact.models import Address, BankAccount
from basesoft.util.ordered_dict import OrderedDict
from basesoft.web.contrib.dashboard.models import Dashboard


def get_configs():
    configs = {}
    for app in settings.INSTALLED_APPS:
        try:
            sa_module = namedAny(app+'.config')
        except AttributeError:
            continue
        if not hasattr(sa_module, 'license_configs'):
            continue
        for l, c in sa_module.license_configs.iteritems():
            if not configs.has_key(l):
                configs[l] = []
            configs[l] += c
    return configs

class LicenseRPCHandler:
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.user_id = request.session['user_id']
        self.request = request
        self.controller = LicenseController(self.sa_session, account_id=self.user_id, request=request)

    def get_domain_settings(self, domain_id):
        return self.controller.get_domains(domain_id).pop().get_dict()

    def save_domain_settings(self, domain_id, args):
        self.controller.save_or_update_domain(domain_id, **args)
        return True

    def get_configs(self, polymorphic_type, license_id=None):
        configs = get_configs()

        if not configs.has_key(polymorphic_type):
            return []

        lc = configs[polymorphic_type]
        return [c.get_config(self.sa_session) for c in lc if c.has_permissions(self.controller)]

    def validate_license_configs(self, values, args):
        configs = get_configs()

        polymorphic_type = args[0]

        # dealing with a polymorphic type without configs
        if not configs.has_key(polymorphic_type):
            return True

        lc = configs[polymorphic_type]
        validators = {}
        for c in lc:
            validators = c.get_form_behaviour(validators, values)

        if len(validators) == 0:
            return True

        form_handler = FormValidator(validators)

        return form_handler.validate(values)

    def permission_toggle_disable_login(self, _id):
        self.controller.permission_toggle_disable_login(_id)
        return True

    def update_license_configs(self, licenses):
        configs = get_configs()
        for l in licenses:
            self.controller.save_license_configs(configs, l['values'], l['license_id'])
        return True

    def update_active_state(self, active, license_id):
        self.controller.update_license_active_state(license_id, active)
        return True

    def get_role(self, id):
        role = self.controller.get_roles(id=id).pop()
        return {
            'name': role.name,
            'permissions': [p.id for p in role.permissions],
            'role_type': role.role_type,
            'dashboard': None if role.dashboard is None else role.dashboard.id,
            'dashboard_view_privs': None if role.dashboard_privileges is None else role.dashboard_privileges & Dashboard.PRIV_VIEW == Dashboard.PRIV_VIEW,
            'dashboard_update_privs': None if role.dashboard_privileges is None else role.dashboard_privileges & Dashboard.PRIV_UPDATE == Dashboard.PRIV_UPDATE,
        }

    def get_domains_licenses_and_groups(self, domains=True, licenses=True, groups=True):
        rv = []
        if domains:
            domains = self.controller.get_domains()
            rv.append(('Domains', [{'label': d.name, 'value': 'domain_%s' % d.id} for d in domains]))
        if licenses:
            licenses = self.controller.get_domain_licenses()
            rv.append(('Licenses', [{'label': l.name, 'value': 'license_%s' % l.id} for l in licenses]))

        # TODO: groups

        return OrderedDict(rv)

    def get_available_groups_for_account(self):
        groups = {}
        for g in self.controller.get_groups():
            if not groups.has_key(g.license.name):
                groups[g.license.name] = []
            groups[g.license.name].append({'label': g.name, 'value': g.id})

        rv = []
        for _license, groups in groups.iteritems():
            rv.append({
                'id': 'license',
                'label': _license,
                'options': groups
            })

        return sorted(rv, key=lambda x: x['label'])

    def get_groups(self, domain_id=None, license_id=None, group_id=None):
        return self.controller.get_domain_license_and_groups(domain_id, license_id, group_id)

    def get_licenses_per_domain(self):
        rv = []
        domains = self.controller.get_domains()
        for d in domains:
            rv.append((d.name, [{'label': l.name, 'value': l.id} for l in d.licenses]))

        return OrderedDict(rv)

    def get_licenses_for_master_data_change(self):
        licenses = self.controller.get_account_licenses_with_permssions('update_license_masterdata')
        rv = []
        for _license in licenses:
            lc = _license.controller(self.sa_session, account_id=self.user_id, request=self.request)
            rv.append({'id': _license.id, 'state': lc.load_master_data_state(_license)})

        return rv

    def get_license_master_data_state(self, _id):
        licenses = self.controller.get_account_licenses_with_permssions('update_license_masterdata')
        _license = [l for l in licenses if l.id==_id]
        if len(_license) == 0:
            raise PublicSecurityException('No access to this license.')

        _license = _license.pop()
        lc = _license.controller(self.sa_session, account_id=self.user_id, request=self.request)
        return lc.load_master_data_state(_license)

    def get_master_data_comparison(self, id):

        # this should also get the translations of the fields.

        license = self.controller.get_licenses(id).pop()
        ic = self.controller.get_instance_class()

        original = self.controller.load_master_data_state(id, True)

        temp = self.controller.load_master_data_state(id)
        tmp = {}
        for i in temp:
            tmp['temp_%s' % i] = temp[i]

        rv = dict_merge(original, tmp)
        rv['license_name'] = license.name
        rv['company_translations'] = ic.company_instance_class.cols_i18n
        rv['contact_person_data_translations'] = ic.person_instance_class.cols_i18n
        rv['invoice_person_data_translations'] = ic.person_instance_class.cols_i18n
        rv['address_data_translations'] = Address.cols_i18n.copy()
        rv['address_data_translations']['country_name'] = 'Country'
        rv['bank_data_translations'] = BankAccount.cols_i18n
        return rv

    def reject_master_data_change(self, id, reason):
        self.controller.reject_master_data_change(id, reason)

    def approve_master_data_change(self, id):
        self.controller.approve_master_data_change(id)

    def save_license_master_data_state(self, id, state):
        if id is None:
            raise PublicSystemException('Incorrect usage of master data change.')
        self.controller.update_master_data(id, state)
        return id

    def get_roles(self):
        rv = []

        for r in self.controller.get_roles():
            rv.append({'label': r.name, 'value': r.id})

        return rv

    def save_role(self, values):

        if values.has_key('id'):
            values['role_type'] = None
        else:
            values['role_type'] = Role.ROLE_TYPE_MANUAL
        values['permissions'] = self.controller.get_permissions(values['permissions'])
        values['dashboard_privileges'] = None
        if values.has_key('dashboard_view_privs') and values['dashboard_view_privs']:
            values['dashboard_privileges'] = Dashboard.PRIV_VIEW
            values['dashboard_privileges'] |= 0 if not values['dashboard_update_privs'] else Dashboard.PRIV_UPDATE
        if values.has_key('dashboard_view_privs'):
            del values['dashboard_view_privs']
        if values.has_key('dashboard_update_privs'):
            del values['dashboard_update_privs']

        self.controller.save_or_update_role(**values)
        return True

    def validate_role_values(self, values):
        form_handler = FormValidator({
            'name': [(plugins.Required, {})],
            'permissions': [(plugins.Required, {})],
        })

        return form_handler.validate(values)

    def load_role(self, _id):
        return self.controller.load_role(_id)

    def apply_role(self, _id, state):
        return self.controller.apply_role(_id, **state)

class ThemesRPCHandler:
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.user_id = request.session['user_id']
        self.ltc = LicenseThemeController(self.sa_session, account_id=self.user_id, request=request)

    def get_available_themes(self):
        """
        Get the list of permissions that are available for the current user based on his license
        """
        available_themes = [{'label': e.name, 'value': e.module_name} for e in self.ltc.get_available_themes()]
        available_themes_values = [e['value'] for e in available_themes]
        default_value = self.ltc.config.get('basesoft.web.core.media.themes')
        #if default_value not in available_themes_values:
        #    default_value = self.ltc.get_default_theme_module()

        options = []
        for at in available_themes:
            opt = at
            opt['default'] = opt['value'] == default_value
            options.append(opt)

        return options

handlers = (ThemesRPCHandler, LicenseRPCHandler, )
