# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

"""
Contains the objects, functions and variables to setup the IQPortal database for license app,
and to view information from the data models stored within it

"""
import datetime
from sqlalchemy.orm import mapper, relation
from sqlalchemy import \
    Column, Integer, String, ForeignKey, Boolean, Date, \
    DECIMAL, DateTime, UnicodeText, SmallInteger, Unicode
from basesoft.sqlalchemy.table import IQTable
from basesoft.web.contrib.contact import models as contact_models
from basesoft.web.contrib.license import models as license_models
from django.utils.translation import ugettext as _

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    """
    """
    return tables

def do_tables(metadata):
    global __tables_defined
    if not __tables_defined:
        tblPermissionCategory = IQTable('tblPermissionCategory', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('name', String(40), nullable=False, unique=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblPermission = IQTable('tblPermission', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            # Permissions are unique!
            Column('name', String(40), nullable=False, unique=True),
            Column('description', UnicodeText),

            # translated string to be used in the admin interfaces (shorter than description). this is the english value
            Column('title', UnicodeText, nullable=True),

            # category name of the permission, based on each application they are defined for
            Column('category_id', Integer, ForeignKey('tblPermissionCategory.id'), nullable=True),

            Column('disable_login', Boolean, default=False, nullable=False, server_default='0'),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblDomain = IQTable('tblDomain', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('name', String(40), nullable=False, unique=True),
            #do not emit this next key as part of the CREATE TABLE definition.
            #Instead, use ALTER TABLE after table creation to add the key
            Column('reseller_id', Integer, ForeignKey('tblLicense.id', use_alter=True, name="tblDomainFK"), nullable=True),
            Column('registrant_role_id', Integer, ForeignKey('tblRole.id', ondelete='CASCADE'), nullable=True),
            Column('settings', Integer, default=0, nullable=False), # bitwise settings. see Domain.SETTING_*
            Column('status', Integer, default=1, nullable=False),
            Column('http_host', String(255), nullable=True),
            Column('virtual_date', Date(), nullable=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblLicense = IQTable('tblLicense', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('name', String(255), nullable=False),
            Column('domain_id', Integer, ForeignKey('tblDomain.id'), nullable=False),
            Column('role_id', Integer, ForeignKey('tblRole.id'), nullable=True),
            Column('is_default', SmallInteger, default=0, nullable=False),
            Column('polymorphic_type', String(255), nullable=False, default="License"),
            Column('company_id', Integer, ForeignKey('tblCompany.id'), nullable=True),
            Column('temp_company_id', Integer, ForeignKey('tblCompany.id'), nullable=True),
            Column('http_host', String(255), nullable=True),
            Column('self_branded', SmallInteger, default=0, nullable=False),
            Column('expiration_date', DateTime(), nullable=True),
            Column('is_active', SmallInteger, nullable=False, default=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblReseller = IQTable('tblReseller', metadata,
            Column('license_id', Integer, ForeignKey('tblLicense.id', ondelete='CASCADE'), primary_key=True),
        )

        tblLicenseConfig = IQTable('tblLicenseConfig', metadata,
            Column('license_id', Integer, ForeignKey('tblLicense.id', ondelete='CASCADE'), primary_key=True),
            Column('config_id', Integer, ForeignKey('tblConfig.id', ondelete='CASCADE'), primary_key=True),
            Column('value', String(40), nullable=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblLicensePermission = IQTable('tblLicensePermission', metadata,
            Column('license_id', Integer,
                   ForeignKey('tblLicense.id', ondelete='CASCADE'),
                   primary_key=True),
            Column('permission_id', Integer,
                   ForeignKey('tblPermission.id', ondelete='CASCADE'),
                   primary_key=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblGroup = IQTable('tblGroup', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('name', String(40), nullable=False),
            Column('license_id', Integer,
                   ForeignKey('tblLicense.id', ondelete='CASCADE')),
            Column('is_default', SmallInteger, default=0, nullable=False),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblGroupPermission = IQTable('tblGroupPermission', metadata,
            Column('group_id', Integer,
                   ForeignKey('tblGroup.id', ondelete='CASCADE'),
                   primary_key=True),
            Column('permission_id', Integer,
                   ForeignKey('tblPermission.id', ondelete='CASCADE'),
                   primary_key=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblUsageOption = IQTable('tblUsageOption', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('name', String(40), nullable=False, unique=True),
            Column('description', UnicodeText),

            # translated string to be used in the admin interfaces (shorter than description). this is the english value
            Column('title', UnicodeText, nullable=True),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblLicenseUsage = IQTable('tblLicenseUsage', metadata,
            Column('license_id', Integer,
                   ForeignKey('tblLicense.id', ondelete='CASCADE'),
                   primary_key=True),
            Column('usage_option_id', Integer,
                   ForeignKey('tblUsageOption.id', ondelete='CASCADE'),
                   primary_key=True),
            Column('credits', Integer, default=0, nullable=False),
            Column('price_per_unit', DECIMAL(10,2), nullable=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblDomainGroup = IQTable('tblDomainGroup', metadata,
            Column('domain_id', Integer,
                   ForeignKey('tblDomain.id', ondelete='CASCADE'),
                   primary_key=True),
            Column('group_id', Integer,
                   ForeignKey('tblGroup.id', ondelete='CASCADE'),
                   primary_key=True),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblPrivilegeObjectHook = IQTable('tblPrivilegeObjectHook', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('domain_id', Integer, ForeignKey('tblDomain.id', ondelete='CASCADE'), nullable=False),
            Column('license_id', Integer, ForeignKey('tblLicense.id', ondelete='CASCADE'), nullable=True),
            Column('group_id', Integer, ForeignKey('tblGroup.id', ondelete='CASCADE'), nullable=True),

            Column('object', String(255), nullable=False),
            Column('domain_privileges', Integer, nullable=True),
            Column('license_privileges', Integer, nullable=True),
            Column('group_privileges', Integer, nullable=True),

            Column('target_group_id', Integer, ForeignKey('tblGroup.id', ondelete='CASCADE'), nullable=True),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )


        tblRole = IQTable('tblRole', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('role_type', SmallInteger, nullable=False, default=license_models.Role.ROLE_TYPE_MANUAL),
            Column('dashboard_id', Integer, ForeignKey('tblDashboard.id', ondelete='RESTRICT'), nullable=True),
            Column('dashboard_privileges', Integer, nullable=True),
            Column('name', Unicode(255), nullable=False, unique=True),
            Column('required_permission', String(255), nullable=False),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
        )

        tblRolePermission = IQTable('tblRolePermission', metadata,
            Column('role_id', Integer, ForeignKey('tblRole.id', ondelete='CASCADE'), primary_key=True),
            Column('permission_id', Integer, ForeignKey('tblPermission.id', ondelete='CASCADE'), primary_key=True),
        )

        tblRoleConfig = IQTable('tblRoleConfig', metadata,
            Column('role_id', Integer, ForeignKey('tblRole.id', ondelete='CASCADE'), primary_key=True),
            Column('config_id', Integer, ForeignKey('tblConfig.id', ondelete='CASCADE'), primary_key=True),
            Column('is_license', Boolean, nullable=False, default=1),
            Column('value', Unicode(255), nullable=False),
        )

        tables['tblPermissionCategory'] = tblPermissionCategory
        tables['tblPermission'] = tblPermission
        tables['tblDomain'] = tblDomain
        tables['tblLicense'] = tblLicense
        tables['tblReseller'] = tblReseller
        tables['tblLicenseConfig'] = tblLicenseConfig
        tables['tblLicensePermission'] = tblLicensePermission
        tables['tblGroup'] = tblGroup
        tables['tblGroupPermission'] = tblGroupPermission
        tables['tblUsageOption'] = tblUsageOption
        tables['tblLicenseUsage'] = tblLicenseUsage
        tables['tblDomainGroup'] = tblDomainGroup
        tables['tblPrivilegeObjectHook'] = tblPrivilegeObjectHook

        tables['tblRole'] = tblRole
        tables['tblRolePermission'] = tblRolePermission
        tables['tblRoleConfig'] = tblRoleConfig

        __tables_defined = True


def do_mappings():
    """
    This function is only called by the django middleware
    when it is needed
    """
    global __mapping_defined
    if not __mapping_defined:
        from basesoft.web.contrib.account import models as account_models
        from basesoft.web.contrib.dashboard import models as dashboard_models

        mapper(license_models.PermissionCategory, tables['tblPermissionCategory'])
        mapper(license_models.Permission, tables['tblPermission'],
               properties = {
                'category': relation(license_models.PermissionCategory),
            })

        domain_mapper = mapper(license_models.Domain, tables['tblDomain'],
            properties = {
                'registrant_role': relation(license_models.Role),
            })

        license_mapper = mapper(license_models.License, tables['tblLicense'],
               properties = {
                        'company': relation(contact_models.Company,
                            primaryjoin=tables['tblLicense'].c.company_id==contact_models.Company.id
                        ),
                        'temp_company': relation(contact_models.Company,
                            primaryjoin=tables['tblLicense'].c.temp_company_id==contact_models.Company.id
                        ),
                        'permissions': relation(license_models.Permission,
                                                secondary=tables['tblLicensePermission']),
                        'role': relation(license_models.Role)
                        },
               polymorphic_on=tables['tblLicense'].c.polymorphic_type,
               polymorphic_identity='License'
            )
        # some properties of the License and Domain mappers depend upon columns present on an as-yet undefined class
        # so need to add those properties after the correspondent classes have been created
        domain_mapper.add_property('reseller', relation(license_models.License,
                                                  primaryjoin=license_models.License.id==license_models.Domain.reseller_id))
        license_mapper.add_property('domain', relation(license_models.Domain, backref='licenses',
                                                  primaryjoin=license_models.License.domain_id==license_models.Domain.id))

        mapper(license_models.LicensePermission, tables['tblLicensePermission'],
               properties = {
                        'license': relation(license_models.License),
                        'permission': relation(license_models.Permission),
                        }
            )
        mapper(license_models.Reseller, tables['tblReseller'],
            polymorphic_identity='Reseller',
            inherits=license_models.License
        )

        mapper(license_models.LicenseConfig, tables['tblLicenseConfig'],
            properties = {
                'license': relation(license_models.License, backref="configs"),
                'config': relation(account_models.Config)
            }
        )
        mapper(license_models.Group, tables['tblGroup'],
               properties = {
                        'license': relation(license_models.License, backref='groups'),
                        'permissions': relation(license_models.Permission,
                                                secondary=tables['tblGroupPermission']),
                        'domains': relation(license_models.Domain,
                                            secondary=tables['tblDomainGroup'])
                        }
            )
        mapper(license_models.GroupPermission, tables['tblGroupPermission'])
        mapper(license_models.DomainGroup, tables['tblDomainGroup'])
        mapper(license_models.UsageOption, tables['tblUsageOption'])
        mapper(license_models.LicenseUsage, tables['tblLicenseUsage'],
               properties = {
                        'license': relation(license_models.License),
                        'usage_option': relation(license_models.UsageOption),
                        }
            )

        mapper(license_models.PrivilegeObjectHook, tables['tblPrivilegeObjectHook'],
            properties = {
                'domain': relation(license_models.Domain),
                'license': relation(license_models.License),
                'group': relation(license_models.Group,
                                  primaryjoin=tables['tblPrivilegeObjectHook'].c.group_id==tables['tblGroup'].c.id),
                'target_group': relation(license_models.Group,
                                         primaryjoin=tables['tblPrivilegeObjectHook'].c.target_group_id==tables['tblGroup'].c.id)
            }
        )


        mapper(license_models.Role, tables['tblRole'],
            properties = {
                'permissions' : relation(license_models.Permission, secondary=tables['tblRolePermission']),
                'dashboard': relation(dashboard_models.Dashboard, backref='roles'),
            }
        )

        mapper(license_models.RolePermission, tables['tblRolePermission'],
            properties = {
                'role' : relation(license_models.Role),
                'permission': relation(license_models.Permission),
            }
        )

        mapper(license_models.RoleConfig, tables['tblRoleConfig'],
            properties = {
                'role' : relation(license_models.Role, backref='configs'),
                'config' : relation(account_models.Config),
            }
        )

        __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'License')

def app_permissions():
    """
    The permissions used within this application. Can be sync'd with the
    `iqsyncpermissions` command.

    Titles are with gettext calls, but in the database they appear in English. We need it like this so we get the
    titles in the language files. Later, in the js interfaces we show them translated.
    """
    return [
            license_models.Permission(u'distribute_group_permissions', u"""
                This gives the right to grant/revoke Permissions to Groups.
                """, _("Grant/revoke group permissions")),
            license_models.Permission(u'distribute_license_permissions', u"""
                This gives the right to grant/revoke Permissions and
                UsageOptions to Licenses.
                """, _("Grant/revoke license permissions and usage options")),

            license_models.Permission(u'view_license', u"""
                This gives the right to view License.
                """, _("View license")),
            license_models.Permission(u'create_license', u"""
                This gives the right to create a License.
                """, _("Create license")),
            license_models.Permission(u'update_license', u"""
                This gives the right to update a License.
                """, _("Update license")),

            # Used to update the company linked to this license. Requires approval.
            license_models.Permission(u'update_license_masterdata', u"""
                This gives the rights to update the license master data.
                """, _('Update license master data')),
            license_models.Permission(u'mark_license_as_deleted', u"""
                This gives the right to mark a License as deleted.
                """, _("Mark license as deleted")),
            license_models.Permission(u'delete_license', u"""
                This gives the right to delete a License.
                """, _("Delete license")),
            license_models.Permission(u'manage_license_configs', u"""
                This gives the rights to manage the Roles and their Permissions.
                """, _("Manage license configs")),
            license_models.Permission(u'license_masterdata_change_approval', u"""
                This gives the rights to approve or reject changes made to the master data.
                """, _('Master data change approval')),

            # do we still need this
            license_models.Permission(u'hook_manager', u"""
                Rights on the Hook Manager interface.
                """, _('Access to the hook manager interface')),

             license_models.Permission(u'view_domain', u"""
                This gives the right to view Domain.
                """, _("View domain")),
            license_models.Permission(u'create_domain', u"""
                This gives the right to create a Domain.
                """, _("Create domain")),
            license_models.Permission(u'update_domain', u"""
                This gives the right to update a Domain.
                """, _("Update domain")),
            license_models.Permission(u'mark_domain_as_deleted', u"""
                This gives the right to mark a Domain as deleted.
                """, _("Mark domain as deleted")),
            license_models.Permission(u'delete_domain', u"""
                This gives the right to delete a Domain.
                """, _("Delete domain")),

            license_models.Permission(u'view_all_domains', u"""
                This gives rights to view all Domains within the
                system. This is used for Sysadmin rights vs. Resellers.
                Reseller will only have access to Domains linked to
                them. Service Manager have rights to domains linked to their
                respective groups.
                """, _("View all domains")),
            license_models.Permission(u'view_all_permissions',u"""
                This gives rights to view all Permissions within the
                system. This is used for Sysadmin rights vs. Resellers.
                Reseller will only have access to Permissions defined in
                Product Templates to which they have rights.
                """, _("View all permissions")),
            license_models.Permission(u'view_all_usage_options',u"""
                This gives rights to view all Usage Options within the
                system. This is used for Sysadmin rights vs. Resellers.
                Reseller will only have access to Usage Options defined in
                Product Templates to which they have rights.
                """, _("View all usage options")),
            license_models.Permission(u'privilege_object_administrator', u"""
                This gives rights to manage ALL privilege objects.
                """, _("Privilege object administrator")),

            license_models.Permission(u'manage_roles', u"""
                This gives the rights to manage the Roles and their Permissions.
                """, _("Manage roles")),

            license_models.Permission(u'create_roles', u"""
                This gives the rights to create Roles and their Permissions.
                """, _("Create roles")),

            license_models.Permission(u'apply_roles', u"""
                This gives the rights to apply Roles to Groups and Licenses.
                """, _("Apply roles")),

            license_models.Permission(u'manage_iqportal_roles', u"""
                This gives the rights to manage the Roles IQPortal roles.
                """, _("Manage IQPortal roles")),

            license_models.Permission(u'manage_license_company_logo', u"""
                This gives the rights to add or change the license logo.
                """, _('Manage license logo')),
        ]

def app_configs():
    from basesoft.web.contrib.account import models as account_models

    configs = []

    config = account_models.Config()
    config.name = 'basesoft.license.company.logo'
    config.type = 'int'
    config.default = None
    config.description = u'Logo for the licensee'
    configs.append(config)

    return configs
