# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.
import datetime

from twisted.python.reflect import namedAny
from sqlalchemy.sql import and_, or_, desc
from django.conf import settings

from basesoft.web.contrib.mailing.models import Mail, MailGroup, MailLocale, Mailing
from basesoft.web.contrib.license.models import Group
from basesoft.web.contrib.i18n.models import Locale
from basesoft.web.contrib.contact.models import ContactList
from basesoft.web.contrib.contact.controllers import ContactController
from basesoft.web.core import exceptions as basesoft_exceptions
from basesoft.web.core.util.controllers import AbstractController, transaction_decorator, AbstractStateController, StateManager
from sqlalchemy.orm.session import make_transient

class AbstractMailStateController(AbstractStateController):
    """
    Extends AbstractStateController and adds some common logic for all mails
    """
    MailLocaleClass = MailLocale
    _locale_attributes = ['subject', 'content']

    def _instance_from_state(self, state):
        """
        Abstract method that takes care of retrieving the common compose objects
        based on a state.

        @access protected
        """

        # Get the object from the db if the id is passed else create a new obj.
        # these are objects like IQuestMail, MockMail, etc.
        self._assert_state(state, 'dbID', (long, int, None))
        if state['dbID'] is None:
            instance = self._instance_class()
        else:
            instance = self.sa_session.query(self._instance_class).get(state['dbID'])

        return instance

    def load_state(self, mail):
        """
        Private load method. This is called from a load_by_* method
        This returns a state dict
        """

        assert isinstance(mail, Mail)
        self.assert_permissions(*self.VIEW_ACCESS_PERMISSIONS)

        state = {
            'mail': {
                'is_wiki': mail.is_wiki,
                'objectName': mail.polymorphic_type,
                'dbID': mail.id,
                'name': mail.name,
                'mail_from_text': mail.mail_from_text,
                'locales': self._load_locales(mail),
                'fallbackLocale': {
                    'value': mail.fallback_locale.locale,
                    'name': mail.fallback_locale.name
                },
            }
        }

        self._load_plugin(mail, state)

        return state

    def save_plugin(self, mail, state):
        self._save_plugin(mail, state)

    def _save_plugin(self, mail, state):
        pass

    def _load_plugin(self, mail, state):
        pass

    def get_replacement_fields(self):
        """
        returns all the replacement fields for the instance_class
        """
        return self._instance_class.replacement_fields()

    def get_js_module(self):
        return self._instance_class.JS_MODULE

class MailBaseController(AbstractController):
    """
    Controller for the mailing app
    """

    MailLocaleClass = MailLocale
    LocaleClass = Locale

    def get_all_mail_types(self):
        """
        Return the type of mails that are available for the current user

        @return: a list of mails to use. eg. [IQuestMail]
        """
        apps = settings.INSTALLED_APPS

        # we only look for the mail type at the beginning
        if hasattr(self, 'available_types'):
            return self.available_types

        self.available_types = []

        for app in apps:
            try:
                mailing_module = namedAny(app+'.mailing')
            except AttributeError:
                # if we do not have a mailing module then the app doesn't have
                # it's own mail types
                continue

            available_state_controllers = mailing_module.MailController._state_controllers
            for key in available_state_controllers:
                c = available_state_controllers[key]
                if self.has_permissions(*c.VIEW_ACCESS_PERMISSIONS):
                    self.available_types.append(c._instance_class)

        return self.available_types

    def _set_state_controllers(self):
        """
        Private method used by the get_controller method to set the _state_controllers dict in case we don't have it
        (eg.: when we create a MailBaseController for rpcs, instead of importing the MailController class from the
        mailing.py module that interests us)
        """
        apps = settings.INSTALLED_APPS

        self._state_controllers = {}

        for app in apps:
            try:
                mailing_module = namedAny(app+'.mailing')
            except AttributeError:
                # if we do not have a mailing module then the app doesn't have
                # it's own mail types
                continue

            available_state_controllers = mailing_module.MailController._state_controllers
            for key in available_state_controllers:
                c = available_state_controllers[key]
                self._state_controllers[key] = c

    def get_replacement_fields(self, mail_type, is_wiki=False):
        """
        Return the dictionary that contains the replacement fields for the specified type
        @param mail_type: the type of the mail for which we want replacements. eg.: 'IQuestMail', 'MockMail', etc
        """
        rf = self.get_controller(mail_type).get_replacement_fields()
        if not is_wiki:
            return rf

        def append_replacement(l, key, value):

            submenu = None
            if type(value) == dict:
                submenu = []
                for k, v in value.iteritems():
                    append_replacement(submenu, k, v)
                value = key

            x = {
                 'key': key,
                 'label': value
            }
            if submenu is not None:
                x['submenu'] = submenu

            l.append(x)

        rv = []
        for k, v in rf.iteritems():
            append_replacement(rv, k, v)

        return rv

    def get_js_module(self, mail_type):
        return self.get_controller(mail_type).get_js_module()

    def get_controller(self, state):
        """
        Create a state controller based on what object we have in the state
        @param state: a state dictionary or the string that represents the
            object type. eg: 'IQuestMail' or a mail object
            (which has the polymorphic_type set, and we use that)
        """
        if isinstance(state, dict):
            objectType = state['objectName']
        else:
            objectType = state

        # if we are using MailBaseController (for rpc for example), we don't have _state_controllers set, so we set it
        if not hasattr(self, '_state_controllers'):
            self._set_state_controllers()

        try:
            c_class = self._state_controllers[objectType]
        except KeyError:
            raise basesoft_exceptions.PublicSystemException('%s: no such controller: %s' % (self, objectType) )

        return c_class(self.sa_session, StateManager(self.locale, self.account_id, None))

    def get_locales(self, mail_type):
        """
        Return the locales that are used for the mail translating
        @param mail_type: the type of the mail for which we want locales. eg.: 'IQuestMail', 'MockMail', etc
        """
        i18n_controller = self.get_controller(mail_type)
        return i18n_controller.locale_controller.get_domain_locales(availability=Locale.AVAILABLE_FOR_COMPOSE)

    def get_locales_state(self, mail_type):
        """
        Return the locales that are used for the mail translating
        @param mail_type: the type of the mail for which we want locales. eg.: 'IQuestMail', 'MockMail', etc
        """
        i18n_controller = self.get_controller(mail_type)
        locales = i18n_controller.locale_controller.get_domain_locales(availability=Locale.AVAILABLE_FOR_COMPOSE)

        return [l.get_state() for l in locales]

    @transaction_decorator
    def save_state(self, state, group=None):
        """
        Takes care of saving the mail state and returns the new/updated mail

        @param state: the state that we need to save
        @param group: the group that created this mail. if not specified, set
            the default group
        @param group_privs: the privileges for the group on object level. if not
            specified, the group will have all privileges
        @param license_privs: the privileges for the license on object level.
            0 if not specified (no privileges)
        @param domain_privs: the privileges for the domain. 0 if not specified
        """

        if group is None: group = self._get_default_group()
        assert isinstance(group, Group)

        state = state['mail']

        assert isinstance(state, dict)
        try:
            assert isinstance(state['objectName'], (unicode, str))
        except AssertionError, e:
            raise e


        # save or update
        controller = self.get_controller(state)
        if isinstance(state['dbID'], (long, int, )):
            self.assert_permissions('update_mail')
            controller.assert_permissions(*controller.UPDATE_ACCESS_PERMISSIONS)
        else:
            self.assert_permissions('create_mail', group_id=group.id)
            controller.assert_permissions(*controller.CREATE_ACCESS_PERMISSIONS)

        mail = controller._instance_from_state(state)

        # check the object privileges in case of update
        if isinstance(state['dbID'], (long, int, )):
            self.assert_privileges_on_object(mail, Mail.PRIV_UPDATE)

        # Save the mail name
        controller._assert_state(state, 'name', unicode)
        mail.name = state['name']
        mail.is_wiki = False if not state.has_key('is_wiki') else state['is_wiki']
        mail.mail_from_text = None if not state.has_key('mail_from_text') else state['mail_from_text']

        # locales
        controller._save_locale(mail, controller.MailLocaleClass, state)
        controller._assert_state(state, 'fallbackLocale', dict)
        controller._assert_state(state['fallbackLocale'], 'value', unicode)
        mail.fallback_locale = controller.locale_controller.get_locale_by_str(state['fallbackLocale']['value'])

        # set other survey properties using the given parameters
        if not isinstance(state['dbID'], (long, int, )):
            self.save_default_privilege_object_group(mail, group)

        controller.save_plugin(mail, state)
        self.sa_session.expire(mail)

        self.sa_session.add(mail)
        self.sa_session.flush()

        # finally return the new saved mail
        return mail

    def load_by_mail_id(self, id):
        """
        Loads and checks permissions for the given mail id
        Returns a state object
        """
        self.assert_permissions('view_mail')

        mail = self.get_mail_by_id(id)
        controller = self.get_controller(mail.polymorphic_type)
        return controller.load_state(mail)

    def get_mail_by_id(self, id):
        """
        Returns the mail model object for the given id
        """
        mail = self.sa_session.query(Mail).get(id)

        if not mail:
            raise basesoft_exceptions.PublicSystemException('Invalid Mail: [%s]' % id)

        self.assert_privileges_on_object(mail, Mail.PRIV_VIEW)

        return mail

    def get_mails(self, id=None, group=None, mail_type=None):
        """
        Returns a list with all Mails in case id is None or a list of Mails with
        supplied id/ids. (if one valid id is supplied, a list with one Mail is
        returned). If group is set, the list returned will be with Contacts on
        which the specified group(s) has(have) privileges.

        @param id: None, list or int
        @param group: None, list with Group objects or Group object
        @param mail_type: None, list or String. The type of the mail we want to get. if None, get all mails;
            if we have a string or list (eg. 'MockMail', 'IQuestMail'), get only mails of that type

        @return: Mail objects
        """
        self.assert_permissions('view_mail')

        filter, fo = self.get_query_filter_on_privileges(Mail, MailGroup, Mail.PRIV_VIEW)

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, Mail.id.in_(id))

        if mail_type is not None:
            if not isinstance(mail_type, list):
                mail_type = [mail_type]
            filter = and_(filter, Mail.polymorphic_type.in_(mail_type))

        # add the filter for group(s) on which the Mail has privileges
        if group is not None:
            if not isinstance(group, list):
                group = [group]
            g_ids = [g.id for g in group]
            filter = and_(filter,
                          or_(Mail.group_id.in_(g_ids),
                              MailGroup.group_id.in_(g_ids)))

        filter = and_(filter, and_(Mail.deleted==False, Mail.archived==False))

        query = self.sa_session.query(Mail).select_from(fo).filter(filter).order_by(desc(Mail.created))

        return query.all()

    def get_mail_locales(self, mail, locale=None, order_locales_by=None):
        """
        Returns a list with MailLocale objects that belong to the specified Mail

        @param mail: the Mail object or mail_id for which we want to get the locales
        @param locale: None, list of Locale objects or Locale ids

        @return: a list containing all the locales (if locale=None), or only the
            specified locales (if locale is a list of locales or just a locale)
        """
        self.assert_permissions('view_mail')
        if not isinstance(mail, Mail):
            mails = self.get_mails(mail)
            if len(mails) == 0:
                raise basesoft_exceptions.SecurityException('Invalid mail id %s' % mail)
            mail = mails[0]

        self.assert_privileges_on_object(mail, Mail.PRIV_VIEW)

        if order_locales_by is None:
            order_locales_by = self.LocaleClass.id

        if locale is None:
            mail_locales = self.sa_session.query(self.MailLocaleClass).join(self.LocaleClass).filter(
                self.MailLocaleClass.mail==mail).order_by(order_locales_by).all()
        else:
            if not isinstance(locale, list): locale = [locale]
            if not isinstance(locale[0], self.LocaleClass):
                i18n_controller = self.get_controller(mail.polymorphic_type)
                locales = i18n_controller.locale_controller.get_locales(locale)
                if len(locales) == 0:
                    raise basesoft_exceptions.SecurityException('Invalid locale(s)')
                locale = locales
            locale_ids = [elem.id for elem in locale]
            mail_locales = self.sa_session.query(self.MailLocaleClass).join(self.LocaleClass).filter(and_(
                self.MailLocaleClass.mail==mail,
                self.LocaleClass.id.in_(locale_ids))).order_by(order_locales_by).all()

        return mail_locales

    @transaction_decorator
    def delete_mail_locales(self, mail, locale=None):
        """
        Delete locale(s) for the given mail

        @param mail: {Mail} the mail object for which we want to delete locales
        @param locale: the locale(s) that we want to delete. can be a Locale
            object, a list of locale objects or None if we want to delete all
        """

        self.assert_permissions('delete_mail')
        controller = self.get_controller(mail.polymorphic_type)
        controller.assert_permissions(*controller.DELETE_ACCESS_PERMISSIONS)

        self.assert_privileges_on_object(mail, Mail.PRIV_DELETE)

        if locale is None:
            mail_locales = self.sa_session.query(self.MailLocaleClass).filter(self.MailLocaleClass.mail==mail).all()
        else:
            if not isinstance(locale, list): locale = [locale]
            for l in locale:
                assert isinstance(l, self.LocaleClass)
            locale_ids = [elem.id for elem in locale]
            mail_locales = self.sa_session.query(self.MailLocaleClass).filter(and_(
                self.MailLocaleClass.mail==mail,
                self.MailLocaleClass.locale_id.in_(locale_ids))).all() # FIXME: bad! shouldnt be referencing fk id's (MailLocale.locale_id)  on models

        for loc in mail_locales:
            self.sa_session.delete(loc)

    @transaction_decorator
    def delete_mail(self, id):
        """
        Delete the specified mail(s)

        @param id: {int} or {List} of ints that represent the ids of the mails we want to delete
        """

        # TODO: refactor so it can directly use mark_as_deleted from abstract controller

        self.assert_permissions('delete_mail')
        if not isinstance(id, list):
            id = [id]
        for i in id:
            mail = self.get_mail_by_id(i)

            # get the proper controller for this
            controller = self.get_controller(mail.polymorphic_type)
            controller.assert_permissions(*controller.DELETE_ACCESS_PERMISSIONS)
            self.mark_as_deleted(Mail, i)

    @transaction_decorator
    def save_mailing(self, mailing):
        """
        Checks permissions and consistency of the mailing object before its saved. This should be called
        from another controller on specific application level (iquest, iqreport, etc).

        @param mailing: {Mailing} object containing all needed info for a mailing like mail, contact lists etc
        """

        # check if we have the correct permissions and correctness
        self.assert_permissions('schedule_mail')
        assert isinstance(mailing, Mailing)
        assert isinstance(mailing.mail, Mail)
        assert isinstance(mailing.contact_list, ContactList)
        assert isinstance(mailing.send_date, datetime)

        controller = self.get_controller(mailing.mail.polymorphic_type)
        controller.assert_permissions(*controller.SCHEDULE_ACCESS_PERMISSIONS)

        self.sa_session.add(mailing)

    def get_mailings(self, id=None):
        """
        Returns a list with all Mailings in case id is None or a list of
        Mailings with the supplied id(s) (if one valid id is supplied, a list
        with one Mailing is returned).

        @param id: None, list or int

        @return: list of Mailing objects
        """
        self.assert_permissions('schedule_mail')

        if id is None:
            mailings = self.sa_session.query(Mailing).all()
        else:
            if not isinstance(id, list): id = [id]
            mailings = self.sa_session.query(Mailing).filter(
                Mailing.id.in_(id)).all()

        return mailings

    @transaction_decorator
    def duplicate_mail(self, original):
        if not isinstance(original, Mail):
            original = self.get_mail_by_id(original)

        controller = self.get_controller(original.polymorphic_type)
        self.assert_permissions(*controller.CREATE_ACCESS_PERMISSIONS)
        new = original.clone()
        self.sa_session.add(new)
        return new

class BaseMailingStateController(AbstractStateController):
    """
    Provides a base state controller for Mailing
    @abstract
    """

    # set this to the inherited Mailing class (e.g. ComposePublicationMailing)
    _instance_class = Mailing

    # set this to the correct mail controller for your application
    _mail_controller_class = None

    def save_instance_mailings(self, instance, instance_column, mailings_state):
        """
        Does the propper chacjs and saves mailings for instance
        @param instance
        @param instance_column: string with column name of the *Mailing object that should be the link to instance
        @ param mailings_state: list with mailings dicts
        """
        assert hasattr(instance, 'mailings')
        mailings = mailings_state
        if isinstance(mailings, list) and len(mailings) > 0:

            ids = [int(m['dbID']) for m in mailings if m['dbID'] is not None]
            todel = []
            for m in instance.mailings:
                if m.id not in ids:
                    if m.status in (Mailing.STATUS_SCHEDULED, Mailing.STATUS_PREPARING_FOR_SENDING):
                        todel.append(m)
                    else:
                        raise basesoft_exceptions.SecurityException('Trying to remove SENDING/SENT mailing %s' % m.id)
            ms = []
            for m in mailings:
                # can't update SEDNING or SENT mailings
                if m['dbID'] is not None and m['status'] not in (Mailing.STATUS_SCHEDULED, Mailing.STATUS_PREPARING_FOR_SENDING):
                    newm = self.sa_session.query(self._instance_class).filter(
                            and_(self._instance_class.mailing_id==m['dbID'],
                                 getattr(self._instance_class, instance_column)==instance)).all()
                    if len(newm) == 0:
                        raise basesoft_exceptions.SecurityException('Invalid mailing id %s' % m['dbID'])
                    newm = newm[0]
                else:
                    newm = self.save(m)
                ms.append(newm)
            instance.mailings = ms
            for m in todel:
                self.sa_session.delete(m)
        else:
            for m in instance.mailings:
                if m.status in (Mailing.STATUS_SCHEDULED, Mailing.STATUS_PREPARING_FOR_SENDING):
                    self.sa_session.delete(m)
                else:
                    raise basesoft_exceptions.SecurityException('Trying to remove SENDING/SENT publication mailing %s' % m.id)
        return instance

    @transaction_decorator
    def save(self, state):
        """
        Extracts the base arguments from the state and constructs a *Mailing object. This object is returned
        and can be saved in the derived save method of this class. (like PublicationMailingStateController.save)
        """
        self._assert_state(state, None, dict)
        self._assert_state(state, 'dbID', (None, long, int))
        self._assert_state(state, 'domID', unicode )
        try:
            state['mail_id'] = int(state['mail_id'])
        except ValueError:
            raise basesoft_exceptions.PublicSystemException("Wrong type for key '%s' in state '%s'." % ('mail_id', state))
        self._assert_state(state, 'mail_id', (long, int))
        self._assert_state(state, 'mail_from_text', unicode)
        self._assert_state(state, 'send_date', datetime.datetime)
        self._assert_state(state, 'contact_lists', list)
        if state.has_key('filter'):
            state['filter'] = int(state['filter'])
            assert state['filter'] in [avf[0] for avf in self.get_available_filters()]

        # contact lists should be ints
        for i in range(len(state['contact_lists'])):
            try:
                state['contact_lists'][i] = int(state['contact_lists'][i])
            except ValueError:
                raise basesoft_exceptions.PublicSystemException("Wrong type for key '%s' in state '%s'." % (i, state['contact_lists']))
            self._assert_state(state['contact_lists'], i, (long, int))

        # now check if there is at least one contact that has an email address in every contact list
        # otherwise it has no point in saving this mailing
        contact_controller = ContactController(self.sa_session, account_id=self.account_id)
        cl_objs = contact_controller.get_contact_lists(id=state['contact_lists'])
        # this will raise a public user exception
        contact_controller.check_contact_lists_email(cl_objs)

        # use the mailing controller to get the mail by id (that also checks permissions)
        # this mail can be of any kind (like IQuestMail, ..)
        mail_controller = self._mail_controller_class(self.sa_session, self.locale, self.account_id)
        mail = mail_controller.get_mail_by_id(state['mail_id'])

        # since we are in the mailing-state class the user needs to have schedule privileges on the mail object to be able to save
        # for this we need to instanciate the mail_state controller (not sure why but ok its there)
        mail_state_controller = mail_controller.get_controller(mail.polymorphic_type)
        mail_state_controller.assert_permissions(*mail_state_controller.SCHEDULE_ACCESS_PERMISSIONS)

        # get or create the mailing
        mailing = AbstractStateController._instance_from_state(self, state)

        # check the status of the mailing.. we only allow saving in scheduled state
        if mailing.status is not None and mailing.status not in [Mailing.STATUS_SCHEDULED, Mailing.STATUS_PREPARING_FOR_SENDING]:
            raise basesoft_exceptions.PublicSystemException('Unable to update Mailing [%s] because its already sent.' % mailing.id)

        # finally all good here.. update properties
        mailing.mail = mail
        mailing.mail_from_text = state['mail_from_text']
        mailing.send_date = state['send_date']
        mailing.contact_lists = cl_objs
        if state.has_key('filter'):
            mailing.filter = state['filter']

        # if we are saving a mail it must be in scheduled mode.. set it explicit
        mailing.status = Mailing.STATUS_SCHEDULED

        # notice that we dont save here but return the mailing object
        return mailing

    def get_available_filters(self):
        """
        Method returns available filters
        """
        return self._instance_class().get_filters()

    def load(self, mailing):
        """
        Loads the mail state for the given mailing object (should be called from another StateController)
        """
        assert isinstance(mailing, self._instance_class)

        state = {
            "dbID": mailing.id,
            "mail_id": mailing.mail_id,
            "mail_from_text": mailing.mail_from_text,
            "mail_name": mailing.mail.name,
            "domID": None,
            "status": mailing.status,
            "filter": mailing.filter,
            "send_date": mailing.send_date,
            'contact_lists': [cl.id for cl in mailing.contact_lists],
        }

        return state

