# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

"""
Mailing tasks..
TODO: since this file only contains abstract tasks this file should be renamed to 'tasks.py'
"""
from sqlalchemy.exc import IntegrityError

import os
import re
import datetime
import smtplib

from BeautifulSoup import BeautifulSoup

from email.message import Message
from email.header import Header
from email.mime.text import MIMEText
from email.mime.multipart import MIMEMultipart
from email.mime.image import MIMEImage
from email.MIMEBase import MIMEBase
from email.Utils import formatdate
from email import Encoders
from email import utils as mail_utils

from sqlalchemy.sql import and_

from basesoft.scheduler.task import ScheduleTask
from basesoft.web.contrib.mailing.models import Mailing, MailResult, Mail
from django.conf import settings
from basesoft.web.contrib.contact.models import ContactListContact, Contact
from basesoft.web.contrib.file.models import File

class MailRecipientRefusedException(Exception): pass
class MailErrorException(Exception): pass
class MailIgnoredException(Exception): pass

from basesoft import logging
logger = logging.getLogger('basesoft.account.schedule')

class MailTask(ScheduleTask):
    """
    Class that provides methods that take care of mailing. This does the actual sending and provides a base for
    all tasks that need to send out emails.
    """

    email_host = settings.EMAIL_HOST
    email_port = settings.EMAIL_PORT

    def _send_mail(self, email, msg, mail_from=None, return_path=None):
        """
        Protected method that takes care of sending the given msg to the given contact. Message must be of the type
        email.Message and can be a mime or whatever type of message. This methods throws an exception if mail failed
        for whatever reason.. so this should be caught.
        """
        # Get the 'to address' from the contact.. if none is available skip to the next contact
        if email.strip() == '':
            logger.info('%s: Ignoring sending because of empty email address' )
            raise MailIgnoredException

        to = email.strip()

        # Some error checking before sending
        assert isinstance(to, (str, unicode))
        assert isinstance(msg, Message)

        if mail_from is None:
            mail_from = settings.EMAIL_FROM

        if return_path is None:
            return_path = mail_from

        if msg['To'] is None:
            msg['To'] = mail_utils.formataddr((email, email))

        if msg['From'] is None:
            msg['From'] = mail_from

        #
        # send out mail
        #
        s = smtplib.SMTP(self.email_host, self.email_port)
        try:
            logger.info('%s: _send_mail: smtp sending mail to %s' % (self, to))
            s.sendmail(return_path, [to], msg.as_string())
            logger.info('%s: _send_mail: successfully sent to %s' % (self, to))
        except smtplib.SMTPRecipientsRefused, e:

            # All recipients were refused. Nobody got the mail. The recipients attribute of the exception object is a
            # dictionary with information about the refused recipients (like the one returned when at least one recipient was accepted).
            # Notice that is not treated as an error in the log.. we'll probably send meelions of these
            logger.info('%s: All recipients were refused to:%s, e:%s' % (self, to, e.message))
            raise MailRecipientRefusedException(e.message)

        except smtplib.SMTPSenderRefused, e:
            logger.error('%s: The server didn\'t accept the from_addr:%s, to:%s, e:%s' % (self, settings.EMAIL_FROM, to, e.message))
            raise MailErrorException(e.message)

        except smtplib.SMTPHeloError, e:
            logger.error('%s: The server didn\'t reply properly to the HELO greeting. to:%s, e:%s' % (self, to, e.message))
            raise MailErrorException(e.message)

        except smtplib.SMTPDataError:
            logger.info('%s: The server replied with an unexpected error code (other than a refusal of a recipient). to:%s, e:%s' % (self, to, e.message))
            raise MailErrorException(e.message)

        except Exception, e:
            logger.error('%s: %s' % (self,e.message) )
            raise MailErrorException(e.message)

        finally:
            s.close()

    def _build_message(self, person, mail, mail_from=None):
        """
        returns a email.Message object of any kind (MIMEText, MIMEMultipart, ..)
        See the documentation at http://docs.python.org/library/email and http://docs.python.org/library/email-examples.html

        By default we just return a simple MIMEText message with the mail content
        """
        if mail_from is None:
            mail_from = settings.EMAIL_FROM
        msg = MIMEText(mail.get_locale_mail(person.preferred_locale).content)
        msg['Subject'] = Header(mail.get_locale_mail(person.preferred_locale).subject, 'utf-8')
        msg['From'] = mail_from
        msg['Date'] = formatdate(localtime=True)
        msg['To'] = person.email
        return msg


class HtmlMailTask(MailTask):
    """
    Base task class for sending html mail(s). This overrides the _build_message to create a html message with
    replacement-fields.
    """

    def __init__(self, *args, **kwargs):
        MailTask.__init__(self, *args, **kwargs)
        self.mail_locales = {}

    def _insert_result(self, contact_id, mailing_id, status=MailResult.STATUS_NEW):
        sql = """
            INSERT INTO tblMailResult (contact_id, status, mailing_id, created, last_updated)
            VALUES (:contact_id, :status, :mailing_id, NOW(), NOW())
        """
        self.sa_session.execute(sql, {
            'contact_id': contact_id,
            'status': status,
            'mailing_id': mailing_id
        })
        return {'contact_id': contact_id, 'mailing_id': mailing_id}

    def _save_status(self, result):
        sql = """
            UPDATE tblMailResult set status=:status, last_updated=NOW()
            WHERE contact_id=:contact_id AND mailing_id=:mailing_id
        """
        self.sa_session.execute(sql, result)

    def _send_mail(self, email, msg, result, mail_from=None, return_path=None):


        # send the mail and catch any exceptions
        try:
            MailTask._send_mail(self, email, msg, mail_from, return_path)
        except MailRecipientRefusedException:
            # recipient refused.. update status
            status = MailResult.STATUS_REFUSED
        except MailIgnoredException:
            # seems that the mail was ignored to send.. return so status wont be saved
            status = MailResult.STATUS_IGNORED
        except:
            # error while sending.. update status.. the error message is in the log
            status = MailResult.STATUS_ERROR
        else:
            # in any other case just return as accepted (for now)
            status = MailResult.STATUS_ACCEPTED

        # save the new result object
        logger.info('%s: Mail result: %s' % (self, status))
        if result is not None:
            result['status'] = status
            self._save_status(result)
        return status

    def send_mail(self, context, msg, result, mail_from=None, return_path=None):
        """
        When this is used for direct mail sending
        """
        email = None
        if isinstance(context, Contact):
            if context.email is not None:
                email = context.email
            else:
                email = context.person.email
        elif context is not None:
            email = context.email

        if email is None:
            return
        return self._send_mail(email, msg, result, mail_from, return_path)

    def _embed_images(self, html):

        from basesoft.web.contrib.file.controllers import AnonymousFileController

        images = []

        pattern = re.compile(Mail.regexp_embed)
        controller = AnonymousFileController(self.sa_session)

        def embed_image(match_obj):

            logger.info('Found: %s' % match_obj.group(0))

            if len(match_obj.groups()) != 2:
                return match_obj.group(0)

            cid = 'image_%s_%s' % (match_obj.group(2), len(images))
            logger.info('Embedding with cid: %s' % cid)

            file = controller.get_file_content(match_obj.group(2))
            msgImage = MIMEImage(file['content'])
            msgImage.add_header('Content-ID', '<%s>' % cid)
            images.append(msgImage)

            return '<img src="cid:%s">' % cid

        html = re.sub(pattern, embed_image, html)

        return html, images

    def _direct_images(self, html):

        pattern = re.compile(Mail.regexp_embed)

        def direct_image(match_obj):
            if len(match_obj.groups()) != 2:
                return match_obj.group(0)

            return '<img src="%s%s" />' % (settings.BASE_URL, match_obj.group(1))

        html = re.sub(pattern, direct_image, html)

        return html

    def _plain_text(self, html):
        text = html

        # replace <br> with \n
        regexp = re.compile(r'<\s*br\s*/?\s*>')
        text = re.sub(regexp, '\n', text)

        # replace all anchors with their href link
        regexp = re.compile('<a[^>]*(:?\s+href=["\']?([^"\'\s]+)\s*["\']?){1}[^>]*/?>([^<]*)</a>')
        def replace(match_obj):
            return match_obj.group(2)
        text = re.sub(regexp, replace, text)

        # replace all starting and ending tags with ''
        regexp = re.compile(r'<[^<]*?/?>')
        text = re.sub(regexp, '', text)
        return text

    def _build_message(self, person, mail, subject=None, html=None, context=None, mail_from=None, postfix_subject=None, attachments=None):
        """
        Override _build_message to build a MIMEMultipart message and do replacement fields.
        context can be anything that can be accessed in steps like replacement_fields
        """

        if mail_from is None:
            mail_from = settings.EMAIL_FROM

        if subject is None or html is None:
            if person.preferred_locale is not None and self.mail_locales.has_key(person.preferred_locale.id):
                ml = self.mail_locales[person.preferred_locale.id]
            else:
                ml = mail.get_locale_mail(person.preferred_locale)
                self.mail_locales[ml.locale.id] = ml

            if subject is None:
                subject = ml.subject
            if html is None:
                html = mail.get_html_mail(person, context, ml.content).encode('utf-8')
        
        if postfix_subject is not None:
            subject = '%s %s' % (subject, postfix_subject)

        result = None
        if context.has_key('mailing'):
            result = self._insert_result(context['contact'].id, mailing_id=context['mailing'].id)

        images = []
        # We do not embed images for the time being: html, images = self._embed_images(html)
        html = self._direct_images(html)

        # Create message container - the correct MIME type is multipart/alternative.
        msg = MIMEMultipart('related')
        msg['Subject'] = Header(subject, 'utf-8')
        msg['From'] = mail_from
        msg['Date'] = formatdate(localtime=True)
        msg.preamble = 'This is a multi-part message in MIME format.'

        msgAlternative = MIMEMultipart('alternative')
        msg.attach(msgAlternative)

        plain_text = self._plain_text(html)

        # specify the encoded data in the head so that beautifulsoup knows the current encoding
        # else it will encode double
        html = '<html><head><meta http-equiv="Content-type" content="text/html; charset=utf-8" /></head><body>%s</body></html>' % html

        soup = BeautifulSoup(html)

        # Record the MIME types of both parts - text/plain and text/html.
        part1 = MIMEText(plain_text, 'plain')
        part1.set_charset('utf-8')
        part2 = MIMEText(soup.prettify(), 'html')
        part2.set_charset('utf-8')

        # Attach parts into message container.
        # According to RFC 2046, the last part of a multipart message, in this case
        # the HTML message, is best and preferred.
        msgAlternative.attach(part1)
        msgAlternative.attach(part2)

        for msgImage in images:
            msg.attach(msgImage)

        if attachments is not None:
            for f in attachments:
                if not isinstance(f, File):
                    logger.error('%s: cannot attach file (%s)' % (self, f))
                    continue
                part = MIMEBase('application', "octet_stream")
                part.set_payload(open(os.path.join(settings.FILE_REPOSITORY, f.file_hash), "rb").read())
                Encoders.encode_base64(part)
                part.add_header('Content-Disposition', 'attachment; filename="%s"' % os.path.basename(f.file_name))
                msg.attach(part)

        # return the msg so it will be sent
        return (msg, result)

    def build_message(self, person, mail, subject=None, html=None, context=None, mail_from=None):
        """
        When this is used for direct mail sending
        """
        return self._build_message(person, mail, subject, html, context, mail_from)

class MailingTask(HtmlMailTask):
    """
    MailingTask extends the HtmlMailTask and provides logic to send out all types of mailings (not single mails).
    """

    # Mailing model to query and send out (like ComposePublicationMailing)
    # by filtering by model in a task we allow different business logic per Mailing type (like replacement fields).
    mailing_model = None

    def get_mailings(self):
        return self.sa_session.query(self.mailing_model).filter( \
            and_(self.mailing_model.status.in_([Mailing.STATUS_SCHEDULED, Mailing.STATUS_PREPARING_FOR_SENDING]), self.mailing_model.send_date <= datetime.datetime.now() )).all()

    def execute(self):

        # check for task integrity
        if self.mailing_model is None:
            raise Exception('self.mailing_model cannot be None')

        # query for all mailing models returns tuples containing mailings but only the mailings
        # that are scheduled and should be send out now.
        mailings = self.get_mailings()

        to_be_sent = []
        for mailing in mailings:
            if not mailing.approved_for_sending():
                if mailing.status != Mailing.STATUS_PREPARING_FOR_SENDING:
                    mailing.status = Mailing.STATUS_PREPARING_FOR_SENDING
                    self.sa_session.add(mailing)
                    self.sa_session.flush()
                continue

            # update mailing status to sending..
            mailing.status = Mailing.STATUS_SENDING
            self.sa_session.add(mailing)
            self.sa_session.flush()

            to_be_sent.append(mailing)

        # Loop mailings, contact-lists.. and process each contact in it
        for mailing in to_be_sent:
            logger.info('%s: Starting of sending out mails for mailing %s' % (self, mailing) )

            # process contact lists / contacts
            for contact_list in mailing.contact_lists:

                logger.info('%s: Processing contact_list %s' % (self, contact_list) )

                for clc in contact_list.contact_list_contacts:

                    # if this contact list contact is set to not receive any mail as per mail setting, skip it.
                    if clc.mail_settings == ContactListContact.CONTACT_RECEIVES_NO_MAIL:
                        continue

                    contact = clc.contact
                    if contact.archived == 1 or contact.deleted == 1:
                        continue

                    logger.info('%s: Sending to contact %s' % (self, contact.id) )
                    # see if the contact doesnt have a result yet (a contact can be in multiple contact lists)
                    if self.sa_session.query(MailResult).filter( and_(MailResult.contact==contact, MailResult.mailing==mailing)).count():
                        logger.info('%s: Already sent to %s' % (self, contact) )
                        continue
                    try:
                        self.__execute_mailing_contact(mailing, contact)
                    except IntegrityError, e:
                        logger.info('%s: Failed sending to %s with error %s' % (self, contact, e.message) )
                        continue

                    # checks if the task is stopped and aborts further execution
                    self.loop_validator()

                logger.info('%s: Finished processing contact_list %s' % (self, contact_list) )

            # update mailing status to sent
            mailing.status = Mailing.STATUS_SENT
            self.sa_session.add(mailing)
            self.sa_session.flush()

    def __execute_mailing_contact(self, mailing, contact):
        """
        Private method called by 'execute' that takes care of sending to a contact. This calls the _send_mail method
        and catches its exceptions so we can save the result back in the db.
        """
        if self._filter_contact(mailing, contact):
            return

        mail_from = None
        if hasattr(mailing, 'mail_from'):
            mail_from = mailing.mail_from(contact)

        license = mailing.mail.groups[0].group.license

        # Build the message for the contact and mail
        msg, result = self._build_message(contact, mailing, mail_from, license)

        # use the contact.email at first try, if that is None, try to use the personal email address. Ticket #3166
        email = contact.email
        if email is None:
            if contact.person is None:
                return
            email = contact.person.email

        self._send_mail(email, msg, result, mail_from, return_path=mailing.mail.mail_return_path())

    def _filter_contact(self, mailing, contact):
        """
        Override this method to filter on contacts (Example: respondents/non-respondents in IQuest)
        """
        return False

    def _build_message(self, contact, mailing, mail_from=None, license=None):
        """
        Mailing overrides _build_message cuz it deals with mailing objects instead of mail.. by default call parent
        with context but subclasses can extend this behaviour.
        """
        return HtmlMailTask._build_message(self, contact.person, mailing.mail, context={
            'mailing': mailing,
            'contact': contact,
            'license': license,
        }, mail_from=mail_from)

    def __repr__(self):
        return 'MailingTask'

tasks = ()
