# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.
import sys

from BeautifulSoup import BeautifulSoup
from basesoft.web.contrib.account.models import Account
from django.utils.formats import number_format
from django.utils.translation import ugettext as _
from sqlalchemy.sql.expression import text

from basesoft.util.ordered_dict import OrderedDict
from basesoft.web.contrib.wiki.formatter import HTMLFormatter, OneLinerFormatter
from basesoft.web.contrib.wiki.parser import Macro, Parser
from basesoft.web.contrib.license.models import Permission

from basesoft.web.contrib.wiki.controllers import WikiController
from basesoft.web.contrib.wiki.models import Wiki, WikiContent
from basesoft.web.core.exceptions import PublicUserException, PublicSecurityException

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from basesoft.web.contrib.manual.models import ManualChapter
from basesoft.web.contrib.i18n.models import Locale

class ManualController(AbstractController, WikiController):

    _context = None
    __words = None

    def get_chapter(self, _id):
        if type(_id) in (str, unicode, ):
            return self.get_chapter_by_handle(_id)
        return self.sa_session.query(ManualChapter).get(_id)

    def get_chapter_by_handle(self, handle):
        return self.sa_session.query(ManualChapter).filter(ManualChapter.handle==handle).first()

    def _assert_access(self, chapter):
        if len(chapter.permissions) == 0:
            return

        if self.get_account() is None:
            raise PublicSecurityException(_('No access to the requested manual chapter.'))

        if self.has_permissions('manage_manual_chapters'):
            return True
        self.assert_permissions(*[p.name for p in chapter.permissions])

    def _has_access(self, chapter):
        if len(chapter.permissions) == 0:
            return True
        if self.has_permissions('manage_manual_chapters'):
            return True
        return self.has_permissions(*[p.name for p in chapter.permissions])

    def get_wiki_from_instance_for_reading(self, handle):
        chapter = self.get_chapter_by_handle(handle)
        if chapter is None:
            return Wiki()
        self._assert_access(chapter)
        return chapter.wiki

    def get_wiki_from_instance(self, handle):
        chapter = self.get_chapter_by_handle(handle)
        if chapter is None:
            return Wiki()
        return chapter.wiki

    def _get_locale(self):
        return 'en_GB' if not self.request.session.has_key('config') else self.request.session['config']['basesoft.web.contrib.locale']

    def load_parsed_chapter(self, handle, locale=None):
        if locale is None:
            locale = self._get_locale()

        locale = self.sa_session.query(Locale).filter(Locale.locale == locale).first()
        chapter = self.get_chapter_by_handle(handle)

        if chapter is None:
            raise PublicUserException(_('Chapter not found'))

        self._context = WikiContext(chapter, locale)

        rv = self.load_wiki_with_fallback(handle, locale)
        rv['id'] = chapter.id
        return rv

    def get_context(self):
        return self._context

    def get_available_languages(self, _id):
        chapter = self.get_chapter(_id);
        if chapter is None:
            return []
        return sorted([{'id': l.id, 'name': l.name, 'locale': l.locale} for l in chapter.wiki.available_locales], key=lambda k: k['name'])

    @transaction_decorator
    def save_or_update_chapter(self, _id, manual_state, wiki_state):
        self.assert_permissions('manage_manual_chapters')

        if _id is None:
            chapter = ManualChapter()
            chapter.wiki = None
        else:
            chapter = self.get_chapter(_id)

        chapter.handle = manual_state['handle']
        chapter.category = manual_state['category']
        chapter.permissions = self.sa_session.query(Permission).filter(Permission.id.in_(manual_state['permissions'])).all()

        chapter.wiki = self._save_wiki(chapter.wiki, wiki_state)

        locale = self.sa_session.query(Locale).filter(Locale.locale == 'en_GB').one()

        # english is leading
        if wiki_state['locale_id'] == locale.id:
            self._update_chapter_links(chapter, wiki_state['content'])

        self.sa_session.add(chapter)
        return chapter

    def _update_chapter_links(self, chapter, content):
        f = HTMLFormatter({}, context=self)
        f.format(content)

        wikis = f.get_linked_wikis()
        if len(wikis) == 0:
            chapter.linked_to_chapters = []
        else:
            chapter.linked_to_chapters = self.sa_session.query(ManualChapter).filter(ManualChapter.handle.in_(set(wikis))).all()

    def load_chapter_state(self, _id):
        chapter = self.get_chapter(_id)

        if chapter is None:
            raise PublicUserException(_('Manual chapter not found'))

        # FIXME, load from config or so
        locale = self.sa_session.query(Locale).filter(Locale.locale == 'en_GB').one()

        return {
            'manual': {
                'handle': chapter.handle,
                'category': chapter.category,
                'permissions': [p.id for p in chapter.permissions]
            },
            'wiki': {
                'fallback_locale_id': locale.id,
                'locale_id': locale.id
            }
        }

    def get_available_chapter_handles(self):
        rv = []
        for chapter in self.sa_session.query(ManualChapter):
            if not self._has_access(chapter):
                continue
            rv.append(chapter.handle)
        return rv

    def get_grouped_chapters(self):
        rv = {}
        for chapter in self.sa_session.query(ManualChapter):
            if not self._has_access(chapter):
                continue
            if not rv.has_key(chapter.category):
                rv[chapter.category] = []
            rv[chapter.category].append(chapter)

        chapters = sorted(rv.keys())

        od = []
        for chapter in chapters:
            od.append((chapter, rv[chapter], ))

        return OrderedDict(od)

    def add_macros(self, parser):
        parser.add_macro('ManualIndex', ManualIndex)
        parser.add_macro('RecentChanges', RecentChanges)
        parser.set_available_wikis(self.get_available_chapter_handles())
        parser.set_highlight_words(self.__words)
        parser.set_context(self)

    def get_chapter_information(self, _id, locale=None):
        if locale is None:
            locale = self._get_locale()
        locale = self.sa_session.query(Locale).filter(Locale.locale == locale).first()

        chapter = self.get_chapter(_id)
        revision = self._get_current_wiki_content(chapter.wiki, locale)

        languages = sorted([l.name for l in chapter.wiki.available_locales])

        # FIXME, use gettext for content

        rv = """
        = Basic information =
            || Wiki title || %s ||
            || Display title || %s ||
            || Category || %s ||
            || Page ID || %d ||
            || Page content language || %s ||
            || Page length (in bytes) || %s ||
            || Available languages || %s ||
        = Edit history =
            || Date of page creation || %s ||
            || Latest editor || %s ||
            || Date of last edit || %s ||
            || Total number of edits || %s ||
        """ % (
            chapter.handle,
            '-' if revision is None else revision.title,
            chapter.category,
            chapter.id,
            '-' if revision is None else revision.locale.name,
            '-' if revision is None else number_format(sys.getsizeof(revision.content), use_l10n=True, force_grouping=True),
            ', '.join(languages),
            chapter.created,
            '-' if revision is None or revision.account is None else revision.account.get_full_name(),
            '-' if revision is None else revision.created,
            '-' if revision is None else revision.revision,
        )

        p = Parser()
        return p.render_to_html(rv)

    def _get_editor_account(self):
        return self.get_account()

    def get_history(self, _id, locale=None):
        if locale is None:
            locale = self._get_locale()
        locale = self.sa_session.query(Locale).filter(Locale.locale == locale).first()

        chapter = self.get_chapter(_id)
        history = self._get_wiki_history(chapter.wiki, locale)

        # FIXME, use gettext for content

        rv = """
        = Change history for %s =
        || || '''Version''' || '''Date''' || '''Author''' ||
        """ % chapter.handle

        for revision in history:
            rv += "|| || [wiki:Special:ViewRevision:%d-%s:%s %d] || %s || %s ||\n" % (
                revision.revision,
                chapter.handle,
                locale.locale,
                revision.revision,
                revision.created,
                '-' if revision.account is None else revision.account.get_full_name()
            )

        p = Parser()
        return p.render_to_html(rv)

    def get_links_from(self, _id, locale=None):
        if locale is None:
            locale = self._get_locale()
        locale = self.sa_session.query(Locale).filter(Locale.locale == locale).first()

        # FIXME, use gettext for content

        chapter = self.get_chapter(_id)
        wiki = self._get_current_wiki_content(chapter.wiki, locale)
        if len(chapter.linked_from_chapters) == 0:
            rv = """== No other pages link here =="""
        else:
            rv = """
                = Pages that link to "%s" =
                === The following pages link to [wiki:%s %s] ===\n""" % (wiki.title, chapter.handle, wiki.title)
            for linked in chapter.linked_from_chapters:
                wr = self._get_current_wiki_content(linked.wiki, locale)
                rv += " * [wiki:%s %s]\n" % (linked.handle, wr.title)

        p = Parser()
        return p.render_to_html(rv)

    def get_recent_changes(self, limit, locale):
        """
        Returns the most recent changes
        """
        chapters = []
        q = self.sa_session.query(ManualChapter).join(Wiki).join(WikiContent).filter(
            WikiContent.locale == locale,
        ).order_by(WikiContent.created.desc())

        for c in q:
            if c in chapters:
                continue
            if not self._has_access(c):
                continue
            chapters.append(c)
            if len(chapters) == limit:
                break

        return chapters

    def search_wiki(self, _id, locale, terms, match_all=False):
        if match_all:
            self.__words = [terms.strip()]
        else:
            self.__words = [w.strip() for w in terms.split(' ') if w.strip() != '']
        if locale is None:
            locale = self._get_locale()
        locale = self.sa_session.query(Locale).filter(Locale.locale == locale).first()

        chapter = self.get_chapter(_id)
        revision = self._get_current_wiki_content(chapter.wiki, locale)
        if revision is None:
            return {
                'matches': 0
            }
        self._context = WikiContext(chapter, locale)

        macros = Parser.macros
        macros.update({'ManualIndex': ManualIndex, 'RecentChanges': RecentChanges})
        f = HTMLFormatter(macros, context=self)
        f.set_available_wikis(self.get_available_chapter_handles())
        f.set_highlight_words(self.__words)
        return {
            'title': revision.title,
            'content': f.format(revision.content),
            'matches': f.get_highlight_iterator().keys()
        }

    def search(self, locale, terms, match_all=False, offset=0):
        if match_all:
            words = [terms.strip()]
        else:
            words = [w.strip() for w in terms.split(' ') if w.strip() != '']
        if locale is None:
            locale = self._get_locale()
        locale = self.sa_session.query(Locale).filter(Locale.locale == locale).first()

        if len(words) == 0:
            return None

        chapters = []
        for chapter in self.sa_session.query(ManualChapter):
            if not self._has_access(chapter):
                continue
            chapters.append(chapter)

        params = {}

        _like = []
        for w in words:
            d = len(_like)
            params['param%d' % d] = '%%%s%%' % w
            _like.append('w.content LIKE :param%d' % d)

        sql = text("""
            SELECT mc.id, mc.handle, w.title, w.content, w.created, w.account_id
            FROM tblManualChapter as mc
            INNER JOIN tblWikiLocaleRevision as w
              ON mc.wiki_id = w.wiki_id
            INNER JOIN (
              SELECT wiki_id, MAX(revision) as rev
              FROM tblWikiLocaleRevision
              GROUP BY wiki_id
            ) as w2
              ON w.wiki_id = w2.wiki_id and w.revision = w2.rev
            WHERE mc.id IN (%s)
              AND w.locale_id = %d
              AND (%s)
        """ % (','.join([str(c.id) for c in chapters]), locale.id, " OR ".join(_like)))

        rv = []
        result = self.sa_session.execute(sql, params).fetchall()

        p = Parser()
        sl = OneLinerFormatter()

        for row in result[:25]:
            excerpt = ""

            by = None
            if row.account_id is not None:
                by = self.sa_session.query(Account).filter(Account.id == row.account_id).first().get_full_name()

            html = p.render_to_html(row.content)

            soup = BeautifulSoup(html)
            all_text = ''.join(soup.findAll(text=True)).lower()

            for w in words:
                # FIXME!

                part = all_text.partition(' %s ' % w)
                if part[1] != '':
                    excerpt = '%s $$%s$$ %s' % (part[0][-50:], w, part[2][:50])
                    break
                part = all_text.partition(' %s' % w)
                if part[1] != '':
                    excerpt = '%s $$%s$$%s' % (part[0][-50:], w, part[2][:50])
                    break
                part = all_text.partition('%s ' % w)
                if part[1] != '':
                    excerpt = '%s$$%s$$ %s' % (part[0][-50:], w, part[2][:50])
                    break
                part = all_text.partition('%s' % w)
                if part[1] == '':
                    break
                excerpt = '%s$$%s$$%s' % (part[0][-50:], w, part[2][:50])

            rv.append({
                'content': '... %s ...' % sl.format(excerpt),
                'title': row.title,
                'created': row.created,
                'handle': row.handle,
                'account': by
            })

        return {
            'offset': offset,
            'content': rv,
            'matches': len(result)
        }

class WikiContext:
    def __init__(self, chapter, locale):
        self.chapter = chapter
        self.locale = locale


class ManualIndex(Macro):

    def execute(self, bitbucket):
        if not isinstance(self.context, ManualController):
            return self.format('==== Manual index preview ====')

        chapters = self.context.get_grouped_chapters()

        rv = ''
        for k, items in chapters.iteritems():
            rv += '=== %s ===\n' % k
            for chapter in items:
                content = self.context._get_current_wiki_content(chapter.wiki, self.context.get_context().locale)
		if content is None:
		    continue
		rv += ' * [wiki:%s %s]\n' % (chapter.handle, content.title)

            rv += '\n'

        return self.format(rv)


class RecentChanges(Macro):

    def execute(self, number_of_items=15):
        if not isinstance(self.context, ManualController):
            return self.format('==== Recent changes preview ====')

        chapters = self.context.get_recent_changes(number_of_items, self.context.get_context().locale)

        rv = {}
        for c in chapters:
            content = self.context._get_current_wiki_content(c.wiki, self.context.get_context().locale)
            if not rv.has_key(content.created.date()):
                rv[content.created.date()] = []
            rv[content.created.date()].append((c.handle, content.title, content.account, ))

        lst = ""
        for k in reversed(sorted(rv.keys())):
            values = rv[k]

            lst += "=== %s ===\n" % k
            for v in values:
                handle, title, account = v
                by = '' if account is None else ' by %s' % account.get_full_name()
                lst += " * [wiki:%s %s]%s\n" % (handle, title, by)
            lst += "\n"

        txt = """{{{
#!highlight hl="blue" title="%s"
%s
        }}}""" % ('The following pages were recently updated', lst)  # FIXME locale from self.context.get_context()

        return self.format(txt)

