# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

import datetime

from django.utils.translation import ugettext as _

from sqlalchemy.orm import mapper, relation, backref
from sqlalchemy import UniqueConstraint, Column, Integer, String, ForeignKey, DateTime, UnicodeText, SmallInteger

from basesoft.sqlalchemy.table import IQTable
from basesoft.web.contrib.account import models as account_models
from basesoft.web.contrib.license import models as license_models
from basesoft.web.contrib.manual import models as manual_models
from basesoft.web.contrib.i18n import models as i18n_models

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    """
    """
    return tables

def do_tables(metadata):
    global __tables_defined
    if not __tables_defined:

        tblManualChapter = IQTable('tblManualChapter', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('wiki_id', Integer, ForeignKey('tblWiki.id', ondelete='CASCADE')),

            Column('handle', String(255), nullable=False, unique=True),
            Column('category', String(255), nullable=False),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
        )

        tblManualChapterPermission = IQTable('tblManualChapterPermission', metadata,
            Column('chapter_id', Integer, ForeignKey('tblManualChapter.id', ondelete='CASCADE'), primary_key=True),
            Column('permission_id', Integer, ForeignKey('tblPermission.id', ondelete='CASCADE'), primary_key=True),
        )

        tblManualChapterLink = IQTable('tblManualChapterLink', metadata,
            Column('chapter_id', Integer, ForeignKey('tblManualChapter.id', ondelete='CASCADE'), primary_key=True),
            Column('target_chapter_id', Integer, ForeignKey('tblManualChapter.id', ondelete='CASCADE'), primary_key=True),
        )

        tables['tblManualChapter'] = tblManualChapter
        tables['tblManualChapterPermission'] = tblManualChapterPermission
        tables['tblManualChapterLink'] = tblManualChapterLink

        __tables_defined = True

def do_mappings():
    """
    This function is only called by the django middleware
    when it is needed
    """
    global __mapping_defined
    if not __mapping_defined:
        from basesoft.web.contrib.wiki import models as wiki_models

        mapper(manual_models.ManualChapter, tables['tblManualChapter'],
            properties = {
                'permissions': relation(license_models.Permission, secondary=tables['tblManualChapterPermission']),
                'wiki': relation(wiki_models.Wiki),
                'linked_to_chapters': relation(manual_models.ManualChapter,
                    secondary=tables['tblManualChapterLink'],
                    primaryjoin=tables['tblManualChapter'].c.id == tables['tblManualChapterLink'].c.chapter_id,
                    secondaryjoin=tables['tblManualChapter'].c.id == tables['tblManualChapterLink'].c.target_chapter_id,
                    backref="linked_from_chapters"
                )
            }
        )

        mapper(manual_models.ManualChapterPermission, tables['tblManualChapterPermission'],
            properties = {
                'chapter': relation(manual_models.ManualChapter),
                'permission': relation(license_models.Permission)
            }
        )

        mapper(manual_models.ManualChapterLink, tables['tblManualChapterLink'],
            properties = {
                'chapter': relation(manual_models.ManualChapter,
                    primaryjoin=tables['tblManualChapterLink'].c.chapter_id == tables['tblManualChapter'].c.id,
                ),
                'target_chapter': relation(manual_models.ManualChapter,
                    primaryjoin=tables['tblManualChapterLink'].c.target_chapter_id == tables['tblManualChapter'].c.id,
                )
            }
        )

        __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'Manual')

def app_permissions():
    """
    The permissions used within this application. Can be sync'd with the
    `iqsyncpermissions` command.
    """
    return [
        license_models.Permission(u'view_manual_chapter', u"""
            Gives the permission to view manual chapters.
            """, _("View manual chapter")),
        license_models.Permission(u'manage_manual_chapters', u"""
            Gives the permission to manage manual chapters.
            """, _("Create manual chapter")),
        license_models.Permission(u'translate_manual_chapter', u"""
            Gives the permission to translate a manual chapter.
            """, _("Translate manual chapter")),
        ]

def app_configs():
    """
    The account configs available for this application. This can be sync'd with the
    `iqsyncpermissions` command.
    """
    configs = []

    return configs
