# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _
from django.utils import simplejson
from sqlalchemy import and_

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from basesoft.web.core.exceptions import PermissionDeniedException
from basesoft.web.core.jsonrpc.dispatcher import json_object_hook, JsonRpcJSONEncoder

from basesoft.web.contrib.license.controllers import LicenseController
from basesoft.web.contrib.license.models import Group
from basesoft.web.contrib.notification.models import Notification, NotificationGroup, NotificationReminderMail
from sqlalchemy.sql.expression import distinct


class NotificationController(AbstractController):

    _instanceClass = Notification

    def __init__(self, sa_session=None, locale=None, account_id=None, request=None):
        AbstractController.__init__(self, sa_session, locale, account_id, request)
        self.license_controller = LicenseController(self.sa_session, None, self.account_id)

    def get_notifications(self, _id=None):
        self.assert_permissions('view_notification')
        return self.get_privileged_object_by_id(Notification, _id, Notification.PRIV_VIEW)

    def get_statuses(self):
        """
        Get notification statuses which the user is already using.
        """
        _filter, fo = self.get_query_filter_on_privileges(Notification, NotificationGroup, Notification.PRIV_VIEW)
        _filter = and_(_filter, Notification.status != None)

        return self.sa_session.query(distinct(Notification.status)).select_from(fo).filter(_filter).all()

    @transaction_decorator
    def update_status(self, _id, status):
        """
        Update status of a notification. This is user specific, should not require permissions
        """
        n = self.get_privileged_object_by_id(Notification, _id, Notification.PRIV_VIEW)
        n[0].status = status
        self.sa_session.add(n[0])

    def get_unread_notifications_count(self):
        """
        Return the number of unread notifications
        """

        _filter, fo = self.get_query_filter_on_privileges(Notification, NotificationGroup, Notification.PRIV_VIEW)
        _filter = and_(_filter, Notification.is_read == False)

        return self.sa_session.query(Notification).select_from(fo).filter(_filter).count()

    def get_unread_notifications(self):
        _filter, fo = self.get_query_filter_on_privileges(Notification, NotificationGroup, Notification.PRIV_VIEW)
        _filter = and_(_filter, Notification.is_read==False)
        return self.sa_session.query(Notification).select_from(fo).filter(_filter).all()

    def load_notification(self, _id, mark_as_read=False):
        """
        Loads a notification for the user.
        """
        n = self.get_notifications(_id)[0]

        if mark_as_read:
            self.mark_as_read(n)

        sender = _("System")
        if n.sender is not None:
            sender = n.sender.license.name + ' - ' + n.sender.name

        action = None
        if n.callback is not None:
            action = simplejson.loads(n.callback, object_hook=json_object_hook)

        return {
            'id': n.id,
            'subject': n.get_subject(),
            'text': n.get_message(),
            'sender': sender,
            'date': n.created,
            'action': action,
            'is_read': n.is_read,
            'status': n.status
        }

    @transaction_decorator
    def mark_as_read(self, notification):
        """
        Marks a notification as read
        """

        notification.is_read = True
        self.sa_session.add(notification)
        return notification

    @transaction_decorator
    def mark_as_archived(self, notification):
        """
        Marks a notification as archived for current user
        """
        if not isinstance(notification, list):
            notification = [notification]
        if not isinstance(notification[0], Notification):
            # got ids
            notification = self.get_notifications(notification)

        for n in notification:
            # also unread notifications can be marked as archived
            na = self.mark_as_read(n)
            na.archived = True
            self.sa_session.add(na)

    def get_available_recipients_groups(self):
        """
        Method returns all recipients the user can send notifications to depending on its permissions.
        """
        self.assert_permissions('create_notification')

        group = self.get_account().default_group
        #get all groups this account has access to
        try:
            groups = self.license_controller.get_groups()
        except PermissionDeniedException:
            groups = [group]

        ids = [g.id for g in groups]
        if self.has_permissions('send_to_license_default_group') and group.license.default_group.id not in ids:
            groups.append(group.license.default_group)
        if self.has_permissions('send_to_domain_default_group') and group.license.domain.default_group.id not in ids:
            groups.append(group.license.domain.default_group)

        ret = []
        for g in groups:
            d = {'id': g.id, 'for_domain': False, 'for_license': False}
            if self.has_permissions("view_all_domains") or self.has_permissions("view_domain"):
                d['name'] = g.license.domain.name + ' - ' + g.license.name + ' - ' + g.name
                d['for_domain'] = True
                d['for_license'] = True
            elif self.has_permissions('view_license'):
                d['name'] = g.license.name + ' - ' + g.name
                d['for_license'] = True
            else:
                d['name'] = g.name
            ret.append(d)
        return ret

    def get_all_recipients(self):
        """
        Returns a list with everybody the user can send a notification too. Unlike get_available_recipients_groups,
        this method just returns one group and the correspondent values for for_license and for_domain in case applicable
        instead of all groups in that license/domain.
        """

        try:
            domains = self.license_controller.get_domains()
            return [{'group': [d.default_group for d in domains],
                     'for_group': False,
                     'for_license': False,
                     'for_domain': True
                     }]
        except PermissionDeniedException:
            group = self.get_account().default_group
            if self.has_permissions('view_license'):
                ret = [{'group': [group.license.default_group],
                        'for_group': False,
                        'for_license': True,
                        'for_domain': False
                        }]
                if self.has_permissions('send_to_domain_default_group'):
                    ret.append({'group': [group.license.domain.default_group],
                                'for_group': True,
                                'for_license': False,
                                'for_domain': False
                                })
            ret = [{'group': [group],
                    'for_group': True,
                    'for_license': False,
                    'for_domain': False
                    }]
            defg = []
            if self.has_permissions('send_to_license_default_group'):
                defg.append(group.license.default_group)
            if self.has_permissions('send_to_domain_default_group'):
                defg.append(group.license.domain.default_group)
            if len(defg) > 0:
                ret.append({'group': defg,
                            'for_group': True,
                            'for_license': False,
                            'for_domain': False
                            })
            return ret

    def set_notification(self, subject, text, instance_class=None, callback=None, status=None, notify=None):
        instance_class = instance_class if instance_class is not None else Notification
        notification = instance_class()
        notification.subject = subject
        notification.text = text
        notification.status = status

        if callback is not None:
            notification.callback = simplejson.dumps(callback, ensure_ascii=False, cls=JsonRpcJSONEncoder)
        if notify is not None:
            notification.notify = simplejson.dumps(notify, ensure_ascii=False, cls=JsonRpcJSONEncoder)
        return notification

    @transaction_decorator
    def create_system_notification(self, notification, recipient_groups, sender=None, for_group=True, for_license=False, for_domain=False):
        """
        Creates a system notification. Recipient should always be a Group (or group id) and set correspondent values for for_group,
        for_license or for_domain in case notification should be for the group, group's whole license or for its whole domain.
        When a notification should be sent to multiple groups/licenses/domains recipient_groups would be a list.

        No permissions checks are done since this is a system notification.
        """
        if not isinstance(recipient_groups, list):
            recipient_groups = [recipient_groups]

        if not isinstance(recipient_groups[0], Group):
            rgs = self.sa_session.query(Group).filter(Group.id.in_(recipient_groups)).all()
            assert len(rgs) == len(recipient_groups)
            recipient_groups = rgs

        for rg in recipient_groups:
            assert isinstance(rg, Group)
            notification = self.save_view_privileges(notification, rg, for_group, for_license, for_domain)

        if sender is not None:
            notification.sender = sender

        self.sa_session.add(notification)

        return notification

    @transaction_decorator
    def create_notification(self, notification, recipient_groups, for_group=True, for_license=False, for_domain=False):
        """
        Creates a notification with current group as sender. All correspondent rights will be checked.
        """
        self.assert_permissions('create_notification')
        assert isinstance(notification, self._instanceClass)
        notification.sender = self.get_account().default_group

        if not isinstance(recipient_groups, list):
            recipient_groups = [recipient_groups]

        for rg in recipient_groups:
            g = self.has_recipient_permissions(rg, for_license, for_domain)
            if g:
                notification = self.save_view_privileges(notification, g, for_group, for_license, for_domain);
        self.sa_session.add(notification)

        self.logger.info('Created notification for recipient_groups=%s' % recipient_groups)

        return notification

    def has_recipient_permissions(self, recipient, for_license, for_domain):
        """
        Checks if current user (sender) has permissions to send notifications to provided recipient.

        The following apply for sender:
         * send_to_license_default_group - can send to default group of its license
         * send_to_domain_default_group - can send to default group of its domain
         * view_all_domains - can send to everybody
         * view_domain - can send to everybody in its own domain
         * view_license - can send to everybody in its own license
         * else, can send to everybody in its own group
        """
        if isinstance(recipient, Group):
            recipient_id = recipient.id
        else:
            recipient_id = int(recipient)

        account = self.get_account()
        domain = account.default_group.license.domain
        license = account.default_group.license
        group = account.default_group

        if for_domain:
            self.assert_any_permissions('view_all_domains', 'view_domain')
        if for_license:
            self.assert_any_permissions('view_all_domains', 'view_domain', 'view_license')

        if recipient_id == group.id:
            return group

        # first check for default groups and correspondent permissions
        if (recipient_id == license.default_group.id and self.has_permissions('send_to_license_default_group')) or\
           (recipient_id == domain.default_group.id and self.has_permissions('send_to_domain_default_group')):
            return self.sa_session.query(Group).filter(Group.id==recipient_id).one()

        # let license controller take care of all next checks
        rgs = self.license_controller.get_groups(recipient_id)
        if len(rgs) != 1:
            raise PermissionDeniedException('Insufficient permissions')

        return rgs[0]

    def save_view_privileges(self, notification, group, for_group, for_license, for_domain):

        # lets always save for default group
        if for_license:
            group = group.license.default_group
        if for_domain:
            group = group.license.domain.default_group

        # this will prevent saving privileges for same license or domain
        go = None
        if notification.id is not None:
            go = self.sa_session.query(NotificationGroup).filter(and_(NotificationGroup.notification==notification,
                                                                 NotificationGroup.group==group,
                                                                 NotificationGroup.privilege==notification.PRIV_VIEW
                                                                 )).first()
        if go is None:
            go = NotificationGroup()
            go.group = group
            go.privilege = notification.PRIV_VIEW
        go.group_has_privilege = for_group
        go.license_has_privilege = for_license
        go.domain_has_privilege = for_domain
        self.sa_session.add(go)
        notification.groups.append(go)

        return notification

    def get_reminder_mail(self):
        return self.get_object_by_priv_in_hierarchy(
            self.get_default_group(),
            NotificationReminderMail,
            NotificationReminderMail.PRIV_USE_FOR_SENDING
        )
