# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.
from basesoft.web.core.util.controllers import get_models_of_type

from django.utils.translation import ugettext as _
from django.utils import simplejson
from sqlalchemy import func, case, or_
from basesoft.web.contrib.license.models import Group, License

from basesoft.web.core.exceptions import PermissionDeniedException
from basesoft.web.core.util.custom_query import CustomQuery, CustomQueryField, CustomQueryModel
from basesoft.web.core.jsonrpc.dispatcher import json_object_hook

from basesoft.web.contrib.notification.models import Notification, NotificationGroup


class NotificationCustomQuery(CustomQuery):
    _type_map = None

    def validate_access(self):
        if not (self.has_permissions('view_notification')):
            raise PermissionDeniedException('You do not have access to the requested resource')

    def tables(self):
        return (
            CustomQueryModel(Notification),
            CustomQueryModel(NotificationGroup, onclause=Notification.id==NotificationGroup.notification_id, use_outerjoin=True),
            CustomQueryModel(Group, onclause=Notification.sender_id==Group.id, use_outerjoin=True),
            CustomQueryModel(License, onclause=Group.license_id==License.id, use_outerjoin=True),
        )

    def fields(self):
        return (
            CustomQueryField(Notification.id, _('ID'), key='id'),
            CustomQueryField(Notification.polymorphic_type, _('Type'), key='type', is_filter=True),
                             # TODO; this requires permission checks
                             #  filter_options=[{'value': k, 'label': l} for k, l in self._get_polymorphic_types().iteritems()]),
            CustomQueryField(Notification.status, _('Status'), key='status', is_filter=True),
            CustomQueryField(case([(Notification.sender_id == None, _('System'))], else_=Group.name),
                             _('From'), key='sender', is_filter=True),
            CustomQueryField(case([(Notification.sender_id == None, _('System'))], else_=License.name),
                             _('License'), key='license', is_filter=True),
            CustomQueryField(case([(func.length(Notification.subject)<=40, Notification.subject)],
                else_=func.substring(Notification.subject, 1, 40)+" ..."), _('Subject'), key='subject', is_filter=False),
            CustomQueryField(Notification.subject, _('Subject'), key='subject_full', is_filter=True, in_select=False),
            CustomQueryField(Notification.callback, _('Action'), key='callback'),
            CustomQueryField(Notification.created, _('Send date'), key='created'),
            CustomQueryField(case([(Notification.is_read == 0, _("Not read"))], else_=_("Read")),
                             _('Is read'), key='is_read', is_filter=True,
                             filter_options=[{
                                                'value': 1,
                                                'label': _("Read"),
                                            },{
                                                'value': 0,
                                                'label': _("Not read"),
                                            }]
            ),
        )
    order_by = (Notification.created.desc(), )
    group_by = (Notification.id, )

    def set_where(self):
        self.where = or_(Notification.archived==False, Notification.archived==None)

    def set_value_processors(self):
        def callback_proc(row, key):
            value = getattr(row, key)
            if value is None:
                return None
            return simplejson.loads(value, object_hook=json_object_hook)

        def type_proc(row, key):
            value = getattr(row, key)
            pmt = self._get_polymorphic_types()
            return pmt[value]

        self._value_processors['callback'] = callback_proc
        self._value_processors['type'] = type_proc

    # in this case we don't need filters for deleted and archived on the Notification object
    def _set_deleted_and_archived_fields(self):
        pass

    def _get_polymorphic_types(self):
        if self._type_map is not None:
            return self._type_map

        self._type_map = {}
        _type = get_models_of_type(Notification)
        for n in _type:
            self._type_map[n['value']] = n['model'].name_i18n
        return self._type_map


custom_queries = (NotificationCustomQuery, )
