# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.
from datetime import datetime
from django.conf import settings

from basesoft.scheduler.crontab import CronTab
from basesoft.web.contrib.mailing.schedule import HtmlMailTask
from basesoft.web.contrib.notification.controllers import NotificationController
from basesoft.web.contrib.notification.models import Notification, NotificationReminderMail
from basesoft.web.core.util.controllers import AbstractController, AccountConfigController

from basesoft import logging
logger = logging.getLogger('basesoft.web.contrib.notification.schedule')


class NotificationReminderMailTask(HtmlMailTask):
    INTERVAL = CronTab('2 0 * * 1-5')  # every weekday at 2:00

    def execute(self):

        if settings.VIRTUAL_DATE is None and not self.today_is_monday(datetime.now().date()):
            return 0

        accounts = []
        account_map = {}

        # get all unread notifications
        notifications = self._get_unread_notifications()

        ac = AbstractController(self.sa_session)
        # for each notification retrieve all distinct accounts
        for n in notifications:
            # unique accounts
            acs = ac.get_accounts_with_privilege_on_instance(n, Notification.PRIV_VIEW, 'view_notification')
            accounts += list(set(acs))
            for a in acs:

                if not account_map.has_key('account_%d' % a.id):
                    account_map['account_%d' % a.id] = []
                account_map['account_%d' % a.id].append(n.get_subject())

        cnt = 0
        for account in accounts:

            ac = AccountConfigController(self.sa_session, account_id=account.id)
            # verify account setting for receiving notification reminder mails.
            cfg = ac.get('basesoft.web.contrib.notification.reminder.mail')
            if cfg == NotificationReminderMail.NO_REMINDER:
                continue
            if cfg == NotificationReminderMail.SEND_MONTHLY and not self.today_is_first_monday_of_month(account.domain.today()):
                continue
            if cfg == NotificationReminderMail.SEND_WEEKLY and not self.today_is_monday(account.domain.today()):
                continue

            cnt += 1
            self._execute_account(account, account_map['account_%d' % account.id])

        return cnt

    @staticmethod
    def today_is_monday(today):
        return today.weekday() == 0

    @staticmethod
    def today_is_first_monday_of_month(today):
        from dateutil import relativedelta
        import datetime

        first_of_month = datetime.datetime(today.year, today.month, 1)

        return today == (first_of_month + relativedelta.relativedelta(weekday=0)).date()  # monday

    def _execute_account(self, account, subjects):
        mail = self._get_mail(account)
        if mail is None:
            logger.warn('NotificationReminderMailTask: no notification reminder mail available for account (%s), aborting' % account)
            return

        person = account.person
        _license = account.default_group.license

        msg, result = self._build_message(person, mail, context={
            'subjects': subjects,
            'license': _license
        }, mail_from=mail.mail_from())
        self._send_mail(person.email, msg, result, mail_from=mail.mail_from(), return_path=mail.mail_return_path())

        logger.info('%s: NotificationReminder mail sent to person_id [%s]' % (self, person.id))

    def _get_mail(self, account):
        controller = NotificationController(self.sa_session, account_id=account.id)
        return controller.get_reminder_mail()

    def _get_unread_notifications(self):
        notifications = self.sa_session.query(Notification).filter(Notification.is_read==False).all()
        return notifications


tasks = (NotificationReminderMailTask, )