'''
Created on Oct 2, 2013

@author: marten
'''
import datetime
from basesoft import logging
from basesoft.web.core.util.controllers import lazy_assert

from django.utils.translation import ugettext, ugettext_lazy as _

from sqlalchemy.orm.session import Session
from sqlalchemy import func

from basesoft.web.core.models import PrivilegeObject
from basesoft.web.contrib.file.models import File
from basesoft.web.core.exceptions import SecurityException
from basesoft.web.contrib.mailing.models import Mail

logger = logging.getLogger("basesoft.web.contrib.order")

class Currency(object):
    """
    @DynamicAttrs
    """

class CountryCurrency(object):
    """
    @DynamicAttrs
    """

class TaxClass(PrivilegeObject):
    """
    @DynamicAttrs
    """
    _name_property = 'name'

    def get_group_object(self):
        return (TaxClassGroup, 'tax_class')


class TaxClassGroup(object):
    """
    @DynamicAttrs
    """

class TaxClassCountry(object):
    """
    @DynamicAttrs
    """

    def get_current_rate(self, for_date):
        session = Session.object_session(self)
        return session.query(func.iqGetCurrentTaxCountryRate(self.tax_class.id, self.country.id, for_date)).one()[0]

class TaxClassCountryRate(object):
    """
    @DynamicAttrs
    """

class GLAccount(object):
    """
    @DynamicAttrs
    """

class InvoiceFile(File):
    """
    @DynamicAttrs
    """
    IS_INVOICE = 0
    IS_CREDIT = 1

    NAME_I18N = _('Invoice')

class NewInvoiceMail(Mail):
    """
    @DynamicAttrs
    """
    NAME_I18N = _("New Invoice Mail")

    PRIV_USE_FOR_SENDING = 64

    JS_MODULE = 'iq_account_mail_plugin'

    def get_privileges(self):
        return Mail.get_privileges(self) + [
            (self.PRIV_USE_FOR_SENDING, ugettext('Use for sending'))
        ]

    @staticmethod
    def replacement_fields():
        """
        Invoice emails don't have person information so just override it and don't call parent
        """
        x = Mail.replacement_fields()
        x.update({
            ugettext('Merchant'): {
                'merchant.name': ugettext('Merchant name'),
            },
            ugettext('Client'): {
                'client.name': ugettext('Client name'),
            },
            ugettext('Order'): {
                'order.id': ugettext('Order ID'),
                #'order.payment_status': ugettext('Payment status'),
            }
        })
        return x

    def _replace_field(self, person, field_name, context):

        tokens = field_name.split('.')
        c = context['order']
        if tokens[0] == 'merchant':
            assert tokens[1] in ['name']
            c = c.merchant
        elif tokens[0] == 'client':
            assert tokens[1] in ['name']
            c = c.client
        elif tokens[0] == 'person':
            c = None
        else:
            assert tokens[1] in ['id']

        if hasattr(c, tokens[1]):
            return unicode(getattr(c, tokens[1]))

        return Mail._replace_field(self, person, field_name, c)

class PaymentMethod(object):
    """
    @DynamicAttrs
    """

    # payment method name that is shown to the user, should be overriden
    name_i18n = None

    # js_module where the payment_method handler can be found
    js_module = None

    def get_payment_method_controller(self):
        raise SecurityException('Cannot get the controller for an Abstract PaymentMethod')

    def get_properties(self):
        """
        properties is a dict that can contain specific payment_method
        """
        return {}

    def __repr__(self):
        return '<PaymentMethod id="%s">' % self.id

class PaymentMethodBankTransfer(PaymentMethod):
    """
    @DynamicAttrs
    """
    payment_type_i18n = _('Bank transfer')

    name_i18n = _('Bank transfer')
    js_module = 'iq_order_payment_banktransfer'

    permissions = ['payment_method_bank_transfer']

    def get_payment_controller(self):
        from basesoft.web.contrib.order.controllers import BankTransferPaymentController
        return BankTransferPaymentController

class PaymentMethodOgone(PaymentMethod):
    """
    @DynamicAttrs
    """
    payment_type_i18n = _('Online payment')

    name_i18n = _('Ogone')
    js_module = 'iq_order_payment_ogone'

    permissions = ['payment_method_ogone']

    def get_payment_controller(self):
        from basesoft.web.contrib.order.controllers import OgonePaymentController
        return OgonePaymentController


class Order(PrivilegeObject):
    """
    @DynamicAttrs
    """
    STATUS_NEW = 0
    STATUS_PENDING  = 1
    STATUS_REJECTED = 2
    STATUS_SUBMITTED = 3 # ready for export, awaits an invoice
    STATUS_CREDITED = 4 # used when a credit is made for this Order

    METHOD_BANK_TRANSFER = 1
    METHOD_DIRECT_DEBIT = 2
    METHOD_PSP = 3

    INVOICE_STAGE_NEW = 0
    INVOICE_STAGE_SENDING = 1
    INVOICE_STAGE_SENT = 2
    INVOICE_STAGE_SKIPPED = 3 # used when the client has disabled receiving of invoice emails.

    PRIV_SUBMIT = 32
    PRIV_PAY = 64 # used for payment through PSP
    PRIV_EXPORT = 128 # used for soap exports
    PRIV_ADD_INVOICE = 256
    PRIV_REMOVE_INVOICE = 512
    PRIV_COMPLETE = 1024 # priv to set order status to complete/reject

    TYPE_DEBIT = 0 # normal orders.
    TYPE_CREDIT = 1 # used for determining auto approval or approval by child order.

    SUBMIT_DONE_BY_USER = 0
    SUBMIT_DONE_BY_SYSTEM = 1

    MERCHANT_LABEL = _('Merchant')
    CLIENT_LABEL = _('Client')

    mark_as_deleted_permission = 'mark_order_as_deleted'
    mark_as_archived_permission = 'mark_order_as_archived'

    all_privileges = (PrivilegeObject.PRIV_VIEW, PrivilegeObject.PRIV_UPDATE, PrivilegeObject.PRIV_ARCHIVE, PRIV_PAY)

    # when auto_submit is True, no intervention needed for an order to be exported
    # then it will be submitted directly on save
    auto_submit = False

    label = None
    name_i18n = None
    readonly = False

    _name_property = 'id'

    def get_name_i18n(self):
        return None

    permissions = None
    any_permissions = False

    js_details_module = 'iq_order_details'

    def get_privileges(self):
        return PrivilegeObject.get_privileges(self) + [
            (self.PRIV_SUBMIT, ugettext('Submit privileges')),
            (self.PRIV_PAY, ugettext('Pay privileges')),
            (self.PRIV_EXPORT, ugettext('Export privileges')),
            (self.PRIV_ADD_INVOICE, ugettext('Add invoice privileges')),
            (self.PRIV_REMOVE_INVOICE, ugettext('Remoe invoice privileges')),
            (self.PRIV_COMPLETE, ugettext('Complete privileges')),
        ]

    def get_group_object(self):
        return (OrderGroup, 'order')

    def get_order_controller(self):
        raise SecurityException('Cannot get the controller for an Abstract Order')

    def get_current_payment_status(self):
        session = Session.object_session(self)
        return session.query(func.iqGetCurrentOrderPaymentStatus(self.id)).one()[0]

    def add_log(self, log_text):
        lazy_assert(log_text, str)

        if self.log is None:
            self.log = ''

        self.log += "|| %s || %s || \n" % (datetime.datetime.now().strftime('%Y-%m-%d %H:%M:%S'), log_text)

        logger.info('Order[%s] add_log(): %s' % (self.id, log_text) )

    def __repr__(self):
        return '<%s id="%s">' % (self.polymorphic_type, self.id)

    def get_submit_dry_run_clone_args(self):

        return {
            'clone_children':{'lines': {'reference_children': []}},
            'clone_similar':['parent_order'],
            'reference_children': ['merchant', 'client', 'currency']
        }

    @staticmethod
    def get_simulate_fields():
        from basesoft.web.contrib.order.rule.conditions import RuleConditionMerchant, RuleConditionGroup
        return [
            RuleConditionMerchant,
            RuleConditionGroup,
        ]

    @staticmethod
    def get_simulate_order_line_types():
        raise Exception('abstract')

    def get_invoice_view_privileges(self):
        # the merchant (license) should be granted view privilege
        return [self.merchant.default_group]


    def print_console(self):

        fields = ['description', 'polymorphic_type', 'quantity', 'price_per_unit', 'line_amount_excluding_vat', 'line_amount_including_vat']
        row_format ="{:>30}" * (len(fields) + 1)
        print row_format.format("", *fields)

        for line in self.lines:
            line_fields = [getattr(line, x) for x in fields]
            print row_format.format("", *line_fields)

        print 'total_amount_excluding_vat: %s' % self.total_amount_excluding_vat
        print 'total_amount_including_vat: %s' % self.total_amount_including_vat


class OrderGroup(object):
    """
    @DynamicAttrs
    """
    def __repr__(self):
        return '<OrderGroup order=%s>' % self.order

class OrderLine(object):
    """
    @DynamicAttrs
    """

    # tax_class can either be None or a string
    # if its None, 0 vat will be applied to the order
    # if its a string, a lookup will be made to TaxClass for its code, and TaxClassCountry for the merchants country when applying vat
    tax_class_code = None

    # same as above but for GL Accounts
    gl_account_code = None

    # No maximum quantity
    MAX_LINE_QUANTITY = None

    # fields that are negated when a credit is created. In some apps (like pvcycle) it will also negate weight and custom_unit
    CREDIT_FIELDS = ('quantity', )

    def clone(self):
        exclude = ['_sa_instance_state', 'id', 'index', 'created', 'last_updated', 'order']
        cls = self.__class__
        new = cls()
        for attr, value in self.__dict__.iteritems():
            if attr in exclude:
                continue
            setattr(new, attr, value)
        return new

    def get_sort(self):
        return lambda k: k.index

    def __repr__(self):
        return '<%s id="%s">' % (self.polymorphic_type, self.id)


    @staticmethod
    def get_simulate_fields():
        from basesoft.web.contrib.order.rule import conditions
        return [
            conditions.RuleConditionLineDescription,
            conditions.RuleConditionLineQuantity,
            conditions.RuleConditionLinePrice,
        ]


class OrderPaymentStatus(object):
    """
    @DynamicAttrs
    """
    PAYMENT_PENDING = 0
    PAYMENT_SUBMITTED = 1 # has done initial payment, but unconfirmed by third party
    PAYMENT_PARTIALLY_RECEIVED = 2 # part of the order has been paid
    PAYMENT_COMPLETED = 3 # payment completed and confirmed
    PAYMENT_REJECTED = 4 # Direct Debit or Credit Card denial


class OrderRule(object):
    """
    @DynamicAttrs
    """
    mark_as_deleted_permission = 'manage_order_rules'
    mark_as_archived_permission = 'manage_order_rules'

    _name_property = 'name'

    def __repr__(self):
        return '<OrderRule name="%s">' % self.name

class OrderRuleCondition(object):
    """
    @DynamicAttrs
    """
    TYPE_TEXT = 1
    TYPE_BOOLEAN = 2
    TYPE_INTEGER = 3
    TYPE_DECIMAL = 4

    OPERATOR_EQUAL = 1
    OPERATOR_GT = 2 # greater than
    OPERATOR_GTE = 3 # greater than or equal
    OPERATOR_LT = 4
    OPERATOR_LTE = 5


class OrderRuleAction(object):
    """
    @DynamicAttrs
    """
    TYPE_NONE = 1
    TYPE_TEXT = 2
    TYPE_BOOLEAN = 3
    TYPE_INTEGER = 4
    TYPE_DECIMAL = 5

class OrderLineRule(OrderLine):
    """
    @DynamicAttrs
    """

class OrderSimulation(PrivilegeObject):
    """
    @DynamicAttrs
    """
    _name_property = 'name'

    def get_group_object(self):
        return (OrderSimulationGroup, 'order_simulation')


class OrderSimulationGroup(object):
    """
    @DynamicAttrs
    """
