# Copyright (c) 2013 BaseSoft B.V.
# See LICENSE for details.
from basesoft.web.contrib.account.decorators import login_required
from basesoft.web.contrib.wiki.parser import Parser
from basesoft.web.core.util.controllers import get_models_of_type, lazy_assert
from basesoft.web.contrib.order.models import Order, PaymentMethod,\
    OrderPaymentStatus, OrderRuleCondition, OrderRuleAction
from basesoft.web.contrib.order.controllers import AbstractOrderController,\
    TaxController, GLAccountController, PaymentController, get_available_order_types
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.input.validation import plugins
from django.utils.translation import ugettext as _
import datetime

from django.conf import settings
from iqportal.iqube.core.controllers import ConfigurableWidgetController,\
    DataSourceController
from django.utils import simplejson


def combobox_filter(options, filter_obj):
    """
    :param options:
    :param filter_obj:
    :return:
    """
    if filter_obj:
        if filter_obj.has_key('filter'):
            options = [o for o in options if unicode(o['label']).lower().find(filter_obj['filter'].lower()) != -1]
        if filter_obj.has_key('value'):
            options = [o for o in options if o['value'] == filter_obj['value']]

    # fixed limit
    options = options[:20]

    return options


class TaxHandler:

    @login_required
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']

        self._controller = TaxController(self.sa_session, account_id = self.account_id, request=request)

    def get_tax_classes(self):
        return [{'label': tax_class.name, 'value': tax_class.id} for tax_class in self._controller.get_tax_classes()]

    def validate_tax_class(self, args):

        form_handler = FormValidator({
            'name': [(plugins.Required, {})],
            'code': [(plugins.Required, {})],
        })
        return form_handler.validate(args)

    def save_tax_class(self, id, state):
        return self._controller.save_tax_class_state(id, state).id

    def load_tax_class(self, id):
        return self._controller.get_tax_class_state(id)

    def validate_save_country_tax(self, args):

        form_handler = FormValidator({
            'country_id': [(plugins.Required, {})],
            'tax_class_id': [(plugins.Required, {})],
            'start_date': [(plugins.Required, {})],
            'rate': [(plugins.Required, {}), (plugins.Number, {})],

        })
        return form_handler.validate(args)

    def load_country_tax(self, id):
        return self._controller.get_tax_class_country_state(id)

    def save_country_tax(self, id, state):
        return self._controller.save_country_tax_state(id, state).id

    def delete_country_taxes(self, ids):
        for id in ids:
            self._controller.delete_country_taxes(id)

class GLAccountHandler:

    @login_required
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']

        self._controller = GLAccountController(self.sa_session, account_id = self.account_id, request=request)

    def validate_gl_account(self, args):

        form_handler = FormValidator({
            'licenses': [(plugins.Required, {})],
            'code': [(plugins.Required, {})],
            'account_number': [(plugins.Required, {})],
        })
        return form_handler.validate(args)

    def save_gl_account(self, id, state):
        return self._controller.save_gl_account_state(id, state).id

    def load_gl_account(self, id):
        return self._controller.get_gl_account_state(id)

    def delete_gl_accounts(self, ids):
        for id in ids:
            self._controller.delete_gl_accounts(id)

    def get_gl_licenses(self):
        return [{'label': license_.name, 'value': license_.id}
            for license_ in self._controller.get_gl_licenses()]


class OrderHandler:

    @login_required
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']
        self.request = request

        self.aoc = AbstractOrderController(self.sa_session, account_id=self.account_id, request=request)
        self.pc = PaymentController(self.sa_session, account_id=self.account_id, request=request)

    def get_js_module(self, polymorphic_type):
        model = get_models_of_type(Order, exclude_parent=True, name=polymorphic_type)
        return model.js_details_module

    def is_credit(self, order_id):
        assert order_id is not None
        order = self.aoc.get_order(order_id).pop()
        return order.type == Order.TYPE_CREDIT

    def _get_payment_controller_instance(self, payment_method_id):
        payment_method = self.pc.get_payment_methods(payment_method_id=payment_method_id)[0]
        return self.pc.get_payment_controller(payment_method_id)(self.sa_session, account_id=self.account_id, request=self.request, payment_method=payment_method)

    def load_order(self, order_id):
        return self.aoc.load_order_state(order_id)

    def remove_invoices(self, order_ids):
        for order_id in order_ids:
            return self.aoc.remove_order_invoice(order_id)

    def get_available_payment_methods(self):
        return self.pc.get_available_payment_methods()

    def get_payment_method_js_module(self, polymorphic_type):
        model = get_models_of_type(PaymentMethod, exclude_parent=True, name=polymorphic_type)
        return model.js_module

    def get_payment_methods(self, order_id):
        assert order_id is not None
        order = self.aoc.get_order(order_id).pop()
        assert order is not None

        rv = []
        for pm in self.pc.get_payment_methods(order.merchant):
            rv.append({
                'id': pm.id,
                'name': pm.label,
                'js_module': pm.js_module
            })

        return rv

    def get_payment_method_state(self, payment_method_id):
        return self.pc.get_payment_method_state(payment_method_id)

    def get_payment_details(self, order_id, payment_method_id):
        order = self.load_order(order_id)
        pc = self._get_payment_controller_instance(payment_method_id)

        return {
            'order': order,
            'payment_vars': pc.get_payment_vars(order_id)
        }

    def submit_payment(self, order_id, payment_method_id):
        self.aoc.submit_payment(order_id, payment_method_id)
        return True

    def save_payment_method_for_merchant(self, polymorphic_type, values):
        pm = self.pc.save_or_update_payment_method_for_merchant(polymorphic_type, **values)
        return pm.id

    def save_payment_status(self, payment_status_id, form_values):

        order_id = int(form_values['order_id'])
        payment_method_id = None if not form_values.has_key('payment_method') else int(form_values['payment_method'])
        payment_status = int(form_values['payment_status'])
        description = form_values['description']

        if payment_method_id is None:
            pc = PaymentController(self.sa_session, account_id=self.account_id, request=self.request)
        else:
            pc = self._get_payment_controller_instance(payment_method_id)

        if payment_status == OrderPaymentStatus.PAYMENT_COMPLETED:
            pc.complete_payment(order_id, description)
        elif payment_status == OrderPaymentStatus.PAYMENT_REJECTED:
            pc.reject_payment(order_id, description)
        else:
            raise Exception('Not implemented payment_status: %s' % payment_status)

    def get_order_log(self, order_id):
        log_text = self.aoc.get_order_log(order_id)

        p = Parser()
        return p.render_to_html(log_text)


        return log_text

    def archive_payment_method(self, ids):
        pc = PaymentController(self.sa_session, account_id=self.account_id, request=self.request)
        pc.mark_as_archived(ids)


class OrderRuleHandler:

    @login_required
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']
        self.request = request

        self._controller = AbstractOrderController(self.sa_session, account_id = self.account_id, request=request)

    def get_order_types(self):
        return [{
            'label': x.name_i18n if x.name_i18n is not None else x.__name__,
            'value': x.__name__} for x in get_available_order_types(self._controller)]

    # rule

    def validate_order_rule(self, args):

        form_handler = FormValidator({
            'name': [(plugins.Required, {})],
            'condition': [(plugins.Required, {})],
            'action': [(plugins.Required, {})],
            'start_date': [(plugins.Required, {})],
        })
        return form_handler.validate(args)

    def load_order_rule(self, id):
        return self._controller.load_rule(id)

    def save_order_rule(self, id, state):
        if state['start_date'] == '':
            state['start_date'] = None
        if state['end_date'] == '':
            state['end_date'] = None
        state['match_all'] = bool(state['match_all'])
        return self._controller.save_rule(id, state).id

    def load_rule(self, id):
        return self._controller.load_rule(id)

    def delete_rule(self, ids):
        for id in ids:
            self._controller.delete_rule(id)

    # conditions

    def validate_order_rule_condition(self, args, cb_args):

        order_type = cb_args[0]

        validation = {
            'condition': [(plugins.Required, {})],
            #'name': [(plugins.Required, {})],
        }

        condition = None
        for c in self._controller.get_available_rule_conditions(order_type):
            if c.name == args['condition']:
                condition = c
        if condition is not None:

            if condition.show_select:
                validation['value_select'] = [(plugins.Required, {})]
            else:

                if condition.type == OrderRuleCondition.TYPE_TEXT:
                    validation['value_text'] = [(plugins.Required, {})]
                elif condition.type == OrderRuleCondition.TYPE_BOOLEAN:
                    validation['value_bool'] = [(plugins.Required, {})]
                elif condition.type == OrderRuleCondition.TYPE_INTEGER:
                    validation['value_int'] = [(plugins.Required, {})]
                elif condition.type == OrderRuleCondition.TYPE_DECIMAL:
                    validation['value_decimal'] = [(plugins.Required, {})]

        form_handler = FormValidator(validation)
        return form_handler.validate(args)

    def get_available_conditions(self, order_type):
        return [{
            'name': c.name,
            'label': c.label,
            'type': c.type,
            'show_operators': c.show_operators,
            'show_select': c.show_select,
        } for c in self._controller.get_available_rule_conditions(order_type)]

    def get_condition_select_options(self, filter_obj, order_type, name, order_form_values):
        lazy_assert(name, str)

        form_values = {}
        for o in order_form_values['conditions']:
            v = o['value_int'] or o['value_decimal'] or o['value_bool'] or o['value_text']
            form_values[o['condition']] = v

        for c in self._controller.get_available_rule_conditions(order_type):
            if c.name == name:
                options = c.get_select_options(form_values)
                return combobox_filter(options, filter_obj)

        raise Exception('invalid order_type: %s' % order_type)

    # actions

    def validate_order_rule_action(self, args, cb_args):

        order_type = cb_args[0]

        validation = {
            'action': [(plugins.Required, {})],
            #'name': [(plugins.Required, {})],
        }

        action = None
        for c in self._controller.get_available_rule_actions(order_type):
            if c.name == args['action']:
                action = c
        if action is not None:

            if action.show_select:
                validation['value_select'] = [(plugins.Required, {})]
            else:

                if action.type == OrderRuleAction.TYPE_NONE:
                    pass
                elif action.type == OrderRuleAction.TYPE_TEXT:
                    validation['value_text'] = [(plugins.Required, {})]
                elif action.type == OrderRuleAction.TYPE_BOOLEAN:
                    validation['value_bool'] = [(plugins.Required, {})]
                elif action.type == OrderRuleAction.TYPE_INTEGER:
                    validation['value_int'] = [(plugins.Required, {})]
                elif action.type == OrderRuleAction.TYPE_DECIMAL:
                    validation['value_decimal'] = [(plugins.Required, {})]

        form_handler = FormValidator(validation)
        return form_handler.validate(args)

    def get_available_actions(self, order_type):
        return [{
            'name': a.name,
            'label': a.label,
            'type': a.type,
            'show_select': a.show_select,
            'show_persistent': a.show_persistent,
        } for a in self._controller.get_available_rule_actions(order_type)]

    def get_action_select_options(self, order_type, name=None):
        lazy_assert(name, str)
        for c in self._controller.get_available_rule_actions(order_type):
            if c.name == name:
                return c.get_select_options()
        raise Exception('invalid order_type: %s' % order_type)

    def get_gl_account_code_options(self):

        gca = GLAccountController(self.sa_session, account_id = self.account_id, request=self.request)
        return [{'label': x[0], 'value': x[0]} for x in gca.get_distinct_gl_accounts()]

    def load_rule_execute_order(self, order_type):
        return {'execute_order': [x.id for x in self._controller.get_rules_by_execute_order(order_type)]}

    def get_rule_execute_order_options(self, order_type):
        return [{'label': '%s - %s' % (x.name, x.execute_order), 'value': x.id} for x in self._controller.get_rules_by_execute_order(order_type)]

    def save_rule_execute_order(self, order_type, values):
        self._controller.save_rule_execute_order(order_type, values['execute_order'])


class OrderSimulationHandler:

    @login_required
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']
        self.request = request

        self._controller = AbstractOrderController(self.sa_session, account_id = self.account_id, request=request)

    def get_order_types(self):
        return [{
            'label': x.name_i18n if x.name_i18n is not None else x.__name__,
            'value': x.__name__} for x in get_available_order_types(self._controller)]

    def get_line_types(self, order_type):
        order_types = get_available_order_types(self._controller)
        for OrderType in order_types:

            if OrderType.__name__ == order_type:
                rv = [];
                for order_line_type in OrderType.get_simulate_order_line_types():
                    rv.append({
                        'label': order_line_type[0],
                        'value': order_line_type[1],
                    })
                return rv

        raise Exception('invalid order_type: %s' % order_type)


    def validate_order_simulation(self, form_values):

        form_behaviour = {
            'name': [(plugins.Required, {})]
        }

        order_type = form_values.get('polymorphic_type', None)
        if order_type:
            for k,c in self._controller.get_simulate_order_fields(order_type).iteritems():
                if c.simulation_required:
                    form_behaviour[k] = [(plugins.Required, {})]

        form_handler = FormValidator(form_behaviour)
        return form_handler.validate(form_values)

    def validate_order_simulation_line(self, form_values, order_type):

        form_behaviour = {
            'polymorphic_type': [(plugins.Required, {})]
        }

        order_line_type = form_values.get('polymorphic_type', None)
        if order_line_type:
            for k,c in self._controller.get_simulate_order_line_fields(order_type, order_line_type).iteritems():
                if c.simulation_required:
                    form_behaviour[k] = [(plugins.Required, {})]


        form_handler = FormValidator(form_behaviour)
        return form_handler.validate(form_values)

    def get_order_fields(self, order_type):

        return [{
            'name': k,
            'label': c.label,
            'type': c.type,
            'show_operators': c.show_operators,
            'show_select': c.show_select,
        } for k,c in self._controller.get_simulate_order_fields(order_type).iteritems()]


    def get_order_line_fields(self, order_type, order_line_type):

        return [{
            'name': k,
            'label': c.label,
            'type': c.type,
            'show_operators': c.show_operators,
            'show_select': c.show_select,
        } for k,c in self._controller.get_simulate_order_line_fields(order_type, order_line_type).iteritems()]

    def get_order_field_select_options(self, filter_obj, order_type, field_name, form_values):
        lazy_assert(order_type, str)
        lazy_assert(field_name, str)

        for k, c in self._controller.get_simulate_order_fields(order_type).iteritems():
            if k == field_name:
                options = c.get_select_options(form_values)
                return combobox_filter(options, filter_obj)

        raise Exception('invalid: %s/%s' % (order_type, field_name) )

    def get_order_line_field_select_options(self, filter_obj, order_type, order_line_type, field_name, form_values):
        lazy_assert(order_type, str)
        lazy_assert(field_name, str)

        for k, c in self._controller.get_simulate_order_line_fields(order_type, order_line_type).iteritems():
            if k == field_name:
                options = c.get_select_options(form_values)
                return combobox_filter(options, filter_obj)

        raise Exception('invalid: %s/%s' % (order_type, field_name) )

    def save_order_simulation(self, simulation_id, state):
        simulation = self._controller.save_order_simulation(simulation_id, state)
        return simulation.id

    def load_order_simulation(self, simulation_id):
        return self._controller.load_order_simulation(simulation_id)

    def simulate_submit(self, simulation_id):

        order_state = self._controller.load_order_simulation(simulation_id)

        # name is only used for saving a simulation object and isnt expected
        # in the simulate state
        order_name = order_state['name']
        del order_state['name']

        return {
            'order_name': order_name,
            'order_result': self._controller.simulate_submit(order_state),
        }



handlers = (OrderHandler, TaxHandler, GLAccountHandler, OrderRuleHandler, OrderSimulationHandler)
