
from decimal import Decimal
from django.utils.translation import ugettext_lazy as _
from basesoft.web.contrib.order.models import OrderLineRule, OrderRuleAction
from basesoft.web.core.util.controllers import transaction_decorator

class RuleActionController(object):

    name = None
    label = None
    type = None

    show_select = False
    show_persistent = False

    def __init__(self, order_controller):
        from basesoft.web.contrib.order.controllers import AbstractOrderController
        assert isinstance(order_controller, AbstractOrderController)

        self._order_controller = order_controller
        self.logger = order_controller.logger

    def _rule_action_add_line(self, order, rule, action, quantity, price, dry_run):
        """
        handy method to add a single line to an order for a rule
        """

        order.add_log('Adding line with quantity %s, price %s, tax_class %s' % (quantity, price, action.tax_class) )
        apply_vat = self._order_controller._apply_vat_rate(order)

        tcc = None if action.tax_class is None else action.tax_class.code

        order_line = self._order_controller._save_or_update_order_line(order, len(order.lines), order.merchant, action.full_description,
                                        quantity=quantity, price=price, polymorphic_type=OrderLineRule, apply_vat=apply_vat,
                                        gl_account_code=action.gl_account_code, tax_class=action.tax_class, dry_run=dry_run,
                                        tax_class_code=tcc)

        return order_line

    def execute(self, order, rule, action, dry_run):
        raise Exception('abstract')


class RuleActionPersistentController(RuleActionController):
    show_persistent = True

    def _rule_action_add_line(self, order, rule, action, *args, **kwargs):
        order_line = super(RuleActionPersistentController, self)._rule_action_add_line(order, rule, action, *args, **kwargs)
        order_line.is_persistent = action.is_persistent

        if order_line.is_persistent:
            order_line.order.add_log('Persistent flag set this order line')

        return order_line


class RuleActionFixedAmountDiscountController(RuleActionPersistentController):

    name = 'fixed_amount_discount'
    label = _('Fixed amount discount')
    type = OrderRuleAction.TYPE_DECIMAL

    def execute(self, order, rule, action, dry_run):
        assert isinstance(action.value_decimal, Decimal)
        assert action.value_decimal > 0

        discount = action.value_decimal
        if order.total_amount_excluding_vat <= discount:
            order.add_log('Discount amount is larger than the order total amount; the discount has not been granted.')
            return

        self._rule_action_add_line(order, rule, action, -1, discount, dry_run)

class RuleActionFixedAmountPenaltyController(RuleActionPersistentController):

    name = 'fixed_amount_penalty'
    label = _('Fixed amount penalty')
    type = OrderRuleAction.TYPE_DECIMAL

    def execute(self, order, rule, action, dry_run):
        assert isinstance(action.value_decimal, Decimal)
        assert action.value_decimal > 0
        self._rule_action_add_line(order, rule, action, 1, action.value_decimal, dry_run)

class RuleActionRelativeAmountDiscountController(RuleActionPersistentController):

    name = 'relative_amount_discount'
    label = _('Relative amount discount')
    type = OrderRuleAction.TYPE_DECIMAL

    def execute(self, order, rule, action, dry_run):
        assert isinstance(action.value_decimal, Decimal)
        assert action.value_decimal > 0

        # relative discount is current total amount times discount percentage
        discount = order.total_amount_excluding_vat * action.value_decimal / Decimal('100')
        line = self._rule_action_add_line(order, rule, action, -1, discount, dry_run)
        assert isinstance(line, OrderLineRule)

        # update ex vat value
        line.line_amount_including_vat = self._order_controller._round_decimal(-1 * order.total_amount_including_vat * action.value_decimal / Decimal('100'))

class RuleActionRelativeAmountPenaltyController(RuleActionPersistentController):

    name = 'relative_amount_penalty'
    label = _('Relative amount penalty')
    type = OrderRuleAction.TYPE_DECIMAL

    def execute(self, order, rule, action, dry_run):
        assert isinstance(action.value_decimal, Decimal)
        assert action.value_decimal > 0

        # relative penalty is current total amount times discount percentage
        penalty = order.total_amount_excluding_vat * action.value_decimal / Decimal('100')
        self._rule_action_add_line(order, rule, action, 1, penalty, dry_run)


class RuleActionMinimumAmountController(RuleActionPersistentController):
    name = 'minimum_amount'
    label = _('Minimum amount')
    type = OrderRuleAction.TYPE_DECIMAL

    def execute(self, order, rule, action, dry_run):
        assert isinstance(action.value_decimal, Decimal)
        assert action.value_decimal > 0

        if order.total_amount_excluding_vat <= action.value_decimal:
            self._rule_action_add_line(order, rule, action, 1, action.value_decimal - order.total_amount_excluding_vat, dry_run)



