
from decimal import Decimal
from django.utils.translation import ugettext as _, ugettext_lazy
from basesoft.web.contrib.i18n.controllers import I18nController
from basesoft.web.contrib.license.controllers import LicenseController
from basesoft.web.contrib.order.models import OrderRuleCondition, Order
from basesoft.web.core.util.controllers import lazy_assert

class RuleConditionController(object):

    name = None
    label = None
    type = None

    show_operators = False
    show_select = False

    simulation_required = True

    def __init__(self, order_controller):
        from basesoft.web.contrib.order.controllers import AbstractOrderController
        assert isinstance(order_controller, AbstractOrderController)

        self._order_controller = order_controller
        self.logger = order_controller.logger

        assert self.name is not None
        assert self.label is not None
        assert self.type is not None

    def matches(self, order, rule, condition):
        raise Exception('abstract')

    def get_select_options(self, form_values):
        raise Exception('abstract')

    def _rule_condition_operator(self, value1, operator, value2):
        self.logger.info('_rule_condition_operator: comparing %s %s %s' % (value1, operator, value2))

        lazy_assert(value1, Decimal if OrderRuleCondition.TYPE_DECIMAL else int, {'zero': True, 'negative': True})
        lazy_assert(value2, Decimal if OrderRuleCondition.TYPE_DECIMAL else int, {'zero': True, 'negative': True})

        if operator == OrderRuleCondition.OPERATOR_EQUAL:
            return value1 == value2
        elif operator == OrderRuleCondition.OPERATOR_GT:
            return value1 > value2
        elif operator == OrderRuleCondition.OPERATOR_GTE:
            return value1 >= value2
        elif operator == OrderRuleCondition.OPERATOR_LT:
            return value1 < value2
        elif operator == OrderRuleCondition.OPERATOR_LTE:
            return value1 <= value2
        else:
            raise Exception('Unknown operator: %s' % operator)

    def simulate_to_object(self, value, order_state):
        return value


class RuleConditionMerchant(RuleConditionController):
    name = 'merchant'
    label = ugettext_lazy('Merchant')
    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def __init__(self, order_controller):
        super(RuleConditionMerchant, self).__init__(order_controller)
        self.label = order_controller._instance_class.MERCHANT_LABEL

    def matches(self, order, rule, condition):
        assert condition.value_int is not None
        return order.merchant.id == condition.value_int

    def get_select_options(self, form_values):
        lc = LicenseController(self._order_controller.sa_session, account_id=self._order_controller.get_account().id)
        return [{'label': l.name, 'value': l.id} for l in lc.get_domain_licenses()]

    def simulate_to_object(self, license_id, order_state):
        lazy_assert(license_id, int)

        lc = LicenseController(self._order_controller.sa_session, account_id=self._order_controller.get_account().id)
        licenses = lc.get_licenses(license_id)
        if len(licenses) == 0:
            raise Exception('Cant get license: %s' % license_id)
        return licenses[0]


class RuleConditionGroup(RuleConditionController):
    name = 'scope'
    label = ugettext_lazy('Group')
    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def matches(self, order, rule, condition):
        raise Exception('fixme: check if group is in groups of order.license')

    def get_select_options(self, form_values):
        lc = LicenseController(self._order_controller.sa_session, account_id=self._order_controller.get_account().id)
        return [{'label': l.name, 'value': l.id} for l in lc.get_groups()]

    def simulate_to_object(self, group_id, order_state):
        lc = LicenseController(self._order_controller.sa_session, account_id=self._order_controller.get_account().id)
        return lc.get_groups(group_id)[0]


class RuleConditionClient(RuleConditionController):
    name = 'client'
    label = ugettext_lazy('Client')
    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def __init__(self, order_controller):
        super(RuleConditionClient, self).__init__(order_controller)
        self.label = order_controller._instance_class.CLIENT_LABEL

    def matches(self, order, rule, condition):
        assert condition.value_int is not None
        return order.client.id == condition.value_int

    def get_select_options(self, form_values):
        lc = LicenseController(self._order_controller.sa_session, account_id=self._order_controller.get_account().id)
        return [{'label': l.name, 'value': l.id} for l in lc.get_domain_licenses()]

class RuleConditionClientCountry(RuleConditionController):
    name = 'client_country'
    label = ugettext_lazy('Client country')
    type = OrderRuleCondition.TYPE_INTEGER

    show_select = True

    def __init__(self, order_controller):
        super(RuleConditionClientCountry, self).__init__(order_controller)
        self.label = _('%(client)s country') % {'client': order_controller._instance_class.CLIENT_LABEL}

    def matches(self, order, rule, condition):
        assert condition.value_int is not None
        return order.client.company.addresses[0].address.country.id == condition.value_int

    def get_select_options(self, form_values):
        i18nc = I18nController(self._order_controller.sa_session, account_id=self._order_controller.get_account().id)
        for c in i18nc.get_domain_countries():
            print 'CC', c
        return [{'label': c.name if dc is None else dc.name, 'value': c.id} for c, dc in i18nc.get_domain_countries()]

class RuleConditionAutoSubmitted(RuleConditionController):
    name = 'auto_submitted'
    label = ugettext_lazy('Auto-submitted')
    type = OrderRuleCondition.TYPE_INTEGER

    show_operators = False
    show_select = True

    def matches(self, order, rule, condition):
        assert condition.value_int in (Order.SUBMIT_DONE_BY_USER, Order.SUBMIT_DONE_BY_SYSTEM)
        return order.auto_submitted == condition.value_int

    def get_select_options(self, form_values):
        return [
            {'label': _('User'), 'value': Order.SUBMIT_DONE_BY_USER},
            {'label': _('System'), 'value': Order.SUBMIT_DONE_BY_SYSTEM},
        ]

class RuleConditionIsResubmit(RuleConditionController):
    name = 'is_resubmit'
    label = ugettext_lazy('Is resubmit')
    type = OrderRuleCondition.TYPE_BOOLEAN

    def matches(self, order, rule, condition):
        assert condition.value_bool is not None

        if condition.value_bool:
            # condition: must be resubmit
            return order.parent_order is not None
        else:
            # condition: must NOT be resubmit
            return order.parent_order is None

class RuleConditionTotalAmountExclVat(RuleConditionController):
    name = 'total_amount_excl_vat'
    label = ugettext_lazy('Total amount (ex VAT)')
    type = OrderRuleCondition.TYPE_DECIMAL

    show_operators = True

    def matches(self, order, rule, condition):
        assert condition.value_decimal is not None
        return self._rule_condition_operator(order.total_amount_excluding_vat, condition.operator, condition.value_decimal)

class RuleConditionTotalAmountInclVat(RuleConditionController):
    name = 'total_amount_incl_vat'
    label = ugettext_lazy('Total amount (incl VAT)')
    type = OrderRuleCondition.TYPE_DECIMAL

    show_operators = True

    def matches(self, order, rule, condition):
        assert condition.value_decimal is not None
        return self._rule_condition_operator(order.total_amount_including_vat, condition.operator, condition.value_decimal)

#
# LINE CONDITIONS
#

class RuleConditionLineQuantity(RuleConditionController):
    name = 'quantity'
    label = ugettext_lazy('Line quantity')
    type = OrderRuleCondition.TYPE_INTEGER
    show_operators = True

    def matches(self, order, rule, condition):
        lazy_assert(condition.value_decimal, Decimal)

        for line in order.lines:
            if self._rule_condition_operator(line.quantity, condition.operator, condition.value_int):
                return True
        return False


class RuleConditionLineDescription(RuleConditionController):
    name = 'description'
    label = ugettext_lazy('Line description')
    type = OrderRuleCondition.TYPE_TEXT

    def matches(self, order, rule, condition):
        raise Exception('can be used in simulation only')

class RuleConditionLinePrice(RuleConditionController):
    name = 'price'
    label = ugettext_lazy('Line price')
    type = OrderRuleCondition.TYPE_DECIMAL

    show_operators = True

    def matches(self, order, rule, condition):
        lazy_assert(condition.value_decimal, Decimal)

        for line in order.lines:
            if self._rule_condition_operator(line.price, condition.operator, condition.value_decimal):
                return True
        return False
