'''
Created on Oct 16, 2013

@author: marten
'''
from datetime import date, datetime

from pyws.functions.register import register

from basesoft.web.contrib.soap import soap_assert_logged_in
from basesoft.web.core.exceptions import SystemException, PublicUserException
from basesoft.web.contrib.soap.args import _types
from pyws.functions.args.types.simple import DateTime
from basesoft.web.contrib.order.controllers import PaymentController,\
    AbstractOrderController
from basesoft.web.contrib.order.models import OrderPaymentStatus
from basesoft.logging import getLogger

def get_logger(context):
    return getLogger('basesoft.web.contrib.order', request=context)

def register_soap_functions(server, controller=None):
    """
    Registers order soap functions to the supplied server.
    """
    if controller is None:
        raise SystemException('Abstract SOAP-Order methods. Please provide a valid Order Controller')

    @register(return_type=_types.OrderDataDict, args=((int, None, ), ), to=server, needs_context=True)
    def get_order_by_id(order_id, context):
        """
        Returns a single order for the provided ID.
        Raises an exception if the current logged in user has no rights to the order.

        payment_status is defined as following:
          PAYMENT_PENDING = 0
          PAYMENT_SUBMITTED = 1 # has done initial payment, but unconfirmed by third party
          PAYMENT_PARTIALLY_RECEIVED = 2 # part of the order has been paid
          PAYMENT_COMPLETED = 3
          PAYMENT_REJECTED = 4 # Direct Debit or Credit Card denial
        """

        get_logger(context).info('soap::get_order_by_id(%s)' % order_id)

        if not order_id:
            raise PublicUserException('Missing required field "order_id"') # TODO: fix in WSDL

        soap_assert_logged_in(context=context)

        c = controller(context.Session(), account_id=context.session['user_id'], request=context)
        return c.get_exported_orders_by_id(order_id)

    @register(return_type=[_types.OrderDataDict], args=((int, None, ), ), to=server, needs_context=True)
    def get_unexported_orders_for_merchant(merchant_id, context):
        """
        Returns a list with Orders for the supplied Merchant where the current logged in user has rights to. Also see get_order_by_id()
        """
        get_logger(context).info('soap::get_unexported_orders_for_merchant(%s)' % merchant_id)

        if not merchant_id:
            raise PublicUserException('Missing required field "merchant_id"') # TODO: fix in WSDL

        soap_assert_logged_in(context=context)

        c = controller(context.Session(), account_id=context.session['user_id'], request=context)
        return c.get_unexported_orders_for_merchant(merchant_id, True)

    @register(return_type=[_types.OrderDataDict], args=((int, None, ), (date, None, ) ), to=server, needs_context=True)
    def get_orders_by_export_date(merchant_id, export_date, context):
        """
        Returns a list with Orders which were previously exported on the supplied Date for the supplied Merchant. Also see get_order_by_id()
        """
        get_logger(context).info('soap::get_orders_by_export_date(%s, %s)' % (merchant_id, export_date))

        if not merchant_id or not export_date:
            raise PublicUserException('Missing required field "merchant_id" and "export_date"') # TODO: fix in WSDL

        soap_assert_logged_in(context=context)

        c = controller(context.Session(), account_id=context.session['user_id'], request=context)
        return c.get_orders_by_export_date(merchant_id, export_date, True)

    @register(return_type=bool, args=((int, None, ), (_types.Document, None, ), (bool, False, ), ), to=server, needs_context=True)
    def attach_invoice_to_order(order_id, attachment, is_credit, context):
        """
        Attaches the provided file to the order id supplied.
        """
        get_logger(context).info('soap::attach_invoice_to_order(%s)' % (order_id))

        if not order_id or not attachment:
            raise PublicUserException('Missing required field "order_id" and "attachment"') # TODO: fix in WSDL

        soap_assert_logged_in(context=context)

        c = controller(context.Session(), account_id=context.session['user_id'], request=context)
        return c.attach_invoice_to_order(order_id, attachment['file_name'], attachment['payload'], is_credit)

    @register(return_type=bool, args=((int, None, ), (int, None, ), (str, None, ), (str, False, ), (DateTime, False, ), ), to=server, needs_context=True)
    def update_order_payment(order_id, payment_status_id, payment_method_type, description, payment_date, context):

        get_logger(context).info('soap::update_order_payment(%s, %s)' % (order_id, payment_status_id))

        sa_session = context.Session()
        account_id = context.session['user_id']

        # get the order so we can get the merchant
        aoc = AbstractOrderController(sa_session, account_id=account_id, request=context)
        try:
            order = aoc.get_order(order_id)[0]
        except IndexError:
            raise PublicUserException('Order not found: %s' % order_id)

        # init abstract payment controller to get payment_method object and controller
        pc = PaymentController(sa_session, account_id=account_id, request=context)

        # try to get a payment method from payment_method_type. This can be empty for credits
        payment_method = None
        if payment_method_type != None and payment_method_type != '':
            try:
                payment_method = pc.get_payment_methods(order.merchant, payment_method_polymorphic_type=payment_method_type)[0]
            except IndexError:
                raise PublicUserException('Payment method %s not found for order %s' % (payment_method_type, order.id))

        if not description:
            description = ''

        # instanciate specific payment_method controller for the payment method
        if payment_method is None:
            pc = PaymentController(sa_session, account_id=account_id, request=context)
        else:
            pc = pc.get_payment_controller(payment_method.id)(sa_session, account_id=account_id, request=context, payment_method=payment_method)

        if payment_status_id == OrderPaymentStatus.PAYMENT_COMPLETED:
            pc.complete_payment(order_id, description, payment_date)
        elif payment_status_id == OrderPaymentStatus.PAYMENT_REJECTED:
            pc.reject_payment(order_id, description, payment_date)
        else:
            raise PublicUserException('Invalid payment_status_id: %s' % payment_status_id)



