# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

import time

from django.conf import settings
from django.utils.cache import patch_vary_headers
from django.utils.http import cookie_date

class SASessionMiddleware(object):
    """
    This middleware is based on the exisiting django session
    middleware, but implements db storage with sqlalchemy
    """
    def process_request(self, request):
        """
        Make the session dict available in request object
        """
        session_key = request.COOKIES.get(settings.SESSION_COOKIE_NAME, None)

        # use redis or db session storage
        if settings.REDIS_HOST is not None:
            from basesoft.web.contrib.session.engines.redis_store import SessionStore as RedisSessionStore
            request.session = RedisSessionStore(session_key)
        else:
            from basesoft.web.contrib.session.engines.db_store import SessionStore as DbSessionStore
            request.session = DbSessionStore(session_key, SASessionMiddleware.get_client_ip(request))

        request.META['IQPORTAL_VERSION'] = settings.VERSION

    def process_response(self, request, response):
        """
        If request.session was modified, or if response.session was set, save
        those changes and set a session cookie.
        """
        try:
            accessed = request.session.accessed
            modified = request.session.modified
        except AttributeError:
            pass
        else:
            if accessed:
                patch_vary_headers(response, ('Cookie',))
            if modified:
                if settings.SESSION_EXPIRE_AT_BROWSER_CLOSE:
                    max_age = None
                    expires = None
                else:
                    if (request.session.has_key('keep_logged_in') and request.session['keep_logged_in']) or \
                        (request.session.has_key('impersonator_keep_logged_in') and request.session['impersonator_keep_logged_in']):
                        max_age = settings.KEEP_SIGNED_COOKIE_AGE
                    else:
                        max_age = settings.SESSION_COOKIE_AGE
                    expires_time = time.time() + max_age
                    expires = cookie_date(expires_time)
                # Save the seesion data and refresh the client cookie.
                request.session.save()
                response.set_cookie(settings.SESSION_COOKIE_NAME,
                        request.session.session_key, max_age=max_age,
                        expires=expires, domain=settings.SESSION_COOKIE_DOMAIN,
                        path=settings.SESSION_COOKIE_PATH,
                        secure=settings.SESSION_COOKIE_SECURE or None)

        # close the database for connection pooling to happen
        request.Session().close()

        return response

    @staticmethod
    def get_client_ip(request):
        # X_FORWARDED_FOR & handling of proxies that sometimes return multiple IP's in the HTTP_X_FORWARDED_FOR header.
        # The real IP is usually at the end of the list, as explained here: http://en.wikipedia.org/wiki/X-Forwarded-For
        x_forwarded_for = request.META.get('HTTP_X_FORWARDED_FOR')
        if x_forwarded_for:
            ip = x_forwarded_for.split(',')[-1].strip()
        else:
            ip = request.META.get('REMOTE_ADDR')
        return ip
