# Copyright (c) 2013 BaseSoft B.V.
# See LICENSE for details.
import sys
import random
import traceback

from hashlib import sha256 as sha
from basesoft.web.core.exceptions import PublicException
from django.conf import settings

from functools import partial

from pyws.protocols.soap import SoapProtocol, get_context_data_from_headers
from pyws.functions.args import TypeFactory

def get_context_data(request, headers_schema=None):
    rv = get_context_data_from_headers(request, headers_schema)
    if rv is None:
        return request
    return rv

class IQPortalSoapProtocol(SoapProtocol):
    def __init__(
            self, service_name, tns, location,
            headers_schema=None, *args, **kwargs):
        """
        ``service_name`` gives a name to the server, it is used to generate
        WSDL file. ``tns`` is a root namespace, where all the stuff lives, SOAP
        is based on XML, so no wonder we need it. ``location`` tells the
        server, where it lives. It is only a piece of information required by
        pyws to generate a proper WSDL file, it is not a real binding to an
        URL. ``headers_schema`` a schema used to retrive context data from
        request headers. Other arguments are passed to parent constructor.
        """

        headers_schema = headers_schema and TypeFactory(headers_schema)
        context_data_getter = partial(get_context_data, headers_schema=headers_schema)

        super(SoapProtocol, self).\
            __init__(context_data_getter, *args, **kwargs)

        self.service_name = service_name
        self.tns = tns
        self.location = location
        self.headers_schema = headers_schema

    def get_error_response(self, error, request):
        if isinstance(error, PublicException) or settings.DEBUG:
            return super(IQPortalSoapProtocol, self).get_error_response(error)

        exception_hash = sha(str(random.random())).hexdigest()[:8]

        from django.core.mail import mail_admins

        exc_info = sys.exc_info()

        subject = '#%s: Error (%s IP): %s' % (exception_hash, (request.META.get('REMOTE_ADDR') in settings.INTERNAL_IPS and 'internal' or 'EXTERNAL'), request.META.get('HTTP_SOAPACTION'))
        request_repr = unicode(request)

        session_data = ""
        for k in request.session.keys():
            session_data += "%s: %s\n" % (k, request.session.get(k))

        traceback_str = '\n'.join(traceback.format_exception(*(exc_info or sys.exc_info())))
        message = "%s\n\n%s\n\n%s" % (traceback_str, session_data, request_repr)
        mail_admins(subject, message, fail_silently=True)

        error.hash = exception_hash
        return super(IQPortalSoapProtocol, self).get_error_response(error)

    def get_error(self, error):
        rv = super(IQPortalSoapProtocol, self).get_error(error)
        if not isinstance(error, PublicException) and hasattr(error, 'hash'):
            rv['name'] = '%s (#%s)' % (rv['name'], error.hash)
        return rv
