# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

import datetime, time

from django.conf import settings
from django.utils.translation import ugettext as _

from basesoft.web.core.util.custom_query import CustomQuery, CustomQueryField, CustomQueryModel, SQLCustomQuery,\
    SQLCustomQueryTable, SQLCustomQueryField, DATA_TYPE_DATE, DATA_TYPE_BOOLEAN
from basesoft.web.contrib.ticketing.models import Ticket
from basesoft.web.contrib.account.models import Account
from basesoft.web.contrib.ticketing.controllers import TicketController
from basesoft.web.core.util.controllers import AccountConfigController

class TicketCustomQuery(CustomQuery):

    def tables(self):
        return (
            CustomQueryModel(Ticket),
            CustomQueryModel(Account, use_outerjoin=True),
        )
    def fields(self):
        return (
            CustomQueryField(Ticket.id, _('ID'), key='ticket_id'),
            CustomQueryField(Account.name, _('Reported by'), key='account_name'),
            CustomQueryField(Ticket.trac_ticket_id, _('Trac ticket id'), key='trac_ticket_id'),
            CustomQueryField(Ticket.trac_ticket_id, _('Summary'), key='summary'),
            CustomQueryField(Ticket.trac_ticket_id, _('Status'), key='status'),
            CustomQueryField(Ticket.created, _('Created')),
            CustomQueryField(Ticket.last_updated, _('Last updated')),
        )

class TicketSQLCustomQuery(SQLCustomQuery):

    connstr = settings.TRAC_DB_CONNSTR

    def tables(self):
        return (
            SQLCustomQueryTable('ticket', alias='t'),
            SQLCustomQueryTable('ticket_custom', alias='estimated', use_outerjoin=True, onclause=[('t', 'id', 'estimated', 'ticket', ), ('estimated', 'name', 'estimatedhours')]),
            SQLCustomQueryTable('ticket_custom', alias='escalation', use_outerjoin=True, onclause=[('t', 'id', 'escalation', 'ticket', ), ('escalation', 'name', 'escalation_id')]),
            SQLCustomQueryTable('ticket_custom', alias='testing', use_outerjoin=True, onclause=[('t', 'id', 'testing', 'ticket', ), ('testing', 'name', 'available_for_testing')]),
            SQLCustomQueryTable('ticket_custom', alias='approved', use_outerjoin=True, onclause=[('t', 'id', 'approved', 'ticket', ), ('approved', 'name', 'verified_and_approved')]),
        )

    def fields(self):
        return (
            SQLCustomQueryField('t', 'id', 'new', key='is_new'),
            SQLCustomQueryField('t', 'id', _('ID'), key='ticket_id', is_filter=True),
            SQLCustomQueryField('t', 'summary', _('Summary'), key='summary', is_filter=True),
            SQLCustomQueryField('t', 'status', _('Status'), is_filter=True),
            SQLCustomQueryField('t', 'version', _('Version'), is_filter=True),
            SQLCustomQueryField('testing', 'value', _('Testable'), key='available_for_testing', _type=DATA_TYPE_BOOLEAN, is_filter=True),
            SQLCustomQueryField('approved', 'value', _('Approved'), key='verified_and_approved', _type=DATA_TYPE_BOOLEAN, is_filter=True),
            SQLCustomQueryField('estimated', 'value', _('Estimated hours'), key='estimated_hours'),
            SQLCustomQueryField('escalation', 'value', _('Escalation ID'), key='escalation_id'),
            SQLCustomQueryField('t', 'changetime', _('Last updated'), key='last_updated', _type=DATA_TYPE_DATE),
            SQLCustomQueryField('t', 'time', _('Created'), key='created', _type=DATA_TYPE_DATE),
        )

    def set_value_processors(self):
        acc = AccountConfigController(self.sa_session, account_id=self.get_account().id)
        last_ticket_retrieval = acc.get('basesoft.web.contrib.ticketing.last.ticket.retrieval')

        def bool_proc(row, key):
            return False if row[key] is None else row[key]

        def dt_proc(row, key):
            if row[key] is None:
                return None
            return datetime.datetime.fromtimestamp(row[key])

        def is_new_proc(row, key):
            if last_ticket_retrieval == 0:
                return True
            return row['last_updated'] > last_ticket_retrieval

        self._value_processors['created'] = dt_proc
        self._value_processors['last_updated'] = dt_proc
        self._value_processors['available_for_testing'] = bool_proc
        self._value_processors['verified_and_approved'] = bool_proc
        self._value_processors['is_new'] = is_new_proc

    def get_data(self, split_by, sort, asc, offset, limit, applied_filters, static_filters, row_count=True):

        # load the normal session as we will be writing the config
        from basesoft.web.core.util.db import Session
        write_sa = Session()

        acc = AccountConfigController(write_sa, account_id=self.get_account().id)
        acc.set('basesoft.web.contrib.ticketing.last.ticket.retrieval', int(time.time()))

        return super(TicketSQLCustomQuery, self).get_data(split_by, sort, asc, offset, limit, applied_filters, static_filters, row_count)

    def set_where(self):
        c = TicketController(self.sa_session, account_id=self.get_account().id)
        tickets = c.get_tickets()
        if len(tickets) > 0:
            params = [str(t.trac_ticket_id) for t in tickets]
            self.add_where('`t`.`id` IN (%s)' % ', '.join(params), None)
        else:
            self.add_where('ISNULL(`t`.`id`)')

    group_by = (('t', 'id'), )

custom_queries = (TicketCustomQuery, TicketSQLCustomQuery, )
