# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.

from basesoft.web.contrib.i18n.models import Locale
from basesoft.web.contrib.wiki.models import Wiki, WikiContent
from basesoft.web.core.util.controllers import transaction_decorator
from sqlalchemy.orm.exc import NoResultFound
from types import NoneType
from basesoft.web.contrib.wiki.parser import Parser


class WikiController:
    """
    WikiController provides wiki methods for saving and loading wiki texts.
    Should be used as a 2nd inheritance class where the child class should implement certain methods
    for retrieving the .wiki object depending on permissions/privileges.

    instance_id in the methods refer to the ID of the object to where the wiki object is attached.
    For example in Announcement there is a .wiki, instance_id is the announcement.id
    In the future Mail can do the same when there is a .wiki on Mail

    """
    def __init__(self):
        raise Exception('Abstract class, can only be used as 2nd inheritance')

    def get_wiki_from_instance(self, instance_id):
        """
        used to get a wiki object from the instance, this is used to edit (load/save) the wiki object itself
        """
        raise Exception('abstract')

    def get_wiki_from_instance_for_reading(self, instance_id):
        """
        retrieves the wiki object from the instance in context where the user is only allowed to view the text
        """
        raise Exception('abstract')


    def load_wiki(self, instance_id, locale_id):
        """
        returns the wiki state for the given instance and locale
        """
        return self._load_wiki(self.get_wiki_from_instance(instance_id), locale_id)

    def load_wiki_translate(self, instance_id):
        """
        returns initial wiki state for translate interface
        """
        wiki = self.get_wiki_from_instance(instance_id)
        locale_id = wiki.fallback_locale.id
        wiki_content = self._get_current_wiki_content(wiki, wiki.fallback_locale)

        # wiki content can be None if the fallback locale was changed without content being saved.
        return {
            'src_locale_id': locale_id,
            'src_title': '' if wiki_content is None else wiki_content.title,
            'src_content': '' if wiki_content is None else wiki_content.content,
            'locale_id': locale_id,
            'title': '' if wiki_content is None else wiki_content.title,
            'content': '' if wiki_content is None else wiki_content.content,
        }

    def add_macros(self, parser):
        # override in child classes
        return

    def load_wiki_with_fallback(self, instance_id, locale):
        """
        returns the parsed wiki html for the given intance object and locale.
        Takes care of fallback handling
        """

        wiki = self.get_wiki_from_instance_for_reading(instance_id)
        assert isinstance(locale, Locale)

        wiki_content = self._load_wiki(wiki, locale)
        wiki_content_fallback = self._load_wiki(wiki, wiki.fallback_locale)

        title = wiki_content['title']
        content = wiki_content['content']

        if title.strip() == "": title = wiki_content_fallback['title']
        if content.strip() == "": content = wiki_content_fallback['content']

        p = Parser()
        self.add_macros(p)
        content = p.render_to_html(content)

        return {'title': title, 'content': content}

    def _load_wiki(self, wiki, locale=None):
        """
        Loads a wiki state for editing the wiki. If no locale is provided, the wiki fallback locale is used.
        Returns empty texts if no content is found.

        May be called by the child class to edit the wiki text.
        """
        assert isinstance(wiki, Wiki)

        if locale is None:
            locale = wiki.fallback_locale
        elif isinstance(locale, (int, long)):
            locale = self.sa_session.query(Locale).filter(Locale.id == locale).one()
        assert isinstance(locale, Locale)

        content = self._get_current_wiki_content(wiki, locale)

        if content is None:
            return {
            'fallback_locale_id': wiki.fallback_locale.id,
            'locale_id': locale.id,
            'title': '',
            'content': '',
        }

        return {
            'fallback_locale_id': wiki.fallback_locale.id,
            'locale_id': locale.id,
            'title': content.title,
            'content': content.content,
        }

    def _get_current_wiki_content(self, wiki, locale):
        """
        Returns the current WikiContent object for the given wiki and locale, None if not found
        """

        assert isinstance(wiki, Wiki)
        assert isinstance(locale, Locale)

        try:
            return self.sa_session.query(WikiContent).filter(
                WikiContent.wiki == wiki,
                WikiContent.locale == locale
            ).order_by(WikiContent.revision.desc()).first()
        except NoResultFound:
            return None

    def _get_wiki_history(self, wiki, locale, limit=None):
        """
        Returns the history (all revisions) of the wiki page for the given locale.
        """

        assert isinstance(wiki, Wiki)
        assert isinstance(locale, Locale)

        q = self.sa_session.query(WikiContent).filter(
            WikiContent.wiki == wiki,
            WikiContent.locale == locale
        ).order_by(WikiContent.revision.desc())
        if limit is not None:
            q = q.limit(limit)
        return q

    def save_wiki_revision(self, instance_id, wiki_state):
        """
        Saves a wiki state
        """
        return self.__save_wiki_revision(self.get_wiki_from_instance(instance_id), wiki_state)

    @transaction_decorator
    def _save_wiki(self, wiki, wiki_state):
        """
        Saves a wiki object from the given state. May be called by child class to save a wiki object on its instance.
        """

        fallback_locale = self.sa_session.query(Locale).filter(Locale.id == wiki_state['fallback_locale_id']).one()

        assert isinstance(wiki, (Wiki, NoneType))
        assert isinstance(fallback_locale, Locale)

        if wiki is None:
            wiki = Wiki()
        wiki.fallback_locale = fallback_locale

        self.sa_session.add(wiki)

        self.__save_wiki_revision(wiki, wiki_state)

        return wiki

    @transaction_decorator
    def __save_wiki_revision(self, wiki, state):
        """
        Saves a wiki revision object from the given state. Does not create a new revision if the texts are the same as current
        """
        assert isinstance(wiki, Wiki)

        assert isinstance(state['locale_id'], (int, long))
        assert isinstance(state['title'], (str, unicode))
        assert isinstance(state['content'], (str, unicode))

        locale = self.sa_session.query(Locale).filter(Locale.id == state['locale_id']).one()
        content = self._get_current_wiki_content(wiki, locale)

        if content is None or content.content != state['content'] or content.title != state['title']:
            c = WikiContent()
            c.wiki = wiki
            c.locale = locale
            c.account = self._get_editor_account()
            c.revision = 1 if content is None else content.revision + 1
            c.title = state['title']
            c.content = state['content']
            self.sa_session.add(c)
            content = c
        return content

    def _get_editor_account(self):
        return None


