# Copyright (c) 2013 BaseSoft B.V.
# See LICENSE for details.

"""
Based on Trac wiki
"""
import locale

import re
import os
import datetime

from decimal import Decimal, InvalidOperation

from basesoft.web.contrib.wiki.provider import WikiProvider
from basesoft.web.core.exceptions import PublicUserException

from django.utils.translation import ugettext as _
from django.utils.dateformat import DateFormat
from django.utils.formats import get_format
from django.conf import settings

class Formatter:
    """
    """
    def __init__(self, macros=None, providers=None, context=None):
        """
        """

        if providers is None:
            providers = self._get_providers()
        self.providers = providers
        self.macros = macros
        self.context = context
        self._available_wikis = None
        self._words = None
        self._highlight_iterator = {}
        self.reset()

    def set_available_wikis(self, wikis):
        self._available_wikis = wikis

    def set_highlight_words(self, words):
        self._words = words

    def has_wiki(self, wiki):
        if self._available_wikis is None:
            return True
        return wiki in self._available_wikis

    def _get_providers(self):
        """
        """
        return [WikiProvider()]

    def get_highlight_iterator(self):
        return self._highlight_iterator

    def set_highlight_iterator(self, hl_iterator):
        self._highlight_iterator = hl_iterator

class HTMLFormatter(Formatter):
    """
    """

    def reset(self):
        """
        """
        self._open_tags = []
        self._anchors = {}
        self.paragraph_open = False
        self.in_list_item = False
        self.in_code_block = 0
        self.in_center = 0
        self.in_table = 0
        self.in_table_row = 0
        self.in_table_cell = 0
        self._list_stack = []
        self._wikis = []
        self._highlight_iterator = {}

    def close(self, items):
        """
        """
        rv = ''
        for item in items:
            close = getattr(self,'close_%s' % item)
            rv += close()
        return rv

    def open_paragraph(self):
        if not self.paragraph_open:
            self.paragraph_open = True
            return '<p>' + os.linesep
        return ''

    def close_paragraph(self):
        """
        """
        if self.paragraph_open:
            rv = ''
            while self._open_tags != []:
                rv += self._open_tags.pop()[1]
            self.paragraph_open = False
            return rv + '</p>' + os.linesep
        return ''

    def replace(self, fullmatch):
        """
        """
        for itype, match in fullmatch.groupdict().items():
            if match is None:
                continue
            if match[0] == '!':
                continue
            if itype in ['hdepth', 'lns', 'ltgt', 'label','macroname', 'inline', 'inline2','ldepth', 'property', 'definition']:
                continue
            if not hasattr(self, '_%s_formatter' % itype):
                raise PublicUserException(_('Incorrect usage of the wiki-editor. (%s)' % itype))
            handler = getattr(self, '_%s_formatter' % itype)
            return handler(match, fullmatch)

    def _lowerthan_formatter(self, match, fullmatch):
        return '&lt;'

    def _bolditalic_formatter(self, match, fullmatch):
        italic = ('<em>', '</em>')
        italic_open = self.has_open_tag(italic)
        tmp = ''
        if italic_open:
            tmp += italic[1]
            self.close_tag(italic[1])
        tmp += self._bold_formatter(match, fullmatch)
        if not italic_open:
            tmp += italic[0]
            self.open_tag(*italic)
        return tmp

    def _italic_formatter(self, match, fullmatch):
        return self.simple_tag_handler(match, '<em>', '</em>')

    def _bold_formatter(self, match, fullmatch):
        return self.simple_tag_handler(match, '<strong>', '</strong>')

    def _highlight_formatter(self, match, fullmatch):
        word = fullmatch.group('hlword')
        if self._highlight_iterator.has_key(word.lower()):
            counter = self._highlight_iterator[word.lower()]
        else:
            counter = len(self._highlight_iterator)
            self._highlight_iterator[word.lower()] = counter

        return '<span class="highlight highlight-%d">%s</span>' % (counter, word)

    def _underline_formatter(self, match, fullmatch):
        return self.simple_tag_handler(match, '<u>', '</u>')

    def _strike_formatter(self, match, fullmatch):
        return self.simple_tag_handler(match, '<del>', '</del>')

    def _center_formatter(self, match, fullmatch):
        if self.in_center:
            self.in_center = 0
            return '</center>'

        self.in_center = 1

        return '<center>'

    def _subscript_formatter(self, match, fullmatch):
        return self.simple_tag_handler(match, '<sub>', '</sub>')

    def _superscript_formatter(self, match, fullmatch):
        return self.simple_tag_handler(match, '<sup>', '</sup>')

    def _pagebreak_formatter(self, match, fullmatch):
        return '<div style="page-break-after:always;" class="print-page-break"></div><div class="page"></div>'

    def _inlinecode_formatter(self, match, fullmatch):
        return '<tt>%s</tt>' % fullmatch.group('inline')

    def _inlinecode2_formatter(self, match, fullmatch):
        return '<tt>%s</tt>' % fullmatch.group('inline2')

    def _feed_formatter(self, match, fullmatch):

        link = '<a href="/%s">%s</a>' % (match, match)
        return '<span class="at-name">@%s</span>' % link

    def _hashtag_formatter(self, match, fullmatch):

        link = '<a href="/search?q=%s">%s</a>' % (match, match)
        return '<span class="hashtag">#%s</span>' % link

    def _lhref_formatter(self, match, fullmatch):
        """
        """
        link = fullmatch.group('ltgt')
        format = fullmatch.group('lns')
        label = fullmatch.group('label') or ('http:%s' % link if format == 'http' else link)

        if format == 'wiki':
            cls = ''
            self._wikis.append(link)
            if not self.has_wiki(link):
                cls = ' class="unavailable"'

            return '<a href="#m=wiki-%s"%s>%s</a>' % (link, cls, label)
        elif format == 'http':
            return '<a href="http:%s" target="_blank">%s</a>' % (link, label)

    def _rel_formatter(self, match, fullmatch):
        link = fullmatch.group('ltgt')
        label = fullmatch.group('label') or 'http:%s' % link
        return '<a href="%s">%s</a>' % (match, label)

    def _parse_heading(self, match, fullmatch, shorten):
        match = match.strip()

        depth = min(len(fullmatch.group('hdepth')), 5)
        anchor = fullmatch.group('hanchor') or ''
        heading_text = match[depth+1:-depth-1-len(anchor)]

        rules = self.providers[0].get_replacement_rules()
        heading_text = re.sub(rules, self.replace, heading_text)

        heading = heading_text

        if anchor:
            anchor = anchor[1:]

        i = 1
        anchor_base = anchor
        while anchor in self._anchors:
            anchor = anchor_base + str(i)
            i += 1
        self._anchors[anchor] = True

        return (depth, heading, anchor)

    def _heading_formatter(self, match, fullmatch):
        depth, heading, anchor = self._parse_heading(match, fullmatch, False)

        fmt = OneLinerFormatter()
        fmt.set_highlight_iterator(self.get_highlight_iterator())
        heading = fmt.format(heading)

        return '<h%d id="%s">%s</h%d>' % (depth, anchor, heading.strip(), depth)

    def _macro_formatter(self, match, fullmatch):
        """
        """
        name = fullmatch.group('macroname')
        if name.lower() == 'br':
            if self.in_table and not self.in_table_cell:
                return ''
            return '<br />'

        args = fullmatch.group('macroargs') or ''

        if not self.macros.has_key(name):
            return '<div class="wiki-error">Unknown Macro %s</div>' % name

        m = self.macros[name](self.context, self)
        return m.execute(args)

    def _replacement_formatter(self, match, fullmatch):
        if self.context is None:
            return '<span class="wiki-error">No context provided</span>'

        assert issubclass(self.context.__class__, object)

        _object = fullmatch.group('replacement')
        _property = fullmatch.group('property') or None

        if hasattr(self.context, _object):
            obj = getattr(self.context, _object)
            if _property is None:
                return '%s' % obj
            if hasattr(obj, _property):
                rv = getattr(obj, _property)

                if isinstance(rv, datetime.date):
                    df = DateFormat(rv)
                    rv = df.format(get_format('SHORT_DATE_FORMAT'))

                if rv is None:
                    return ''
                return '%s' % rv
        return '<span class="wiki-error">Unsupported replacement field used `%s`.</span>' % _property

    def _list_formatter(self, match, fullmatch):
        """
        """
        ldepth = len(fullmatch.group('ldepth'))
        listid = match[ldepth]
        self.in_list_item = True
        class_ = start = None
        if listid in '-*':
            type_ = 'ul'
        else:
            type_ = 'ol'
            idx = '01iI'.find(listid)
            if idx >= 0:
                class_ = ('arabiczero', None, 'lowerroman', 'upperroman')[idx]
            elif listid.isdigit():
                start = match[ldepth:match.find('.')]
            elif listid.islower():
                class_ = 'loweralpha'
            elif listid.isupper():
                class_ = 'upperalpha'
        return self._set_list_depth(ldepth, type_, class_, start)

    def _get_list_depth(self):
        """Return the space offset associated to the deepest opened list."""
        return self._list_stack and self._list_stack[-1][1] or 0

    def _set_list_depth(self, depth, new_type, list_class, start):
        def open_list():
            self.close_paragraph()
            self._list_stack.append((new_type, depth))
            class_attr = (list_class and ' class="%s"' % list_class) or ''
            start_attr = (start and ' start="%s"' % start) or ''
            return '<'+new_type+class_attr+start_attr+'><li>'
        def close_list(tp):
            self._list_stack.pop()
            return '</li></%s>' % tp

        rv = ''
        # depending on the indent/dedent, open or close lists
        if depth > self._get_list_depth():
            rv += open_list()
        else:
            while self._list_stack:
                deepest_type, deepest_offset = self._list_stack[-1]
                if depth >= deepest_offset:
                    break
                rv += close_list(deepest_type)
            if depth > 0:
                if self._list_stack:
                    old_type, old_offset = self._list_stack[-1]
                    if new_type and old_type != new_type:
                        rv += close_list(old_type)
                        rv += open_list()
                    else:
                        if old_offset != depth: # adjust last depth
                            self._list_stack[-1] = (old_type, depth)
                        rv += '</li><li>'
                else:
                    rv += open_list()
        return rv

    def close_list(self):
        return self._set_list_depth(0, None, None, None)

    def simple_tag_handler(self, match, open_tag, close_tag):
        """
        """
        if self.has_open_tag((open_tag, close_tag)):
            return self.close_tag(close_tag)
        else:
            self.open_tag(open_tag, close_tag)
        return open_tag

    def has_open_tag(self, tag):
        """
        """
        return tag in self._open_tags

    def open_tag(self, open, close):
        """
        """
        self._open_tags.append((open, close))

    def close_tag(self, tag):
        """
        """
        tmp =  ''
        for i in xrange(len(self._open_tags)-1, -1, -1):
            tmp += self._open_tags[i][1]
            if self._open_tags[i][1] == tag:
                del self._open_tags[i]
                for j in xrange(i, len(self._open_tags)):
                    tmp += self._open_tags[j][0]
                break
        return tmp

    def handle_code_block(self, line):
        if line.strip() == WikiProvider.STARTBLOCK:
            self.in_code_block += 1
            if self.in_code_block == 1:
                self.code_processor = None
                self.code_buf = []
            else:
                self.code_buf.append(line)
                if not self.code_processor:
                    self.code_processor = WikiProcessor(self, 'default')
        elif line.strip() == WikiProvider.ENDBLOCK:
            self.in_code_block -= 1
            if self.in_code_block == 0 and self.code_processor:
                self.close_table()
                self.close_paragraph()
                if self.code_buf:
                    self.code_buf.append('')
                code_text = os.linesep.join(self.code_buf)
                processed = self.code_processor.process(code_text)
                return processed
            else:
                self.code_buf.append(line)
        elif not self.code_processor:
            match = WikiProvider._processor_re.match(line)
            if match:
                name = match.group(1)
                args = WikiProvider._processor_param_re.split(line[len(name):])
                del args[::3]
                keys = [str(k) for k in args[::2]] # used as keyword parameters
                values = [(v and v[0] in '"\'' and [v[1:-1]] or [v])[0]
                          for v in args[1::2]]
                args = dict(zip(keys, values))
                self.code_processor = WikiProcessor(self, name, args)
            else:
                self.code_buf.append(line)
                self.code_processor = WikiProcessor(self, 'default')
        else:
            self.code_buf.append(line)
        return ''

    def close_code_blocks(self):
        rv = ''
        while self.in_code_block > 0:
            rv += self.handle_code_block(WikiProvider.ENDBLOCK)
        return rv

    def _highlight_words(self, text):

        def replace(match):
            return '$$%s$$' % match.group(0)

        for word in self._words:
            match = re.search(r'\b%s\b' % word, text, re.IGNORECASE)
            if match is None:
                continue
            self._highlight_iterator[word.lower()] = len(self._highlight_iterator)
            text = re.sub(re.compile(r'\b%s\b' % word, re.IGNORECASE), replace, text)

        return text

    def format(self, text, skip_highlight=False):
        """
        """
        self._text = text
        for provider in self.providers:
            rules = provider.rules

        if not skip_highlight and self._words is not None:
            text = self._highlight_words(text)

        rv = ''
        for line in text.splitlines():
            if self.in_code_block or line.strip() == WikiProvider.STARTBLOCK:
                rv += self.handle_code_block(line)
                continue
            elif line[0:4] == '----':
                rv += self.close(['table','paragraph','list'])
                rv += '<hr class="wiki" />'+os.linesep
                continue
            elif line == '':
                # new paragraph
                rv += self.close(['table','paragraph', 'list'])
                continue

            line = line.replace('\t', ' '*8)

            if not isinstance(self, OneLinerFormatter) and line.strip()[0:2] != '||' and line.strip()[0:2] != '==':
                line += ' [[BR]]'

            skip_paragraph = False

            if self.in_center:
                skip_paragraph = True

            self.in_list_item = False

            result = re.sub(rules, self.replace, line)

            if self.in_center or not skip_paragraph:
                skip_paragraph = False

            #if not self.in_list_item and not self.in_table:
            #    rv += self.close_list()+'<br />'+os.linesep

            if self.in_table and line.strip()[0:2] != '||':
                rv += self.close_table()

#            self.in_list_item = False
#            result = re.sub(rules, self.replace, line)

#            if not self.in_list_item:
#                rv += self.close_list()

            if len(result) and not self.in_list_item and not self.in_center and not skip_paragraph:
                rv += self.open_paragraph()

            rv += result
            rv += self.close_table_row()
            rv += os.linesep
        rv += self.close(['table', 'paragraph', 'list', 'code_blocks'])
        return rv

    def get_original_text(self):
        return self._text

    def get_linked_wikis(self):
        return self._wikis

    # Table

    def _last_table_cell_formatter(self, match, fullmatch):
        return ''

    def _table_cell_formatter(self, match, fullmatch):
        rv = self.open_table()
        rv += self.open_table_row()
        if self.in_table_cell:
            return rv + '</td><td>'
        else:
            self.in_table_cell = 1
            return rv + '<td>'

    def open_table(self):
        rv = ''
        if not self.in_table:
            rv += self.close_paragraph()
            rv += self.close_list()
            self.in_table = 1
            return '<table class="wiki">' + os.linesep
        return ''

    def open_table_row(self):
        if not self.in_table_row:
            rv = self.open_table()
            self.in_table_row = 1
            return rv + '<tr>'
        return ''

    def close_table_row(self):
        rv = ''
        if self.in_table_row:
            self.in_table_row = 0
            if self.in_table_cell:
                self.in_table_cell = 0
                rv += '</td>'

            rv += '</tr>'
        return rv

    def close_table(self):
        if self.in_table:
            rv = self.close_table_row()
            self.in_table = 0
            return rv + '</table>' + os.linesep
        return ''


class OneLinerFormatter(HTMLFormatter):
    def _one_line_formatter(self, match, fullmatch):
        return match

    def open_paragraph(self):
        return ''

    def close_paragraph(self):
        return ''

    _list_formatter = _one_line_formatter
    _macro_formatter = _one_line_formatter
    _indent_formatter = _one_line_formatter
    _heading_formatter = _one_line_formatter
    _definition_formatter = _one_line_formatter
    _table_cell_formatter = _one_line_formatter
    _last_table_cell_formatter = _one_line_formatter


class WikiProcessor(object):

    def __init__(self, formatter, code_type, kwargs=None):
        """
        """
        self._formatter = formatter
        self._type = code_type
        self._kwargs = {} if kwargs is None else kwargs

    def process(self, text):
        return getattr(self, '_process_%s' % self._type)(text)

    def _process_default(self, text):
        import cgi
        return '<pre class="wiki">%s</pre>' % cgi.escape(text).encode('utf8', 'xmlcharrefreplace')

    def _parse(self, text):
        return self._formatter.format(text, True)

    def _process_version(self, text):

        if len(self._kwargs) == 0 or not self._kwargs.has_key('op') or not self._kwargs.has_key('v'):
            return '<div class="wiki-error">Incorrect usage of Version processor</div>'

        highlight = True if not self._kwargs.has_key('hl') else self._kwargs['hl'] == 'True'
        hide = True if not self._kwargs.has_key('hide') else self._kwargs['hide'] == 'True'
        title = None if not self._kwargs.has_key('title') else self._kwargs['title']

        p = self._kwargs['op']
        if p not in ['lt', 'eq', 'gt', 'lte', 'gte']:
            return '<div class="wiki-error">Incorrect usage of Version processor</div>'

        version = Decimal(self._kwargs['v'])
        release = settings.VERSION.split('-')[0]
        if release == 'trunk':
            release = '99.9'
        # FIXME, no third level versioning support
        release = release.split('.')
        try:
            majorminor = Decimal('%s.%s' % (release[0], release[1]))
        except InvalidOperation:
            return ''

        txt = None
        cls = 'green'

        if p == 'lt' and majorminor < version:
            txt = self._parse(text)
        elif p == 'lte' and majorminor <= version:
            txt = self._parse(text)
        elif p == 'eq' and majorminor == version:
            cls = 'blue'
            txt = self._parse(text)
        elif p == 'gt' and (majorminor > version):
            cls = 'purple'
            txt = self._parse(text)
        elif p == 'gte' and majorminor >= version:
            cls = 'orange'
            txt = self._parse(text)

        if txt is None:
            return '' if hide else self._parse(text)

        if title is not None:
            title = '<h1 class="wiki-highlight-title">%s</h1>' % title
        else:
            title = ''

        return '<div class="%(highlight)s wiki-highlight-%(cls)s">%(title)s%(formatted_text)s</div>' % {
            'title': title,
            'cls': cls,
            'highlight': 'wiki-highlight' if highlight else '',
            'formatted_text': txt
        }

    def _process_domain(self, text):
        from basesoft.web.core.util.controllers import AbstractController
        if not isinstance(self._formatter.context, AbstractController):
            return self._parse(text)

        if len(self._kwargs) == 0 or not self._kwargs.has_key('domain'):
            return '<div class="wiki-error">Incorrect usage of Domain processor</div>'

        domain = self._kwargs['domain']
        c = self._formatter.context
        account = c.get_account()
        if account is None:
            return ''

        equal = domain == account.default_group.license.domain.name
        exclude = False if not self._kwargs.has_key('exclude') else self._kwargs['exclude'] == 'True'
        if not exclude and equal:
            return self._parse(text)
        if exclude and not equal:
            return self._parse(text)
        return ''

    def _process_permission(self, text):
        from basesoft.web.core.util.controllers import AbstractController
        if not isinstance(self._formatter.context, AbstractController):
            return self._parse(text)

        if len(self._kwargs) == 0 or not self._kwargs.has_key('permission'):
            return '<div class="wiki-error">Incorrect usage of Permission processor</div>'

        permissions = self._kwargs['permission'].split(';')
        has_any = False if not self._kwargs.has_key('any') else self._kwargs['any'] == 'True'
        title = None if not self._kwargs.has_key('title') else self._kwargs['title']
        # TODO: option to exclude on having the permission
        if has_any:
            has = self._formatter.context.has_any_permissions(*permissions)
        else:
            has = self._formatter.context.has_permissions(*permissions)

        if not has:
            return ''
        if title is not None:
            text = '=== %s ===\n%s' % (title, text)
        return self._parse(text)

    def _process_highlight(self, text):
        allowed_hls = ['green', 'blue', 'orange', 'purple']

        title = None if not self._kwargs.has_key('title') else self._kwargs['title']
        cls = 'green' if not self._kwargs.has_key('hl') else self._kwargs['hl']
        if cls not in allowed_hls:
            cls = 'green'

        if title is not None:
            title = '<h1 class="wiki-highlight-title">%s</h1>' % title
        else:
            title = ''

        rv = '<div class="wiki-highlight wiki-highlight-%(cls)s">%(title)s%(formatted_text)s</div>' % {
            'cls': cls,
            'title': title,
            'formatted_text': self._parse(text)
        }

        return rv




