# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.
"""
Exceptions module
"""

HTTP_BAD_REQUEST = 400
HTTP_FORBIDDEN = 403

HTTP_INTERNAL_SERVER_ERROR = 500
HTTP_NOT_IMPLEMENTED = 501


class PublicException(Exception):
    """
    Public Exceptions
    Only exceptions inheriting this type are being sent to the client (browser / soap / other external source).
    """

    http_status = HTTP_BAD_REQUEST

class PublicSystemException(PublicException):
    """
    Public System Exceptions
    Contain system related information and are useful for debugging the external client (like jsonrpc or soap)
    """

    http_status = HTTP_INTERNAL_SERVER_ERROR

class PublicUserException(PublicException):
    """
    Public User Exceptions
    End user triggered exceptions.
    """

class PublicOutDatedSourcesException(PublicUserException):
    """
    The sources on the client are outdated and the user should refresh.
    """

class PublicSecurityException(PublicException):
    """
    This class is used for security exceptions that get sent to the client.
    This is extended for more specific cases.
    """
    http_status = HTTP_FORBIDDEN

class PublicInactiveAccountException(PublicSecurityException):
    """
    This is triggered when the Permission Controller gets instantiated with an account
    which is no longer active at any group
    """

class PublicForbiddenSecurityException(PublicSecurityException):
    """
    This is used when the user tries to access something he's not supposed to.
    Eg. incorrect user/pass combination at login, trying to access a page without
    being logged in, etc.
    """

class PublicInsufficientCreditsSecurityException(PublicSecurityException):
    """
    This can be raised when a Security Exception is caught on the assert_credits
    method on the PermissionController to let the User know there are
    insufficient credits for their action.
    """

class SecurityException(Exception):
    """
    This is private and does not get sent to the client. We use this for hacking
    attempts, etc. This can also be hooked into OSSEC.
    """
    http_status = HTTP_FORBIDDEN

class SystemException(Exception):
    """
    Private exception but indicating that the problem comes from our system.
    """
    http_status = HTTP_INTERNAL_SERVER_ERROR

class PermissionDeniedException(PublicSecurityException):
    """
    This class is used to raise Permission Denied exceptions in case of insufficient privileges or
    permissions.
    """

class IconException(Exception):
    """
    Exception raised when icons don't meet certain conditions to be showed in the iconbar.
    """

class LoginRequiredException(PublicSecurityException):
    """
    Exception raised when a login is required.. typically fired from a decorator
    """
    pass

class SCSSCompileException(SystemException):
    pass
