# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

"""
Module for creating form behaviours that can be used for validating fields from
forms or soap, where forms (or soap) need to know the value of all fields, in
order to validate certain fields based on the values of other fields
"""

from django.utils.translation import ugettext as _
from django.forms.util import ValidationError
from django.forms.forms import NON_FIELD_ERRORS

from basesoft.web.core.exceptions import SystemException

def RaiseValidationError(message, fieldName=NON_FIELD_ERRORS):
    """
    Raise the Validation error with the specified message and associate it to
    the proper field

    @param message: the message that needs to be displayed as error
    @param fieldName: the name of the field to which the error should be
        associated. if it's not specified, the error won't be bound to a field
    @raise ValidationError
    """
    e = ValidationError(message)
    e.fieldName = fieldName
    raise e

class Validation:

    debug = False

    def __init__(self, siblings):
        """
        Initialize the validation with the value it needs to check, and a dictionary
        of all the other fields and their values

        @param value: the value that needs to be checked
        @param siblings: the dictionary that contains all the other fields and
            their values (it's the responsibility of the form or soap to provide
            this dictionary)
        """
        self._handles = {}
        self._siblings = siblings

    def check(self):
        pass

class Required(Validation):
    """
    Check if the field needs to have a value or not, based on the values of other fields

    Usage:
    validation = ((group_plugins.Required, {'f1': {'f2': True}}), )
        if f2 has a value, f1 is required

    validation = ((group_plugins.Required, {'f1': {'f2': False}}), )
        if f2 doesn't have a value, f1 is required

    validation = ((group_plugins.Required, {'f1': {'f2': 'abc123'}}), )
        if f2 has value 'abc123', f1 is required (used for fields that depend on radios mostly)

    validation = ((group_plugins.Required, {'f1': {'f2': True, 'f3': False}}), )
        if f2 has value and f3 doesn't, f1 is required (used for fields that depend on more then one field)

    validation = ((group_plugins.Required, {'f1': {'f2': False}, 'f2': {'f1': False}}), )
        if f2 doesn't have a value, f1 is required,
        and if f1 doesn't have a value, f2 is required

    For more documentation check:
    http://trac.office.basesoft.com/iqportal/wiki/ValidationRequired
    """

    def __init__(self, *args, **kwargs):
        """
        Initialization for the plugin. The plugin needs to receive the dictionary
        containing fields and values from the form or whoever it's using the plugin

        @param fieldValues: dictionary containing the names of the fields as keys
            and as values the field values
        """
        Validation.__init__(self, *args, **kwargs)
        self._handles['required_value'] = _('A value is required')

    def check(self, params=None):
        """
        Check if the field is required or not, based on what dependencies it has

        @param params: a dictionary containing as keys the fields that need to
            be checked if they're required or not, and as values they have other
            dictionaries, that contain as keys the names of the dependency fields
            and as values a boolean indicating if the dependencies need to have
            a value or not in order to mark the target field as required
        @raise ValidationError: raised if the dependency is not met. The
            ValidationError object that's raised can also have an aditional field
            called "fieldName" that contains the name of the field to which this
            error is associated

        Example for params:
        {'f1': {'f2': False}} - f1 is required if f2 doesn't have a value
        {'f1': {'f2': True}} - f1 is required if f2 has a value
        {'f1': {'f2': True, 'f3': True}} - f1 is required if both f2 and f3 have values
        {'f1': {'f2': 'abc123'}} - f1 is required if f2 has value 'abc123'
        """
        if not params:
            raise SystemException('Incorrect usage of Required group plugin, no parameters supplied')

        for param in params:
            target_field = param
            dependencies = params[target_field]

            # check the value of the dependencies to see if it's required or not
            if not type(dependencies) == dict:
                raise SystemException('Incorrect usage of Required group plugin, no dictionary supplied')

            # no need to check further if the target field is not in siblings
            # or it already has a value
            if not self._siblings.has_key(target_field):
                skip = True
                for key in dependencies.keys():
                    if self._siblings.has_key(key):
                        skip = False
                if skip:
                    continue
                RaiseValidationError(_('The %(target_field)s field is required due to dependencies.') % {'target_field': target_field}, target_field)

            if self._siblings[target_field]:
                continue

            # how many times have we triggered the dependency ? we only want to raise the
            # error if all of the fields are required
            error_count = 0

            for dep in dependencies:
                try:
                    if isinstance(dependencies[dep], bool):
                        # we have True or False
                        if dependencies[dep]:
                            if self._siblings[dep]: error_count = error_count + 1
                        else:
                            if not self._siblings[dep]: error_count = error_count + 1
                    else:
                        if str(self._siblings[dep]) == str(dependencies[dep]): error_count = error_count + 1
                except KeyError:
                    raise SystemException('Incorrect usage of Required group plugin, unknown dependency')

            if error_count == len(dependencies):
                RaiseValidationError(self._handles['required_value'], target_field)
