
"""
This module can be used to create JSON-RPC service
by creating a server and either installing functions, or a
class instance.
________________________________________________________________________
# Example: Defining a JsonRPC view using simple methods

from basesoft.web.core.jsonrpc.dispatcher import JsonRpcDispatcher
from django.http import HttpResponse

def simple_view(request):
    def multiply(a, b):
        return a*b

    def echo(str):
        return str;

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_function(multiply, 'multiply')
    dispatcher.register_function(echo, 'echo')

    return dispatcher.handle_request(request)

_________________________________________________________________________
# Example: Defining a JsonRPC view using a class dispatcher
# FIXME
_________________________________________________________________________
# Example: How to call a JSONRPC view from Javascript
# Note: Make sure the iq_broker module is loaded.

// prepare a broker object that can handle multiple calls to the specified view
var broker = new BASESOFT.util.JsonBroker('/path/to/my/view/');

// call the multiply method with arguments 2 and 8
broker.call('multiply', [2, 8],
    function(result) {
        // print the result to console
        console.debug(result)
    },
    null,  // default error handler will show errors in a popup
    this); // scope in which the callback is executed

"""
from basesoft.web.core.exceptions import SecurityException, PublicUserException

import sys
import traceback
import datetime, decimal
import random

from hashlib import sha256 as sha

from SimpleXMLRPCServer import resolve_dotted_attribute

from django.http import HttpResponse, UnreadablePostError
from django.utils import simplejson
from django.utils.functional import Promise

from django.conf import settings
import pytz

from basesoft import logging
from decimal import Decimal
from simplejson.decoder import JSONDecodeError
logger = logging.getLogger('basesoft.core.jsonrpc')

def json_object_hook(d):
    """
    This method is hooked on to the de-serialisation of json strings. Check
    if the object has some kind of special var set and then it deserialises
    it in its specific way.
    """

    if '__jsonclass__' in d:
        constructor, values = d['__jsonclass__']

        # datetime is always sent in utc
        if constructor == 'datetime':

            # convert values into a datetime with server timezone
            d_utc = d = datetime.datetime(tzinfo=pytz.utc, *values)
            tz = pytz.timezone(settings.TIME_ZONE)
            dt_local = tz.normalize(d_utc.astimezone(tz))
            return datetime.datetime(*[i for i in dt_local.timetuple()[0:6]])

        if constructor == 'file_hashes':
            # return the temp file objects
            from basesoft.web.contrib.file.controllers import UploadFileController
            from basesoft.web.core.util.db import Session
            controller = UploadFileController(Session(), None)
            d = controller.get_hash_temp_files(values)
            return d

        if constructor == 'decimal':
            return Decimal(values)

    return d


class JsonRpcDispatcher:
    """JsonRPC Dispatcher"""

    def __init__(self):
        self.funcs = {}
        self.instance = None

    def register_function(self, func, name=None):
        """Registers a function to respond to JSON-RPC requests.

        The optional name argument can be used to set a Unicode name
        for the function.
        """
        if name is None:
            name = func.__name__
        self.funcs[name] = func

    def register_instance(self, instance):
        """Registers an instance to respond to JSON-RPC requests.

        Only one instance can be installed at a time."""
        self.instance = instance

    def _marshaled_dispatch(self, data, dispatch_method = None):
        """Dispatches an JSON-RPC method from marshalled data."""

        # decode data json string
        try:
            data = simplejson.loads(data.decode('utf-8'), object_hook=json_object_hook)
        except JSONDecodeError:
            raise PublicUserException('Invalid request: data cannot be decoded')

        method = data['method']
        params = data['params']

        if dispatch_method is not None:
            response = dispatch_method(method, params)
        else:
            response = self._dispatch(method, params)

        return response

    def _dispatch(self, method, params):
        """Dispatches the JSON-RPC method.

        JSON-RPC calls are forwarded to a registered function that
        matches the called JSON-RPC method name. If no such function
        exists then the call is forwarded to the registered instance,
        if available."""

        func = None
        try:
            # check to see if a matching function has been registered
            func = self.funcs[method]
        except KeyError:
            if self.instance is not None:
                # check for a _dispatch method
                if hasattr(self.instance, '_dispatch'):
                    return self.instance._dispatch(method, params)
                else:
                    # call instance method directly
                    try:
                        func = resolve_dotted_attribute(
                            self.instance,
                            method,
                            False
                            )
                    except AttributeError:
                        pass

        if func is not None:
            if type(params) == list:
                return func(*params)
            elif type(params) == dict:
                return func(**params)
            else:
                raise Exception('invalid params type: %s' % type(params))
        else:
            raise Exception('method "%s" is not supported on %s' % (method, self.instance))

    def handle_request(self, request):
        try:
            response = self._marshaled_dispatch(request.raw_post_data)
            return JsonRpcResponse(request, response)
        except UnreadablePostError, e:
            if 'request data read error' in e.message:
                # Ignore POST exceptions. This means the client disconnected before the post was completed
                return JsonRpcResponseException(request, PublicUserException('request data read error'))
            else:
                raise
        except Exception, e:
            # the JsonRpcResponseException will handle exception filtering
            # (to send to the client)
            return JsonRpcResponseException(request, e)

class JsonRpcJSONEncoder(simplejson.JSONEncoder):

    def default(self, obj):

        # handle dates
        if (isinstance(obj, datetime.datetime)) or (isinstance(obj, datetime.date)):
            if isinstance(obj, datetime.datetime):
                vals = [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]
            else:
                vals = [obj.year, obj.month, obj.day, 0, 0, 0]
            tz = pytz.timezone(settings.TIME_ZONE)
            obj = tz.localize(datetime.datetime(*vals))
            return {
                '__jsonclass__': ['datetime', [i for i in obj.utctimetuple()[0:6]] ]
            }

        # handle decimals
        if isinstance(obj, decimal.Decimal):
            return {
                '__jsonclass__': ['decimal', unicode(format(obj, 'f'))]
            }

        # force unicode, this will trigger translations for lazy_gettext
        if isinstance(obj, Promise):
            return unicode(obj)

        # obj must be unicode already.. return
        return obj

from django.utils.http import http_date
import time

class JsonRpcResponse(HttpResponse):
    """Extends HttpResponse and json encodes the object passed"""
    def __init__(self, request, obj={}, stat='ok'):
        content = simplejson.dumps(obj, ensure_ascii=False, cls=JsonRpcJSONEncoder, use_decimal=False)
        super(JsonRpcResponse, self).__init__(content, mimetype='text/plain')

        # expire directly
        self['Expires'] = http_date(time.time())

class JsonRpcResponseException(JsonRpcResponse):
    """JSON-RPC failure object"""

    # by setting the status_code we make sure this arrives in the 'failure' corben
    # by default set it to HTTP_INTERNAL_SERVER_ERROR. This can change if the exception received
    # has its own http_status variable
    status_code = 500

    def __init__(self, request, exception, extra_params={}):

        from basesoft.web.core.exceptions import PublicException
        from django.utils.translation import ugettext as _

        # type is something like 'LoginRequiredException'
        # so we can map it clientside to some kind of Error object if needed
        exception_type = exception.__class__.__name__
        exception_message = exception.message

        # try to get http_status from the exception
        if hasattr(exception, 'http_status'):
            assert isinstance(exception.http_status, int)
            self.status_code = exception.http_status

        # if the exception is a SecurityException we want to have a 403 status code.
        if isinstance(exception, SecurityException):
            exception_message = _('Permission denied')

        elif isinstance(exception, NotImplementedError):
            exception_message = '%s - %s' % (_('Not implemented'), exception_message)

        # in case of public exception, send the exception type and message to client
        # else send a generic message
        elif not isinstance(exception, PublicException):

            exception_hash = sha(str(random.random())).hexdigest()[:8]

            # When DEBUG is False, send an error message to the admins. (similar to how django reports errors)
            if not settings.DEBUG:
                from django.core.mail import mail_admins

                exc_info = sys.exc_info()
                subject = '#%s: Error (%s IP): %s' % (exception_hash, (request.META.get('REMOTE_ADDR') in settings.INTERNAL_IPS and 'internal' or 'EXTERNAL'), request.path)
                try:
                    request_repr = repr(request)
                except:
                    request_repr = "Request repr() unavailable"

                session_data = ""
                for k in request.session.keys():
                    session_data += "%s: %s\n" % (k, request.session.get(k))

                traceback_str = '\n'.join(traceback.format_exception(*(exc_info or sys.exc_info())))
                message = "%s\n\n%s\n\n%s" % (traceback_str, session_data, request_repr)
                mail_admins(subject, message, fail_silently=True)

            # Override exception info to user for security reason
            exception_type = "SystemException"
            exception_message = _("We apologize for the inconvenience, but we are unable to process your request at this time. Our engineers have been notified of this problem and will work to resolve it. Reference #%s" % exception_hash)

        # error log backtrace anyway
        logger.error(traceback.format_exc())

        errdata = {
            'exception': {
                'type': str(exception_type),
                'message': exception_message,
            }
        }
        errdata.update(extra_params)
        JsonRpcResponse.__init__(self, request, errdata, "fail")
