# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

"""
Contains a command object to reset the database for all INSTALLED_APPS
"""

from basesoft import logging
logger = logging.getLogger("basesoft.django.commands.iqresetdb")

from django.core.management.base import BaseCommand, CommandError
from django.conf import settings
from twisted.python.reflect import namedAny

class Command(BaseCommand):

    help = 'Drops and recreates the database tables for all IQPortal applications'
    args = '[appname ...]'

    # override this as we do not use django models
    # and this just gets in the way
    requires_model_validation = False
    output_transaction = False

    def handle(self, *app_names, **options):

        from sqlalchemy import create_engine, MetaData
        from sqlalchemy.orm import clear_mappers

        show_traceback = options.get('traceback', False)

        apps_to_load = list(app_names)
        if len(apps_to_load) >= 1:
            raise CommandError('iqresetdb can only be done for all IQPortal installed apps')

        # get the list of apps
        apps_to_reset = settings.INSTALLED_APPS

        logger.info('IQResetDB: resetting apps %s' % settings.INSTALLED_APPS)

        db_engine = create_engine(settings.DB_CONNSTR, echo=False)
        conn = db_engine.connect()

        #the list of all the tables that will be reset
        table_list = []
        metadata = MetaData(db_engine)

        from sqlalchemy.sql import text
        # to clear the database, we have to disable foreign key checks cuz we have circular references.
        conn.execute(text('SET FOREIGN_KEY_CHECKS=0'))
        result = conn.execute(text('SHOW TABLES'))
        for i in result:
            if i[0][:4] == 'view':
                conn.execute(text('DROP VIEW %s' % i[0]))
            else:
                conn.execute(text('DROP TABLE %s' % i[0]))
        conn.execute(text('SET FOREIGN_KEY_CHECKS=1'))

        clear_mappers()

        # see if there is a sa_datamapper module in the app
        # if there is import it and get the list of the tables that will be reset
        for app in apps_to_reset:

            try:
                sa_module = namedAny(app+'.sa_datamapper')
            except AttributeError, e:
                continue

            if hasattr(sa_module, 'reset'):
                sa_module.reset()

            if hasattr(sa_module, 'do_tables'):
                sa_module.do_tables(metadata)

            if hasattr(sa_module, 'get_tables'):
                table_list.extend(sa_module.get_tables().values())

        from basesoft.sqlalchemy.table import IQTable
        for i in table_list:
            assert isinstance(i, IQTable), "not a IQTable: %s" % i

        try:
            metadata.create_all()
        except Exception, e:
                if show_traceback:
                    raise
                error = "Reset failed:  %s" % e
                raise CommandError(error)

        fill_data = []
        for app in apps_to_reset:
            try:
                sa_module = namedAny(app+'.sa_datamapper')
            except AttributeError, e:
                continue

            if hasattr(sa_module, 'do_mappings'):
                logger.debug('iqresetdb: %s: executing do_mappings()' % sa_module);
                sa_module.do_mappings()

            if hasattr(sa_module, 'fill_data'):
                logger.debug('iqresetdb: %s: attaching to fill_data()' % sa_module);
                fill_data.append(sa_module)

        from basesoft.web.core.util.db import Session
        from basesoft.web.core.management.commands.iqsyncdb import Command as SyncDB
        from basesoft.web.contrib.license.db_sync import sync as PermissionSync

        session = Session()
        PermissionSync(session)

        # load initial data
        for sa_module in fill_data:
            if hasattr(sa_module, 'fill_data'):
                logger.info('iqresetdb: %s: executing fill_data' % sa_module);
                sa_module.fill_data(conn)

        self.__add_sysadmin(session)
        SyncDB().handle(session=session)

        # post reset
        for app in apps_to_reset:
            try:
                sa_module = namedAny(app+'.sa_datamapper')
            except AttributeError, e:
                continue

            if hasattr(sa_module, 'post_reset'):
                logger.info('iqresetdb: %s: calling post_reset()' % sa_module);
                sa_module.post_reset()



    def __add_sysadmin(self, session):
        """
        This adds a complete sysadmin account setup to the database.
        """
        from basesoft.web.contrib.contact import models as contact_models
        from basesoft.web.contrib.i18n import models as i18n_models
        from basesoft.web.contrib.license import models as license_models
        from basesoft.web.contrib.account import models as reg_models

        domain = license_models.Domain()
        domain.name = 'Resellers'

        person = contact_models.Person()
        person.first_name = u'Paul'
        person.middle_name = u'de'
        person.last_name = u'Graaf'
        person.email = u'paul@basesoft.com'

        netherlands = session.query(i18n_models.Country).filter(
                                    i18n_models.Country.iso_3166_a2=='NL').one()

        address = contact_models.Address()
        address.country = netherlands
        address.postcode = '1017 CH'
        address.street = u'Herengracht 564'
        address.city = u'Amsterdam'
        address.state = u'Noord-Holland'

        company = contact_models.Company()
        company.contact_person = person
        company.website = u'http://www.basesoft.com'
        company.vat_no = 'NL123456782B90'
        company.name = u'BaseSoft Solutions B.V.'

        ca = contact_models.CompanyAddress()
        ca.address = address
        ca.company = company
        ca.is_invoice = True
        ca.is_delivery = True

        bank_account = contact_models.BankAccount()
        bank_account.number = '1234567890'
        bank_account.iban = 'IBAN BaseSoft Solutions B.V.'
        bank_account.bic = 'BIC BaseSoft Solutions B.V.'

        cba = contact_models.CompanyBankAccount()
        cba.is_default = True
        cba.company = company
        cba.bank_account = bank_account

        license_manager_permissions = session.query(
            license_models.Permission).join(license_models.PermissionCategory).filter(
            license_models.PermissionCategory.name.in_(["Core",
                                                        "Dashboard",
                                                        "License",
                                                        "Account",
                                                        "Mailing",
                                                        "Mockup",
                                                        "Manual",
                                                        "Notification"])).all()
        license = license_models.License()
        license.is_default = True
        license.name = settings.SITE_OWNER_LICENSE_NAME
        license.domain = domain
        license.contact_person = person
        license.company = company
        license.http_host = u'iqportal.basesoft.com' # http:// is not required
        license.self_branded = True
        license.permissions = license_manager_permissions

        group = license_models.Group()
        group.name = u'System Administrators'
        group.license = license
        group.permissions = license_manager_permissions
        group.is_default = True

        # pass: test_password
        account = reg_models.Account()
        account.name = u'sysadmin'
        account.password = '10a6e6cc8311a3e2bcc09bf6c199adecd5dd59408c343e926b129c4914f3cb01'
        account.locale = session.query(i18n_models.Locale).filter(
                                       i18n_models.Locale.locale=='en_GB').one()

        ga = reg_models.GroupAccount()
        ga.account = account
        ga.group = group
        ga.active = True
        ga.is_default = True

        session.add(ga)
        session.add(ca)
        session.add(cba)
        session.flush()
