# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

"""
TODO: can we add this command to the IQube app?
This command is used to synchronized "datasource" & "cubes" from different apps with the Mondrian setup
"""
from twisted.python.reflect import namedAny
from sqlalchemy import and_

from django.core.management.base import BaseCommand
from django.conf import settings

from basesoft.web.core.util.db import Session
from basesoft.util.bitwise import get_bits

from basesoft import logging
from iqportal.iqube.core.models import DataSource, DataSourceGroup, DataSourceCube
from basesoft.web.contrib.license.models import Group, License
from sqlalchemy.orm.exc import NoResultFound
from iqportal.iqube.core.exceptions import IQubePublicException
from sqlalchemy.exc import IntegrityError
logger = logging.getLogger("basesoft.django.commands.iqubesync")

class Command(BaseCommand):
    """
    """
    help = "Synchronizes datasource & cube data for the supplied apps"
    args = "[appname...]"

    def handle(self, sa_session=None, *app_names, **options):
        """
        """
        if len(app_names) == 0:
            app_names = settings.INSTALLED_APPS

        if 'iqportal.iqube.core' not in settings.INSTALLED_APPS or settings.IQUBE_SERVLET is None:
            print 'IQube not enabled, skipping IQube Sync'
            return

        sa_session = Session()

        from iqportal.iqube.core.controllers import IQubeController
        controller = IQubeController()

        for app in app_names:
            try:
                sync_module = namedAny(app+'.iqube_sync')
            except AttributeError:
                # if we do not have a sync module, we don't care for this app, continue...
                continue

            # if there is no sync command, we have nothing to do.
            if not hasattr(sync_module, 'projects'):
                continue

            logger.info('Synchronizing datasources & cubes for app %s' % app)
            for project in sync_module.projects:
                if not issubclass(project, ProjectBase):
                    logger.info('Project [%s] is not a ProjectBase subclass in app %s' % (project, app))
                    continue
                project(sa_session, controller).execute()

        # finally call clean_datasources on iqube to trash all old datasources
        controller.clean_datasources()


class ProjectBase:
    protocol = 'MySQL'
    host = 'localhost'
    port = '3306'
    description = ''
    additional_graphs = None
    client_module = None
    calculated_member_hook_js_module = None
    role_level = DataSource.ROLE_LEVEL_NONE

    # license, group
    owner = ('BaseSoft Solutions B.V.', 'System Administrators')
    group_privileges = 0
    license_privileges = 0
    domain_privileges = 0
    world_privileges = 0

    def __init__(self, sa_session, controller):
        self.sa_session = sa_session
        self.controller = controller

    def __attach_groups(self, obj, group):

        # set group, license and domain privs.
        group_privs = get_bits(self.group_privileges)
        license_privs = get_bits(self.license_privileges)
        domain_privs = get_bits(self.domain_privileges)

        world_privs = self.world_privileges
        if settings.DEBUG:
            world_privs = world_privs | DataSource.PRIV_VIEW

        world_privs = get_bits(world_privs)
        gobj, backref = obj.get_group_object()

        used_privs = group_privs
        [used_privs.append(p) for p in license_privs if p not in used_privs]
        [used_privs.append(p) for p in domain_privs if p not in used_privs]
        [used_privs.append(p) for p in world_privs if p not in used_privs]
        for priv in used_privs:
            dsg = gobj()
            dsg.group = group
            dsg.privilege = priv
            if dsg.privilege in group_privs:
                dsg.group_has_privilege = True
            else:
                dsg.group_has_privilege = False
            if dsg.privilege in license_privs:
                dsg.license_has_privilege = True
            if dsg.privilege in domain_privs:
                dsg.domain_has_privilege = True
            if dsg.privilege in world_privs:
                dsg.world_has_privilege = True
            obj.groups.append(dsg)

    def _attach_cube(self, datasource, name):
        try:
            cube = self.sa_session.query(DataSourceCube).filter(and_(DataSourceCube.datasource==datasource, DataSourceCube.name==name)).one()
        except NoResultFound, e:
            cube = DataSourceCube()
            cube.datasource = datasource
            cube.name = name

            license_name, group_name = self.owner

            group = self.sa_session.query(Group).join(License).filter(and_(License.name==license_name,Group.name==group_name)).one()
            self.__attach_groups(cube, group)
            self.sa_session.add(cube)

        return cube

    def _db_from_connstr(self, connstr):
        import re
        p_def = re.compile('(?P<protocol>\w+):\/\/(?P<user_name>[\w_\d]+):(?P<password>.+)@(?P<host>[\w\-\d\.]+):(?P<port>\d+)\/(?P<database_name>[\w\d_\-]+)')

        m = p_def.search(connstr)
        self.host = m.group('host')
        self.port = m.group('port')
        self.db_name = m.group('database_name')
        self.db_user = m.group('user_name')
        self.db_password = m.group('password')
        self.protocol = m.group('protocol')

    def execute(self):
        # add the project to IQPortal, prefix the name
        self.project_name = '%s.%s' % (settings.SETTINGS_MODULE, self.project_name)

        logger.info('Adding/updating mondrian project "%s"' % self.project_name)
        # add/update the project on the mondrian server through the IQube servlet.
        db = {
            "protocol": self.protocol,
            "host": self.host,
            "port": self.port,
            "db_name": self.db_name,
            "user": self.db_user,
            "password": self.db_password
        }
        try:
            self.controller.call("updateDatasource", self.project_name, db)
        except IQubePublicException, e:
            logger.info(' - Unable to add project')
            logger.error(e)
            return False

        try:
            ds = self.sa_session.query(DataSource).filter(DataSource.name==self.project_name).one()
        except NoResultFound, e:
            ds = DataSource()
            ds.type = DataSource.DATASOURCE_TYPE_DATABASE

            license_name, group_name = self.owner

            group = self.sa_session.query(Group).join(License).filter(and_(License.name==license_name,Group.name==group_name)).one()
            self.__attach_groups(ds, group)

        ds.name = self.project_name
        ds.mondrian_project = self.project_name
        ds.client_module = self.client_module
        ds.role_level = self.role_level

        self.sa_session.add(ds)

        for cube in self.get_cubes():
            self._attach_cube(ds, cube.name)

        self.sa_session.flush()

        from iqportal.iqube.core.controllers import ProjectController
        pc = ProjectController(self.sa_session)
        pc.flush_schema(ds.id)

        if self.additional_graphs is not None:
            from iqportal.iqreport.core.models import IQReportGraphType, IQReportGraphTypeDatasource
            for graph_name in self.additional_graphs:
                try:
                    gt = self.sa_session.query(IQReportGraphType).filter(IQReportGraphType.name==graph_name).one()
                except:
                    logger.warning(' - Graph type `%s` not found' % graph_name)
                    continue

                gtds = IQReportGraphTypeDatasource()
                gtds.datasource = ds
                gtds.graph_type = gt
                self.sa_session.add(gtds)

                try:
                    self.sa_session.flush()
                except IntegrityError:
                    logger.info(' - IQReportGraphTypeDatasource already exists')
                    continue

    def get_shared_dimensions(self):
        return []

    def get_cubes(self):
        return []

    def get_virtual_cubes(self):
        return []

    def get_roles(self):
        return []

class SharedDimensionBase:
    xml = None

class CubeBase:
    xml = None

    def __init__(self, sa_session):
        self.sa_session = sa_session

class RoleBase(object):
    xml = None

    def __init__(self, sa_session, ds):
        self.sa_session = sa_session
        self.ds = ds
