/*
 * Copyright (c) 2008 BaseSoft B.V.
 * See LICENSE for details.
 */

/**
 * @module iq_validation
 */

// namespace
if (typeof BASESOFT.validation == 'undefined')
	BASESOFT.validation = {};

/**
 * FormValidator
 * Uses JsonBroker to send validation requests
 * @class FormValidator
 * @constructor
 * @namespace BASESOFT.validation
 * @param {String} form_id Id of the form that needs validation
 */
BASESOFT.validation.FormValidator = function(form, config) {

	BASESOFT.validation.FormValidator.superclass.constructor.call(this, form, config);

	BASESOFT.env.growler.register({
		application: "FormValidation",
		allowed: ['ValidationError']
	});

	this.form = this.get('element');
	console.debug(this.get('rpc_url'));

	this.broker = new BASESOFT.util.JsonBroker(this.get('rpc_url'));

	//dictionary with fields error messages
	this.messages = {};

	//prepare the submit events
	this.onSubmitSuccess = new YAHOO.util.CustomEvent('onSubmitSuccess');
	this.onSubmitFailure = new YAHOO.util.CustomEvent('onSubmitFailure');
	this.onRPCSubmit = new YAHOO.util.CustomEvent('onRPCSubmit');

	this._addEvents();
};

YAHOO.extend(BASESOFT.validation.FormValidator, YAHOO.util.Element, {

	initAttributes: function(attr) {
		attr = attr || {};

		BASESOFT.validation.FormValidator.superclass.initAttributes.call(this, attr);

		this.setAttributeConfig('rpc_url',{
			value: attr.rpc_url || window.location+"rpc/"
		});

		this.setAttributeConfig('error_box_container', {
			value: attr.error_box_container || null
		});
	},
	/*
	* Return a dictionary with all form fields; name and value pairs
	*/
	_get_form_fields: function()
	{
		var form_elements = this.form.elements;
		var elements = {};
		for (var e = 0; e < form_elements.length; e++)
		{
			var el = form_elements[e];
			if(el.name == undefined || el.name === "")
				continue;

			//send the value of the radio only in case it is checked
			if (el.type.toLowerCase() == 'radio')
			{
				if (el.checked)
				{
					elements[el.name] = el.value;
					continue;
				}
				//don't overwrite if there's already a value (case of RadioSelect)
				if(elements[el.name] == undefined)
					elements[el.name] = "";
				continue;
			}

			//also for checkboxes, send to the server only the checked values
			if (el.type.toLowerCase() == 'checkbox')
			{
				if(elements[el.name] != undefined)
				//this is a multiple checkbox select, so we make a list with the values and send that to the server
				{
					if(typeof elements[el.name].push != 'function')
						elements[el.name] = (elements[el.name] === "")?[]:[elements[el.name]];
					if (el.checked)
						elements[el.name].push(el.value);
					continue;
				}
				//simple checkbox input
				elements[el.name] = (el.checked)?el.value:"";
				continue;
			}

			elements[el.name] = el.value;
		}
		return elements;
	},

	/*
	* Function for adding onBlur event to each form input and onsubmit event to the form
	*/
	_addEvents: function()
	{
		/* ---- form event
		 * add a custom listener for the onsubmit event that will send a rpc submit request to the server.
		*/
		YAHOO.util.Event.addListener(this.form, "submit",
			function(e, args){
				//first prevent the form from submitting
				YAHOO.util.Event.preventDefault(e);
				//fire our custom onRPCSubmit event
				this.onRPCSubmit.fire({});
			},[], this);

		var form_elements = this.form.elements;

		//add listener for the custom rpc submit event
		this.onRPCSubmit.subscribe(
				function(e, elems){
						// for editors, we need to call saveHTML()
						// NOTICE: variable 'e' below hides function argument
						for (e = 0; e < form_elements.length; e++) {
							if (form_elements[e].name != undefined) {
								var instance = BASESOFT.env.widgetTransformer.getInstance(form_elements[e].id);
								if (instance != undefined && typeof instance.saveHTML == 'function')
									instance.saveHTML();
							}
						}

						//elements that will be attached to the list of elements sent to the server on submit
						var prepared_elements = {};
						if (YAHOO.lang.isArray(elems) && YAHOO.lang.isObject(elems[0]))
							prepared_elements = elems[0];

						//for fancy select we need to get the selected options in the prepared_elements
						var els = YAHOO.util.Dom.getElementsBy(
									function(el) {
										return (el.tagName.toLowerCase() == 'div' && el.id !== undefined && el.id !== '');
									  });

						// NOTICE: variable 'e' below hides function argument
						for (e = 0; e < els.length; e++)
						{
							var fs_instance = BASESOFT.env.widgetTransformer.getInstance(els[e].id);
							if(fs_instance != undefined && typeof fs_instance.addOptions == 'function') {
								var selectedopts = fs_instance.get('selectedoptions');
								var selectedopts_list = [];
								for(var opt in selectedopts) {
									selectedopts_list.push(selectedopts[opt]['value']);
								}
								prepared_elements[fs_instance.get('name')] = selectedopts_list;
							}
						}
						this._submit_form(prepared_elements);
					},this,true);

		/* ---- field events: use event delegation for this */

		//handler for field events
		function blurHandler(e) {

			//get the resolved (non-text node) target:
			var elTarget = YAHOO.util.Event.getTarget(e);

			//walk up the DOM tree looking for inputs
			while (elTarget != this.form) {
				if((elTarget.nodeName.toUpperCase() == "INPUT" || elTarget.nodeName.toUpperCase() == "SELECT" || elTarget.nodeName.toUpperCase() == "TEXTAREA") && elTarget.type.toUpperCase != 'HIDDEN' && elTarget.type.toUpperCase != 'BUTTON')
				{
					this._validate_field(elTarget);
					break;
				} else {
					elTarget = elTarget.parentNode;
				}
			}
		}

		//attach blurHandler as a listener for any blur on the form
		if(YAHOO.env.ua.ie)
			/*
			  Since the "blur" event doesn't bubble, and IE doesn't support
			  listening for the capture phase of events, subscribe to the
			  "focusout" event, which DOES bubble.
			*/
			YAHOO.util.Event.addListener(this.form, "focusout",blurHandler,[],this);
		else
		{
			 /*
			   Since the "blur" event doesn't bubble, we'll use
			   event capture for Gecko, Webkit and Opera.
			*/
			var save_this = this;
			this.form.addEventListener('blur',function(e) {blurHandler.apply(save_this,[e]);}, true);
		}

	},

	/*
	* Calls the 'validate' method on the server for field validation.
	 * Param for the server validation method is the form contents(all elements),
	 * so also field group plugins errors will be returned in the response.

	* @param field: form field.
	*/
	_validate_field: function(field){
		var elements = this._get_form_fields();
		this.broker.call('validate', [elements], function(result){
			this._resolve_field_validation(result, field);
		}, function(failure){
			console.debug('failure: ', failure);
		}, this);
	},

	/*
	 * Calls the 'submit' method on the server for field validation.
	 * Param for the server method is the form contents + the prepared_elements in the arg

	 * rpc submit request can have one of the following responses:
	   * the form has validation errors: the result of the request is a dictionary with the form's errors -> notification box
	   * there was an exception thrown when trying to save form contents: failure -> fire the onSubmitFailure event
	   * saving the form contents went ok: result is empty -> fire onSubmitSuccess event

	 * @param prepared_elements: dictionary with prepared elements that will be attached to the [elements] list sent to the server
	*/
	_submit_form: function(prepared_elements)
	{
		var elements = this._get_form_fields();
		for (var el in prepared_elements)
			elements[el] = prepared_elements[el];

		this.broker.call('submit', [elements],
		   function(result){
				this._resolve_form_validation(result);
				if (BASESOFT.util.is_empty(result))
					this.onSubmitSuccess.fire();

		   },
		   function(failure) {
				console.debug('failure: ',failure);
				//this will remove any notification with fields errors messages
				this._resolve_form_validation({});
				this.onSubmitFailure.fire(failure);
			},
		   this);
	},

	/*
	* Handles server response for field validation
	* @param result: server response.
	* @param field: form field.
	*/
	_resolve_field_validation: function(result,field)
	{
		YAHOO.util.Dom.removeClass(field,'error');
		YAHOO.util.Dom.removeClass(field,'valid');
		if(result[field.name] != undefined)
		{
			YAHOO.util.Dom.addClass(field,'error');
			this.messages[field.name] = result[field.name];
		}
		else
		{
			YAHOO.util.Dom.addClass(field,'valid');
			delete(this.messages[field.name]);
		}
		//show the panel with the error messages (if there were any)
		this._show_error_box();

	},

	/*
	* Handles server response for form validation
	* @param result: server response.
	*/
	_resolve_form_validation: function(result)
	{
		//in case of submit, the dictionary with the error messages will be the response dictionary
		this.messages = result;

		if (!BASESOFT.util.is_empty(result))
		{
			var form_elements = this.form.elements;
			for (var e = 0; e < form_elements.length; e++) {
				if (form_elements[e].name != undefined && form_elements[e].name !== "" && form_elements[e].getAttribute('disabled') == undefined) {
					YAHOO.util.Dom.removeClass(form_elements[e], 'error');
					YAHOO.util.Dom.removeClass(form_elements[e], 'valid');
					if (result[form_elements[e].name] != undefined)
						YAHOO.util.Dom.addClass(form_elements[e], 'error');
					else
						YAHOO.util.Dom.addClass(form_elements[e], 'valid');
				}
			}
		}
		//show the panel with the error messages (if there were any)
		this._show_error_box();

	},

	/*
	* Manages the panel with the error messages.
	*/
	_show_error_box: function()
	{
		if (BASESOFT.util.is_empty(this.messages))
			return;
		var mss = "";
		for (var i in this.messages) {
			if (i == '__all__')
				mss = mss + this.messages[i] + "<br />";
			else {
				var el_id = YAHOO.util.Dom.getElementsBy(function(el){
					var tag_name = el.tagName.toLowerCase();
					return (el.name == i);
				});
				var el_label = YAHOO.util.Dom.getElementsBy(function(el){
					return (el.tagName.toLowerCase() == 'label') && (el.htmlFor == el_id[0].id);
				});
				mss = mss + el_label[0].innerHTML + ": " + this.messages[i] + "<br />";
			}
		}
		BASESOFT.env.growler.growl('FormValidation',
			new BASESOFT.widget.Growl({
				name: 'ValidationError',
				title: gettext('Please correct the following errors'),
				message: mss,
				priority: 2, // high priority
				lifetime: 15000
		}));
	}
});

YAHOO.register("iq_validation", null, {version: "1.0.0", build: "1"});
