# Copyright (c) 2009 BaseSoft B.V.
# See LICENSE for details.


import os
import os.path
import re
from subprocess import check_call
from twisted.python.hashlib import md5
from twisted.python.reflect import namedAny

import iqportal
from django.conf import settings
from django.template.loader import render_to_string

from basesoft import logging
from basesoft.web.core.exceptions import SCSSCompileException
logger = logging.getLogger("basesoft.core")

from django.utils import importlib
import gettext as gettext_module
from django.utils.translation import activate, to_locale, get_language
from django.views.i18n import PluralIdx, SimplePlural
from django.utils.text import javascript_quote

import types
import sass

__modules_loaded = False
modules = {}

from threading import Lock

# Use a lock for each load_modules() call. In apache there are multiple threads working in the same module
# environment. The lock prevents multiple threads filling the 'modules' dict at the same time. When the first worker
# is done it will release the lock and other threads will read the __modules_loaded as True
# also see http://trac.office.basesoft.com/iqportal/ticket/2574
load_modules_lock = Lock()

def load_modules():
    global __modules_loaded, modules, module_lock

    # lock it
    load_modules_lock.acquire()

    if not __modules_loaded:

        modules = {}
        for app in settings.INSTALLED_APPS:

            # every app has a i18n module.. add the module here and make every normal module in this app depend on it
            i18n_module_name = 'i18n_'+app.replace('.', '_')
            modules[i18n_module_name] = {
                'type': 'i18n',
                'app': app,
                'requires': []
            }

            try:
                js_modules = namedAny(app+'.js_modules')
            except AttributeError:
                continue

            if not hasattr(js_modules, 'modules'):
                continue

            # module path
            media_path = os.path.dirname(js_modules.__file__) + '/media'

            try:

                # modules can either be a method or private but must return a dict
                # pass in the request object when calling modules
                if callable(js_modules.modules):
                    temp_modules = js_modules.modules()
                else:
                    temp_modules = js_modules.modules

                assert isinstance(temp_modules, dict)

                # Merge in modules
                for module_name, data in temp_modules.iteritems():

                    if modules.has_key(module_name):
                        raise Exception('Module already defined: "%s"' % (module_name))

                    if not data.has_key('type'):
                        raise Exception('type not found in %s' % module_name)

                    if not data.has_key('path'):
                        raise Exception('No path defined for module "%s"' % (module_name))

                    if not data.has_key('requires'):
                        data['requires'] = []

                    # copy requires so we dont get into trouble with references
                    requires = data['requires'][:]

                    if type(data['path']) == types.FunctionType:
                        fullpath = data['path']
                    else:
                        # path can either be absolute or relative
                        if data['path'].startswith('/'):
                            fullpath = data['path']
                        else:
                            # rewrite path to the media dir where the module came from
                            # we assume that the path is relative to the module's media dir
                            fullpath = media_path + '/' + data['path']

                        # assert module definitions dont contain -debug
                        assert fullpath.find('-debug') == -1, 'debug in filename definition: %s' % fullpath

                        # load debug files in debug mode
                        if settings.YUI_DEBUG_FILES:
                            fullpath = fullpath.replace('-min.', '-debug.')

                        # when yui debug is off, expect the compiled version to exist
                        if not settings.YUI_DEBUG_FILES and data['type'] == 'es6':
                            fullpath = fullpath.replace('-min.es6', '-min-compiled.js')

                        # check if the file exists
                        if not os.path.exists(fullpath):

                            # some yui files (like reset) have -min files defined but physically
                            # only exists reset.css.. wierness!
                            if settings.YUI_DEBUG_FILES:
                                fullpath = fullpath.replace('-debug.', '.')
                            else:
                                fullpath = fullpath.replace('-min.', '.')

                            if not os.path.exists(fullpath):
                                raise Exception('File not found: %s' % fullpath)

                        # Skinning is setup to be compatible with YUI skinning and works as follows. Whenever a module has
                        # the 'skinnable' boolean set to true it means it supports skinning. In that case a new module is injected
                        # where for now we assume it loads the an extra css file with the 'sam' skin from the assets/skins/sam
                        # directory
                        if data.has_key('skinnable') and data['skinnable']:

                            # the base path of the skin is the module's dir.. from there on take the module name skin file
                            skin_path = os.path.dirname(fullpath) + '/assets/skins/sam/' + module_name + '-min.css'
                            skin_module_name = 'skin-'+module_name

                            if settings.YUI_DEBUG_FILES:
                                skin_path = skin_path.replace('-min.', '.')

                            if not os.path.exists(skin_path):
                                raise Exception('module is skinnable but skin css not found: %s' % skin_path)


                            # merge in the skin module
                            modules[skin_module_name] = {
                                'type': 'css',
                                'fullpath': skin_path,
                                'requires': []
                            }

                            # add dependency on the main module to the new skin module
                            requires += [skin_module_name]

                            # add dependency on the main module to the new skin module
                            requires += [skin_module_name]


                    # all good now.. merge the module in
                    modules[module_name] = {
                        'type': data['type'],
                        'requires': requires + [i18n_module_name],
                        'fullpath': fullpath
                    }

            except Exception, e:
                print ('Could not load js_modules.py from "%s": %s' % (app, e.message))
                raise




        __modules_loaded = True

    # release the lock
    load_modules_lock.release()

    return modules


# list used in the loader_get() view which caches all md5 filenames from
# files that have already been served so we know they exist
_cached_loaders = []


def loader_get_url(request, content_type, exclude_modules, requires, use_cache=True):
    assert isinstance(exclude_modules, list)
    assert isinstance(requires, list)
    assert content_type in ('js', 'css')

    # a theme must be set in session, else fallback
    if not request.session.has_key('theme_module_name'):
        request.session['theme_module_name'] = 'basesoft_theme_default'
        logger.info('session has no theme set, falling back to basesoft_theme_default')

    # add base requires before any other requires
    # this is special sass logic for variables and themeing and shouldnt include real css code
    base_requires = ['default_variables_scss']
    base_requires += ['%s_scss' % request.session['theme_module_name']]
    requires = base_requires + requires

    to_hash = ''.join(requires + exclude_modules)
    if content_type == 'js': # we need to cache per language for JS files.
        to_hash = '%s_%s' % (to_hash, get_language())

    # make hashes unique per version to avoid bad caching
    to_hash += settings.VERSION

    # create md5 filename from requires
    file_name = '%s.%s' % (md5(to_hash).hexdigest(), content_type)
    assert os.path.exists(settings.STATIC_SQUASHED_ROOT), 'STATIC_SQUASHED_ROOT does not exist: %s' % settings.STATIC_SQUASHED_ROOT
    abs_file = os.path.join(settings.STATIC_SQUASHED_ROOT, file_name)

    # create file if its not already known and when doesnt exist yet
    # but when debug is on always create a new file
    if settings.YUI_DEBUG_FILES or (not file_name in _cached_loaders and not os.path.exists(abs_file)) or not use_cache:

        c = CodeSquasher(requires, exclude_modules, request, locale=request.LANGUAGE_CODE)

        if content_type == 'js':
            code = c.js_code_as_string()
        elif content_type == 'es6':
            code = c.es6_code_as_string()
        elif content_type == 'css':
            try:
                code = c.css_code_as_string()
            except Exception, e:

                raise Exception('Failed to compile: %s: %s' % (unicode(requires), e.message ))
        else:
            raise Exception('Invalid bootstrap type: %s' % content_type)

        f = open(abs_file, 'w')
        f.write(code.encode('utf-8'))
        f.close()

        if use_cache:
            _cached_loaders.append(file_name)

    return '/static/%s' % file_name



class CodeSquasher:
    """
    Give this class a list of modules to load and a list of modules to be excluded and using the
    js_code_as_string and css_code_as_string it returns the squashed contents of these modules in one string.
    """

    def __init__(self, load_modules, exclude_modules=None, request=None, locale='en'):
        """
        @param load_modules: list of modules (like ['yahoo', 'dom', 'event']
        @param exclude_modules: list of modules to be excluded
        @param request: optional django request object which will be passed into the modules call for additional permission checking
        """

        if exclude_modules is None:
            exclude_modules = []

        self._load_modules = load_modules
        self._exclude_modules = exclude_modules
        self._request = request
        self._locale = locale

        self._modules = self._init_modules()

    def _init_modules(self):
        """
        This loops all applications and  creates a CLIENT_MODULES variable used in the CodeSquasher class. This is not
        a module method so that any dynamic modules can be inserted from applications depending on privs or whatever.
        """
        return load_modules()

    def _process_depends(self, module_name):
        """
        recursive method gets called by 'module_list' and returns a list of modules with dependencies
        in the correct order to be loaded. It takes the 'exclude_modules' into account.
        """

        if not self._modules.has_key(module_name):
            raise Exception('module "%s" does not exist' % module_name)

        # recursive processes dependencies
        depends = []
        for dependency in self._modules[module_name]['requires']:

            if callable(dependency):
                dependency = dependency()
            if dependency is None:
                continue

            # and only process dependency if not in excluded
            if dependency in self._exclude_modules:
                continue

            # add parent dependencies
            depends += self._process_depends(dependency)

        # finally add the current module as last.. but only if not excluded
        if  module_name not in self._exclude_modules:
            return depends + [module_name]
        else:
            return depends

    def module_list(self, type):
        """
        returns the full list of modules to be loaded including dependencies
        """

        rv = []
        for module_name in self._load_modules:

            # load the dependency list
            for dependency in self._process_depends(module_name):
                if dependency not in rv:

                    module_type = self._modules[dependency]['type']

                    # switch type.. if its js then also include i18n dependencies
                    if type == 'js':
                        if  module_type == 'js' or module_type == 'i18n' or module_type == 'es6':
                            rv.append(dependency)
                    elif type == 'css' and module_type == 'css':
                        rv.append(dependency)
                    elif type == 'scss' and module_type == 'scss':
                        rv.append(dependency)

        return rv

    def _get_code(self, module_name):
        """
        returns the full sourcecode for the given module_name
        """
        module = self._modules[module_name]

        # threat i18n modules a bit different.. they dont have a path
        if module['type'] == 'i18n':
            return self._get_i18n_code(module_name)

        if module['type'] == 'es6':
            return self._get_es6_code(module_name)

        # get css and js contents
        if type(module['fullpath']) == types.FunctionType:
            contents = module['fullpath'](self._request)
        else:
            path = module['fullpath']
            f = open(path, 'r')
            contents = f.read()
            f.close()

        # fix all urls in the css to use absolute paths. We assume that all relative images are in yui files
        # so just point those to the sam skin path
        if module['type'] == 'css':

            url = '/site_media/basesoft.web.contrib.yui/assets/skins/sam/'
            p = re.compile('(?:url)\((?!/)([^\)]*)')
            contents =  p.sub(r'url(%s\1)' % url, contents)

            # the regex above fixes urls but cant figure out relative paths so use a quick hack to replace this..
            contents = contents.replace('/../../../../assets/skins/sam', '')

            # it seems that urls are double closed.. regex needs fixin? well this workies
            contents = contents.replace('))', ')')

        return contents

    def _get_i18n_code(self, module_name):
        """
        returns the i18n catalog javascript code for the given module_name the same way django does
        """

        # current module
        module = self._modules[module_name]

        # basic vars
        t = {}
        domain = 'djangojs'
        package = module['app']

        # sets the current locale so things like get_language() and fallback locals work
        activate(self._locale)
        default_locale = to_locale(settings.LANGUAGE_CODE)
        locale = to_locale(get_language())

        # first load all english languages files for defaults
        p = importlib.import_module(package)
        path = os.path.join(os.path.dirname(p.__file__), 'locale')
        try:
            catalog = gettext_module.translation(domain, path, ['en'])
            t.update(catalog._catalog)
        except IOError:
            # 'en' catalog was missing. This is harmless.
            pass

        # next load the settings.LANGUAGE_CODE translations if it isn't english
        if default_locale != 'en':
            try:
                catalog = gettext_module.translation(domain, path, [default_locale])
            except IOError:
                catalog = None
            except Exception, e:
                raise Exception('Cannot load i18n files from %s, %s, %s: %s' % (path, domain, default_locale, e.message))

            if catalog is not None:
                t.update(catalog._catalog)

        # last load the currently selected language, if it isn't identical to the default.
        if locale != default_locale:
            try:
                catalog = gettext_module.translation(domain, path, [locale])
            except IOError:
                catalog = None
            except Exception, e:
                raise Exception('Cannot load i18n files from %s, %s, %s: %s' % (path, domain, default_locale, e.message))

            if catalog is not None:
                t.update(catalog._catalog)

        # rest is similar to the way django handles translations..
        # see package django.views.i18n.javascript_catalog
        src = []
        plural = None
        if '' in t:
            for l in t[''].split('\n'):
                if l.startswith('Plural-Forms:'):
                    plural = l.split(':',1)[1].strip()
        if plural is not None:
            # this should actually be a compiled function of a typical plural-form:
            # Plural-Forms: nplurals=3; plural=n%10==1 && n%100!=11 ? 0 : n%10>=2 && n%10<=4 && (n%100<10 || n%100>=20) ? 1 : 2;
            plural = [el.strip() for el in plural.split(';') if el.strip().startswith('plural=')][0].split('=',1)[1]
            src.append(PluralIdx % plural)
        else:
            src.append(SimplePlural)
        csrc = []
        pdict = {}
        for k, v in t.items():
            if k == '':
                continue
            if type(k) in (str, unicode):
                csrc.append("catalog['%s'] = '%s';\n" % (javascript_quote(k), javascript_quote(v)))
            elif type(k) == tuple:
                if k[0] not in pdict:
                    pdict[k[0]] = k[1]
                else:
                    pdict[k[0]] = max(k[1], pdict[k[0]])
                csrc.append("catalog['%s'][%d] = '%s';\n" % (javascript_quote(k[0]), k[1], javascript_quote(v)))
            else:
                raise TypeError, k
        csrc.sort()
        for k,v in pdict.items():
            src.append("catalog['%s'] = [%s];\n" % (javascript_quote(k), ','.join(["''"]*(v+1))))

        src.extend(csrc)
        src = ''.join(src)

        return src

    def _get_es6_code(self, module_name):
        """
        Gets the javascript compiled code for the given ES6 module.

        ES6 modules must be defined - similar to normal JS files - with the -min postfix in js_modules and a its file
        called with -debug.js. For example, a file called 'mymodule-debug.es6' should have path 'mymodule-min.es6'

        When YUI_DEBUG_FILES is on the code below compiles directly from mymodule-debug.es6 to mymodule-debug-compiled.js
        and that is served to the client.

        When YUI_DEBUG_FILES are off, below commands are either ran by the buildbot makefile or build_deb.sh
         * compiles_es6.sh compiles mymodule-debug.es6 -> mymodule-debug-compiled.js
         * yui/compress.sh compresses mymodule-debug-compiled.js -> mymodule-min-compiled.js
        """

        module = self._modules[module_name]

        if not settings.ES6_COMPILE:
            if not settings.DEBUG:
                raise Exception('native es6 can only be on in DEBUG')

            f = open(module['fullpath'], 'r')
            contents = f.read()
            f.close()

            return contents


        path, filename = os.path.split(module['fullpath'])
        filename, extension = os.path.splitext(filename)


        if settings.YUI_DEBUG_FILES:
            compiled_file = os.path.join(path, '%s-compiled.js' % filename)
        else:
            compiled_file = os.path.join(path, '%s.js' % filename)

        if settings.YUI_DEBUG_FILES:

            if not os.path.exists(compiled_file) or os.path.getmtime(module['fullpath']) > os.path.getmtime(compiled_file):

                os.chdir(os.path.join(os.path.dirname(os.path.abspath(iqportal.__file__)), '..'))

                cmd = [
                    './node_modules/.bin/babel',
                    '--source-maps',
                    '--out-file', compiled_file,
                    module['fullpath']
                ]
                print 'executing: %s' % ' '.join(cmd)
                check_call(cmd)

        assert os.path.exists(compiled_file), "path exists: %s" % compiled_file
        f = open(compiled_file, 'r')
        contents = f.read()
        f.close()

        return contents


    def _template_vars(self, type):
        """
        builds the template variables dict that can either go into the css or js template
        """
        modules = self.module_list(type)

        if type == 'scss':
            code = ''.join([ ("\n /* ---- BEGIN %s--- */\n\n" %  module_name) + self._get_code(module_name) + ("\n /* ---- END %s ---- */\n\n" %  module_name) for module_name in modules])

            try:
                output_style = 'compressed'
                if settings.DEBUG:
                    output_style = 'nested'
                code = sass.compile(string=code, output_style=output_style)
            except Exception, e:

                if settings.DEBUG:

                    # write failed compiled string to a file
                    import random, string
                    filename = ''.join(random.SystemRandom().choice(string.ascii_uppercase + string.digits) for _ in range(8)) + '.scss'
                    filename = os.path.join(settings.WRITABLE_DIRECTORY, filename)
                    with open(filename, 'w') as f:
                        f.write(code)
                        print '!! libsass pre-compiled output written to %s' % filename

                    regex = r'^:([\d]+):'
                    m = re.search(regex, str(e))
                    if m is not None:
                        raise SCSSCompileException('Failed to compile. Error: %s, code: %s' % (e.message, "\n".join(code.split("\n")[int(m.group(1))-5:int(m.group(1))+5])))
                raise SCSSCompileException('Failed to compile. Error: %s, code:' % (e.message))

            modules_names = ','.join(modules)

            return {'modules': [{'name': modules_names, 'code': code, 'code_len': len(code)/1024}]}

        template_vars = {'modules': []}
        for module_name in modules:

            code = self._get_code(module_name)

            template_vars['modules'].append({
                'name': module_name,
                'code': code,
                'code_len': len(code) / 1024
            })

        return template_vars

    def js_code_as_string(self):
        """
        Returns a string of code for all requested JavaScript modules squashed into one
        """
        return render_to_string('squash_js.html', self._template_vars('js'))

    def css_code_as_string(self):
        """
        Returns a string of code for all requested CSS modules squashed into one
        """
        m = self._template_vars('scss')
        m['modules'] = self._template_vars('css')['modules'] + m['modules']

        return render_to_string('squash_css.html', m)

