# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

from threading import Lock
from datetime import date, datetime

from django.conf import settings
from django.http import HttpResponse, HttpResponsePermanentRedirect
from django.shortcuts import render_to_response
from django.template.context import RequestContext
from django.utils.cache import add_never_cache_headers, patch_response_headers
from django.views.decorators.cache import cache_page
from django.utils.translation import get_language
from twisted.python.reflect import namedAny

from basesoft.web.core.jsonrpc.dispatcher import JsonRpcDispatcher
from basesoft.web.core.util.controllers import AbstractController, lazy_assert
from basesoft.web.contrib.account.decorators import login_required
from basesoft.web.core.util.parse import CodeSquasher, loader_get_url

from basesoft import logging
from basesoft.web.core.exceptions import PublicSystemException, PublicUserException

logger = logging.getLogger("basesoft.core")

from hashlib import md5, sha1

import random
import subprocess

from pyPdf import PdfFileReader

def blank(request):
    return HttpResponse('')

def bootstrap(request, version, content_type, extra_requires):
    """
    View for type=js|css that returns the initial JS code that is executed at first when the page loads.
    This includes all global objects included in all applications.
    """

    # parse extra_requires
    if extra_requires == '':
        extra_requires = []
    else:
        extra_requires = extra_requires.split(',')

    # setup requires
    requires = [
        'iq_core',
        'iq_loader',
        'iqportal_init',
        'iq_growl',
    ] + extra_requires

    return loader_get(request, version, content_type, '', ','.join(requires))


def loader_get(request, version, content_type, exclude_modules, requires):

    """
    Loader url using a GET request to ask for the source for the given requested requires.
    @param content_type: string either 'css' or 'js'
    @param exclude_modules: string of comma separated modules string that already have been loaded clientside
    @param requires: string of comma seperated modules strings that are requested to be loaded
    """

    if exclude_modules == '':
        exclude_modules = []
    else:
        exclude_modules = exclude_modules.split(',')

    if requires == '':
        requires = []
    else:
        requires = requires.split(',')

    url = loader_get_url(request, content_type, exclude_modules, requires)
    lazy_assert(url, str)

    return HttpResponsePermanentRedirect(url)


import os
__bootstrap_registry_defined = False
bootstrap_registry = ''

def js_module_boot_register(request):
    """
    Loops all apps and squashes the bootstrap_register.js files. This is loaded at startup of the page so that the
    bootloader knows which applications are enabled.
    """
    global __bootstrap_registry_defined
    global bootstrap_registry
    if not __bootstrap_registry_defined:
        apps = settings.INSTALLED_APPS

        for app in apps:

            init_module = namedAny(app+'.__init__')
            path = os.path.dirname(init_module.__file__) + '/media'

            if os.path.exists(path):

                file = path + '/boot_register-min.js'
                if settings.YUI_DEBUG_FILES:
                    file = file.replace('-min.', '-debug.')

                if os.path.exists(file):
                    f = open(file, 'r')
                    bootstrap_registry += f.read()

        __bootstrap_registry_defined = True

    return bootstrap_registry

from basesoft.web.core.util.custom_query import CustomQueryRpcHandler

def custom_query(request, name):
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(CustomQueryRpcHandler(name, request.SessionReadOnly(), account_id=request.session['user_id']))
    return dispatcher.handle_request(request)

_handlers = {}
_handlers_loaded = False


# load_handlers_lock protects _load_handlers to be called simultaniously in different threads
load_handlers_lock = Lock()
def _load_handlers():

    load_handlers_lock.acquire()

    global _handlers_loaded, _handlers
    if _handlers_loaded:
        load_handlers_lock.release()
        return

    for app in settings.INSTALLED_APPS:
        init_module = namedAny(app+'.__init__')
        path = os.path.dirname(init_module.__file__) + '/rpc.py'

        if not os.path.exists(path):
            continue
        module = __import__(app + '.rpc', {}, {}, [''])

        if not hasattr(module, 'handlers'):
            raise Exception('%s - rpc module has no attribute handlers' % app)

        assert isinstance(module.handlers, (list, tuple))
        for handler in module.handlers:
            _handlers['%s.%s' % (app, handler.__name__)] = handler

    _handlers_loaded = True

    load_handlers_lock.release()

def rpc(request, name):

    _load_handlers()
    dispatcher = JsonRpcDispatcher()
    try:
        HandlerClass = _handlers[name]
    except IndexError:
        raise Exception('View not found: %s' % name)

    dispatcher.register_instance(HandlerClass(request))
    return dispatcher.handle_request(request)

def objects_rpc(request):
    controller = AbstractController(request.Session(), account_id=request.session['user_id'])

    class Handler:

        def mark_as_deleted(self, object_name, ids):
            return controller.mark_as_deleted(object_name, ids)

        def mark_as_archived(self, object_name, ids):
            return controller.mark_as_archived(object_name, ids)

        def unmark_as_deleted(self, object_name, ids):
            return controller.unmark_as_deleted(object_name, ids)

        def unmark_as_archived(self, object_name, ids):
            return controller.unmark_as_archived(object_name, ids)

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(Handler())
    return dispatcher.handle_request(request)

# ------------------------------------------------

@login_required
def download_xls(request, hash):
    """
    Download xls exported for customdatatables
    """

    path = os.path.join(settings.TMP_FILE_REPOSITORY, 'data_%s.xls' % hash)

    if not os.path.exists(path):
        raise PublicUserException('XLS file no longer available, please try exporting again.')

    _file = open(path, 'r')

    response = HttpResponse(_file, mimetype='application/excel')
    response['Content-Disposition'] = 'attachment; filename=data_%s.xls' % date.strftime(datetime.now(), '%d-%m-%y_%H-%I')

    # remove the file
    os.unlink(path)
    return response


def download_pdf(request, url_path):
    """
    Creates a pdf document for the given url_path. This could be a compose runner in print mode
    """

    # if pdf doesnt work, enable code below:
    #return render_to_response('pdf_unavailable.html', {
    #    'url_path': url_path
    #}, context_instance=RequestContext(request))

    # fix for #3080: in apache environment double slashes get stripped so the url_path will arrive without first slash
    if url_path[0] != '/':
        url_path = '/' + url_path

    # build url hased on url_path. if we need external urls in the future we can add a check for this.
    protocol = 'http://'

    # #5044: https disabled for wkhtml
    #if request.is_secure():
    #    protocol = 'https://'

    url = ''.join((protocol, request.META['HTTP_HOST'], url_path))

    temp_file = os.path.join(settings.TMP_FILE_REPOSITORY, 'download_%s.pdf' % sha1(str(random.random())).hexdigest())

    footer_html = os.path.abspath(os.path.dirname(__file__)) + '/templates/pdf_footer.html'

    # Prepare the pdf convert command..
    #
    # Important: Javascript delay is set to 2 mins which should never be reached. It gives the document time to render
    # images n stuff. When the document is done it should call window.print() which tells the pdf printer to convert to pdf
    # All documents should call window.print so also headers and footers
    #

    cmd = [settings.WKHTMLTOPDF_BIN,
           "--print-media-type",
           "--javascript-delay", "120000", # 2 mins
           "--footer-right", "[page]/[toPage]",
           "--cookie","sessionid", request.session.session_key,
           "--debug-javascript",
           "--footer-html", footer_html,
           "--load-error-handling", "ignore",
           url,
           temp_file
    ]

    logger.info('download_pdf: cmd: %s' % " ".join(cmd))

    # spawn the converter process. very important is here not to use shell=True for command injection
    proc = subprocess.Popen(cmd)

    # wait till program is done
    # we dont check for error code because that seems to be unreliable. If the html contains a non existing image (404) or
    # any error code then the wkhtmltopdf returns an error code. Instead just check if it generated any pdf file.
    proc.wait()

    if not os.path.exists(temp_file):
        raise Exception('Failed to create pdf')

    # use a pdf reader to extract filename from created pdf file (its taken from html title)
    pdf_reader = PdfFileReader(open(temp_file, "r"))
    filename = "%s.pdf" % pdf_reader.getDocumentInfo().title

    # present as download
    response = HttpResponse(open(temp_file, 'r'), mimetype='application/pdf')
    response['Content-Disposition'] = 'attachment; filename="%s"' % filename # date.strftime(datetime.now(), '%d-%m-%y_%H-%I')

    return response

def download_screenshot(request, url_path):

    if settings.WKHTMLTOIMAGE_BIN is None:
        raise PublicSystemException('Cannot generate screenshot')

    # fix for #3080: in apache environment double slashes get stripped so the url_path will arrive without first slash
    if url_path[0] != '/':
        url_path = '/' + url_path

    # build url hased on url_path. if we need external urls in the future we can add a check for this.
    protocol = 'http://'
    if request.is_secure():
        protocol = 'https://'
    url = ''.join((protocol, request.META['HTTP_HOST'], url_path))

    temp_file = os.path.join(settings.TMP_FILE_REPOSITORY, 'download_%s.png' % sha1(str(random.random())).hexdigest())

    cmd = [settings.WKHTMLTOIMAGE_BIN,
           "--javascript-delay", "30000", # 1 min
           "--cookie","sessionid", request.session.session_key,
           "--debug-javascript",
           url,
           temp_file
    ]

    # spawn the converter process. very important is here not to use shell=True for command injection
    proc = subprocess.Popen(cmd)

    # wait till program is done
    # we dont check for error code because that seems to be unreliable. If the html contains a non existing image (404) or
    # any error code then the wkhtmltopdf returns an error code. Instead just check if it generated any pdf file.
    proc.wait()

    if not os.path.exists(temp_file):
        raise Exception('Failed to create screenshot')

    response = HttpResponse(open(temp_file, 'r'), mimetype='image/png')
    response['Content-Disposition'] = 'attachment; filename="screenshot.png"'

    return response
