# Copyright (c) 2009 BaseSoft B.V.
# See LICENSE for details.

"""
This contains the controllers for mockup compose
"""

from sqlalchemy import and_
from basesoft.web.core.util.controllers import AbstractController, transaction_decorator

from basesoft.web.dev.compose.models import Mockup, MockupCategory, MockupCategoryGroup
from basesoft.web.contrib.license.models import Group
from basesoft.web.core.exceptions import SecurityException

from basesoft import logging
logger = logging.getLogger("basesoft.web.dev.compose")

class MockupComposeController(AbstractController):
    """
    This should be called from a view
    """

    def get_mockups(self, id=None, order_by=None, category=None):
        """
        Return the mockups the user has access to
        @param id: None, int or list of ints. the id(s) of the mockups we want to return or None if we want all
        @param order_by: the column after which we want to make the ordering, or None we order by last_updated
        @return: a list of Mockup objects
        """
        self.assert_permissions('develop_tools')

        if order_by is None:
            order_by = Mockup.last_updated

        filter = None
        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = Mockup.id.in_(id)

        if category is not None:
            assert isinstance(category, MockupCategory)
            if filter is None:
                filter = Mockup.category==category
            else:
                filter = and_(filter, Mockup.category==category)

        return self._get_privileged_object(Mockup, filter, Mockup.PRIV_VIEW, order_by=order_by)

    def get_categories(self, id=None):
        """
        Return the mockup categories the user has access to
        @param id: None, int or list of ints. the id(s) of the mockup categories we want to return or None if we want all
        @return: a list of MockupCategory objects
        """
        self.assert_permissions('develop_tools')
        return self.get_privileged_object_by_id(MockupCategory, id, MockupCategory.PRIV_VIEW)

    def get_category_by_label(self, label):
        """
        Return the mockup categories the user has access to
        @return: a list of MockupCategory objects
        """
        self.assert_permissions('develop_tools')
        filter, fo = self.get_query_filter_on_privileges(MockupCategory, MockupCategoryGroup, MockupCategory.PRIV_VIEW)
        filter = and_(filter, MockupCategory.label==label)
        return self.sa_session.query(MockupCategory).select_from(fo).filter(filter).all()

    def save_or_update_category(self, id, label, group=None,
                               group_privs=MockupCategory.all_privileges,
                               license_privs=MockupCategory.no_privileges,
                               domain_privs=MockupCategory.no_privileges):
        """
        Saves or updates a mockup category
        """
        self.assert_permissions('develop_tools')

        if group is None:
            group = self._get_default_group()
        else:
            assert isinstance(group, Group)

        if id is None:
            mc = MockupCategory()
        else:
            mc = self.get_categories(id)
            if len(mc) == 0:
                raise SecurityException("Invalid MockupCategory id supplied")
            mc = mc[0]
            self.assert_privileges_on_object(mc, MockupCategory.PRIV_UPDATE)
        mc.label = label

        mc.group = group
        mc.group_privileges = group_privs
        mc.license_privileges = license_privs
        mc.domain_privileges = domain_privs

        self.sa_session.add(mc)
        self.sa_session.flush()
        return mc

    @transaction_decorator
    def save_state(self, state, group=None,
                   group_privs=MockupCategory.all_privileges,
                   license_privs=MockupCategory.no_privileges,
                   domain_privs=MockupCategory.no_privileges):
        """
        Receives a state and saves it. Returns the Mockup obj.
        @param state: the state that we need to save
        """
        self.assert_permissions('develop_tools')
        from django.utils import simplejson
        from basesoft.web.core.jsonrpc.dispatcher import JsonRpcJSONEncoder
        state = state['mockup']

        if group is None:
            group = self._get_default_group()
        else:
            assert isinstance(group, Group)

        if state['dbID'] == None:
            f = Mockup()
        else:
            f = self.get_mockups(state['dbID'])[0]
            self.assert_privileges_on_object(f, Mockup.PRIV_UPDATE)

        f.name = state['name']
        cat = self.get_category_by_label(state['category']['label'])
        if len(cat) > 0:
            cat = cat[0]
        else:
            cat = self.save_or_update_category(None, state['category']['label'])
        f.category = cat
        state['category']['id'] = cat.id
        f.state = simplejson.dumps(state, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        f.group = group
        f.group_privileges = group_privs
        f.license_privileges = license_privs
        f.domain_privileges = domain_privs

        self.sa_session.add(f)
        self.sa_session.flush()
        return f

    def load_state(self, mockup_id):
        """
        Returns a state of the mockup for the given mockup_id
        """
        from django.utils import simplejson
        from basesoft.web.core.jsonrpc.dispatcher import json_object_hook

        f = self.get_mockups(mockup_id)
        if len(f) == 0:
            raise SecurityException("load_state: Invalid Mockup id supplied")
        f = f[0]
        state = simplejson.loads(f.state, object_hook=json_object_hook)
        state['dbID'] = f.id
        return {'mockup': state}

    def delete_mockups(self, id):
        """
        Method for removing mockups with ids
        @param ids array or int of mockup id(s) to be removed
        """
        self.assert_permissions('develop_tools')
        mockups = self.get_mockups(id)
        for mockup in mockups:
            self.mark_as_deleted(Mockup, mockup.id)
        self.sa_session.flush()

    def __repr__(self):
        return '<MockupComposeController>'