'''
Created on Nov 20, 2013

@author: martijn
'''

import os
import re
import xml.etree.ElementTree as ET
import csv
import cStringIO
import codecs

from django.conf import settings
from sqlalchemy.orm import outerjoin
from sqlalchemy.sql.expression import or_, and_

from basesoft.scheduler.task import ScheduleTask
from basesoft.web.contrib.license.models import License
from basesoft.web.contrib.i18n.models import Country, DomainCountry, Language, DomainLanguage
from basesoft.web.contrib.notification.controllers import NotificationController


class ArvatoImportXml(ScheduleTask):

    INTERVAL = 3600

    def execute(self):

        license_ = self.sa_session.query(License).filter(License.name == settings.ARVATO_LICENSE_NAME).first()
        if license_ is None:
            return

        # get mapping #4273
        fo = outerjoin(Country, DomainCountry, onclause=(and_(Country.id == DomainCountry.country_id, DomainCountry.domain_id == license_.domain.id)))
        q = self.sa_session.query(Country, DomainCountry).select_from(fo).filter(or_(DomainCountry.active == None, DomainCountry.active == 1))
        countries = q.order_by(Country.name).all()
        country_map = {}
        for c, local in countries:
            name = c.name if local is None else local.name
            country_map[name] = c.iso_3166_a2

        language_map = {}
        for language in self.sa_session.query(Language).all():
            language_map[language.name] = language.iso_639_1

        aliases = self.sa_session.query(DomainLanguage).filter(DomainLanguage.domain==license_.domain).all()
        for dl in aliases:
            language_map[dl.name] = dl.language.iso_639_1

        # http://trac.office.basesoft.com/iqportal/ticket/3853
        self._value_map = {
            'ContactCountry': country_map,
            'ContactLanguage': language_map
        }

        assert os.path.exists(settings.ARVATO_LOCAL_PATH)

        # execute moving of files
        exit_code = os.system(settings.ARVATO_GET_XML_CMD)
        if exit_code != 0:
            raise Exception('Command errored with exit code %s: %s' % (exit_code, settings.ARVATO_GET_XML_CMD))

        for on in os.listdir(settings.ARVATO_LOCAL_PATH):

            fn = os.path.join(settings.ARVATO_LOCAL_PATH, on)
            if os.path.isfile(fn):
                self._xml2csv(fn, on)
                os.remove(fn)

    def _xml2csv(self, fn, filename):

        license_ = self.sa_session.query(License).filter(License.name == settings.ARVATO_LICENSE_NAME).one()
        license_id = license_.id

        # create contact dir if not exists
        if not os.path.exists(settings.CONTACT_IMPORT_DIR):
            os.mkdir(settings.CONTACT_IMPORT_DIR)

        # create license dir if not exists
        license_dir = os.path.join(settings.CONTACT_IMPORT_DIR, unicode(license_id))
        if not os.path.exists(license_dir):
            os.mkdir(license_dir)

        # csv destination is the original filename + .csv
        target_csv = os.path.join(license_dir, os.path.basename(fn)+'.csv')

        # open the xml
        tree = ET.parse(fn)
        root = tree.getroot()

        # a list with import errors.
        notify = []

        # start writing to the csv
        with open(target_csv, 'wb') as csvfile:

            writer = UnicodeWriter(csvfile)
            columns = [
                'ContactGender',
                'ContactFirstname',
                'ContactLastname',
                'ContactEmail',
                'ContactLanguage',
                'ContactCountry',
                'CaseId',
                'ContactChannel',
                'ContactCompany',
                'CaseCreated',
                'CaseCreatedTeam',
                'CaseModified',
                'CaseModifiedTeam',
                'CaseStatus',
                'Complaint',
                'Taxonomy1',
                'Taxonomy2',
                'Taxonomy3'
            ]

            # write csv header
            writer.writerow(columns)

            for xml_contact in list(root):

                values = [''] * len(columns)

                add_row = True
                for xml_kv in list(xml_contact):

                    # use regex to strip namespace from tag
                    tag = re.sub('^\{.*\}', '', unicode(xml_kv.tag))
                    value = xml_kv.text

                    if self._value_map.has_key(tag):
                        try:
                            value = self._value_map[tag][value]
                        except KeyError:
                            msg = 'No mapping found for column "%s" and value "%s"' % (tag, value)
                            if msg not in notify:
                                notify.append(msg)
                            add_row = False
                            continue

                    values[ columns.index(tag) ] = value

                if add_row:
                    writer.writerow(values)

        if len(notify) > 0:
            # we have to send a notification to the Arvato group.
            nc = NotificationController(self.sa_session)
            notification = nc.set_notification(
                'The XML import skipped some lines',
                'The following errors were encountered in the XML file (%(filename)s):\n%(lines)s' % {
                    'filename': filename,
                    'lines': '\n'.join(notify)
                }
            )
            nc.create_system_notification(notification, license_.default_group)


class UnicodeWriter:
    """
    from http://docs.python.org/2/library/csv.html
    """

    def __init__(self, f, dialect=csv.excel, encoding="utf-8", **kwds):
        self.queue = cStringIO.StringIO()
        self.writer = csv.writer(self.queue, dialect=dialect, **kwds)
        self.stream = f
        self.encoder = codecs.getincrementalencoder(encoding)()

    def writerow(self, row):
        self.writer.writerow([s.encode("utf-8") for s in row])
        data = self.queue.getvalue()
        data = data.decode("utf-8")
        data = self.encoder.encode(data)
        self.stream.write(data)
        self.queue.truncate(0)

    def writerows(self, rows):
        for row in rows:
            self.writerow(row)

tasks = (ArvatoImportXml, )

