# Copyright (c) 2009 BaseSoft B.V.
# See LICENSE for details.

from django.utils import simplejson


from sqlalchemy import and_
from sqlalchemy.orm.exc import NoResultFound

from basesoft.web.core.jsonrpc.dispatcher import json_object_hook, JsonRpcJSONEncoder
from basesoft.web.contrib.compose.models import ComposeInvitee, ComposePublicationGroup, ComposePublicationLocale
from basesoft.web.core import exceptions as basesoft_exceptions
from basesoft.web.core.util.controllers import AbstractController, transaction_decorator, AbstractStateController, StateManager

from iqportal.iqreport.core import models as iqreport_models
from iqportal.iqube.core.controllers import DataSourceController, ConfigurableWidgetController

from basesoft import logging
logger = logging.getLogger("iqportal.iqreport.compose")

class ComposeController(AbstractController):
    """
    """
    @transaction_decorator
    def save_state(self, state, group_id=None):
        """
        """
        # reset global id_map
        AbstractStateController.reset_state_map()

        state_controller = ReportStateController(self.sa_session,
                                    StateManager(self.locale, self.account_id, iqreport_models.Report.MODE_COMPOSE))
        report = state_controller.save(state, group_id)

        logger.info('%s: Saved %s' % (self, report) )

        # At this point the whole state has been procesed and modules have been saved to the db. What is left to
        # do is flush the delete_pool because we are sure that these objects arent used anymore.
        AbstractStateController.commit_delete_pool(self.sa_session)

        # return the id map
        return AbstractStateController.save_state_id_map

    def get_reports(self, id=None, order_by=None, priv=iqreport_models.Report.PRIV_VIEW):
        """
        Return the reports the user has access to
        @param id: None, int or list of ints. the id(s) of the surveys we want to return or None if we want all
        @param order_by: the column after which we want to make the ordering, or None we order by name
        @return: a list of survey objects
        """
        self.assert_permissions('view_report')

        filter_, fo = self.get_query_filter_on_privileges(iqreport_models.Report, iqreport_models.ReportGroup, priv)

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter_ = and_(filter_, iqreport_models.Report.id.in_(id))

        filter_ = and_(filter_, iqreport_models.Report.deleted==False)

        if order_by is None:
            order_by = iqreport_models.Report.name
        return self.sa_session.query(iqreport_models.Report).select_from(fo).filter(filter_).order_by(order_by).all()

    def get_themes(self, id=None):
        """
        returns the themes to which the current user has permissions to save reports for
        """
        filter, fo = self.get_query_filter_on_privileges(iqreport_models.ReportTheme, iqreport_models.ReportThemeGroup, iqreport_models.ReportTheme.PRIV_VIEW)

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, iqreport_models.ReportTheme.id.in_(id))

        return self.sa_session.query(iqreport_models.ReportTheme).select_from(fo).filter(filter).all()

    def get_report_publication(self, publication_id):

        filter, fo = self.get_query_filter_on_privileges(iqreport_models.ReportPublication, ComposePublicationGroup, iqreport_models.ReportPublication.PRIV_VIEW)
        filter = and_(filter, iqreport_models.ReportPublication.publication_id==publication_id)
        return self.sa_session.query(iqreport_models.ReportPublication).select_from(fo).filter(filter).one()

    def load_state(self, report_id, publication_id=None):
        """
        Load state is be called by the preview runner.
        If a publication is passed load the state for that publication which is cached but only if the publication is online
        """

        publication = None
        if publication_id is not None:
            publication = self.get_report_publication(publication_id)

            if publication.status < iqreport_models.ReportPublication.STATUS_ONLINE:
                publication = None

        state_controller = ReportStateController(self.sa_session,
                                    StateManager(self.locale, self.account_id, iqreport_models.Report.MODE_COMPOSE))


        state = state_controller.load_by_report_id(report_id, publication)

        state['report']['has_active_publications'] = False
        # check if report has at least one online publication
        q = self.sa_session.query(iqreport_models.ReportPublication).filter(and_(iqreport_models.ReportPublication.report_id==report_id,
            iqreport_models.ReportPublication.status.in_([iqreport_models.ReportPublication.STATUS_ONLINE,
            iqreport_models.ReportPublication.STATUS_FINISHED]))).all()
        if len(q) > 0:
            state['report']['has_active_publications'] = True

        return state

    @transaction_decorator
    def mark_as_deleted(self, id):
        """
        This will mark the report as deleted if the user has the correct rights.
        """
        self.assert_permissions('mark_report_deleted')

        reports = self.get_reports(id=id, priv=iqreport_models.Report.PRIV_MARK_AS_DELETED)
        if len(reports) != 1:
            raise basesoft_exceptions.PublicSecurityException('Survey can not be marked as deleted')

        report = reports.pop()
        report.deleted = True
        self.sa_session.add(report)

    def __repr__(self):
        return '<ComposeController "Report">'

class ReportStateController(AbstractStateController):
    """
    Report controller
    """

    _instance_class = iqreport_models.Report
    _locale_attributes = ['title']

    def __init__(self, session, manager):
        AbstractStateController.__init__(self, session, manager)
        self.blockStateController = BlockStateController(self.sa_session, manager)

    @transaction_decorator
    def save(self, state, group_id=None):
        state = state['report']

        self._assert_state(state, None, dict)
        self._assert_state(state, 'domID', unicode)
        self._assert_state(state, 'objectName', unicode, 'Report')
        self._assert_state(state, 'theme', dict)
        self._assert_state(state['theme'], 'module', (unicode, None))
        self._assert_state(state['theme'], 'name', unicode)
        self._assert_state(state['theme'], 'cssClass', (unicode, None))
        self._assert_state(state, 'children', dict)
        self._assert_state(state['children'], 'blocks', list)

        # save or update
        if state['dbID']:
            self.assert_permissions('update_report')
        else:
            self.assert_permissions('create_report', group_id=group_id)

        # call parent which takes care of common compose logic
        report = AbstractStateController._instance_from_state(self, state)

        # check the object privileges in case of update
        if state['dbID']:
            self.assert_privileges_on_object(report, iqreport_models.Report.PRIV_UPDATE)

        # Save report name
        self._assert_state(state, 'name', unicode)
        report.name = state['name']

        # theme
        try:
            theme = self.sa_session.query(iqreport_models.ReportTheme).filter(iqreport_models.ReportTheme.name==state['theme']['name']).one()
        except NoResultFound:
            raise basesoft_exceptions.PublicUserException('No such theme: %s' % state['theme']['name'])
        report.report_theme = theme
        self.assert_privileges_on_object(theme, iqreport_models.ReportTheme.PRIV_VIEW)

        self._assert_state(state, 'fallbackLocale', dict)
        self._assert_state(state['fallbackLocale'], 'locale', unicode)
        report.fallback_locale = self.locale_controller.get_locale_by_str(state['fallbackLocale']['locale'])

        if state['dbID'] in [None, '']:
            self.save_default_privilege_object_group(report, target_group=group_id)
        else:

            # get all report publications that have a state saved (and are offline) and reset that
            publications = self.sa_session.query(iqreport_models.ReportPublication) \
                            .join(ComposePublicationLocale) \
                            .filter(and_(
                                iqreport_models.ReportPublication.report==report,
                                iqreport_models.ReportPublication.status.in_([
                                    iqreport_models.ReportPublication.STATUS_NEW,
                                    iqreport_models.ReportPublication.STATUS_SCHEDULED]), \
                                )) \
                            .all()

            for p in publications:
                self._save_publication_state(p, None, locale=None)

        # call parent save that actually writes to the db
        AbstractStateController.save(self, report, state)
        self._save_available_locales(report, iqreport_models.ReportLocale, 'report', state)
        # delete any removed blocks and process
        self._delete_children(report.blocks, state['children']['blocks'])
        for i in range(len(state['children']['blocks'])):
            self.blockStateController.save(report, state['children']['blocks'][i], i)

        self._canvas_post_save()

        # finally return the new saved report
        return report

    def load_by_report_id(self, report_id, publication=None):
        """
        Loads and checks permissions for the given report_id
        Returns a state dict
        """
        self.assert_permissions('view_report')
        report = self._instance_from_id(report_id)
        self.assert_privileges_on_object(report, iqreport_models.Report.PRIV_VIEW)

        if publication is not None:
            return self.load_and_save_state_by_publication(publication)
        return self._load_state(report)

    def load_by_invitee(self, invitee):
        if not isinstance(invitee, ComposeInvitee):
            invitee = self.sa_session.query(ComposeInvitee).filter(ComposeInvitee.id==invitee).one()

        publication = invitee.publication
        return self.load_and_save_state_by_publication(publication)

    @transaction_decorator
    def load_and_save_state_by_publication(self, publication):

        # FIXME: should come from contact or something
        locale = publication.report.fallback_locale

        # try to get state from cache else load it
        try:
            state = self.sa_session.query(ComposePublicationLocale) \
                .filter(ComposePublicationLocale.publication==publication, ComposePublicationLocale.locale==locale).one().state

            state = simplejson.loads(state, object_hook=json_object_hook)

        except NoResultFound:

            state = self._load_state(publication.report)
            self._save_publication_state(publication, state, locale)

        return state


    def _load_state(self, report):
        """
        Private load method whithout permission checking.. should be called from a load_by_* method
        This returns a state dict
        """

        gc = None
        if report.report_theme.graph_theme is not None and report.report_theme.graph_theme.graph_config is not None:
            gc = simplejson.loads(report.report_theme.graph_theme.graph_config, object_hook=json_object_hook)

        state = {
            'report': {
                'objectName': 'Report',
                'dbID': report.id,
                'name': report.name,
                'theme': {
                    'name': report.report_theme.name,
                    'cssClass': report.report_theme.css_class,
                    'module': report.report_theme.module,
                    'graph_config': gc
                },
                'availableLocales': [locale.locale for locale in report.available_locales],
                'fallbackLocale': {'locale': report.fallback_locale.locale, 'name': report.fallback_locale.name},
                'children': {
                    'blocks': []
                }
            }
        }

        for block in report.blocks:
            state['report']['children']['blocks'].append(self.blockStateController.load(block.id))

        return state

class BlockStateController(AbstractStateController):
    """
    Takes care of block state logic
    """

    _instance_class = iqreport_models.ReportBlock

    def __init__(self, session, manager):
        AbstractStateController.__init__(self, session, manager)
        self.columnStateController = ColumnStateController(self.sa_session, manager)

    @transaction_decorator
    def save(self, report, state, position):
        """
        """
        assert isinstance(report, iqreport_models.Report)
        self._assert_state(state, None, dict)
        assert isinstance(position, int)
        self._assert_state(state, 'objectName', unicode, 'ReportBlock')

        self._assert_state(state, 'pageEnd', bool)
        self._assert_state(state, 'children', dict)
        self._assert_state(state['children'], 'columns', list)

        # call parent which takes care of common compose logic
        block = AbstractStateController._instance_from_state(self, state)

        # reference method arguments
        block.report = report
        block.position = position
        block.page_end = state['pageEnd']

        # call parent save that actually writes to the db
        AbstractStateController.save(self, block, state)

        # delete any removed columns
        self._delete_children(block.columns, state['children']['columns'])

        # continue processing columns
        for i in range(len(state['children']['columns'])):
            self.columnStateController.save(block, state['children']['columns'][i], i)

        # expire survey (so it will update survey.blocks)
        self.sa_session.expire(report)

    def load(self, id):
        block = self._instance_from_id(id)

        state = {
            "objectName": u"ReportBlock",
            "dbID": block.id,
            "pageEnd": bool(block.page_end),
            "children": {
                'columns': []
            }
        }

        for column in block.columns:
            state['children']['columns'].append(self.columnStateController.load(column.id))

        return state

class ColumnStateController(AbstractStateController):
    """
    Takes care of state handling for columns
    """

    _instance_class = iqreport_models.ReportBlockColumn

    @transaction_decorator
    def save(self, block, state, position):
        assert isinstance(block, iqreport_models.ReportBlock)
        self._assert_state(state, None, dict)
        assert isinstance(position, int)
        self._assert_state(state, 'objectName', unicode, 'ReportBlockColumn')

        self._assert_state(state, 'children', dict)
        self._assert_state(state['children'], 'items', list)

        # call parent which takes care of common compose logic
        column = AbstractStateController._instance_from_state(self, state)

        column.block = block
        column.width_percentage = state['width']
        column.position = position

        # call parent save that actually writes to the db
        AbstractStateController.save(self, column, state)

        # delete any removed items
        self._delete_children(column.items, state['children']['items'])

        # process items
        for i in range(len(state['children']['items'])):
            self.__save_item_state(column, state['children']['items'][i], i)

        # expire
        self.sa_session.expire(block)

    def __save_item_state(self, column, state, position):
        """
        Takes care of the saving state of items by mapping the item controller and instanciating it.
        """
        assert isinstance(column, iqreport_models.ReportBlockColumn)
        self._assert_state(state, None, dict)

        try:
            # TODO: add doc
            ItemController = item_controller_mapping[state['objectName']]
        except KeyError:
            logger.error('ColumnStateController: Ignoring item saving. No mapper found for %s' % state['objectName'])
        else:
            ic = ItemController(self.sa_session, self.manager)
            ic.save(column, state, position)

    def load(self, id):
        column = self._instance_from_id(id)

        state = {
            "objectName": u"ReportBlockColumn",
            "dbID": column.id,
            "width": column.width_percentage,
            "children": {
                'items': []
            }
        }

        for item in column.items:
            try:
                stateControllerClass = item_controller_mapping[item.__class__.__name__]
            except KeyError:
                logger.error('ColumnStateController: Ignoring item loading. No mapper found for %s' % item.__class__.__name__)
            else:
                stateController = stateControllerClass(self.sa_session, self.manager)
                state['children']['items'].append(stateController.load(item.id))

        return state


class ItemStateController(AbstractStateController):
    """
    """
    """
    Abstract item controller.

    Takes care of common item logic like column, position and item type.
    """

    @transaction_decorator
    def save(self, item, column, state, position):
        assert isinstance(column, iqreport_models.ReportBlockColumn)
        self._assert_state(state, None, dict)
        assert isinstance(position, int)

        item.column = column
        item.position = position

        # call parent save that actually writes to the db
        AbstractStateController.save(self, item, state)

        # expire
        self.sa_session.expire(column)

    def load(self, id):
        pass

class ItemHtmlStateController(ItemStateController):
    """
    Controller for the ItemHtml model
    """

    _instance_class = iqreport_models.ReportItemHtml
    _locale_attributes = ['html']

    @transaction_decorator
    def save(self, column, state, position):
        self._assert_state(state, 'objectName', unicode, 'ReportItemHtml')

        # call super to get/create the iqreport_models.ReportItem and set its properties
        item = AbstractStateController._instance_from_state(self, state)
        ItemStateController.save(self, item, column, state, position)

        # save locales in locale table
        self._save_locale(item, iqreport_models.ReportItemHtmlLocale, state)

        return item


    def load(self, id):
        item = self._instance_from_id(id)

        state = {
            "objectName": u"ReportItemHtml",
            "dbID": item.id,
            'locales': self._load_locales(item),
            'isWiki': item.is_wiki,
        }

        return state

class ItemDataSourceWidgetController(ItemStateController):
    """
    Controller for the QAItem
    """

    _instance_class = iqreport_models.ReportItemDataSourceWidget

    def __init__(self, session, manager):
        ItemStateController.__init__(self, session, manager)
        self.ds_controller = DataSourceController(session, account_id=manager.account_id)
        self.cwc = ConfigurableWidgetController(session, account_id=manager.account_id)

    @transaction_decorator
    def save(self, column, state, position):
        item = ItemStateController._instance_from_state(self, state)

        # if datasource_widget is None we are dealing with 'item results', else normal CW behaviour
        if state['datasource_widget_id'] is None:
            item = self.cwc.save_cube_config(item, state)
            assert item.datasource_cube is not None
        else:
            item.datasource_widget = self.cwc.get_configurable_widgets(state['datasource_widget_id']).pop()

        item.title = state['title']
        item.show_title = state['showTitle']
        item.applied_config = simplejson.dumps(state['appliedConfig'], ensure_ascii=False, cls=JsonRpcJSONEncoder)

        if len(state['graphConfig']) > 0:
            item.graph_config = simplejson.dumps(state['graphConfig'], ensure_ascii=False, cls=JsonRpcJSONEncoder)

            # TODO: verify graph config

        if len(state['tableConfig']) > 0:
            item.table_config = simplejson.dumps(state['tableConfig'], ensure_ascii=False, cls=JsonRpcJSONEncoder)

        if state.has_key('linkedItems'):
            item.temp_linked_items = state['linkedItems']
        else:
            item.temp_linked_items = []

        ItemStateController.save(self, item, column, state, position)

        return item

    def load(self, id):
        item = self._instance_from_id(id)
        gc = {}
        if item.graph_config is not None:
            gc = simplejson.loads(item.graph_config, object_hook=json_object_hook)
        tc = {}
        if item.table_config is not None:
            tc = simplejson.loads(item.table_config, object_hook=json_object_hook)
        li = {}
        if item.linked_items is not None:
            li = simplejson.loads(item.linked_items, object_hook=json_object_hook)
        # check for item result, else normal cw
        if item.datasource_widget is None:
            cfg = self.cwc.load_cube_config(item)
        else:
            cfg = self.cwc.get_state(item.datasource_widget)[0]

        requires = []
        for i in cfg['dimensions']:
            if i.has_key('fancy_select_module') and i['fancy_select_module'] not in requires:
                requires.append(i['fancy_select_module'])

        state = {
            "objectName": u'ReportItemDataSourceWidget',
            "loaderName": 'configurable_report_item',
            "requires": requires,
            "dbID": item.id,
            'datasource_widget_id': None if item.datasource_widget is None else item.datasource_widget.id,
            'title': item.title,
            'showTitle': item.show_title,
            'config': cfg,
            'appliedConfig': simplejson.loads(item.applied_config, object_hook=json_object_hook),
            'graphConfig': gc,
            'tableConfig': tc,
            'linkedItems': li,
        }

        return state

    def canvas_post_save(self, item):
        if len(item.temp_linked_items) == 0:
            item.linked_items = None
            self.sa_session.add(item)
            return

        links = {}

        for db_id, axis in item.temp_linked_items.iteritems():
            if unicode(db_id).startswith('yui-'):
                db_id = AbstractStateController.save_state_id_map[db_id]

            links[db_id] = axis

        item.linked_items = simplejson.dumps(links, ensure_ascii=False, cls=JsonRpcJSONEncoder)
        self.sa_session.add(item)
        return

class ItemDynamicTableController(ItemStateController):
    """
    Controller for the QAItem
    """

    _instance_class = iqreport_models.ReportItemDynamicTable

    def __init__(self, session, manager):
        ItemStateController.__init__(self, session, manager)

    @transaction_decorator
    def save(self, column, state, position):
        item = ItemStateController._instance_from_state(self, state)

        item.data = simplejson.dumps(state['data'], ensure_ascii=False, cls=JsonRpcJSONEncoder)
        item.title = state['title']
        item.show_title = state['showTitle']
        item.show_table = state['showTable']
        item.show_graph = state['showGraph']

        if len(state['graphConfig']) > 0:
            item.graph_config = simplejson.dumps(state['graphConfig'], ensure_ascii=False, cls=JsonRpcJSONEncoder)

        ItemStateController.save(self, item, column, state, position)

        return item

    def load(self, id):
        item = self._instance_from_id(id)
        gc = {}

        if item.graph_config is not None:
            gc = simplejson.loads(item.graph_config, object_hook=json_object_hook)

        state = {
            "objectName": u'ReportItemDynamicTable',
            "loaderName": 'dynamic_table_report_item',
            "data": simplejson.loads(item.data, object_hook=json_object_hook),
            'title': item.title,
            'showTitle': item.show_title,
            'graphConfig': gc,
            'showTable': item.show_table,
            'showGraph': item.show_graph,
            "dbID": item.id,
        }

        return state

item_controller_mapping = {
    'ReportItemHtml': ItemHtmlStateController,
    'ReportItemDynamicTable': ItemDynamicTableController,
    'ReportItemDataSourceWidget': ItemDataSourceWidgetController,
}
