# Copyright (c) 2009 BaseSoft B.V.
# See LICENSE for details.

from django.utils import simplejson

from basesoft.web.core.jsonrpc.dispatcher import json_object_hook, JsonRpcJSONEncoder
from basesoft.web.core.util.controllers import AbstractController, transaction_decorator, StateManager
from basesoft.web.core.exceptions import PublicUserException
from basesoft.web.contrib.dashboard.controllers import DashboardWidgetStateController
from basesoft.web.contrib.compose.controllers import PublicationStateController

from iqportal.iqreport.core.models import IQReportDashboardDynamicTable, IQReportGraphPreset,\
    ReportItemDataSourceWidget
from iqportal.iqreport.core.models import ReportPublication
from iqportal.iqreport.core.models import Report
from iqportal.iqreport.compose.controllers import ReportStateController
from iqportal.iqube.core.models import DataSourceWidget

from basesoft import logging
from iqportal.iqube.core.controllers import AbstractResultController
logger = logging.getLogger('iqportal.iqreport.core')

class DashboardDynamicTableController(DashboardWidgetStateController):
    """
    Controller for the QAItem
    """

    _instance_class = IQReportDashboardDynamicTable

    def __init__(self, session, manager):
        DashboardWidgetStateController.__init__(self, session, manager)

    @transaction_decorator
    def save(self, column, state, position):
        widget = DashboardWidgetStateController._instance_from_state(self, state)

        widget.data = simplejson.dumps(state['data'], ensure_ascii=False, cls=JsonRpcJSONEncoder)

        widget.show_table = state['showTable']
        widget.show_graph = state['showGraph']

        if len(state['graphConfig']) > 0:
            widget.graph_config = simplejson.dumps(state['graphConfig'], ensure_ascii=False, cls=JsonRpcJSONEncoder)
            # TODO: verify graph config


        DashboardWidgetStateController.save(self, column, state, position, widget)

        return widget

    def load(self, id):
        widget = self._instance_from_id(id)
        gc = {}

        if widget.graph_config is not None:
            gc = simplejson.loads(widget.graph_config, object_hook=json_object_hook)

        state = {
            "objectName": u'DynamicTableWidget',
            "loaderName": 'db_dynamic_table',
            "data": simplejson.loads(widget.data, object_hook=json_object_hook),
            'graphConfig': gc,
            'showTable': widget.show_table,
            'showGraph': widget.show_graph,
            'title': widget.title,
            "dbID": widget.id,
        }

        return state

class ReportPublicationStateController(PublicationStateController):

    _instance_class = ReportPublication

    @transaction_decorator
    def set_publication(self, state, group_id=None):

        self._assert_state(state, 'report_id', (long, int))

        if state['dbID'] == None:
            self.assert_permissions('create_report_publication', group_id=group_id)

            # if the publication is directly set to scheduled on creation, need to check for that perm also
            if state['status'] == ReportPublication.STATUS_SCHEDULED:
                self.assert_permissions('schedule_report_publication')
        else:
            self.assert_permissions('update_report_publication')

        self.assert_permissions('view_report')
        # query report based on id
        self.report = self.sa_session.query(Report).get(state['report_id'])
        self.assert_privileges_on_object(self.report, Report.PRIV_VIEW)

        publication = PublicationStateController.set_publication(self, state, group_id)

        if state['dbID'] is not None:
            # scheduling a publication (setting status from NEW to SCHEDULED) requires
            # 'schedule_report_publication' perm (privs are checked in parent)
            if (publication.status == ReportPublication.STATUS_NEW and \
                state['status'] == ReportPublication.STATUS_SCHEDULED):
                self.assert_permissions('schedule_report_publication')

        publication.report = self.report
        # TODO: temporary solution for using the publishers account id as role :S #3237
        publication.cache_role_name = str(self.get_account().id)
        return publication

    @transaction_decorator
    def save(self, state, group_id=None):

        publication = self.set_publication(state, group_id)

        if state['dbID'] is None:
            self.save_default_privilege_object_group(publication, target_group=group_id)

        self.sa_session.add(self.report)
        self.sa_session.flush()

        # expire report so it will update the publications
        self.sa_session.expire(self.report)
        return publication

    def load(self, id):
        """
        Loads and returns the state for the given publication_id
        """

        self.assert_permissions('view_report')
        self.assert_permissions('view_report_publication')

        publication = self._instance_from_id(id)

        self.assert_privileges_on_object(publication.report, Report.PRIV_VIEW)
        self.assert_privileges_on_object(publication, ReportPublication.PRIV_VIEW)

        state = PublicationStateController.load(self, publication)
        state["report_id"] = publication.report_id

        self.publication = publication

        return state

    @transaction_decorator
    def reset_cache(self, *ids):
        self.assert_permissions('update_report_publication')

        status = [ReportPublication.STATUS_NOT_CACHED, ReportPublication.STATUS_CACHE_REPROCESS]

        publications = self.get_privileged_object_by_id(ReportPublication, [id for id in ids], priv=ReportPublication.PRIV_UPDATE)
        rv = []
        for p in publications:
            if p.cache_status in status:
                logger.info('reset_cache: not recaching publication: %s' % p.id)
                continue
            p.cache_status = ReportPublication.STATUS_CACHE_REPROCESS
            self.sa_session.add(p)
            logger.info('reset_cache: recached set for publication: %s' % p.id)
            rv.append(p.id)

        self.sa_session.flush()
        return rv

class PublicationCacheController(AbstractController):

    def generate_cache(self, loop_validator=None):
        """
        This will (re)generate the cache for all Stale report publications
        """

        status = [ReportPublication.STATUS_NOT_CACHED, ReportPublication.STATUS_CACHE_REPROCESS]
        publications = self.sa_session.query(ReportPublication).filter(ReportPublication.cache_status.in_(status)).all()

        for p in publications:
            if p.cache_status not in status:
                continue

            new_status = ReportPublication.STATUS_CACHE_PROCESSING
            if p.cache_status == ReportPublication.STATUS_CACHE_REPROCESS:
                new_status = ReportPublication.STATUS_CACHE_REPROCESSING
            # set the status to processing so no other task picks this up
            p.cache_status = new_status
            self.sa_session.add(p)
            self.sa_session.flush()

            self._cache_publication(p)
            p.cache_status = ReportPublication.STATUS_CACHE_PROCESSED

            self.sa_session.add(p)
            self.sa_session.flush()

            if loop_validator is not None:
                loop_validator()

    def _get_cachable_items_from_state(self, state):
        cachable_items = ['ReportItemDataSourceWidget']
        items = []
        r = state['report']
        for b in r['children']['blocks']:
            for c in b['children']['columns']:
                for i in c['children']['items']:
                    if i['objectName'] in cachable_items:
                        items.append(i)
        return items

    def _cache_publication(self, publication):

        manager = StateManager(self.locale, None, Report.MODE_VIEW)
        controller = ReportStateController(self.sa_session, manager)
        state = controller.load_and_save_state_by_publication(publication)

        items = self._get_cachable_items_from_state(state)

        # if (re-)caching a publication, clear it first

        # for each invitee, we trigger the cache, it could be individual response
        for invitee in publication.invitees:
            for item in items:
                # some items might unset the contact from the contexts, so refill it here.
                context = {'publication': publication, 'contact': invitee.contact, 'force_cache': False, 'force_recache': True}
                if item['objectName'] == 'ReportItemDataSourceWidget':
                    self._cache_configurable_widget(item, publication, context)

    def _cache_configurable_widget(self, item, publication, context):
        from copy import deepcopy

        # get the applied config from the item
        applied_config = deepcopy(item['appliedConfig'])

        if item.has_key('datasource_widget_id') and item['datasource_widget_id'] not in ['', None]:
            widget = self.sa_session.query(DataSourceWidget).filter(DataSourceWidget.id==item['datasource_widget_id']).one()
            datasource = widget.datasource
            context = {'context': context, 'run_mode': 1}
        else:
            widget = self.sa_session.query(ReportItemDataSourceWidget).filter(ReportItemDataSourceWidget.item_id == item['dbID']).one()
            datasource = widget.datasource_cube.datasource
            context = {}

        abstract_result_controller = AbstractResultController(self.sa_session, account_id=self.account_id)
        try:
            abstract_result_controller.cache_widget(datasource, item['config'], applied_config, publication.cache_role_name, context)
        except PublicUserException, e:
            logger.info('PublicUserException while generating cache for ConfigurableWidgetItem, %s' % e.message)

class GraphPresetController(AbstractController):
    """
    Graph presets
    """
    def get(self, id=None, graph_lib=None, priv=IQReportGraphPreset.PRIV_VIEW):

        where = None
        if id is not None:
            if type(id) != list:
                id = [id]
            where = IQReportGraphPreset.id.in_(id)

        if graph_lib is not None:
            if type(graph_lib) != list:
                graph_lib = [graph_lib]
            s = IQReportGraphPreset.graph_lib.in_(graph_lib)
            if where is None:
                where = s
            else:
                where = and_(where, s)

        gps = self._get_privileged_object(IQReportGraphPreset, where, priv=priv)
        rv = []
        for gp in gps:
            gp.config = simplejson.loads(gp.config, object_hook=json_object_hook)
            rv.append(gp)
        return rv

    def save_or_update(self, id, name, graph_lib, config):

        if id is None:
            self.assert_permissions('create_graph_preset')
            gp = IQReportGraphPreset()
        else:
            self.assert_permissions('update_graph_preset')
            gp = self.get(id, priv=IQReportGraphPreset.PRIV_UPDATE).pop()

        gp.name = name
        gp.graph_lib = graph_lib
        gp.config = simplejson.dumps(config, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        if id is None:
            self.save_default_privilege_object_group(gp)

        self.sa_session.add(gp)
        return gp

