# Copyright (c) 2009 BaseSoft B.V.
# See LICENSE for details.

from sqlalchemy.sql import and_
from django.utils.translation import ugettext, ugettext_lazy as _
from sqlalchemy.orm.session import Session

from basesoft.web.core.models import PrivilegeObject, ComposePrivilegeObject
from basesoft.web.contrib.dashboard.models import DashboardWidget
from basesoft.web.contrib.compose.models import ComposePublication, ComposeInvitee, ComposeMail

class IQReportMail(ComposeMail):
    """
    Model for a repotr mail, this inherits from the mailing app Mail
    @DynamicAttrs
    """

    NAME_I18N = _("IQReport e-mail")

    def __repr__(self):
        return '<IQReportMail id=%s>' % self.id

    @staticmethod
    def replacement_fields():
        x = ComposeMail.replacement_fields()
        x.update({
            ugettext('Report'): {
                'iqreport.report.link': ugettext('Report link'),
                'iqreport.report.name': ugettext('Report name'),
                'iqreport.report.publication_label': ugettext('Publication label')
            }
        })
        return x

    def _replace_field(self, person, field_name, context):
        """
        Method called to replace every given replacement field with the current context. here we deal with specific
        IQuest replacement fields.. contact stuff is handled upstream.
        """
        from sqlalchemy.orm.exc import NoResultFound

        session = Session.object_session(self)

        # context is set in the MailingTask
        if context.has_key('publication'):
            publication = context['publication']
        else:
            assert context.has_key('mailing')
            publication = context['mailing'].publication
        contact = context['contact']

        if field_name == 'iqreport.report.name':
            return publication.report.name

        if field_name == 'iqreport.report.publication_label':
            return publication.label

        # Process the link to survey.. this depends if the survey is anonymous or not
        # if the survey hasn't been published we just throw an exception which should abort the task (untested)
        if field_name == 'iqreport.report.link':

            # first we check if we can find the contact as part of the publication.
            try:
                invitee = session.query(ComposeInvitee).filter(
                    and_(ComposeInvitee.contact==contact,
                         ComposeInvitee.publication==publication)).one()
                invitee_hash = invitee.invitee_hash
            except NoResultFound:
                # apparently this contact was not included as a non-anonymous member of the publication
                # Lets see if this publication also is anonymous.
                if publication.anonymous:
                    try:
                        invitee_hash = publication.invitees[0].invitee_hash
                    except IndexError:
                        raise Exception('Publication %s, Contact: %s: Cannot setup invite email because report hasnt been published yet.' %
                                        (publication.id, contact.id) )
                else:
                    raise Exception('Publication %s, Contact: %s: This contact is not part of this non-anonymous publication.' %
                                    (publication.id, contact.id) )

            # return the invite hash url in a <a>
            link = self._base_url(context) + 'iqportal.iqreport.display/%s/' % invitee_hash
            return '<a href="%s">%s</a>' % (link, link)

        return ComposeMail._replace_field(self, person, field_name, context)

class Report(ComposePrivilegeObject):
    """
    Model for the Report
    @DynamicAttrs
    """

    MODE_VIEW = ComposePrivilegeObject.MAX_MODE_BIT*2

    mark_as_deleted_permission = 'delete_report'
    mark_as_archived_permission = 'update_report'

    _name_property = 'name'

    def get_group_object(self):
        return (ReportGroup, 'report')

    def __get_items(self):
        """
        Easy access to all the ItemQA's in the survey when you dont care in which block/column they are.
        """
        session = Session.object_session(self)
        return session.query(ReportItem) \
            .join(ReportBlockColumn) \
            .join(ReportBlock) \
            .join(Report) \
            .filter(Report.id==self.id) \
            .all()

    items = property(__get_items)

    def __repr__(self):
        return '<Report id=%s>' % self.id

class ReportGroup(object):
    """
    Model representing the link between reports and groups.
    @DynamicAttrs
    """

class ReportLocale(object):
    """
    Model for the Report Locale, this will include the title of the report
    per Locale.
    @DynamicAttrs
    """

class ReportBlock(object):
    """
    Model for a Block within a report.
    A block can contain items of various types.
    @DynamicAttrs
    """

    def __repr__(self):
        return '<ReportBlock id=%s>' % self.id

class ReportBlockColumn(object):
    """
    Block contains Columns
    @DynamicAttrs
    """
    def _get_items(self):
        """
        This creates a property .items on the ReportBlock object. When .items
        is requested the query below is executed and returns a list of all (*)
        extensions of ReportItem.
        @see http://www.sqlalchemy.org/docs/04/mappers.html#advdatamapping_mapper_inheritance
        """
        session = Session.object_session(self)
        return session.query(ReportItem).with_polymorphic('*').filter_by(column_id=self.id).order_by('position').all()

    items = property(_get_items)

class ReportItem(object):
    """
    Model for the Report Item.
    @DynamicAttrs
    """

    def __repr__(self):
        return '<ReportItem id=%s>' % self.id

class ReportItemHtml(ReportItem):
    """
    @DynamicAttrs
    """

    def __repr__(self):
        return '<ReportItemHtml id=%s>' % self.id

class ReportItemHtmlLocale(object):
    """
    @DynamicAttrs
    """

class ReportItemDynamicTable(ReportItem):
    pass

class ReportItemDataSourceWidget(ReportItem):
    pass

class IQReportGraphType(object):
    """
    @DynamicAttrs
    """
    pass

class IQReportGraphTypeDatasource(object):
    """
    @DynamicAttrs
    """
    pass

class ReportTheme(PrivilegeObject):
    """
    @DynamicAttrs
    """

    _name_property = 'name'

    def get_group_object(self):
        return (ReportThemeGroup, 'report_theme')

    def __repr__(self):
        return '<ReportTheme name=%s>' % self.name

class ReportThemeGroup(object):
    """
    @DynamicAttrs
    """
    pass

class IQReportDashboardDynamicTable(DashboardWidget):
    """
    @DynamicAttrs
    """
    def __repr__(self):
        return '<IQReportDashboardDynamicTable widget_id=%s>' % self.widget_id

class ReportPublication(ComposePublication):
    """
    Model representing periods in which the report is available.
    @DynamicAttrs
    """

    STATUS_NOT_CACHED = 0
    STATUS_CACHE_PROCESSING = 1
    STATUS_CACHE_PROCESSED = 2
    STATUS_CACHE_REPROCESS = 3
    STATUS_CACHE_REPROCESSING = 4

    NAME_I18N = _('IQReport report publication')

    permissions = ('use_iqreport_report_publication', )
    mark_as_deleted_permission = 'delete_report_publication'
    mark_as_archived_permission = 'update_report_publication'

    _name_property = 'label'
    js_module =  'iqreport_publication'

    def get_controller_instance(self):
        from iqportal.iqreport.core.controllers import ReportPublicationStateController
        return ReportPublicationStateController

    def filter_contact(self, mailing, contact):
        return False

    def approved_for_sending(self):
        return (self.cache_status == self.STATUS_CACHE_PROCESSED)

    def __repr__(self):
        return '<ReportPublication id=%s>' % (self.id)

class IQReportGraphPreset(PrivilegeObject):
    """
    @DynamicAttrs
    """

    _name_property = 'name'

    mark_as_deleted_permission = 'delete_graph_preset'
    mark_as_archived_permission = 'archive_graph_preset'

    def get_group_object(self):
        return (IQReportGraphPresetGroup, 'graph_preset')

    def __repr__(self):
        return '<GraphPreset name=%s>' % self.name

class IQReportGraphPresetGroup(object):
    """
    @DynamicAttrs
    """
    pass
