# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.
from sqlalchemy import and_

from simplejson.decoder import JSONDecodeError

from django.utils.translation import ugettext as _
from django.utils import simplejson

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from basesoft.web.core.exceptions import PublicSecurityException, PublicUserException
from basesoft.web.core.jsonrpc.dispatcher import JsonRpcJSONEncoder, json_object_hook

from iqportal.iqube.core.models import MdxCalculatedMember, MdxMethod, MdxDataSourceCalculatedMember, DataSource
from iqportal.iqube.core.controllers import DataSourceController, MdxMethodController, MondrianMemberSerialize
from iqportal.iqube.compose.models import MdxFormatString

class MdxCalculatedMemberController(AbstractController):
    """
    This controller takes care of the business logic of calulated members

    Note that this is a merge of two previously separated objects:
     * DataSourceCalcualtedMember
     * MdxSavedMethod
    """
    def _assert_mdx_method(self, _id):
        """
        Verifies that the user has access to the used MdxMethod
        """
        m = self.get_privileged_object_by_id(MdxMethod, _id, priv=MdxMethod.PRIV_VIEW)
        if len(m) == 0:
            raise PublicSecurityException(_('No access to provided MDX Method'))
        return True

    def _assert_calculated_member(self, _id):
        m = self.get_privileged_object_by_id(MdxCalculatedMember, _id, priv=MdxMethod.PRIV_VIEW)
        if len(m) == 0:
            raise PublicSecurityException(_('No access to provided MDX Calculated Member'))
        return True

    def _get_mdx_method(self, _id):
        """
        Used to parse the Mdx string from the state.
        """
        return self.sa_session.query(MdxMethod).filter(MdxMethod.id==_id).first()

    def _validate_repeaters(self, state, repeaters, parent=None, repeater=None):
        if not state.has_key('inputs'):
            return

        c = None if parent is None else repeaters['children'][parent]

        if state.has_key('repeaterID'):
            repeater = state['repeaterID']

        def check_parent(parents, repeater, repeaters):
            if repeater in parents:
                raise PublicUserException('Cannot connect a repeater to a repeater that is connected to it.')

            for p in parents:
                if not repeaters['parents'].has_key(p):
                    continue
                check_parent(repeaters['parents'][p], repeater, repeaters)

        for input in state['inputs']:
            if input is None:
                continue
            if not input.has_key('repeater'):
                self._validate_repeaters(input, repeaters, parent, repeater)
                continue
            if input['repeater'] == repeater:
                raise PublicUserException(_('Cannot link repeater to itself.'))

            if c is not None:
                c.append(input['repeater'])

            if repeaters['parents'].has_key(repeater):
                check_parent(repeaters['parents'][repeater], input['repeater'], repeaters)

            if parent is not None and not repeaters['parents'].has_key(input['repeater']):
                repeaters['parents'][input['repeater']] = [parent]


    def _validate_state(self, state, used_obj=None):
        """
        Validates the state and the access to the used MdxMethods
        """
        if state is None:
            return True

        if type(state) != dict:
            return True

        if used_obj is None:
            used_obj = []


        #if state of repeater has repeater in used_cms exception
        if state.has_key('repeaters'):
            repeaters = {'parents': {}, 'children': {}}

            for repeater in state['repeaters']:
                repeaters['children'][repeater['repeaterID']] = []
                self._validate_repeaters(repeater, repeaters, repeater['repeaterID'])

            ordered_repeaters = []
            for repeater, references in repeaters['children'].iteritems():
                for r in references:
                    if r in ordered_repeaters:
                        continue
                    ordered_repeaters.insert(0, r)
                ordered_repeaters.append(repeater)
            new_repeaters = [None for i in state['repeaters']]
            for repeater in state['repeaters']:
                new_repeaters[ordered_repeaters.index(repeater['repeaterID'])] = repeater

            state['repeaters'] = new_repeaters

        if state.has_key('calculated_member_id'):
            if state['calculated_member_id'] in used_obj:
                raise PublicUserException(_('You cannot use calculated members recursively.'))
            used_obj.append(state['calculated_member_id'])
            self._assert_calculated_member(state['calculated_member_id'])
            nested = self.load_state(state['calculated_member_id'])
            self._validate_state(nested['children'], used_obj)

        if state.has_key('method_id'):
            self._assert_mdx_method(state['method_id'])

        if state.has_key('inputs'):
            for i in state['inputs']:
                self._validate_state(i, used_obj)

        return True

    def _get_input_count(self, state, cnt=None):
        """
        Returns the number of open inputs in the formula.
        """
        if cnt is None:
            cnt = 0

        if state is None:
            cnt += 1
            return cnt

        if type(state) != dict:
            return cnt

        if state.has_key('repeaters'):
            for i in state['repeaters']:
                cnt = self._get_input_count(i, cnt)

        if state.has_key('inputs'):
            for i in state['inputs']:
                cnt = self._get_input_count(i, cnt)
        return cnt

    def _get_calculated_member_type(self, state):
        # FIXME backward compat with MdxSavedMethod
        if state['return_type'] == 'set':
            return MdxCalculatedMember.NAMED_SET

        return MdxCalculatedMember.CALCULATED_MEMBER

    def get_calculated_member(self, _id=None, priv=MdxCalculatedMember.PRIV_VIEW, order_by=None):
        """
        Wrapper method to retrieve a MdxCalculatedMember from the DB including privileges check.
        """
        self.assert_permissions('view_calculated_member')
        return self.get_privileged_object_by_id(MdxCalculatedMember, _id, priv=priv, order_by=order_by)

    def get_calculated_member_by_type(self, _type=MdxCalculatedMember.CALCULATED_MEMBER, dimension=None):
        _where = MdxCalculatedMember.type==_type
        return self._get_privileged_object(MdxCalculatedMember, _where, priv=MdxCalculatedMember.PRIV_VIEW, order_by=MdxCalculatedMember.label)

    def get_calculated_members_for(self, datasource_id, cube, _type=MdxCalculatedMember.CALCULATED_MEMBER, dimension=None):
        """
        Method to retrieve all template CMs and datasource specific ones.
        """

        _where = and_(MdxCalculatedMember.type==_type, MdxCalculatedMember.dimension==dimension)

        # first we get only template cms
        rv = self._get_privileged_object(MdxCalculatedMember,
                                         and_(_where, MdxCalculatedMember.polymorphic_type=='CalculatedMember'),
                                         priv=MdxCalculatedMember.PRIV_VIEW,
                                         order_by=MdxCalculatedMember.label)

        # now we get datasource specific calculations
        if not isinstance(datasource_id, DataSource):
            dsc = DataSourceController(self.sa_session, account_id=self.get_account().id)
            datasource_id = dsc.get_datasource(datasource_id)

        _where = and_(_where, MdxDataSourceCalculatedMember.datasource==datasource_id, MdxDataSourceCalculatedMember.cube==cube)

        rv += self._get_privileged_object(MdxDataSourceCalculatedMember,
                                          and_(_where, MdxCalculatedMember.polymorphic_type=='DataSourceCalculatedMember'),
                                          priv=MdxDataSourceCalculatedMember.PRIV_VIEW,
                                          order_by=MdxDataSourceCalculatedMember.label)

        return rv

    @transaction_decorator
    def save_or_update_calculated_member(self, _id, label, state, datasource_id=None, cube=None, constraints=None, dimension=None, format_string=None):
        """
        Saves or updates a MdxCalulcatedMember, this method determines whether the supplied formula is a Calculated Member or a Named Set.
        """
        assert isinstance(state, dict)

        polymorphic_type = MdxCalculatedMember
        if datasource_id is not None:
            polymorphic_type = MdxDataSourceCalculatedMember

        used_cms = []
        if _id is None:
            self.assert_permissions('create_calculated_member')
            mcm = polymorphic_type()
        else:
            self.assert_permissions('update_calculated_member')
            mcm = self.get_calculated_member(_id, priv=MdxCalculatedMember.PRIV_UPDATE).pop()
            assert isinstance(mcm, polymorphic_type) # we can not change polymorphic types
            used_cms.append(_id)

        self._validate_state(state, used_cms)

        if datasource_id is not None:
            dsc = DataSourceController(self.sa_session, account_id=self.get_account().id)
            mcm.datasource = dsc.get_datasource(datasource_id)
            mcm.cube = cube
            mcm.constraints = None if constraints is None else simplejson.dumps(constraints, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        mcm.label = label
        mcm.dimension = dimension
        mcm.type = self._get_calculated_member_type(state)
        mcm.formula = simplejson.dumps(state, ensure_ascii=False, cls=JsonRpcJSONEncoder)
        mcm.input_count = self._get_input_count(state)
        if(mcm.type!=MdxCalculatedMember.CALCULATED_MEMBER):
            format_string = None

        mcm.format_string = format_string

        if _id is None:
            self.save_default_privilege_object_group(mcm)

        self.sa_session.add(mcm)

        return mcm

    def get_parsed(self, _id, inputs=None, include_type=False, is_top=True):
        """
        This returns a Mdx string, if inputs are supplied it will apply those to the open inputs.
        """

        mcm = self.get_calculated_member(_id).pop()

        if inputs is not None:
            if not is_top:
                new_inputs = []
                for i in range(mcm.input_count):
                    new_inputs.append(inputs.pop(0))
                inputs = new_inputs
            assert len(inputs) == mcm.input_count

        state = self.load_state(mcm)

        if not include_type:
            return self._parse_from_state(state['children'], inputs)

        return self._parse_from_state(state['children'], inputs), mcm.type, mcm.dimension

    def _parse_from_state(self, state, inputs, usable_repeaters=None):
        """
        This parse part of the state and replaces the inputs accordingly.
        """
        if usable_repeaters is None:
            usable_repeaters = {}
        if state is None:
            if type(inputs) == list and len(inputs) is 0:
                return None
            if inputs is not None:
                rv = inputs.pop(0)
                if type(rv) == list:
                    parsed = [simplejson.loads(x) for x in rv]
                    if len(rv) == 1:
                        rv = '%s' % ','.join([x['member_unique_name'] for x in parsed])
                    else:
                        rv = '{%s}' % ','.join([x['member_unique_name'] for x in parsed])
                return rv
            return '$var'

        if type(state) != dict:
            return state

        add = []
        repeaters = []
        if state.has_key('repeaters'):
            for i in state['repeaters']:
                if i['return_type'] != 'set':
                    repeater_name = '[Measures].[%s]' % i['repeaterID']
                else:
                    repeater_name = '[%s]' % i['repeaterID']

                usable_repeaters[i['repeaterID']] = repeater_name
                repeaters.append((repeater_name, self._parse_from_state(i, inputs, usable_repeaters)))

        if state.has_key('inputs'):
            for i in state['inputs']:
                string = self._parse_from_state(i, inputs, usable_repeaters)
                if string is None:
                    continue
                elif type(string) == list:
                    repeaters += string[:-1]
                    string = string.pop()
                add.append(string)

        if state.has_key('repeater'):
            return usable_repeaters[state['repeater']]


        if state.has_key('calculated_member_id'):
            return self.get_parsed(state['calculated_member_id'], add, is_top=False)

        if state.has_key('operator'):
            assert len(add) == 2
            rv = '(%s)' % state['operator'].join(add)
        elif state.has_key('comparator'):
            assert len(add) == 2
            rv = '(%s)' % state['comparator'].join(add)
        elif not state.has_key('is_property'):
            rv = ','.join([str(x) for x in add])

        if state.has_key('method_id'):
            m = self._get_mdx_method(state['method_id'])

            required_count = len([f for f in m.fields if f.required_field])

            assert required_count <= len(add) <= len(m.fields)
            if len(rv) == 0 and required_count == 0 and state.has_key('value'):
                rv = '%(method)s(%(inputs)s)' % {'method': m.name, 'inputs': state['value']}
            else:
                rv = '%(method)s(%(inputs)s)' % {'method': m.name, 'inputs': rv}

            if m.name == 'Axis': # This is a workaround as Mondrian does not have axis implemented.
                name = '[Axis%s]' % state['value']
                repeaters.append((name, '%s' % rv))
                rv = name

        elif state.has_key('is_tuple'):
            rv = '(%s)' % rv

        elif state.has_key('is_property'):
            rv = '%s.%s' % (add.pop(0), str(state['value']))
            if len(add) > 0:
                rv = '%s(%s)' % (rv, ','.join([str(x) for x in add]))
        elif state.has_key('value'):
            if state['return_type'] == 'string':
                rv += "'%s'" % str(state['value'])
            else:
                rv += str(state['value'])

        if state.has_key('selection'):
            rv += self._parse_selection(state)

        if len(repeaters) != 0:
            repeaters.append(rv)
            return repeaters

        return rv

    def _parse_selection(self, state):
        if state['return_type'] == 'set':
            parsed = [simplejson.loads(x) for x in state['selection']]
            return '{%s}' % ','.join([x['member_unique_name'] for x in parsed])
        else:
            if type(state['selection']) == list:
                assert len(state['selection']) == 1
                return simplejson.loads(state['selection'][0])['member_unique_name']
            if type(state['selection']) in (str, unicode, ):
                try:
                    return simplejson.loads(state['selection'])['member_unique_name']
                except JSONDecodeError:
                    return state['selection']

        return ''

    parse_from_state = _parse_from_state

    def load_state(self, _id, with_methods=False, with_calculated_members=False):
        """
        Loads the state (formula) from the stored JSON string for the supplied MdxCalculatedMember
        """
        if isinstance(_id, MdxCalculatedMember):
            mcm = _id
        else:
            mcm = self.get_calculated_member(_id)
            if len(mcm) == 0:
                raise PublicSecurityException(_('You do not have permission to access the requested resource'))
            mcm = mcm.pop()


        rv = {
            'name': mcm.label,
            'dbID': mcm.id,
            'input_count': mcm.input_count,
            'type': mcm.type,
            'dimension': mcm.dimension,
            'format_string': mcm.format_string,
            'children': simplejson.loads(mcm.formula, object_hook=json_object_hook)
        }
        if with_methods:

            def _append_method(child):
                if type(child) != dict:
                    return
                if child.has_key('method_id'):
                    child['method'] = self._get_mdx_method_state(child['method_id'])
                if child.has_key('inputs'):
                    for i in child['inputs']:
                        if i is not None:
                            _append_method(i)

            if rv['children'].has_key('inputs'):
                _append_method(rv['children'])

        if with_calculated_members:
            def _append_cm(child):
                if child.has_key('calculated_member_id'):
                    return self.load_state(child['calculated_member_id'], with_methods, True)['children']
                if child.has_key('inputs'):
                    j = 0
                    for i in child['inputs']:
                        if i is not None:
                            child['inputs'][j] = _append_cm(i) # make it a reference
                        j+=1
                return child

            if rv['children'].has_key('inputs'):
                rv['children'] = _append_cm(rv['children'])

        if isinstance(mcm, MdxDataSourceCalculatedMember):
            rv['datasource_id'] = mcm.datasource.id
            rv['cube'] = mcm.cube
            rv['constraints'] = None if mcm.constraints is None else simplejson.loads(mcm.constraints, object_hook=json_object_hook)

        return rv

    def load_state_from_compose(self, state):

        if not state.has_key('children'):
            raise PublicUserException(_('Cannot preview an undefined Calculated Member'))

        def _append_method(child):
            if type(child) != dict:
                return
            if child.has_key('method_id'):
                child['method'] = self._get_mdx_method_state(child['method_id'])
            if child.has_key('inputs'):
                for i in child['inputs']:
                    if i is not None:
                        _append_method(i)

        if state['children'].has_key('inputs'):
            _append_method(state['children'])

        state['type'] = self._get_calculated_member_type(state['children'])

        return state

    def _get_mdx_method_state(self, _id):
        mmc = MdxMethodController(self.sa_session, account_id=self.get_account().id)
        m = mmc.get_mdx_method(_id).pop()
        return {
            'label': m.label,
            'name': m.name,
            'return_type': m.return_type,
            'category': m.category,
            'fields': [{
                'field_type': f.field_type,
                'required_field': f.required_field
            } for f in m.fields]
        }

    def get_with_statement(self, name, _type, formulas, dimension, settings=None, include_type=True):

        if settings is None:
            settings = {}

        rv = []
        if type(formulas) == list:
            formula = formulas.pop()
            for f in formulas:
                if f[0].find('].[') == -1:
                    rv.append("SET %s AS '%s'" % f)
                else:
                    rv.append("MEMBER %s AS '%s'" % f)
        else:
            formula = formulas

        format_string = None if not settings.has_key('format_string') else settings['format_string']

        if settings.has_key('label') and settings['label']:
            name = '[%s]' % settings['label']
        if _type == MdxCalculatedMember.CALCULATED_MEMBER:
            dimension = '[%s]' % ('Measures' if dimension is None else dimension)
            name = '%s.%s' % (dimension, name)
            _with = '%s AS %s' % (name, formula)
            if format_string is not None and format_string != '':
                if format_string[0] == '#':
                    _with = "%s, FORMAT_STRING='%s'" % (_with, format_string.replace('{value}', name))
                else:
                    _with = "%s, FORMAT_STRING=%s" % (_with, format_string.replace('{value}', name))
            if include_type:
                _with = 'MEMBER %s' % _with
        else:
            dimension = 'FORMULA'
            _with = '%s AS %s' % (name, formula)
            if include_type:
                _with = 'SET %s' % _with

        rv.append(_with)

        return name, rv, dimension

    def get_preview_result(self, datasource_id, cube, state, replacements, settings=None):

        # retrieve the correct MDX formula
        self._validate_state(state['children'])
        formula = self._parse_from_state(state['children'], replacements)

        use_columns = False
        _type = MdxCalculatedMember.CALCULATED_MEMBER

        name = '[Preview]'
        if state['children']['return_type'] != 'set':
            use_columns = True
        else:
            _type = MdxCalculatedMember.NAMED_SET

        use_slicers = False
        name, _with, dimension = self.get_with_statement(name, _type, formula, dimension=state['dimension'], settings=settings)
        if use_columns and dimension != '[Measures]':
            use_slicers = True

        rows = [{'hierarchy_unique_name': dimension, 'member_unique_name': name}]

        if use_columns:
            columns = [{'hierarchy_unique_name': dimension, 'member_unique_name': name}]
            rows = []

        dsc = DataSourceController(self.sa_session, account_id=self.account_id)
        ds = dsc.get_datasource(datasource_id)

        client = ds.connect(self.sa_session, self.get_account().id)
        measure = None if settings is None else settings['measure']


        if measure is None:
            measures = client.get_measures(cube)
            measure = client.get_member_tree_op(cube, measures[0]['measure_unique_name'], client.MDTREEOP_SELF)[0]
        else:
            measure = simplejson.loads(measure)

        slicers = None
        if not use_columns:
            columns = [measure]
        elif use_slicers:
            slicers = [measure]

        mdx = client.build_select_statement(columns=columns, rows=rows, cube=cube, slicers=slicers, nonempty=False, mdx_with=_with, allow_empty_rows=True)
        return client.execute_mdx(mdx).get_dict(), mdx

    def _state_from_mdx(self, clc, children, x, y, repeaters, client=None, cube=None):
        """
        Recursive method which builds a valid state to be used in the Calculated Member designer from a provided MDX string.
        TODO: time members, levels
        TODO: comparators
        """

        # this means we are dealing with multiple selected values
        if type(clc) == list:

            # if we do not have a client instance, this means we cannot lookup any members
            if client is None:
                return [None for i in clc]

            members = []
            dimension = None
            for unique_name in clc:
                # load the members from Mondrian
                d = client.get_member_tree_op(cube, unique_name['description'], client.MDTREEOP_SELF)[0]
                if dimension is None:
                    dimension = d['level_unique_name']
                members.append(MondrianMemberSerialize(d))

            # return a datasource SET expression object -> the list type indicated multiple selections thus a set.
            return {
                'objectName': 'MdxDataSourceExpressionItem',
                'loaderName': 'iqube_datasource_expression_item',
                'return_type': 'set',
                'selection': members,
                'dimension': dimension,
                'type': 'set',
                'posXY': [x, y]
            }

        # we need a dict in order to continue
        if type(clc) != dict:
            return None

        # we are dealing with a numerical calculation -> operator.
        if clc.has_key('operator'):
            rv = {
                'objectName': 'MdxOperatorItem',
                'loaderName': 'iqube_mdx_operator_item',
                'return_type': 'numeric',
                'operator': '%s' % clc['operator'],
                'posXY': [x, y]
            }
            inputs = []
            x = x - 375
            y -= 200
            # each operator always has a left and a right.
            for i in ['left', 'right']:
                x += 250
                inputs.append(self._state_from_mdx(clc[i], children, x, y, repeaters, client, cube))
            rv['inputs'] = inputs
            return rv

        # no description, no idea what to do
        if not clc.has_key('description'):
            return None
        try:
            value = int(clc['description'])
            return {
                'objectName': 'MdxNumericalInputItem',
                'loaderName': 'iqube_mdx_numerical_input_item',
                'return_type': 'numeric',
                'value': value,
                'posXY': [x, y]
            }
        except ValueError:
            pass

        m = self.sa_session.query(MdxMethod).filter(MdxMethod.name==clc['description']).first()
        # it appears as if the description is available as an MDX Method.
        if m is not None:
            self.assert_privileges_on_object(m, MdxMethod.PRIV_VIEW)

            meth = {
                'objectName': 'MdxMethodItem',
                'loaderName': 'iqube_mdx_method_item',
                'return_type': 'set',
                'method_id': m.id,
                'posXY': [x, y],
                'inputs': []
            }

            inputs = []
            # we know how much fields we require, so we can put the "x" coordinates accordingly.
            x = x - (167*len(m.fields))
            y -= 200
            for i in range(len(m.fields)):
                x += 250
                myvar = children.pop(0)
                inputs.append(self._state_from_mdx(myvar, children, x, y, repeaters, client, cube))

            meth['inputs'] = inputs

            return meth

        # we are dealing with a set or a member
        if clc['description'].endswith(']'):
            # if the description matches a previously entered repeater, reference to it.
            if clc['description'] in repeaters:
                return {'repeater': clc['description']}

            # if we do not have a client instance, this means we cannot lookup any members
            if client is None:
                return None

            tokens = clc['description'].split('].[')
            last = tokens.pop()[:-1]
            # if the last value is None, its an bogus member
            if last == 'None':
                return None

            rv = {
                'objectName': 'MdxDataSourceExpressionItem',
                'loaderName': 'iqube_datasource_expression_item',
                'selection': None,
                'dimension': None,
                'posXY': [x, y]
            }

            d = client.get_member_tree_op(cube, clc['description'], client.MDTREEOP_SELF)[0]
            # we are dealing with a measure
            if tokens[0][1:] == 'Measures':
                rv['selection'] = MondrianMemberSerialize(d)
                rv['return_type'] = 'numeric'
                rv['type'] = 'measure'
                return rv

            # FIXME: time member
            rv['dimension'] = d['level_unique_name']
            rv['return_type'] = 'member'
            rv['type'] = 'member'
            rv['selection'] = [MondrianMemberSerialize(d)]

            return rv

        return None

    def state_from_mdx(self, mdx, datasource_id=None, cube=None, x=450, y=650):
        """
        This takes an MDX string and tries to create a Calculated Member state.
        """
        mdx = mdx.strip()
        # first check conditions -> WITH statement required.
        if mdx[:4] != 'WITH':
            raise PublicUserException(_('Invalid MDX provided to parse as a Calculated Member'))

        import re
        # get all the members and sets from the with statement.
        matches = re.findall(r'^(?:(?:MEMBER|SET)\s(.*))+$', mdx[5:], re.MULTILINE)

        client = None
        if datasource_id is not None:
            dsc = DataSourceController(self.sa_session, account_id=self.get_account().id, request=self.request)
            ds = dsc.get_datasource(datasource_id)
            client = ds.connect(self.sa_session, self.get_account().id)

        # use the iqube.xmla.grammar to retrieve the parsed with statements.
        from iqportal.iqube.xmla.mdx import MdxParser
        parser = MdxParser()

        rv = {
            'children': {},
            'repeaters': [],
            'datasource_id': datasource_id,
            'cube': cube
        }

        final = matches.pop()
        repeaters = []
        # if we still have matches then it means we have a repeater
        for mdx in matches:
            # build state from bottom to top.
            myclc = parser.get_dict(mdx)
            clc = myclc['calculation']
            repeaters.append('[%s]' % myclc['member_name'])
            repeater = self._state_from_mdx(clc.pop(0), clc, x, y, repeaters, client, cube)
            repeater['repeaterID'] = myclc['member_name']
            rv['repeaters'].append(repeater)

        # and the actual eventual output.
        clc = parser.get_dict(final)['calculation']
        rv['children'] = self._state_from_mdx(clc.pop(0), clc, x, y, repeaters, client, cube)
        return rv

    def auto_configure_mdx_function(self, calculated_member_id):
        """
        We try to auto configure calculated members, this can be done in case they have no required inputs.
        Used in iqube.core.reporters and iqube.core.controllers
        """
        cm = self.get_calculated_member(calculated_member_id)
        if len(cm) == 0:
            raise PublicSecurityException(_('Incorrect Calculated Member'))

        cm = cm.pop()
        if cm.input_count > 0:
            raise PublicUserException(_('Cannot use unconfigured Calculated Member: %(member)s' % {'member': cm.label}))

        formula_config = {
            'calculated_member_id': cm.id,
            'replacements': [],
            'settings': {
                'label': cm.label
            }
        }
        return formula_config

class MdxFormatStringController(AbstractController):
    def get_format_strings(self, _id=None, priv=MdxFormatString.PRIV_VIEW):
        return self.get_privileged_object_by_id(MdxFormatString, _id, priv)

    def get_format_string(self, _id, priv=MdxFormatString.PRIV_VIEW):
        fmt = self.get_format_strings(_id=_id, priv=priv)
        if len(fmt) == 0:
            raise PublicSecurityException(_('No access to requested format string'))

        return fmt.pop()

    @transaction_decorator
    def save_or_update_format_string(self, _id, name, format_string):
        self.assert_permissions('manage_format_strings')

        # TODO: do we want format string validation here?

        if _id is None:
            fmt = MdxFormatString()
        else:
            fmt = self.get_format_string(_id, priv=MdxFormatString.PRIV_UPDATE)

        fmt.name = name
        fmt.format_string = format_string

        if _id is None:
            self.save_default_privilege_object_group(fmt)

        self.sa_session.add(fmt)
        return fmt

