# Copyright (c) 2014 BaseSoft B.V.
# See LICENSE for details.
from django.utils import simplejson

from basesoft.web.contrib.wiki.parser import Parser

from iqportal.iqube.core.controllers import MdxMethodController, DataSourceController
from iqportal.iqube.compose.controllers import MdxCalculatedMemberController, MdxFormatStringController
from iqportal.iqube.core.models import MdxCalculatedMember
from iqportal.iqube.core import MDX_METHODS, get_properties_map
from basesoft.web.core.exceptions import PublicUserException

class IQubeComposeRPCHandler:
    def __init__(self, request, **kwargs):
        self._request = request
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']

    def get_compose_vars(self):
        """
        called before the manager is instanciated and returns all information that is needed for that..
        like localeLists, themes, last saved survey etc
        """
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        last_id = mcmc.config.get('iqube.compose.calculated.member.id')
        return {
            'lastID': last_id
        }

    def get_mdx_method_example(self, method_id):

        mmc = MdxMethodController(self.sa_session, account_id=self.account_id, request=self._request)
        m = mmc.get_mdx_method(method_id).pop()

        if not MDX_METHODS.has_key(m.name):
            raise PublicUserException(_('Incorrect method supplied'))
        entry = MDX_METHODS[m.name]

        return {
            'label': m.label,
            'name': m.name,
            'return_type': m.return_type,
            'description': Parser().render_to_html(entry.description),
            'example': Parser().render_to_html(entry.example_description),
            'example_mdx': entry.example_mdx,
            'fields': [{'method_id': f.method_id,
                 'field_type': f.field_type,
                 'required_field': f.required_field
                 } for f in m.fields]
        }

    def get_mdx_property_example(self, property_name, category):

        _map = get_properties_map()
        if not _map.has_key(category):
            raise PublicUserException(_('Unsupported property category supplied'))

        p = _map[category][property_name]

        fields = [{
            'field_type': category,
            'required_field': True
        }]
        if p.fields is not None:
            for f in p.fields:
                fields.append({
                    'field_type': f[0],
                    'required_field': f[1]
                })

        return {
            'label': p.label,
            'name': p.name,
            'return_type': p.return_type,
            'description': Parser().render_to_html(p.description),
            'example': Parser().render_to_html(p.example_description),
            'example_mdx': p.example_mdx,
            'fields': fields
        }

    def save(self, state, group_id):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        mcm = mcmc.save_or_update_calculated_member(
            state['dbID'],
            state['name'],
            state['children'],
            datasource_id=state['datasource_id'],
            cube=state['cube'],
            constraints=state['constraints'],
            dimension=state['dimension'],
            format_string=state['format_string'],
        )
        return {state['domID']: mcm.id}

    def load(self, _id, with_methods=False, with_calculated_members=False):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        return mcmc.load_state(_id, with_methods=with_methods, with_calculated_members=with_calculated_members)

    def load_state_from_compose(self, state):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        return mcmc.load_state_from_compose(state)

    def get_parsed_formula(self, state):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        return mcmc.parse_from_state(state, None)

    def get_calculated_members(self, datasource_id=None, cube=None, _type=None, dimension=None):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        rv = []
        if datasource_id is None:
            cms = mcmc.get_calculated_member()
        else:
            cms = mcmc.get_calculated_members_for(datasource_id, cube, _type, dimension)
        for cm in cms:
            rv.append({'label': cm.label, 'value': cm.id})
        return rv

    def get_last_modified_cms(self, limit=5):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        return [{'name': cm.label, 'id': cm.id} for cm in mcmc.get_calculated_member(order_by=MdxCalculatedMember.last_updated.desc())[:limit]]

    def get_named_sets(self, datasource_id=None, cube=None, dimension=None, selected=None):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        rv = []
        for cm in mcmc.get_calculated_members_for(datasource_id, cube, _type=MdxCalculatedMember.NAMED_SET, dimension=dimension):
            rv.append({'label': cm.label, 'value': cm.id})

        if selected is not None:
            for opt in selected:
                unserialized = simplejson.loads(opt)
                rv.append({'label': unserialized['settings']['label'], 'value': opt})

        return rv

    def get_members(self, datasource_id=None, cube=None, dimension=None, selected=None):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        rv = []
        for cm in mcmc.get_calculated_members_for(datasource_id, cube, _type=MdxCalculatedMember.CALCULATED_MEMBER, dimension=dimension):
            rv.append({'label': cm.label, 'value': cm.id})

        if selected is not None:
            for opt in selected:
                unserialized = simplejson.loads(opt)
                rv.append({'label': unserialized['settings']['label'], 'value': opt})

        return rv

    def get_datasource_name(self, datasource_id):
        if datasource_id is None:
            return '-'
        dsc = DataSourceController(self.sa_session, account_id=self.account_id, request=self._request)
        ds = dsc.get_datasource(datasource_id)
        return ds.name

    def get_preview_result(self, datasource_id, cube, state, replacements, settings=None):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        return mcmc.get_preview_result(datasource_id, cube, state, replacements, settings)

    def get_state_from_mdx(self, mdx, datasource_id=None, cube=None):
        mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id, request=self._request)
        return mcmc.state_from_mdx(mdx, datasource_id=datasource_id, cube=cube)

class MdxFormatStringRPCHandler:
    def __init__(self, request, **kwargs):
        self._request = request
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']
        self._controller = MdxFormatStringController(self.sa_session, account_id=self.account_id, request=self._request)

    def get_format_strings(self):
        return [{'label': fs.name, 'value': fs.id} for fs in self._controller.get_format_strings()]

    def get_format_string_state(self, _id):
        fs = self._controller.get_format_string(_id)
        return {
            'format_string': fs.format_string,
            'id': fs.id,
            'name': fs.name
        }

    def save_or_update_format_string(self, _id, name, format_string):
        fs = self._controller.save_or_update_format_string(_id, name, format_string)
        return fs.id

handlers = (IQubeComposeRPCHandler, MdxFormatStringRPCHandler, )
