# Copyright (c) 2009 BaseSoft B.V.
# See LICENSE for details.
from basesoft.util.ordered_dict import OrderedDict

import re
import xmlrpclib
import os
import commands

from sqlalchemy import and_, create_engine
from twisted.python.reflect import namedAny
from xml.dom import minidom

from django.utils.translation import ugettext as _
from django.conf import settings
from django.utils import simplejson

from basesoft.web.core.jsonrpc.dispatcher import json_object_hook, JsonRpcJSONEncoder
from basesoft.web.core.exceptions import SystemException, PublicUserException, PublicSystemException
from basesoft.web.core.util.controllers import AbstractController, transaction_decorator, AbstractStateController
from basesoft.web.contrib.license.models import Group
from basesoft.web.core.management.commands.iqubesync import ProjectBase

from basesoft import logging
from basesoft.web.contrib.dashboard.controllers import DashboardWidgetStateController
from basesoft.web.contrib.license.controllers import LicenseController
from basesoft.web.core.util.db import DatabaseSession, Session

from iqportal.iqube.core.models import DataSource, DataSourceGroup,\
    CalculatedMember, MdxMember, IncorrectCalculatedMember, DataSourceWidget, DataSourceDashboardWidget, DataSourceCube,\
    MdxMethod, MdxMethodField
from iqportal.iqube.core.exceptions import IQubeSystemException, IQubePublicException
from iqportal.iqube.core.db_sync import AbstractConfigurableWidget


logger = logging.getLogger("iqportal.iqube.core")

def MondrianMemberSerialize(member):

    if type(member) != dict:
        return simplejson.dumps(member, ensure_ascii=False, separators=(', ', ': '))

    n = {}
    if member.has_key('calculated_member_id'):
        keep = ['calculated_member_id', 'replacements', 'settings']
        settings = ['axis', 'format_string', 'label']
        ns = {}
        for s in settings:
            if member['settings'].has_key(s):
                ns[s] = member['settings'][s]
        member['settings'] = ns
    else:
        keep = ['member_unique_name', 'hierarchy_unique_name', 'depth', 'member_name', 'level_unique_name', 'dimension_unique_name']
    for i in keep:
        if member.has_key(i):
            n[i] = member[i]
    return simplejson.dumps(n, ensure_ascii=False, separators=(', ', ': '))

class IQubeController:
    """
    Used to interact with iqube servlet
    """
    def __init__(self):
        if settings.IQUBE_SERVLET is not None:
            self.proxy = xmlrpclib.ServerProxy(settings.IQUBE_SERVLET)

    def call(self, method, *params):
        call_method = getattr(self.proxy.IQubeServer, method)
        try:
            rv = call_method(*params)
        except Exception, e:
            logger.error('Exception raised %s' % e)
            raise IQubePublicException(_("An unexpected exception occured"))
        if rv['status'] == "OK":
            return rv['message']
        if rv['status'] == "IQubePublicException":
            raise IQubePublicException(rv['message'])
        if rv['status'] == "IQubeSystemException":
            raise IQubeSystemException(rv['message'])

    def clear_xmla_cache(self, project_name, minutes=0):
        """
        removes xmla cache file for the given datasource older than minutes
        """

        # check if caching is enabled and if the path is writable
        cache_path = os.path.join(settings.CACHE_DIR, 'xmla', project_name)
        if settings.MONDRIAN_XMLA_CACHE_ENABLED:

            if not os.path.exists(cache_path):
                os.makedirs(cache_path)

            if not os.access(cache_path, os.W_OK):
                raise Exception('cache_path is not writable (%s)' % cache_path)

            # remove any cache files found older than given minutes
            command = "find '%s' -cmin +%s -type f -exec rm {} \;" % (cache_path, minutes)
            status, output = commands.getstatusoutput(command)
            if status != 0:
                raise Exception('failed to clean up: "%s": "%s"' % (command, output))

        # notice: dont call clear_mondrian_cache() here. in some cases (like tnt) the aggregator calls
        # clear_mondrian_member_cache() for only a specific cell region

    def clear_mondrian_cache(self, project_name):
        """
        clears all mondrian cache for the given project name
        """

        self.call("clearCache", project_name)
        logger.info("clear_mondrian_cache: cleared cache for %s, minutes" % (project_name) )


    def clear_mondrian_member_cache(self, project_name, cube_name, members_list):
        """
        clears a part of the mondrian cache for the given project, cube and member list
        fixme: example
        """

        self.call("clearMemberCache", project_name, cube_name, members_list)
        logger.info("clear_mondrian_member_cache: cleared cache for %s,%s,%s" % (project_name, cube_name, members_list) )

    def get_optiq_metadata(self, model_json_path):
        """
        clears all mondrian cache for the given project name
        """

        return self.call("getOptiqMetadata", model_json_path)

    def get_members_captions(self, ds_id, cube, member_str, with_ancestors=False):
        """
        clears all mondrian cache for the given project name
        """
        dsc = DataSourceController()
        schema_obj = dsc.get_datasources(ds_id).pop()
        return self.call("getMembersCaptions", schema_obj.name, cube, member_str, with_ancestors)

    def clean_datasources(self):
        """
        removes all datasources from iqube that are not in the iqportal db (anymore)
        """

        sa_session = Session()

        exclude = [ds.mondrian_project for ds in sa_session.query(DataSource)]
        self.call('cleanupDatasources', exclude)



class DataSourceController(AbstractController):
    """
    This is used to manage Datasources. A Datasource is a link to a project
    on the IQube server. It also includes the username and password to connect
    with.
    """

    def __init__(self, *args, **kwargs):

        AbstractController.__init__(self, *args, **kwargs)
        self.iqube_controller = IQubeController()

    def get_datasource_by(self, field=None, values=None, ds_type=None, priv=DataSource.PRIV_VIEW):
        """
        """

        fo = DataSource
        filter = None

        if self.account_id:
            self.assert_permissions('view_datasource')

            # only query datasources that the user is privileged to.
            filter, fo = self.get_query_filter_on_privileges(DataSource,
                                                          DataSourceGroup,
                                                          priv)
            filter = and_(filter, and_(DataSource.deleted==False, DataSource.archived==False))
        if field is not None:
            assert field in ['mondrian_project', 'name', 'id']

        if values is not None:
            if not isinstance(values, list):
                values = [values]
            if filter is None:
                filter = getattr(DataSource, field).in_(values)
            else:
                filter = and_(filter, getattr(DataSource, field).in_(values))

        if ds_type is not None:
            if filter is None:
                filter = filter, DataSource.type==ds_type
            else:
                filter = and_(filter, DataSource.type==ds_type)

        q = self.sa_session.query(DataSource).select_from(fo).filter(filter).order_by(DataSource.name)
        return q.all()

    def get_datasource(self, id=None, priv=DataSource.PRIV_VIEW):
        """
        This returns a list of data source to which the current user has
        privileges.
        """
        try:
            return self.get_datasources(id, priv=priv).pop()
        except IndexError:
            raise PublicUserException(_('Unknown datasource or no access: %s' % id))

    def get_datasources(self, id=None, ds_type=None, priv=DataSource.PRIV_VIEW):
        """
        This returns a list of data source to which the current user has
        privileges.
        """
        if id is not None:
            return self.get_datasource_by('id', id, ds_type=ds_type, priv=priv)

        return self.get_datasource_by(ds_type=ds_type, priv=priv)

    def get_cubes(self, datasource, priv=DataSourceCube.PRIV_VIEW):

        if type(datasource)==int:
            datasource = self.get_datasource(datasource)
        elif type(datasource) in [str, unicode]:
            datasource = self.get_datasource_by(field='mondrian_project', values=datasource).pop()

        where = DataSourceCube.datasource==datasource
        return self._get_privileged_object(DataSourceCube, where, priv, order_by=DataSourceCube.name)

    def get_datasource_cube(self, datasource_id, cube_name):
        """
        returns a DatasourceCube instance
        """

        assert isinstance(datasource_id, (int, long))
        assert isinstance(cube_name, (str, unicode))
        ds = self.get_datasource(datasource_id)
        where = and_(DataSourceCube.datasource==ds, DataSourceCube.name==cube_name)
        return self._get_privileged_object(DataSourceCube, where, DataSourceCube.PRIV_VIEW, order_by=DataSourceCube.name)[0]

    def get_last_aggregation_update(self, datasource_id, cube):
        c = self.sa_session.query(DataSourceCube).filter(and_(DataSourceCube.datasource_id==datasource_id, DataSourceCube.name==cube)).first()
        if c is None:
            return None

        return c.last_aggregation_update

    def load_datasource_state(self, id):
        """
        This returns a datasource state from IQube server
        """
        ds = self.get_datasource_by('id', id)
        if len(ds) == 0:
            return {}
        ds = ds[0]

        info = self.iqube_controller.call("getDatasourceInfo", ds.name)
        ret = info


        if info['db_connection'].startswith('mysql://'):
            try:

                # ret: {'db_connection': 'mysql://localhost:3306/iqportal?user=root&password=f00b4r', 'definition': 'file:///var/lib/tomcat-6/webapps/iqube-testing/WEB-INF/queries/basesoft.settings.martijn.IQuest.xml', 'name': 'basesoft.settings.martijn.IQuest', 'datasource_info': 'Provider=mondrian;JdbcDrivers=com.mysql.jdbc.Driver;Jdbc=jdbc:mysql://localhost:3306/iqportal?user=root&password=f00b4r;Catalog=file:///var/lib/tomcat-6/webapps/iqube-testing/WEB-INF/queries/basesoft.settings.martijn.IQuest.xml;UseContentChecksum=true;'}

                # FIXME: code below should use a regex #3250

                conn = info['db_connection'].split("/")

                ret['protocol'] = conn[0].replace(":", "")

                hp = conn[2].split(":")
                ret['host'] = hp[0]
                ret['port'] = hp[1]

                db = conn[3].split("?")
                ret['database'] = db[0]

                up = db[1].split("&")
                ret['user'] = up[0].replace('user=', "")
                ret['password'] = up[1].replace('password=', "")
            except IndexError:
                raise SystemException("Unable to parse mysql connection")

        elif info['db_connection'].startswith("Jdbc='jdbc:luciddb"):

            # ret: ret: {'db_connection': "Jdbc='jdbc:luciddb:http://localhost:8034", 'definition': 'file:///var/lib/tomcat-6/webapps/iqube-testing/WEB-INF/queries/ble5.xml', 'name': 'ble5', 'datasource_info': "Provider=mondrian;JdbcDrivers=org.luciddb.jdbc.LucidDbClientDriver;Jdbc='jdbc:luciddb:http://localhost:8034;schema=LOCALDB.schema_ble5';JdbcUser=sa;JdbcPassword=;Catalog=file:///var/lib/tomcat-6/webapps/iqube-testing/WEB-INF/queries/ble5.xml;UseContentChecksum=true;"}

            # FIXME: fill in info below #3251
            ret['user'] = ''
            ret['password'] = ''
            ret['database'] = ret['db_connection']
            ret['host'] = ''
            ret['port'] = ''
            ret['protocol'] = 'luciddb'


        else:
            raise SystemException("Unknown db connection")

        return ret

    def get_datasource_with_cube_privs(self, id=None):
        """
        This returns the datasources on which the user has view right and also create_cube.
        """
        self.assert_permissions('view_datasource', 'datasource_create_cube')

        filter, fo = self.get_query_filter_on_privileges(DataSource,
                                                      DataSourceGroup,
                                                      DataSource.PRIV_CREATE_CUBE)

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, DataSource.id.in_(id))

        return self.sa_session.query(DataSource).select_from(fo).filter(filter).all()

    def get_datasource_with_delete_cube_privs(self, id=None):
        """
        This returns the datasources on which the user has view right and also delete_cube.
        """
        self.assert_permissions('view_datasource', 'datasource_delete_cube')
        return self.get_privileged_object_by_id(DataSource, id=id, priv=DataSource.PRIV_DELETE_CUBE)

    @transaction_decorator
    def set_stale(self, identifier):
        """
        Set the datasource as stale
        @param identifier: id or name of the datasource
        @return: the datasource object
        """
        if isinstance(identifier, (str, unicode)):
            datasource = self.get_datasource_by('name', identifier)
        else:
            datasource = self.get_datasource_by('id', identifier)

        if len(datasource) == 0:
            logger.error("Datasource with identifier '%s' doesn't exist." % identifier)
            raise IQubeSystemException("Datasource with identifier '%s' doesn't exist." % identifier)

        d = datasource[0]
        d.has_stale_definitions = True
        self.sa_session.add(d)

        logger.info("Datasource with identifier '%s' was marked as stale." % identifier)

    @transaction_decorator
    def save_or_update_cube(self, id, name, datasource):
        cube = None
        if id is not None:
            cube = self.get_cube(id)
        else:
            cubes = self.get_cubes(datasource)
            for c in cubes:
                if c.name == name:
                    cube = c
                    break

        if cube is not None:
            self.assert_privileges_on_object(cube, DataSourceCube.PRIV_UPDATE)
        else:
            cube = DataSourceCube()
            cube.name = name
            cube.datasource = datasource
            self.save_default_privilege_object_group(cube)

        self.sa_session.add(cube)
        return cube

    @transaction_decorator
    def save_or_update_datasource(self, id, name, description, mondrian_project, \
                                  protocol, user, password, host, port, database, \
                                  ds_type=DataSource.DATASOURCE_TYPE_DATABASE, group=None):
        """
        This saves or updates a data source definition. Sends all info to the IQube servlet and if all saved ok on mondrian,
        it also save in db.
        """
        if id is not None:
            self.assert_permissions('update_datasource')
            ds = self.sa_session.query(DataSource).get(id)
            self.assert_privileges_on_object(ds, DataSource.PRIV_UPDATE)
        else:
            self.assert_permissions('create_datasource')
            # name must be unique
            ds = self.get_datasource_by('name', name)
            if len(ds) > 0:
                raise PublicUserException(_("Datasource name already exists."))
            ds = DataSource()

        if group is None:
            group = self._get_default_group()
        else:
            assert isinstance(group, Group)

        ds.name = name
        ds.type = ds_type
        if id is None:

            # save mondrian project if given else set it to temp cuz we set it with the id later
            ds.mondrian_project = mondrian_project if not mondrian_project is None else '__temp__'

            ds.group = group
            ds.group_privileges = DataSource.all_privileges
            ds.license_privileges = DataSource.no_privileges
            ds.domain_privileges = DataSource.no_privileges

        self.sa_session.add(ds)

        # save so ds has an id
        self.sa_session.flush()

        if mondrian_project is None:
            ds.mondrian_project = mondrian_project = 'ds_%s' % ds.id
            self.sa_session.add(ds)

        # if database is a callback call it to get the real database name
        # for example for luciddb it uses the datasource.id to get the name
        if hasattr(database, '__call__'):
            database = database(ds)
        assert isinstance(database, (unicode, str))

        # perform connection tests
        if protocol == 'mysql':
            self.test_connection(protocol, user, password, host, port, database)

        # send to mondrian
        db = {"protocol": protocol,
              "host": host,
              "db_name": database,
              "user": user,
              "password": password}

        self.iqube_controller.call("updateDatasource", mondrian_project, db)

        return ds

    def get_datasource_objects(self, id):
        ds = self.get_datasource_by('id', id)
        if len(ds) == 0:
            return {}
        ds = ds[0]
        objects = {"type": ds.type}

        if ds.type == DataSource.DATASOURCE_TYPE_FILE:
            raise Exception('fixme')
        else:
            objects["tables"] = self.get_datasource_tables(ds.id)
        return objects

    def _get_metadata(self, id):
        from sqlalchemy import MetaData
        info = self.load_datasource_state(id)
        connstr = '%s://%s:%s@%s:%s/%s' % (info['protocol'], info['user'], info['password'],
                                           info['host'], info['port'], info['database'])
        return MetaData(bind=connstr, reflect=True)

    def get_datasource_tables(self, id):
        """
        Returns a list with all datasource's tables and their description.
        @param id: id of the datasource
        """
        metadata = self._get_metadata(id)
        tables = []
        for table in metadata.tables:
            tables.append(table)
        return sorted(tables)

    def get_datasource_table_definition(self, id, table_name):
        """
        Returns a list with all datasource's tables and their description.
        @param id: id of the datasource
        """
        from basesoft.sqlalchemy.wrappers import SQLAlchemyTableWrapper
        metadata = self._get_metadata(id)
        return SQLAlchemyTableWrapper(metadata.tables[table_name]).toJson()

    @transaction_decorator
    def delete_datasource(self, id):
        """
        This will only be deleted if no reports are attached to the data source.
        TODO: remove from mondrian also #3254
        """
        self.assert_permissions('delete_datasource')
        self.mark_as_deleted(DataSource, id)

    def test_connection(self, protocol, user, password, host, port, database):
        """
        """
        from sqlalchemy import create_engine
        connstr = '%s://%s:%s@%s:%s/%s' % (protocol, user, password, host, port, database)

        db_engine = create_engine(connstr, echo=False)
        db_engine.connect()

    def _get_datasource_project(self, mondrian_project):
        for app in settings.INSTALLED_APPS:
            try:
                sync_module = namedAny(app+'.iqube_sync')
            except AttributeError:
                continue

            if not hasattr(sync_module, 'projects'):
                continue

            for Project in sync_module.projects:
                if not issubclass(Project, ProjectBase):
                    logger.error('Project [%s] is not a ProjectBase subclass in app %s' % (Project, app))
                    continue

            if '%s.%s' % (settings.SETTINGS_MODULE, Project.project_name) == mondrian_project:
                return Project

        return None

    def get_calculated_member_hook(self, datasource_id):
        datasource = self.get_datasource(datasource_id)

        Project = self._get_datasource_project(datasource.mondrian_project)
        if Project is None:
            return None
        return Project.calculated_member_hook_js_module

    def get_iqube_role_json(self, datasource_id, account_id):
        """
        Gets called by the view get_iqube_role() and returns a list of dicts with iqube role information
        for the given datasource_id and account_id
        """

        datasource = self.sa_session.query(DataSource).filter(DataSource.id == datasource_id).one()

        rv = []

        Project = self._get_datasource_project(datasource.mondrian_project)
        project = Project(self.sa_session, self)

        for Role in project.get_roles():

            role = Role(self.sa_session, datasource)
            rv += role.get_role_list(account_id)

        return rv

class AbstractResultController(AbstractController):
    def cache_widget(self, datasource, config, applied_config, role_name, context):
        from iqportal.iqube.core.reporters import ReporterRpcHandler

        for o, c in context.iteritems():
            config[o] = c

        reporter = ReporterRpcHandler(self.sa_session, datasource, config, account_id=role_name)
        od = reporter.get_mdx_config(applied_config)
        for label, cfg in od.iteritems():
            mdx = {}
            for i, s in cfg['mdx_config'].iteritems():
                mdx[str(i)] = s

            client = datasource.connect(self.sa_session, account_id=role_name)
            if mdx.has_key('context'):
                client.set_context(mdx['context'])
                del mdx['context']
            if mdx.has_key('result_full_headers'):
                del mdx['result_full_headers']

            mdx = client.build_select_statement(**mdx)
            client.execute_mdx(mdx)

class ConfigurableWidgetController(AbstractResultController):

    def get_simple_dimension_options(self, ds, cube, config, values=None):
        """
        This method will return 4 values, it expects the supplied dimension level to only have TWO members (boolean)
        """
        client = ds.connect(self.sa_session, self.get_account().id)

        assert len(config['children']) == 1
        child = config['children'][0]

        ms = getattr(client, child['method'])(cube, child['method_argument'])
        if len(ms) > 2:
            raise PublicSystemException(_('Incorrectly configured widget, please contact the Service Provider.'))

        rv = []
        members = []
        set_ = u"SET [All %(label)s (Members)] AS 'ORDER({%(members)s}, %(level)s.CurrentMember.Caption, BASC)'"

        for d in ms:
            members.append(d['member_unique_name'])
            rv.append({
                'value': MondrianMemberSerialize(d),
                'label': d['member_caption']
            })

        rv.append({
            'value': set_ % {
                'label': config['caption'],
                'members': ','.join(members),
                'level': child['unique_name']
            },
            'label': _('Show for both')
        })

        rv.append({
            'value': '',
            'label': _("Don't add as filter")
        })

        return rv

    def get_dimension_options(self, ds, cube, config, values=None):
        from iqportal.iqube.compose.controllers import MdxCalculatedMemberController
        if values is None:
            values = []

        client = ds.connect(self.sa_session, self.get_account().id)
        rv = OrderedDict()
        selected = OrderedDict()

        def serialize(x):
            # separators used to be compact like the JS one.
            return {'label': x['member_caption'], 'value': MondrianMemberSerialize(x)}

        def deserialize(x):
            try:
                return simplejson.loads(x)
            except simplejson.JSONDecodeError:
                return x

        def add_selected(k, x, selected, y):
            if not selected.has_key(k):
                selected[k] = []
            selected[k].append([x, y])

        for y in values:
            x = deserialize(y)
            if type(x) == dict and x.has_key('calculated_member_id'):
                add_selected('Calculated members', x, selected, y)
            elif type(x) == dict and not x.has_key('format_string'):
                if not x.has_key('level_unique_name'):
                    raise PublicUserException(_('Incorrect selection (possible invalid All-member setup) provided, please recreate the widget.'))
                add_selected(x['level_unique_name'], x, selected, y)
            else:
                add_selected('Custom', x, selected, y)

        def get_sorted_members(unique_name, members):

            # sort by member name
            sort = {}

            for member in members:
                # when we are dealing with ids, ordering should be done as int not as str
                try:
                    key = int(member['member_name'])
                except ValueError:
                    key = member['member_name']
                except KeyError:
                    key = member['member_caption']
                sort[key] = serialize(member)

            # when there are values set with the 'Use .Children' fancy select option, those values
            # are not in default options, so we need to add them
            if selected.has_key(unique_name):
                for i in selected[unique_name]:
                    if type(i[0]) in (str, unicode, ):
                        continue
                    try:
                        key = int(i[0]['member_name'])
                    except ValueError:
                        key = i[0]['member_name']
                    except KeyError:
                        try:
                            key = i[0]['member_caption']
                        except KeyError:
                            key = None

                    if key is None or (sort.has_key(key) and sort[key]['value'] != i[1]):
                        member_unique_name = i[0]['member_unique_name']

                        # check if lookup by .Children or .Members
                        if member_unique_name.endswith('.Children') or member_unique_name.endswith('.Members'):

                            if member_unique_name.endswith('.Children'):
                                lookup = i[0]['member_unique_name'][:-9]
                            else:
                                lookup = i[0]['member_unique_name'][:-8]
                            m = client.get_member_tree_op(cube, lookup, client.MDTREEOP_SELF)
                            if len(m) == 0:
                                logger.warn("get_dimension_options: member '%s' not found in cube '%s'" % (lookup, cube))
                                m = client.get_level_info(cube, i[0]['level_unique_name'])
                                m[0]['member_caption'] = m[0]['level_caption']
                                if len(m) == 0:
                                    continue

                            if member_unique_name.endswith('.Children'):
                                caption = '%s (Children)' % m[0]['member_caption']
                            else:
                                caption = '%s (Members)' % m[0]['member_caption']
                            m[0]['member_caption'] = caption

                        else:
                            lookup = i[0]['member_unique_name']
                            # else assume a lookup by member_unique name
                            m = client.get_member_tree_op(cube, lookup, client.MDTREEOP_SELF)
                            if len(m) == 0:
                                logger.warn("get_dimension_options: member '%s' not found in cube '%s'" % (lookup, cube))
                                continue
                            caption = m[0]['member_caption']

                        sort[m[0]['member_caption']] = {'label': caption, 'value': i[1]}

                del selected[unique_name]
            return sort

        # when filters are set, only get the members that are in those filters, plus their children depending
        # on the 'children' property value
        if config.has_key('filters') and len(config['filters']['members']) > 0:

            levels = OrderedDict()
            # only show those members that have their level in the dimension's children and show_members is True for it
            for child in config['children']:
                if child['show_members']:
                    levels.update([(child['unique_name'], {'caption': child['caption'], 'members': []})])

            for f in config['filters']['members']:

                if f.endswith('.Children') or f.endswith('.Members'):
                    continue

                # in case we have a predefined CM or Set, add it to the "Custom" entries
                if f[:4] == 'SET ' or f[:7] == 'MEMBER ':
                    if not levels.has_key('Custom'):
                        levels.update([('Custom', {'caption': 'Custom', 'members': []})])
                    cm = self._make_calculated_member(f)
                    levels['Custom']['members'].append({'label': cm.name, 'value': MondrianMemberSerialize(str(cm))})
                    continue

                member = client.get_member_tree_op(cube, f, client.MDTREEOP_SELF, "MEMBER_UNIQUE_NAME")
                # either doesn't exist or user does not have permission on it
                if len(member) == 0:
                    continue

                member = member[0]
                # only add it if its level name is in the config dimension's children
                if member.has_key('level_unique_name') and member['level_unique_name'] in levels.keys():
                    levels[member['level_unique_name']]['members'].append(member)

                # depending on the children status set, we get the member children
                # No children needed, continue
                if config['filters']['children'] == DataSourceWidget.FILTERS_CHILDREN_NONE:
                    continue

                # only direct descendants of this member.
                if config['filters']['children'] == DataSourceWidget.FILTERS_CHILDREN_DIRECT_DESCENDANTS:
                    tree_op = client.MDTREEOP_CHILDREN

                # all descendants of this member, children, grand children, etc.
                if config['filters']['children'] == DataSourceWidget.FILTERS_CHILDREN_ALL_DESCENDANTS:
                    tree_op = client.MDTREEOP_DESCENDANTS

                members = client.get_member_tree_op(cube, f, tree_op, "MEMBER_UNIQUE_NAME")
                for m in members:
                    if m.has_key('level_unique_name') and m['level_unique_name'] in levels.keys():
                        levels[m['level_unique_name']]['members'].append(m)

            for k, v in levels.items():
                if k == 'Custom':
                    rv.update([('Custom', v['members'])])
                    continue
                if len(v['members']) > 0:
                    sort = get_sorted_members(k, v['members'])
                    rv.update([(v['caption'], [sort[member] for member in sorted(sort)], )])

        # if no filters are set, we get all members depending on children
        else:
            for child in config['children']:
                all = []
                depth = None
                if child.has_key('depth'):
                    depth = child['depth']

                if not child['caption']:
                    member = client.get_level_info(cube, child['unique_name'])
                    if len(member) > 0:
                        child['caption'] = member[0]['level_caption']

                if child.has_key('has_all_member_available') and child['has_all_member_available']:
                    # Bit of a hack: this is to check if this dimension/level/hierarchy actually has the ALL member
                    # else we get a Member not found except when selected.
                    # Also, this way it uses the ACTUAL proper ALL name as defined in the cube.

                    q = client.get_member_tree_op(cube, child['unique_name'], client.MDTREEOP_SELF, "LEVEL_UNIQUE_NAME")
                    if len(q) == 0:
                        raise Exception('Cannot get childs for member: ', child['unique_name'])
                    lvl = q[0]

                    if lvl.has_key('parent_unique_name') and lvl['parent_unique_name'].split('].[')[1][:4] == 'All ':
                        all.append({'depth': depth, 'member_unique_name': '%s' % lvl['parent_unique_name'],
                                    'member_caption': 'All %s' % child['caption'],
                                    'level_unique_name': lvl['level_unique_name'],
                                    'dimension_unique_name': child['unique_name'].split('.')[0]}),

                if child.has_key('has_children_available') and child['has_children_available']:
                    member = client.get_level_info(cube, child['unique_name'])
                    if len(member) > 0:
                        member = member[0]
                        member['member_unique_name'] = '%s.Members' % child['unique_name']
                        member['member_caption'] = '%s (Members)' % child['caption']
                        member['member_name'] = child['unique_name']
                        #member['level_unique_name'] = 'Members'
                        all.append(member)

                if len(all) > 0:
                    rv.update([('All (%s)' % child['caption'], [serialize(x) for x in all],)])

                if child.has_key('show_members') and child['show_members']:
                    ms = getattr(client, child['method'])(cube, child['method_argument'])

                    if child['unique_name'].find('[Measures]') == 0:
                        # mondrian will send back all measures plus all calculated members, so we need to filter that
                        members = [m for m in ms if m['dimension_unique_name'] == '[Measures]']
                    else:
                        members = ms

                    if len(members) > 0:
                        sort = get_sorted_members(child['unique_name'], members)
                        rv.update([(child['caption'], [member for member in sorted(sort.values(), key=lambda m: m['label'].lower())], )])

        for k,v in selected.iteritems():
            supply = []
            is_custom = False
            is_cm = False
            for i in v:
                if type(i[0]) in (str, unicode, ):
                    if i[0][:4] == 'SET ' or i[0][:7] == 'MEMBER ':
                        cm = self._make_calculated_member(i[0])
                        is_custom = True
                        supply.append({'label': cm.name, 'value': MondrianMemberSerialize(str(cm))})
                        continue

                if type(i[0]) in (int, ):
                    mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.get_account().id)
                    formula_config = mcmc.auto_configure_mdx_function(i[0])
                    is_cm = True
                    supply.append({'label': formula_config['settings']['label'], 'value': MondrianMemberSerialize(formula_config)})
                    continue

                if i[0].has_key('calculated_member_id'):
                    is_cm = True
                    supply.append({'label': i[0]['settings']['label'], 'value': MondrianMemberSerialize(i[0])})
                    continue

                if i[0].has_key('format_string'):
                    is_custom = True
                    supply.append({'label': i[0]['member_caption'], 'value': i[1]})
                    continue

                if  i[0]['member_unique_name'].endswith('.Children'):
                    lvl = client.get_level_info(cube, i[0]['level_unique_name'] ).pop()
                    m = {'member_caption': lvl['level_caption'] + ' (Children)'}
                    #i[0]['member_name'] = lvl['level_unique_name']

                elif  i[0]['member_unique_name'].endswith('.Members'):
                    lvl = client.get_level_info(cube, i[0]['level_unique_name'] ).pop()
                    m = {'member_caption': lvl['level_caption'] + ' (Members)'}
                    #i[0]['member_name'] = lvl['level_unique_name']

                else:
                    m = client.get_member_tree_op(cube, '%s.[%s]' % (i[0]['level_unique_name'], i[0]['member_name']), client.MDTREEOP_SELF)[0]

                i[0]['member_caption'] = m['member_caption']
                supply.append(i[0])
            if is_custom:
                rv.update([('Custom', supply)])
                continue
            if is_cm:
                rv.update([('Calculated members', supply)])
                continue
            sort = get_sorted_members(k, supply)
            child = client.get_level_info(cube, k)[0]
            rv.update([(child['level_caption'], [sort[member] for member in sorted(sort)], )])


        if config.has_key('show_calculated_members') and config['show_calculated_members']:
            cube_cms = [] if not rv.has_key('Calculated members') else rv['Calculated members']
            dimension = None
            if config['children'][0]['unique_name'] != '[Measures].[MeasuresLevel]':
                cube_cms = [serialize(x) for x in client.get_calculated_members(cube, config['caption'])]
                dimension = config['caption'] # FIXME
            user_cms = []

            mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.get_account().id)
            for cm in mcmc.get_calculated_members_for(ds, cube, dimension=dimension):
                user_cms.append({'value': cm.id, 'label': cm.label})

            rv.update([(
                'Calculated members',
                cube_cms+user_cms,
            )])

        return rv

    def _make_calculated_member(self, member):
        if isinstance(member, dict):
            return CalculatedMember(**member)
        if isinstance(member, (str, unicode)):
            return CalculatedMember(member_string=member)
        return None

    def get_level_descendants(self, ds, cube, config, filter, selected=None):
        """
        filtered: members for which to retrieve the descendents
        selected: (optional): already selected members which should also be returned
        """

        client = ds.connect(self.sa_session, self.get_account().id)
        used = {}
        temp = {}
        level_order = {}

        def append_used(member, used, serialized, member_label):
            # we are dealing with a Custom option
            if not member.has_key('level_unique_name'):
                lvl = 'Custom'
            else:
                lvl = member['level_unique_name']

            if not used.has_key(lvl):
                used[lvl] = []

            used[lvl].append((member, serialized, member_label))


        for serialized in filter:
            member = simplejson.loads(serialized)

            member_q = client.get_member_tree_op(cube, '%s.[%s]' % (member['level_unique_name'], member['member_name']), client.MDTREEOP_SELF)
            if len(member_q) == 0:
                raise PublicUserException(_("Can't get info for: %s") % member['member_unique_name'])
            member_q = member_q[0]

            # we first get the level data (the level itself should be in the result still)
            append_used(member_q, used, MondrianMemberSerialize(member_q), member_q['member_caption'])

            descendants = client.get_member_tree_op(cube, member['member_unique_name'], client.MDTREEOP_DESCENDANTS)

            if len(descendants) == 0: # no access on parent levels in role defined
                descendants = client.get_member_tree_op(cube, '%s.[%s]' % (member['level_unique_name'], member['member_name']), client.MDTREEOP_DESCENDANTS)

            for d in descendants:
                append_used(d, used, MondrianMemberSerialize(d), d['member_caption'])

        if selected is not None:

            # For the selected ones we do not filter down to the descendants
            for serialized in selected:
                unserialized = simplejson.loads(serialized)
                # We need to reload these from mondrian as serialized doesn't contain member_caption
                member_unique_name = unserialized['member_unique_name']

                # check if lookup by .Children or .Members
                if member_unique_name.endswith('.Children') or member_unique_name.endswith('.Members'):

                    m = client.get_member_tree_op(cube, unserialized['level_unique_name']+'.'+unserialized['member_name'], client.MDTREEOP_SELF)
                    if len(m) == 0:
                        logger.warn('WARN: member "%s" not found in clube %s' % (unserialized['hierarchy_unique_name']+'.'+unserialized['member_name'], cube))
                        continue

                    if member_unique_name.endswith('.Children'):
                        caption = '%s (Children)' % m[0]['member_caption']
                    else:
                        caption = '%s (Members)' % m[0]['member_caption']

                else:
                    # else assume a lookup by member_unique name
                    m = client.get_member_tree_op(cube, '%s.[%s]' % (unserialized['level_unique_name'], unserialized['member_name']), client.MDTREEOP_SELF)
                    if len(m) == 0:
                        logger.warn('WARN: member "" not found in cube %s' % (member_unique_name, cube))
                        continue
                    caption = m[0]['member_caption']

                append_used(m[0], used, serialized, caption)

        # TODO: always append Custom (grouped?) at the end. #3252
        for level, members in used.iteritems():
            lvl = client.get_level_info(cube, level).pop()
            level_order[lvl['level_number']] = lvl['level_caption']

            temp[lvl['level_caption']] = []
            for x in members:
                member, serialized, member_label = x
                temp[lvl['level_caption']].append( {'value': serialized, 'label': member_label} )

        return OrderedDict([(level_order[index], temp[level_order[index]], ) for index in sorted(level_order)])



    def get_configurable_widgets(self, id=None, priv=DataSourceWidget.PRIV_VIEW, availability=None):
        self.assert_permissions('view_datasource_widget')
        widgets = self.get_privileged_object_by_id(DataSourceWidget, id, priv)
        if availability is None:
            return widgets
        return [widget for widget in widgets if widget.is_available(availability)]

    def get_configurable_widget_by_label(self, label):
        self.assert_permissions('view_datasource_widget')
        where = DataSourceWidget.label==label
        widgets = self._get_privileged_object(DataSourceWidget, where, priv=DataSourceWidget.PRIV_VIEW)
        return widgets

    def get_state(self, widget):

        if not isinstance(widget, DataSourceWidget):
            widget = self.get_configurable_widgets(widget).pop()

        config = simplejson.loads(widget.config, object_hook=json_object_hook)
        config['label'] = widget.label
        config['category'] = widget.category
        config['datasource_id'] = widget.datasource.id
        config['cube'] = widget.cube
        config['id'] = widget.id
        applied_config = None
        if widget.applied_filters is not None:
            applied_config = simplejson.loads(widget.applied_filters, object_hook=json_object_hook)
        return (config, applied_config, )

    def get_sync_source_definition(self, widget):
        if not isinstance(widget, DataSourceWidget):
            widget = self.get_configurable_widgets(widget).pop()

        from iqportal.iqube.core import configurable_widget_class_definition
        config, applied_filters = self.get_state(widget)

        def dump(_config):
            return simplejson.dumps(_config, ensure_ascii=False, indent=4) \
                .replace(': true', ': True') \
                .replace(': false', ': False')

        definition = configurable_widget_class_definition % {
            'class_name': config['label'].replace(' ', ''),
            'category': widget.category,
            'cube': widget.cube,
            'datasource': widget.datasource.name.replace('%s.' % settings.SETTINGS_MODULE, ''),
            'config': dump(config),
            'applied_filters': dump(applied_filters),
        }
        return (config['label'], definition)

    @transaction_decorator
    def copy_widget(self, id, label, category, applied_filters):
        widget = self.get_configurable_widgets(id).pop()
        config = self.get_state(widget)[0]
        return self.save_or_update_configurable_widget(None, widget.datasource, widget.cube, label, category, config, applied_filters)

    @transaction_decorator
    def update_configurable_widget_default_filters(self, id, applied_filters):
        widget = self.get_configurable_widgets(id).pop()
        widget.applied_filters = simplejson.dumps(applied_filters, ensure_ascii=False, cls=JsonRpcJSONEncoder)
        self.sa_session.add(widget)
        return widget

    @transaction_decorator
    def save_or_update_configurable_widget(self, id, datasource, cube, label, category, config, applied_filters, privileges=None, availability=None):

        assert isinstance(datasource, DataSource)
        if id is None:
            self.assert_permissions('create_datasource_widget')
            widget = DataSourceWidget()
            widget.datasource = datasource
            widget.cube = cube
        else:
            self.assert_permissions('update_datasource_widget')
            widget = self.get_configurable_widgets(id, priv=DataSourceWidget.PRIV_UPDATE).pop()
            assert widget.datasource == datasource
            assert widget.cube == cube

        widget.label = label
        widget.category = category
        widget.config = simplejson.dumps(config, ensure_ascii=False, cls=JsonRpcJSONEncoder)
        widget.applied_filters = simplejson.dumps(applied_filters, ensure_ascii=False, cls=JsonRpcJSONEncoder)
        if availability is not None:
            widget.availability = availability

        if id is None:
            self.save_default_privilege_object_group(widget)

        if privileges is not None:
            g = privileges['group_has_privilege']
            l = False
            if privileges.has_key('license_has_privilege'):
                l = privileges['license_has_privilege']
            d = False
            if privileges.has_key('domain_has_privilege'):
                d = privileges['domain_has_privilege']
            lc = LicenseController(self.sa_session, account_id=self.get_account().id)
            groups = lc.get_groups(privileges['groups'])
            for group in groups:
                self.grant_privilege_to(widget, DataSourceWidget.PRIV_VIEW, g, l, d, group)

        self.sa_session.add(widget)
        return widget

    def cache_widget(self, widget, role_name, applied_config=None, context=None):

        config, ac = self.get_state(widget)
        if applied_config is None:
            applied_config = ac

        AbstractResultController.cache_widget(self, widget.datasource, config, applied_config, role_name, context)


    def save_cube_config(self, item, state):
        """
        Gets called for 'item result' from dashboard item state controller and iqreport.
        It gets the CW config from the state and saves it to the cube
        """

        dsc = DataSourceController(self.sa_session, account_id=self.account_id)

        if type(state['config']['datasource_id']) == str:
            state['config']['datasource_id'] = int(state['config']['datasource_id'])
        cube = dsc.get_datasource_cube(state['config']['datasource_id'], state['config']['cube'])
        if cube.default_config is None:
            cube.default_config = simplejson.dumps(state['config'], ensure_ascii=False)
            self.sa_session.add(cube)

        item.datasource_cube = cube
        return item

    def load_cube_config(self, item):
        """
        Gets called for 'item result' from dashboard item state controller and iqreport.
        Returns the cube CW config
        """

        if item.datasource_cube is None:
            return {'dimensions': []}
        return simplejson.loads(item.datasource_cube.default_config)

class ConfigurableWidgetStateController(DashboardWidgetStateController):
    """
    Controller for Configurable Dashboard Report Widgets
    """

    _instance_class = DataSourceDashboardWidget

    def __init__(self, session, manager):
        DashboardWidgetStateController.__init__(self, session, manager)
        self.ds_controller = DataSourceController(session, account_id=manager.account_id)
        self.cwc = ConfigurableWidgetController(session, account_id=manager.account_id)

    @transaction_decorator
    def save(self, column, state, position):

        widget = DashboardWidgetStateController._instance_from_state(self, state)

        # if 'datasource_widget_id' isnt passed in the state, then this is a 'item result' widget
        if state['datasource_widget_id'] is None:
            widget = self.cwc.save_cube_config(widget, state)
            assert widget.datasource_cube is not None

        # else: normal configurable widget
        else:
            widget.datasource_widget = self.cwc.get_configurable_widgets(state['datasource_widget_id']).pop()

        widget.applied_config = simplejson.dumps(state['appliedConfig'], ensure_ascii=False, cls=JsonRpcJSONEncoder)

        # TODO: verify graphConfig?

        if len(state['graphConfig']) > 0:
            widget.graph_config = simplejson.dumps(state['graphConfig'], ensure_ascii=False, cls=JsonRpcJSONEncoder)

        if len(state['tableConfig']) > 0:
            widget.table_config = simplejson.dumps(state['tableConfig'], ensure_ascii=False, cls=JsonRpcJSONEncoder)

        widget.temp_linked_items = state['linkedItems']

        DashboardWidgetStateController.save(self, column, state, position, widget)

    def load(self, id):
        widget = self._instance_from_id(id)
        gc = {}
        if widget.graph_config is not None:
            gc = simplejson.loads(widget.graph_config, object_hook=json_object_hook)
        tc = {}
        if widget.table_config is not None:
            tc = simplejson.loads(widget.table_config, object_hook=json_object_hook)

        li = {}
        if widget.linked_items is not None:
            li = simplejson.loads(widget.linked_items, object_hook=json_object_hook)

        # check for item result, else normal cw
        if widget.datasource_widget is None:
            cfg = self.cwc.load_cube_config(widget)
        else:
            cfg = self.cwc.get_state(widget.datasource_widget)[0]

        requires = []
        for i in cfg['dimensions']:
            if i.has_key('fancy_select_module') and i['fancy_select_module'] not in requires:
                requires.append(i['fancy_select_module'])

        state = {
            "objectName": widget.object_name,
            "loaderName": widget.object_loader,
            "requires": requires,
            "dbID": widget.id,
            "position": widget.position,
            "title": widget.title,
            "show_header": widget.show_header,
            "hide_margins": widget.hide_margins,
            'datasource_widget_id': None if widget.datasource_widget is None else widget.datasource_widget.id,
            'config': cfg,
            'appliedConfig': simplejson.loads(widget.applied_config, object_hook=json_object_hook),
            'graphConfig': gc,
            'tableConfig': tc,
            'linkedItems': li,
            "text": widget.text,
            "text_location": widget.text_location,
        }

        return state

    def canvas_post_save(self, widget):
        if len(widget.temp_linked_items) == 0:
            widget.linked_items = None
            self.sa_session.add(widget)
            return

        links = {}

        for db_id, axis in widget.temp_linked_items.iteritems():
            if unicode(db_id).startswith('yui-'):
                db_id = AbstractStateController.save_state_id_map[db_id]

            links[db_id] = axis

        widget.linked_items = simplejson.dumps(links, ensure_ascii=False, cls=JsonRpcJSONEncoder)
        self.sa_session.add(widget)
        return

class ProjectController(DataSourceController):
    """
    This is used to manage Mondrian Schema (project). A schema is a collection of cubes.
    """

    @transaction_decorator
    def flush_schema(self, datasource_id):
        """
        this calls writeSchema method on the iqube servlet which re-writes the complete schema file
        Since we have set UseContentChecksum=true in datasources.xml it will reload it automatically
        """

        # TODO: implement 'flush' permission #3253

        # this method can either be called from the iqubesync script (no user there) or from the interface by the user
        # so only call get_datasources() which does the permission check in that case
        if self.account_id is not None:
            ds = self.get_datasources(datasource_id)[0]
        else:
            ds = self.sa_session.query(DataSource).get(datasource_id)

        ds.has_stale_definitions = False
        self.sa_session.add(ds)

        fallback = self.get_schema_string_from_mondrian(ds.mondrian_project)

        if ds.type == DataSource.DATASOURCE_TYPE_FILE:
            logger.info("flush_schema: file based datasource %s" % ds)
            xml = self.rebuild_xml_for_file(ds)
        else:
            xml = self.get_schema_string_from_iqube_sync(ds)

        self.write_schema(ds.mondrian_project, xml, validate=False)

        # also clear cache when flushing schema
        try:
            self.clear_cache(datasource_id)
        except IQubeSystemException, e:
            logger.info("flush_schema: reverting for %s" % ds)
            self.write_schema(ds.mondrian_project, fallback, validate=False)
            raise PublicUserException(e)

        logger.info("flush_schema: flushed for datasource %s" % ds)

    def write_schema(self, mondrian_project, xml, validate=True):
        """
        validates then writes the given xml to the given mondrian project. project must exist already
        """

        from lxml import etree

        dtd = etree.XMLSchema ( etree.XML(open(os.path.join(os.path.dirname(__file__), 'mondrian.xsd')).read()) )
        root = etree.XML(xml)

        if validate:
            valid = dtd.validate(root)

            if not valid:
                print xml
                errors = dtd.error_log.filter_from_errors()
                raise PublicUserException(_("Invalid schema: %s") % errors)

        self.iqube_controller.call("writeSchema", mondrian_project, xml)

        logger.info('write_schema: wrote new schema for mondrian_project: %s' % (mondrian_project))


    def get_schema_string_from_mondrian(self, mondrian_project):
        """
        returns the schema contents of what is currently on the mondrian server
        """
        return self.iqube_controller.call("getSchema", mondrian_project)

    def get_mondrian_cube_xml(self, mondrian_project, cube_name):
        """
        extracts the cube xml from the schema for the given mondrian_project and cube_name
        """

        schema = self.get_schema_string_from_mondrian(mondrian_project)
        xschema = minidom.parseString(schema)
        xcubes = xschema.getElementsByTagName('Cube')

        for xcube in xcubes:
            if xcube.getAttribute('name') == cube_name:
                return xcube.toprettyxml()

        raise Exception('No such cube: %s' % cube_name)


    def rebuild_xml_for_file(self, datasource):
        schema = self.get_schema_string_from_mondrian(datasource.mondrian_project)

        xschema = minidom.parseString(schema)
        xcubes = xschema.getElementsByTagName('Cube')

        for xcube in xcubes:
            members = []
            sets = []
            # remove all current calculated members first:
            cms = xcube.getElementsByTagName('CalculatedMember')
            for cm in cms:
                cm.parentNode.removeChild(cm)
            cms = xcube.getElementsByTagName('NamedSet')
            for cm in cms:
                cm.parentNode.removeChild(cm)

            for i in [members, sets]:
                for tag in i:
                    xcube.appendChild(tag)

        return xschema.toxml()

    def save_mondrian_cube_xml(self, datasource, cube_name, cube_xml):
        """
        saves a new cube to mondrian or updates existing cube to the existing schema
        """

        mondrian_project = datasource.mondrian_project

        # get schema string from mondrian and feed it to the xml dom parser
        schema = self.get_schema_string_from_mondrian(mondrian_project)
        xschema = minidom.parseString(schema)

        # make another dom parser from the new cube_xml
        xnewcube = minidom.parseString(cube_xml)

        xcube = None
        xcubes = xschema.getElementsByTagName('Cube')

        # if new cube, make sure the same cube doesnt already exist
        if cube_name is None:
            new_cube_name = xnewcube.getElementsByTagName('Cube')[0].getAttribute('name')
            for xcube_ in xcubes:
                if xcube_.getAttribute('name') == new_cube_name:
                    raise PublicUserException(_('Cube already exists: %s') % new_cube_name)

        # else if existing cube, make sure the requested cube exists in the current schema
        else:
            for xcube_ in xcubes:
                if xcube_.getAttribute('name') == cube_name:
                    xcube = xcube_
            if xcube is None:
                raise Exception('No such cube: %s' % cube_name)

        # remove the old cube if update
        if xcube is not None:
            xcube.parentNode.removeChild(xcube)

        # add the updated cube to the current schema
        xschema.getElementsByTagName("Schema")[0].appendChild(xnewcube.getElementsByTagName('Cube')[0])

        # send the new schema to mondrian
        self.write_schema(mondrian_project, xschema.toprettyxml())

        # clear cache so members are reloaded
        self.clear_cache(datasource.id)

        logger.info('save_mondrian_cube_xml: updated mondrian_project: %s cube: %s' % (mondrian_project, cube_name))

    def delete_mondrian_cube(self, datasource_id, mondrian_project, cube_name):

        # first we need to make sure the cube is not being used check tblDataSourceWidget
        used = self.sa_session.query(DataSourceWidget).filter(and_(DataSourceWidget.cube == cube_name,
            DataSourceWidget.datasource_id == datasource_id)).all()
        if len(used):
            raise PublicUserException(_('Cube %s is in use by a DataSourceWidget') % cube_name)

        # get schema string from mondrian and feed it to the xml dom parser
        schema = self.get_schema_string_from_mondrian(mondrian_project)
        xschema = minidom.parseString(schema)

        xcube = None
        xcubes = xschema.getElementsByTagName('Cube')

        for xcube_ in xcubes:
            if xcube_.getAttribute('name') == cube_name:
                xcube = xcube_

        if xcube is None:
            raise PublicUserException(_('Cube does not exist: %s') % cube_name)

        # delete cube
        xcube.parentNode.removeChild(xcube)

        # send the new schema to mondrian
        self.write_schema(mondrian_project, xschema.toprettyxml(), validate=False)

        logger.info('delete_mondrian_cube: deleted cube in mondrian_project: %s cube: %s' % (mondrian_project, cube_name))


    def get_schema_string_from_iqube_sync(self, ds):
        """
        returns the complete schema xml for the datasource
        """

        schema_name = ds.mondrian_project

        app_names = settings.INSTALLED_APPS

        controller = IQubeController()

        for app in app_names:
            try:
                sync_module = namedAny(app+'.iqube_sync')
            except AttributeError:
                # if we do not have a sync module, we don't care for this app, continue...
                continue

            # if there is no sync command, we have nothing to do.
            if not hasattr(sync_module, 'projects'):
                continue

            for Project in sync_module.projects:
                if not issubclass(Project, ProjectBase):
                    continue


                _schema_name = '%s.%s' % (settings.SETTINGS_MODULE, Project.project_name)
                if schema_name == _schema_name:

                    xml = '<Schema name="%s" >' % schema_name

                    project = Project(self.sa_session, controller)

                    for Cube in project.get_cubes():
                        cube = Cube(self.sa_session)
                        if cube.xml is None:
                            continue
                        xml += cube.xml

                    for Role in project.get_roles():
                        role = Role(self.sa_session, ds)
                        if role.xml is None:
                            continue
                        xml += role.xml

                    xml += '</Schema>'
                    return xml


        raise Exception('No such schema: %s' % schema_name)

    def get_cubes_by_local_table(self, mondrian_project, table):
        """
        Search the schema for cubes built from local table
        """
        # get schema string from mondrian and feed it to the xml dom parser
        schema = self.get_schema_string_from_mondrian(mondrian_project)
        xschema = minidom.parseString(schema)

        xcubes = xschema.getElementsByTagName('Cube')
        cubes = []

        for xcube_ in xcubes:
            for child in xcube_.getElementsByTagName("Table"):
                if child.getAttribute('name') == table:
                    cubes.append(xcube_.getAttribute('name'))
        return cubes

    def delete_local_table(self, datasource, table):

        if len(self.get_cubes_by_local_table(datasource.mondrian_project, table)):
            raise PublicUserException(_('Cannot delete table as it is in use by cubes: %s') % table)

        raise Exception('fixme')

    @transaction_decorator
    def clear_cache(self, datasource_id):
        """
        Calls the clearCache method on the iqube servlet. That loops all cubes and clears the cache
        """

        ds = self.get_datasources(datasource_id)[0]
        self.sa_session.add(ds)

        # clear xmla cache files
        self.iqube_controller.clear_xmla_cache(ds.mondrian_project)

        # tells mondrian to clear cache
        self.iqube_controller.clear_mondrian_cache(ds.mondrian_project)


class MdxMemberController(AbstractController):

    @transaction_decorator
    def save_mdx_member(self, datasource_id, cube, definition, members=[]):
        """
        Save an mdx member obj.
        Can never update a member since it can be used in multiple places

        @param members: list with member definition
        """

        if isinstance(datasource_id, DataSource):
            datasource = datasource_id
        else:
            dsc = DataSourceController(self.sa_session, account_id=self.account_id)
            datasource = dsc.get_datasources(datasource_id)

        # check if we already have it
        q = self.sa_session.query(MdxMember).filter(and_(MdxMember.datasource==datasource,
            MdxMember.cube==cube, MdxMember.definition==definition)).all()
        # search grouped members also
        if len(q) > 0:
            for m in q:
                if len(m.members) != len(members):
                    continue
                same_members = True
                for mm in m.members:
                    # members order does not count
                    if mm.definition not in members:
                        same_members = False
                        break
                if same_members:
                    return m

        mdx_member = MdxMember()
        mdx_member.datasource = datasource
        mdx_member.cube = cube
        mdx_member.definition = definition

        g_members = []
        for m in members:
            member = self.save_mdx_member(datasource, cube, m)
            g_members.append(member)
        mdx_member.members = g_members

        self.sa_session.add(mdx_member)
        return mdx_member

    def save_from_str(self, datasource_id, cube, str):
        """
        Parse str and saves correspondent mdx members objs
        """
        if str[0:6] != 'MEMBER':
            return self.save_mdx_member(datasource_id, cube, str)
        else:
            regexp = "MEMBER ([\[\w\s\-\.\]]+\.\[.*\]) AS '(.*)'"
            pattern = re.compile(regexp)
            match = re.match(pattern, str)
            if match is None:
                raise IncorrectCalculatedMember()
            defs = match.group(2).split('+')
            return self.save_mdx_member(datasource_id, cube, match.group(1), defs)

    def load_str(self, member):
        """
        Returns mdx member definition
        """
        if isinstance(member, MdxMember):
            mdx_member = member
        else:
            mdx_member = self.sa_session.query(MdxMember).filter(MdxMember.id==member).one()
        if mdx_member.members is None or len(mdx_member.members) == 0:
            return mdx_member.definition
        definition = "MEMBER %s AS '" % mdx_member.definition
        for m in mdx_member.members:
            definition = definition + m.definition + '+'
        definition = definition[:-1] + "'"
        return definition


class IQubeRole(object):
    """
    Object that contains role information. Goes into IQubeRoleController grant methods
    """
    MEMBER_TYPE_CUBE = 1
    MEMBER_TYPE_DIMENSION = 2
    MEMBER_TYPE_HIERARCHY = 3
    MEMBER_TYPE_MEMBER = 4

    ACCESS_NONE = 1
    ACCESS_CUSTOM = 2
    ACCESS_RESTRICTED = 3
    ACCESS_ALL_DIMENSIONS = 4
    ACCESS_ALL = 5

    ROLLUP_POLICY_FULL = 1
    ROLLUP_POLICY_PARTIAL = 2
    ROLLUP_POLICY_HIDDEN = 3

    cube = None
    unique_name = None
    type = None
    access = None
    rollup_policy = None
    top_level = None
    bottom_level = None
    order = None

    def to_dict(self):

        assert self.cube is not None
        assert self.type is not None

        if self.type == self.MEMBER_TYPE_CUBE:
            assert self.unique_name is None
        else:
            assert self.unique_name is not None

        assert self.access is not None


        return {
            'cube': self.cube,
            'unique_name': self.unique_name,
            'type': self.type,
            'access': self.access,
            'rollup_policy': self.rollup_policy,
            'top_level': self.top_level,
            'bottom_level': self.bottom_level,
            'order': self.order,
        }

    def __repr__(self):
        return '<IQubeRole %s=%s>' % (self.unique_name, self.access)



class IQubeRoleController(AbstractController):

    def __init__(self, sa_session, account, datasource, cube):
        self._log('__init__ for account [%s] / [%s] / [%s]' % (account, datasource, cube))

        self.sa_session = sa_session
        self.account = account
        self.datasource = datasource
        self.cube = cube

        self._cubes = []
        self._dimensions = []
        self._hierarchies = []
        self._members = []

    @transaction_decorator
    def save(self):
        raise Exception('obsolete')

    def to_list(self):

        rv = []

        for t in [self._cubes, self._dimensions, self._hierarchies, self._members]:
            order = 0
            for dsr in t:
                dsr.datasource = self.datasource
                dsr.account = self.account
                dsr.cube = self.cube
                dsr.order = order
                order = order + 1

                rv.append(dsr.to_dict())

        # return result ordered by .type, .order
        return sorted(rv, key=lambda k: (k['type'], k['order']))

    def grant_cube(self, access):
        self._log('- grant_cube: %s' % (access))
        self._cubes.append(self.grant(access, _type=IQubeRole.MEMBER_TYPE_CUBE))

    def grant_dimension(self, dimension, access):
        self._log('- grant_dimension: %s, %s' % (dimension, access))
        self._dimensions.append(self.grant(access, dimension, IQubeRole.MEMBER_TYPE_DIMENSION))

    def grant_hierarchy(self, hierarchy, access, rollup_policy=None, top_level=None, bottom_level=None):
        self._log('- grant_hierarchy: %s, %s' % (hierarchy, access))
        self._hierarchies.append(self.grant(access, hierarchy, IQubeRole.MEMBER_TYPE_HIERARCHY, rollup_policy=rollup_policy, top_level=top_level, bottom_level=bottom_level))

    def grant_member(self, member, access):
        self._log('- grant_member: %s, %s' % (member, access))
        self._members.append(self.grant(access, member, IQubeRole.MEMBER_TYPE_MEMBER))

    def grant(self, access, member=None, _type=None, rollup_policy=None, top_level=None, bottom_level=None):
        dsr = IQubeRole()
        dsr.access = access
        dsr.unique_name = member
        dsr.type = _type
        dsr.rollup_policy = rollup_policy
        dsr.top_level = top_level
        dsr.bottom_level = bottom_level
        return dsr

    def _log(self, message):
        logger.debug('IQubeRoleController(): %s' % message)

class IQubeAggregator(AbstractController):
    """
    Abstract aggregator
    """

    _table = ""
    _columns = ""
    _select_sql = ""
    _affected_cubes = None

    def __init__(self):
        self.db_engine = create_engine(settings.DB_CONNSTR, echo=False).connect()
        self.db_engine_ro = create_engine(settings.DB_CONNSTR_READONLY, echo=False, isolation_level='READ COMMITTED').connect()

    def run(self, loop_validator=None):
        pass

    def release(self):
        self.db_engine.close()
        self.db_engine_ro.close()

    def _work(self, where_args, loop_validator=None):

        agg_rows = self.db_engine_ro.execute(self._select_sql, where_args)

        self.db_engine.execute("BEGIN")

        self._delete(where_args)

        def flush():
            """
            flushes max 1000 records
            """

            if len(flush.buffer) == 0:
                return 0

            sql = "REPLACE INTO `"+self._table+"` ("+ ','.join(('`%s`' % col for col in self._columns))  + ") VALUES "

            values = []
            for agg_row in flush.buffer:
                sql += "("+','.join(['%s'] * len(self._columns))+"),"

                [values.append(agg_row[col]) for col in self._columns]
            sql = sql[:-1]

            res = self.db_engine.execute(sql, *values)

            # empty buffer
            flush.buffer = []

            if loop_validator is not None:
                loop_validator()
            return res.rowcount

        flush.buffer = []
        ret = 0
        for agg_row in agg_rows:
            flush.buffer.append(agg_row)
            if len(flush.buffer) >= 1000:
                ret = ret + flush()
        ret = ret + flush()

        agg_rows.close()

        self.db_engine.execute("COMMIT")

        if self._affected_cubes is not None:
            self._update_affected_cubes()

        return ret

    def _delete(self, where_args):
        pass

    def _update_affected_cubes(self):
        import datetime
        sa_session = DatabaseSession().Session
        for cube in self._affected_cubes:
            c = sa_session.query(DataSourceCube).join(DataSource).filter(
                and_(DataSourceCube.name==cube['cube'], DataSource.name=='%s.%s' % (settings.SETTINGS_MODULE, cube['datasource']))).first()
            c.last_aggregation_update = datetime.datetime.now()
            sa_session.add(c)
        sa_session.flush()

class MdxMethodController(AbstractController):

    def get_mdx_method(self, _id=None):
        return self.get_privileged_object_by_id(MdxMethod, _id, priv=MdxMethod.PRIV_VIEW)

    def get_mdx_methods(self):
        return self._get_privileged_object(MdxMethod, priv=MdxMethod.PRIV_VIEW, filter_marks=False, order_by=MdxMethod.label)

    def get_method_fields(self, _id):
        return self.sa_session.query(MdxMethodField).filter(MdxMethodField.method_id == _id).order_by(MdxMethodField.order).all()

class ItemResultController(AbstractController):

    def _get_applied_config(self, dimensions, measure_obj):
        d = {}
        for i in range(len(dimensions)):
            if dimensions[i] == 'Measures':
                measures_index = i
            d[dimensions[i]] = str(i)

        return {
            "dimensions": d,

            "axes_%s" % measures_index: "columns",
            "dimension_%s" % measures_index: [measure_obj],

            "column_order": [
                'Measures'
            ],
            "row_order": [
            ],

            "show_table": True,
            "show_slicers": True,
            "hide_empty_fields": True,
            "initialRefresh": False,

        }

    def make_configurable_widget_config(self, datasource_id, cube_name):
        """
        Generates a Configurable Widget config and appliedConfig by parsing the cube xml
        """

        logger.info('make_configurable_widget_config: datasource=%s, cube=%s' % (datasource_id, cube_name))

        assert isinstance(datasource_id, (int, long))
        assert isinstance(cube_name, (str, unicode))

        dc = DataSourceController(self.sa_session, account_id=self.account_id)
        datasource = dc.get_datasource(datasource_id, priv=DataSource.PRIV_USE_IN_RESULT)
        cube = dc.get_datasource_cube(datasource_id, cube_name)

        # We only have to connect to IQube and parse the XML if we havent done so previously.
        if not settings.DEBUG and cube.default_config is not None:
            cfg = simplejson.loads(cube.default_config)
            client = datasource.connect()
            measure = client.get_level_members(cube_name, '[Measures].[MeasuresLevel]')[0]
            dimensions = [d['caption'] for d in cfg['dimensions']]
            return {
                'config': cfg,
                'applied_filters': self._get_applied_config(dimensions, MondrianMemberSerialize(measure))
            }


        pc = ProjectController(self.sa_session, account_id=self.account_id)

        schema = pc.get_schema_string_from_mondrian(datasource.mondrian_project)
        xschema = minidom.parseString(schema)
        xcubes = xschema.getElementsByTagName('Cube')

        xcube = None
        for xcube_ in xcubes:
            if xcube_.getAttribute('name') == cube_name:
                xcube = xcube_
        if xcube is None:
            raise PublicUserException('Cube not found:' + cube_name)

        default_measure = xcube.getAttribute('defaultMeasure')

        config_dimensions = []
        dimensions = []

        xdimensions = xcube.getElementsByTagName('Dimension')
        for xdimension in xdimensions:

            dimension_name = xdimension.getAttribute('name')
            dimension_caption = xdimension.getAttribute('caption') or dimension_name

            dimensions.append(dimension_caption)

            xhierarchies = xdimension.getElementsByTagName('Hierarchy')

            config_hierarchies = []

            for xhierarchy in xhierarchies:

                hierarchy_name = xhierarchy.getAttribute('name')
                hierarchy_caption = xhierarchy.getAttribute('caption') or hierarchy_name or dimension_caption

                first_level_name = None
                has_multiple_levels = False

                # default rpc method
                rpc_method = 'get_dimension_options'

                # check if its a time hierarchy, then override the rpc method
                xTables = xhierarchy.getElementsByTagName('Table')
                for xTable in xTables:
                    if xTable.getAttribute('name') == 'date_dimension':
                        rpc_method = 'get_time_children'

                xlevels = xhierarchy.getElementsByTagName('Level')
                for xlevel in xlevels:
                    if first_level_name is None:
                        first_level_name = xlevel.getAttribute('name')
                    else:
                        has_multiple_levels = True

                if hierarchy_name == '':
                    member_name = "[%s].[%s]" % (dimension_name, first_level_name)
                else:
                    member_name = "[%s.%s].[%s]" % (dimension_name, hierarchy_name, first_level_name)

                config_hierarchies.append({
                    "unique_name": member_name,
                    "caption": hierarchy_caption,
                    "has_all_member_available": False,
                    "has_children_available": (not rpc_method == 'get_time_children'),
                    "show_members": True,
                    "method_argument": member_name,
                    "method": "get_level_members",
                    "has_drill_enabled": has_multiple_levels,
                })

            config_dimension = {
                "caption": dimension_caption,
                "available_on_columns": True,
                "available_on_rows": True,
                "available_as_slicer": True,
                "has_select_as_group": True,
                "show_named_sets": False,
                "available_as_split_by": False,
                "required": False,
                "fancy_select_optionsRpcUrl": 'iqportal.iqube.core.IQubeRPCHandler',
                "fancy_select_module": "level_filter_fancy_select",
                "fancy_select_RpcMethod": rpc_method,
                "show_calculated_members": True,
                "has_result_full_headers": True,
                "children": config_hierarchies
            }
            config_dimensions.append(config_dimension)


        xmeasures = xcube.getElementsByTagName('Measure')
        for xmeasure in xmeasures:
            if not default_measure:
                default_measure = xmeasure.getAttribute('name')

        config_dimensions.append({
            "caption": "Measures",
            "available_on_rows": True,
            "show_named_sets": False,
            "has_select_as_group": False,
            "available_as_split_by": False,
            "required": False,
            "available_as_slicer": True,
            "fancy_select_module": "iq_fancyselect",
            "available_on_columns": True,
            "show_calculated_members": True,
            "children": [{
                "unique_name": "[Measures].[MeasuresLevel]",
                "has_children_available": True,
                "caption": "Measures",
                "has_all_member_available": False,
                "show_members": True,
                "method_argument": "[Measures].[MeasuresLevel]",
                "method": "get_level_members"
            }]
        })
        dimensions.append('Measures')

        config = {
            "reporter": "iqportal.iqube.core.ConfigurableWidgetReporter",
            "has_quick_filter": True,
            "show_hide_empty": True,
            "transposable": True,
            "datasource_id": datasource_id,
            "label": 'Results',
            "has_calculated_member": True,
            "show_excel_export": True,
            "cube": cube_name,
            "has_graph": True,
            "table_object_name": "datatablecube",
            "table_module": "iqreport_graph_datatablecube",
            "has_table": True,
            "has_hierarchy_names": True,
            "graph_config": {},
            "has_drill_position": False,
            "has_drill_replace": False,
            "has_drill_through": False,
            "dimensions": config_dimensions,
        }

        measure_uname = "[Measures].[%s]" % default_measure

        acw = AbstractConfigurableWidget(self.sa_session, datasource=datasource, cube=cube_name)
        measure_obj = simplejson.dumps(acw.get_modified_filter(measure_uname), ensure_ascii=False)

        return {'config': config, 'applied_filters': self._get_applied_config(dimensions, measure_obj)}
