# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from django.utils.translation import ugettext as _

from sqlalchemy.orm import aliased
from sqlalchemy.sql.expression import func

from basesoft.web.core.util.custom_query import CustomQuery, CustomQueryField, CustomQueryModel
from iqportal.iqube.core.models import DataSource, DataSourceWidget, DataSourceCube,\
    MdxCalculatedMember, MdxDataSourceCalculatedMember

class DataSourceCustomQuery(CustomQuery):
    permissions = ('view_datasource',)

    def tables(self):
        return (
            CustomQueryModel(DataSource),
            CustomQueryModel(MdxDataSourceCalculatedMember, onclause=DataSource.id==MdxDataSourceCalculatedMember.datasource_id, priv=None, use_outerjoin=True),
            CustomQueryModel(DataSourceWidget, onclause=DataSource.id==DataSourceWidget.datasource_id, priv=None, use_outerjoin=True),
            CustomQueryModel(DataSourceCube, onclause=DataSource.id==DataSourceCube.datasource_id, use_outerjoin=True),
        )

    def fields(self):
        return (
            CustomQueryField(DataSource.id, _('ID'), key='id'),
            CustomQueryField(DataSource.name, _('Name'), key='name', is_filter=True),
            CustomQueryField(DataSource.type, _('Type'), key='type', is_filter=True,
                filter_options=[{'value': DataSource.DATASOURCE_TYPE_DATABASE,
                                 'label': DataSource.type_i18n[DataSource.DATASOURCE_TYPE_DATABASE]
                                },
                                {'value': DataSource.DATASOURCE_TYPE_FILE,
                                 'label': DataSource.type_i18n[DataSource.DATASOURCE_TYPE_FILE]
                                }]
            ),
            CustomQueryField(func.count(DataSourceCube.id.distinct()), _('Cubes'), key='cubes'),
            CustomQueryField(func.count(DataSourceWidget.id.distinct()), _('Available widgets'), key='widgets'),
            CustomQueryField(func.count(MdxDataSourceCalculatedMember.id.distinct()), _('Calculated members'), key='calculated_members'),
            CustomQueryField(DataSource.id, _('Tables'), key='objects'),
            CustomQueryField(DataSource.id, _('Schema'), key='schema'),
            CustomQueryField(DataSource.has_stale_definitions, _('Stale'), key='has_stale_definitions'),
            CustomQueryField(DataSource.id, _('Cache'), key='cache'),
            CustomQueryField(DataSource.created, _('Created')),
            CustomQueryField(DataSource.last_updated, _('Last updated')),
        )

    group_by = (DataSource.id, )

class DataSourceCubeCustomQuery(CustomQuery):
    permissions = ('datasource_view_cube',)

    def tables(self):
        return (
            CustomQueryModel(DataSourceCube),
            CustomQueryModel(DataSource),
        )

    def fields(self):
        return (
            CustomQueryField(DataSourceCube.id, _('ID'), key="id"),
            CustomQueryField(DataSourceCube.name, _('Name'), key="name"),
            CustomQueryField(DataSource.id, _('Datasource'), key='datasource_id'),
            CustomQueryField(DataSource.type, _('Datasource Type'), key='datasource_type'),
        )

class ConfigurableWidgetCustomQuery(CustomQuery):
    permissions = ('view_datasource_widget',)

    def tables(self):
        return (
            CustomQueryModel(DataSourceWidget),
            CustomQueryModel(DataSource, priv=None),
        )

    def fields(self):
        return (
            CustomQueryField(DataSourceWidget.id, _('ID'), key='id'),
            CustomQueryField(DataSourceWidget.label, _('Label'), key='label'),
            CustomQueryField(DataSourceWidget.category, _('Category'), key='category'),
            CustomQueryField(DataSourceWidget.id, _('Default settings'), key='default_settings'),
            CustomQueryField(DataSourceWidget.id, _('Sync. Source'), key='sync_source'),
            CustomQueryField(DataSourceWidget.availability, _('Availability'), key='availability'),
            CustomQueryField(DataSourceWidget.cube, _('Cube'), key='cube'),
            CustomQueryField(DataSource.name, _('Datasource'), key='datasource'),
            CustomQueryField(DataSourceWidget.datasource_id, _('Datasource ID'), key='datasource_id'),
        )

MCM = aliased(MdxCalculatedMember)

class MdxCalculatedMemberCustomQuery(CustomQuery):
    permissions = ('view_calculated_member', )

    def tables(self):
        return (
            CustomQueryModel(MCM, with_polymorphic='CalculatedMember'),
            CustomQueryModel(MdxDataSourceCalculatedMember, priv=None, use_outerjoin=True, onclause=MCM.id==MdxDataSourceCalculatedMember.calculated_member_id),
            CustomQueryModel(DataSource, priv=None, use_outerjoin=True, onclause=MdxDataSourceCalculatedMember.datasource_id==DataSource.id),
        )

    def fields(self):
        return (
            CustomQueryField(MCM.id, _('ID'), key='id'),
            CustomQueryField(MCM.label, _('Label'), key='label', is_filter=True),

            CustomQueryField(DataSource.id, _('DataSource'), key='datasource_id', in_select=False, is_filter=True, filter_options=DataSource.name),
            CustomQueryField(DataSource.name, _('Datasource'), key='datasource'),
            CustomQueryField(MdxDataSourceCalculatedMember.cube, _('Cube'), key='cube', is_filter=True),
        )

custom_queries = (
    DataSourceCustomQuery,
    DataSourceCubeCustomQuery,
    ConfigurableWidgetCustomQuery,
    MdxCalculatedMemberCustomQuery,
)
