# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.
import os
import datetime

from django.conf import settings
from django.utils import simplejson
from django.utils.translation import ugettext as _

from simplejson.decoder import JSONDecodeError

from basesoft.web.core.exceptions import PublicUserException
from basesoft.util.ordered_dict import OrderedDict
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.input.validation import plugins

from iqportal.iqube.core.controllers import DataSourceController, ConfigurableWidgetController, MdxMethodController,\
    ItemResultController, MondrianMemberSerialize, IQubeController
from iqportal.iqube.core.reporters import ReporterRpcHandler
from iqportal.iqube.core.models import DataSourceWidget, DataSource
from iqportal.iqube.xmla.handler import XmlaException
from iqportal.iqreport.core.controllers import GraphPresetController
from iqportal.iqube.compose.controllers import MdxCalculatedMemberController
from iqportal.iqube.core import get_properties_map

class IQubeRPCHandler:
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']
        self.dsc = DataSourceController(self.sa_session, account_id=self.account_id)

    def get_last_aggregation_update(self, datasource_id, cube):
        return self.dsc.get_last_aggregation_update(datasource_id, cube)

    def get_datasources(self, ds_type=None, restrict=False):
        """
        Returns dictionary with id and name for all datasources the user has rights on.
        """
        priv = DataSource.PRIV_VIEW
        if restrict:
            priv = DataSource.PRIV_USE_IN_RESULT
        return [{'value': d.id, 'label': d.name} for d in self.dsc.get_datasources(None, ds_type, priv=priv)]

    def get_calculated_member_hook(self, datasource_id):
        return self.dsc.get_calculated_member_hook(datasource_id)

    def get_datasource_cubes(self, datasource_id):
        if datasource_id is None:
            return []
        ds = self.dsc.get_datasource(datasource_id)

        cubes = [c.name for c in self.dsc.get_cubes(ds)]

        client = ds.connect(account_id=self.account_id)
        rv = [{'name': c['cube_name'], 'description': c['description'], 'type': c['cube_type']}
                for c in client.get_cubes() if c['cube_name'] in cubes]
        if len(rv):
            rv[0]['default'] = True
        return rv

    def get_dimensions(self, datasource_id, cube):
        if datasource_id is None or cube is None:
            return []

        ds = self.dsc.get_datasource(datasource_id)
        client = ds.connect(self.sa_session, self.account_id)
        dims = client.get_dimensions(cube)

        rv = []
        for dim in dims:
            x = client.get_dimension_levels(cube, dim['dimension_unique_name'])
            l = x.pop(0)
            if l['level_type'] == '1':
                l = x.pop(0)
            rv.append({'label': dim['dimension_caption'], 'value': l['level_unique_name']})
        return sorted(rv, key=lambda k: k['label'])

    def get_time_dimensions(self, datasource_id, cube):
        if datasource_id is None or cube is None:
            return []

        ds = self.dsc.get_datasource(datasource_id)
        client = ds.connect(self.sa_session, self.account_id)
        dims = client.get_dimensions_by_type(cube, client.MD_DIMTYPE_TIME)
        rv = []
        for dim in dims:
            x = client.get_dimension_levels(cube, dim['dimension_unique_name'])
            l = x.pop(0)
            if l['level_type'] == '1':
                l = x.pop(0)
            rv.append({'label': dim['dimension_caption'], 'value': l['level_unique_name']})
        return sorted(rv, key=lambda k: k['label'])

    def get_dimension_levels(self, datasource_id, cube, level_unique_name):
        if datasource_id is None or cube is None:
            return []

        ds = self.dsc.get_datasource(datasource_id)
        client = ds.connect(self.sa_session, self.account_id)
        dimension = client.get_level_info(cube, level_unique_name)
        rv = []

        dim_levels = client.get_dimension_levels(cube, dimension[0]['dimension_unique_name'])
        for l in dim_levels:
            rv.append({'label': l['level_caption'], 'value': l['level_unique_name']})

        return rv

    def get_dimensions_and_measures(self, datasource_id, cube):
        if datasource_id is None or cube is None:
            return []

        ds = self.dsc.get_datasource(datasource_id)
        client = ds.connect(self.sa_session, self.account_id)
        rv = []
        for i in client.get_hierarchies(cube):
            rv.append((i['hierarchy_name'], client.get_hierarchy_levels(cube, i['hierarchy_unique_name']),))
        return OrderedDict(rv)

    def get_default_measure(self, datasource_id, cube):
        ds = self.dsc.get_datasource(datasource_id)
        client = ds.connect(self.sa_session, self.account_id)

        try:
            measures = client.get_measures(cube)
        except XmlaException, e:

            # workaround for get_measure bug
            if (e.message.find('The Mondrian XML: mondrian.olap4j.MondrianOlap4jMember cannot be cast to org.olap4j.metadata.Measure') != -1):
                return client.get_level_members(cube, '[Measures].[MeasuresLevel]')[0]

        if len(measures) == 0:
            raise PublicUserException(_('Cube must have at least one measure'))

        return client.get_member_tree_op(cube, measures[0]['measure_unique_name'], client.MDTREEOP_SELF)[0]

    def get_measures(self, datasource_id, cube, unique_name_only=False):
        ds = self.dsc.get_datasource(datasource_id)
        client = ds.connect(self.sa_session, self.account_id)
        rv = []
        for m in client.get_level_members(cube, '[Measures].[MeasuresLevel]'):
            rv.append({'label': m['member_caption'], 'value': m['member_unique_name'] if unique_name_only else MondrianMemberSerialize(m)})
        return rv

    def get_simple_dimension_options(self, datasource, cube, config, values=None):
        ds = self.dsc.get_datasource(datasource)
        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.dsc.get_account().id)
        return cwc.get_simple_dimension_options(ds, cube, config, values)

    def get_dimension_options(self, datasource, cube, config, values):
        ds = self.dsc.get_datasource(datasource)
        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.dsc.get_account().id)
        return cwc.get_dimension_options(ds, cube, config, values)

    def get_filtered_options(self, datasource, cube, config, filter, selected=None):
        # returns the same as the dimension options, but filtered on the selected parent nodes.
        ds = self.dsc.get_datasource(datasource)
        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.dsc.get_account().id)
        return cwc.get_level_descendants(ds, cube, config, filter, selected)

    def get_level_members(self, datasource, cube, level_unique_name):
        ds = self.dsc.get_datasource(datasource)
        client = ds.connect(self.sa_session, self.account_id)
        return client.get_level_members(cube, level_unique_name)

    def get_member_children(self, datasource, cube, member_unique_name):
        ds = self.dsc.get_datasource(datasource)
        client = ds.connect(self.sa_session, self.account_id)
        return client.get_member_tree_op(cube, member_unique_name, client.MDTREEOP_CHILDREN)

    def get_member_children_or_self(self, datasource, cube, member_unique_name):
        """
        used for drilldown. if the member doesnt have children it only returns itself
        """
        ds = self.dsc.get_datasource(datasource)
        client = ds.connect(self.sa_session, self.account_id)
        children = client.get_member_tree_op(cube, member_unique_name, client.MDTREEOP_CHILDREN)
        if len(children) == 0:
            return client.get_member_tree_op(cube, member_unique_name, client.MDTREEOP_SELF)
        else:
            return children

    def get_member_parent(self, datasource, cube, member_unique_name):
        ds = self.dsc.get_datasource(datasource)
        client = ds.connect(self.sa_session, self.account_id)
        return client.get_member_tree_op(cube, member_unique_name, client.MDTREEOP_PARENT)

    def get_member_tree_op(self, datasource, cube, members):
        if type(members) != list:
            members = [members]
        rv = []

        ds = self.dsc.get_datasource(datasource)
        client = ds.connect(self.sa_session, self.account_id)

        for m in members:
            # only one entry
            rv.append(client.get_member_tree_op(cube, '%s.[%s]' % (m['level_unique_name'], m['member_name']), client.MDTREEOP_SELF).pop())
        return rv

    def get_member_descendants(self, datasource, cube, member_unique_name):
        ds = self.dsc.get_datasource(datasource)
        client = ds.connect(self.sa_session, self.account_id)
        return client.get_member_tree_op(cube, member_unique_name, client.MDTREEOP_DESCENDANTS)

    def get_configurable_widget_applied_config(self, config, applied_config):
        ds = self.dsc.get_datasource(config['datasource_id'])
        reporter = ReporterRpcHandler(self.sa_session, ds, config, account_id=self.dsc.get_account().id)
        return reporter.get_mdx_config(applied_config)

    def save_configurable_widget(self, datasource, cube, label, category, as_dashboard_widget, as_report_item, config, default_filters, privileges=None):
        ds = self.dsc.get_datasource(datasource)
        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.dsc.get_account().id)

        availability = 0
        if as_dashboard_widget:
            availability += DataSourceWidget.AVAILABLE_AS_DASHBOARD_WIDGET
        if as_report_item:
            availability += DataSourceWidget.AVAILABLE_AS_REPORT_ITEM

        widget = cwc.save_or_update_configurable_widget(None, ds, cube, label, category, config, default_filters, privileges, availability=availability)
        return widget.id

    def update_configurable_widget_default_filters(self, id, default_filters):
        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.dsc.get_account().id)
        widget = cwc.update_configurable_widget_default_filters(id, default_filters)
        return widget.id

    def get_configurable_widget(self, id):
        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.dsc.get_account().id)
        config, applied_config = cwc.get_state(id)
        return {'config': config, 'applied_config': applied_config}

    def get_configurable_widget_source(self, id):

        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.dsc.get_account().id)
        label, source = cwc.get_sync_source_definition(id)

        return {'label': label, 'source': source}

    def copy_widget(self, id, label, category, applied_filters):
        cwc = ConfigurableWidgetController(self.sa_session, account_id=self.dsc.get_account().id)
        widget = cwc.copy_widget(id, label, category, applied_filters)
        return widget.id

    def get_named_sets(self, datasource, cube, dimension):
        return []

    def get_cm_composition(self, calculation, datasource=None, cube=None):

        redefine = True
        try:
            cm = simplejson.loads(calculation)
            mcmc = MdxCalculatedMemberController(self.sa_session, account_id=self.account_id)
            if type(cm) in (int, ):
                cm = mcmc.auto_configure_mdx_function(cm)
            if type(cm) in (unicode, str, ):
                parse = cm[7:]
            else:
                if not cm.has_key('calculated_member_id'):
                    return None
                formula, _type, dimension = mcmc.get_parsed(cm['calculated_member_id'], inputs=cm['replacements'], include_type=True)
                cm['settings']['format_string'] = None
                name, parse, hierarchy = mcmc.get_with_statement('CM', _type, formula, dimension, cm['settings'], include_type=False)
                redefine = False
                parse = parse[0] # TODO, make this work for multi-layered Calculations (Repeaters, Axis)
        except JSONDecodeError:
            parse = calculation[7:]

        client = None

        from iqportal.iqube.xmla.mdx import MdxParser
        from pyparsing import ParseException

        if datasource is not None:
            ds = self.dsc.get_datasource(datasource)
            client = ds.connect(self.sa_session, self.account_id)

        try:
            return MdxParser().get_dict(parse, client, cube), parse, redefine
        except ParseException:
            return None

    def get_raw_result(self, datasource, mdx, render_table=False):
        ds = self.dsc.get_datasource_by('name', datasource).pop()
        client = ds.connect(account_id=self.account_id)
        response = client.execute_mdx(mdx, raw=True)
        table_data = False
        if render_table:
            table_data = client.execute_mdx(mdx).get_dict()

        return response, table_data

    def write_excel(self, data):
        import xlwt
        import random
        from hashlib import sha256 as sha

        workbook = xlwt.Workbook(encoding="UTF-8")
        sheet = workbook.add_sheet('data')

        if len(data[0]) > 255:
            raise PublicUserException(_('Too many columns provided to make an Excel file.'))

        for y, row in enumerate(data):
            for x, col in enumerate(row):
                sheet.write(y, x, col)

        hash = sha(str(random.random())).hexdigest()[:8]
        workbook.save(os.path.join(settings.TMP_FILE_REPOSITORY, 'data_%s.xls' % hash))

        return hash

    def save_preset(self, name, graph_lib, config):
        gpc = GraphPresetController(self.sa_session, account_id=self.account_id)
        gp = gpc.save_or_update(None, name, graph_lib, config)
        return gp.id

    def load_preset(self, id):
        gpc = GraphPresetController(self.sa_session, account_id=self.account_id)
        gp = gpc.get(id=id).pop()
        return {'id': gp.id, 'config': gp.config}

    def get_presets(self, graph_lib=None):
        gpc = GraphPresetController(self.sa_session, account_id=self.account_id)
        return [{'value': gp.id, 'label': gp.name} for gp in gpc.get(graph_lib=graph_lib)]

    def get_mdx_method(self, _id, include_fields=False):
        mmc = MdxMethodController(self.sa_session, account_id=self.account_id)
        m = mmc.get_mdx_method(_id).pop()
        return {
            'method_id': m.id,
            'name': m.name,
            'label': m.label,
            'return_type': m.return_type,
            'category': m.category,
            'example': m.example,
            'fields': None if not include_fields else self.get_method_fields(m.id)
        }

    def get_mdx_methods(self, include_fields=False):
        mmc = MdxMethodController(self.sa_session, account_id=self.account_id)
        return [{'id': m.id,
                 'name': m.name,
                 'label': m.label,
                 'return_type': m.return_type,
                 'category': m.category,
                 'example': m.example,
                 'fields': None if not include_fields else self.get_method_fields(m.id)
                 } for m in mmc.get_mdx_methods()]

    def get_method_fields(self, _id):
        mmc = MdxMethodController(self.sa_session, account_id=self.account_id)
        return [{'method_id': f.method_id,
                 'field_type': f.field_type,
                 'required_field': f.required_field
                 } for f in mmc.get_method_fields(_id)]

    def get_mdx_properties(self, category):
        _map = get_properties_map()
        if not _map.has_key(category):
            raise PublicUserException(_('Unsupported property category supplied'))

        return [{'label': p.label, 'value': p.name} for p in _map[category].values()]

    def get_members_captions(self, datasource_id, cube, members_str, with_ancestors=False):
        qc = IQubeController()
        return qc.get_members_captions(datasource_id, cube, members_str, with_ancestors)

    def get_time_children(self, datasource_id, cube, config, values):
        """
        """

        # number of years to show extra in case no end_date is specified
        EXTRA_YEARS = 4

        min_year = 2012
        max_year = datetime.datetime.now().year + EXTRA_YEARS

        # use the original get_dimension_options() to get the result. but then filter per group for min/max years
        dsc = DataSourceController(self.sa_session, account_id=self.account_id)
        ds = dsc.get_datasource_by('id', datasource_id).pop()

        cwc = ConfigurableWidgetController(self.sa_session, account_id=dsc.get_account().id)
        grouped_rv = cwc.get_dimension_options(ds, cube, config, values)

        # group_name can be year, quarter, week, etc.. check the first member and see
        # if its part of the year level. then limit those years to the ones for this survey
        for group_name, members in grouped_rv.iteritems():

            if len(members) > 0:
                member_0 = simplejson.loads(members[0]['value'])
                if member_0['level_unique_name'].endswith('.[Year]'):

                    new_members = []
                    for member_str in members:

                        year = int( simplejson.loads(member_str['value'])['member_name'])
                        if year >= min_year and year <= max_year:
                            new_members.append(member_str)

                    grouped_rv[group_name] = new_members

        return grouped_rv

    def get_serialized(self, member):
        return MondrianMemberSerialize(member)

    def validate_ds_method_page(self, args):
        required = {}
        for key in args.keys():
            if key == 'format_string':
                continue
            required[key] = [(plugins.Required, {})]

        form_handler = FormValidator(required)
        return form_handler.validate(args)

class ItemResultHandler:

    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']

    def make_configurable_widget_config(self, datasource_id, cube_name):
        c = ItemResultController(self.sa_session, account_id=self.account_id)
        return c.make_configurable_widget_config(datasource_id, cube_name)


handlers = (IQubeRPCHandler, ItemResultHandler, )
