# Copyright (c) 2012 BaseSoft B.V.
# See LICENSE for details.

from django.conf import settings
from django.utils.translation import ugettext as _

import os, re

from basesoft.logging import getLogger
from iqportal.iqube.core.models import DataSource
from basesoft.web.core.exceptions import PublicUserException, SecurityException
from iqportal.iqube.core.controllers import DataSourceController,\
    ProjectController
from iqportal.iqube.optiq.controllers import OptiqCSVController, make_valid_name
from iqportal.iqube.optiq.importers import AbstractImporter
logger = getLogger('iqportal.iqube.optiq')


class SchemaRPCHandler:
    def __init__(self, request, **kwargs):
        self.sa_session = request.Session()
        self.account_id = request.session['user_id']

    def edit_local_table(self, datasource_id, local_table_name, truncate, tmp_file):
        """
        refreshes or appends data to the given table
        """
        assert isinstance(datasource_id, (int, long))

        if len(tmp_file) != 1:
            raise PublicUserException(_('Upload one file at the time'))
        tmp_file = tmp_file[0]

        ds_controller = DataSourceController(self.sa_session, account_id=self.account_id)
        datasource = ds_controller.get_datasource_by('id', datasource_id)[0]


    def pre_import(self, tmp_files, import_options, datasource_name):
        """
        receives an uploaded file and datasource name from autoselect (either new or existing),
        returns information about the guessed columns
        """
        if len(tmp_files) != 1:
            raise PublicUserException(_('Upload one file at the time'))
        tmp_file = tmp_files[0]

        ds_controller = ProjectController(self.sa_session, account_id=self.account_id)

        controller = OptiqCSVController(self.sa_session, account_id=self.account_id)

        # try to get datasource by name
        datasources = ds_controller.get_datasource_by('name', datasource_name, ds_type = DataSource.DATASOURCE_TYPE_OPTIQ)
        if len(datasources) == 0:
            datasource = controller.create_datasource(datasource_name)

        elif len(datasources) == 1:
            datasource = datasources.pop()
        else:
            raise Exception('too many datasources')

        file_name, file_extension = os.path.splitext(tmp_file.file_name)

        importer = AbstractImporter.get_importer(tmp_file)

        columns = importer.get_file_columns()

        table_name = make_valid_name(file_name.strip())
        rows = []

        rv = {
            'columns': columns,
            'rows': rows,
            'datasource_id': datasource.id,
            'skip_csv_properties': file_extension in ('.xls', '.xlsx'),
            'table_name': table_name
        }

        return rv


    def create_table(self, datasource_id, file_, table_name):
        """
        called by the schema wizard to import a new table with the given column definition
        it could be that the table already exists but columns have been changed
        """

        # create importer for file
        importer = AbstractImporter.get_importer(file_[0])

        # detect get columns
        columns = importer.get_file_columns()

        # make a datasource
        ds_controller = ProjectController(self.sa_session, account_id=self.account_id)
        datasource = ds_controller.get_datasource_by('id', datasource_id, ds_type = DataSource.DATASOURCE_TYPE_OPTIQ)[0]

        # import as a csv file
        controller = OptiqCSVController(self.sa_session, account_id=self.account_id)
        controller.do_import(datasource, make_valid_name(table_name), importer, columns)

    def build_cube(self, datasource_id, cube_xml, name):
        """
        reveives a single cube xml block which will be written as a schema
        """

        controller = ProjectController(self.sa_session, account_id=self.account_id)
        ds = controller.get_datasources(datasource_id)
        if len(ds) == 0:
            raise SecurityException('Invalid datasource id %s' % datasource_id)
        ds = ds[0]

        controller.save_or_update_cube(None, name, ds)

        cube_xml = '<Schema name="%s">%s</Schema>' % (ds.name, cube_xml)

        controller.write_schema(ds.mondrian_project, cube_xml)



    def get_column_types(self):
        """
        """

        return [
            {'text': 'Varchar', 'value': 'VARCHAR', 'PRECISION': 1024},
            #{'text': 'Boolean', 'value': 'BOOLEAN'},
            {'text': 'Tiny int (<128)', 'value': 'TINYINT'},
            {'text': 'Small int (<65k)', 'value': 'SMALLINT'},
            {'text': 'Integer (<2G)', 'value': 'INTEGER'},
            {'text': 'Big Integer', 'value': 'BIGINT'},
            {'text': 'Decimal', 'value': 'DECIMAL', 'PRECISION': 12, 'DEC_DIGITS': 4},
            {'text': 'Date', 'value': 'DATE'},
            {'text': 'Time', 'value': 'TIME'},
        ]

    def get_tables_state(self, datasource_id):
        """
        returns the 'tables state' required by the cube designer for the given datasource_id
        """

        controller = OptiqCSVController(self.sa_session, account_id=self.account_id)
        return controller.get_tables_state(datasource_id)

    def get_cube_xml(self, datasource_id, cube_name):
        """
        extracts the given cube_name from the schema for the given datasource_id and cube_name
        """

        ds_controller = ProjectController(self.sa_session, account_id=self.account_id)
        datasource = ds_controller.get_datasources(datasource_id)[0]

        return ds_controller.get_mondrian_cube_xml(datasource.mondrian_project, cube_name)


    def save_cube_xml(self, datasource_id, cube_name, cube_xml):

        ds_controller = ProjectController(self.sa_session, account_id=self.account_id)
        datasource = ds_controller.get_datasources(datasource_id)[0]

	# for now only optiq cubes can be edited
        assert datasource.type == DataSource.DATASOURCE_TYPE_OPTIQ

        ds_controller.save_mondrian_cube_xml(datasource, cube_name, cube_xml)




handlers = (SchemaRPCHandler, )
