# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

from basesoft.web.core.exceptions import PublicUserException

class MdxProvider:
    COLUMN = 1
    ROW = 2
    SLICER = 4

    def _modify_hierarchies(self, hierarchies, union, order, axis_type):
        pass

    def process_crossjoin(self, axis, axis_type, hierarchize, process_union=False):

        if len(axis) == 0:
            return ''
        order = []
        hierarchies = {}
        rv = []
        union = []

        for member in axis:
            if type(member) == list and process_union:
                union.append(self._process(member, axis_type, is_union=True))
                continue
            # check for depth
            hname = member['hierarchy_unique_name']
            if hname is None:
                hname = member['member_unique_name']
            if member.has_key('depth'):
                depth = member['depth']
            else:
                depth = 0
            # if we have disabled hierarchizing, set the depth hard to 0 for all.
            if not hierarchize:
                depth = 0
            if not hierarchies.has_key(hname):
                hierarchies[hname] = {}
                order.append(hname)
            if not hierarchies[hname].has_key('depth_%s' % depth):
                hierarchies[hname]['depth_%s' % depth] = []

            hierarchies[hname]['depth_%s' % depth].append(member['member_unique_name'])

        self._modify_hierarchies(hierarchies, union, order, axis_type)

        for hname in order:
            depths = hierarchies[hname]
            hier = []
            for depth, value in depths.iteritems():
                if axis_type == self.SLICER and len(value) > 1:
                    hier.append(self._make_calculated_member(hname, value))
                else:
                    hier.append(','.join(value))

            hierstr = ''
            # don't hierarchize on slicers, that doesn't make sense.
            if axis_type != self.SLICER and len(hier) > 1:
                hierstr = 'HIERARCHIZE({%s})' % ','.join(hier)
            else:
                hierstr = '{%s}' % ','.join(hier)
            rv.append(hierstr)

        result = ''
        # multiple hierarchies means Crossjoining
        for hier in rv:
            if result == '':
                result = '%s' % hier
                continue
            result = 'CROSSJOIN(%s, %s)' % (result, hier)
        return result, union

class MdxBuilder(MdxProvider):
    """
    Mdx Builder is used to create MDX Select-statements from provided member information.
    """


    def __init__(self, cube, columns, rows, slicers, nonempty=False, sort=None, is_graph=False, account_id=None, hierarchize=True):
        self.cube = cube
        self.columns = columns
        self.rows = rows
        self.slicers = slicers
        self.nonempty = nonempty
        self.sort = sort
        self.with_ = None
        self.is_graph = is_graph
        self.account_id = account_id
        self.allow_empty_rows = False
        self.hierarchize = hierarchize

    def set_calculated_members(self, calculated_members):
        print 'MdxBuilder: set_calculated_members is deprecated, use set_with()'
        self.set_with(calculated_members)

    def set_with(self, with_):
        self.with_ = with_

    def _make_calculated_member(self, hierarchy_name, values):
        if self.with_ is None:
            self.with_ = []

        label = '+'.join([i[i.rfind('[')+1:-1] for i in values])
        name = '%s.[%s]' % (hierarchy_name, label)

        self.with_.append(" MEMBER %s AS 'AGGREGATE({%s})'" % (name, ', '.join(values)))
        return name

    def _process_with(self, columns, rows):
        if self.with_ is None or len(self.with_) == 0:
            return '', columns, rows

        regex = r"SET \[Axis(?:0|1)\] AS 'Axis\((0|1)\)'"
        import re

        # This code is here for the IQPortal implementation of the Axis() method.
        # Can be removed when mondrian has Axis() implemented
        rv = []
        for with_ in self.with_:
            # verify axis
            matches = re.match(regex, with_)
            if matches is None:
                rv.append(unicode(with_))
                continue

            axis = matches.group(1)
            if int(axis) == self.COLUMN-1:
                direction = columns
            else:
                direction = rows

            nonempty = self.nonempty == True or (self.nonempty & (axis+1) == (axis+1))
            a = "SET [Axis%s] AS '%s'" % (axis, direction[10:] if nonempty else direction)
            if int(axis) == self.COLUMN-1:
                columns = '%s[Axis%s]' % ('NON EMPTY ' if nonempty else '', axis)
            else:
                rows = '%s[Axis%s]' % ('NON EMPTY ' if nonempty else '', axis)
            rv.append(unicode(a))

        return 'WITH %s' % '\n '.join(rv), columns, rows

    def _process(self, axis, axis_type, is_union=False):
        """
        Check the members on the axes and parse accordingly.
        """
        result, union = self.process_crossjoin(axis, axis_type, self.hierarchize, process_union=True)

        if axis_type == self.ROW and self.sort is not None:
            member, direction = self.sort
            result = 'ORDER(%s, %s, %s)' % (result, member['member_unique_name'], direction)

        if len(union) > 0:
            for u in union:
                if result == '':
                    result = '%s' % u
                    continue
                result = 'UNION(%s, %s)' % (result, u)
            result = 'HIERARCHIZE(%s)' % result

        if not is_union and axis_type != self.SLICER and self.nonempty != False:
            if self.nonempty == True or (self.nonempty & axis_type == axis_type):
                result = 'NON EMPTY %s' % result

        return result

    def _process_where(self):
        """
        Process slicers.
        """
        if self.slicers is None or len(self.slicers) == 0:
            return ''

        self.slicers = sorted(self.slicers, key=lambda k: k['member_unique_name'])

        return 'WHERE %s' % self._process(self.slicers, self.SLICER)

    def get_result(self):
        if not self.allow_empty_rows:
            if len(self.columns) == 0 or len(self.rows) == 0:
                raise PublicUserException('At least one row and one column item are required')
        elif len(self.columns) == 0:
                raise PublicUserException('At least one column item is required')

        columns = self._process(self.columns, self.COLUMN)

        rows = None
        if self.rows is not None and len(self.rows) > 0:
            rows = self._process(self.rows, self.ROW)

        slicers = self._process_where()
        with_statement, columns, rows = self._process_with(columns, rows)

        mdx = """
            %s
            SELECT %s ON COLUMNS
                %s
            FROM [%s]
            %s""" % (with_statement, columns, '' if rows is None else ', %s ON ROWS' % rows, self.cube, slicers)

        return mdx

class MdxParser:
    def __init__(self):
        from iqportal.iqube.xmla.grammar import CalculatedMemberGrammar
        self._grammar = CalculatedMemberGrammar()

    def parse(self, member):
        g = self._grammar.get_mdx().parseString(member)
        return g

    def get_dict(self, member, client=None, cube=None):
        import pyparsing

        c = self.parse(member)
        cc = self._grammar.calculator

        es = []
        fs = []

        for i, op in cc.stack.iteritems():

            if len(op[0]) > 0:
                s = cc.get_stack(op[0])
                es.append(s.get_dict(client, cube))
            if len(op[1]) > 0:
                es += [cc.get_stack([o]).get_dict(client, cube) for o in op[1]]

        hierarchy = 'FORMULA'
        if len(c.name) == 1:
            name = c.name[0]
        else:
            hierarchy = c.name[0]
            name = c.name[1]

        rv = {
            'hierarchy': hierarchy,
            'member_name': name,
            'calculation': es + fs
        }

        return rv


