# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.
import re

from basesoft.web.core.exceptions import PublicUserException

from copy import copy
import types

RETURN_TYPE_CDATA = 0
RETURN_TYPE_INT = 1
RETURN_TYPE_FLOAT = 2

class MdxException(PublicUserException):
    """
    When the user supplies wrong MDX which the provider cannot process this is raised.
    """

class MdxProcessor:
    """
    Default processor to parse XML data.
    """
    child = None
    data = None
    child_obj = None
    properties = None

    def __init__(self, attr):
        self.children = []

    def process_child(self, tag, attr):
        child = self.get_child(tag, attr)
        if child is not None:
            self.child = child
            self.children.append(child)

        if self.child is not None:
            self.child.process_child(tag, attr)

    def set_data(self, data):
        if self.child is not None:
            self.child.set_data(data)

    def get_child(self, tag, attr):
        if self.child_obj is not None and self.child_obj.tag == tag:
            return self.child_obj(attr)

    def __repr__(self):
        return '<MdxProcessor %s>' % self.children

class MdxMemberProcessor(MdxProcessor):
    tag = 'Member'
    _caption = False
    _uname = False
    _lname = False
    _lnum = False

    def __init__(self, attr):
        self.hierarchy = attr.getValue((None, 'Hierarchy'))
        self.caption = ''
        self.uname = ''
        self.lname = ''
        self.lnum = ''

    def process_child(self, tag, attr):
        if tag == 'Caption':
            self._caption = True
        if tag == 'UName':
            self._uname = True
        if tag == 'LName':
            self._lname = True
        if tag == 'LNum':
            self._lnum = True

    def set_data(self, data):
        if self._caption:
            self.caption = data
            self._caption = False
        if self._uname:
            self.uname = data
            self._uname = False
        if self._lname:
            self.lname = data
            self._lname = False
        if self._lnum:
            self.lnum = data
            self._lnum = False

    def get_dict(self):
        # ticket #5103: for closure tables, lnum seems to be 1 always. So instead build a lnum based on the number of levels detected in the member
        lnum = self.lnum if self.uname == '' else len(self.uname.split('].[')) - 1
        return {'hierarchy': self.hierarchy, 'caption': self.caption, 'uname': self.uname, 'lname': self.lname, 'lnum': lnum}

    def __repr__(self):
        return '<Member hierarchy="%s" caption="%s" uname="%s">' % (self.hierarchy, self.caption, self.uname)

class MdxTupleProcessor(MdxProcessor):
    tag = 'Tuple'
    child_obj = MdxMemberProcessor

    def get_list(self):
        return [m.get_dict() for m in self.children]

    def __repr__(self):
        return '<Tuple %s>' % self.children

class MdxAxisProcessor(MdxProcessor):
    tag = 'Axis'
    child_obj = MdxTupleProcessor

    def __init__(self, attr):
        MdxProcessor.__init__(self, attr)
        self.name = attr.getValue((None, 'name'))

        self._unique_index = {}

    def get_list(self):
        rv = [t.get_list() for t in self.children]

        for set_ in rv:

            # unique index used for colors

            if not len(set_):
                continue

            uname = set_[-1]['uname']
            if not self._unique_index.has_key(uname):
                self._unique_index[uname] = len(self._unique_index)
            set_[-1]['uindex'] = self._unique_index[uname]


        return rv

    def __repr__(self):
        return '<Axis name="%s" children="%s">' % (self.name, self.children)

class MdxAxesProcessor(MdxProcessor):
    tag = 'Axes'
    child_obj = MdxAxisProcessor

    def __repr__(self):
        return '<Axes %s>' % self.children

class MdxCellProcessor(MdxProcessor):
    tag = 'Cell'
    _value = False
    _formatted_value = False
    _format_string = False

    def __init__(self, attr):
        self.ordinal = attr.getValue((None, 'CellOrdinal'))
        self.value = ''
        self.formatted_value = ''
        self.format_string = ''
        self.type = RETURN_TYPE_CDATA

    def process_child(self, tag, attr):
        if tag == 'Value':
            self._type = attr.getValue(('http://www.w3.org/2001/XMLSchema-instance', 'type'))
            self._value = True
        if tag == 'FmtValue':
            self._formatted_value = True
        if tag == 'FormatString':
            self._format_string = True

    def set_data(self, data):
        if self._value:
            # TODO: this is ugly, cant we use OBJECTIFY from lxml for this? #3258
            if  self._type == 'xsd:int':
                self.value = int(data)
                self.type = RETURN_TYPE_INT
            elif  self._type == 'xsd:double':
                if data == 'Infinity':
                    self.value = '&#8734;' # infinity sign code
                    self.type = RETURN_TYPE_CDATA
                elif data == 'NaN':
                    self.value = 'NaN'
                    self.type = RETURN_TYPE_CDATA
                else:
                    self.value = float(data)
                    self.type = RETURN_TYPE_FLOAT
            else:
                self.value = data
                self.type = RETURN_TYPE_CDATA
            self._value = False
        if self._formatted_value:
            if data == 'Infinity%':
                self.formatted_value = ''
            else:
                self.formatted_value = data
            self._formatted_value = False
        if self._format_string:
            data = data.strip()
            self.format_string = data
            self._format_string = False

    def get_dict(self):
        return {'value': self.value, 'cellOrdinal': self.ordinal, 'formattedValue': self.formatted_value, 'type': self.type, 'formatString': self.format_string}

    def __repr__(self):
        return '<Cell value="%s", formatted_value="%s" ordinal=%s>' % (self.value, self.formatted_value, self.ordinal)

class MdxCellDataProcessor(MdxProcessor):
    tag = 'CellData'
    child_obj = MdxCellProcessor

    def __repr__(self):
        return '<CellData %s>' % self.children

class MdxErrorProcessor(MdxProcessor):
    tag = 'Error'
    _code = False
    _desc = False

    def __init__(self, attr):
        self.code = ''
        self.desc = ''

    def process_child(self, tag, attr):
        if tag == 'code':
            self._code = True
        if tag == 'desc':
            self._desc = True

    def set_data(self, data):
        if self._code:
            self.code = data
            self._code = False
        if self._desc:
            # strip the non-necessary mondrian references from the error.
            self.desc = data.replace('The Mondrian XML: Mondrian Error:', '')
            self._desc = False

    def __repr__(self):
        return '<Error code="%s" desc="%s">' % (self.code, self.desc)

class MdxResultProxy:
    """
    Smorgles the data from the ExecuteHandler as responded by the XMLA provider.
    """
    error = None

    def __init__(self):
        self.result = MondrianMdxResult()

    def get_processor(self, tag, attrib):
        if hasattr(self, 'get_%s_processor' % tag.lower()):
            return getattr(self, 'get_%s_processor' % tag.lower())(attrib)

    def get_axes_processor(self, attrib):
        self.axes = MdxAxesProcessor(attrib)
        return self.axes

    def get_celldata_processor(self, attrib):
        self.celldata = MdxCellDataProcessor(attrib)
        return self.celldata

    def get_error_processor(self, attrib):
        self.error = MdxErrorProcessor(attrib)
        return self.error

    def get_result(self):
        if self.error is not None:
            raise MdxException('%s (code: %s)' % (self.error.desc, self.error.code))
        self.result.process_headers(self.axes)
        self.result.process_data(self.celldata)
        return self.result

    def __repr__(self):
        return '<MdxResultProxy>'

class MondrianMdxResult:
    def process_headers(self, axes):
        self.headers = {'columns': [], 'rows': []}
        self.slicers = []

        for axis in axes.children:
            if axis.name == 'SlicerAxis':
                self.slicers = axis.get_list()
                continue
            if axis.name == 'Axis0':
                self.headers['columns'] = axis.get_list()
                continue
            self.headers['rows'] = axis.get_list()

    def process_full_headers(self, dimensions, ds_id, cube, with_ancestors=True):

        if dimensions is None:
            return

        from iqportal.iqube.core.controllers import IQubeController
        headers = self.get_headers()
        axisCheck = 0
        rows = headers['rows']
        cols = headers['columns']
        member_str = []
        qc = IQubeController()

        for a, i in [(rows, 1), (cols, 2)]:
            for r in a:
                for mem in r:
                    h = '[%s]' % mem.get('hierarchy')
                    if h in dimensions and mem.get('uname') not in member_str:
                        member_str.append(mem.get('uname'))
                        axisCheck |= i

        full_headers = qc.get_members_captions(ds_id, cube, member_str, with_ancestors=True)

        for a, i in [(rows, 1), (cols, 2)]:
            if axisCheck & i != i:
                continue

            for r in a:
                for mem in r:
                    h = '[%s]' % mem.get('hierarchy')
                    if h in dimensions and full_headers.has_key(mem.get('uname')):
                        mem['caption'] = ', '.join(full_headers.get(mem.get('uname')))

    def process_data(self, data):
        self.data = [d.get_dict() for d in data.children]

    def get_headers(self):
        return self.headers

    fetch_headers = get_headers

    def get_slicers(self):
        return self.slicers

    fetch_slicers = get_slicers

    def get_data(self):
        return self.data

    fetch_all = get_data

    def fetch_chart_data(self):
        """
        We need a seperate method when getting the result for a chart, as we need to 'massage'
        percentage values
        """
        data = []
        for i in self.data:
            if isinstance(i["formattedValue"], types.StringTypes) and i["formattedValue"].endswith('%'):
                try:
                    i["value"] = float(i["value"]) * 100
                except:
                    pass
            data.append(i)
        return data

    def get_dict(self):
        return {'columns': self.headers['columns'], 'rows': self.headers['rows'], 'slicers': self.slicers, 'data': self.data}

    def __repr__(self):
        return '<MondrianMdxResult>'

class MdxResult:

    """
    NOTE: not sure if we still require this information as provided like this. Currently this is
    here for backwards compatibility with the previous IQube server software.
    """

    def process_data(self, data):
        rv = []
        self.ordinal_cnt = 0

        def process_col(col, data, rv):
            if not col.has_key('children'):

                key_str = ''.join(col['key'])

                if len(data.children) == 0:
                    rv[key_str] = None
                elif int(data.children[0].ordinal) != self.ordinal_cnt:
                    rv[key_str] = None
                else:
                    rv[key_str] = data.children.pop(0).value
                self.ordinal_cnt += 1
                return rv
            for child in col['children']:
                process_col(child, data, rv)
            return rv

        def process_row(row, data, rv):
            if not row.has_key('children'):

                key_str = ''.join(row['key'])

                r = {'data': {}, 'row_key': key_str}
                for col in self.headers['columns']:
                    process_col(col, data, r['data'])
                rv.append(r)
                return rv
            for child in row['children']:
                process_row(child, data, rv)
            return rv

        for row in self.headers['rows']:
            process_row(row, data, rv)
        self.data = rv

    def process_headers(self, axes):
        self.slicers = []
        rv = {'columns': [], 'rows': []}

        def process_children(children):
            rv = []

            base_indents = [None] * range(len(children[0].children))

            # Every tuple is a new row/column; all members within the tuple are on the same row/column.
            def process_members(children, result, references, key=None, x=None, first=False):
                member = children.pop(0)
                if x is None:
                    x = 0
                else:
                    x = x + 1

                if key is None:
                    key = []

                if base_indents[x] is None:
                    base_indents[x] = int(member.lnum)

                indent = int(member.lnum) - base_indents[x]

                key.append(member.uname)
                key_str = ''.join(key)

                # this doesn't cover all situations
                #if first and len(children)!=0 and references.has_key(key_str):
                if references.has_key(key_str):
                    data = references[key_str]
                else:
                    data = {
                        'dimension': member.hierarchy,
                        'label': member.caption,
                        'key': key_str,
                        'key_list': key,
                        'level': member.lname,
                        'indent': indent,
                        'uname': member.uname
                    }

                    references[key_str] = data
                    result.append(data)

                if len(children) != 0:
                    if not data.has_key('children'):
                        data['children'] = []
                    process_members(children, data['children'], references, copy(key), x)

            references = {}
            for tuple in children:
                if len(tuple.children) == 0:
                    continue
                process_members(tuple.children, rv, references, None, None, True)
            return rv

        for axis in axes.children:
            if axis.name == 'SlicerAxis':
                self.slicers = process_children(axis.children)
                continue
            if axis.name == 'Axis0': # columns
                rv['columns'] = process_children(axis.children)
                continue
            rv['rows'] = process_children(axis.children)

        self.headers = rv

    def get_headers(self):
        return self.headers

    fetch_headers = get_headers

    def get_slicers(self):
        return self.slicers

    fetch_slicers = get_slicers

    def get_data(self):
        return self.data

    fetch_all = get_data

    def __repr__(self):
        return '<MdxResult>'


class MockMdxResult:

    def __init__(self, data):
        self._ordinal = 0
        self._set_data([[i for i in d[1:]] for d in data[1:]])
        self._set_headers(data[0][1:], [d[0] for d in data[1:]])
        self.slicers = None

    def _set_data(self, data):
        d = []

        def make_data(i):
            cell = {'cellOrdinal': '%s' % self._ordinal, 'formattedValue': i, 'type': 2, 'value': i}
            self._ordinal = self._ordinal + 1
            return cell

        for row in data:
            [d.append(make_data(i)) for i in row]

        self.data = d

    def _set_headers(self, columns, rows):

        def make_header(i, axis, uindex):
            uindex[0] = uindex[0]+1
            return [{'hierarchy': axis, 'caption': i, 'uindex' : uindex[0], 'lname': axis, 'lnum': '0', 'uname': '[%s].[%s]' % (axis, i)}]

        ci = [-1]
        ri = [-1]

        self.headers = {
            'columns': [make_header(d, 'columns', ci) for d in columns],
            'rows': [make_header(d, 'rows', ri) for d in rows]
        }

    def get_headers(self):
        return self.headers

    fetch_headers = get_headers

    def get_slicers(self):
        return self.slicers

    fetch_slicers = get_slicers

    def get_data(self):
        return self.data

    fetch_all = get_data
    # We do not 'massage' data here as we do in MondrianMdxResult
    fetch_chart_data = get_data

    def get_dict(self):
        return {'columns': self.headers['columns'], 'rows': self.headers['rows'], 'slicers': self.slicers, 'data': self.data}

    def __repr__(self):
        return '<MockMdxResult>'


