# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

"""
This contains the controllers for survey compose
"""

import re
import copy

from sqlalchemy import and_
from sqlalchemy.orm.exc import NoResultFound
from django.utils.translation import ugettext as _
from django.utils import simplejson
from basesoft.web.core.jsonrpc.dispatcher import JsonRpcJSONEncoder, json_object_hook

from basesoft.web.contrib.compose import controllers as compose_controllers
from basesoft.web.core.exceptions import PublicSecurityException, PublicUserException, PublicSystemException, SecurityException,\
    SystemException
from basesoft.web.core.util.controllers import AbstractController, transaction_decorator, AbstractStateController, StateManager

from basesoft.web.contrib.mailing.controllers import BaseMailingStateController
from basesoft.web.contrib.i18n.models import Locale

from iqportal.iquest.core.models import SurveyPublication, Survey, IQuestTheme, IQuestThemeGroup, SurveyItemQALayoutAttribute,\
    SurveyItemQALocale, SurveyItemQA, IQuestQAFilter
from iqportal.iquest.core import models as iquest_models
from iqportal.iquest.core.mailing import MailController
from iqportal.iquest.core.qa_filters.controllers import SurveyItemQAFilterController
from iqportal.iquest.core.controllers import ItemPoolController, IQuestItemHtmlStateController,\
    IQuestItemQAStateController, IQuestItemElementController

from basesoft import logging
from basesoft.web.contrib.compose.models import ComposePublicationLocale,\
    ComposePublication
from types import NoneType
logger = logging.getLogger("iqportal.iquest.compose")

class ComposeController(AbstractController):
    """
    This should be called from a view
    """

    def __init__(self, session, locale=None, account_id=None):
        AbstractController.__init__(self, session, locale, account_id)
        manager = StateManager(locale, account_id, Survey.MODE_COMPOSE)
        self.state_controller = SurveyStateController(self.sa_session, manager)

    def get_surveys(self, id=None, order_by=None, priv=Survey.PRIV_VIEW):
        """
        Return the surveys the user has access to
        @param id: None, int or list of ints. the id(s) of the surveys we want to return or None if we want all
        @param order_by: the column after which we want to make the ordering, or None we order by name
        @return: a list of survey objects
        """
        self.assert_permissions('view_survey')

        if order_by is None:
            order_by = Survey.name

        return self.get_privileged_object_by_id(Survey, id, priv, order_by=order_by)

    def get_qa_locales(self, survey, locale=None):
        if not isinstance(survey, Survey):
            survey = self.get_surveys(survey).pop()
        if locale is None:
            locale = survey.fallback_locale


        ids = [i.id for i in survey.itemQAs]
        return self.sa_session.query(SurveyItemQALocale).filter(
            and_(SurveyItemQALocale.locale==locale, SurveyItemQALocale.item_id.in_(ids))).all()

    def _mark_as(self, id, mark, priv):
        """
        Overridden method for custom logic and checks.
        """
        surveys = self.get_surveys(id=id, priv=priv)
        if len(surveys) != 1:
            raise PublicSecurityException('Survey can not be marked as deleted')

        survey = surveys.pop()

        # check if there are online publications for survey
        op = [p for p in survey.publications if p.status==iquest_models.SurveyPublication.STATUS_ONLINE]
        if len(op):
            raise PublicUserException(_("Survey '%s' has online publications and cannot be deleted") % survey.name)

        # expire all publications for this survey before marking it as deleted
        for p in survey.publications:
            p.status = iquest_models.SurveyPublication.STATUS_FINISHED

        setattr(survey, mark, True)
        self.sa_session.add(survey)

    @transaction_decorator
    def mark_as_deleted(self, id):
        """
        This will mark the survey as deleted if the user has the correct rights.
        """
        self.assert_permissions('delete_survey')

        self._mark_as(id, 'deleted', Survey.PRIV_DELETE)

    @transaction_decorator
    def mark_as_archived(self, id):
        """
        This will mark the survey as archived, this survey will no longer show up in overviews by default.
        """
        self.assert_permissions('update_survey')

        self._mark_as(id, 'archived', Survey.PRIV_ARCHIVE)

    def get_publications(self, id=None, priv=SurveyPublication.PRIV_VIEW):
        """
        """
        self.assert_permissions('view_survey_publication')
        return self.get_privileged_object_by_id(SurveyPublication, id, priv)

    def get_themes(self, id=None):
        """
        returns the themes to which the current user has permissions to save surveys for
        """
        filter, fo = self.get_query_filter_on_privileges(IQuestTheme, IQuestThemeGroup, IQuestTheme.PRIV_VIEW)

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, IQuestTheme.id.in_(id))

        return self.sa_session.query(IQuestTheme).select_from(fo).filter(filter).all()

    def _check_state(self, new_children, old_children):
        """
        Recursive method that checks if new state children are same as old state children.
        (For when updating online surveys)
        """
        for type in new_children:
            if len(new_children[type]) != len(old_children[type]):
                raise SecurityException('Invalid action for online survey')
            old_ids = [c['dbID'] for c in old_children[type]]
            for i in range(len(new_children[type])):
                child = new_children[type][i]
                old_child = old_children[type][i]
                if child['dbID'] not in old_ids:
                    raise SecurityException('Invalid action for online survey')

                if child.has_key('filters'):
                    if child['filters'] != old_child['filters']:
                        raise SecurityException('Invalid action for online survey')

                if child.has_key('type'):
                    if child['type'] != old_child['type']:
                        raise SecurityException('Invalid action for online survey')

                if child.has_key('required'):
                    if child['required'] != old_child['required']:
                        raise SecurityException('Invalid action for online survey')

                if not child.has_key('children') or len(child['children'])==0:
                    continue
                self._check_state(child['children'], old_child['children'])

    @transaction_decorator
    def save_state(self, state, group_id=None):
        """
        Receives a state and saves it. Returns the a mapping of db id's

        @param state: the state that we need to save
        """
        # check the state in case the survey is online
        if state['survey']['dbID'] is not None:
            survey_id = state['survey']['dbID']
            old_state = self.load_state(survey_id)
            if old_state['survey']['has_active_publications']:
                if not old_state['survey']['alter_published_survey']:
                    self._check_state(state['survey']['children'], old_state['survey']['children'])
                # in case the user cannot alter published surveys:
                # only changes possibles are related to translations, fallback
                # locale or survey locales. In this case we need to set the survey to stale for the aggregators
                # to repopulate the tables with the changes
                state['survey']['stale'] = 1

        # reset global id_map
        AbstractStateController.reset_state_map()

        # if survey has the 'save new items to pool' enabled, notify items
        if state['survey'].has_key('savePoolItems') and state['survey']['savePoolItems']:
            c = None
            if state['survey'].has_key('poolItemsCategory'):
                c = state['survey']['poolItemsCategory']
            self._notify_items(state['survey']['children'], c)

        survey = self.state_controller.save(state, group_id)

        logger.info('%s: Saved %s' % (self, survey) )

        # At this point the whole state has been procesed and modules have been saved to the db. What is left to
        # do is flush the delete_pool because we are sure that these objects arent used anymore.
        AbstractStateController.commit_delete_pool(self.sa_session)

        id_map = AbstractStateController.save_state_id_map
        AbstractStateController.reset_state_map()

        if self.has_permissions('create_conditional_branch'):
            controller = BranchController(self.sa_session, self.account_id)
            controller.save(state, id_map)

        # return the id map
        return id_map

    def _notify_items(self, children, category_name):
        """
        Recursive method that finds the SurveyItems in the state.
        """
        for type in children:
            for child in children[type]:
                if child['objectName'] in iquest_models.SurveyItemQA.available_object_names+[u'SurveyItemHtml']:
                    child['savePoolItem'] = True
                    child['poolItemCategory'] = category_name
                    continue
                if not child.has_key('children') or len(child['children'])==0:
                    continue
                self._notify_items(child['children'], category_name)

    def load_state(self, survey_id, locale=None):
        """
        Returns a state of the survey for the given survey_id
        """
        state = self.state_controller.load_by_survey_id(survey_id, locale=locale)

        state['survey']['has_active_publications'] = False
        # check if survey has at least one online publication
        q = self.sa_session.query(SurveyPublication).filter(and_(SurveyPublication.survey_id==survey_id,
            SurveyPublication.status.in_([SurveyPublication.STATUS_ONLINE, SurveyPublication.STATUS_FINISHED]))).all()
        if len(q) > 0:
            state['survey']['has_active_publications'] = True
        return state

    def clone(self, survey_id, new_label, make_pool_items=False):
        """
        This method clones a survey.

        This is done by loading the survey state, removing the dbID's and saving it again as new.
        In case the make_pool_items boolean is positive, all (QA) items which are not pool items
        yet will be added to the Item Pool.
        """
        return self.state_controller.clone(survey_id, new_label, make_pool_items)

    def has_all_qas_as_pool_items(self, survey_id):
        """
        Returns a boolean whether or not ALL the QA's in the survey are pool items.
        """
        state = self.state_controller.load_by_survey_id(survey_id)
        return self.state_controller.has_all_qas_as_pool_items(state['survey'])

    def __repr__(self):
        return '<ComposeController>'

class SurveyStateController(AbstractStateController):
    """
    Survey controller
    """
    _instance_class = iquest_models.Survey
    _locale_attributes = ['title']

    def __init__(self, session, manager):
        AbstractStateController.__init__(self, session, manager)
        self.blockStateController = BlockStateController(self.sa_session, manager)

    @transaction_decorator
    def save(self, state, group_id=None):
        """
        Takes care of saving the survey state.
        Returns the new/updated survey

        @param state: the state that we need to save
        """

        # for now we are only intestested in the canvas state
        state = state['survey']

        # assert state
        self._assert_state(state, None, dict)
        self._assert_state(state, 'domID', unicode)
        self._assert_state(state, 'objectName', unicode, 'Survey')
        self._assert_state(state, 'theme', dict)
        self._assert_state(state['theme'], 'module', unicode)
        self._assert_state(state['theme'], 'name', unicode)
        self._assert_state(state['theme'], 'cssClass', unicode)
        self._assert_state(state, 'children', dict)
        self._assert_state(state['children'], 'blocks', list)

        # save or update
        if state['dbID']:
            self.assert_permissions('update_survey')
        else:
            self.assert_permissions('create_survey', group_id=group_id)
            # check if sufficient credits
            self.assert_credits("max_number_of_surveys")
            # new survey, so remove one credit
            self.substract_usage_credits("max_number_of_surveys", 1)

        # call parent which takes care of common compose logic
        survey = AbstractStateController._instance_from_state(self, state)

        # check the object privileges in case of update
        if state['dbID']:
            self.assert_privileges_on_object(survey, iquest_models.Survey.PRIV_UPDATE)

        # Save survey name
        self._assert_state(state, 'name', unicode)
        survey.name = state['name']

        if state.has_key('stale'):
            survey.stale = state['stale']

        # theme
        try:
            theme = self.sa_session.query(iquest_models.IQuestTheme).filter(iquest_models.IQuestTheme.module==state['theme']['module']).one()
        except NoResultFound:
            raise PublicSystemException('No such theme: %s' % state['theme']['module'])
        survey.theme = theme
        self.assert_privileges_on_object(theme, IQuestTheme.PRIV_VIEW)

        # assert fallback locale
        self._assert_state(state, 'fallbackLocale', dict)
        self._assert_state(state['fallbackLocale'], 'locale', unicode)
        survey.fallback_locale = self.locale_controller.get_locale_by_str(state['fallbackLocale']['locale'])

        # call parent save that actually writes to the db
        AbstractStateController.save(self, survey, state)
        self._save_available_locales(survey, iquest_models.SurveyLocale, 'survey', state)

        # assert that fallback locale is in available locales
        if not survey.fallback_locale.locale in state['availableLocales']:
            raise PublicSystemException('Fallback locale must be in available locales')

        # delete any removed blocks and process
        self._delete_children(survey.blocks, state['children']['blocks'])
        for i in range(len(state['children']['blocks'])):
            self.blockStateController.save(survey, state['children']['blocks'][i], i)

        if state['dbID'] in [None, '']:
            self.save_default_privilege_object_group(survey, target_group=group_id)
        else:

            # get all survey publications that have a state saved (no matter the status) and reset that
            publications = self.sa_session.query(iquest_models.SurveyPublication) \
                            .join(ComposePublicationLocale) \
                            .filter(iquest_models.SurveyPublication.survey==survey) \
                            .all()

            for p in publications:
                self._save_publication_state(p, None, locale=None)

        # finally return the new saved survey
        return survey

    def clone(self, survey_id, new_label, make_pool_items=False):
        state = self.load_by_survey_id(survey_id)
        self._clone_dom_counter = 0

        if make_pool_items:
            # update the original survey
            self._alter_state_for_clone(state['survey'], make_pool_items)
            self._clone_dom_counter = 0
            self.save(state)
            AbstractStateController.reset_state_map()
            state = self.load_by_survey_id(survey_id)

        self._alter_state_for_clone(state['survey'])
        state['survey']['name'] = new_label

        survey = self.save(state)

        AbstractStateController.reset_state_map()

        return survey


    def has_all_qas_as_pool_items(self, state, type='survey'):

        if type == 'items' and state['poolItemID'] is None:
            return False

        if (state.has_key('children')):
            for child_attr in ['blocks', 'columns', 'items']:
                if state['children'].has_key(child_attr):
                    for child in state['children'][child_attr]:
                        if not self.has_all_qas_as_pool_items(child, child_attr):
                            return False
        return True

    def _alter_state_for_clone(self, state, make_pool_items=False):
        """
        recursive unsets dbID's for the given state
        """
        state['domID'] = 'dom-%s' % self._clone_dom_counter
        self._clone_dom_counter = self._clone_dom_counter + 1

        if not make_pool_items:
            state['dbID'] = None

        if make_pool_items and state['objectName'] in iquest_models.SurveyItemQA.available_object_names:
            state['savePoolItem'] = True
            state['poolItemCategory'] = None

        if (state.has_key('children')):
            for child_attr in ['blocks', 'columns', 'items', 'questions', 'answers']:
                if state['children'].has_key(child_attr):
                    for child in state['children'][child_attr]:
                        self._alter_state_for_clone(child, make_pool_items)

    def _load_state(self, survey, locale=None):
        """
        Private load method whithout permission checking.. should be called from a load_by_* method
        This returns a state dict
        """

        if survey == None:
            raise PublicUserException(_('Survey not found'))
        if survey.deleted == True:
            raise PublicSecurityException('Invalid survey (deleted)')

        alter_published_survey = False
        if self.has_permissions('update_published_survey') and self.has_privileges_on_object(survey, Survey.PRIV_UPDATE_PUBLISHED_SURVEY):
            alter_published_survey = True

        state = {
            'survey': {
                'objectName': u'Survey',
                'dbID': survey.id,
                'name': survey.name,
                'alter_published_survey': alter_published_survey,
                'theme': {
                    'name': survey.theme.name,
                    'cssClass': survey.theme.css_class,
                    'module': survey.theme.module,
                },
                'availableLocales': [locale_.locale for locale_ in survey.available_locales],
                'fallbackLocale': {'locale': survey.fallback_locale.locale, 'name': survey.fallback_locale.name},
                'children': {
                    'blocks': []
                }
            }
        }

        for block in survey.blocks:
            state['survey']['children']['blocks'].append(self.blockStateController.load(block.id))

        if locale is not None:

            # strip the state from other locales than what the user needs
            state['survey'] = self._locale_strip_state(state['survey'], locale, state['survey']['fallbackLocale']['locale'])

            # override fallback locale, this makes the runner choose this locale as default (there are no others anway)
            state['survey']['fallbackLocale'] = {'name': locale.name, 'locale': locale.locale}

        return state

    def load_by_survey_id(self, survey_id, locale=None):
        """
        Loads and checks permissions for the given survey_id
        Returns a state object
        """
        self.assert_permissions('view_survey')

        survey = self._instance_from_id(survey_id)
        self.assert_privileges_on_object(survey, iquest_models.Survey.PRIV_VIEW)

        return self._load_state(survey, locale)

    def load_survey_variable_replaced_state(self, survey_state, set_id):
        """
        Having a survey state, it replaces all variables in it for the given set and and for all locales the set
        has values for.
        Returns the new survey state.
        """

        # since this can be called when switching variable sets in preview mode, selecting
        # no variable set should just return the survey state
        if set_id is None or set_id == '':
            return survey_state

        variable_controller = SurveyVariableStateController(self.sa_session, self.manager)

        # method for finding and replacing vars in texts
        def replace_vars(obj, label):
            """
            Searches in all locales for the given obj and label for survey variables. When found they get added
            to the 'return_vars' list.
            """
            for key in obj["locales"]:
                value = obj["locales"][key]
                var_list = variable_controller.load_variables(set_id, key)

                text = value[label]
                vars = re.findall('\{\{[_|\w]+\}\}', text) # like {{var_name}}
                for var in vars:
                    var_handler = var[2:][:-2]
                    if var_list[var_handler] is not None:
                        value[label] = value[label].replace(var, var_list[var_handler])
                    # else it will just return {{a_var}} and the user would see in preview mode which variables
                    # don't have a value set

        state = survey_state['survey']

        # search in state for vars to replace
        for block in state['children']['blocks']:
            for column in block['children']['columns']:
                for item in column['children']['items']:
                    if item['objectName'] == u'SurveyItemHtml':
                        replace_vars(item, 'html')
                    if item['objectName'] in iquest_models.SurveyItemQA.available_object_names:
                        # search and replace in ItemQA title, questions and answers
                        replace_vars(item, 'label')
                        for q in item['children']['questions']:
                            replace_vars(q, 'label')
                        for a in item['children']['answers']:
                            replace_vars(a, 'label')

        # finally return the new survey state
        return survey_state


    @transaction_decorator
    def load_by_invitee_id(self, invitee_id):
        """
        Used for retrieving a state for the given invitee_id.. this is used by the ResponseController to retrieve the state.
        The survey state returned has all the variables replaced for the invitee's publication set id and
        invitee's locale
        """

        from basesoft.web.contrib.compose.models import ComposeInvitee
        publication = self.sa_session.query(iquest_models.SurveyPublication). \
            join(ComposeInvitee). \
            filter(ComposeInvitee.id == invitee_id).one()
        set_id = publication.variable_set_id

        survey = self.sa_session.query(iquest_models.Survey). \
            join(iquest_models.SurveyPublication). \
            join(ComposeInvitee). \
            filter(ComposeInvitee.id == invitee_id).one()

        locale = self.locale
        if locale is None:
            locale = survey.fallback_locale

        # try to load survey state from cache else load it
        try:
            survey_state = self.sa_session.query(ComposePublicationLocale) \
                .filter(ComposePublicationLocale.publication==publication, ComposePublicationLocale.locale==locale).one().state

            survey_state = simplejson.loads(survey_state, object_hook=json_object_hook)

        except NoResultFound:

            # state not cached.. load it
            survey_state = self._load_state(survey, locale=locale)
            branch_cache = self._load_branch_cache(survey)

            # cache it for next invitee with same locale
            self._save_publication_state(publication, survey_state, locale, branch_cache)

        # send to client
        return self.load_survey_variable_replaced_state(survey_state, set_id)


    def _save_publication_state(self, publication, state, locale, branch_cache=None):

        if branch_cache is None:
            publication.branch_cache = None
        else:
            publication.branch_cache = simplejson.dumps(branch_cache, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        AbstractStateController._save_publication_state(self, publication, state, locale)


    def _load_branch_cache(self, survey):
        # get a state which lists all the branches and the unreachable items for those branches.

        bc = BranchController(self.sa_session, account_id=None)
        return bc.get_branch_cache(survey)

    def load_branching(self, survey_id):
        """
        """
        survey = self.sa_session.query(iquest_models.Survey).filter(iquest_models.Survey.id==survey_id).one()
        return [block.id for block in survey.blocks if len(block.branches) > 0]

    def is_same(self, source_survey, target_survey, strict=False):
        self.logger.info('Comparing survey states')
        if self.get_account() is not None:
            # for now this can only be used with clone perms or from task
            self.assert_permissions('clone_survey_publication_responses')

        source_state = self._load_state(source_survey)['survey']
        target_state = self._load_state(target_survey)['survey']

        locale = source_state['fallbackLocale']['locale']

        self.logger.info(' +- Checking locales')

        # All locales in the source state should be in target_state
        if not all(val in target_state['availableLocales'] for val in source_state['availableLocales']):
            return False, 'Not all locales are the same.'

        source_blocks = source_state['children']['blocks']
        target_blocks = target_state['children']['blocks']

        self.logger.info(' +- Checking blocks')
        # The target state should have at least the same number of blocks.
        if len(source_blocks) > len(target_blocks):
            return False, 'The number of blocks do not match (%d in source vs %d in target).' % (len(source_blocks), len(target_blocks))

        target_blocks = iter(target_blocks)

        # We loop the blocks and compare the content
        bi = 1
        for sb in source_blocks:
            self.logger.info('   +- Checking block [%d]' % bi)
            tb = next(target_blocks)

            source_columns = sb['children']['columns']
            target_columns = tb['children']['columns']

            # The target state should have at least the same number of columns.
            if len(source_columns) > len(target_columns):
                return False, 'The number of columns in block number %d do not match.' % bi

            target_columns = iter(target_columns)

            ci = 1
            for sc in source_columns:
                self.logger.info('    +- Checking columns [%d]' % ci)
                tc = next(target_columns)

                source_qas = [qa for qa in sc['children']['items'] if qa['objectName'] == u'SurveyItemQA']
                if len(source_qas) == 0:
                    ci += 1
                    continue

                self.logger.info('    +- Checking QAs; total of %d' % len(source_qas))
                target_qas = [qa for qa in tc['children']['items'] if qa['objectName'] == u'SurveyItemQA']
                if len(source_qas) != len(target_qas):
                    return False, 'The number of QAs in column number %d in block number %d do not match.' % (ci, bi)

                target_qas = iter(target_qas)
                for sqa in source_qas:
                    tqa = next(target_qas)
                    label = "No translation - QA title hidden?"
                    if sqa['locales'].has_key(locale):
                        label = sqa['locales'][locale]['label']
                    self.logger.info('     +- Checking QA; %s' % label)

                    # verify QA type
                    if sqa['type'] != tqa['type']:
                        return False, 'The QA type of "%s" is different.' % label

                    # verify obligatory
                    self.logger.info('     +- Checking required setting')
                    if sqa['required'] != tqa['required']:
                        return False, 'The QA mandatory state of "%s" is different.' % label

                    # verify question length
                    self.logger.info('     +- Checking questions')
                    if len(sqa['children']['questions']) != len(tqa['children']['questions']):
                        return False, 'The number of questions in QA "%s" do not match.' % label

                    # verify answer length
                    self.logger.info('     +- Checking answers')
                    if len(sqa['children']['answers']) != len(tqa['children']['answers']):
                        return False, 'The number of answers in QA "%s" do not match.' % label

                    self.logger.info('     +- Checking filters')

                    if len(sqa['filters']) != len(tqa['filters']):
                        return False, 'The number of filters for QA "%s" do not match.' % label

                    # verify filters
                    if len(sqa['filters']) == 0:
                        continue

                    qmap = {None: None}  # question map for filtering checks
                    tqs = iter(tqa['children']['questions'])
                    for q in sqa['children']['questions']:
                        tq = next(tqs)
                        qmap[q['dbID']] = tq['dbID']
                    amap = {None: None}  # answer map for filtering checks
                    tas = iter(tqa['children']['answers'])
                    for a in sqa['children']['answers']:
                        ta = next(tas)
                        amap[a['dbID']] = ta['dbID']

                    # TODO: do/should we care about the order of the filters?
                    tfs = tqa['filters']
                    for sf in sqa['filters']:
                        tf = None
                        found = None
                        _filter = self.sa_session.query(IQuestQAFilter).filter(IQuestQAFilter.id==sf['dbID']).first()
                        for f in tfs:
                            if f['dbID'] != sf['dbID']:  # this is the id in the DB in tblIQuestQAFilter, not tblSurveyItemQAFilter!
                                continue
                            tf = f
                            self.logger.info('      +- Checking filter %s' % _filter.name)
                            # check for the affected question/answers
                            if qmap[sf['questionDbID']] != tf['questionDbID']:
                                continue

                            if amap[sf['answerDbID']] != tf['answerDbID']:
                                continue
                            # if we get here, that can only happen once for the same filter for the same Q/A as the DB
                            # constraint is item_id+filter_id+question_id+answer_id

                            # check for the filter attributes.
                            for k, v in sf['attributes'].iteritems():
                                if not tf['attributes'].has_key(k):
                                    return False, 'The filter "%s" on QA "%s" misses the attribute "%s".' % (_filter.name, label, k)
                                if v != tf['attributes'][k]:
                                    return False, 'The attribute "%s" for filter "%s" on QA "%s" does not match. (%s vs %s)' % (
                                        k, _filter.name, label, v, tf['attributes'][k]
                                    )

                            found = True
                        if found is None:
                            return False, 'The filter "%s" on QA "%s" could not be matched.' % (_filter.name, label)
                ci += 1

            bi += 1

        if strict:  # if strict mode, we also expect the inverse of source and target to return True
            return self.is_same(target_survey, source_survey, False)

        return True, None

class SurveyVariableStateController(AbstractStateController):
    """
    Controller for SurveyVariableSet
    """
    _instance_class = iquest_models.SurveyVariableSet

    def get_variable_sets(self, survey_id=None, id=None):
        """
        """
        self.assert_permissions('view_survey_variable_set')
        filter, fo = self.get_query_filter_on_privileges(iquest_models.SurveyVariableSet,
                                                     iquest_models.SurveyVariableSetGroup,
                                                     iquest_models.SurveyVariableSet.PRIV_VIEW)
        if survey_id is not None:
            # check surveys
            compose_controller = ComposeController(self.sa_session, self.locale, self.account_id)
            survey_ids = [s.id for s in compose_controller.get_surveys(survey_id)]
            filter = and_(filter, iquest_models.SurveyVariableSet.survey_id.in_(survey_ids))

        if id is None:
            return self.sa_session.query(iquest_models.SurveyVariableSet).outerjoin(
                    iquest_models.SurveyVariableSetGroup).filter(filter).all()
        if not isinstance(id, list):
            id = [id]

        return self.sa_session.query(iquest_models.SurveyVariableSet).select_from(fo).filter(
                    and_(filter, iquest_models.SurveyVariableSet.id.in_(id))).all()

    def load(self, set_id):
        """
        Returns the state for the given set_id
        """

        set = self.get_variable_sets(survey_id=None, id=set_id)
        if len(set) == 0:
            raise SecurityException('Invalid variable set id %s' % set_id)
        set = set[0]

        return {
            'dbID': set.id,
            'domID': None,
            'survey_id': set.survey_id,
            'name': set.name
        }


    @transaction_decorator
    def save(self, state):
        """
        Inserts or update the basic information for a SurveyVariableSet. The variables themselfs can be updated using
        update_variables()
        """
        # assert state
        self._assert_state(state, None, dict)
        self._assert_state(state, 'name', unicode)

        # get variable set instance
        variable_set = AbstractStateController._instance_from_state(self, state)

        # check create/update permissions
        if state['dbID'] == None:
            self.assert_permissions('create_survey_variable_set')
        else:
            self.assert_permissions('update_survey_variable_set')
            self.assert_privileges_on_object(variable_set, iquest_models.SurveyVariableSet.PRIV_UPDATE)

        # query survey based on id
        survey = self.sa_session.query(iquest_models.Survey).get(state['survey_id'])

        # on survey level we only need the view rights
        self.assert_permissions('view_survey')
        self.assert_privileges_on_object(survey, iquest_models.Survey.PRIV_VIEW)

        # perform actual updating
        variable_set.survey = survey
        variable_set.name = state['name']

        if state['dbID'] is None:
            self.save_default_privilege_object_group(variable_set)

        # save object
        self.sa_session.add(variable_set)
        self.sa_session.flush()

        # return the variable_set's id
        return variable_set.id

    def list_variables(self, survey_id, locale_str):
        """
        Returns all possible variable handles for the given survey_id as a list of strings
        """

        # there are no privs required to get the list of all variables. This method is also
        # used when variables get replaces in a participant's survey and there are not privs checks
        # for a survey invitee

        # get survey
        survey = self.sa_session.query(iquest_models.Survey).get(survey_id)

        # get locale obj
        locale = self.sa_session.query(Locale).filter_by(locale=locale_str).one()

        # return value
        return_vars = []

        # method for finding vars in texts
        def find_vars(obj, label):
            """
            Searches in all locales for the given obj and label for survey variables. When found they get added
            to the 'return_vars' list.
            """
            for obj_locale in obj.locales:
                if obj_locale.locale == locale:
                    text = getattr(obj_locale, label)
                    if text is None:
                        break
                    vars = re.findall('\{\{[_|\w]+\}\}', text) # like {{var_name}}
                    for var in vars:
                        var = var[2:][:-2] # string {{..}}
                        if var not in return_vars:
                            return_vars.append(var)

        # continue to items
        for block in survey.blocks:
            for column in block.columns:
                for item in column.items:
                    if isinstance(item, iquest_models.SurveyItemHtml):
                        find_vars(item, 'html')
                    if isinstance(item, iquest_models.SurveyItemQA):
                        find_vars(item, 'label')
                        for q in item.questions:
                            find_vars(q, 'label')
                            find_vars(q, 'trailing_label')
                        for a in item.answers:
                            find_vars(a, 'label')

        # finally return the list of vars
        return return_vars

    @transaction_decorator
    def update_variables(self, set_id, locale_str, variables):
        """
        Updates the db with the given variables for the given set_id and locale_str
        """

        assert isinstance(set_id, (int,long) )
        assert isinstance(locale_str, (unicode) )
        assert isinstance(variables, (dict) )

        # get variable_set
        try:
            variable_set = self.sa_session.query(iquest_models.SurveyVariableSet).get(set_id)
        except NoResultFound:
            raise PublicSystemException('No such variable_set: [%s]' % set_id)

        self.assert_permissions('update_survey_variable_set')
        self.assert_privileges_on_object(variable_set, iquest_models.SurveyVariableSet.PRIV_UPDATE)

        # get locale obj
        locale = self.sa_session.query(Locale).filter_by(locale=locale_str).one()

        # delete all existing so we dont duplicate when inserting
        for obj in self.sa_session.query(iquest_models.SurveyVariableLocale).filter(and_(
            iquest_models.SurveyVariableLocale.variable_set==variable_set,
            iquest_models.SurveyVariableLocale.locale==locale)):
            self.sa_session.delete(obj)

        # insert variables
        for key,value in variables.items():

            if value is None or value.strip() == '':
                continue

            obj = iquest_models.SurveyVariableLocale()
            obj.variable_set = variable_set
            obj.locale = locale
            obj.variable = key
            obj.text = value.strip()
            self.sa_session.add(obj)

        # send to db
        self.sa_session.flush()

    def load_variables(self, set_id, locale_str):
        """
        Returns a dictionary with keys of variable names and values of texts for the given set_id and locale_str. This
        also includes empty (non-filled) texts for variables.
        """

        # there are no privs required to get the list of all variables and values. This method is also
        # used when variables get replaces in a participant's survey and there are not privs checks
        # for a survey invitee

        # get variable_set
        try:
            variable_set = self.sa_session.query(iquest_models.SurveyVariableSet).get(set_id)
        except NoResultFound:
            raise PublicSystemException('No such variable_set: [%s]' % set_id)

        # get locale obj
        locale = self.sa_session.query(Locale).filter_by(locale=locale_str).one()

        # use the list_variables() result as a base of the return value
        rv = {}
        for i in self.list_variables(variable_set.survey.id, locale.locale):
            rv[i] = None

        # and merge it with values in the db
        for obj in variable_set.variables:
            if obj.locale == locale:
                rv[obj.variable] = obj.text

        # and return
        return rv

class PublicationStateController(compose_controllers.PublicationStateController):
    """
    """

    _instance_class = iquest_models.SurveyPublication

    def get_publications(self, id=None, priv=iquest_models.SurveyPublication.PRIV_VIEW):
        self.assert_permissions('view_survey_publication')
        return compose_controllers.PublicationStateController.get_publications(self, id, priv)

    @transaction_decorator
    def save_branch_cache(self, publication):
        bc = BranchController(self.sa_session, account_id=None)
        branch_cache = bc.get_branch_cache(publication.survey)

        if branch_cache is None:
            publication.branch_cache = None
        else:
            publication.branch_cache = simplejson.dumps(branch_cache, ensure_ascii=False, cls=JsonRpcJSONEncoder)

        self.sa_session.add(publication)
        self.sa_session.flush()

    @transaction_decorator
    def reopen_publication(self, _id):
        self.assert_permissions('update_survey_publication', 'schedule_survey_publication')
        publication = self.get_publications(_id, priv=iquest_models.SurveyPublication.PRIV_UPDATE).pop()
        if publication.status != iquest_models.SurveyPublication.STATUS_FINISHED:
            raise PublicUserException(_('Only finished publications can be reopened.'))

        publication.status = iquest_models.SurveyPublication.STATUS_ONLINE
        publication.end_date = None # set the end date to None else it will immediately close again.
        self.sa_session.add(publication)
        self.sa_session.flush()

        self.logger.info('Survey publication [%s] has been reopened.' % publication.id)

        return publication

    @transaction_decorator
    def save(self, state, group_id=None):

        self._assert_state(state, 'survey_id', (long, int))
        # query survey based on id
        survey = self.sa_session.query(iquest_models.Survey).get(state['survey_id'])

        # on survey level we only need the view right to set a publication
        self.assert_permissions('view_survey')
        self.assert_privileges_on_object(survey, iquest_models.Survey.PRIV_VIEW)

        if state['dbID'] == None:
            self.assert_permissions('create_survey_publication', group_id=group_id)

            # check if sufficient credits
            self.assert_credits("max_number_of_survey_publications")
            # new survey publication, so remove one credit
            self.substract_usage_credits("max_number_of_survey_publications", 1)

            # if the publication is directly set to scheduled on creation, need to check for that perm also
            if state['status'] == iquest_models.SurveyPublication.STATUS_SCHEDULED:
                self.assert_permissions('schedule_survey_publication')
        else:
            self.assert_permissions('update_survey_publication')

        publication = self.set_publication(state, group_id)

        if state['dbID'] is not None:
            # scheduling a publication (setting status from NEW to SCHEDULED) requires
            # 'schedule_survey_publication' perm (privs are checked in parent)
            if (publication.status == iquest_models.SurveyPublication.STATUS_NEW and \
                state['status'] == iquest_models.SurveyPublication.STATUS_SCHEDULED):
                self.assert_permissions('schedule_survey_publication')

        # iquest specific, survey variable set
        if state.has_key('variable_set_id') and state['variable_set_id'] is not None:
            if publication.id is not None and publication.status == iquest_models.SurveyPublication.STATUS_ONLINE and \
                publication.variable_set_id != state['variable_set_id']:
                raise PublicUserException(_('You cannot change the variable set of a publication that is online.'))
            vs_controller = SurveyVariableStateController(self.sa_session, self.manager)
            vs = vs_controller.get_variable_sets(id=state['variable_set_id'])
            if len(vs) > 0:
                publication.variable_set = vs[0]

        publication.survey = survey
        publication.show_end_message = False
        if state.has_key('minimum_respondents') and state['minimum_respondents'] is not None:
            publication.minimum_respondents = int(state['minimum_respondents'])
        if state.has_key('maximum_market_share') and state['maximum_market_share'] is not None:
            publication.maximum_market_share = int(state['maximum_market_share'])
        if state.has_key('multi_session_setting') and state['multi_session_setting'] is not None:
            publication.multi_session_setting = state['multi_session_setting']
        if state.has_key('minimum_responses_for_respondent') and state['minimum_responses_for_respondent'] is not None:
            publication.minimum_responses_for_respondent = state['minimum_responses_for_respondent']
        if state.has_key('show_end_message') and state['show_end_message'] is not None:
            publication.show_end_message = state['show_end_message']

        if state['dbID'] is None:
            self.save_default_privilege_object_group(publication, target_group=group_id)

        self.sa_session.add(survey)
        self.sa_session.flush()

        # expire survey so it will update the publications
        self.sa_session.expire(survey)
        return publication

    def load(self, id):
        """
        Loads and returns the state for the given publication_id
        """
        publication = self._instance_from_id(id)

        self.assert_permissions('view_survey')
        self.assert_permissions('view_survey_publication')
        self.assert_privileges_on_object(publication.survey, iquest_models.Survey.PRIV_VIEW)
        self.assert_privileges_on_object(publication, iquest_models.SurveyPublication.PRIV_VIEW)

        state = compose_controllers.PublicationStateController.load(self, publication)
        state["survey_id"] = publication.survey_id
        state["variable_set_id"] = publication.variable_set_id
        state["minimum_respondents"] = publication.minimum_respondents
        state["maximum_market_share"] = publication.maximum_market_share
        state["multi_session_setting"] = publication.multi_session_setting
        state["minimum_responses_for_respondent"] = publication.minimum_responses_for_respondent
        state["show_end_message"] = publication.show_end_message

        return state

class BlockStateController(AbstractStateController):
    """
    Takes care of block state logic
    """

    _instance_class = iquest_models.SurveyBlock

    def __init__(self, session, manager):
        AbstractStateController.__init__(self, session, manager)
        self.columnStateController = ColumnStateController(self.sa_session, manager)

    @transaction_decorator
    def save(self, survey, state, position):
        assert isinstance(survey, iquest_models.Survey)
        self._assert_state(state, None, dict)
        assert isinstance(position, int)
        self._assert_state(state, 'objectName', unicode, 'SurveyBlock')

        self._assert_state(state, 'pageEnd', bool)
        self._assert_state(state, 'children', dict)
        self._assert_state(state['children'], 'columns', list)

        # call parent which takes care of common compose logic
        block = AbstractStateController._instance_from_state(self, state)

        # reference method arguments
        block.survey = survey
        block.position = position
        block.page_end = state['pageEnd']

        # call parent save that actually writes to the db
        AbstractStateController.save(self, block, state)

        # delete any removed columns
        self._delete_children(block.columns, state['children']['columns'])

        # continue processing columns
        for i in range(len(state['children']['columns'])):
            self.columnStateController.save(block, state['children']['columns'][i], i)

        # expire survey (so it will update survey.blocks)
        self.sa_session.expire(survey)

    def load(self, id):
        block = self._instance_from_id(id)

        state = {
            "objectName": u"SurveyBlock",
            "dbID": block.id,
            "pageEnd": bool(block.page_end),
            "children": {
                'columns': []
            }
        }

        for column in block.columns:
            state['children']['columns'].append(self.columnStateController.load(column.id))

        return state


class ColumnStateController(AbstractStateController):
    """
    Takes care of state handling for columns
    """

    _instance_class = iquest_models.SurveyBlockColumn

    def __init__(self, session, manager):
        AbstractStateController.__init__(self, session, manager)
        self.pool_controller = ItemPoolController(self.sa_session, manager.locale, manager.account_id)

    @transaction_decorator
    def save(self, block, state, position):
        assert isinstance(block, iquest_models.SurveyBlock)
        self._assert_state(state, None, dict)
        assert isinstance(position, int)
        self._assert_state(state, 'objectName', unicode, 'SurveyBlockColumn')

        self._assert_state(state, 'children', dict)
        self._assert_state(state['children'], 'items', list)

        # call parent which takes care of common compose logic
        column = AbstractStateController._instance_from_state(self, state)

        column.block = block
        column.width_percentage = state['width']
        column.position = position

        # call parent save that actually writes to the db
        AbstractStateController.save(self, column, state)

        # delete any removed items
        self._delete_children(column.items, state['children']['items'])

        # process items
        for i in range(len(state['children']['items'])):
            self.__save_item_state(column, state['children']['items'][i], i)

        # expire
        self.sa_session.expire(block)

    def __save_item_state(self, column, state, position):
        """
        Takes care of the saving state of items by mapping the item controller and instanciating it.
        """
        assert isinstance(column, iquest_models.SurveyBlockColumn)
        self._assert_state(state, None, dict)

        try:
            # different controllers depending on item object name (SurveyItemQA, ItemQANPS etc)
            ItemController = item_controller_mapping[state['objectName']]
        except KeyError:
            logger.error('ColumnStateController: Ignoring item saving. No mapper found for %s' % state['objectName'])
        else:
            if state.has_key('savePoolItem') and state['savePoolItem'] and state['poolItemID'] is None:
                c = iquest_models.IQuestItemPoolCategory.DEFAULT_CATEGORY_NAME
                if state.has_key('poolItemCategory') and state['poolItemCategory']:
                    c = state['poolItemCategory']
                category = self.pool_controller.get_category_by_name(c)
                if category is None:
                    category = self.pool_controller.save_or_update_category(None, c)

                pool_state = copy.deepcopy(state)
                # pool items are always saved as new at this stage, so remove dbID from the state if set
                # (it is the dbID of the survey item anyway)
                pool_state = self.pool_controller.remove_item_state_dbIDs(pool_state)
                self.pool_controller.save_or_update_item(category, pool_state)

                # currently the id map of this item's domID is set the the newly created pool item id. This will
                # be updated when saving the survey item with the survey item's id

                # get the id of the saved pool item and set is as pool item id for the survey item
                state['poolItemID'] = AbstractStateController.save_state_id_map[state['domID']]
                if state['objectName'] in iquest_models.SurveyItemQA.available_object_names:
                    for q in state['children']['questions']:
                        q['poolElementID'] = AbstractStateController.save_state_id_map[q['domID']]
                    for a in state['children']['answers']:
                        a['poolElementID'] = AbstractStateController.save_state_id_map[a['domID']]
                # need to update id_map so that survey items' poolItemID would get updated
                # and also clear save_state_id_map because parent's save method checks for double domIDs in it
                if not AbstractStateController.save_state_id_map.has_key('poolItems'):
                    AbstractStateController.save_state_id_map['poolItems'] = {}
                AbstractStateController.save_state_id_map['poolItems'][state['domID']] = AbstractStateController.save_state_id_map[state['domID']]
                del AbstractStateController.save_state_id_map[state['domID']]

                if state['objectName'] in iquest_models.SurveyItemQA.available_object_names:
                    for q in state['children']['questions']:
                        AbstractStateController.save_state_id_map['poolItems'][q['domID']] = AbstractStateController.save_state_id_map[q['domID']]
                        del AbstractStateController.save_state_id_map[q['domID']]
                    for a in state['children']['answers']:
                        AbstractStateController.save_state_id_map['poolItems'][a['domID']] = AbstractStateController.save_state_id_map[a['domID']]
                        del AbstractStateController.save_state_id_map[a['domID']]

            ic = ItemController(self.sa_session, self.manager)
            ic.save(column, state, position)

    def load(self, id):
        column = self._instance_from_id(id)

        state = {
            "objectName": u"SurveyBlockColumn",
            "dbID": column.id,
            "width": column.width_percentage,
            "children": {
                'items': []
            }
        }

        for item in column.items:
            try:
                # use the correct controller depending on the item object type (SurveyItemQA, ItemQANPS etc)
                stateControllerClass = item_controller_mapping[item.__class__.__name__]
            except KeyError:
                logger.error('ColumnStateController: Ignoring item loading. No mapper found for %s' % item.__class__.__name__)
            else:
                stateController = stateControllerClass(self.sa_session, self.manager)
                state['children']['items'].append(stateController.load(item.id))

        return state

class ItemHtmlStateController(IQuestItemHtmlStateController):

    """
    Controller for the ItemHtml model
    """
    _container = 'column'
    _container_class = iquest_models.SurveyBlockColumn
    _instance_class = iquest_models.SurveyItemHtml
    _locale_attributes = ['html']
    _locale_model_class = iquest_models.SurveyItemHtmlLocale

class ItemQAStateController(IQuestItemQAStateController):
    """
    Controller for the QAItem
    """
    _container = 'column'
    _container_class = iquest_models.SurveyBlockColumn
    _instance_class = iquest_models.SurveyItemQA
    _locale_attributes = ['label']
    _locale_model_class = iquest_models.SurveyItemQALocale
    _layout_attribute_class = SurveyItemQALayoutAttribute
    _layout_attribute_item_column = 'item_id'

    def __init__(self, session, manager):
        IQuestItemQAStateController.__init__(self, session, manager)
        self.qaStateController = ItemElementController(self.sa_session, manager)
        self.filterController = SurveyItemQAFilterController(self.sa_session, self.locale, self.account_id, manager)
        self.branchController = BranchController(self.sa_session, self.account_id)

    def load(self, id):
        state = IQuestItemQAStateController.load(self, id)
        state['branching'] = []
        if self.has_permissions('view_conditional_branch'):
            state['branching'] = self.branchController.load(state['dbID'])
        return state

class ItemElementController(IQuestItemElementController):
    """
    Shares common state logic between questions and answers
    """
    _instance_class = iquest_models.SurveyItemQAElement
    _locale_attributes = ['label', 'trailing_label']
    _locale_model_class = iquest_models.SurveyItemQAElementLocale
    _objectName = u'SurveyItemQAElement'
    _item_class = iquest_models.SurveyItemQA

item_controller_mapping = {
    'SurveyItemHtml': ItemHtmlStateController,
    'SurveyItemQA': ItemQAStateController,

    #specific QAs - single choice QA, but client side needs specific checks
    'ItemQANPS': ItemQAStateController
}

class BranchController(AbstractController):

    def get_branch_cache(self, survey):
        """
        Returns a dictionary with a list of unreachable items per branch.
        """
        reachable_items = self.get_reachable_items(survey)
        all = survey.itemQAs

        rv = {}
        # unreachable items are all items which are not in the reachable items, logically :p
        for branch, items in reachable_items.iteritems():
            rv[branch] = [i.id for i in all if i not in items]

        return rv

    def get_reachable_items(self, survey):
        """
        Returns a list per branch with all the possible reachable items. This includes the items of "previous" pages.
        """
        # indexed per "start block" for a page.
        pages = {}
        # list in the correct order of the pages.
        ordered_pages = []
        # a page can have multiple blocks, keep track of the page's starting block.
        start_block = None

        # loop all the blocks in the survey
        for b in survey.blocks:
            # if we dont have a start block, the current block must be the start block.
            if start_block is None:
                start_block = b.id
                ordered_pages.append(start_block)

            # this page has not been added to the dict.
            if not pages.has_key('block_%s' % start_block):
                pages['block_%s' % start_block] = []

            # get all the QA items for each column in the current block.
            for c in b.columns:
                [pages['block_%s' % start_block].append(i) for i in c.items if isinstance(i, SurveyItemQA)]

            # this was the last block on the page, reset.
            if b.page_end:
                start_block = None

        # a dictionary with all the items which have one or more branches, contains the branch id and the end block.
        branched_items = {}
        # a dictionary of blocks which are end points for branches, contains the branch id.
        branched_blocks = {}
        # a dictionary with items per page.
        from_this_page = {}
        # items in block
        items_for_block = {}
        # block from which a branch came.
        branch_from_block = {}

        # loop the blocks
        for b in survey.blocks:
            # check if the block is an end point.
            if b.branches is None:
                continue
            # we are dealing with an end point.
            branched_blocks['block_%s' % b.id] = []
            # loop all the branches to this block.
            for x in b.branches:
                # fill the dict.
                branched_blocks['block_%s' % b.id].append(x.id)
                # add the items involved.
                if not branched_items.has_key('item_%s' % x.item.id):
                    branched_items['item_%s' % x.item.id] = []
                branched_items['item_%s' % x.item.id].append((x.id, b.id, ))

        rv = {}

        def _get_items_for_block(block, rv):

            if items_for_block.has_key('block_%s' % block):
                return items_for_block['block_%s' % block]

            x = [] # items for this iteration
            b = [] # branches for this iteration
            next_page = True

            for i in pages['block_%s' % block]:
                x.append(i)
                # no branch for this item, continue to the next one.
                if not branched_items.has_key('item_%s' % i.id):
                    continue

                # If the item is required , we have to see if every answer/question has a branch.
                if i.required and len(branched_items['item_%s' % i.id]) == len(i.questions)*len(i.answers):
                    next_page = False

                for bb in branched_items['item_%s' % i.id]:
                    branch, bblock = bb
                    branch_from_block['branch_%s' % branch] = block
                    # add the branch to merge with this pages items in the next stage.
                    b.append('branch_%s' % branch)
                    if rv.has_key('branch_%s' % branch):
                        raise SystemException('Endless recursion detected')

                    # the branch should get all the following items for other pages.
                    rv['branch_%s' % branch] = [item for item in _get_items_for_block(bblock, rv)]

            # we should add the items from the current page also to all the branches on this page.
            for y in b:
                [rv[y].append(item) for item in x if item not in rv[y]]

            # there are no obligated branches, so it is possible to go to the next page.
            if next_page:
                if not from_this_page.has_key('block_%s' % block):
                    from_this_page['block_%s' % block] = []
                # we store the current items of this page to later on add based on possible "previous" pages
                from_this_page['block_%s' % block] = [a for a in x if a not in from_this_page['block_%s' % block]]
                index = ordered_pages.index(block)+1
                # is this the last page?
                if len(ordered_pages) > index:
                    [x.append(item) for item in _get_items_for_block(ordered_pages[index], rv) if item not in x]

            # This block is an end point for one or more branches, this means that all previous items are possibly answered!
            if branched_blocks.has_key('block_%s' % block):
                for branch in branched_blocks['block_%s' % block]:
                    if rv.has_key('branch_%s' % branch):
                        [x.append(item) for item in rv['branch_%s' % branch] if item not in x]

            # we also should add items from each possible branch from this page.
            for y in b:
                [x.append(item) for item in rv[y] if item not in x]

            items_for_block['block_%s' % block] = [i for i in x]

            return x

        _get_items_for_block(ordered_pages[0], rv)

        # reset, we're now going to add previous items to this block
        items_for_block = {}

        # if there is a previous page that should have been accessed (ie the previous page could go to the next without branches)
        def _get_items_before_branch(block, rv):

            x = [] # items to add for the branch

            if items_for_block.has_key('block_%s' % block):
                return items_for_block['block_%s' % block]

            for i in pages['block_%s' % block]:
                x.append(i)

            index = ordered_pages.index(block)-1
            # we are not on the first page and it is possible to go to the next page from this one.
            if index >= 0 and from_this_page.has_key('block_%s' % ordered_pages[index]) and from_this_page['block_%s' % ordered_pages[index]]:
                # add items in this page.
                [x.append(item) for item in from_this_page['block_%s' % ordered_pages[index]] if item not in x]
                # check how far we can go back
                [x.append(item) for item in _get_items_before_branch(ordered_pages[index], rv)]


            # we check if this block is an end point for any branches.
            if branched_blocks.has_key('block_%s' % block):
                # get all the items for the original branches' block.
                for branch in branched_blocks['block_%s' % block]:
                    [x.append(item) for item in _get_items_before_branch(branch_from_block['branch_%s' % branch], rv) if item not in x]

            items_for_block['block_%s' % block] = [i for i in x]

            # now we have to merge these items with the current pages' branches.

            for i in pages['block_%s' % block]:
                if not branched_items.has_key('item_%s' % i.id):
                    continue

                # loop the branches for the current item
                for bb in branched_items['item_%s' % i.id]:
                    branch, block = bb
                    # should already exist?
                    if not rv.has_key('branch_%s' % branch):
                        rv['branch_%s' % branch] = []
                    # only add items that dont already exist.
                    [rv['branch_%s' % branch].append(item) for item in x if item not in rv['branch_%s' % branch]]

            return x

        for page in ordered_pages:
            _get_items_before_branch(page, rv)

        return rv

    def load(self, item_id):
        """
        """
        state = []

        branches = self.sa_session.query(iquest_models.SurveyItemQABranch).filter(iquest_models.SurveyItemQABranch.item_id==item_id).all()
        for branch in branches:
            qid = None
            if branch.question is not None:
                qid = branch.question.id
            aid = None
            if branch.answer is not None:
                aid = branch.answer.id
            bid = branch.branch_block.id
            state.append({
                'question': {'dbID': qid},
                'answer': {'dbID': aid},
                'block': {'dbID': bid}
            })

        return state

    @transaction_decorator
    def save(self, state, id_map):
        """
        All QA's plus blocks that are no longer availables already deleted their linked branches.
        All other QA's should delete all their branches and insert new ones.
        """
        self.id_map = id_map
        self._process_children(state['survey']['children'])

    @transaction_decorator
    def _delete_branches(self, item_id):
        """
        """
        branches = self.sa_session.query(iquest_models.SurveyItemQABranch).filter(iquest_models.SurveyItemQABranch.item_id==item_id).all()
        for branch in branches:
            self.sa_session.delete(branch)
        self.sa_session.flush()

    def _save_qa_branch(self, child):
        """
        Save the branching settings that are attached to this item qa.
        """
        item_id = self.id_map[child['domID']]
        item = self.sa_session.query(iquest_models.SurveyItemQA).filter(iquest_models.SurveyItemQA.item_id==item_id).one()

        if child.has_key('branching'):
            # server does not send domID to client, so in case no changes were made on client (which update domIDs),
            # the state will not have domID. Don't continue since nothing has changed

            # even if some of the elements from branching were removed, because of delete cascade, the correspondent
            # branching entry will automatically be removed also
            for b in child['branching']:
                if not b['block'].has_key('domID'):
                    return

        # delete currently attached branches from this item.
        self._delete_branches(item_id)
        if not child.has_key('branching'):
            return

        for b in child['branching']:
            branch = iquest_models.SurveyItemQABranch()
            branch.item = item

            if b.has_key('question'):
                question_id = None
                if b['question']['dbID'] is not None:
                    question_id = b['question']['dbID']
                elif b['question'].has_key('domID') and b['question']['domID'] is not None:
                    # this question was removed from canvas without being saved first: user adds question, then branching
                    # for that question then removes that question then saved
                    if not self.id_map.has_key(b['question']['domID']):
                        continue
                    question_id = self.id_map[b['question']['domID']]
                if question_id is not None:
                    q = self.sa_session.query(iquest_models.SurveyItemQAElement).filter(iquest_models.SurveyItemQAElement.id==question_id).first()
                    # this question was removed
                    if q is None:
                        continue
                    branch.question = q

            if b.has_key('answer'):
                answer_id = None
                if b['answer']['dbID'] is not None:
                    answer_id = b['answer']['dbID']
                elif b['answer'].has_key('domID') and b['answer']['domID'] is not None:
                    # this answer was removed from canvas without being saved first: user adds answer, then branching
                    # for that answer then removes that answer then saved
                    if not self.id_map.has_key(b['answer']['domID']):
                        continue
                    answer_id = self.id_map[b['answer']['domID']]
                if answer_id is not None:
                    a = self.sa_session.query(iquest_models.SurveyItemQAElement).filter(iquest_models.SurveyItemQAElement.id==answer_id).first()
                    # this answer was removed
                    if a is None:
                        continue
                    branch.answer = a

            block_id = None
            if b['block']['dbID'] is not None:
                block_id = b['block']['dbID']
            elif b['block'].has_key('domID') and b['block']['domID'] is not None:
                # this block was removed from canvas without being saved first
                if not self.id_map.has_key(b['block']['domID']):
                    continue
                block_id = self.id_map[b['block']['domID']]

            if block_id is None:
                continue
            bl = self.sa_session.query(iquest_models.SurveyBlock).filter(iquest_models.SurveyBlock.id==block_id).first()
            # this block was removed
            if bl is None:
                continue
            branch.branch_block = bl
            self.sa_session.add(branch)

    def _process_children(self, children):
        """
        Recursive method that finds the SurveyItemQAs in the state.
        """
        for type in children:
            for child in children[type]:
                if child['objectName'] in iquest_models.SurveyItemQA.available_object_names:
                    self._save_qa_branch(child)
                    continue
                if not child.has_key('children') or len(child['children'])==0:
                    continue
                self._process_children(child['children'])
