# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

from basesoft.web.contrib.i18n.controllers import I18nController
from basesoft.web.contrib.i18n.models import Locale
from basesoft.web.contrib.license.controllers import LicenseController

from basesoft.web.core.util.controllers import StateManager

from basesoft.web.contrib.account.decorators import login_required, assert_user_logged_in
from basesoft.web.core.forms.handler import FormValidator
from basesoft.web.core.jsonrpc.dispatcher import JsonRpcDispatcher

from iqportal.iquest.compose.controllers import ComposeController, SurveyStateController, \
    SurveyVariableStateController
from iqportal.iquest.core.models import Survey, SurveyPublication
from iqportal.iquest.core.controllers import ItemPoolController

from iqportal.iquest.core.qa_filters.controllers import IQuestQAFilterController
from iqportal.iquest.core.qa_filters import attributes

from basesoft import logging
from django.shortcuts import render_to_response
from django.template.context import RequestContext
from django.template import TemplateDoesNotExist
from django.utils import simplejson
from basesoft.web.contrib.compose.views import compose_runner
logger = logging.getLogger("iquest.core")


class ComposeRPCHandler:
    """
    Handles rpc calls from the survey compose application. Required is to be logged in
    url: /iqportal.iquest.compose/rpc/
    """

    def __init__(self, session, locale, account_id, group_id):

        self.account_id = account_id
        self.session = session
        self.group_id = group_id
        self.locale = locale

        # instanciate a compose controller
        self.controller = ComposeController(session, locale, account_id)

        # a license controller we'll use to get the group
        self.lc = LicenseController(session, locale, account_id)

    def get_compose_vars(self):
        """
        called before the manager is instanciated and returns all information that is needed for that..
        like localeLists, themes, last saved survey etc
        """

        i18n_controller = I18nController(self.session, self.locale, account_id=self.account_id)

        # get last survey_id
        last_survey_id = self.controller.config.get('iquest.compose.survey.id')
        # it can be that the survey was removed, or revoked permissions for this person.
        try:
            self.controller.get_surveys(last_survey_id).pop()
        except IndexError:
            last_survey_id = 0
            # don't try this again.
            self.controller.config.set('iquest.compose.survey.id', 0)

        # iquest filters
        # this a dictionary that has filter names as keys, and as values we have dictionaries containing the id and
        # the extended id of the filter objects
        filter_controller = IQuestQAFilterController(self.session)
        available_filters = filter_controller.existing_filters() # list of all filter objects

        existing_filters = {}
        for f in available_filters:
            existing_filters[f.name] = {'id': f.id, 'extended_filter_id': f.extended_filter_id}

        # return vars
        return {
            'lastID': last_survey_id,
            'localeList': [{'name': locale.name, 'locale': locale.locale, 'id': locale.id}
                           for locale in i18n_controller.get_domain_locales(availability=Locale.AVAILABLE_FOR_COMPOSE)],
            'themeList': [{
                'name': i.name,
                'module': i.module,
                'cssClass': i.css_class
            } for i in self.controller.get_themes()],
            'existingFilters': existing_filters
        }

    def get_surveys(self):
        """
        Returns a list of surveys that the current group has access to. Used for selections in forms.
        """
        return [{'label': survey.name, 'value': survey.id} for survey in self.controller.get_surveys()]

    def get_last_modified_surveys(self, limit=5):
        return [{'name': survey.name, 'id': survey.id} for survey in self.controller.get_surveys(order_by=Survey.last_updated.desc())[:limit]]

    def get_active_surveys(self):
        """
        Returns only active (with online or finished publications) surveys.
        """
        surveys = self.controller.get_surveys()
        ret = []
        for s in surveys:
            a = False
            for p in s.publications:
                if p.status in [SurveyPublication.STATUS_ONLINE, SurveyPublication.STATUS_FINISHED]:
                    a = True
                    break
            if a:
                ret.append({'label': s.name, 'value': s.id})
        return ret

    def clone(self, survey_id, new_label, make_pool_items=False):
        survey = self.controller.clone(survey_id, new_label, make_pool_items)
        return survey.id

    def has_all_qas_as_pool_items(self, survey_id):
        return self.controller.has_all_qas_as_pool_items(survey_id)

    def save(self, state, group_id=None):
        """
        Receives a state object and saves the survey using the controller
        It returns a dict of domId's with new/updated id's back to the
        application. This is interface specicic logic so should go here.
        """
        return self.controller.save_state(state, group_id)

    def mark_as_deleted(self, survey_id):
        """
        Mark the supplied survey as deleted.
        """
        return self.controller.mark_as_deleted(survey_id)

    def load(self, survey_id):
        """
        Receives a suvey_id and loads the survey using the controller.
        It returns a state of the survey
        """
        return self.controller.load_state(survey_id)

    def load_preview(self, survey_id, locale_id=None, set_id=None):
        """
        Receives a suvey_id and loads the survey for the preview.
        It returns a state of the survey with all variable replaced for the given set_id.
        """

        locale = None
        if locale_id is not None:
            locale = self.session.query(Locale).filter(Locale.id==locale_id).one()

        state = self.controller.load_state(survey_id, locale)

        manager = StateManager(self.locale, self.account_id, None)
        state_controller = SurveyStateController(self.session, manager)
        state = state_controller.load_survey_variable_replaced_state(state, set_id)
        state['survey']['branching'] = state_controller.load_branching(survey_id)

        # besides the state also return the compose context so the
        # preview can be loaded from dashboard
        return [state, self.get_compose_vars()]

    def load_publication_preview(self, publication_id, locale_id):

        locale = None
        if locale_id is not None and locale_id != '':
            locale = self.session.query(Locale).filter_by(id=locale_id).one()

        publication = self.controller.get_publications(id=publication_id)[0]
        state = self.controller.load_state(publication.survey_id, locale)
        manager = StateManager(self.locale, self.account_id, None)
        state_controller = SurveyStateController(self.session, manager)


        set_id = None
        if publication.variable_set is not None:
            set_id = publication.variable_set.id

        state = state_controller.load_survey_variable_replaced_state(state, set_id)
        state['survey']['show_page_nrs'] = publication.show_page_nrs
        state['survey']['branching'] = state_controller.load_branching(publication.survey_id)

        # besides the state also return the compose context so the
        # preview can be loaded from dashboard
        return [state, self.get_compose_vars()]

    def get_survey_locales(self, survey_id):
        survey = self.controller.get_surveys(survey_id)[0]
        return [{'label': l.name, 'value': l.id} for l in survey.available_locales]

    def get_survey_qa_items(self, survey_id=None):
        if survey_id is None or survey_id == '':
            return []
        return [{'value': i.item.id, 'label': i.label} for i in self.controller.get_qa_locales(survey_id)]






@assert_user_logged_in
def jsonrpc_view(request):
    """View that handles all rpc calls"""

    dispatcher = JsonRpcDispatcher()

    locale = request.session['config']['basesoft.web.contrib.locale']

    if request.session.has_key('user_id'):
        account_id = request.session['user_id']
    else:
        account_id = None

    if request.session.has_key('group_id'):
        group_id = request.session['group_id']
    else:
        group_id = None

    dispatcher.register_instance(ComposeRPCHandler(request.Session(), locale, account_id, group_id))

    return dispatcher.handle_request(request)

#
# survey variables
#

@login_required
def survey_variables_rpc(request):
    """View that handles rpc calls for a survey's variable sets"""

    class Handler:

        def __init__(self, session, locale, account_id):

            self.account_id = account_id
            self.session = session
            self.locale = locale

            manager = StateManager(self.locale, self.account_id, None)

            self._controller = SurveyVariableStateController(self.session, manager)

        def validate(self, form_fields):

            from basesoft.web.core.input.validation import plugins
            form_handler = FormValidator(
                fields = {
                    'name': [(plugins.Required, {})],
                    'locale_str': [(plugins.Required, {})],
                })
            return form_handler.validate(form_fields)

        def validate_variables(self, form_fields):

            from basesoft.web.core.input.validation import plugins

            fields = {}
            for vals in form_fields.items():
                fields[vals[0]] = [(plugins.Required, {})]

            form_handler = FormValidator(fields=fields)
            return form_handler.validate(form_fields)

        def save(self, variable_set_state):
            """
            returns set_id
            """
            return self._controller.save(variable_set_state)

        def list_variables(self, survey_id, locale_str):
            return self._controller.list_variables(survey_id, locale_str)

        def update_variables(self, set_id, locale_str, variables):
            return self._controller.update_variables(set_id, locale_str, variables)

        def load_variables(self, set_id, locale_str):
            return self._controller.load_variables(set_id, locale_str)

        def save_state_property(self, set_id, state_properties):
            """
            Method receives a set_id and a state properties dict.
            The state_properties dict contains state keys and values to be changed.
            """
            state = self._controller.load(set_id)

            # update state
            for key in state_properties:
                state[key] = state_properties[key]

            self._controller.save(state)
            return state

        def load_state(self, set_id):
            return self._controller.load(set_id)

        def get_variable_sets(self, survey_id):
            """
            return all variable sets for a certain survey (if set)
            """
            return [{'value': vs.id, 'label':vs.name} for vs in self._controller.get_variable_sets(survey_id=survey_id)]

    # account_id is available cuz we use the login_required decorator
    account_id = request.session['user_id']
    locale = request.session['config']['basesoft.web.contrib.locale']

    # dispatch reques
    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(Handler(request.Session(), locale, account_id))
    return dispatcher.handle_request(request)

@login_required
def itemqa_filters_rpc(request):
    """
    rpc view for the itemQA filters panel
    """

    sa_session = request.Session()
    account_id = request.session['user_id']

    fc = IQuestQAFilterController(sa_session, account_id=account_id)

    class ManagerRpc:
        """
        """
        def get_filters(self):
            """
            Return a list of all filter info, as defined in the db, tblIQuestQAFilter, column name
            @return the list of filter names to which the user has access to. The list contains objects with name, id,
                extended_filter_id
            """
            available_filters = fc.get_filters()
            filters = [{'name': f.name, 'id': f.id, 'extended_filter_id': f.extended_filter_id} for f in available_filters]
            return filters

        def load_initial_data(self, itemqa_id):
            """
            Return the list of existing filters and what filters were assigned to the itemqa that interests us
            @param itemqa_id: the id of the item we're working with
            @return: an object containing the list of filters (result of get_filters), a list of item_filters
                (result of get_item_filters) and a list of attribute names and translations
            """
            # FIXME: what's with those attributes?
            return {'filters': self.get_filters(), 'attributes': attributes}

    m = ManagerRpc()

    dispatcher = JsonRpcDispatcher()
    dispatcher.register_instance(m)
    return dispatcher.handle_request(request)


def __preview(request, _vars, survey, title):
    """
    private common preview for previewing survey and publication
    """

    licensee = survey.groups[0].group.license

    _vars['preview'] = True
    _vars['account_id'] = request.session['user_id']
    _vars['compose_name'] = survey.name

    return compose_runner(request, ['iquest_response_init'], _vars, title, licensee)

def preview_survey(request, params_str):
    """
    loads the response runner for previewing or printing
    """

    _vars = simplejson.loads(params_str)

    locale = request.session['config']['basesoft.web.contrib.locale']
    account_id = request.session['user_id']

    c = ComposeController(request.Session(), locale, account_id)
    survey = c.get_surveys(id=_vars['survey_id'])[0]
    title = 'IQuest - %s' % (survey.name)

    return __preview(request, _vars, survey, title)


def preview_publication(request, params_str):
    """
    loads the response runner for previewing or printing
    """

    _vars = simplejson.loads(params_str)

    locale = request.session['config']['basesoft.web.contrib.locale']
    account_id = request.session['user_id']
    c = ComposeController(request.Session(), locale, account_id)
    publication = c.get_publications(id=_vars['publication_id'])[0]
    title = 'IQuest - %s - %s' % (publication.survey.name, publication.label)

    _vars['publication_label'] = publication.label

    return __preview(request, _vars, publication.survey, title)

