# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.

from sqlalchemy import and_

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator, AbstractStateController, StateManager
from iqportal.iquest.core.models import IQuestItemPoolCategory, IQuestItemPoolItem, IQuestItemPoolItemHtml,\
    IQuestItemPoolItemHtmlLocale, IQuestItemPoolItemQA, IQuestItemPoolItemQALocale, IQuestItemPoolItemQAElement,\
    IQuestItemPoolItemQAElementLocale, IQuestItemPoolItemQALayoutAttribute, Survey, SurveyItemQAWeightFactor
from iqportal.iquest.core.qa_filters.controllers import IQuestItemQAFilterController
from basesoft.web.core.exceptions import SecurityException, PublicSystemException,\
    PublicUserException

from basesoft import logging
logger = logging.getLogger("iqportal.iquest.core")

class QorARemovedException(Exception):
    pass

class ItemPoolController(AbstractController):
    """
    This controller is used to manage Item Pool entries. This include categories as well as polymorphic items.
    """
    def get_categories(self, id=None, priv=IQuestItemPoolCategory.PRIV_VIEW):
        self.assert_permissions('view_iquest_pool_category')
        return self.get_privileged_object_by_id(IQuestItemPoolCategory, id, priv)

    def get_category_by_name(self, name):
        self.assert_permissions('view_iquest_pool_category')
        result = self._get_privileged_object(IQuestItemPoolCategory, IQuestItemPoolCategory.name==name,
                                             IQuestItemPoolCategory.PRIV_VIEW)
        if len(result)==0:
            return None
        return result.pop()

    def get_items(self, id=None, type=None):
        if type is None:
            type = IQuestItemPoolItem

        cids = [c.id for c in self.get_categories()]
        filter = type.pool_category_id.in_(cids)

        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, type.id.in_(id))

        return self.sa_session.query(type).filter(filter).all()

    def get_elements(self, id, is_question=None, parent_item=None):
        q = self.sa_session.query(IQuestItemPoolItemQAElement)
        filter = IQuestItemPoolItemQAElement.id==id

        if is_question is not None:
            filter = and_(filter, IQuestItemPoolItemQAElement.is_question==is_question)

        if parent_item is not None:
            filter = and_(filter, IQuestItemPoolItemQAElement.pool_item==parent_item)

        return q.filter(filter).all()

    def get_items_by_category(self, category_id, type=None):
        category = self.get_categories(category_id)[0]
        if type is None:
            type = IQuestItemPoolItem
        return self.sa_session.query(type).filter(type.pool_category==category).all()

    @transaction_decorator
    def save_or_update_category(self, id, name):
        if id is None:
            self.assert_permissions('create_iquest_pool_category')
            category = IQuestItemPoolCategory()
        else:
            self.assert_permissions('update_iquest_pool_category')
            category = self.get_categories(id, IQuestItemPoolCategory.PRIV_UPDATE)[0]
        category.name = name
        if id is None:
            self.save_default_privilege_object_group(category)
        self.sa_session.add(category)
        return category

    @transaction_decorator
    def update_pool_item(self, id, name, category_id):
        """
        Save name and/or category for a pool item
        """
        item = self.get_items(id)
        if len(item) == 0:
            raise SecurityException('Invalid item id %s for account %s' % (id, self.account_id))
        item = item[0]
        if item.pool_category_id != category_id:
            category = self.get_categories(category_id)
            if len(category) == 0:
                raise SecurityException('Invalid category id %s for account %s' % (category_id, self.account_id))
            item.pool_category = category[0]
        item.name = name
        self.sa_session.add(item)
        return item

    def save_standalone_item(self, category, state):
        """
        This method uses the normal save method but it also commits the delete pool and returns the id map
        """
        # for now this can only be used to save new pool items, so make sure all dbIDs are None
        state = self.remove_item_state_dbIDs(state)

        AbstractStateController.reset_state_map()
        poolitem = self.save_or_update_item(category, state)
        if poolitem is None:
            return {}
        else:
            AbstractStateController.commit_delete_pool(self.sa_session)
            return AbstractStateController.save_state_id_map

    def save_or_update_item(self, category, state):
        if not isinstance(category, IQuestItemPoolCategory):
            category = self.get_categories(category)
        try:
            ItemController = item_controller_mapping[state['objectName']]
        except KeyError:
            logger.error('ItemPoolController: Ignoring item saving. No mapper found for %s' % state['objectName'])
        else:
            manager = StateManager(self.locale, self.account_id, None)
            ic = ItemController(self.sa_session, manager)
            return ic.save(category, state, None)
        return None

    def load_item_state(self, id):
        item = self.get_items(id)[0]
        try:
            ItemController = item_controller_mapping[item.polymorphic_type]
        except KeyError:
            logger.error('ItemPoolController: Ignoring item loading. No mapper found for %s' % item.polymorphic_type)
        else:
            manager = StateManager(self.locale, self.account_id, None)
            ic = ItemController(self.sa_session, manager)
            return ic.load(id)
        return {}

    def remove_item_state_dbIDs(self, state):
        state['dbID'] = None
        # also for item Elements (questions and answers)
        if state.has_key('children'):
            for key in state['children']:
                childs = state['children'][key]
                for child in childs:
                    child['dbID'] = None
        # and filters
        if state.has_key('filters'):
            for filter in state['filters']:
                filter['questionDbID'] = None
                filter['answerDbID'] = None
        return state

class IQuestItemStateController(AbstractStateController):
    """
    Abstract item controller.
    """
    _container = 'pool_category'
    _container_class = IQuestItemPoolCategory

    @transaction_decorator
    def save(self, item, container, state, position):
        assert isinstance(container, self._container_class)
        self._assert_state(state, None, dict)

        setattr(item, self._container, container)
        if position is not None:
            assert isinstance(position, int)
            item.position = position
        if hasattr(item, 'name') and state.has_key('poolItemName'):
            item.name = state['poolItemName']

        if hasattr(item, 'pool_item') and state.has_key('poolItemID'):
            if state['poolItemID'] is None:
                item.pool_item = None
            else:
                pc = ItemPoolController(self.sa_session, self.locale, self.account_id)
                pool_item = pc.get_items(state['poolItemID'])
                if len(pool_item) > 0:
                    item.pool_item = pool_item[0]
                else:
                    # Raise exception
                    raise SecurityException('No PoolItem found for id: %s, item: %s' % (state['poolItemID'], item))

        # call parent save that actually writes to the db
        AbstractStateController.save(self, item, state)

        # expire
        self.sa_session.expire(container)

    def load(self, id):
        pass

class IQuestItemHtmlStateController(IQuestItemStateController):

    """
    Controller for the ItemHtml model
    """

    _instance_class = IQuestItemPoolItemHtml
    _locale_attributes = ['html']
    _locale_model_class = IQuestItemPoolItemHtmlLocale

    @transaction_decorator
    def save(self, container, state, position):
        self._assert_state(state, 'objectName', unicode, 'SurveyItemHtml')

        # call super to get/create the SurveyItem and set its properties
        item = AbstractStateController._instance_from_state(self, state)
        IQuestItemStateController.save(self, item, container, state, position)

        # save locales in locale table
        self._save_locale(item, self._locale_model_class, state)

        self.sa_session.expire(item)
        return item

    def load(self, id):
        item = self._instance_from_id(id)

        state = {
            "objectName": u"SurveyItemHtml",
            "dbID": item.id,
            "poolItemID": (hasattr(item, 'pool_item') and item.pool_item is not None) and item.pool_item.id or None,
            'locales': self._load_locales(item),
            'isWiki': item.is_wiki,
        }

        return state


class IQuestItemQAStateController(IQuestItemStateController):
    """
    Controller for the QAItem
    """

    _instance_class = IQuestItemPoolItemQA
    _locale_attributes = ['label']
    _locale_model_class = IQuestItemPoolItemQALocale
    _layout_attribute_class = IQuestItemPoolItemQALayoutAttribute
    _layout_attribute_item_column = 'pool_item_id'

    def __init__(self, session, manager):
        AbstractStateController.__init__(self, session, manager)
        self.qaStateController = IQuestItemElementController(self.sa_session, manager)
        self.filterController = IQuestItemQAFilterController(self.sa_session, self.locale, self.account_id, manager)

    @transaction_decorator
    def save(self, container, state, position):

        self._assert_state(state, 'children', dict)
        self._assert_state(state, 'questionsAlign', int)
        self._assert_state(state, 'answersAlign', int)
        self._assert_state(state['children'], 'questions', list)
        self._assert_state(state['children'], 'answers', list)
        self._assert_state(state, 'type', int)
        self._assert_state(state, 'trailingQuestions', bool)
        self._assert_state(state, 'showTitle', bool )
        if state.has_key('sum_questions'):
            self._assert_state(state, 'sum_questions', bool)
        if state.has_key('sum_answers'):
            self._assert_state(state, 'sum_answers', bool)
        if state.has_key('validateMarketShare'):
            self._assert_state(state, 'validateMarketShare', bool )
        else:
            state['validateMarketShare'] = False
        self._assert_state(state, 'layout', (int, long, None) )
        self._assert_state(state, 'layoutAttributes', dict )
        self._assert_state(state, 'required', bool )
        self._assert_state(state, 'filters', list)
        self._assert_state(state, 'objectName', (str, unicode))
        assert state['objectName'] in self._instance_class.available_object_names
        questionsAlign = state['questionsAlign']
        answersAlign = state['answersAlign']

        #NPS QA needs specific permission and must always have 11 answers
        if state['objectName'] == u'ItemQANPS':
            self.assert_permissions('iquest_nps_access')
            assert len(state['children']['answers']) == 11

        ALIGN_TOP = self._instance_class.ALIGN_TOP
        ALIGN_RIGHT = self._instance_class.ALIGN_RIGHT
        ALIGN_BOTTOM = self._instance_class.ALIGN_BOTTOM
        ALIGN_LEFT = self._instance_class.ALIGN_LEFT

        # assert alignment
        try :
            assert (
                (answersAlign in (ALIGN_TOP, ALIGN_BOTTOM) and questionsAlign not in (ALIGN_TOP, ALIGN_BOTTOM)) or
                (answersAlign in (ALIGN_RIGHT, ALIGN_LEFT) and questionsAlign not in (ALIGN_RIGHT, ALIGN_LEFT))
            )
        except AssertionError:
            raise PublicSystemException('Incorrect question/answer alignment in state %s' % state)

        item_qa = AbstractStateController._instance_from_state(self, state)
        item_qa.type = state['type']
        item_qa.object_name = state['objectName']
        item_qa.questions_align = state['questionsAlign']
        item_qa.answers_align = state['answersAlign']
        item_qa.trailing_questions = state['trailingQuestions']
        item_qa.show_title = state['showTitle']
        item_qa.validate_market_share = state['validateMarketShare']

        if state.has_key('sum_questions'):
            item_qa.sum_questions = state['sum_questions']
        if state.has_key('sum_answers'):
            item_qa.sum_answers = state['sum_answers']

        if state.has_key('alwaysShowQuestionLabels'):
            item_qa.always_show_question_labels = state['alwaysShowQuestionLabels']
        if state.has_key('alwaysShowAnswerLabels'):
            item_qa.always_show_answer_labels = state['alwaysShowAnswerLabels']

        item_qa.layout = state['layout']
        item_qa.required = state['required']

        IQuestItemStateController.save(self, item_qa, container, state, position)

        # save layout attributes
        attrs = self.sa_session.query(self._layout_attribute_class).filter(
                getattr(self._layout_attribute_class, self._layout_attribute_item_column)==item_qa.id).all()
        keys = []
        for attr in attrs:
            key = "%s" % attr.key_id
            keys.append(key)
            # remove from db if an attr is not in state
            if not state["layoutAttributes"].has_key(key):
                self.sa_session.delete(attr)
            else: # else update the value
                attr.value = state["layoutAttributes"][key]
                self.sa_session.add(attr)
        # now add all the attrs that are new
        for k, v in state["layoutAttributes"].items():
            try:
                i = keys.index(k)
            except ValueError:
                attr = self._layout_attribute_class()
                if hasattr(attr, 'item'):
                    attr.item = item_qa
                else:
                    attr.pool_item = item_qa
                attr.key_id = k
                attr.value = v
                self.sa_session.add(attr)
        self.sa_session.flush()

        #locales
        self._save_locale(item_qa, self._locale_model_class, state)

        # process questions
        self._delete_children(item_qa.questions, state['children']['questions'])
        try:
            assert len(state['children']['questions']) > 0
        except AssertionError:
            raise PublicSystemException('ItemQA %(iqa)s must contain atleast one question' % {'iqa': item_qa})

        for i in range(len(state['children']['questions'])):
            self.qaStateController.save(item_qa, state['children']['questions'][i], i, True)

        # process answers
        self._delete_children(item_qa.answers, state['children']['answers'])
        try:
            assert len(state['children']['answers']) > 0
        except AssertionError:
            raise PublicSystemException('ItemQA %(iqa)s must contain atleast one answer' % {'iqa': item_qa})

        for i in range(len(state['children']['answers'])):
            self.qaStateController.save(item_qa, state['children']['answers'][i], i, False)

        if self.has_permissions('update_itemqa_filter') and self.has_permissions('delete_itemqa_filter'):
            self._save_item_filters(state, item_qa)

        if self.has_permissions('manage_itemqa_weight_factor'):
            self._save_weight_factors(state, item_qa)


        self.sa_session.expire(item_qa)
        return item_qa

    def __get_q_and_a_id_from_substate(self, s, item_qa):
        question_id = s['questionDbID']
        answer_id = s['answerDbID']
        if question_id is None:
            if s.has_key('questionDomID') and s['questionDomID'] is not None:
                if not self.save_state_id_map.has_key(s['questionDomID']):
                    raise QorARemovedException()
                question_id = self.save_state_id_map[s['questionDomID']]
        if answer_id is None:
            if s.has_key('answerDomID') and s['answerDomID'] is not None:
                if not self.save_state_id_map.has_key(s['answerDomID']):
                    raise QorARemovedException()
                answer_id = self.save_state_id_map[s['answerDomID']]

        # Ticket #2523: the answer was removed. We cannot set a filter for it.
        if answer_id is not None:
            if not answer_id in [a.id for a in item_qa.answers]:
                raise QorARemovedException()
        # Ticket #2523: the question was removed. We cannot set a filter for it.
        if question_id is not None and question_id not in self.save_state_id_map:
            if not question_id in [q.id for q in item_qa.questions]:
                raise QorARemovedException()

        return question_id, answer_id

    def _save_weight_factors(self, state, item_qa):
        # we first delete all existing weight factors for this item QA
        self.sa_session.query(SurveyItemQAWeightFactor).filter(SurveyItemQAWeightFactor.item==item_qa).delete()

        # no weight factors defineds
        if not state.has_key('weight_factors'):
            return

        for s in state['weight_factors']:
            try:
                question_id, answer_id = self.__get_q_and_a_id_from_substate(s, item_qa)
            except QorARemovedException:
                continue

            sw = SurveyItemQAWeightFactor()
            sw.item = item_qa
            sw.question_id = question_id
            sw.answer_id = answer_id
            sw.weight = s['weight']
            self.sa_session.add(sw)

    def _save_item_filters(self, state, item_qa):
        # process filters only if the user can update and delete them
        # delete the existing item filters for this itemQA, so we can have our new ones
        existing_item_filters = self.filterController.get_item_filters(item_qa.id)
        if existing_item_filters:
            self.filterController.delete_item_filters(existing_item_filters)

        # save the new filters
        for f in state['filters']:
            # if we receive questionDbID, it means we had the filter saved previously and we have the same id for
            # question. if we don't receive this, we look at questionDomID, that can be null, which means we
            # attached the filter globally or to an answer, or if it's not null, we have filter attached to a
            # new question, or cell
            try:
                question_id, answer_id = self.__get_q_and_a_id_from_substate(f, item_qa)
            except QorARemovedException:
                continue

            # get the filter object so we can use it for saving the filters
            filters = self.filterController.qaFilterController.get_filters(f['dbID'])
            if not filters:
                raise PublicUserException(_("You are trying to use a non-existing filter or one you don't have access to."))
            _filter = filters[0]

            self.filterController.save_or_update_item_filter(None, _filter, item_qa,attributes=f['attributes'],
                                                             question_id=question_id, answer_id=answer_id)

    def load(self, id):
        item = self._instance_from_id(id)
        la = {}
        for attr in item.layout_attributes:
            la[attr.key_id] = attr.value

        if isinstance(item, IQuestItemPoolItemQA):
            id = None
            pool_item_id = item.id
        else:
            id = item.id
            pool_item_id = (hasattr(item, 'pool_item') and item.pool_item is not None) and item.pool_item.id or None

        state = {
            "objectName": item.object_name,
            "dbID": id,
            "poolItemID": pool_item_id,
            "type": item.type,
            "questionsAlign": item.questions_align,
            "answersAlign": item.answers_align,
            "trailingQuestions": bool(item.trailing_questions),
            "locales": self._load_locales(item),
            "showTitle": bool(item.show_title),
            "validateMarketShare": bool(item.validate_market_share),
            "sum_questions": bool(item.sum_questions),
            "sum_answers": bool(item.sum_answers),
            "alwaysShowQuestionLabels": bool(item.always_show_question_labels),
            "alwaysShowAnswerLabels": bool(item.always_show_answer_labels),
            "layout": item.layout,
            "layoutAttributes": la,
            "required": bool(item.required),
            "filters": [],
            "weight_factors": [],
            "children": {
                'questions': [],
                'answers': []
            }
        }

        # TODO: if a pool item, validate the number of questions and answers to match ! - #3275

        for element in item.questions:
            state['children']['questions'].append(self.qaStateController.load(element.id))
        for element in item.answers:
            state['children']['answers'].append(self.qaStateController.load(element.id))

        weights = self.sa_session.query(SurveyItemQAWeightFactor).filter(SurveyItemQAWeightFactor.item_id==id).all()
        for w in weights:
            state['weight_factors'].append({
                'questionDbID': w.question_id,
                'answerDbID': w.answer_id,
                'weight': w.weight,
            })

        add_filters = False
        # pool items don't have a mode
        if self.manager.mode is None:
            if self.has_permissions('view_itemqa_filter'):
                add_filters = True
        elif self.manager.has_mode(self.manager.STRICT_MODE_IS_EQUAL, Survey.MODE_RESPONSE) or\
            (self.manager.has_mode(self.manager.STRICT_MODE_IS_EQUAL, Survey.MODE_COMPOSE) and\
             self.has_permissions('view_itemqa_filter')):
            add_filters = True

        if add_filters:
            # list of filter configurations
            filters = []
            existing_item_filters = self.filterController.get_item_filters(item.id)
            for e in existing_item_filters:
                attributes = {}
                for a in e.attributes:
                    attributes[a.attribute] = a.value
                filters.append({
                    'dbID': e.filter_id,
                    'questionDbID': e.question_id,
                    'answerDbID': e.answer_id,
                    'attributes': attributes
                })
            state['filters'] = filters
        return state

class IQuestItemElementController(AbstractStateController):
    """
    Shares common state logic between questions and answers
    """
    _instance_class = IQuestItemPoolItemQAElement
    _locale_attributes = ['label', 'trailing_label']
    _locale_model_class = IQuestItemPoolItemQAElementLocale
    _objectName = u'SurveyItemQAElement'
    _item_class = IQuestItemPoolItemQA

    @transaction_decorator
    def save(self, item_qa, state, position, is_question):

        assert isinstance(item_qa, self._item_class)
        assert state.has_key('objectName')
        assert state['objectName'] in [u'SurveyQuestion', u'SurveyAnswer']

        q_or_a = AbstractStateController._instance_from_state(self, state)
        if hasattr(q_or_a, 'item'):
            q_or_a.item = item_qa
        else:
            q_or_a.pool_item = item_qa
        q_or_a.position = position
        q_or_a.is_question = is_question

        if hasattr(q_or_a, 'pool_element') and state.has_key('poolElementID'):
            if state['poolElementID'] is None:

                # we should have a poolElementID in case the item_qa has a PoolItem.
                if item_qa.pool_item is not None:
                    raise SecurityException('No PoolElement found while QA has a PoolItem parent: %s' % item_qa.pool_item)

                q_or_a.pool_element = None
            else:
                pc = ItemPoolController(self.sa_session, self.locale, self.account_id)
                pool_element = pc.get_elements(state['poolElementID'], is_question, item_qa.pool_item)
                if len(pool_element) > 0:
                    q_or_a.pool_element = pool_element[0]
                else:
                    # Raise exception
                    raise SecurityException('No PoolElement found for id: %s, question: %s, parent: %s' % (state['poolElementID'], is_question, item_qa.pool_item))

        # call parent save that actually writes to the db
        AbstractStateController.save(self, q_or_a, state)

        # save locales
        self._save_locale(q_or_a, self._locale_model_class, state)

        # expire the QA
        self.sa_session.expire(item_qa)

    def load(self, id):
        q_or_a = self._instance_from_id(id)

        if isinstance(q_or_a, IQuestItemPoolItemQAElement):
            id = None
            pool_element_id = q_or_a.id
        else:
            id = q_or_a.id
            pool_element_id = (hasattr(q_or_a, 'pool_element') and q_or_a.pool_element is not None) and q_or_a.pool_element.id or None

        state = {
            "objectName": q_or_a.is_question and u'SurveyQuestion' or u'SurveyAnswer',
            "dbID": id,
            "poolElementID": pool_element_id,
            'locales': self._load_locales(q_or_a),
        }

        return state

item_controller_mapping = {
    'SurveyItemHtml': IQuestItemHtmlStateController,
    'SurveyItemQA': IQuestItemQAStateController,
    'IQuestItemPoolItemHtml': IQuestItemHtmlStateController,
    'IQuestItemPoolItemQA': IQuestItemQAStateController,

    # specific type QAs - single choice QA, but client side needs specific checks
    'ItemQANPS': IQuestItemQAStateController
}
