# Copyright (c) 2008 BaseSoft B.V.
# See LICENSE for details.

"""
This file contains methods which will be executed by the scheduler on a regular interval. They take care of time related
issues like publishing and finishing of surveys.

Notice that they do not run in a 'web session' so they should not depend on anything like that. Think of it that they
run as a 'root' user.

This file contains two subclasses of ScheduleTask and are added to the 'tasks' list in this same module. The scheduler
loads these tasks and instanciates them for each interval.
"""

from datetime import datetime

from sqlalchemy.sql import and_

from basesoft.web.contrib.compose import schedule as compose_schedule
from basesoft.web.contrib.compose.models import ComposePublicationMailing
from iqportal.iquest.core.models import SurveyPublication
from iqportal.iquest.response.controllers import ResponseController
from basesoft import logging
logger = logging.getLogger("iquest.schedule")

class PublishSurveysTask(compose_schedule.PublishComposeTask):
    INTERVAL = 600 # every ten minutes
    model = SurveyPublication

    def execute(self):
        """
        Overwrite in order to set survey to stale
        """

        rv = []

        # get all scheduled publications in the current period
        result = self._get_publications()

        surveys = []
        for publication in result:
            logger.info('PublishSurveysTask: Publication [%s] changing status to STATUS_ONLINE.' % (publication.id))
            rv.append({'publication_id': publication.id, 'msg': 'changed status to STATUS_ONLINE'})

            publication.status = self.model.STATUS_ONLINE

            self._create_anonymous_invitee(publication)

            self.sa_session.add(publication)

            if publication.survey not in surveys:
                surveys.append(publication.survey)

        # set all surveys that went online to stale - so the aggregators will fill correspondent tables
        for s in surveys:
            s.stale = 1
            self.sa_session.add(s)

        self.sa_session.flush()

        # immediately send mails to invitees
        # this will also send reminder emails for online publications
        PublicationMailing().execute()

        return rv

class CreatePublicationInviteesTask(compose_schedule.CreatePublicationInviteesTask):
    INTERVAL = 90
    model = SurveyPublication
    controller = ResponseController

class FinishSurveysTask(compose_schedule.FinishPublicationsTask):
    model = SurveyPublication

class PublicationMailing(compose_schedule.PublicationMailing):
    """
    Implements a MailingTask that takes care of sending out mails to invitees.
    This will be ran immediately after the PublishSurveysTask has finished (it will not start automaticly).
    """
    publication_model = SurveyPublication

    def _filter_contact(self, mailing, contact):
        """
        This filters the contacts based on the applied filter.
        """
        if mailing.filter == ComposePublicationMailing.FILTER_NONE:
            return False

        rc = ResponseController(self.sa_session)
        is_respondent = rc.is_respondent(contact, mailing.publication)
        if mailing.filter == ComposePublicationMailing.FILTER_NON_RESPONDENTS:
            return not is_respondent
        else:
            return is_respondent

# publish tasks so the scheduler can pick them up
tasks = (PublishSurveysTask, CreatePublicationInviteesTask, FinishSurveysTask, )
