# Copyright (c) 2009 BaseSoft B.V.
# See LICENSE for details.

"""
Reporters are definitions of Reports based on the provided publication.
"""

from django.conf import settings
from django.utils.translation import ugettext as _

from sqlalchemy.sql.expression import and_

from iqportal.iqube.core.reporters import ConfigurableWidgetReporter
from iqportal.iqube.core.models import IncorrectCalculatedMember
from iqportal.iqreport.core.models import Report

from basesoft.web.core.exceptions import PublicUserException, SystemException, SecurityException, PermissionDeniedException

class AbstractReporter:
    """
    Abstract Reporter should be extended by the different Reporters.
    """
    def __init__(self, sa_session, account, publication):
        self.sa_session = sa_session
        self.account = account
        self.publication = publication
        self.survey = publication.survey

class DefaultIQReporter(AbstractReporter):
    """
    The Default IQReporter is used for creating IQReport based reports. They should be able to
    return the state of a report that can be saved with the Compose Controller of IQReport.
    """
    def get_state(self):
        """
        Get the state based on the definitions of the blocks, columns and items.
        """

        state = {
            'objectName': 'Report',
            'dbID': None,
            'name': 'Default Report for %s' % self.survey.name,
            'availableLocales': [locale.locale for locale in self.survey.available_locales],
            'fallbackLocale': {
                'locale': self.survey.fallback_locale.locale,
                'name': self.survey.fallback_locale.name
            },
            'children': {
                'blocks': self._get_blocks_state()
            }
        }

        return {'report': state}

    def _get_datasource(self):
        """
        Get the datasource for IQuest projects.
        """
        from iqportal.iqube.core.controllers import DataSourceController
        from basesoft.web.core.exceptions import PublicSystemException
        dsc = DataSourceController(self.sa_session, account_id=self.account)
        ds = dsc.get_datasource_by('mondrian_project', '%s.IQuest' % settings.SETTINGS_MODULE)
        if len(ds) == 0:
            raise PublicSystemException('No datasource found for IQuest')

        return ds.pop()

    def _get_blocks_state(self):
        """
        Implement this method in the child classes
        """
        return []

class DefaultReporter(DefaultIQReporter):
    """
    The Default Reporter is a simple IQReporter, it only has one block with a full width column.
    All items are added to the column with their correct MDX Statement.
    """

    def _get_blocks_state(self):
        """
        Get the blocks for this report, one block, one column.
        """
        # we want one block, with one column, with all Result Items.
        return [{
            'objectName': 'ReportBlock',
            'dbID': None,
            'children': {
                'columns': [{
                    'objectName': 'ReportColumn',
                    'dbID' : None,
                    'width': 100,
                    'children': {
                        'items': self._get_items_state()
                    }
                }]
            }
        }]

    def _get_items_state(self):
        """
        Loop through the qa items of the survey and add them as items to the state.
        """
        items = []
        for qa in self.survey.itemQAs:
            items.append(self._get_item_default(qa))

        return items

    def get_mdx_for_item(self, item):
        """
        This method should be called before displaying. MDX should always be created server side for
        permission/privilege purposes.
        """
        from sqlalchemy.orm.exc import NoResultFound
        from iqportal.iquest.core.models import SurveyItemQA, SurveyItemQAFilter, IQuestQAFilter

        # Check the type, check the filters.
        measures = ['Response Responses Int Sum']
        numeric = False
        if item.type == SurveyItemQA.TYPE_SINGLE_TEXT_INPUT:
            measures = ['Response Response Count']

        # check the filters; global only, we do not want the complexity to display the default report
        # with row/column or even cell based filters.
        try:
            filter = self.sa_session.query(SurveyItemQAFilter).join(IQuestQAFilter).filter(
                SurveyItemQAFilter.item_id==item.id).filter(
                SurveyItemQAFilter.question_id==None).filter(
                SurveyItemQAFilter.answer_id==None).filter(
                IQuestQAFilter.name=='Numeric').one()

            type = None
            for attr in filter.attributes:
                if attr.attribute != 'domain':
                    continue
                if attr.value == 'int':
                    type = 'Int'
                    break
                type = 'Float'
                break
            if type is None:
                raise Exception('missing domain for numberic attribute')
            numeric = True
            measures = ['Response Responses %s %s' % (type, i) for i in ('Sum', 'Avg')]
            measures.insert(0, 'Response Response Count')
        except NoResultFound: # no filter found, nothing to see, move along
            pass

        cols = []
        for m in measures:
            cols.append({'member_unique_name': u'[Measures].[%s]' % m, 'hierarchy_unique_name': '[Measures]', 'depth': '0'})

        locale_id = self.survey.fallback_locale.id

        mdx = {
            'cube': 'VirtualResponse',
            'columns': [{'member_unique_name': '[Question].[%s].[%s].Children' % (item.id, locale_id), 'hierarchy_unique_name': '[Question items]', 'depth': '0'}] + cols,
            'rows': [{'member_unique_name': '[Answer].[%s].[%s].Children' % (item.id, locale_id), 'hierarchy_unique_name': '[Answer items]', 'depth': '0'}],
            'slicers': [{'member_unique_name': '[Publication].[%s]' % self.publication.id, 'hierarchy_unique_name': '[Publication]', 'depth': '1'}],
            'nonempty': False
        }

        return (numeric, mdx)

    def _get_item_default(self, qa):
        """
        Get the default qa item setup. Super simple MDX statement.
        """
        graph_obj = {'config': {'show_title': True}, 'module': 'iquest_graph_table', 'name': 'iquest_table'}

        numeric, mdx = self.get_mdx_for_item(qa)
        if numeric:
            graph_obj = {'config': {'show_title': True}, 'module': 'iquest_graph_numericaltable', 'name': 'iquest_numerical_table'}

        title = None
        for locale in qa.locales:
            if locale.locale == self.survey.fallback_locale:
                title = locale.label
                break

        if not qa.show_title or (not qa.always_show_question_labels and len(qa.questions) == 1):
            graph_obj['config']['show_title'] = False

        rv = {
            'objectName': 'ReportItemResult',
            'dbID': None,
            'datasource_id': self._get_datasource().id,
            'mdx_statement': mdx,
            'title': title,
            'graphs': [graph_obj]
        }

        return rv

    def __repr__(self):
        return '<DefaultReporter>'

class IQuestReporter(ConfigurableWidgetReporter):

    RESPONSE_MEASURES = ('Responses Int Sum', 'Responses Float Sum', )
    AVERAGE_MEASURES = ('Responses Int Avg', 'Responses Float Avg', )
    RESPONDENT_AVERAGE_MEASURES = ('Responses Int Avg per Respondent', 'Responses Float Avg per Respondent', )
    RESPONDENTS_MEASURES = ('Respondents', )
    COMPANIES_MEASURES = ('Companies', )
    RESPONSE_COUNT_MEASURES = ('Response Count', )
    WEIGHTED_MEASURES = ('Weighted Int Sum', 'Weighted Float Sum', )
    WEIGHTED_AVERAGE_MEASURES = ('Weighted Int Avg', 'Weighted Float Avg', )

    _individual_measure = None
    _show_individual = False
    _use_measure = None
    _basket_analysis = None

    def __init__(self, sa_session, datasource, account_id=None, config=None):
        ConfigurableWidgetReporter.__init__(self, sa_session, datasource, account_id, config)
        if self._config.has_key('run_mode') and self._config['run_mode']:
            self._config['cube'] = 'VirtualFilteredResponse'

    def process_market_share(self, applied_config, field, measures_member, fmt_string):
        # this gets called by the parents' process_measures
        # we are not showing market share, no reason to calculate anything.
        if not applied_config.has_key('show_%s_measure' % field) or not applied_config['show_%s_measure' % field]:
            return

        if not self._individual_measure:
            return

        dec = 1
        if applied_config.has_key('market_share_decimals') and applied_config['market_share_decimals']:
            try:
                dec = int(applied_config['market_share_decimals'])
            except ValueError:
                pass

        fmt_string = ", FORMAT_STRING='0.%s%%'" % ('0' * dec)

        self._extra_with.append("MEMBER "+measures_member+" as '("+self._individual_measure+"/"+self._use_measure+")'"+fmt_string)
        self.all_measures.append(measures_member)
        self.show_measures[field] = measures_member

    def process_context(self):
        if self._individual_measure is not None and self._show_individual:
            return

        # if the reporter is not configured to show by individual response, we should remove the contact from the context.
        # this way we don't store duplicates on a per contact base
        if self._config.has_key('context') and self._config['context'].has_key('contact'):
            self._config['context']['contact'] = None

    def _set_basket_analysis(self, applied_config, options):
        # setting simple basket analysis based on one QA's question and answer.
        fields = ['basket_analysis_answers', 'basket_analysis_questions', 'basket_analysis_qa']

        # All fields are required for now in order to apply the basket analysis:
        for i in fields:
            if not applied_config.has_key(i) or applied_config[i] is None or applied_config[i] == '':
                return
            if type(applied_config[i]) in [str, unicode]:
                applied_config[i] = [self.deserialize(applied_config[i])]
            if len(applied_config[i]) != 1: # we can only do basket analysis at the moment with one q and one a.
                return # TODO: raise exception

        q = applied_config['basket_analysis_questions'][0]['member_unique_name']
        a = applied_config['basket_analysis_answers'][0]['member_unique_name']
        # we also need to use the Item itself to get results in case item is also in the rest of the MDX statement.
        qa = applied_config['basket_analysis_qa'][0]['member_unique_name']

        self._extra_with.append("SET [BasketAnalysis] AS '%s * %s * %s'" % (qa, q, a))

        self._basket_analysis = {}

        default = "SUM([Participant].Children, IIF(SUM([BasketAnalysis]) > 0, %s, 0))"

        def get_config(o):

            measures, label = o

            members = None
            for m in measures:
                m = '[Measures].[%s]' % m
                if members is None:
                    members = m
                    continue
                members = members + '+' + m
            return members, label

        members, rlabel = get_config(options['response'])
        options['response'] = (default % members, rlabel, )

        members, clabel = get_config(options['response_count'])
        options['response_count'] = (default % '1', clabel, )

        members, label = get_config(options['average'])
        options['average'] = ("[Measures].[%s] / [Measures].[%s]" % (rlabel, clabel), label, )

        # TODO: fix weighted for basket analysis
        #members, rlabel = get_config(options['weighted'])
        #options['weighted'] = (default % members, rlabel, )

        #members, label = get_config(options['weighted_average'])
        #options['weighted_average'] = ("[Measures].[%s] / [Measures].[%s]" % (rlabel, clabel), label, )

        # TODO: not sure if respondents and companies can work like this.
        members, respondentlabel = get_config(options['respondents'])
        options['respondents'] = (default % members, respondentlabel, )

        members, clabel = get_config(options['companies'])
        options['companies'] = (default % members, clabel, )

        members, label = get_config(options['respondent_average'])
        options['respondent_average'] = ("[Measures].[%s] / [Measures].[%s]" % (rlabel, respondentlabel), label, )

    def _set_individual(self, applied_config, measures_member, fmt_string, field):
        if not self._use_measure:
            raise SystemException('No measure to use for individual')

        if self._individual_measure is not None:
            return

        self._individual_measure = measures_member

        # we need the correct permissions to use this option in compose mode
        if not self._config.has_key('run_mode') or not self._config['run_mode']:
            if not self.has_permissions('create_individual_reports'):
                return

            # check if user has correct privileges for the supplied company
            if applied_config.has_key('use_individual_company') and applied_config['use_individual_company'] not in ['', None]:
                from basesoft.web.contrib.contact.models import Contact, ContactGroup
                # FIXME what if multiple contacts with same company?

                filters, fo = self.get_query_filter_on_privileges(Contact, ContactGroup, Contact.PRIV_VIEW)
                filters = and_(Contact.company_id==applied_config['use_individual_company'], filters)

                c = self.sa_session.query(Contact).select_from(fo).filter(filters).all()
                if len(c) == 0:
                    raise PermissionDeniedException(_('No privileges on contact'))

        # if we are not showing this measure, no need to check.
        elif not applied_config.has_key('show_%s_measure' % field) or not applied_config['show_%s_measure' % field]:
            applied_config['use_individual_company'] = None

        elif self._config.has_key('context') and self._config['context'].has_key('contact'):
            if self._config['context']['contact'] is None or self._config['context']['contact'].company is None:
                raise PublicUserException(_('Widget unavailable for current viewer'))
            applied_config['use_individual_company'] = self._config['context']['contact'].company.id

        m = self._use_measure

        # the user of IQReport can select an individual Company to verify data, this happens in the preview runner
        # this can also be done from the IQuest Reporters global filter panel in compose mode.
        if applied_config.has_key('use_individual_company') and applied_config['use_individual_company']:
            self._extra_with.append("MEMBER %s as '([Company].[%s], %s)'%s" % (measures_member, applied_config['use_individual_company'], m, fmt_string))
        else:
            self._extra_with.append("MEMBER %s as '0'%s" % (measures_member, fmt_string))

        self.all_measures.append(measures_member)

    def process_individual(self, applied_config, field, measures_member, fmt_string):
        # this gets called by the parents' process_measures
        self._set_individual(applied_config, measures_member, fmt_string, field)

        if applied_config.has_key('show_%s_measure' % field) and applied_config['show_%s_measure' % field]:
            self._show_individual = True
            self.show_measures[field] = measures_member

    def _add_measures_with(self, applied_config, use_measure='response', additional=None):

        # these are default IQuest fields that are available from all widgets (based on permissions)
        fields = ['response', 'average', 'respondent_average', 'weighted_average', 'weighted', 'respondents', 'companies', 'response_count', 'individual', 'market_share']

        options = {
            'response': (IQuestReporter.RESPONSE_MEASURES, 'Response', ), # merge measures from float & int
            'average': (IQuestReporter.AVERAGE_MEASURES, 'Average', ), # merge measures from float & int
            'respondent_average': (IQuestReporter.RESPONDENT_AVERAGE_MEASURES, 'Average per Respondent', ), # merge measures from float & int
            'weighted_average': (IQuestReporter.WEIGHTED_AVERAGE_MEASURES, 'Weighted Average', ), # merge measures from float & int
            'weighted': (IQuestReporter.WEIGHTED_MEASURES, 'Weighted', ), # merge measures from float & int
            'respondents': (IQuestReporter.RESPONDENTS_MEASURES, '# of Respondents', ),
            'companies': (IQuestReporter.COMPANIES_MEASURES, '# of Companies', ),
            'response_count': (IQuestReporter.RESPONSE_COUNT_MEASURES, '# of Responses', ),
            'individual': (None, 'Individual', ),
            'market_share': (None, 'Market share', )
        }

        # if optional additional fields are supplied, add them to the list and dict.
        if additional is not None:
            for i, settings in additional.iteritems():
                fields.append(i)
                options[i] = settings

        self._set_basket_analysis(applied_config, options)

        # process the measures (parent object).
        self.process_measures(applied_config, fields, options, use_measure)
        self.process_context()

class DefaultQAItemReporter(IQuestReporter):

    def get_mdx_config(self, config, applied_config):
        applied_config['hierarchize'] = False

        self._extra_with = []
        self._add_measures_with(applied_config)

        totals = {
            'question_total': ('Question', 'SUM', 'Total', ),
            'answer_total': ('Answer', 'SUM', 'Total',  ),
            'question_average': ('Question', 'AVG', 'Average',  ),
            'answer_average': ('Answer', 'AVG', 'Average',  ),
        }
        members = {}
        for t, settings in totals.iteritems():
            i, operator, label = settings

            if not applied_config.has_key(t) or not applied_config[t]:
                continue

            if applied_config.has_key('%s_label' % t) and applied_config['%s_label' % t]:
                label = applied_config['%s_label' % t]

            index = applied_config['dimensions'][i]
            if not members.has_key(i):
                members[i] = '+'.join([j['member_unique_name'] for j in applied_config['dimension_%s' % index]])

            member = '[%s].[%s]' % (i, label)
            if operator == 'SUM':
                m = "MEMBER %s as '%s'" % (member, members[i])
            else:
                m = "MEMBER %s as '(%s)/%d'" % (member, members[i], len(applied_config['dimension_%s' % index]))
            self._extra_with.append(m)
            applied_config['dimension_%s' % index].append(member)

        return ConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)

class MultiQAItemReporter(IQuestReporter):
    pass

class GrowthItemReporter(IQuestReporter):

    def process_growth(self, applied_config, field, measures_member, fmt_string):
        self.__process_growth(applied_config, field, measures_member, fmt_string)

    def process_individual_growth(self, applied_config, field, measures_member, fmt_string):
        # set the default Individual Measure. In case there already is one set it wont do it again.
        self._set_individual(applied_config, '[Measures].[Individual]', fmt_string, field)
        if applied_config.has_key('show_%s_measure' % field) and applied_config['show_%s_measure' % field]:
            self._show_individual = True
            self.__process_growth(applied_config, field, measures_member, fmt_string)

    def __process_growth(self, applied_config, field, measures_member, fmt_string):
        # TODO: make the iterator to calculate the growth over dynamic as per ticket #2821

        # This gets called from the parent's process_measure method in case we have a "growth" measure set.

        # we are not showing Growth, so nothing to do here.
        if not applied_config.has_key('show_%s_measure' % field) or not applied_config['show_%s_measure' % field]:
            return

        # format string for Growth is a percentage.
        dec = 1
        if applied_config.has_key('%s_decimals' % field) and applied_config['%s_decimals' % field]:
            try:
                dec = int(applied_config['%s_decimals' % field])
            except ValueError:
                pass

        fmt_string = ", FORMAT_STRING='0.%s%%'" % ('0' * dec)

        if not self.__prepared:
            # make a set with members to use in the growth calculation.
            survey_index = applied_config['dimensions']['Survey']
            survey_members = applied_config['dimension_%s' % survey_index]

            new_survey_members = []
            # all the survey members are required for the sets we use to calculate the growth.
            for c in survey_members:
                try:
                    new_survey_members.append(self._make_mdx_from_cm(c))
                    self._extra_with.append(str(c))
                except IncorrectCalculatedMember:
                    new_survey_members.append(self._make_mdx(c))

            # we make two sets of surveys, one to iterate over, the other to determine the previous survey.
            self._extra_with.append("SET [SurveySet] as '{%s}'" % ','.join([m['member_unique_name'] for m in new_survey_members]))
            self._extra_with.append("SET [SurveySet2] as '{%s}'" % ','.join([m['member_unique_name'] for m in new_survey_members]))
            applied_config['dimension_%s' % survey_index] = ['[SurveySet]']

        if field == 'growth':
            m = self._use_measure
        else:
            m = self._individual_measure
        self._extra_with.append("""MEMBER """+measures_member+""" AS
            '(("""+m+""", [SurveySet2].Item([SurveySet].CurrentOrdinal))-("""+m+""", [SurveySet2].Item([SurveySet].CurrentOrdinal-1)))/("""+m+""", [SurveySet2].Item([SurveySet].CurrentOrdinal-1))'"""+fmt_string)

        self.all_measures.append(measures_member)
        self.show_measures[field] = measures_member

    def get_mdx_config(self, config, applied_config):
        self.__prepared = False
        applied_config['hierarchize'] = False

        self._extra_with = []
        # We can have an additional growth item.
        self._add_measures_with(applied_config, additional={'growth': (None, 'Growth', ), 'individual_growth': (None, 'Individual growth', )})

        return ConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)

class QATotalsItemReporter(IQuestReporter):

    def get_mdx_config(self, config, applied_config):

        # having Survey .Children on slicers will error with java.lang.NullPointerException
        index = applied_config['dimensions']['Survey']
        if applied_config['axes_%s' % index] == 'slicers':
            for d in applied_config['dimension_%s' % index]:
                if d['member_unique_name'][-9:] == '.Children':
                    d['member_unique_name'] = d['member_unique_name'][:-9]

        applied_config['hierarchize'] = False

        self._extra_with = []
        self._add_measures_with(applied_config)

        if applied_config.has_key('row_total') and applied_config['row_total']:
            t = 'Total'
            if applied_config.has_key('total_label'):
                t = applied_config['total_label']

            item_totals = '[Item].[%s]' % t

            index = applied_config['dimensions']['Item']
            self._extra_with.append("MEMBER %s as '%s'" % (item_totals, '+'.join([m['member_unique_name'] for m in applied_config['dimension_%s' % index]])))
            applied_config['dimension_%s' % index].append(item_totals)

        return ConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)

class NPSItemReporter(IQuestReporter):

    def process_nps_kpi(self, applied_config, field, measures_member, fmt_string):
        if not self.__get_answers_for_nps(applied_config, field, measures_member, fmt_string):
            return

        # % in format_string automatically multiplies the expression by 100
        fmt_string = ", FORMAT_STRING='0.%s%%'" % ('0' * self.__decimals)
        all_nps_member = "MEMBER [Measures].[All NPS] AS '%s'" % '+'.join(self.__answers_mdx)
        self._extra_with.append(all_nps_member)

        member = "MEMBER "+measures_member+" AS '((%s+%s) - (%s+%s+%s+%s+%s+%s+%s))/[Measures].[All NPS]' %s" %\
            (self.__answers_mdx[9], self.__answers_mdx[10], self.__answers_mdx[0], self.__answers_mdx[1], self.__answers_mdx[2],\
             self.__answers_mdx[3], self.__answers_mdx[4], self.__answers_mdx[5], self.__answers_mdx[6], fmt_string)
        self._extra_with.append(member)
        self.all_measures.append(measures_member)
        self.show_measures[field] = measures_member

    def process_nps_detractors_perc(self, applied_config, field, measures_member, fmt_string):
        if not self.__get_answers_for_nps(applied_config, field, measures_member, fmt_string):
            return

        fmt_string = ", FORMAT_STRING='0.%s%%'" % ('0' * self.__decimals)

        all_nps_member = "MEMBER [Measures].[All NPS] AS '%s'" % '+'.join(self.__answers_mdx)
        self._extra_with.append(all_nps_member)

        member ="MEMBER "+measures_member+" AS '(%s+%s+%s+%s+%s+%s+%s)/[Measures].[All NPS]' %s" %\
                    (self.__answers_mdx[0], self.__answers_mdx[1], self.__answers_mdx[2], self.__answers_mdx[3],\
                     self.__answers_mdx[4], self.__answers_mdx[5], self.__answers_mdx[6], fmt_string)
        self._extra_with.append(member)
        self.all_measures.append(measures_member)
        self.show_measures[field] = measures_member

    def process_nps_passives_perc(self, applied_config, field, measures_member, fmt_string):
        if not self.__get_answers_for_nps(applied_config, field, measures_member, fmt_string):
            return

        fmt_string = ", FORMAT_STRING='0.%s%%'" % ('0' * self.__decimals)

        all_nps_member = "MEMBER [Measures].[All NPS] AS '%s'" % '+'.join(self.__answers_mdx)
        self._extra_with.append(all_nps_member)

        member = "MEMBER "+measures_member+" AS '(%s+%s)/[Measures].[All NPS]' %s" %\
                    (self.__answers_mdx[7], self.__answers_mdx[8], fmt_string)
        self._extra_with.append(member)
        self.all_measures.append(measures_member)
        self.show_measures[field] = measures_member

    def process_nps_promotors_perc(self, applied_config, field, measures_member, fmt_string):
        if not self.__get_answers_for_nps(applied_config, field, measures_member, fmt_string):
            return

        fmt_string = ", FORMAT_STRING='0.%s%%'" % ('0' * self.__decimals)

        all_nps_member = "MEMBER [Measures].[All NPS] AS '%s'" % '+'.join(self.__answers_mdx)
        self._extra_with.append(all_nps_member)

        member = "MEMBER "+measures_member+" AS '(%s+%s)/[Measures].[All NPS]' %s" %\
                (self.__answers_mdx[9], self.__answers_mdx[10], fmt_string)
        self._extra_with.append(member)
        self.all_measures.append(measures_member)
        self.show_measures[field] = measures_member

    def process_nps_detractors(self, applied_config, field, measures_member, fmt_string):
        if not self.__get_answers_for_nps(applied_config, field, measures_member, fmt_string):
            return

        member = "MEMBER "+measures_member+" AS '%s+%s+%s+%s+%s+%s+%s'" %\
                    (self.__answers_mdx[0], self.__answers_mdx[1], self.__answers_mdx[2], self.__answers_mdx[3],\
                     self.__answers_mdx[4], self.__answers_mdx[5], self.__answers_mdx[6])
        self._extra_with.append(member)
        self.all_measures.append(measures_member)
        self.show_measures[field] = measures_member

    def process_nps_passives(self, applied_config, field, measures_member, fmt_string):
        if not self.__get_answers_for_nps(applied_config, field, measures_member, fmt_string):
            return

        member = "MEMBER "+measures_member+" AS '%s+%s'" % (self.__answers_mdx[7], self.__answers_mdx[8])
        self._extra_with.append(member)
        self.all_measures.append(measures_member)
        self.show_measures[field] = measures_member

    def process_nps_promotors(self, applied_config, field, measures_member, fmt_string):
        if not self.__get_answers_for_nps(applied_config, field, measures_member, fmt_string):
            return

        member = "MEMBER "+measures_member+" AS '%s+%s'" % (self.__answers_mdx[9], self.__answers_mdx[10])
        self._extra_with.append(member)
        self.all_measures.append(measures_member)
        self.show_measures[field] = measures_member

    def __get_answers_for_nps(self, applied_config, field, measures_member, fmt_string):
        if not applied_config.has_key('show_%s_measure' % field) or not applied_config['show_%s_measure' % field]:
            return False

        self.__decimals = 1
        if applied_config.has_key('%s_decimals' % field) and applied_config['%s_decimals' % field]:
            try:
                self.__decimals = int(applied_config['%s_decimals' % field])
            except ValueError:
                pass

        if not self.__answers_mdx:
            # make a set with members to use in the growth calculation.
            item_index = applied_config['dimensions']['Item']
            item_members = applied_config['dimension_%s' % item_index]
            assert len(item_members) == 1
            item_member = item_members[0]

            from iqportal.iquest.core.models import SurveyItemQAElement
            from sqlalchemy import and_

            # for selected item, need to get the answers in order to build the nps measure
            item_id = item_member['member_name']
            locale_id = item_member['member_unique_name'].split('[')[2][:-2]
            answers = self.sa_session.query(SurveyItemQAElement).filter(and_(SurveyItemQAElement.item_id==item_id,
                        SurveyItemQAElement.is_question==0)).order_by(SurveyItemQAElement.position).all()
            if len(answers) != 11:
                raise SecurityException('Invalid NPS item')

            self.__answers_mdx = ['[Answer].[%s].[%s].[%s]' % (item_id, locale_id, a.id) for a in answers]
        return True

    def get_mdx_config(self, config, applied_config):

        self.__answers_mdx = None
        applied_config['hierarchize'] = False

        self._extra_with = []
        # We can have an additional growth item.
        self._add_measures_with(applied_config,
                additional={'nps_kpi': (None, 'NPS %', ),
                            'nps_detractors': (None, 'NPS Detractors', ),
                            'nps_passives': (None, 'NPS Passives', ),
                            'nps_promotors': (None, 'NPS Promotors', ),
                            'nps_detractors_perc': (None, 'NPS Detractors %', ),
                            'nps_passives_perc': (None, 'NPS Passives %', ),
                            'nps_promotors_perc': (None, 'NPS Promotors %', )
                            })

        return IQuestReporter.get_mdx_config(self, config, applied_config)

class PublicationStatisticsReporter(ConfigurableWidgetReporter):
    def get_mdx_config(self, config, applied_config):
        applied_config['hierarchize'] = False

        self._extra_with = []
        self._add_measures_with(applied_config)
        return ConfigurableWidgetReporter.get_mdx_config(self, config, applied_config)

    def _add_measures_with(self, applied_config, use_measure='respondents', additional=None):

        # these are default IQuest fields that are available from all widgets (based on permissions)
        fields = ['invitees', 'respondents', 'participants', 'response_percentage']

        options = {
            'invitees': (('PubStats Invitees', ), 'Invitees', ),
            'respondents': (('PubStats Respondents', ), 'Respondents', ),
            'participants': (('PubStats Participants', ), 'Participants', ),
            'response_percentage': (('PubStats Response %', ), 'Response %', )
        }

        # process the measures (parent object).
        self.process_measures(applied_config, fields, options, use_measure)

reporters = (DefaultQAItemReporter, MultiQAItemReporter, GrowthItemReporter, QATotalsItemReporter, NPSItemReporter, PublicationStatisticsReporter)

