# Copyright (c) 2010 BaseSoft B.V.
# See LICENSE for details.

"""
This contains the controllers for survey response
"""
import time
import datetime

from decimal import Decimal

import sqlalchemy.orm
from basesoft.web.contrib.account.models import Account
from basesoft.web.contrib.contact.models import ContactList, Company
from basesoft.web.contrib.license.models import License
from basesoft.web.contrib.notification.controllers import NotificationController
from basesoft.web.core.exceptions import PublicUserException
from iqportal.iquest.response.models import SurveyParticipant, SurveyParticipantSession, SurveyQAResponse
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy import and_, func, select

from django.utils.translation import ugettext as _

from django.utils import simplejson
from basesoft.web.core.jsonrpc.dispatcher import json_object_hook

from basesoft.web.core import exceptions as basesoft_exceptions
from basesoft.web.core.util.controllers import AbstractController, transaction_decorator, StateManager,\
    LicenseConfigController

from iqportal.iquest.core import models as core_models
from iqportal.iquest.response import models as response_models
from basesoft.web.contrib.compose.models import ComposeInvitee, ComposePublicationGroup,\
    ComposePublication

from basesoft.web.contrib.compose.controllers import ComposeInviteeController
from iqportal.iquest.compose.controllers import SurveyStateController, PublicationStateController

from basesoft import logging
from iqportal.iquest.core.models import SurveyItemQAFilter, SurveyItemQA, Survey, SurveyPublication

logger = logging.getLogger("iqportal.iquest.response")

class ResponseController(ComposeInviteeController):
    """
    IQuest Response controller
    """
    def is_respondent(self, contact, publication, minimum_responses=1,
    respondent_types=[response_models.SurveyParticipant.RESPONSES_INTERNET, response_models.SurveyParticipant.RESPONSES_BACK_OFFICE]):
        """
        Checks if the contact is a respondent in the publication. Respondent when number of responses >= minimum
        """
        minimum_responses = publication.minimum_responses_for_respondent

        responses = self.sa_session.query(response_models.SurveyQAResponse).join(
            response_models.SurveyParticipant).join(ComposeInvitee).filter(
            and_(ComposeInvitee.publication==publication, ComposeInvitee.contact==contact,
                 response_models.SurveyParticipant.is_respondent.in_(respondent_types))).all()
        return len(responses) >= minimum_responses

class AbstractResponseCheckController:
    """
    Controller used for validating itemQA (filters, required, etc). This is not used by itself.
    """
    SKIP_INVALID_BLOCKS = False
    def __init__(self, sa_session, responses, skip_items=None):
        self.sa_session = sa_session
        self._responses = responses # [{'item_id': 1, 'responses': [{}]}]
        self._result = [] # result list with {item_id: id, result: {}}
        self._skip_items = []
        if skip_items is not None:
            self._skip_items = skip_items # these items should be skipped because branching tells us to do so.

    def is_valid(self, blocks=None):
        """
        Validates all the supplied responses and returns a boolean. If blocks is supplied, validation will be made
        only for the items in those blocks
        """
        # loop the responses
        for qa in self._responses:
            if qa['item_id'] in self._skip_items:
                continue
            item = self.sa_session.query(SurveyItemQA).filter(SurveyItemQA.id==qa['item_id']).one()
            if self.SKIP_INVALID_BLOCKS and blocks is not None and item.column.block.id not in blocks:
                continue
            responses = self._map_responses(qa['item_id'], qa['responses'])
            try:
                self._validate_response_type(item, qa['responses'], responses)
            except AssertionError:
                self._result.append({'item_id': qa['item_id'], 'messages': _('Incorrect response for this QA type')})
                continue

            should_validate = blocks is None or item.column.block.id in blocks
            self._validate_item(qa['item_id'], responses, should_validate)

        return len(self._result)==0 # return boolean

    def _validate_response_type(self, item, supplied, mapped):
        if item.type == SurveyItemQA.TYPE_SINGLE_CHOICE:
            for i in supplied:
                c = 0
                for x in mapped['question_%s' % i['q']]['responses']:
                    if x['v'] == True:
                        c = c+1

                assert c <= 1
                assert i['v'] in [None, True, False, u"1", "1"]

        if item.type == SurveyItemQA.TYPE_MULTIPLE_CHOICE:
            for i in supplied:
                assert i['v'] in [None, True, False, u"1", "1"]

    def _validate_item(self, item_id, responses, should_validate=True):
        """
        Validate a single QA item on it's responses.
        """
        raise Exception("This controller can't be used by itself, it must be extended.")

    def get_result(self):
        """
        Returns a list with all the results. This list is empty if all responses passed their checks.
        """
        return self._result

    def _map_responses(self, item_id, responses):
        """
        This maps the responses for easy processing on the filters.
        """
        rv = {'responses': []}

        for response in responses:
            q = 'question_%s' % response['q']
            a = 'answer_%s' % response['a']
            # global level
            # pass response as we need it as reference for filtered result
            rv['responses'].append(response)
            # question level
            if not rv.has_key(q):
                rv[q] = {'responses': []}
            # answer level
            if not rv.has_key(a):
                rv[a] = {'responses': []}
            # cell level
            if not rv[q].has_key(a):
                rv[q][a] = {'responses': []}
            rv[q]['responses'].append(response)
            rv[a]['responses'].append(response)
            rv[q][a]['responses'].append(response)

        return rv

    def __repr__(self):
        return '<AbstractResponseCheckController>'

class ResponseFilterController(AbstractResponseCheckController):
    """
    This controller is used to apply filters to given responses for a certain publication
    """
    def _validate_item(self, item_id, responses, should_validate=True):
        """
        Validate a single QA item on it's responses.
        TODO: Add correct way of validating based on priority, see ticket #866
        """
        from django.forms.util import ValidationError
        from basesoft.web.core.exceptions import PublicSecurityException
        from iqportal.iquest.core.qa_filters import filters

        result = {'failures': [], 'q': [], 'a': [], 'qa': []}
        fail = False

        # get all the filters for supplied item_id
        qa_filters = self.sa_session.query(SurveyItemQAFilter).filter(SurveyItemQAFilter.item_id==item_id).all()

        # if we don't have any filters, we shouldn't continue checking them.
        if len(qa_filters) == 0:
            return

        # loop the filters
        for qa_filter in qa_filters:
            a = None
            q = None
            # check if the filters even exists.
            object_name = qa_filter.filter.name
            if not hasattr(filters, object_name):
                raise PublicSecurityException('Using a filter "%s" that does not exist' % object_name)

            # load the attributes.
            attributes = {}
            for attr in qa_filter.attributes:
                attributes[attr.attribute] = attr.value

            f = getattr(filters, object_name)
            if not should_validate and not f.ALWAYS_CHECK:
                continue

            attributes['should_validate'] = should_validate

            # global filter
            if qa_filter.question is None and qa_filter.answer is None:
                resp = responses['responses']
                f = f(resp)
                try:
                    f.validate(**attributes)
                except ValidationError, e:
                    fail = True
                    result['failures'] += e.messages
                continue
            # answer level filter
            if qa_filter.question is None:
                a = qa_filter.answer.id
                resp = responses['answer_%s' % a]['responses']
                f = f(resp)
                try:
                    f.validate(**attributes)
                except ValidationError, e:
                    fail = True
                    result['a'].append({'id': a, 'failures': e.messages})
                continue
            # question level filter
            if qa_filter.answer is None:
                q = qa_filter.question.id
                resp = responses['question_%s' % q]['responses']
                f = f(resp)
                try:
                    f.validate(**attributes)
                except ValidationError, e:
                    fail = True
                    result['q'].append({'id': q, 'failures': e.messages})
                continue
            # cell level filter
            a = qa_filter.answer.id
            q = qa_filter.question.id
            resp = responses['question_%s' % q]['answer_%s' % a]['responses']
            f = f(resp)
            try:
                f.validate(**attributes)
            except ValidationError, e:
                fail = True
                result['qa'].append({'qid': q, 'aid': a, 'failures': e.messages})

        # if there are any failures, append them to the result.
        if fail:
            self._result.append({'item_id': item_id, 'messages': result})

    def __repr__(self):
        return '<ResponseFilterController result="%s">' % self._result


class ResponseRequiredController(AbstractResponseCheckController):
    """
    This controller is used to check that the given responses have proper values for itemQAs that are required.
    """
    SKIP_INVALID_BLOCKS = True

    def _validate_item(self, item_id, responses, should_validate=True):
        """
        Validate a single QA item on it's required setting.
        """
        from django.forms.util import ValidationError

        if not should_validate:
            return

        # check if we have the object as required
        q = self.sa_session.query(core_models.SurveyItemQA).filter(
            and_(core_models.SurveyItemQA.item_id==item_id, core_models.SurveyItemQA.required==True))

        if q.count() == 0:
            # we don't do anything if we don't have the item with the required flag set
            return

        # get the item object
        item_qa = q.one()

        result = {'failures': [], 'q': [], 'a': [], 'qa': []}
        fail = False
        filled = {} # a dictionary where we keep how many answers have been filled for each question
        if len(responses['responses']) == 0:
            self._result.append({'item_id': item_id, 'messages': _('This question is required')})
            return

        for r in responses['responses']:
            if not filled.has_key(r['q']):
                filled[r['q']] = {'answers':0, 'filled': 0}

            filled[r['q']]['answers'] = filled[r['q']]['answers'] + 1
            if r['v'] is not None:
                filled[r['q']]['filled'] = filled[r['q']]['filled'] + 1
            else:
                # we don't have an answer, but if we're dealing with an itemQA that needs all answers for a question,
                # then we can add the messages directly here
                if item_qa.type in [core_models.SurveyItemQA.TYPE_SINGLE_TEXT_INPUT]:
                    fail = True
                    e = ValidationError(_('This answer is required'))
                    result['qa'].append({'qid': r['q'], 'aid': r['a'], 'failures': e.messages})

        if fail:
            # it means we had an item that requires all answers, so we append the error messages and exit
            self._result.append({'item_id': item_id, 'messages': result})
            return

        # if it's something that requires only one answer per question, we proceed
        for key, value in filled.items():
            # for the rest, is enough to have at least a single answer for the question to be considered filled in
            if value['filled'] == 0:
                fail = True
                e = ValidationError(_('You must provide at least an answer for this question'))

                # need to check if the item has more questions, and if not, show the message at item level
                if len(filled) == 1:
                    # just one question, show the message at item level
                    result['failures'] += e.messages
                else:
                    # just show it for the question with all the answers missing
                    result['q'].append({'id': key, 'failures': e.messages})

        if fail:
            self._result.append({'item_id': item_id, 'messages': result})

    def __repr__(self):
        return '<ResponseRequiredController result="%s">' % self._result

class BranchController:
    """
    Controller that takes care of branching related response issues.
    """
    def __init__(self, sa_session, responses):
        self.sa_session = sa_session
        self._responses = responses

    def get_next_blocks(self, current_blocks=None):
        """
        Get the next block based on the supplied responses. The first branch to be found as valid
        will be returned.
        """
        blocks = []
        # the first matching branch is leading.
        for item_response in self._responses:
            branch = self.sa_session.query(core_models.SurveyItemQABranch).filter(core_models.SurveyItemQABranch.item_id==item_response['item_id']).all()
            if len(branch) == 0:
                continue
            if current_blocks is not None:
                if branch[0].item.column.block.id not in current_blocks:
                    continue
            for r in item_response['responses']:
                rv = self._branch(branch, r)
                if rv is not None and rv not in blocks:
                    blocks.append(rv)
        return blocks

    def _branch(self, branch, r, return_branch=False):
        """
        Determines if the supplied response contains a branch.
        """
        if r['v'] is None:
            return None
        for b in branch:
            if b.question is not None:
                if r['q'] == b.question.id and (b.answer is None or r['a'] == b.answer.id):
                    if return_branch:
                        return b
                    return b.branch_block.id
                continue
            if b.answer is not None:
                if r['a'] != b.answer.id:
                    continue
            if return_branch:
                return b
            return b.branch_block.id
        return None

    def filter_invalid_responses(self, publication):
        if publication.branch_cache is None:
            return [], []

        c = simplejson.loads(publication.branch_cache, object_hook=json_object_hook)

        branches = []
        for item_response in self._responses:
            branch = self.sa_session.query(core_models.SurveyItemQABranch).filter(core_models.SurveyItemQABranch.item_id==item_response['item_id']).all()
            if len(branch) == 0:
                continue
            for r in item_response['responses']:
                b = self._branch(branch, r, True)
                if b is not None:
                    branches.append(b.id)
        map = {}

        for i in publication.survey.itemQAs:
            map['i_%s' % i.id] = i

        remove_qas = []
        skip_items = []

        for b in branches:
            for i in c['branch_%s' % b]:
                item = map['i_%s' % i]
                skip_items.append(i)
                for q in item.questions:
                    for a in item.answers:
                        remove_qas.append((q.id, a.id,))

        return remove_qas, skip_items

class ParticipantController(AbstractController):
    """
    Access point from the view of a participant
    """

    # gets set to the participant object
    _participant = None

    # participant session
    _participant_session = None

    # survey which the participant tries to access
    _survey = None

    def __init__(self, sa_session, locale, invitee_hash=None, participant_session_id=None, verify_status=True):
        """
        This method should be called from a view where the participant has access to. To be able to initialise this
        controller a valid invitee_hash must arrive in this class. Then if all goes well a participant_session_id is
        generated which can be used to initialise this constructor while the session is active. (over http requests)

        In the end we create or retrieve a Participant to which we can set the session and we cna save responses to.

        @param invitee_hash: string that the gives us the invitee_id (tblComposeInvitee)
        @param participant_session: int of the participant-session
        """

        # assert arguments
        if invitee_hash: assert isinstance(invitee_hash, (unicode, str))
        if invitee_hash: assert len(invitee_hash) == 8
        if participant_session_id: assert isinstance(participant_session_id, long)

        self._verify_status = verify_status

        if invitee_hash is None and participant_session_id is None:
            raise Exception('invitee_hash and participant_id not provided')

        # call parent constructor
        AbstractController.__init__(self, sa_session, locale)

        # if provided a participant_session_id we can use that to retrieve the invitee_id
        if participant_session_id:

            invitee = self._get_invitee_from_session_id(participant_session_id)
            if verify_status and invitee.publication.status != core_models.SurveyPublication.STATUS_ONLINE:
                raise basesoft_exceptions.PublicUserException(_('The survey requested is currently not online'))
            # if the invitee_hash is given make sure that it belongs to the same session cuz else we need to make a new one
            if invitee_hash is None or invitee_hash == invitee.invitee_hash:
                self._participant_session = self._init_participant_session(participant_session_id)
                self._participant = self._participant_session.participant

        # try to retrieve the invitee_id from the hash if given
        if self._participant is None:
            try:
                invitee = self.sa_session.query(ComposeInvitee).filter(
                    ComposeInvitee.invitee_hash == invitee_hash).one()
            except sqlalchemy.orm.exc.NoResultFound:
                raise basesoft_exceptions.PublicUserException('Invalid hash supplied')

            if verify_status and invitee.publication.status != core_models.SurveyPublication.STATUS_ONLINE:
                raise basesoft_exceptions.PublicUserException(_('The survey requested is currently not online'))

            invitee_id = invitee.id

            # assert we have the invitee_id now and initialise the participant
            assert isinstance(invitee_id, long)
            self._participant = self._init_participant(invitee_id)

            # start a new session
            self._participant_session = self._init_participant_session(None)

        # initialise survey for the participant
        self._init_survey()

        # if we do not want to update the participant used locale (in update_responses for example) we skip it by supplying False
        if self.locale is not False:
            # if locale was passed
            if self.locale is None:
                self.locale = self.get_default_locale()
            assert self.locale is not None

            # update the locale to the participant
            self._participant.locale = self.locale
            self.sa_session.add(self._participant);
            self.sa_session.flush()

    def _init_participant(self, invitee_id):
        """
        Method called from the constructor to setup the controller to have a valid participant to work with.
        """

        participant = None

        # retrieve invitee from id
        invitee = self.sa_session.query(ComposeInvitee).get(invitee_id)
        if not invitee:
            raise basesoft_exceptions.SystemException('Unable to retrieve invitee for [%s]' % invitee_id)

        # if there is no contact attached to the invitee we assume the survey is anonymous
        if invitee.contact is None:
            participant = self._create_anonymous_participant(invitee)
        else:
            # there is a contact attached.. see if there is already a participant for this contact and invitee_id
            # else create one
            try:
                participant = self.sa_session.query(response_models.SurveyParticipant).filter(response_models.SurveyParticipant.invitee==invitee).one()
            except sqlalchemy.orm.exc.NoResultFound:
                participant = self._create_participant(invitee)

        # error out if we still didnt find a participant
        if not participant:
            raise basesoft_exceptions.SystemException('Unable to initialise participant for invitee_id [%s]' % invitee_id)

        logger.info('ParticipantController: initialised for participant [%s]' % participant.id)

        return participant

    def _create_participant(self, invitee):
        """
        Creates a participant for the given invitee
        """

        participant = response_models.SurveyParticipant()
        participant.invitee = invitee

        p = invitee.publication
        # when a participant is created we also have to update the number_of_participants on the survey publication (#2745)
        if p.number_of_participants is None:
            p.number_of_participants = 0
        p.number_of_participants = p.number_of_participants + 1

        self.sa_session.add_all([participant, p])
        self.sa_session.flush()

        logger.info('ParticipantController: Created participant [%s]' % participant.id)

        return participant

    def _create_anonymous_participant(self, invitee):
        """
        Creates a new participant for the given invitee which is used for anonymous surveys
        """
        # check if the survey is anonymous
        publication = invitee.publication
        if not publication.anonymous:
            raise basesoft_exceptions.SystemException('Cannot create anonymous participant because Publication [%s] is not anonymous' % publication.id)

        # all good to create a new participant
        return self._create_participant(invitee)

    def _init_survey(self):
        """
        Performs checks if the participant is allowed to access a survey and returns it
        """
        self._publication = self._participant.invitee.publication
        survey = self._publication.survey

        if self._verify_status and self._publication.status != core_models.SurveyPublication.STATUS_ONLINE:
            logger.info('ParticipantController: Cannot initialize survey [%s] because publication [%s] status is not online'
                        % (survey.id, self._publication.id))
            raise basesoft_exceptions.PublicUserException(_('The survey requested is currently not online'))

        if self._publication.deleted:
            logger.info('ParticipantController: Cannot initialize survey [%s] because publication [%s] is marked as deleted'
                        % (survey.id, self._publication.id))
            raise basesoft_exceptions.PublicSystemException(_('Invalid survey publication (deleted)'))

        # check for invitee expiry
        licensee = self._publication.groups[0].group.license.domain.default_license
        lcc = LicenseConfigController(self.sa_session, licensee)
        expire_days = lcc.get('basesoft.web.response.invitee_expiry_days')
        if expire_days != 0:
            expired = self._participant.invitee.created + datetime.timedelta(days=7)
            if expired < datetime.datetime.now():
                logger.info('ParticipantController: Cannot initialize survey [%s] because invitee hash [%s] is expired'
                            % (survey.id, self._participant.invitee.invitee_hash))
                raise basesoft_exceptions.PublicSystemException(_('The requested URL has expired.'))


        logger.info('ParticipantController: initialised survey [%s]' % survey.id)

    def _get_invitee_from_session_id(self, session_id):
        """
        Returns the invitee_id for the given session_id
        """
        try:
            invitee = self.sa_session.query(ComposeInvitee).join(
                response_models.SurveyParticipant).join(
                response_models.SurveyParticipantSession).filter(
                response_models.SurveyParticipantSession.id == session_id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise basesoft_exceptions.SystemException('Unable to retrieve invitee from session_id [%s]' % session_id)

        return invitee

    def _init_participant_session(self, session_id):
        """
        Initialises the participant session object by either continueing or creating a new session
        """

        # start a new session if no session_id is supplied or else continue
        if session_id is None:
            logger.debug('ParticipantController: creating new session for participant [%s]' % self._participant.id)
            session = response_models.SurveyParticipantSession()
            session.participant = self._participant
        else:
            logger.debug('ParticipantController: continueing session [%s]' % session_id)
            session = self.sa_session.query(response_models.SurveyParticipantSession).get(session_id)
            if not session:
                raise basesoft_exceptions.SystemException('Unable to continue session_id [%s]: No session found.' % session_id)

        # we should have a session now.. explicit update its time.. is this needed?
        session.last_update = datetime.datetime.now()

        self.sa_session.add(session)
        self.sa_session.flush()

        logger.info('ParticipantController: initialised participant session [%s]' % session.id)

        return session

    def get_participant_session_id(self):
        """
        Returns a participant_session_id which should be used for initialisation this constructor during the same
        user session. This could happen in a http session for example where the controller is initialised over multiple
        http requests.
        """
        if not self._participant_session:
            raise basesoft_exceptions.SystemException('No active participant session')

        return self._participant_session.id

    def delete_responses(self):
        """
        Deletes all responses for the participant and saves them in the responses history table.
        """
        responses = self.sa_session.query(response_models.SurveyQAResponse).filter(
                                 response_models.SurveyQAResponse.participant_id==self._participant.id).all()
        for response in responses:
            # first add them to the responses history table
            response_h = response_models.SurveyQAResponseHistory()
            response_h.participant = response.participant
            response_h.question = response.question
            response_h.answer = response.answer
            response_h.participant_session = response.participant_session
            response_h.response_text = response.response_text
            response_h.response_int = response.response_int
            response_h.response_float = response.response_float
            response_h.response_created = response.created
            response_h.response_last_updated = response.last_updated
            self.sa_session.add(response_h)

            # now delete from the responses table
            self.sa_session.delete(response)
        self._participant.is_respondent = response_models.SurveyParticipant.NO_RESPONSES

        # we are resetting the state of this participant to being non-respondent
        p = self._participant.invitee.publication
        p.number_of_respondents = p.number_of_respondents - 1

        self.sa_session.add_all([self._participant, p])

        self.sa_session.flush()

    @transaction_decorator
    def update_responses(self, responses, is_responent=response_models.SurveyParticipant.RESPONSES_INTERNET, blocks=None):
        """
        Receives a complex dictionary with question and answers responses which updates the participant's responses.
        If blocks is not None, filters will be checked only for the supplied blocks.
        """

        assert isinstance(responses, list)

        time_a = time.time()

        # filter the list from any impossible responses, due to branching.
        bc = BranchController(self.sa_session, responses)
        delete_responses_for_q_a, skip_items = bc.filter_invalid_responses(self._publication)

        logger.info('filter invalid responses took %s seconds' % (time.time() - time_a))
        #time_a = time.time()

        vr = [] # validation response list

        # check required
        rrc = ResponseRequiredController(self.sa_session, responses, skip_items)
        if not rrc.is_valid(blocks):
            vr.extend(rrc.get_result())

        # check filters
        rfc = ResponseFilterController(self.sa_session, responses, skip_items)
        if not rfc.is_valid(blocks):
            vr.extend(rfc.get_result())

        # if we have any error messages (either from items not having answers when required, or filters), return them
        if len(vr) > 0:
            return vr

        for item_qa_response in responses:
            for response in item_qa_response['responses']:
                if response['v'] is None:
                    delete_responses_for_q_a.append((response['q'], response['a'], ))
                    continue

                assert isinstance(response['q'], (int, long))
                assert isinstance(response['a'], (int, long))
                #used_q_a['q%s_a%s' % (response['q'], response['a'])] = (response['q'], response['a'], )
                self._process_response(response['q'], response['a'], response['v'])

        self._delete_responses(delete_responses_for_q_a)
        self.sa_session.flush()

        # check if the participant made a response and then set the is_respondent flag
        r = self.sa_session.query(response_models.SurveyQAResponse).filter(
                response_models.SurveyQAResponse.participant==self._participant).all()
        s = is_responent
        if len(r) == 0:
            s = response_models.SurveyParticipant.NO_RESPONSES
        if s != self._participant.is_respondent:
            p = self._participant.invitee.publication

            # the participant used to be respondent, but no longer has any responses
            if s == response_models.SurveyParticipant.NO_RESPONSES:
                p.number_of_respondents = p.number_of_respondents - 1

            # the participant used to be non-respondent, but now becomes a respondent
            elif self._participant.is_respondent == response_models.SurveyParticipant.NO_RESPONSES:
                p.number_of_respondents = p.number_of_respondents + 1

            self._participant.is_respondent = s
            self.sa_session.add_all([self._participant, p])

        self.sa_session.flush()

        return None

    def _delete_responses(self, qas):
        if len(qas) == 0:
            return
        q = """
        DELETE FROM tblSurveyQAResponse WHERE participant_id = %s AND (%s)
        """

        s = "(question_id = %s AND answer_id = %s)"

        responses = [s % qa for qa in qas]
        q = q % (self._participant.id, ' OR '.join(responses))
        self.sa_session.execute(q)

    def _process_response(self, question_id, answer_id, answer_value):
        """
        Gets called for every single response which needs to be validated and saved
        """
        if answer_value is None:
            return

        question = self.sa_session.query(core_models.SurveyItemQAElement).filter(and_(
                    core_models.SurveyItemQAElement.id==question_id,
                    core_models.SurveyItemQAElement.is_question==True)).first()
        if not question: raise basesoft_exceptions.SystemException('Question [%s] not found' % question_id)

        answer = self.sa_session.query(core_models.SurveyItemQAElement).filter(and_(
                    core_models.SurveyItemQAElement.id==answer_id,
                    core_models.SurveyItemQAElement.is_question==False)).first()
        if not answer: raise basesoft_exceptions.SystemException('Answer [%s] not found' % answer_id)

        # verify that the question and answer belong to the same QA
        if question.item != answer.item:
            raise basesoft_exceptions.SystemException('Question [%s] and Answer [%s] do not share the same ItemQA' % (question_id, answer_id))

        # as we are possibly updating the responses, query
        try:
            response = self.sa_session.query(response_models.SurveyQAResponse).filter(
                response_models.SurveyQAResponse.participant==self._participant).filter(
                response_models.SurveyQAResponse.question==question).filter(
                response_models.SurveyQAResponse.answer==answer).one()
        except NoResultFound:
            # build a basic response objec
            response = response_models.SurveyQAResponse()
            response.participant = self._participant
            response.question = question
            response.answer = answer

        response.participant_session = self._participant_session

        if isinstance(answer_value, str) or isinstance(answer_value, unicode):
            response.response_text = answer_value
        elif isinstance(answer_value, int) or isinstance(answer_value, long):
            response.response_int = answer_value
        elif isinstance(answer_value, float) or isinstance(answer_value, Decimal):
            response.response_float = answer_value
        else:
            raise basesoft_exceptions.SystemException('Invalid answer_value %s type %s' % (answer_value, type(answer_value)))

        self.sa_session.add(response)

        logger.debug('ParticipantController: saved response:  [%s, %s, %s, %s]' % (self._participant.id, question_id, answer_id, answer_value))

    def get_publication(self):
        return self._publication

    def get_participant_session(self):
        return self._participant_session

    def get_participant(self):
        return self._participant

    def load_survey_state(self):
        """
        Loads and returns the survey state for the current participant
        """
        manager = StateManager(self.locale, self.account_id, core_models.Survey.MODE_RESPONSE)
        controller = SurveyStateController(self.sa_session, manager)
        state = controller.load_by_invitee_id(self._participant.invitee.id)

        # override the show_page_nrs for the current publication settings
        # this is an attribute only available for the survey runner
        state['survey']['show_page_nrs'] = self._publication.show_page_nrs
        state['survey']['branching'] = controller.load_branching(state['survey']['dbID'])

        # get print options from publication
        state['survey']['viewer_print'] = self._publication.viewer_print & ComposePublication.VIEWER_PRINT_PRINTER == ComposePublication.VIEWER_PRINT_PRINTER
        state['survey']['viewer_pdf'] = self._publication.viewer_print & ComposePublication.VIEWER_PRINT_PDF == ComposePublication.VIEWER_PRINT_PDF

        # get the multi session setting from publication
        state['survey']['multi_session_setting'] = self._publication.multi_session_setting

        state['survey']['show_end_message'] = self._publication.show_end_message

        # exit url
        state['survey']['exit_url'] = self._publication.exit_url

        return state

    def get_responses(self):
        """
        returns responses for the current participant as following: {
            q_id: {
              a_id: <v>,
              a_id: <v>,
              ...
            },
            ...
        }
        """
        responses = self.sa_session.query(response_models.SurveyQAResponse) \
            .filter(response_models.SurveyQAResponse.participant == self._participant).all()

        rv = {}
        for r in responses:
            if not rv.has_key(r.question.id):
                rv[r.question.id] = {}
            rv[r.question.id][r.answer.id] = self._response_value(r)
        return rv

    def _response_value(self, response):
        """
        returns the value for the given response
        """
        rv = response.response_text
        if rv is None:
            rv = response.response_int
        if rv is None:
            # convert to str is fix for #4042 till we have proper decimal support
            rv = str(float(response.response_float))
        return rv


    def get_default_locale(self):
        """
        returns the locale for the current survey based on contact (if applicable) or the fallback locale
        """

        invitee = self._participant.invitee
        survey = self._publication.survey

        if invitee.contact is None or invitee.contact.person is None:
            locale = survey.fallback_locale
        else:
            locale = invitee.contact.person.preferred_locale

            if locale not in survey.available_locales:
                locale = survey.fallback_locale

        return locale




class PublicationResponseController(ParticipantController):
    """
    Methods for survey publication responses control. Contains methods used by a manager for controlling participants
    responses.
    """

    def __init__(self, sa_session, account_id):
        """
        """
        AbstractController.__init__(self, sa_session, None, account_id)

    def get_publication_invitees_and_participants(self, publication_id):
        """
        Returns a list with all invitees and participants
        """

        self.assert_permissions('view_survey_publication_response')
        filter, fo = self.get_query_filter_on_privileges(core_models.SurveyPublication,
                                                     ComposePublicationGroup,
                                                    core_models.SurveyPublication.PRIV_VIEW_RESPONSE)
        filter = and_(filter, ComposeInvitee.publication_id==publication_id)

        fop = fo.join(ComposeInvitee)
        fop = fop.join(response_models.SurveyParticipant)

        # this will also include anonymous participant (same invitee id, different participant id)
        participants = self.sa_session.query(response_models.SurveyParticipant).select_from(fop).filter(filter).all()
        participant_invitees_ids = [p.invitee.id for p in participants]

        # don't include anonymous invitees (that are not also participants)
        filter = and_(filter, ComposeInvitee.contact!=None)
        filter = and_(filter, ~ ComposeInvitee.id.in_(participant_invitees_ids))
        invitees = self.sa_session.query(ComposeInvitee).select_from(fo).filter(filter).all()
        return (participants, invitees)

    def get_publication_details(self, publication_id):
        """
        Get detail information of the publication, like survey name, invitees, participant, respondents
        """
        self.assert_permissions('view_survey_publication_response')
        from iqportal.iquest.core.sa_datamapper import get_tables as survey_tables
        from iqportal.iquest.response.sa_datamapper import get_tables as response_tables
        from basesoft.web.contrib.compose.sa_datamapper import get_tables as compose_tables

        columns = [
            survey_tables()['tblSurvey'].c.name,
            func.count(compose_tables()['tblComposeInvitee'].c.contact_id.distinct()).label("invitees"),
            func.count(response_tables()['tblSurveyParticipant'].c.id.distinct()).label("participants"),
            func.count(response_tables()['tblSurveyQAResponse'].c.participant_id.distinct()).label("respondents")
        ]

        filter, fo = self.get_query_filter_on_privileges(core_models.SurveyPublication,
                                                                ComposePublicationGroup,
                                                                core_models.SurveyPublication.PRIV_VIEW_RESPONSE)

        fo = fo.join(survey_tables()['tblSurvey'], core_models.SurveyPublication.survey_id==survey_tables()['tblSurvey'].c.id)
        fo = fo.join(compose_tables()['tblComposeInvitee'], isouter=True)
        fo = fo.join(response_tables()['tblSurveyParticipant'], isouter=True)
        fo = fo.join(response_tables()['tblSurveyQAResponse'], isouter=True)

        filter = and_(core_models.SurveyPublication.id==publication_id)

        q = self.sa_session.connection().execute(select(columns = columns, whereclause = filter, from_obj = fo,
                                                        group_by = survey_tables()['tblSurveyPublication'].c.publication_id))
        return q.fetchone()

    def update_responses(self, invitee_id, responses, blocks=None, participant_id=None, create_for_anonymous=False):
        """
        Updates responses for an invitee - in case of anonymous invitees, participant id must also be supplied
        because cannot create responses for it, only update
        """
        try:
            invitee = self.sa_session.query(ComposeInvitee).filter(
                ComposeInvitee.id == invitee_id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise basesoft_exceptions.SecurityException('Invalid invitee id supplied %s' % invitee_id)

        if invitee.contact is None and participant_id is None and not create_for_anonymous:
            raise basesoft_exceptions.SecurityException('Cannot create responses for anonymous invitee %s' % invitee_id)

        create_new = False # weather or not this is update o create responses
        participant = None
        if participant_id is not None:
            try:
                participant = self.sa_session.query(response_models.SurveyParticipant).filter(
                                    response_models.SurveyParticipant.id==participant_id).one()
                if participant.invitee.id != invitee.id:
                    raise basesoft_exceptions.SecurityException('Participant %s not matching invitee %s' % (participant_id, invitee_id))
            except sqlalchemy.orm.exc.NoResultFound:
                raise basesoft_exceptions.SecurityException('Invalid participant id %s' % participant_id)

        if participant is None:
            if invitee.contact is None and create_for_anonymous:
                create_new = True
            else:
                try:
                    participant = self.sa_session.query(response_models.SurveyParticipant).filter(response_models.SurveyParticipant.invitee==invitee).one()
                except sqlalchemy.orm.exc.NoResultFound:
                    # don't directly create the participant, need to check for perms and privs first
                    create_new = True

        if create_new:
            priv = core_models.SurveyPublication.PRIV_CREATE_RESPONSE
            self.assert_permissions('create_survey_publication_response')
        else:
            priv = core_models.SurveyPublication.PRIV_UPDATE_RESPONSE
            self.assert_permissions('update_survey_publication_response')

        self._publication = invitee.publication
        self.assert_privileges_on_object(self._publication, priv)

        if create_new:
            participant = self._create_participant(invitee)

        self._participant = participant
        logger.info('PublicationResponseController: initialised for participant [%s]' % self._participant.id)

        self._participant_session = self._init_participant_session(None)
        return ParticipantController.update_responses(self, responses, response_models.SurveyParticipant.RESPONSES_BACK_OFFICE, blocks)

    def delete_responses(self, participant_id):
        """
        deletes responses for participant with participant_id
        """
        self.assert_permissions('delete_survey_publication_response')
        try:
            self._participant = self.sa_session.query(response_models.SurveyParticipant).filter(
                   response_models.SurveyParticipant.id == participant_id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise InvalidInviteeHashException('PublicationResponseController: invalid participant id supplied')

        self._publication = self._participant.invitee.publication
        self.assert_privileges_on_object(self._publication, core_models.SurveyPublication.PRIV_DELETE_RESPONSE)

        if self._participant.is_respondent == response_models.SurveyParticipant.NO_RESPONSES:
            return None

        self._participant_session = self._init_participant_session(None)
        ParticipantController.delete_responses(self)
        return self._participant.id

    def load_survey_state(self, participant_id):
        """
        Loads and returns the survey state for the given participant
        """
        self.assert_permissions('view_survey_publication_response')
        try:
            self._participant = self.sa_session.query(response_models.SurveyParticipant).filter(
                   response_models.SurveyParticipant.id == participant_id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise InvalidInviteeHashException('PublicationResponseController: invalid participant id supplied %s' % participant_id)

        self._publication = self._participant.invitee.publication
        self.assert_privileges_on_object(self._publication, core_models.SurveyPublication.PRIV_VIEW_RESPONSE)

        self._participant_session = self._init_participant_session(None)
        return ParticipantController.load_survey_state(self)

    def load_survey_state_by_invitee(self, invitee_id):
        try:
            invitee = self.sa_session.query(ComposeInvitee).filter(ComposeInvitee.id == invitee_id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise InvalidInviteeHashException('PublicationResponseController: invalid invitee id supplied %s' % invitee_id)

        manager = StateManager(self.locale, self.account_id, core_models.Survey.MODE_RESPONSE)
        controller = SurveyStateController(self.sa_session, manager)
        state = controller.load_by_invitee_id(invitee.id)

        # override the show_page_nrs for the current publication settings
        # this is an attribute only available for the survey runner
        state['survey']['show_page_nrs'] = invitee.publication.show_page_nrs
        state['survey']['branching'] = controller.load_branching(state['survey']['dbID'])

        return state

    def get_responses(self, participant_id):
        """
        Loads and returns the given participant's responses
        """
        self.assert_permissions('view_survey_publication_response')
        try:
            self._participant = self.sa_session.query(response_models.SurveyParticipant).filter(
                   response_models.SurveyParticipant.id == participant_id).one()
        except sqlalchemy.orm.exc.NoResultFound:
            raise InvalidInviteeHashException('PublicationResponseController: invalid participant id supplied')

        self._publication = self._participant.invitee.publication
        self.assert_privileges_on_object(self._publication, core_models.SurveyPublication.PRIV_VIEW_RESPONSE)

        self._participant_session = self._init_participant_session(None)
        return ParticipantController.get_responses(self)


class SurveyCloneResponseController(AbstractController):

    def get_licenses_with_survey_permissions(self):
        self.assert_permissions('view_all_domains')
        return self.get_licenses_with_permissions('create_survey')

    def get_surveys_by_license(self, license_id):
        self.assert_permissions('view_all_domains')
        # get all groups for the license

        groups = self.get_groups_in_license_with_permissions('create_survey', license_id=license_id)

        group_obj = Survey().get_group_object()[0]

        _filter, fo = self.get_query_filter_on_privileges(Survey, group_obj, Survey.PRIV_VIEW, groups=groups)
        _filter = and_(_filter, Survey.archived == False)

        # get all non-archived surveys with publications for these groups...
        return self.sa_session.query(Survey).select_from(fo).filter(_filter).all()

    def get_distinct_respondent_companies(self, survey_id):
        self.assert_permissions('view_all_domains')
        # This method retrieves all the distinct companies which were used for responding to all non-archived publications

        # get invitees for publications of this survey, where is_respondent != NO_RESPONSES .
        invitees =self.sa_session.query(ComposeInvitee).join(SurveyParticipant).join(SurveyPublication, ComposeInvitee.publication_id == SurveyPublication.id).filter(
            and_(
                SurveyParticipant.is_respondent != SurveyParticipant.NO_RESPONSES,
                SurveyPublication.survey_id == survey_id,
                SurveyPublication.archived == False
            )
        ).all()
        return sorted(list(set([i.contact.company for i in invitees if i.contact is not None])), key=lambda k: k.name)# TODO, should we raise an exception if anon?

    def get_company_matches(self, survey_id, contact_list_id):
        if contact_list_id is None:
            return None, None
        self.assert_permissions('view_all_domains', 'clone_survey_publication_responses')
        cl = self.sa_session.query(ContactList).filter(ContactList.id == contact_list_id).first()

        respondents = self.get_distinct_respondent_companies(survey_id)
        companies = list(set([c.company for c in cl.contacts]))

        # the number of respondents has to be smaller or equal to the number of distinct companies in the contact list.
        if len(respondents) > len(companies):
            raise PublicUserException('The number of distinct respondents (%d) is larger than the number of available companies (%d) in the selected contact list' % (
                len(respondents),
                len(companies)
            ))

        available = sorted([{'value': c.id, 'label': c.name} for c in companies], key=lambda k: k['label'].strip())

        matches = []
        for respondent in respondents:
            match = None
            for company in companies:
                # we try and match on the company name.
                if company.name == respondent.name:
                    match = company
                    break
            matches.append({
                'source_company': {'id': respondent.id, 'name': respondent.name},
                'target_company': None if match is None else match.id
            })
        return available, matches

    def get_source_information(self, survey_id):
        self.assert_permissions('view_all_domains')
        source_survey = self.sa_session.query(Survey).filter(Survey.id == survey_id).first()
        companies = self.get_distinct_respondent_companies(source_survey.id)
        return {
            'companies': sorted([c.name for c in companies]),
            'publications': sorted([p.label for p in source_survey.publications]),
        }

    def get_available_contact_lists(self, license_id):
        self.assert_permissions('view_all_domains')

        groups = self.get_groups_in_license_with_permissions('view_contact_list', license_id=license_id)

        group_obj = ContactList().get_group_object()[0]

        _filter, fo = self.get_query_filter_on_privileges(ContactList, group_obj, ContactList.PRIV_VIEW, groups=groups)
        _filter = and_(_filter, ContactList.archived == False)

        return self.sa_session.query(ContactList).select_from(fo).filter(_filter).all()

    def get_summary(self, source, target, matches):
        self.assert_permissions('view_all_domains')
        source_survey = self.sa_session.query(Survey).filter(Survey.id == source['survey_id']).first()
        source_license = self.sa_session.query(License).filter(License.id == source['license_id']).first()

        companies = self.get_distinct_respondent_companies(source_survey.id)

        target_survey = self.sa_session.query(Survey).filter(Survey.id == target['survey_id']).first()
        target_license = self.sa_session.query(License).filter(License.id == target['license_id']).first()

        return {
            'source_survey': source_survey.name,
            'source_license': source_license.name,
            'publications': [p.label for p in source_survey.publications],
            'target_survey': target_survey.name,
            'target_license': target_license.name,
            'matches': [{
                'company': c.name,
                'target': self.sa_session.query(Company).filter(Company.id == matches['matches']['source_company_%d' % c.id]).first().name
            } for c in companies]
        }

    def is_same(self, source_survey, target_survey):
        if not isinstance(source_survey, Survey):
            source_survey = self.sa_session.query(Survey).filter(Survey.id == source_survey).first()

        if not isinstance(target_survey, Survey):
            target_survey = self.sa_session.query(Survey).filter(Survey.id == target_survey).first()

        account = None if self.get_account() is None else self.get_account().id

        ssc = SurveyStateController(self.sa_session, StateManager(None, account, Survey.MODE_RESPONSE))
        return ssc.is_same(source_survey, target_survey)

    def clone_responses(self, source_survey_id, target_survey_id, target_contact_list_id, company_matches):
        self.assert_permissions('view_all_domains', 'clone_survey_publication_responses')
        self._clone_responses(source_survey_id, target_survey_id, target_contact_list_id, company_matches)

    def task_clone_responses(self, source_survey_id, target_survey_id, target_contact_list_id, company_matches, account_id):
        assert self.get_account() is None  # this should only be called from the task runner
        ts, ss = self._clone_responses(source_survey_id, target_survey_id, target_contact_list_id, company_matches)

        # we send a notification to the initiator.
        nc = NotificationController(self.sa_session)
        notification = nc.set_notification('Responses have been cloned to %s' % ts.name,
            'Responses for all non-archived publications from "%s" have been cloned to "%s".' % (ss.name, ts.name))

        account = self.sa_session.query(Account).filter(Account.id==account_id).first()
        nc.create_system_notification(notification, account.default_group, for_license=True)

    @transaction_decorator
    def _clone_responses(self, source_survey_id, target_survey_id, target_contact_list_id, company_matches):
        # should not be the same surveys
        assert source_survey_id != target_survey_id

        ss = self.sa_session.query(Survey).filter(Survey.id == source_survey_id).first()
        ts = self.sa_session.query(Survey).filter(Survey.id == target_survey_id).first()

        domain = ts.groups[0].group.license.domain
        default_group = domain.default_group
        account = None
        for ga in default_group.group_accounts:
            if ga.active:
                account = ga.account
                break

        if account is None:
            raise PublicUserException('Cannot find an active account on the target domain "%s" default group.' % domain.name)

        # verify access to the target survey by the default group of the target domain
        # the default group is used for creation of the new publications.
        if not self.has_privileges_on_object(ts, Survey.PRIV_VIEW, groups=[default_group]):
            raise PublicUserException('The default group of the domain "%(domain)s" does not have view privileges on the target survey' % {
                'domain': domain.name
            })

        cl = self.sa_session.query(ContactList).filter(ContactList.id == target_contact_list_id).first()
        if not self.has_privileges_on_object(cl, ContactList.PRIV_VIEW, groups=[default_group]):
            raise PublicUserException('The default group of the domain "%(domain)s" does not have view privileges on the target contact list' % {
                'domain': domain.name
            })

        self.logger.info('Clone of survey publication responses initiated')

        is_same, message = self.is_same(ss, ts)
        if not is_same:
            raise PublicUserException('The states of the source and target survey are not the same. %s' % message)

        tqas = iter(ts.itemQAs)
        qa_map = {}
        for qa in ss.itemQAs:
            tqa = next(tqas)
            tqs = iter(tqa.questions)
            am = {}
            tas = iter(tqa.answers)
            for a in qa.answers:
                ta = next(tas)
                am[a.id] = ta.id
            for q in qa.questions:
                tq = next(tqs)
                qa_map[q.id] = {
                    'item_id': tqa.id,
                    'target': tq.id,
                    'answers': am
                }

        self.logger.info(' - Structure and setup are the same, continuing with cloning...')

        # loop all publications
        for publication in ss.publications:  # this only includes non-deleted, non-archived publications
            self._clone_publication(publication, ts, account, cl, company_matches, qa_map)
        return ts, ss

    def _clone_publication(self, publication, target_survey, account, contact_list, company_matches, qa_map):
        self.logger.info(' - Cloning publication "%s" (%d) responses' % (publication.label, publication.id))

        state = {
            'survey_id': target_survey.id,
            "dbID": None,
            "domID": u'publish_1',
            "begin_date": publication.begin_date,
            "end_date": None,  # we set the end date to None as we have to supply responses.
            "status": SurveyPublication.STATUS_ONLINE,
            "label": publication.label,
            "show_page_nrs": publication.show_page_nrs,
            "viewer_print": publication.viewer_print == 1,
            "show_compose_label": publication.show_compose_label,
            "show_publication_label": publication.show_publication_label,
            "exit_url": publication.exit_url,
            "anonymous": False,  # can never be anonymous
            "contact_lists": [contact_list.id],
            "mailings": []  # no mailings necessary
        }

        # instantiate the controller with the account from the target domain.
        psc = PublicationStateController(self.sa_session, StateManager(None, account.id, Survey.MODE_RESPONSE))
        np = psc.save(state)
        self.logger.info(' - New publication "%s" created, status: %s' % (np, np.status))

        # create the invitees
        cic = ComposeInviteeController(self.sa_session)
        cic.save_or_update_invitees(np)

        self.logger.info(' - Invitees added for publication')

        # loop all respondents
        respondents = self.sa_session.query(SurveyParticipantSession).join(SurveyParticipant).join(ComposeInvitee).filter(
            and_(
                ComposeInvitee.publication == publication,
                SurveyParticipant.is_respondent > 0
            )
        ).all()
        self.sa_session.expire(np)

        self.logger.info(' - Number of respondents: %d' % len(respondents))

        for participant_session in respondents:
            participant = participant_session.participant
            company = participant.invitee.contact.company

            self.logger.info(' - Cloning responses for %s' % company)

            target_company_id = company_matches['source_company_%d' % company.id]
            invitee = None
            for inv in np.invitees:
                if inv.contact.company.id == target_company_id:
                    invitee = inv
                    break
            if invitee is None:
                raise PublicUserException('Could not find matching company for respondent %s' % company.name)

            self.logger.info(' - Invitee %s with publication %s' % (invitee, invitee.publication))

            pc = ParticipantController(self.sa_session, locale=participant.locale, participant_session_id=participant_session.id, verify_status=False)
            responses = pc.get_responses()
            new_responses = self.__convert_response_state(responses, qa_map)
            npc = ParticipantController(self.sa_session, locale=participant.locale, invitee_hash=invitee.invitee_hash)

            npc.update_responses(new_responses)
            new_participant = npc.get_participant()

            # we fix the last_updated columns for data aggregation & reporting purposes
            self.logger.info(' - Fixing last_updated columns')

            # this column is used for participant_time_id in the aggregate table
            new_participant.last_updated = participant.last_updated

            last_updated = self.sa_session.query(SurveyQAResponse).filter(SurveyQAResponse.participant==participant).order_by(
                SurveyQAResponse.last_updated.desc()).first().last_updated
            self.sa_session.execute("""
                UPDATE tblSurveyQAResponse SET last_updated = '%s' WHERE participant_id = %d
            """ % (last_updated.strftime('%Y-%m-%d %H:%M:%S'), new_participant.id))

        np.status = SurveyPublication.STATUS_FINISHED
        self.sa_session.add(np)

    def __convert_response_state(self, responses, qa_map):
        self.logger.info(' - Converting response %s' % responses)
        items = {}
        for qid, answers in responses.iteritems():
            _map = qa_map[qid]
            if not items.has_key(_map['item_id']):
                items[_map['item_id']] = {'item_id': _map['item_id'], 'responses': []}

            for aid, value in answers.iteritems():
                items[_map['item_id']]['responses'].append({
                    'q': _map['target'],
                    'a': _map['answers'][aid],
                    'v': value
                })
        return items.values()

# Exceptions
class InvalidInviteeHashException(basesoft_exceptions.SecurityException): pass

