# Copyright (c) 2015 BaseSoft B.V.
# See LICENSE for details.

import datetime

from sqlalchemy.orm import mapper, relation, backref
from sqlalchemy.schema import UniqueConstraint
from sqlalchemy import Integer, Column, UnicodeText, DateTime, ForeignKey, SmallInteger

from basesoft.web.contrib.license.hooks import PrivilegeObjectMapper
from basesoft.web.contrib.license import models as license_models
from basesoft.sqlalchemy.table import IQTable

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    return tables

def do_tables(metadata):
    """
    This function is called by Django middleware when needed.
    """
    global __tables_defined
    if __tables_defined:
        return

    closureWEEEProductLevel = IQTable('closureWEEEProductLevel', metadata,
        Column('parent_level_id', Integer, nullable=False, primary_key=True, autoincrement=False),
        Column('product_level_id', Integer, nullable=False, primary_key=True, autoincrement=False),
        Column('distance', Integer, nullable=False),
    )

    tblWEEEProductLevel = IQTable('tblWEEEProductLevel', metadata,
        Column('id', Integer, primary_key=True, autoincrement=True),
        Column('parent_level_id', ForeignKey('tblWEEEProductLevel.id'), nullable=True),

        Column('position', Integer, nullable=False),
        Column('label', UnicodeText, nullable=False),
        Column('description', UnicodeText, nullable=False),

        Column('deleted', SmallInteger, nullable=False, default=0),
        Column('archived', SmallInteger, nullable=False, default=0),

        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now),
        UniqueConstraint('parent_level_id', 'position')
    )

    tblWEEEProductLevelGroup = IQTable('tblWEEEProductLevelGroup', metadata,
        Column('product_level_id', Integer, ForeignKey('tblWEEEProductLevel.id', ondelete="CASCADE"), primary_key=True),
        Column('group_id', Integer, ForeignKey('tblGroup.id'), primary_key=True),
        Column('privilege', Integer, nullable=False, primary_key=True, autoincrement=False),
        Column('group_has_privilege', SmallInteger, default=0),
        Column('license_has_privilege', SmallInteger, default=0),
        Column('domain_has_privilege', SmallInteger, default=0),
        Column('world_has_privilege', SmallInteger, default=0),
        Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
        Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now, onupdate=datetime.datetime.now)
    )

    tblWEEEProductLevelLocale = IQTable('tblWEEEProductLevelLocale', metadata,
        Column('product_level_id', Integer, ForeignKey('tblWEEEProductLevel.id', ondelete="CASCADE"), primary_key=True),
        Column('locale_id', Integer, ForeignKey('tblLocale.id', ondelete="CASCADE"), primary_key=True),

        Column('title', UnicodeText, nullable=False),
        Column('description', UnicodeText, nullable=False),
    )

    tables['closureWEEEProductLevel'] = closureWEEEProductLevel

    tables['tblWEEEProductLevel'] = tblWEEEProductLevel
    tables['tblWEEEProductLevelGroup'] = tblWEEEProductLevelGroup
    #tables['tblWEEEProductLevelLocale'] = tblWEEEProductLevelLocale

    __tables_defined = True

def do_mappings():
    """
    This function is called by Django middleware when needed.
    """
    global __mapping_defined
    if __mapping_defined:
        return

    from basesoft.web.contrib.i18n import models as i18n_models
    from iqportal.weee.core import models as core_models


    # NOTE: the mapper sublevels are NOT privileged checked, use with caution.
    mapper(core_models.WEEEProductLevel, tables['tblWEEEProductLevel'],
        extension = PrivilegeObjectMapper(),
        properties = {
            'groups': relation(core_models.WEEEProductLevelGroup, backref='product_level', cascade="all", passive_deletes=True),
            'sublevels': relation(
                core_models.WEEEProductLevel, backref=backref(
                    'parent_level', remote_side=[tables['tblWEEEProductLevel'].c.id]
                ), order_by=tables['tblWEEEProductLevel'].c.position
            ),
        }
    )

    mapper(core_models.WEEEProductLevelGroup, tables['tblWEEEProductLevelGroup'],
        properties = {
            'group': relation(license_models.Group)
        }
    )

    #mapper(core_models.WEEEProductLevelLocale, tables['tblWEEEProductLevelLocale'],
    #    properties = {
    #        'product_level': relation(core_models.WEEEProductLevel),
    #        'locale': relation(i18n_models.Locale)
    #    }
    #)

    __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'WEEE Core')

def app_permissions():
    from django.utils.translation import ugettext as _
    return [
        license_models.Permission(u'view_weee_product_levels', u"""
            This gives the right to view the WEEE product levels.
            """, _("View WEEE product levels")),
        license_models.Permission(u'manage_weee_product_levels', u"""
            This gives the right to manage the WEEE product levels.
            """, _("Manage WEEE product levels")),
    ]