'''
Created on Sep 2, 2011

@author: martijn
'''

import os

import xml.dom.minidom

from datetime import datetime

from basesoft.web.core.util.controllers import AbstractController, transaction_decorator
from iqportal.xbrl.core.models import XBRLProject, XBRLProjectGroup, XBRLContextEntityIdentifier,\
    XBRLContextEntity, XBRLPeriod, XBRLContext, XBRLElement, XBRLUnit, XBRLFact,\
    XBRLContextSegment

from basesoft.web.contrib.file.models import File
from sqlalchemy.orm.exc import NoResultFound
from sqlalchemy.sql.expression import and_
from basesoft.web.core.exceptions import PublicUserException, SecurityException
from django.utils.translation import ugettext as _

from basesoft import logging
logger = logging.getLogger('iqportal.xbrl.core')

UNIT_TO_DATA_FIELD_MAP = {
    'iso4217:EUR': 'float_data',
    'iso4217:USD': 'float_data',
    'shares': 'float_data',
    'pure': 'float_data',
}

class XBRLController(AbstractController):

    def get_project(self, id=None, priv=XBRLProject.PRIV_VIEW):
        self.assert_permissions('view_xbrl_project')
        return self.get_privileged_object_by_id(XBRLProject, id, priv)

    def get_project_by_name(self, name):
        self.assert_permissions('view_xbrl_project')
        filter, fo = self.get_query_filter_on_privileges(XBRLProject, XBRLProjectGroup, XBRLProject.PRIV_VIEW)
        filter = and_(filter, XBRLProject.name==name)
        return self.sa_session.query(XBRLProject).select_from(fo).filter(filter).all()

    def get_identifier(self, project_id, id=None):
        project = self.get_project(project_id)
        if len(project) == 0:
            raise PublicUserException(_('No available projects'))
        project = project[0]
        filter = XBRLContextEntityIdentifier.project==project
        if id is not None:
            if not isinstance(id, list):
                id = [id]
            filter = and_(filter, XBRLContextEntityIdentifier.id.in_(id))
        return self.sa_session.query(XBRLContextEntityIdentifier).filter(filter).all()

    def get_period(self, project_id):
        project = self.get_project(project_id)
        if len(project) == 0:
            raise PublicUserException(_('No available projects'))
        project = project[0]
        # TODO add id filter
        return self.sa_session.query(XBRLPeriod).filter(XBRLPeriod.project==project).all()

    @transaction_decorator
    def update_element_name(self, id, name):
        self.assert_permissions('update_xbrl_project')
        element = self.sa_session.query(XBRLElement).filter(XBRLElement.id==id).all()
        if len(element) == 0:
            raise SecurityException('Invalid element id %s' % id)
        element = element[0]
        # check privs on project
        project = self.get_project(element.project.id, priv=XBRLProject.PRIV_UPDATE)
        if len(project) == 0:
            raise SecurityException('No project update rights - element id %s' % id)
        element.name = name
        self.sa_session.add(element)

    def create_project(self, name, xsd_content):
        """
        """
        self.assert_permissions('create_xbrl_project')
        project = XBRLProject()
        project.name = name
        project.namespace = 'bier'
        self.save_default_privilege_object_group(project)
        self.sa_session.add(project)

        dom = xml.dom.minidom.parseString(xsd_content)
        xml_elements = dom.getElementsByTagNameNS('*', 'element')
        for xml_element in xml_elements:
            # FIXME: there's elements in element whihc have no name attr, Check that.
            if xml_element.getAttribute('name') in [None, ""]:
                continue
            element = XBRLElement()
            element.project = project
            element.handle = xml_element.getAttribute('name')
            element.name = xml_element.getAttribute('name')
            self.sa_session.add(element)

        self.sa_session.flush()

        return project

    def import_xbrl(self, project, files):

        self.assert_permissions('update_xbrl_project')
        if not isinstance(project, XBRLProject):
            p = self.get_project(project, priv=XBRLProject.PRIV_UPDATE)
            if len(p) == 0:
                raise SecurityException('Invalid project %s' % project)
            project = p[0]

        for file in files:
            assert os.path.exists(file), 'File not found: %s' % file

        for file in files:

            logger.info('working on file %s' % file)

            dom = xml.dom.minidom.parseString(open(file).read())

            xml_contexts = dom.getElementsByTagNameNS('*', 'context')
            context_map = {}

            for xml_context in xml_contexts:

                context_id = xml_context.getAttribute('id')

                context = self._import_context(project, xml_context)
                context_map[context_id] = context

            xml_units = dom.getElementsByTagNameNS('*', 'unit')
            unit_map = {}
            for xml_unit in xml_units:
                unit = self._import_unit(project, xml_unit)
                unit_map[xml_unit.getAttribute('id')] = unit


            for element in project.elements:
                xml_facts = dom.getElementsByTagNameNS('*', element.handle)

                for xml_fact in xml_facts:
                    self._import_fact(project, element, xml_fact, context_map, unit_map)



        self.sa_session.flush()

    def _import_context(self, project, xml_context):

        xml_entity = xml_context.getElementsByTagNameNS('*', 'entity')[0]
        xml_period = xml_context.getElementsByTagNameNS('*', 'period')[0]

        xml_entity_identifier = xml_entity.getElementsByTagNameNS('*', 'identifier')[0]
        identifier_content = xml_entity_identifier.childNodes[0].data

        # segments
        segments = []
        xml_segments = xml_entity.getElementsByTagNameNS('*', 'segment')
        if len(xml_segments) == 0:
            pass
        elif len(xml_segments) == 1:
            xml_segment = xml_segments[0]
            xml_explicit_members = xml_segment.getElementsByTagNameNS('*', 'explicitMember')
            for xml_explicit_member in xml_explicit_members:

                segment_name = xml_explicit_member.childNodes[0].data
                segment_dimension = xml_explicit_member.getAttribute('dimension')

                try:
                    segment = self.sa_session.query(XBRLContextSegment) \
                        .filter(
                            and_(XBRLContextSegment.dimension==segment_dimension,
                                XBRLContextSegment.project==project,
                                XBRLContextSegment.name==segment_name,
                            )
                        ).one()
                except NoResultFound:
                    segment = XBRLContextSegment()
                    segment.dimension = segment_dimension
                    segment.name = segment_name
                    segment.project = project
                    self.sa_session.add(segment)
                segments.append(segment)
        else:
            raise Exception('unexpected segment length: %s' % len(xml_segments))


        try:
            identifier = self.sa_session.query(XBRLContextEntityIdentifier) \
                .filter(
                    and_(XBRLContextEntityIdentifier.identifier==identifier_content,
                         XBRLContextEntityIdentifier.project==project) \
                ).one()

        except NoResultFound:
            identifier = XBRLContextEntityIdentifier()
            identifier.identifier = identifier_content
            identifier.project = project
            self.sa_session.add(identifier)

        entities = self.sa_session.query(XBRLContextEntity) \
            .filter(and_(XBRLContextEntity.identifier==identifier, XBRLContextEntity.project==project)).all()

        entity_found = False
        entity = None
        for entity in entities:

            if len(entity.segments) != len(segments):
                continue

            segment_matches = True
            for segment in entity.segments:
                if segment not in segments:
                    segment_matches = False
                    break

            if segment_matches:
                entity_found = True
                break

        if not entity_found:
            entity = XBRLContextEntity()
            entity.identifier = identifier
            entity.project = project
            entity.segments = segments

            self.sa_session.add(entity)

        xml_period_instant = xml_period.getElementsByTagNameNS('*', 'instant')

        year = quarter = month = week = day = None

        try:
            xml_period_instant = xml_period_instant[0]
        except IndexError:

            xml_start_date = xml_period.getElementsByTagNameNS('*', 'startDate')
            xml_end_date = xml_period.getElementsByTagNameNS('*', 'endDate')
            try:
                xml_start_date = xml_start_date[0]
                xml_end_date = xml_end_date[0]
            except:
                raise Exception('forever?')
            else:
                start_date = datetime.strptime(xml_start_date.childNodes[0].data, '%Y-%m-%d').date()
                end_date = datetime.strptime(xml_end_date.childNodes[0].data, '%Y-%m-%d').date()

            # we have to determine the Year,Quarter,etc

            if start_date.year == end_date.year:
                year = start_date.year

                sq = (start_date.month-1)/3+1
                eq = (end_date.month-1)/3+1
                if sq == eq:
                    quarter = sq

                if start_date.month == end_date.month:
                    month = start_date.month

                    # TODO: week & proper weekday
                    if start_date == end_date:
                        day = start_date.isoweekday()

        else:
            start_date = end_date = datetime.strptime(xml_period_instant.childNodes[0].data, '%Y-%m-%d').date()
            year = start_date.year
            quarter = (start_date.month-1)/3+1
            month = start_date.month
            week = None # TODO: isocalendar, but that can be week 1 for 2003-12-29 for example... :S
            day = start_date.isoweekday()


        try:
            period = self.sa_session.query(XBRLPeriod).filter(and_(
                XBRLPeriod.project == project,
                XBRLPeriod.start_date == start_date,
                XBRLPeriod.end_date == end_date,
            )).one()
        except NoResultFound:
            period = XBRLPeriod()
            period.project = project
            period.start_date = start_date
            period.end_date = end_date
            period.the_year = year
            period.quarter = quarter
            period.month_of_year = month
            period.week_of_year = week
            period.day_of_month = day
            self.sa_session.add(period)

        try:
            context = self.sa_session.query(XBRLContext).filter(and_(
                XBRLContext.project == project,
                XBRLContext.period == period,
                XBRLContext.entity == entity,
            )).one()
        except:
            context = XBRLContext()
            context.project = project
            context.period = period
            context.entity = entity
            self.sa_session.add(context)


        return context



    def _import_unit(self, project, xml_unit):

        xml_measure = xml_unit.getElementsByTagNameNS('*', 'measure')[0].childNodes[0].data

        try:
            unit = self.sa_session.query(XBRLUnit).filter(and_(
                XBRLUnit.project == project,
                XBRLUnit.measure == xml_measure,
            )).one()
        except NoResultFound:
            unit = XBRLUnit()
            unit.project = project
            unit.measure = xml_measure
            self.sa_session.add(unit)

        return unit

    def _import_fact(self, project, element, xml_fact, context_map, unit_map):
        fact = XBRLFact()
        fact.element = element

        fact.context = context_map[xml_fact.getAttribute('contextRef')]

        unit_ref = xml_fact.getAttribute('unitRef')
        if unit_ref == '':
            # TODO: no unit_ref found probably means the fact contains a string or something
            return

        fact.unit = unit_map[unit_ref]

        try:
            field = UNIT_TO_DATA_FIELD_MAP[unit_map[unit_ref].measure]
        except KeyError:
            field = 'int_data'

        setattr(fact, field, xml_fact.childNodes[0].data)
        self.sa_session.add(fact)





