# Copyright (c) 2011 BaseSoft B.V.
# See LICENSE for details.
from decimal import Decimal

import datetime

from sqlalchemy.orm import mapper, relation
from sqlalchemy import DECIMAL, Column, BigInteger, Integer, ForeignKey, DateTime, UnicodeText, Unicode, SmallInteger, Date
from django.utils.translation import ugettext as _

from basesoft.sqlalchemy.table import IQTable
from basesoft.web.contrib.license.hooks import PrivilegeObjectMapper
from basesoft.web.contrib.license import models as license_models

__mapping_defined = False
__tables_defined = False
tables = {}

def reset():
    global __mapping_defined
    global __tables_defined

    __mapping_defined = False
    __tables_defined = False

def get_tables():
    """
    """
    return tables

def do_tables(metadata):
    global __tables_defined
    if not __tables_defined:

        tblXBRLProject = IQTable('tblXBRLProject', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('name', Unicode(255), nullable=False),
            Column('namespace', Unicode(255), nullable=False),

            Column('deleted', SmallInteger, nullable=False, default=0),
            Column('archived', SmallInteger, nullable=False, default=0),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tblXBRLProjectGroup = IQTable('tblXBRLProjectGroup', metadata,
            Column('project_id', Integer, ForeignKey('tblXBRLProject.id', ondelete="CASCADE"), primary_key=True),
            Column('group_id', Integer, ForeignKey('tblGroup.id'), primary_key=True),
            Column('privilege', Integer, nullable=False, primary_key=True, autoincrement=False),
            Column('group_has_privilege', SmallInteger, default=0),
            Column('license_has_privilege', SmallInteger, default=0),
            Column('domain_has_privilege', SmallInteger, default=0),
            Column('world_has_privilege', SmallInteger, default=0),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tblXBRLElement = IQTable('tblXBRLElement', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('project_id', Integer, ForeignKey('tblXBRLProject.id', ondelete="CASCADE"), nullable=False),
            Column('handle', Unicode(255), nullable=False),
            Column('name', Unicode(255), nullable=False),
            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tblXBRLContextEntityIdentifier =  IQTable('tblXBRLContextEntityIdentifier', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('project_id', Integer, ForeignKey('tblXBRLProject.id', ondelete="CASCADE"), nullable=False),

            Column('identifier', Unicode(255), nullable=False),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )


        tblXBRLContextEntity = IQTable('tblXBRLContextEntity', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('project_id', Integer, ForeignKey('tblXBRLProject.id', ondelete="CASCADE"), nullable=False),

            Column('identifier_id', Integer, ForeignKey('tblXBRLContextEntityIdentifier.id', ondelete="RESTRICT"), nullable=False),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tblXBRLContextSegment = IQTable('tblXBRLContextSegment', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('project_id', Integer, ForeignKey('tblXBRLProject.id', ondelete="CASCADE"), nullable=False),

            Column('dimension', Unicode(255), nullable=False),
            Column('name', Unicode(255), nullable=False),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tblXBRLContextEntitySegment = IQTable('tblXBRLContextEntitySegment', metadata,
            Column('entity_id', Integer, ForeignKey('tblXBRLContextEntity.id'), primary_key=True),
            Column('segment_id', Integer, ForeignKey('tblXBRLContextSegment.id'), primary_key=True),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tblXBRLPeriod = IQTable('tblXBRLPeriod', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('project_id', Integer, ForeignKey('tblXBRLProject.id', ondelete="CASCADE"), nullable=False),

            Column('start_date', Date(), nullable=False),
            Column('end_date', Date(), nullable=False),

            Column('the_year', SmallInteger, nullable=True),
            Column('quarter', SmallInteger, nullable=True),
            Column('month_of_year', SmallInteger, nullable=True),
            Column('week_of_year', SmallInteger, nullable=True),
            Column('day_of_month', SmallInteger, nullable=True),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tblXBRLContext = IQTable('tblXBRLContext', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('project_id', Integer, ForeignKey('tblXBRLProject.id', ondelete="CASCADE"), nullable=False),

            Column('entity_id', Integer, ForeignKey('tblXBRLContextEntity.id', ondelete="CASCADE"), nullable=False),
            Column('period_id', Integer, ForeignKey('tblXBRLPeriod.id', ondelete="CASCADE"), nullable=False),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tblXBRLUnit = IQTable('tblXBRLUnit', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),
            Column('project_id', Integer, ForeignKey('tblXBRLProject.id', ondelete="CASCADE"), nullable=False),

            Column('measure', Unicode(255), nullable=False),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tblXBRLFact = IQTable('tblXBRLFact', metadata,
            Column('id', Integer, primary_key=True, autoincrement=True),

            Column('element_id', Integer, ForeignKey('tblXBRLElement.id', ondelete="CASCADE"), nullable=False),
            Column('context_id', Integer, ForeignKey('tblXBRLContext.id', ondelete="CASCADE"), nullable=False),
            Column('unit_id', Integer, ForeignKey('tblXBRLUnit.id', ondelete="CASCADE"), nullable=True),

            Column('int_data', BigInteger, nullable=True),
            Column('float_data', DECIMAL(25,10), nullable=True),
            Column('string_data', UnicodeText, nullable=True),

            Column('created', DateTime(), nullable=False, default=datetime.datetime.now),
            Column('last_updated', DateTime(), nullable=False, default=datetime.datetime.now)
        )

        tables['tblXBRLProject'] = tblXBRLProject
        tables['tblXBRLProjectGroup'] = tblXBRLProjectGroup
        tables['tblXBRLElement'] = tblXBRLElement
        tables['tblXBRLContextEntity'] = tblXBRLContextEntity
        tables['tblXBRLContextEntityIdentifier'] = tblXBRLContextEntityIdentifier
        tables['tblXBRLContextSegment'] = tblXBRLContextSegment
        tables['tblXBRLContextEntitySegment'] = tblXBRLContextEntitySegment
        tables['tblXBRLPeriod'] = tblXBRLPeriod
        tables['tblXBRLContext'] = tblXBRLContext
        tables['tblXBRLUnit'] = tblXBRLUnit
        tables['tblXBRLFact'] = tblXBRLFact

        __tables_defined = True


def do_mappings():
    """
    This function is only called by the django middleware
    when it is needed
    """
    global __mapping_defined
    if not __mapping_defined:
        from iqportal.xbrl.core import models as core_models
        from basesoft.web.contrib.file import models as file_models

        mapper(core_models.XBRLProject, tables['tblXBRLProject'],
            extension = [PrivilegeObjectMapper()],
            properties = {
                'groups': relation(core_models.XBRLProjectGroup, backref='project', cascade="all", passive_deletes=True),
                'elements': relation(core_models.XBRLElement, backref='project', cascade="all", passive_deletes=True)
            }
        )

        mapper(core_models.XBRLProjectGroup, tables['tblXBRLProjectGroup'],
            properties = {
                'group': relation(license_models.Group)
            }
        )


        mapper(core_models.XBRLElement, tables['tblXBRLElement'],
            properties = {
            }
        )


        mapper(core_models.XBRLContextEntityIdentifier, tables['tblXBRLContextEntityIdentifier'],
            properties = {
                'project': relation(core_models.XBRLProject)
            }
        )

        mapper(core_models.XBRLContextEntity, tables['tblXBRLContextEntity'],
            properties = {
                'project': relation(core_models.XBRLProject),
                'identifier': relation(core_models.XBRLContextEntityIdentifier),
                'segments': relation(core_models.XBRLContextSegment, secondary=tables['tblXBRLContextEntitySegment'])
            }
        )

        mapper(core_models.XBRLContextSegment, tables['tblXBRLContextSegment'],
            properties = {
                'project': relation(core_models.XBRLProject)
            }
        )

        mapper(core_models.XBRLContextEntitySegment, tables['tblXBRLContextEntitySegment'],
            properties = {
                'entity': relation(core_models.XBRLContextEntity),
                'segment': relation(core_models.XBRLContextSegment)
            }
        )

        mapper(core_models.XBRLPeriod, tables['tblXBRLPeriod'],
            properties = {
                'project': relation(core_models.XBRLProject)
            }
        )

        mapper(core_models.XBRLContext, tables['tblXBRLContext'],
            properties = {
                'project': relation(core_models.XBRLProject),
                'entity': relation(core_models.XBRLContextEntity),
                'period': relation(core_models.XBRLPeriod),
            }
        )

        mapper(core_models.XBRLUnit, tables['tblXBRLUnit'],
            properties = {
                'project': relation(core_models.XBRLProject)
            }
        )
        mapper(core_models.XBRLFact, tables['tblXBRLFact'],
            properties = {
                'element': relation(core_models.XBRLElement),
                'context': relation(core_models.XBRLContext),
                'unit': relation(core_models.XBRLUnit),
            }
        )

        __mapping_defined = True

def fill_data(conn):
    pass

def app_permissions_category():
    """
    Returns a category for all permissions defined for this app
    """
    return license_models.PermissionCategory(u'XBRL')

def app_permissions():
    return [

        license_models.Permission(u'create_xbrl_project', u"""
            This gives the right to create a XBRL Project.
            """, _("Create XBRL Project")),
        license_models.Permission(u'view_xbrl_project', u"""
            This gives the right to view a XBRL Project.
            """, _("View XBRL Project")),
        license_models.Permission(u'update_xbrl_project', u"""
            This gives the right to update a XBRL Project.
            """, _("Update XBRL Project")),
        license_models.Permission(u'delete_xbrl_project', u"""
            This gives the right to delete a XBRL Project.
            """, _("Delete XBRL Project"))
        ]
